#include <zls.h>


int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
// Std SOD Shock problem	
	
  if (prob != NULL) {

    prob->gX0 =    0.0; 
    prob->gX1 = 9999.0; 
    
    prob->gP0  =   1.0;
    prob->gd0  =   1.0; 
    prob->gv0  =   0.0; 
    
    prob->gP1  =   0.1;
    prob->gd1  = 0.125; 
    prob->gv1  =   0.0; 
    
    prob->gP2  =   0.1;
    prob->gd2  = 0.125; 
    prob->gv2  =   0.0;
    
  }
	
  return (err);
}


int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
	
  *prob = (ProblemPtr)malloc(sizeof(Problem));
	
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    fprintf(stderr, " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
	
  return (err);
}


int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
	
  if ( prob != NULL ) {
    free(prob);
  }
	
  return (err);
}




int zlsp_InitZoneData( ZonePtr z){
  //
  // here we know the zone has BCs because we compile zls with that option.
  //
  int err = noErr;
	
  WHD       zoneW;
  ACoordPtr  xx;
  ProblemPtr prob = NULL;
  
	
  if (z != NULL){
		
    err     = zlsp_GetZoneWHD(&zoneW, z);
		
		
    if (err == noErr) {
      
      prob = gLevels[0].levelProblem;  // get the user's problem parameters
      
      Real x0  = prob->gX0;
      Real x1  = prob->gX1;
      
      Real d0  = prob->gd0;
      Real P0  = prob->gP0;
      Real v0  = prob->gv0;
      
      Real d1  = prob->gd1;
      Real P1  = prob->gP1;
      Real v1  = prob->gv1;
      
      Real d2  = prob->gd2;
      Real P2  = prob->gP2;
      Real v2  = prob->gv2;
      
      
#ifdef SYM1D
      Integer    i ;
      xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      
        for(i = 0; i< nti; i++){
          
          if (xx[i].l < x0){
            zoneW.d [i]      =  d0;
            zoneW.P [i]      =  P0;
            zoneW.vx[i]      =  v0;
            
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  0.0;
#endif
          } else if (xx[i].l < x1){
            zoneW.d [i]      =  d1;
            zoneW.P [i]      =  P1;
            zoneW.vx[i]      =  v1;
            
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  1.0;
#endif
          } else {
            zoneW.d [i]      =  d2;
            zoneW.P [i]      =  P2;
            zoneW.vx[i]      =  v2;
            
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  0.0;
#endif
          }
          
        }//i
        
      
        // set wind density after outburst to low value for BC  updates
#endif SYM1D
      
#ifdef SYM2D
       Integer    i, j ;
       xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      
    for(j = 0; j < ntj; j++){
	   for(i = 0; i< nti; i++){
        
      if (xx[i].l < x0){
        zoneW.d [j][i]      =  d0;
        zoneW.P [j][i]      =  P0;
        zoneW.vx[j][i]      =  v0;
        zoneW.vy[j][i]      =  0.0;
          
#if (NTRACERS > 0)    
        zoneW.tr[0][j][i]   =  0.0;
#endif
      } else if (xx[i].l < x1) {
        zoneW.d [j][i]      =  d1;
        zoneW.P [j][i]      =  P1;
        zoneW.vx[j][i]      =  v1;
        zoneW.vy[j][i]      =  0.0;
        
#if (NTRACERS > 0)    
        zoneW.tr[0][j][i]   =  1.0;
#endif
      }  else  {
         zoneW.d [j][i]      =  d2;
         zoneW.P [j][i]      =  P2;
         zoneW.vx[j][i]      =  v2;
         zoneW.vy[j][i]      =  0.0;
         
#if (NTRACERS > 0)    
         zoneW.tr[0][j][i]   =  0.0;
#endif
       }
       
	}//i
       
      } //j
        
        // set wind density after outburst to low value for BC  updates
#endif SYM2D
            
#ifdef SYM3D
        Integer    i, j, k ;
     xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      
      for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
	for(i = 0; i< nti; i++){
        
    if (xx[i].l < x0){
	    zoneW.d [k][j][i]      =  d0;
	    zoneW.P [k][j][i]      =  P0;
	    zoneW.vx[k][j][i]      =  v0;
	    zoneW.vy[k][j][i]      =  0.0;
	    zoneW.vz[k][j][i]      =  0.0;
      
#if (NTRACERS > 0)    
	    zoneW.tr[0][k][j][i]   =  0.0;
#endif
    } else if (xx[i].l < x1){
        zoneW.d [k][j][i]      =  d1;
        zoneW.P [k][j][i]      =  P1;
        zoneW.vx[k][j][i]      =  v1;
        zoneW.vy[k][j][i]      =  0.0;
        zoneW.vz[k][j][i]      =  0.0;
        
#if (NTRACERS > 0)    
        zoneW.tr[0][k][j][i]   =  1.0;
#endif
      } else {
        zoneW.d [k][j][i]      =  d2;
        zoneW.P [k][j][i]      =  P2;
        zoneW.vx[k][j][i]      =  v2;
        zoneW.vy[k][j][i]      =  0.0;
        zoneW.vz[k][j][i]      =  0.0;
          
#if (NTRACERS > 0)    
        zoneW.tr[0][k][j][i]   =  0.0;
#endif
          }
          
	}//i
       
      } //j
      } //k
        
        // set wind density after outburst to low value for BC  updates
#endif SYM3D
      zlsp_ReleaseZoneWHD(zoneW);
        
    } // zoneW
		
  } //z
	
	
  return(err);
}



int zlsp_InitLevelZones(Integer level){
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if (z != NULL){
      err = zlsp_InitZoneData( z );
      if (err != noErr ) {return(err);}
      err = zls_SetZoneConstantBCValues( z );
      if (err != noErr ) {return(err);}
    } else { err = zonErr; }
  } // zone
  
  return(err);
  
}

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
	
  int err = noErr;
  struct ccl_t	config;
  const char		*str1;

  Integer level;
	
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " FYRIS Hydro Problem Initialising from %s \n", confFile);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      
      // Default Values: std SOD tube
      err = zlsp_InitProblem(prob);
      
      Real g = kZLS_eos_gamma;
      if ((str1= ccl_get( &config,"SOD_G"))){
        g = atof( str1);
      }
      if ( g != kZLS_eos_gamma){
         PhysicsPtr phys = gLevels[0].levelPhysics;
         zlsp_InitGlobalEOS_Gamma( phys, g);
      }
      
      if ((str1= ccl_get( &config,"SOD_X0"))){
        prob->gX0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_X1"))){
        prob->gX1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P0"))){
        prob->gP0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D0"))){
        prob->gd0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V0"))){
        prob->gv0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P1"))){
        prob->gP1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D1"))){
        prob->gd1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V1"))){
        prob->gv1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P2"))){
        prob->gP2 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D2"))){
        prob->gd2 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V2"))){
        prob->gv2 = atof( str1);
      }
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
      for (level = 0; level < gNLevels; level++){
        err = zlsp_InitLevelZones( level ); // uses levelProblem
      }
      
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
      
      ccl_release(&config);
      
    }// config	
    
  } else {
    err = memErr;
    fprintf( stderr, " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

#ifdef ZLS_USER_BC_ROUTINE
//
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  Integer err = noErr;
	return (err);
}
#endif ZLS_USER_BC_ROUTINE


#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z){
  //
  // This routine is called just before sweeps are done, and after edge boundaries are set up.
  // The user has an opportunity to include any customised zone values, eg a stellar wind update or
  // coordinate evolution for a star.  Basicly for all boundary conditions that don't fit in the
  // grid boundary direction updates scheme.
  //  Many simulations will have nothing to do here so it is best to 
  // comment out as all the code unless a zone update is actually needed.
  // Define the global macro in zls_defs.h ZLS_USER_UPDATE_ROUTINE to enable this
  // routine.
  //
  int err = noErr;
  
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    
    prob = gLevels[0].levelProblem;  // get the user's problem parameters
    phys = gLevels[0].levelPhysics;  // get the global eos parameters
        
  } else { err = memErr; }//z
  
  
  return(err);
}
#endif ZLS_USER_UPDATE_ROUTINE


//
// customised user text IO routine for the 1D shock tube tests
// includes code from the 1D Riemann code riemann.c
//

typedef struct w{
  LRQ1      d; 
  LRQ1      P;
  LRQ1      v;
} W1D; 


int WaveLocations  ( zFile theFileRef, Real t, Real x0, W1D * wStar );
  int WaveLocations( zFile theFileRef, Real t, Real x0, W1D * wStar ){
  int err = 0;

    Real gam = (kZLS_eos_gamma);
    Real g0  = (kZLS_eos_gamma);
    Real g1  = (kZLS_eos_gamma-1.0)/(2.0*kZLS_eos_gamma);
    Real g2  = (kZLS_eos_gamma+1.0)/(2.0*kZLS_eos_gamma);
    Real g3  = (2.0*kZLS_eos_gamma)/(kZLS_eos_gamma-1.0);
    Real g4  = 2.0/(kZLS_eos_gamma-1.0);
    Real g5  = 2.0/(kZLS_eos_gamma+1.0);
    Real g6  = (kZLS_eos_gamma-1.0)/(kZLS_eos_gamma+1.0);
    Real g7  = (kZLS_eos_gamma-1.0)/2.0;
    
  Real pm = wStar->P.a;
  Real um = wStar->v.a;
  
  Real pl = wStar->P.al;
  Real ul = wStar->v.al;
  Real dl = wStar->d.al;
  Real cl = sqrt(gam*pl/dl);
  
  Real pr = wStar->P.ar;
  Real ur = wStar->v.ar;
  Real dr = wStar->d.ar;
  Real cr = sqrt(gam*pr/dr);
  
  
  Real rmr, rml;
  
  Real shl, shr, sr, sl, pml, pmr, cml, cmr, stl , str;
  
  zlsio_ZFPrintf ( theFileRef, "#--------------------------------------------------------------\n");
  zlsio_ZFPrintf ( theFileRef, "# Exact Riemann Solution\n");
  zlsio_ZFPrintf ( theFileRef, "#--------------------------------------------------------------\n");
  zlsio_ZFPrintf ( theFileRef, "# Gamma = %#11.6g\n",gam);
  zlsio_ZFPrintf ( theFileRef, "# x0    = %#11.6g\n",x0);
  zlsio_ZFPrintf ( theFileRef, "# t     = %#11.6g\n",t);
  zlsio_ZFPrintf ( theFileRef, "#--------------------------------------------------------------\n");
  
  zlsio_Printf ( "#--------------------------------------------------------------\n");
  zlsio_Printf ( "# Exact Riemann Solution\n");
  zlsio_Printf ( "#--------------------------------------------------------------\n");
  zlsio_Printf ( "# Gamma = %#11.6g\n",gam);
  zlsio_Printf ( "# x0    = %#11.6g\n",x0);
  zlsio_Printf ( "# t     = %#11.6g\n",t);
  zlsio_Printf ( "#--------------------------------------------------------------\n");
  
  if (pm<=pl) {
    shl = ul - cl;
    cml = pow(cl*(pm/pl),g1);
    stl = um - cml;
    zlsio_Printf   ( "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shl*t, x0+stl*t);
    zlsio_ZFPrintf ( theFileRef, "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shl*t, x0+stl*t);
    rml = pow(dl*(pm/pl),(1.0/g0));
  } else { 
    pml = pm/pl;
    sl  = ul - cl*sqrt(g2*pml + g1);
    zlsio_Printf   (  "#           S-Wave x = %#11.6g \n",x0+sl*t);
    zlsio_ZFPrintf ( theFileRef, "#           S-Wave x = %#11.6g \n",x0+sl*t);
    rml = dl*(pml+g6)/(pml*g6+1.0);
  }
  
  zlsio_ZFPrintf ( theFileRef, "# Contact Discont. x = %#11.6g \n",x0+um*t);
  zlsio_Printf   (  "# Contact Discont. x = %#11.6g \n",x0+um*t);
  
  if(pm>pr) {
    pmr = pm/pr;
    sr  = ur + cr*sqrt(g2*pmr + g1);
    zlsio_Printf   (  "#           S-Wave x = %#11.6g \n",x0+sr*t);
    zlsio_ZFPrintf ( theFileRef, "#           S-Wave x = %#11.6g \n",x0+sr*t);
    rmr = dr*(pmr+g6)/(pmr*g6+1.0);
  } else {
    shr = ur + cr;
    cmr = pow(cr*(pm/pr),g1);
    str = um + cmr;
    zlsio_Printf   ( "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shr*t, x0+str*t);
    zlsio_ZFPrintf ( theFileRef, "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shr*t, x0+str*t);
    rmr = pow(dr*(pm/pr),(1.0/g0));
  }
  
  
  zlsio_ZFPrintf ( theFileRef, "#--------------------------------------------------------------\n");     
  zlsio_ZFPrintf ( theFileRef, "# d = %#11.6g  %#11.6g  %#11.6g %#11.6g\n",dl, rml, rmr, dr);
  zlsio_ZFPrintf ( theFileRef, "# P = %#11.6g       %#11.6g       %#11.6g\n",pl, pm, pr);
  zlsio_ZFPrintf ( theFileRef, "# v = %#11.6g       %#11.6g       %#11.6g\n",ul, um, ur);
  zlsio_ZFPrintf ( theFileRef, "#--------------------------------------------------------------\n");
  
  zlsio_Printf ( "#--------------------------------------------------------------\n");
  zlsio_Printf ( "# d = %#11.6g %#11.6g %#11.6g %#11.6g\n",dl, rml, rmr, dr);
  zlsio_Printf ( "# P = %#11.6g       %#11.6g       %#11.6g\n",pl, pm, pr);
  zlsio_Printf ( "# v = %#11.6g       %#11.6g       %#11.6g\n",ul, um, ur);
  zlsio_Printf ( "#--------------------------------------------------------------\n");

  return(err);

  }


void RiemanSample(Real * d0, Real *p0, Real * u0, Real s,  W1D * wStar);
  void RiemanSample(Real * d0, Real *p0, Real * u0, Real s,  W1D * wStar){
//
//     Riemann solution sampler at speed = s
//     returns d0, p0, u0 
//     Use s = 0.0 for Eulerian Fluxes
//
  Real shl, shr, c, sr, sl, pml, pmr, cml, cmr, stl , str;
  
  Real gam = (kZLS_eos_gamma);
  Real g0  = (kZLS_eos_gamma);
  Real g1  = (kZLS_eos_gamma-1.0)/(2.0*kZLS_eos_gamma);
  Real g2  = (kZLS_eos_gamma+1.0)/(2.0*kZLS_eos_gamma);
  Real g3  = (2.0*kZLS_eos_gamma)/(kZLS_eos_gamma-1.0);
  Real g4  = 2.0/(kZLS_eos_gamma-1.0);
  Real g5  = 2.0/(kZLS_eos_gamma+1.0);
  Real g6  = (kZLS_eos_gamma-1.0)/(kZLS_eos_gamma+1.0);
  Real g7  = (kZLS_eos_gamma-1.0)/2.0;
  
// cache arguments:	
  
  Real pm = wStar->P.a;
  Real um = wStar->v.a; 
  
  Real dl = wStar->d.al; 
  Real pl = wStar->P.al;
  Real ul = wStar->v.al; 
  
  Real cl = sqrt(gam*pl/dl);
  
  Real dr = wStar->d.ar; 
  Real pr = wStar->P.ar;
  Real ur = wStar->v.ar; 
  
  Real cr = sqrt(gam*pr/dr);
  
  if (s<=um) {
  //
  //    to the left of the CD
  //      
    if (pm <= pl) {
    //  left rarefaction      
      shl = ul - cl;
      if (s <= shl) {
//  in left state           
        *u0 = ul;
        *p0 = pl;
        *d0 = dl;
      } else {
        cml = cl*pow((pm/pl), g1);
        stl = um - cml;
        if (s>stl) {
  //  in left star state           
          *d0 = dl*pow((pm/pl), (1.0/gam));
          *p0 = pm;
          *u0 = um;
        }else{
  //  in left fan state    
          *u0 = g5*(cl + g7*ul+s);
          c   = g5*(cl + g7*(ul-s));
          *d0 = dl*pow((c/cl), g4);
          *p0 = pl*pow((c/cl), g3);
        }
      }
    } else {
    //  left shock
      pml = pm/pl;
      sl  = ul - cl*sqrt(g2*pml + g1);
      if (s<=sl) {
//  in left state     
        *d0 = dl;
        *p0 = pl;
        *u0 = ul;  
      } else {
//  in left star state
        *d0 = dl*(pml+g6)/(pml*g6+1.0);
        *u0 = um;
        *p0 = pm;
      }
    }
  } else{
  //
  //    to the right of the CD
  //      
    if(pm>pr) {
    //   right shock
      pmr = pm/pr;
      sr  = ur + cr*sqrt(g2*pmr + g1);
      if (s>=sr){
//   in right state
        *d0 = dr;
        *u0 = ur;
        *p0 = pr; 
      } else {
//   in right star state
        *d0 = dr*(pmr + g6)/(pmr*g6 + 1.0);
        *u0 = um;
        *p0 = pm;
      }
    } else {
    //   right rarefaction
      shr = ur + cr;
      if (s>=shr) {
//   in right state
        *d0 = dr;
        *u0 = ur;
        *p0 = pr;
      } else {
        cmr =cr*pow((pm/pr),g1);
        str = um + cmr; 
        if (s<=str) {
  //  in right star state
          *d0 = dr*pow((pm/pr),(1.0/gam));
          *u0 = um;
          *p0 = pm;
        } else {
  //  in right fan state    
          *u0 = g5*(-cr + g7*ur+s);
          c  = g5*( cr - g7*(ur-s));
          *d0 = dr*pow((c/cr),g4);
          *p0 = pr*pow((c/cr),g3);
        }
      }
    }
  }
  
}

int GG88RiemannW (W1D * wStar);
  int GG88RiemannW (W1D * wStar){
  int err = 0;
  
  Integer i, is, ie;
  Integer nti;
  
  Integer iter;
  Real iter_err;
  
  
  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real vl; // velocity
  Real vr;
  
  Real z, ul, ur; // initial speed guess
  
// Riemann vars	
  Real newtonUstar;
  Real newtonStep;
  
  Real Ustar = 0.0; 
  Real Pstar = 0.0;
  
  Real Pstarl, Pstarr;
  Real dPstarl, dPstarr;
  Real astarl, astarr;
  
  Real al, ar; // Eulerian sound speed
  
  Real cl, cr; // Lagrangian sound speed
  Real sl, sr; // intermediate wave slopes
  Real zl, zr; // shock mach number
  
  Real alphal; // (gamma+1)/(4a), predivide and collect constants
  Real alphar; // predivide and collect constants
  
  dl  = wStar->d.al;
  dr  = wStar->d.ar;  
  
  Pl  = wStar->P.al;
  Pr  = wStar->P.ar;
  
  vl  = wStar->v.al;
  vr  = wStar->v.ar;
  
  Real gam = (kZLS_eos_gamma);
  Real g0  = (kZLS_eos_gamma);
  Real g1  = (kZLS_eos_gamma-1.0)/(2.0*kZLS_eos_gamma);
  Real g2  = (kZLS_eos_gamma+1.0)/(2.0*kZLS_eos_gamma);
  Real g3  = (2.0*kZLS_eos_gamma)/(kZLS_eos_gamma-1.0);
  Real g4  = 2.0/(kZLS_eos_gamma-1.0);
  Real g5  = 2.0/(kZLS_eos_gamma+1.0);
  Real g6  = (kZLS_eos_gamma-1.0)/(kZLS_eos_gamma+1.0);
  Real g7  = (kZLS_eos_gamma-1.0)/2.0;
  
  Real alpha  = ((gam+1.0)/4.0);
  Real beta   = ((gam-1.0)/2.0);
  Real delta  =  (2.0*gam)/(gam-1.0);
  
  al = sqrt(gam*Pl/dl);  // eulerian speed
  ar = sqrt(gam*Pr/dr);
  
  alphal = alpha/al; // (gamma+1)/(4a), predivide and collect constants
  alphar = alpha/ar; // predivide and collect constants
  
  cl = dl*al; // lagrangian speed
  cr = dr*ar;	
  
// initial guess
  
  ul = vl + al/beta;
  ur = vr - ar/beta;
  
  z = (ar/al)*pow((Pl/Pr), 1.0/delta);
  newtonUstar = (ul*z + ur)/(1.0+z);
  
  iter_err = 1.0;
  for (iter = 0; ((iter < 20)&&( iter_err > 1.0e-12)); iter++) {
    
    Ustar  = newtonUstar;
    
    if ( Ustar <= vl ) { // left wave is shock
      
      sl      = (newtonUstar - vl)*alphal;
      zl      = sl - sqrt(1.0+(sl*sl));
      Pstarl  = Pl + cl*(newtonUstar - vl)*zl;
      dPstarl = 2*cl*zl*zl*zl/(1.0+zl*zl);
      
    } else { // left wave is rarefaction 
      
      astarl  = al - beta*(newtonUstar - vl);
      Pstarl  = Pl * pow(astarl/al, delta);
      dPstarl = -gam*Pstarl/astarl;
      
    }
    
    if ( Ustar >= vr ) { // right wave is shock
      
      sr      = (newtonUstar - vr)*alphar;
      zr      = sr + sqrt(1.0+(sr*sr));
      Pstarr  = Pr + cr*(newtonUstar - vr)*zr;
      dPstarr = 2*cr*zr*zr*zr/(1.0+zr*zr);
      
    } else { // right wave is rarefaction
      
      astarr  = ar + beta*(newtonUstar - vr); // note +
      Pstarr = Pr * pow(astarr/ar, delta);
      dPstarr = gam*Pstarr/astarr;
      
    }
    
    
    register Real deltaP     = ( Pstarl -  Pstarr);
    register Real deltaDP    = (dPstarl - dPstarr);
    newtonStep = deltaP/deltaDP;
    
    newtonUstar  -= newtonStep; 
    iter_err      = fabs(newtonStep); // absolute error, faster as long as delta V is normalised
//
  } // iter
  
  Ustar  = newtonUstar;
  Pstar  = 0.5*(Pstarl+Pstarr);
  
  wStar->P.a   = Pstar; 
  wStar->v.a   = Ustar; 
  
  return (err);   
}


int zlsp_TextZoneOutput( zFile f, ZonePtr z);
  int zlsp_TextZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD               zoneW = z->zoneW;

    ProblemPtr prob = gLevels[0].levelProblem;  // get the user's problem parameters
    
    zlsio_ZFPrintf( f, " Zone ID = %d\n", z->zoneID);
    Real t      = zlst_GlobalTime;
    
#if (ZLS_VERBOSE > 1)
    zls_ZPrintZoneCoords( f,  z);
#endif (ZLS_VERBOSE > 1)
    
    ACoordPtr xx  = z->zoneGrid.x;
    
    Integer i;
    
    Real gam = (kZLS_eos_gamma);
    
    Real x      = 0.0;
    
    Real d         = 0.0;
    Real P         = 0.0;
    Real v         = 0.0;
    Real E         = 0.0;
    
    Real dL        = 0.0;
    Real PL        = 0.0;
    Real vL        = 0.0;
    Real dR        = 0.0;
    Real PR        = 0.0;
    Real vR        = 0.0;
    Real dC        = 0.0;
    Real PC        = 0.0;
    Real vC        = 0.0;
    
    Real d0        = 0.0;
    Real P0        = 0.0;
    Real v0        = 0.0;

    Real e0        = 0.0;
    Real sumD      = 0.0;
    Real sumP      = 0.0;
    Real sumV      = 0.0;
    Real sumE      = 0.0;
    Real sumND     = 0.0;
    Real sumNP     = 0.0;
    Real sumNV     = 0.0;
    Real sumNE     = 0.0;
    
    W1D wStar; 
    
    Real x0 = prob->gX0;
    wStar.d.al = prob->gd0;
    wStar.P.al = prob->gP0;
    wStar.v.al = prob->gv0;
    wStar.d.ar = prob->gd1;
    wStar.P.ar = prob->gP1;
    wStar.v.ar = prob->gv1;
    
    GG88RiemannW ( &wStar ); // get exact solution in wStar
    WaveLocations( f, t, x0, &wStar);
    
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    
    zlsio_ZFPrintf( f, "x.c \t d \t P \t v \t E \t d0 \t P0 \t v0 \t E0 \n");
    
    Integer nsub = 5;
    Real    isub = 0.2;
    
    for(i = is; i< ie; i++){
      
      // integral average of exact solution over cell
      
      d0 = 0.0;
      P0 = 0.0;
      v0 = 0.0;
      Real dx  = xx[i].dx*isub;
      Integer j;
      
      for ( j = 0; j < nsub; j++){
      x      = xx[i].l + dx*(j+0.5);
      Real speed = (x-x0)/t;
      RiemanSample( &dC, &PC, &vC, speed, &wStar);
      d0 += dC;
      P0 += PC;
      v0 += vC;
      }
      
      d0 *= isub;
      P0 *= isub;
      v0 *= isub;
      
        x       = xx[i].c;
        d       = zoneW.d [i];

        sumD   += zls_abs((d-d0));
        sumND  += zls_abs(d0);
        
        P       = zoneW.P [i];
        
        sumP   += zls_abs((P-P0));
        sumNP  += zls_abs(P0);
        
        v       = zoneW.vx[i];
        
        sumV   += zls_abs((v-v0));
        sumNV  += zls_abs(v0);
        
        E      = P/(d*(gam-1.0));
        e0     = P0/(d0*(gam-1.0));  
        
        sumE   += zls_abs((E-e0));
        sumNE  += zls_abs(e0);
        
        zlsio_ZFPrintf( f, "%#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g\n",
                        x, d, P, v, E, d0, P0, v0, e0 );
        
      } //i
    
    zlsio_Printf(  "%s\n",  DIVIDERLINE );
    zlsio_Printf(  " norm L1 Den    = \t %#12.6g percent\n",100.0*sumD/sumND);
    zlsio_Printf(  " norm L1 P      = \t %#12.6g percent\n",100.0*sumP/sumNP);
    zlsio_Printf(  " norm L1 v      = \t %#12.6g percent\n",100.0*sumV/sumNV);
    zlsio_Printf(  " norm L1 E      = \t %#12.6g percent\n",100.0*sumE/sumNE);
    zlsio_Printf(  "%s\n",  DIVIDERLINE );
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " norm L1 Den    = \t %#12.6g percent\n",100.0*sumD/sumND);
    zlsio_ZFPrintf( f, " norm L1 P      = \t %#12.6g percent\n",100.0*sumP/sumNP);
    zlsio_ZFPrintf( f, " norm L1 v      = \t %#12.6g percent\n",100.0*sumV/sumNV);
    zlsio_ZFPrintf( f, " norm L1 E      = \t %#12.6g percent\n",100.0*sumE/sumNE);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return(err);
}

