#ifndef ZLS_PROBLEM_H
#define ZLS_PROBLEM_H
#include <zls.h>


#ifdef SYM2D
#define FYRIS_PROBLEMNAME "2D Kelvin Helmholtz"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
#define KH2D
#endif SYM2D
#ifdef SYM3D
#define FYRIS_PROBLEMNAME "3D Kelvin Helmholtz"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
#define KH3D
#endif SYM3D
//#define ZLS_USER_BOUNDARIES
//#define ZLS_USER_UPDATE_ROUTINE

// set up a custom text output routine, defined in problen.c
#define TEXTOUTPUT(f, z) zlsp_KHZoneOutput(f, z)
int zlsp_KHZoneOutput( zFile f, ZonePtr z);


typedef struct prob {
//
// define global problem parameters here.  These will be available to all the code
// via the global level[0] structure
//  
//
  
  Real gD0;       // density of ambient
  Real gD1;       // density of slab
  
  Real gP0;       // pressure of ambient
  Real gP1;       // pressure of slab
  
  Real gv0;       // velocity of ambient
  Real gv1;       // velocity of slab
  
  Real gDeltaY;       // slab thickness
  Real gNoise;       // velocity perturbation amplitude
  
  Real gUseFreq;       // use sinewave pertubation instead of noise
  Real gFreq;       // sinewave  velocity perturbation frequency
  
} Problem, * ProblemPtr;


int   zlsp_InitProblem    ( ProblemPtr   prob);
int   zlsp_NewProblem     ( ProblemPtr * prob);
int   zlsp_DisposeProblem ( ProblemPtr   prob);

int   zlsp_InitZoneData          ( ZonePtr z );
int   zlsp_InitLevelZones        ( Integer level );

int   zlsp_SetUpProblemFromConfigFile(char * confFile);

#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z);
#endif ZLS_USER_UPDATE_ROUTINE


#endif ZLS_PROBLEM_H
