#ifndef ZLS_PROBLEM_H
#define ZLS_PROBLEM_H
#include <zls.h>


#define FYRIS_PROBLEMNAME "2D Wave Convergence Test"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
//#define ZLS_USER_BOUNDARIES
//#define ZLS_USER_UPDATE_ROUTINE

// set up a custom text output routine, defined in problem.c
#define TEXTOUTPUT(f, z) zlsp_2DWaveZoneOutput(f, z)
int zlsp_2DWaveZoneOutput( zFile f, ZonePtr z);

typedef struct prob {
//
// define global problem parameters here.  
// These will be available to all the code
// via the global level[0] structure
//
// d = 1.0 +0.2 sin( pi(x + y - t(vx +vy))) 
// P, vx, vy constant
// periodic boundaries needed
//
  Real gA;           // amplitude of wave
  Real gP ;          // pressure
  Real gvx;          // x velocity
  Real gvy;          // y velocity 
  
} Problem, * ProblemPtr;


int   zlsp_InitProblem    ( ProblemPtr   prob);
int   zlsp_NewProblem     ( ProblemPtr * prob);
int   zlsp_DisposeProblem ( ProblemPtr   prob);

int   zlsp_InitZoneData          ( ZonePtr z );
int   zlsp_InitLevelZones        ( Integer level );

int   zlsp_SetUpProblemFromConfigFile(char * confFile);

#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z);
#endif ZLS_USER_UPDATE_ROUTINE


#endif ZLS_PROBLEM_H
