#ifndef ZLS_PROBLEM_H
#define ZLS_PROBLEM_H
#include <zls.h>


#define FYRIS_PROBLEMNAME "2D Explosion"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
//#define ZLS_USER_BOUNDARIES
//#define ZLS_USER_UPDATE_ROUTINE
#define ZLS_USERIO_EXPLOSION

enum {
  kDontSmooth = 0,
  kSmooth     = 1
};

typedef struct prob {
//
// define global problem parameters here.  
// These will be available to all the code
// via the global level[0] structure

// initial overpressure region
  Real gR;          // radius
  
  Real gD0;          // density of ambient medium
  Real gP0;          // pressure of ambient medium
  Real gD1;          // density of sphere
  Real gP1;          // pressure of ambient medium
  
  
  Integer gAntiAlias; // whether to smooth the pressuren-sphere by computing cell/sphere overlaps
  Real    gSmooth; // Gaussian sigma (Cells) for smoothing, default = 1 cell.
  
} Problem, * ProblemPtr;


int   zlsp_InitProblem    ( ProblemPtr   prob);
int   zlsp_NewProblem     ( ProblemPtr * prob);
int   zlsp_DisposeProblem ( ProblemPtr   prob);

int   zlsp_InitZoneData          ( ZonePtr z );
int   zlsp_InitLevelZones        ( Integer level );

int   zlsp_SetUpProblemFromConfigFile(char * confFile);

#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z);
#endif ZLS_USER_UPDATE_ROUTINE


#endif ZLS_PROBLEM_H
