#ifndef ZLS_PROBLEM_H
#define ZLS_PROBLEM_H
#include <zls.h>


#define FYRIS_PROBLEMNAME "Wind Bubbles"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
#define BUBBLES
#define STELLAR_WIND_TIMING
//#define DRIVEN_ISM
//#define ZLS_USER_BOUNDARIES
#define ZLS_USER_UPDATE_ROUTINE

// set up a custom text output routine, defined in problen.c
#define TEXTOUTPUT(f, z) zlsp_TextZoneBubbleAnalysis(f, z)
int zlsp_TextZoneBubbleAnalysis( zFile f, ZonePtr z);

#define MAXNSTARS 8

typedef struct prob {
//
// define global problem parameters here.  These will be available to all the code
// via the global level[0] structure
//  
  Real gVISM[NDIMS];
//
// set in user's SetUpProblemFromConfigFile routine
// up to 8 stars at a time
//
  Integer gNstars;
  
  Real gRWindR0[MAXNSTARS];
  
#ifdef SYM1D
  Real gXStar[MAXNSTARS];
#endif SYM1D
#ifdef SYM2D
  Real gXStar[MAXNSTARS];
  Real gYStar[MAXNSTARS];
#endif SYM2D
#ifdef SYM3D
  Real gXStar[MAXNSTARS];
  Real gYStar[MAXNSTARS];
  Real gZStar[MAXNSTARS];
#endif SYM3D
  Real gVWind[MAXNSTARS];
  Real gDWind[MAXNSTARS];
  Real gPWind[MAXNSTARS];
  
   // Hot gas
  // Input T, N, Turb, Kepler in cgs or fractions, others derived and normalised
  
  Real gD0;       // density of hot component (normalised)
  Real gP0;       // pressure of hot component (normalised)
  
} Problem, * ProblemPtr;


int   zlsp_InitProblem    ( ProblemPtr   prob);
int   zlsp_NewProblem     ( ProblemPtr * prob);
int   zlsp_DisposeProblem ( ProblemPtr   prob);

int   zlsp_InitZoneData          ( ZonePtr z );
int   zlsp_InitLevelZones        ( Integer level );

int   zlsp_SetUpProblemFromConfigFile(char * confFile);

#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z);
#endif ZLS_USER_UPDATE_ROUTINE


#endif ZLS_PROBLEM_H
