#ifndef ZLS_ALGORITHM_H
#define ZLS_ALGORITHM_H
#include <zls.h>

//====================================================================================
//
// Sweep and Advance all zones in the simulation, single level or
// multi-level. calls the above as needed, depending on gNlevelsm and
// UNIFORM_SWEEPS or REFINED_SWEEEPS.
//
// To sweep an entire simulation call, zlsa_Sweep(kSweepMainLevel);
//

int zlsa_Sweep(Integer level);

//====================================================================================
//
// Sweep and Advance all zones in level, including any refined zones on
// the level below ***and pass results to super level***.  Recursively
// applied to sub levels. Call RefinedSweep(0,0) to process whole
// structure.  Presently calls SplitSweeps on each zone, and each zone
// has a dt set before a call to refined sweep. Iteration count used in
// lagrangian recursive sweeps, set to 0 when calling zlsa_RefinedSweep
// from elsewhere, only the code will set iteration to other than 0
// interally to zlsa_RefinedSweep. and then only if gNLevels > 1
//====================================================================================

int zlsa_RefinedSweep(Integer level, int iteration);

//====================================================================================
//
// Sweep and Advance all zones in level, including any refined zones on
// the level below advancing *all levels*, one step.  requires that all
// the timesteps be set correctly to the common smallest timestep.
//
//====================================================================================

int zlsa_UniformSweep(Integer level, Integer iteration);

//====================================================================================
//
// Sweep and Advance all zones in level, without processing any refined
// zones on the level below or updating the level above.  Less
// overhead, useful for simple tiled levels without refinement. 
// Iteration count used in lagrangian recursive sweeps, set to 0
// normally ie to sweep a single level simulation call,
// zlsa_DirectSweep(0,0);
//
//========================================

int zlsa_DirectSweep(Integer level, int iteration); 



#endif ZLS_ALGORITHM_H
