
#ifndef ZLS_H
#define ZLS_H

/*                                             */
/* standard UNIX sys libs                      */
/*                                             */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/times.h>
#include <time.h>


/*  Configuration file opensource libs      */


#include <ccl.h>

#include "zls_copyright.h"

/*                                           */
/* Multi-Level Zone Library: ZLS             */
/*                                           */
/*  Operating system settings                */
/*  and code settings                        */
/*                                           */

#include "zls_OSdefs.h"
#include "zls_defs.h"

#ifdef ZLIB_COMPRESSED_IO
#include <zlib.h>
#endif ZLIB_COMPRESSED_IO

#ifdef ZLIB_COMPRESSED_IO
typedef gzFile zFile;
#else
typedef FILE * zFile;
#endif !ZLIB_COMPRESSED_IO
/*                                           */
/* type for files that are never compressed */
/*                                           */
typedef FILE * zFileNC;

#ifdef PTHREADS
#include  <pthread.h>
#endif PTHREADS

#ifdef MPI
#include  <mpi.h>
extern int  gMPIrank;
extern int  gMPIsize;
extern MPI_Status  gMPIstatus;
extern int  gMPIBlocking;
#endif MPI

/*                                           */
/* constants and global utilities            */
/*                                           */

#include <zls_const.h>
#include <zls_inlines.h>
#include <zls_utility.h>

/*                                           */
/* low level structures and routines         */
/*                                           */

#include <zls_arrs.h>
#include <zls_grids.h>
#include <zls_qs.h>
#include <zls_vars.h>

/*                                           */
/* ZLS fundamental zone structure            */
/*                                           */

#include <zls_zones.h>

/* Parallel code if MPI is used              */

#include <zls_MPI.h>

/*                                           */
/* ZLS graphics                              */
/*                                           */

#include <zls_tiffio.h>
#include <zls_jpegio.h>
#include <zls_pngio.h>
#include <zls_graph.h>

/*                                           */
/*  ZLS IO                                   */
/*                                           */
 
#include <zls_io.h>
#include <zls_log.h>
#include <zls_data.h>
#include <zls_fits.h>
#include <zls_html.h>
#include <zls_images.h>
#include <zls_userio.h>
#include <zls_conf.h>
#include <zls_init.h>
#include <zls_restart.h>


/* primitive variable hydro utility routines */
#include <zls_hydro.h>

/* physics routines                          */
 
#include <zls_physics_const.h>
#include <zls_cooling.h>
#include <zls_gravity.h>
#include <zls_physics.h>

/*                                           */
/* user defined problem and definitions      */
/* placed here so that integration routines  */
/* can see them.                             */
/*                                           */

#include <zls_problem.h>

/*                                           */
/* general ZLS simulation routines           */
/* and structures                            */
/*                                           */

#include <zls_timing.h>
#include <zls_levels.h>

/*                                           */
/* inlines here use gLevels, so come after   */
/*  zls_levels.h                             */
/*                                           */

#include <zls_physinlines.h>


/*                                           */
/* algorithm dependent boundary conditions   */
/*                                           */

#include <zls_boundary.h>

/*                                           */
/* hydro routines                            */
/*                                           */

#include <zls_riemann.h>
#include <zls_ppm.h>
#include <zls_cellupdate.h>
#include <zls_sweeps.h>

/*                                           */
/* entry points from main.c                  */
/*                                           */

#include <zls_algorithm.h>

#endif
