#ifndef ZLS_OSDEFS_H
#define ZLS_OSDEFS_H

//================================================================================
// Computer/Achitecture related defines for ZLS, 
// set OS, multi-processing,
// precisions etc plus debug level.
//================================================================================
//
#define ZLS_DEBUG   0
#define ZLS_VERBOSE 1
//
#define OS_GENERIC
//#define OS_GENERICLINUX
//#define OS_CYGWIN
//#define OS_SOLARISSPARC
//#define OS_SOLARISAMD64
//#define OS_MACOSXG4
//#define OS_MACOSXG5
//#define OS_MACOSXCORE
// uncomment the following if running in MacOS X 10.5 or newer.
//#define OS_MACOSX105TIMING
//
//
//================================================================================
//
//  Parallel Processing Options
//
// Enable zone by zone parallel operation using the standard MPI library
// with MPI in this case each zone can be computed by a separate process
// and the processes coordinate at the beginning of each sweep to compute
// the overall timestep and to exchange any common boundaries.
//
// the grid.conf file can specify a process id for each zone, or they are
// allocated in order. 
//
// To use MPI the external MPI libraries and processes must be installed
// and running - this is up to the user to implement.
//
// MPI parallelises over mutliple processes with or without shared memory.
//
// Best if nCPUs > = nZones
//
//#define MPI
//#define MPI_BLOCKALLOCATION
//#define MPI_MODALLOCATION
//#define MPI_DIVALLOCATION
//
// In the PTHREADS implementation, a very local modification is made to
// zls_algorithm.c where each zone in a level sweep is assigned to a
// sweeper thread after BCs are set. The threads are queued up and after
// the zone loop for the level the threads are run and released. This
// creates and destroys threads on the fly, but typically the sweeps are
// the long part of the process.  This model works best on
// systems/grid.confs  with 1 cpu per zone.  Other setups may actually
// run slower.  The pthreads run within a single zls process in a shared
// memory environment using mutex's and wait conditions to synchronise.
//
// Set NCPU to the actual number of cpus involved to make the timing cell
// rates reflect the true cell rates.  NCPU has no effect currently on the
// actual threading, it is merely cosmetic at this point, but it may be
// used in future to optimise the threading strategy. The number of
// threads depends only on the number of zones per level, and can be more
// or less than the number of cpus, but equal to the number of cpus may be
// most efficient on some systems. Best to trial ZLS_MAXPTHREADS == NCPUs
// to start with and experiment within a factor of 2 or so. if nZones = 1
// then it probably doens't help to thread at all.  Keep ZLS_MAXPTHREADS >
// 1 >= NCPUs If you are setting ZLS_MAXPTHREADS to 1 then consider
// turning PTHREADS off altogether.
//
//#define PTHREADS
#define NCPUs 1
#define ZLS_MAXPTHREADS 1
//
// Enable additional code to sort zones by previous computation timing
// and compute sweeps in order of hardest to easiest.  
// Important for parallel algorithms with many zones and sweeps in 
// batches of n < nZones.  Fairly lightweight algorithm so it is probably 
// safe to just leave it on, as the worst case only adds a few micro
// seconds to a sweep
//
#define ZLS_LOADSORTING
//
//================================================================================
// SINGLE_PRECISION = 32 bit floating pt, fast, but low range/accuracy, less memory
// DOUBLE_PRECISION = 64 bit floating pt, best range/accuracy, more memory
//#define SINGLE_PRECISION
#define DOUBLE_PRECISION
//
//================================================================================
//
// Noise Control - default on.
//
// This activates additional fp code to control floating point 'noise'
// in the form of time and space coordinate bit variations and
// unphysical velocity perturbations. sub types allow it to be turned
// on and off in different areas of calculation for testing
//
// Default: ACTIVE_FP_NOISE_CONTROL + CONTROL_EVOLVE for speed and
// effectiveness. 
//
#define ACTIVE_FP_NOISE_CONTROL
#define CONTROL_EVOLVE
//#define CONTROL_REMAP
//
//================================================================================
//
// Optimizations
//
// An important optimisation that works well in many simulations, the
// cell velocities and positions are checked for a row before
// remapping. turn off for a simulation with multi-dimensional movement
// everywhere as it requires additional ifs that are unecessary in that
// case. Leave on by default, works well in most cases, especially in
// early stages. Especially effective for shocks in 3D when most
// movement is in the x-direction. Default is on (defined) as it will
// always work and adds only a little in the worst case, which may be
// very rare.
//
#define PRETEST_W_VEL
//
// Here a row of equal values have constant PPM coefficients set
// without going through the full grid interpolations.  Only works well
// with floating point noise control in force, as exact equality is
// used.  Will be modified to use zls_tolx next
// 
#define PRETESTPPMCOEFFS
//
//================================================================================
//
// Storage Settings - default ZONEDATA_ARE_MULTIDARRAYS
//
// ZONEDATA_ARE_PLAINBLOCKS: 
// 1D 2D 3D zonedata blocks are single malloc blocks, with or without BCs above.
//
// ZONEDATA_ARE_MULTIDARRAYS:
// 2D and 3D arrays with precompted offsets in lower dim blocks.  Each
// array is nD mallocs use special getters and setters to extract the main
// nD block for IO 1D data are still a single malloc block.  If zone data
// are Vars then they are always multi-d arrays
//
// #define ZONEDATA_ARE_PLAINBLOCKS
#define ZONEDATA_ARE_MULTIDARRAYS
//
//  if USEZONEVARS is in force, ZONEDATA_ARE_MULTIDARRAYS  is set automatically
//
//================================================================================
//
// Output Prefs
//
//================================================================================
//
// Performance Measurement Intervals
//
// compute cycles between cpu/cell/cycles progress printouts, used in zls_timing.
//
#define ZLS_PROGRESS_INTERVAL 64
//
// compute cycles between cpu/cell/cycles performance checks 
// must be an integer multiple of ZLS_PROGRESS_INTERVAL
//
#define ZLS_PERFORM_INTERVAL 256
//
//================================================================================
//
// Enable zlib/gzip compressed io for tif dat raw and all binary io
//
#define ZLIB_COMPRESSED_IO
//
// if compressed IO is set then it can be overridden for just the
// TIFF/JPEG IO allowing large (ie 3D) data cubes to be compressed, and
// the 2D TIffs to be uncompressed and easy to monitor (ie in the
// MacOSX Finder) ZLIB_UNCOMPRESSED_IMAGES is ignored if
// ZLIB_COMPRESSED_IO is not set
//
#define ZLIB_UNCOMPRESSED_IMAGES
//
// whether to keep the BCs in the output images/cubes or not.  Usually not
//
//================================================================================
//
#define FILES_NO_BCS
//#define FILES_HAVE_BCS
//
//
//================================================================================
// Do not edit below here, or if you do then it is at your own risk :)
//
//

#if defined(SINGLE_PRECISION)
typedef float  Real;
#elif defined(DOUBLE_PRECISION)
typedef double Real;
#else
#error "Not a valid precision flag"
#endif


#ifdef OS_GENERIC
typedef int      Integer;
typedef short    SInt16;
typedef char     SInt8;  // char treated as signed
typedef unsigned int    Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef unsigned short   UInt16;
typedef unsigned char    UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef unsigned long     Size;  
#endif OS_GENERIC

#ifdef OS_GENERICLINUX
typedef int      Integer;
typedef short    SInt16;
typedef char     SInt8;  // char treated as signed
typedef unsigned int    Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef unsigned short   UInt16;
typedef unsigned char    UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef unsigned long     Size;  
#endif OS_GENERICLINUX

#ifdef OS_CYGWIN
typedef int      Integer;
typedef short    SInt16;
typedef char     SInt8;  // char treated as signed
typedef unsigned int    Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef unsigned short   UInt16;
typedef unsigned char    UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef unsigned long     Size;  
#endif OS_CYGWIN

#ifdef OS_MACOSXG4
typedef int32_t    Integer;
typedef int16_t    SInt16;
typedef int8_t     SInt8;  // char treated as signed
typedef u_int32_t  Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef u_int16_t  UInt16; // unsigned 16 bit for 2D image IO , Tiffs and Raw.
typedef u_int8_t   UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef size_t     Size;  // 32 or 64 bit depending on OS, not used in saved structures. 
              // For memory and file size allocations only
#endif OS_MACOSXG4

#ifdef OS_MACOSXG5
typedef int32_t    Integer;
typedef int16_t    SInt16;
typedef int8_t     SInt8;  // char treated as signed
typedef u_int32_t  Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef u_int16_t  UInt16;
typedef u_int8_t   UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef size_t     Size;  // 32 or 64 bit depending on OS, not used in saved structures. 
              // For memory and file size allocations only
#endif OS_MACOSXG5

#ifdef OS_MACOSXCORE
typedef int32_t    Integer;
typedef int16_t    SInt16;
typedef int8_t     SInt8;  // char treated as signed
typedef u_int32_t  Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef u_int16_t  UInt16;
typedef u_int8_t   UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef size_t     Size;  // 32 or 64 bit depending on OS, not used in saved structures. 
              // For memory and file size allocations only
#endif OS_MACOSXCORE

#ifdef OS_SOLARISSPARC
typedef int32_t    Integer;
typedef int16_t    SInt16;
typedef int8_t     SInt8;  // char treated as signed
typedef uint32_t   Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef uint16_t   UInt16;
typedef uint8_t    UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef size_t     Size;  // 32 or 64 bit depending on OS, not used in saved structures. 
              // For memory and file size allocations only
#endif OS_SOLARISSPARC

#ifdef OS_SOLARISAMD64
typedef int32_t    Integer;
typedef int16_t    SInt16;
typedef int8_t     SInt8;  // char treated as signed
typedef uint32_t   Counter; // all 32 bit counters, ie cycles etc, and keeps binary dumps predictable.
typedef uint16_t   UInt16;
typedef uint8_t    UInt8;  // char treated as unsigned
typedef Counter UInt32; // convenience alias for Counter in low level routines
typedef Integer SInt32; // convenience alias for Integer in low level routines
typedef size_t     Size;  // 32 or 64 bit depending on OS, not used in saved structures. 
              // For memory and file size allocations only
#endif OS_SOLARISAMD64

typedef struct {
  Real  re, im;
} Complex;
#define c_re(c)  ((c).re)
#define c_im(c)  ((c).im)

#ifndef FLT_MAX
#define FLT_MAX (3.40282347E+38F)
#endif

#ifndef FLT_MIN
#define FLT_MIN (1.17549435E-38F)
#endif

#ifndef FLT_EPSILON
#define FLT_EPSILON (1.17549435E-8F)
#endif

#ifndef DBL_MAX
#define DBL_MAX (1.7976931348623157E+308)
#endif

#ifndef DBL_MIN
#define DBL_MIN (2.2250738585072014E-308)
#endif

#ifndef DBL_EPSILON
#define DBL_EPSILON (2.2204460492503131E-016)
#endif

#ifndef MAX_INTEGER
#define MAX_INTEGER (2L<<30)
#endif MAX_INTEGER

#ifdef DOUBLE_PRECISION
#ifdef OS_MACOSXCORE
#define      zls_isnan( x )   __isnand ((x))
#endif OS_MACOSXCORE

#ifdef OS_MACOSXG5
#define      zls_isnan( x )   __isnand ((x))
#endif OS_MACOSXG5

#ifdef OS_MACOSXG4
#define      zls_isnan( x )   __isnand ((x))
#endif OS_MACOSXG4

#ifdef OS_SOLARISSPARC
#define      zls_isnan( x )   isnand ((x))
#endif OS_SOLARISSPARC

#ifdef OS_SOLARISAMD64
#define      zls_isnan( x )   isnand ((x))
#endif OS_SOLARISAMD64

#ifdef OS_GENERIC
#define      zls_isnan( x )   isnan ((x))
#endif OS_GENERIC

#ifdef OS_GENERICLINUX
#define      zls_isnan( x )   isnan ((x))
#endif OS_GENERICLINUX

#ifdef OS_CYGWIN
#define      zls_isnan( x )   isnan ((x))
#endif OS_CYGWIN

#endif DOUBLE_PRECISION

#ifdef SINGLE_PRECISION
#ifdef OS_MACOSXCORE
#define      zls_isnan( x )   __isnanf ((x))
#endif OS_MACOSXCORE

#ifdef OS_MACOSXG5
#define      zls_isnan( x )   __isnanf ((x))
#endif OS_MACOSXG5

#ifdef OS_MACOSXG4
#define      zls_isnan( x )   __isnanf ((x))
#endif OS_MACOSXG4

#ifdef OS_SOLARISSPARC
#define      zls_isnan( x )   isnanf ((x))
#endif OS_SOLARISSPARC

#ifdef OS_SOLARISAMD64
#define      zls_isnan( x )   isnanf ((x))
#endif OS_SOLARISAMD64

#ifdef OS_GENERIC
#define      zls_isnan( x )   isnan ((x))
#endif OS_GENERIC

#ifdef OS_GENERICLINUX
#define      zls_isnan( x )   isnan ((x))
#endif OS_GENERICLINUX

#ifdef OS_CYGWIN
#define      zls_isnan( x )   isnan ((x))
#endif OS_CYGWIN

#endif SINGLE_PRECISION

//
// Sanity check if PTHREADS are not defined and make NCPUs default to 1
// if NCPUs is still not defined then set it to 1
//
#ifndef PTHREADS
#ifdef NCPUs
#undef NCPUs
#endif
#define NCPUs 1
#endif PTHREADS

#ifndef NCPUs
#define NCPUs 1
#endif

//
// flags for general data IO.  TIFFIO has it's own internal endian flags.
//

#ifdef OS_GENERICLINUX
#define DATA_LITENDIAN_OS
#endif OS_GENERICLINUX

#ifdef OS_CYGWIN
#define DATA_LITENDIAN_OS
#endif OS_CYGWIN

#ifdef OS_SOLARISAMD64
#define DATA_LITENDIAN_OS
#endif OS_SOLARISAMD64

#ifdef OS_SOLARISSPARC
#define DATA_BIGENDIAN_OS
#endif OS_SOLARISSPARC

#ifdef OS_MACOSXG4
#define DATA_BIGENDIAN_OS
#endif OS_MACOSXG4

#ifdef OS_MACOSXG5
#define DATA_BIGENDIAN_OS
#endif OS_MACOSXG5

#ifdef OS_MACOSXCORE
#define DATA_LITENDIAN_OS
#endif OS_MACOSXCORE

// default DATA_LITENDIAN_OS if nothing matches

#ifndef DATA_BIGENDIAN_OS
#ifndef DATA_LITENDIAN_OS
#define DATA_LITENDIAN_OS
#endif
#endif

//
//========================================
//
//  Binary Magic Tag  [ makes most sense in bignendian :) ]
//
// The Magic Tag is used in binary data headers. On systems with Big
// Endian or Little Endian byte order the MACIGTAG will appear in the
// file header in  normal or reversed format. Comparing it with the
// other endian constants allows the program to determine the file
// format independent of origin and decide any strategy at that point.
//
//
#define ZLS_MAGICTAG            0x5a4c5321
#define CURRENT_ENDIAN_MAGIC    0x5a4c5321
#define OPPOSITE_ENDIAN_MAGIC   0x21534c5a
//
//  On many systems the IO endianness can be set at the compiler
//  level so if many files are to be swapped it may be better to 
//  recompile.  On the other hand, reading in a swapped file and writing it
//  out again can produce a new set of files for other purposes.
//
//
//========================================
//
//  Here down are automatic/consistency checks, do not edit.
//
//
// if zones use vars then they are always multi-d arrays, this bit of macro gibberish 
// ensures this even if the settings are inconsistent above.
//
#ifdef USEZONEVARS
#ifdef ZONEDATA_ARE_PLAINBLOCKS 
#undef ZONEDATA_ARE_PLAINBLOCKS
#endif
#ifndef ZONEDATA_ARE_MULTIDARRAYS 
#define ZONEDATA_ARE_MULTIDARRAYS
#endif
#endif USEZONEVARS

#ifdef MPI
// make sure at least one allocation scheme is in force
#ifndef MPI_BLOCKALLOCATION
#ifndef MPI_MODALLOCATION
#ifndef MPI_DIVALLOCATION

#define MPI_BLOCKALLOCATION

#endif
#endif
#endif
#endif MPI

#endif ZLS_OSDEFS_H
