#include "zls.h"

#define rend_min( x , y ) ((x)<(y)?(x):(y))
#define rend_max( x , y ) (!((x)<(y))?(x):(y))

#define rend_min3( x, y , z ) (rend_min(rend_min((x),(y)),(z)))
#define rend_max3( x, y , z ) (rend_max(rend_max((x),(y)),(z)))

// internal render modes

enum { 
  kNoRenderMode      = 0x00, // don't draw
  kNormalRenderMode  = 0x01, // single channel
  kRGBRenderMode     = 0x03  // 3 channel RGB
};

enum {  k8Bit  = 0x00, //
  k16Bit  = 0x01,  //
  kS16Bit  = 0x02  //
};
// output resolutions - bits per pixel, formats


static   int renderMode    = kNoRenderMode ;
static   int outBitsPerPix = k8Bit ;
static   int nxImageOut   = 0;
static   int nyImageOut   = 0;

static UInt8 * image8 = NULL;
static UInt16 * image16 = NULL;
static SInt16 * imageS16 = NULL;

static UInt8 * image8R = NULL;
static UInt8 * image8G = NULL;
static UInt8 * image8B = NULL;

static UInt16 * image16R = NULL;
static UInt16 * image16G = NULL;
static UInt16 * image16B = NULL;

static SInt16 * imageS16R = NULL;
static SInt16 * imageS16G = NULL;
static SInt16 * imageS16B = NULL;

int zlsg_Reset     (void){
  int err = noErr;
  
  renderMode    = kNoRenderMode ;
  outBitsPerPix = k8Bit ;
  nxImageOut   = 0;
  nyImageOut   = 0;
  
  image8   = NULL;
  image16  = NULL;
  imageS16 = NULL;
  
  image8R = NULL;
  image8G = NULL;
  image8B = NULL;
  
  image16R = NULL;
  image16G = NULL;
  image16B = NULL;
  
  imageS16R = NULL;
  imageS16G = NULL;
  imageS16B = NULL;
  return(err);
}  

int zlsg_SetImage8     (UInt8   * img, int nxOut, int nyOut ){
  int err = noErr;
  image8 = img;
  image16 = NULL;
  imageS16 = NULL;
  
  image8R = NULL;
  image8G = NULL;
  image8B = NULL;
  
  image16R = NULL;
  image16G = NULL;
  image16B = NULL;
  
  imageS16R = NULL;
  imageS16G = NULL;
  imageS16B = NULL;
  
  renderMode    = kNormalRenderMode ;
  outBitsPerPix = k8Bit ;
  nxImageOut =   nxOut;
  nyImageOut =   nyOut;
  
  return(err);
  
}

int zlsg_SetImage16    (UInt16  * img, int nxOut, int nyOut ){
  int err = noErr;
  image8  = NULL;
  image16 = img;
  imageS16 = NULL;
  
  image8R = NULL;
  image8G = NULL;
  image8B = NULL;
  
  image16R = NULL;
  image16G = NULL;
  image16B = NULL;
  
  imageS16R = NULL;
  imageS16G = NULL;
  imageS16B = NULL;
  
  renderMode    = kNormalRenderMode ;
  outBitsPerPix = k16Bit ;
  nxImageOut =   nxOut;
  nyImageOut =   nyOut;
  
  return(err);
  
}


int zlsg_SetImage8RGB  (UInt8   * imgR, UInt8   * imgG, UInt8   * imgB, int nxOut, int nyOut )
{
  int err = noErr;
  
  image8  = NULL;
  image16 = NULL;
  imageS16 = NULL;
  
  image8R = imgR;
  image8G = imgG;
  image8B = imgB;
  
  image16R = NULL;
  image16G = NULL;
  image16B = NULL;
  
  imageS16R = NULL;
  imageS16G = NULL;
  imageS16B = NULL;
  
  renderMode    = kRGBRenderMode ;
  outBitsPerPix = k8Bit ;
  nxImageOut =   nxOut;
  nyImageOut =   nyOut;
  
  return(err);
  
}

int zlsg_SetImage16RGB (UInt16  * imgR, UInt16  * imgG, UInt16  * imgB, int nxOut, int nyOut ) {
  int err = noErr;
  
  image8  = NULL;
  image16 = NULL;
  imageS16 = NULL;
  
  image8R = NULL;
  image8G = NULL;
  image8B = NULL;
  
  image16R = imgR;
  image16G = imgG;
  image16B = imgB;
  
  imageS16R = NULL;
  imageS16G = NULL;
  imageS16B = NULL;
  
  renderMode    = kRGBRenderMode ;
  outBitsPerPix = k16Bit ;
  nxImageOut =   nxOut;
  nyImageOut =   nyOut;
  
  return(err);
  
}

int DrawPixel( int i, int j){
  int err = noErr;
  if (renderMode > kNoRenderMode) {
    int jRev  = nyImageOut-j-1;
    int maxij = nxImageOut*nyImageOut;
    int ij = i + nxImageOut*jRev;
    
    if ((ij >= 0) && (ij < maxij)) {
      
      if (renderMode == kRGBRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8R [ij]  = 0xFF;
          image8G [ij]  = 0xFF;
          image8B [ij]  = 0xFF;
          
        }
        if (outBitsPerPix == k16Bit)  {
          image16R[ij]  = 0xFFFF;
          image16G[ij]  = 0xFFFF;
          image16B[ij]  = 0xFFFF;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16R[ij]  = 0x7FFF;
          imageS16G[ij]  = 0x7FFF;
          imageS16B[ij]  = 0x7FFF;
        }
      }
      
      if (renderMode == kNormalRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8  [ij]  = 0xFF;
        }
        if (outBitsPerPix == k16Bit)  {
          image16 [ij]  = 0xFFFF;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16 [ij]  = 0x7FFF;
        }
      }
      
    }
  }
  return(err);
}

int DrawVisiblePixel( int i, int j){
  int err = noErr;
  if (renderMode > kNoRenderMode) {
    
    int jRev  = nyImageOut-j-1;
    int maxij = nxImageOut*nyImageOut;
    int ij = i + nxImageOut*jRev;
    
    if ((ij > 0) && (ij < maxij)) {
      
      if (renderMode == kRGBRenderMode) {
        if (outBitsPerPix == k8Bit) {
          unsigned char v = image8R [ij];
          v = v < 0xAA?0xFF:0x00;
          image8R [ij]  = v;
          
          v = image8G [ij];
          v = v < 0xAA?0xFF:0x00;
          image8G [ij]  = v;
          
          v = image8B [ij];
          v = v < 0xAA?0xFF:0x00;
          image8B [ij]  = v;
        }
        if (outBitsPerPix == k16Bit)  {
          unsigned short v = image16R [ij];
          v = v < 0xAAAA?0xFFFF:0x0000;
          image16R[ij]  = v;
          
          v = image16G [ij];
          v = v < 0xAAAA?0xFFFF:0x0000;
          image16G[ij]  = v;
          
          v = image16B [ij];
          v = v < 0xAAAA?0xFFFF:0x0000;
          image16B[ij]  = v;
        }
        if (outBitsPerPix == kS16Bit)  {
          short v = imageS16R [ij];
          v = v < 20000?32000:0x0000;
          imageS16R[ij]  = v;
          
          v = imageS16G [ij];
          v = v < 20000?32000:0x0000;
          imageS16G[ij]  = v;
          
          v = imageS16B [ij];
          v = v < 20000?32000:0x0000;
          imageS16B[ij]  = v;
        }
      }
      
      if (renderMode == kNormalRenderMode) {
        if (outBitsPerPix == k8Bit) {
          unsigned char v = image8 [ij];
          v = v < 0xAA?0xFF:0x00;
          image8 [ij]  = v;
        }
        if (outBitsPerPix == k16Bit)  {
          unsigned short v = image16 [ij];
          v = v < 0xAAAA?0xFFFF:0x0000;
          image16[ij]  = v;
        }
        if (outBitsPerPix == kS16Bit)  {
          short v = imageS16 [ij];
          v = v < 20000?32000:0x0000;
          imageS16[ij]  = v;
        }
      }
      
    }
  }
  return(err);
}

int SetPixelInteger( int i, int j, Integer idx){
  int err = noErr;
  if (renderMode > kNoRenderMode) {
    
    int jRev  = nyImageOut-j-1;
    int maxij = nxImageOut*nyImageOut;
    int ij = i + nxImageOut*jRev;
    
    if ((ij > 0) && (ij < maxij)) {
      
      if (renderMode == kRGBRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8R [ij]  = (UInt8)0x000000FF&idx;
          image8G [ij]  = (UInt8)0x000000FF&idx;
          image8B [ij]  = (UInt8)0x000000FF&idx;
        }
        if (outBitsPerPix == k16Bit)  {
          image16R[ij]  = (UInt16)0x0000FFFF&idx;
          image16G[ij]  = (UInt16)0x0000FFFF&idx;
          image16B[ij]  = (UInt16)0x0000FFFF&idx;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16R[ij]  = (SInt16)0x0000FFFF&idx;
          imageS16G[ij]  = (SInt16)0x0000FFFF&idx;
          imageS16B[ij]  = (SInt16)0x0000FFFF&idx;
        }
      }
      
      if (renderMode == kNormalRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8  [ij]  = (UInt8)0x000000FF&idx;
        }
        if (outBitsPerPix == k16Bit)  {
          image16 [ij]  = (UInt8)0x0000FFFF&idx;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16 [ij]  = (SInt8)0x0000FFFF&idx;
        }
      }
      
    }
  }
  return(err);
}

int SetPixelValue( int i, int j, float v){
  int err = noErr;
  if (renderMode > kNoRenderMode) {
    
    int jRev  = nyImageOut-j-1;
    int maxij = nxImageOut*nyImageOut;
    int ij = i + nxImageOut*jRev;
    
    float f = 0.0F;
    f  = v < 0.0F ? 0.0F: v;
    f  = v > 1.0F ? 1.0F: v;
    
    if ((ij > 0) && (ij < maxij)) {
      
      if (renderMode == kRGBRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8R [ij]  = 0xFF*f;
          image8G [ij]  = 0xFF*f;
          image8B [ij]  = 0xFF*f;
          
        }
        if (outBitsPerPix == k16Bit)  {
          image16R[ij]  = 0xFFFF*f;
          image16G[ij]  = 0xFFFF*f;
          image16B[ij]  = 0xFFFF*f;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16R[ij]  = 0x7FFF*f;
          imageS16G[ij]  = 0x7FFF*f;
          imageS16B[ij]  = 0x7FFF*f;
        }
      }
      
      if (renderMode == kNormalRenderMode) {
        if (outBitsPerPix == k8Bit) {
          image8  [ij]  = 0xFF*f;
        }
        if (outBitsPerPix == k16Bit)  {
          image16 [ij]  = 0xFFFF*f;
        }
        if (outBitsPerPix == kS16Bit)  {
          imageS16 [ij]  = 0x7FFF*f;
        }
      }
      
    }
  }
  return(err);
}

int SetRGBPixelValue( int i, int j, float vR, float vG, float vB);
int SetRGBPixelValue( int i, int j, float vR, float vG, float vB){
  int err = noErr;
  if (renderMode > kNoRenderMode) {
    
    int jRev  = nyImageOut-j-1;
    int maxij = nxImageOut*nyImageOut;
    int ij = i + nxImageOut*jRev;
    float f = 0.0F;
    
    
    if ((ij > 0) && (ij < maxij)) {
      
      if (renderMode == kRGBRenderMode) {
        if (outBitsPerPix == k8Bit) {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          image8R [ij]  = 0xFF*f;
          f  = vG < 0.0F ? 0.0F: vG;
          f  = vG > 1.0F ? 1.0F: vG;
          image8G [ij]  = 0xFF*f;
          f  = vB < 0.0F ? 0.0F: vB;
          f  = vB > 1.0F ? 1.0F: vB;
          image8B [ij]  = 0xFF*f;
          
        }
        if (outBitsPerPix == k16Bit)  {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          image16R[ij]  = 0xFFFF*f;
          f  = vG < 0.0F ? 0.0F: vG;
          f  = vG > 1.0F ? 1.0F: vG;
          image16G[ij]  = 0xFFFF*f;
          f  = vB < 0.0F ? 0.0F: vB;
          f  = vB > 1.0F ? 1.0F: vB;
          image16B[ij]  = 0xFFFF*f;
        }
        if (outBitsPerPix == kS16Bit)  {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          imageS16R[ij]  = 0x7FFF*f;
          f  = vG < 0.0F ? 0.0F: vG;
          f  = vG > 1.0F ? 1.0F: vG;
          imageS16G[ij]  = 0x7FFF*f;
          f  = vB < 0.0F ? 0.0F: vB;
          f  = vB > 1.0F ? 1.0F: vB;
          imageS16B[ij]  = 0x7FFF*f;
        }
      }
      
      if (renderMode == kNormalRenderMode) {
        if (outBitsPerPix == k8Bit) {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          image8  [ij]  = 0xFF*f;
        }
        if (outBitsPerPix == k16Bit)  {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          image16 [ij]  = 0xFFFF*f;
        }
        if (outBitsPerPix == kS16Bit)  {
          f  = vR < 0.0F ? 0.0F: vR;
          f  = vR > 1.0F ? 1.0F: vR;
          imageS16 [ij]  = 0x7FFF*f;
        }
      }
      
    }
  }
  return(err);
}

int zlsg_DrawLine(int i1, int j1, int i2, int j2){
  
  int temp = i1;
  if ( i1 > i2) {
    i1 = i2;
    i2 = temp;
  }
  
  temp = j1;
  if ( j1 > j2) {
    j1 = j2;
    j2 = temp;
  }
  
  int dj = j2 - j1;
  int di = i2 - i1;
  
  int i = i1;
  int j = j1;
  
  
  if ( dj == 0) {
    
    for ( i = i1; i <= i2; i++)  DrawPixel( i , j1);
    
  } else if ( di == 0) {
    
    for ( j = j1; j <= j2; j++) DrawPixel( i1, j );
    
  } else  if ( dj > di) {
    float dij = (float)di/(float)dj;
    float fi  = (float)i1;
    DrawPixel( i1 , j1 );
    for ( j = j1; j < j2; j++) {
      fi += dij;
      i = (int)fi; 
      DrawPixel( i , j );
    }
    DrawPixel( i2 , j2 );
    
  } else {
    float dji = (float)dj/(float)di;
    float fj  = (float)j1;
    DrawPixel( i1 , j1 );
    for ( i = i1; i < i2; i++) {
      fj += dji;
      j = (int)fj; 
      DrawPixel( i , j );
    }
    DrawPixel( i2 , j2 );
    
  }
  
  return(0);
}

inline int DrawLCChar( int i, int j, char c);
inline int DrawLCChar( int i, int j, char c){
  
  char lc = tolower(c);
  
  switch (lc ) {
    case 'a':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+0);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      
      break;
      
    case 'b':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      break;
      
    case 'c':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j+1);
      
      break;
      
    case 'd':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+7);
      DrawVisiblePixel(i+4, j+6);
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+0);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      
      break;
      
    case 'e':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+1, j+3);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j+1);
      break;
      
    case 'f':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+1, j+6);
      
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      break;
      
    case 'g':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+5);
      
      DrawVisiblePixel(i+4, j-1);
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+1, j-2);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+3, j-2);
      
      break;
      
    case 'h':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j  );
      
      break;
      
    case 'i':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+1, j+7);
      
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i+2, j  );
      
      break;
      
    case 'j':
      
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+5);
      
      DrawVisiblePixel(i+4, j+7);
      
      DrawVisiblePixel(i+4, j-1);
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+1, j-2);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+3, j-2);
      
      break;
      
    case 'k':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+4, j  );
      
      break;
      
    case 'l':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+1, j+6);
      DrawVisiblePixel(i+1, j+7);
      
      DrawVisiblePixel(i+2, j  );
      
      break;
      
    case 'm':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j+5);
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      break;
      
    case 'n':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j  );
      
      break;
      
    case 'o':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      break;
      
    case 'p':
      
      DrawVisiblePixel(i  , j-2);
      DrawVisiblePixel(i  , j-1);
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      break;
      
    case 'q':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+5);
      
      DrawVisiblePixel(i+4, j-2);
      DrawVisiblePixel(i+4, j-1);
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      break;
      
    case 'r':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case 's':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      break;
      
      
    case 't':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j+1);
      
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+1, j+6);
      DrawVisiblePixel(i+1, j+7);
      
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      break;
      
    case 'u':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j  );
      
      break;
      
    case 'v':
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j+3 );
      DrawVisiblePixel(i+3, j+2 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      
      break;
      
    case 'w':
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+2 );
      
      DrawVisiblePixel(i+1, j+1 );
      DrawVisiblePixel(i+1, j   );
      
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+2 );
      
      DrawVisiblePixel(i+3, j+1 );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      
      break;
      
    case 'x':
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j   );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+1, j+1 );
      
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+2 );
      
      DrawVisiblePixel(i+3, j+4 );
      DrawVisiblePixel(i+3, j+1 );
      
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    case 'y':
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j+3 );
      DrawVisiblePixel(i+3, j+2 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      
      DrawVisiblePixel(i  , j-2 );
      DrawVisiblePixel(i+1, j-1 );
      
      break;
      
    case 'z':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j   );
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i+1, j+5 );
      DrawVisiblePixel(i+2, j+5 );
      DrawVisiblePixel(i+3, j+5 );
      DrawVisiblePixel(i+4, j+5 );
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+3, j+4 );
      
      break;
      
    default:
      ;
  }
  return(0);
  
}

inline int DrawUCChar( int i, int j, char c);
inline int DrawUCChar( int i, int j, char c){
  
  char uc = toupper(c);
  
  switch (uc ) {
      
    case 'A':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+1, j+4 );
      
      DrawVisiblePixel(i+2, j+5 );
      DrawVisiblePixel(i+2, j+6 );
      DrawVisiblePixel(i+2, j+7 );
      
      DrawVisiblePixel(i+3, j+4 );
      DrawVisiblePixel(i+3, j+3 );
      
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j   );
      
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+3, j+2 );
      break;
      
    case 'B':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+4 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+5 );
      
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+1 );
      
      break;
      
    case 'C':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j+1 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      
      break;
      
    case 'D':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j+1 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+6 );
      
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      
      break;
      
    case 'E':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j   );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+4 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'F':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+4 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'G':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+3, j+3 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      
      break;
      
    case 'H':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+4 );
      
      DrawVisiblePixel(i+4, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    case 'I':
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+5 );
      DrawVisiblePixel(i+2, j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      break;
      
    case 'J':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+7 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      break;
      
    case 'K':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+1, j+4 );
      
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+2, j+5 );
      
      DrawVisiblePixel(i+3, j+1 );
      DrawVisiblePixel(i+3, j+6 );
      
      DrawVisiblePixel(i+4, j   );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'L':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    case 'M':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+6 );
      DrawVisiblePixel(i+1, j+5 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+6 );
      DrawVisiblePixel(i+3, j+5 );
      
      DrawVisiblePixel(i+4, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    case 'N':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+6 );
      DrawVisiblePixel(i+1, j+5 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+3, j+2 );
      DrawVisiblePixel(i+3, j+1 );
      
      DrawVisiblePixel(i+4, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    case 'O':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      break;
      
    case 'P':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+3, j+3 );
      
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      break;
      
    case 'Q':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      DrawVisiblePixel(i+2, j-1 );
      DrawVisiblePixel(i+3, j-2 );
      DrawVisiblePixel(i+4, j-2 );
      
      break;
      
    case 'R':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+3, j+3 );
      
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+3, j+1 );
      DrawVisiblePixel(i+4, j );
      
      break;
      
    case 'S':
      
      DrawVisiblePixel(i  , j+1 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+3, j+4 );
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      
      break;
      
    case 'T':
      
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+5 );
      DrawVisiblePixel(i+2, j+6 );
      
      DrawVisiblePixel(i  , j+7 );
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'U':
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+2 );
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+2 );
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'V':
      
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+3 );
      DrawVisiblePixel(i+1, j+2 );
      
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j   );
      
      DrawVisiblePixel(i+3, j+3 );
      DrawVisiblePixel(i+3, j+2 );
      
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'W':
      
      DrawVisiblePixel(i  , j+3 );
      DrawVisiblePixel(i  , j+4 );
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i  , j+7 );
      
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+1, j+1 );
      DrawVisiblePixel(i+1, j   );
      
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+5 );
      
      DrawVisiblePixel(i+3, j+2 );
      DrawVisiblePixel(i+3, j+1 );
      DrawVisiblePixel(i+3, j   );
      
      DrawVisiblePixel(i+4, j+3 );
      DrawVisiblePixel(i+4, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+4, j+7 );
      
      break;
      
    case 'X':
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i  , j+7 );
      DrawVisiblePixel(i  , j+6 );
      
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+1, j+5 );
      
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+4 );
      
      DrawVisiblePixel(i+3, j+2 );
      DrawVisiblePixel(i+3, j+5 );
      
      DrawVisiblePixel(i+4, j   );
      DrawVisiblePixel(i+4, j+1 );
      DrawVisiblePixel(i+4, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      
      break;
      
    case 'Y':
      
      DrawVisiblePixel(i  , j+7 );
      DrawVisiblePixel(i  , j+6 );
      DrawVisiblePixel(i+1, j+5 );
      DrawVisiblePixel(i+1, j+4 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j+5 );
      DrawVisiblePixel(i+3, j+4 );
      DrawVisiblePixel(i+4, j+7 );
      DrawVisiblePixel(i+4, j+6 );
      
      break;
      
    case 'Z':
      
      DrawVisiblePixel(i  , j+7 );
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+4, j+7 );
      
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+3, j+5 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i  , j+1 );
      
      DrawVisiblePixel(i  , j   );
      DrawVisiblePixel(i+1, j   );
      DrawVisiblePixel(i+2, j   );
      DrawVisiblePixel(i+3, j   );
      DrawVisiblePixel(i+4, j   );
      
      break;
      
    default:
      ;
  }
  return(0);
  
}


inline int DrawNum( int i, int j, char c);
inline int DrawNum( int i, int j, char c){
  
  switch (c ) {
      
    case '1':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j  );
      
      DrawVisiblePixel(i+2, j+1 );
      DrawVisiblePixel(i+2, j+2 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+2, j+4 );
      DrawVisiblePixel(i+2, j+5 );
      DrawVisiblePixel(i+2, j+6 );
      DrawVisiblePixel(i+2, j+7 );
      
      DrawVisiblePixel(i  , j+5 );
      DrawVisiblePixel(i+1, j+6 );
      
      break;
      
    case '2':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j  );
      
      DrawVisiblePixel(i  , j+1 );
      DrawVisiblePixel(i+1, j+2 );
      DrawVisiblePixel(i+2, j+3 );
      DrawVisiblePixel(i+3, j+4 );
      DrawVisiblePixel(i+4, j+5 );
      DrawVisiblePixel(i+4, j+6 );
      DrawVisiblePixel(i+3, j+7 );
      DrawVisiblePixel(i+2, j+7 );
      DrawVisiblePixel(i+1, j+7 );
      DrawVisiblePixel(i  , j+6 );
      
      break;
      
    case '3':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+3, j+4);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i  , j+6);
      
      break;
      
    case '4':
      
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+6);
      DrawVisiblePixel(i+3, j+7);
      
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+4, j+3);
      
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+6);
      
      break;
      
    case '5':
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+4, j+7);
      break;
      
    case '6':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i+1, j+6);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      break;
      
    case '7':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+2);
      
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      DrawVisiblePixel(i+4, j+7);
      
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      break;
      
    case '8':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      break;
      
    case '9':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+3, j+3);
      
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      break;
      
    case '0':
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+5);
      
      break;
      
    default:
      ;
  }
  
  return(0);
  
}

inline int DrawMisc( int i, int j, char c);
inline int DrawMisc( int i, int j, char c){
  
  switch (c ) {
      
    case '`': //squared
      
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+3, j+7);
      
      DrawVisiblePixel(i+1, j+8);
      DrawVisiblePixel(i+2, j+8);
      
      DrawVisiblePixel(i  , j+7);
      
      break;
      
    case '~': //pi
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+5);
      
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      
      break;
      
    case '!': //alpha
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+5);
      
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+4, j+1);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case '@': //beta
      
      DrawVisiblePixel(i  , j-2);
      DrawVisiblePixel(i  , j-1);
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+3, j+3);
      
      DrawVisiblePixel(i+1, j+6);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+3, j+6);
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case '#': //gamma
      
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j-2);
      DrawVisiblePixel(i+1, j-1);
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+2, j+2);
      
      DrawVisiblePixel(i+3, j-1);
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case '$': //rho
      
      DrawVisiblePixel(i  , j-2);
      DrawVisiblePixel(i  , j-1);
      DrawVisiblePixel(i  , j  );
      
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      
      break;
      
    case '%': //phi
      
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+5);
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+5);
      
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      break;
      
    case '^': //theta
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j  );
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+3, j+3);
      
      DrawVisiblePixel(i+1, j+6);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+3, j+6);
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case '&': //mu
      
      DrawVisiblePixel(i  , j-2);
      DrawVisiblePixel(i  , j-1);
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      DrawVisiblePixel(i+4, j+5);
      
      break;
      
    case '*': //degree
      
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+2, j+4);
      
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+6);
      
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      
      break;
      
    case '(': 
      
      DrawVisiblePixel(i+4, j-3);
      DrawVisiblePixel(i+3, j-2);
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+4, j+8);
      
      break;
      
    case ')': 
      
      DrawVisiblePixel(i+1, j-3);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+3, j-1);
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+3);
      DrawVisiblePixel(i+4, j+4);
      
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+6);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+1, j+8);
      
      break;
      
    case '[': 
      
      DrawVisiblePixel(i+3, j-3);
      DrawVisiblePixel(i+4, j-3);
      DrawVisiblePixel(i+2, j-3);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+4, j+7);
      
      break;
      
    case ']': 
      
      DrawVisiblePixel(i+1, j-3);
      DrawVisiblePixel(i+2, j-3);
      DrawVisiblePixel(i+3, j-3);
      DrawVisiblePixel(i+3, j-2);
      DrawVisiblePixel(i+3, j-1);
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+6);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+1, j+7);
      
      break;
      
    case '{': 
      
      DrawVisiblePixel(i+4, j-3);
      DrawVisiblePixel(i+3, j-3);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+4, j+7);
      
      break;
      
    case '}': 
      
      DrawVisiblePixel(i  , j-3);
      DrawVisiblePixel(i+1, j-3);
      DrawVisiblePixel(i+2, j-3);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i  , j+7);
      
      break;
      
    case '<': // Lambda
      
      DrawVisiblePixel(i-1, j  );
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+1, j+4);
      DrawVisiblePixel(i+1, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+4, j+2);
      DrawVisiblePixel(i+4, j+1);
      DrawVisiblePixel(i+4, j  );
      DrawVisiblePixel(i+5, j  );
      
      break;
      
    case '>': // Gamma
      
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i  , j+4);
      DrawVisiblePixel(i  , j+5);
      DrawVisiblePixel(i  , j+6);
      DrawVisiblePixel(i  , j+7);
      DrawVisiblePixel(i+1, j+7);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+4, j+7);
      DrawVisiblePixel(i+4, j+6);
      
      break;
      
    case '|': // integral
      
      DrawVisiblePixel(i  , j-3);
      DrawVisiblePixel(i  , j-2);
      DrawVisiblePixel(i+1, j-3);
      DrawVisiblePixel(i+1, j-2);
      DrawVisiblePixel(i+2, j-2);
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+2, j+6);
      DrawVisiblePixel(i+2, j+7);
      DrawVisiblePixel(i+3, j+7);
      DrawVisiblePixel(i+3, j+8);
      DrawVisiblePixel(i+4, j+7);
      DrawVisiblePixel(i+4, j+8);
      
      break;
      
    case '?': // sqrt
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i  , j  );
      DrawVisiblePixel(i  , j-1);
      
      DrawVisiblePixel(i+1, j-3);
      DrawVisiblePixel(i+1, j-2);
      
      DrawVisiblePixel(i+2, j-1);
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+3, j+4);
      
      DrawVisiblePixel(i+4, j+5);
      DrawVisiblePixel(i+4, j+6);
      DrawVisiblePixel(i+4, j+7);
      
      DrawVisiblePixel(i+5 , j+8);
      DrawVisiblePixel(i+6 , j+8);
      DrawVisiblePixel(i+7 , j+8);
      DrawVisiblePixel(i+8 , j+8);
      DrawVisiblePixel(i+9 , j+8);
      DrawVisiblePixel(i+10, j+8);
      DrawVisiblePixel(i+11, j+8);
      DrawVisiblePixel(i+12, j+8);
      
      break;
      
    case '=':
      
      DrawVisiblePixel(i  , j+1);
      DrawVisiblePixel(i+1, j+1);
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+4, j+1);
      
      DrawVisiblePixel(i  , j+3);
      DrawVisiblePixel(i+1, j+3);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+3, j+3);
      DrawVisiblePixel(i+4, j+3);
      
      break;
      
    case '+':
      
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+4, j+2);
      
      DrawVisiblePixel(i+2, j+0);
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+2, j+3);
      DrawVisiblePixel(i+2, j+4);
      break;
      
    case '-':
      
      DrawVisiblePixel(i  , j+2);
      DrawVisiblePixel(i+1, j+2);
      DrawVisiblePixel(i+2, j+2);
      DrawVisiblePixel(i+3, j+2);
      DrawVisiblePixel(i+4, j+2);
      
      break;
      
    case '.':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j  );
      
      break;
      
    case ':':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j  );
      
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+4);
      
      break;
      
    case ',':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j-1);
      DrawVisiblePixel(i+2, j-2);
      
      break;
      
    case ';':
      
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j  );
      DrawVisiblePixel(i+3, j-1);
      DrawVisiblePixel(i+2, j-2);
      
      DrawVisiblePixel(i+2, j+4);
      DrawVisiblePixel(i+2, j+5);
      DrawVisiblePixel(i+3, j+5);
      DrawVisiblePixel(i+3, j+4);
      
      break;
      
    case ' ':
      
      break;
      
    default:
      DrawVisiblePixel(i+2, j  );
      DrawVisiblePixel(i+2, j+1);
      DrawVisiblePixel(i+3, j+1);
      DrawVisiblePixel(i+3, j  );
  }
  
  return(0);
  
}

int zlsg_DrawChar( int i, int j, char c){
  
  if ( isalpha(c) ) {
    if ( isupper(c) ) DrawUCChar( i, j, c);
    if ( islower(c) ) DrawLCChar( i, j, c);
  } else if ( isdigit(c) ){
    DrawNum( i, j, c);
  } else {
    DrawMisc( i, j, c);
  }
  return(0);
}

int zlsg_DrawString( int i, int j, char * c){
  
  char * ct = c;
  
  while (*ct){
    
    zlsg_DrawChar( i, j, *ct);
    
    i += 6;
    ct++;
    
  }
  
  return(0);
  
}

int zlsg_DrawBar( int i1, int j1, int i2, int j2){
  // 1.0/256
#define eps 0.00390625
  
  int dj = j2 - j1;
  int di = i2 - i1;
  
  int i = i1;
  int j = j1;
  double dc = 1.0;
  
  if ( di >= dj) {
    dc = (1.0-(2.0*eps))/di;
  } else {
    dc = (1.0-(2.0*eps))/dj;
  }  
  
  if ( dj == 0) {
    
    for ( i = i1; i <= i2; i++)  SetPixelValue( i , j1, eps+(i-i1)*dc);
    
  } else if ( di == 0) {
    
    for ( j = j1; j <= j2; j++) SetPixelValue( i1, j , eps+(j-j1)*dc);
    
  } else  if (( dj > di)&&(dj > 0)) {
    
    for ( i = i1; i < i2; i++) {
      for ( j = j1; j <= j2; j++) {
        SetPixelValue( i , j , eps+(j-j1)*dc);
      }
    }
    
  } else if (( di > dj)&&(di > 0)) {
    
    for ( j = j1; j < j2; j++) {
      for ( i = i1; i <= i2; i++) {
        SetPixelValue( i , j , eps+(i-i1)*dc);
      }
      // SetPixelValue( i2 , j2 , 1.0-eps);
    }
    
  }
  
  return(0);
}

int zlsg_ClearRect ( int i1, int j1, int i2, int j2){
  int i, j;
  
  for ( j = j1; j < j2; j++) {
    for ( i = i1; i < i2; i++) {
      SetPixelValue( i , j ,  0.0F);
    }
  }
  
  return(0);
}

int zlsg_FrameRect ( int i1, int j1, int i2, int j2){
  
  zlsg_DrawLine(i1, j1, i2, j1);
  zlsg_DrawLine(i2, j1, i2, j2);
  zlsg_DrawLine(i2, j2, i1, j2);
  zlsg_DrawLine(i1, j2, i1, j1);
  
  return(0);
}

