#include "zls.h"

int zlst_HTMLElapsedRateSummary( zFileNC f, CPUCoordPtr ct, double nCells, Integer level, Real speedup ){
  
  int err = noErr;
#ifdef OS_GENERIC
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_GENERICLINUX
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_CYGWIN
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_MACOSXG4
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXG5
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXCORE
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_SOLARISSPARC
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_SOLARISAMD64
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
  double usr, tot, rea, etc, etf, model, cycps, ccycps, ro;
  Counter cyc;
  
  if ( f != NULL) {
	  if (zlst_GlobalCycles == 0) {
      // reset the clock to remove setup overhead
      zlst_StartCPUTimes( ct );
	  }
	  usr = ((double)(ct->cpuusr[1]-ct->cpuusr[0]))*spc;
	  tot = ((double)(ct->cputot[1]-ct->cputot[0]))*spc;
	  rea = ((double)(ct->realtime[1]-ct->realtime[0]))*spc; // real time is real, so don't scale by NCPUs
	  model = ((ct->modeltime[1]-ct->modeltime[0])); // model time is real, so don't scale by NCPUs
	  cyc  = ((ct->modelcycles[1]-ct->modelcycles[0])); // model in cycles is real, so don't scale by NCPUs
	  
    etc   = 0.0;
    etf   = 0.0;
    ccycps = 0.0;
    cycps = 0.0;
    ro    = 0.0;
    
    // note: these are all real time values, not system timesliced values
    
    if ( zlst_GlobalTime > 0 ) {
      etf    = rea*(zlst_GlobalFrameTime/zlst_GlobalTime);
      etc    = rea*((zlst_GlobalEndTime/zlst_GlobalTime)-1.0);
      cycps  = cyc/rea;
      ccycps = nCells*cyc/rea;
      ro     = ( model /rea ); 
    }
#ifdef PTHREADS    
    
	  zlsio_FPrintf( f, "<code>\n");
	  
	  if ( zlst_GlobalCycles > 0 ) {
      char tsofar[80];
      strcpy(tsofar, zls_TimeIntervalString(rea));
      char togo[80];
      strcpy(togo, zls_TimeIntervalString(etc));
      char toframe[80];
      strcpy(toframe, zls_TimeIntervalString(etf));
      zlsio_FPrintf( f,  "<b>Elapsed Time : %s  Estimated Completion : %s</b><br> Estimated Next Frame : %s <br><br>\n", tsofar, togo, toframe);
    }
	  
	  if ( f_t0 != 1.0 ) {
      zlsio_FPrintf( f, " <b>Cycle:</b> %10d\n", zlst_GlobalCycles);
      zlsio_FPrintf( f, " <b>Time :</b> %#10.5g = %#10.5g %s = %#10.5g %s\n", 
                    zlst_GlobalTime    , zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr<br>");
      zlsio_FPrintf( f, " <b>Delta:</b> %#10.5g = %#10.5g %s = %#10.5g %s\n",
                    zlst_GlobalTimeStep, zlst_GlobalSecondsStep, s_Time, zlst_GlobalYearsStep, "yr<br>");
	  } else {
      zlsio_FPrintf( f, " <b>Cycle:</b> %10d   Time: %#10.5g   Delta: %#10.5g<br>\n",
                    zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTimeStep);      
	  }
    
	  if ( zlst_GlobalTime > 0 ) {
		  zlsio_FPrintf( f,  "Overall Cycles/s : %#10.5g <br>\n", cycps);
		  zlsio_FPrintf( f,  "Overall Cell Cycles/s : %#10.5g <br>\n", ccycps);
		  if ( speedup > 1.01 ) zlsio_FPrintf( f,  "Effective Cell Cycles/s : %#10.5g <br>\n", ccycps*speedup);
		  zlsio_FPrintf( f, " <b>Reality Overdrive : %10.4g  (model/real time)</b> <br>\n",  ro );
	  }
	  zlsio_FPrintf( f, "</code>\n");
#else
	  
#ifdef MPI
	  if (gMPIrank == level) {
#endif MPI
		  
      zlsio_FPrintf( f, "<code>\n");
      if ( zlst_GlobalCycles > 0 ) {
        char tsofar[80];
        strcpy(tsofar, zls_TimeIntervalString(rea));
        char togo[80];
        strcpy(togo, zls_TimeIntervalString(etc));
        char toframe[80];
        strcpy(toframe, zls_TimeIntervalString(etf));
        if ( etc > 1.0e-6 ) {
          zlsio_FPrintf( f,  "<b>Elapsed Time : %s  Estimated Completion : %s</b><br> Estimated Next Frame : %s <br><br>\n", tsofar, togo, toframe);
        } else {
          zlsio_FPrintf( f,  "<b>Completed. Elapsed Time : %s Last Frame : %s <br><br>\n", tsofar, toframe);
        }
      }
      
      if ( f_t0 != 1.0 ) {
        zlsio_FPrintf( f, " <b>Cycle:</b> %10d\n", zlst_GlobalCycles);
        zlsio_FPrintf( f, " <b>Time :</b> %#10.5g = %#10.5g %s = %#10.5g %s\n", 
                      zlst_GlobalTime    , zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr<br>");
        zlsio_FPrintf( f, " <b>Delta:</b> %#10.5g = %#10.5g %s = %#10.5g %s\n",
                      zlst_GlobalTimeStep, zlst_GlobalSecondsStep, s_Time, zlst_GlobalYearsStep, "yr<br>");
      } else {
        zlsio_FPrintf( f, " <b>Cycle:</b> %10d   Time: %#10.5g   Delta: %#10.5g<br>\n",
                      zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTimeStep);      
      }
      
      if ( zlst_GlobalTime > 0 ) {
        
        zlsio_FPrintf( f,  "Overall Cycles/s : %#10.5g :: s/Cycle : %#10.5g <br>\n", cycps, 1.0/cycps);
        zlsio_FPrintf( f,  "Overall Cell Cycles/s :  %#10.5g <br>\n", ccycps);
        if ( speedup > 1.01 ) zlsio_FPrintf( f,  "Effective Cell Cycles/s : %#10.5g <br>\n", ccycps*speedup);
        zlsio_FPrintf( f, " <b>Reality Overdrive :%10.4g  (model/real time)</b> <br>\n",  ro );
        
      }
      zlsio_FPrintf( f, "</code>\n");
      
#ifdef MPI
	  }
#endif MPI
	  
#endif !PTHREADS
	  
  } // f!= NULL
  
  return(err);
}


int  zlsio_WriteHTMLImageDescription ( char * fname, IOSettingsPtr io, Integer ioFrameNumber, Integer ioIndex, Integer nxOut, Integer nyOut ){
  
  int err = noErr;
  char srcfile[PATHLEN];
  char textline[PATHLEN];
  char textline2[PATHLEN];
  double nCells = 0.0;
  double nEqivCells = 0.0;
  Integer nZones = 0;
  Integer level = 0;
  
  zFileNC f;
  zFileNC  hf;
  
  f =  zlsio_openNC ( fname, "w");
  
  sprintf( srcfile,"%s%s%s%s",DATADIR, PATHSEP, HTMLHEAD, HTMLEXT);
  hf =  zlsio_openNC ( srcfile, "r");
  if (( f != NULL ) && ( hf != NULL )){
    
    while (!feof(hf)){
      fgets(textline, PATHLEN, hf);
      fputs( textline, f);
    }
    
    zlsio_closeNC( hf );
  };
  
  Integer percentage =  100*(zlst_GlobalTime/zlst_GlobalEndTime);
  
  zlsio_FPrintf( f,  "<H1>%s Simulation Progress</H1><br>\n", FYRIS );
 
  zlsio_FPrintf( f,  "<code><b>%s</b></code><br>\n", zlst_TheCurrentRealTime());
  
  zlsio_FPrintf( f, "<h3>Simulation: %s, %4d&#37; Completed</h3>\n", FYRIS_PROBLEMNAME, percentage);
  
  if ( ioFrameNumber >= 0){
    zlsio_FPrintf( f, "<h5>\n");
    char prevname[PATHLEN];
    char nextname[PATHLEN];
    
    sprintf(prevname, "%s%4.4d%s", io->outMonitorName, 0, HTMLEXT );
    zlsio_FPrintf( f, "<a href=%s> Start </a> |\n",prevname);
    
    if (ioFrameNumber > 0){
      if(ioFrameNumber>=10) { 
        sprintf(prevname, "%s%4.4d%s", io->outMonitorName, (ioFrameNumber-10), HTMLEXT );
        zlsio_FPrintf( f, "<a href=%s> Previous-10 </a> |\n",prevname);
      } else {
        zlsio_FPrintf( f, " Previous-10 |\n");
      }
      sprintf(prevname, "%s%4.4d%s", io->outMonitorName, (ioFrameNumber-1), HTMLEXT );
      zlsio_FPrintf( f, "<a href=%s> Previous </a> |\n",prevname);
    } else {
      zlsio_FPrintf( f, " Previous-10 |\n");
      zlsio_FPrintf( f, " Previous |\n");
    }
    
    if ( ( 1.0-(zlst_GlobalTime/zlst_GlobalEndTime)) > 1.0e-6 ) {
      sprintf(nextname, "%s%4.4d%s", io->outMonitorName, (ioFrameNumber+1), HTMLEXT );
      zlsio_FPrintf( f, "<a href=%s> Next </a> |\n",nextname);
      sprintf(nextname, "%s%4.4d%s", io->outMonitorName, (ioFrameNumber+10), HTMLEXT );
      zlsio_FPrintf( f, "<a href=%s> Next+10 </a> |\n",nextname);
    }
    zlsio_FPrintf( f, "<a href=%s%s> Current </a>\n",io->outMonitorName, HTMLEXT );
    zlsio_FPrintf( f, "</h5>\n");
  } else {
    zlsio_FPrintf( f, "<h5>\n");
    zlsio_FPrintf( f, "<a href=%s%4.4d%s>Sequence</a>\n",io->outMonitorName, io->ioState.ioFrameCount, HTMLEXT );
    zlsio_FPrintf( f, "</h5>\n");
  }
  
  zlsio_FPrintf( f, "<code>\n");
  zlsio_FPrintf( f,  "<b>Run: </b>%s<b>, Step: </b>%4.4d <br> <br>\n",  io->outRunName, io->ioState.ioFrameCount );
  
  zls_CalcAllLevelsEffectiveCellNumber ( stdout, &nCells, kZLS_DontShowCellCount );
  zls_CalcAllLevelsZoneNumber          ( stdout, &nZones, kZLS_DontShowCellCount );
  nEqivCells = nCells;
  zls_CalcAllLevelsEquivNumberCellNumber( &nEqivCells);
  Integer nR  = gLevels[gNLevels-1].levelTotalRefinement;
  
#ifdef SYM1D
#ifdef CARTESIAN
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cartesian (<i>x</i>) \n");
#endif CARTESIAN
#ifdef SPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Spherical (<i>r</i>) \n");
#endif SPHERICAL
#ifdef POLARSPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Spherical (<i>r</i>) \n");
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cylindrical (<i>z</i>) \n");
#endif CYLINDRICAL
#endif SYM1D
#ifdef SYM2D
#ifdef CARTESIAN
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cartesian (<i>x</i>, <i>y</i>) \n");
#endif CARTESIAN
#ifdef SPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Spherical (<i>r</i>, <i>phi</i>) \n");
#endif SPHERICAL
#ifdef POLARSPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Spherical (<i>r</i>, <i>phi</i>) \n");
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cylindrical (<i>z</i>, <i>r</i>) \n");
#endif CYLINDRICAL
#ifdef CIRCULAR
  zlsio_FPrintf( f, " <b>Coordinates:</b> Circular (<i>r</i>, <i>phi</i>) \n");
#endif CIRCULAR
#endif SYM2D
#ifdef SYM3D
#ifdef CARTESIAN
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cartesian (<i>x</i>, <i>y</i>, <i>z</i>) \n");
#endif CARTESIAN
#ifdef SPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Spherical (<i>r</i>, <i>phi</i>, <i>theta</i>) \n");
#endif SPHERICAL
#ifdef POLARSPHERICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Polar Spherical (<i>r</i>, <i>theta</i>, <i>phi</i>) \n");
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
  zlsio_FPrintf( f, " <b>Coordinates:</b> Cylindrical (<i>z</i>, <i>r</i>, <i>phi</i>) \n");
#endif CYLINDRICAL
#endif SYM3D
  zlsio_FPrintf( f, "</code>\n");
  
  switch (io->v[ioIndex].imageSlice){
    case kSliceX:
      strcpy( textline, "X-Row");
      break;
    case kSliceY:
      strcpy( textline, "Y-Row");
      break;
    case kSliceZ:
      strcpy( textline, "Z-Row");
      break;
    case kSliceXY:
      strcpy( textline, "XY-Slice");
      break;
    case kSliceXZ:
      strcpy( textline, "XZ-Slice");
      break;
    case kSliceYZ:
      strcpy( textline, "YZ-Slice");
      break;
    case kSliceYX:
      strcpy( textline, "YX-Slice");
      break;
    case kSliceZX:
      strcpy( textline, "ZX-Slice");
      break;
    case kSliceZY:
      strcpy( textline, "ZY-Slice");
      break;
    case kSliceXYZ:
      strcpy( textline, "XYZ-Volume");
      break;
    case kSliceXSum:
      strcpy( textline, "X-Row Summed");
      break;
    case kSliceYSum:
      strcpy( textline, "Y-Row Summed");
      break;
    case kSliceZSum:
      strcpy( textline, "Z-Row Summed");
      break;
    case kSliceXYSum:
      strcpy( textline, "XY-Slices Summed");
      break;
    case kSliceXZSum:
      strcpy( textline, "XZ-Slices Summed");
      break;
    case kSliceYZSum:
      strcpy( textline, "YZ-Slices Summed");
      break;
    case kSliceYXSum:
      strcpy( textline, "YX-Slices Summed");
      break;
    case kSliceZXSum:
      strcpy( textline, "ZX-Slices Summed");
      break;
    case kSliceZYSum:
      strcpy( textline, "ZY-Slices Summed");
      break;
    default:
      strcpy( textline, "Unknown!");
  }
  switch (io->v[ioIndex].transformType){
    case kZLS_IOLinearTransform:
      strcpy( textline2, "Linear Scaling");
      break;
    case kZLS_IOLog10Transform:
      strcpy( textline2,  "Log10 Scaling");
      break;
    case kZLS_IOExp10Transform:
      strcpy( textline2,  "Exp10 Scaling");
      break;
    case kZLS_IOPowTransform:
      strcpy( textline2,  "Powerlaw Scaling");
      break;
    case kZLS_IOLog10AbsTransform:
      strcpy( textline2,  "Log10|x| Scaling");
      break;
    default:
      strcpy( textline2, "Unknown!");
  }
  zlsio_FPrintf( f, "<h4>View: Level %2.2d, %s, %s, %s </h4>\n", io->monitorLevel, io->v[ioIndex].varCode, textline, textline2);
  
  if (( nxOut < 100) && ( nyOut < 100)){
    zlsio_FPrintf( f, "<img src=\"%s%4.4d%s\" width=\"%d\" height=\"%d\" alt=\"Current Frame %4.4d\"><br>\n", 
                  io->outMonitorName, io->ioState.ioFrameCount, MONEXT, nxOut<<2, nyOut<<2, io->ioState.ioFrameCount);
  } else if (( nxOut < 200) && ( nyOut < 200)){
    zlsio_FPrintf( f, "<img src=\"%s%4.4d%s\" width=\"%d\" height=\"%d\" alt=\"Current Frame %4.4d\"><br>\n", 
                  io->outMonitorName, io->ioState.ioFrameCount, MONEXT, nxOut<<1, nyOut<<1, io->ioState.ioFrameCount);
  } else if (( nxOut > 1000) || ( nyOut > 1000)){
    zlsio_FPrintf( f, "<img src=\"%s%4.4d%s\" width=\"%d\" height=\"%d\" alt=\"Current Frame %4.4d\"><br>\n", 
                  io->outMonitorName, io->ioState.ioFrameCount, MONEXT, nxOut>>1, nyOut>>1, io->ioState.ioFrameCount);
  } else {
    zlsio_FPrintf( f, "<img src=\"%s%4.4d%s\" width=\"%d\" height=\"%d\" alt=\"Current Frame %4.4d\"><br>\n", 
                  io->outMonitorName, io->ioState.ioFrameCount, MONEXT, nxOut, nyOut, io->ioState.ioFrameCount);
  }
  
  zlsio_FPrintf( f, "<b>Min: %#10.5g \n",io->v[ioIndex].dataMin);
  
  switch (io->v[ioIndex].imageSlice){
    case kSliceX:
    case kSliceY:
    case kSliceZ:
    case kSliceXSum:
    case kSliceYSum:
    case kSliceZSum:
      zlsio_FPrintf( f, " to ");
      break;
    default:
      if (io->v[ioIndex].varTIFFCTab != NULL ){ // if using a ctab scale:
        // total hack to make ctab scale jpg file name from TIFF ctab file name
        Counter len = strlen( io->monitorCTABFile )-5;
        strcpy( srcfile, io->monitorCTABFile);
        srcfile[len] = '.';
        srcfile[len+1] = 'j';
        srcfile[len+2] = 'p';
        srcfile[len+3] = 'g';
        srcfile[len+4] = '\00';
        zlsio_FPrintf( f, "<img src=\"%s\" width=\"%d\" height=\"%d\" alt=\"%s\">\n",  srcfile,  256, 32, srcfile );
      }
      break;
  }
  
  zlsio_FPrintf( f, "Max: %#10.5g </b><br>\n",io->v[ioIndex].dataMax);
  zlsio_FPrintf( f, "<b>Frame: %d :  %4d x %4d pixels</b><br><br>\n",io->ioState.ioFrameCount, nxOut, nyOut);
  
  zlst_EndCPUTimes( &gZLS_GlobalCPUTimer );
  
  zlst_HTMLElapsedRateSummary( f, &gZLS_GlobalCPUTimer, nCells, io->monitorLevel, (nEqivCells/nCells) );
  
  zlsio_FPrintf( f, "<br>\n" );
  zlsio_FPrintf( f, "<code>\n");
  
#ifdef SYM1D
  zlsio_FPrintf( f, "<b>Grid:</b> %1dD (%3d) %#10.5g cells/cycle : %2d zones <br>\n",
                gLevels[0].levelDims,
                gLevels[0].levelGrid.n[kXDim],
                nCells, gLevels[level].nZones);
  
  if (nEqivCells>nCells)  {
    zlsio_FPrintf( f, "<b>Equivalent Grid:</b> %1dD (%3d) %#10.5g cells/cycle : Speedup: %#10.5g <br>\n",
                  gLevels[0].levelDims,  
                  gLevels[0].levelGrid.n[kXDim]*nR,
                  nEqivCells, nEqivCells/nCells);
  }
#endif SYM1D
#ifdef SYM2D
  zlsio_FPrintf( f, "<b>Grid:</b> %1dD (%3dx%3d) %#10.5g cells/cycle : %2d zones <br>\n",
                gLevels[0].levelDims,  
                gLevels[0].levelGrid.n[kXDim], gLevels[0].levelGrid.n[kYDim],
                nCells, gLevels[level].nZones);
  
  if (nEqivCells>nCells)  {
    zlsio_FPrintf( f, "<b>Equivalent Grid:</b> %1dD (%3dx%3d) %#10.5g cells/cycle : Speedup: %#10.5g <br>\n",
                  gLevels[0].levelDims,   
                  gLevels[0].levelGrid.n[kXDim]*nR, gLevels[0].levelGrid.n[kYDim]*nR,
                  nEqivCells, nEqivCells/nCells);
  }
#endif SYM2D
#ifdef SYM3D
  zlsio_FPrintf( f, "<b>Grid:</b> %1dD (%3dx%3dx%3d) %#10.5g cells/cycle : %2d zones <br>\n",
                gLevels[0].levelDims, 
                gLevels[0].levelGrid.n[kXDim], gLevels[0].levelGrid.n[kYDim], gLevels[0].levelGrid.n[kZDim],
                nCells, gLevels[level].nZones);
  
  if (nEqivCells>nCells)  {
    zlsio_FPrintf( f, "<b>Equivalent Grid:</b> %1dD (%3dx%3dx%3d) %#10.5g cells/cycle : Speedup: %#10.5g <br>\n",
                  gLevels[0].levelDims,  
                  gLevels[0].levelGrid.n[kXDim]*nR, gLevels[0].levelGrid.n[kYDim]*nR, gLevels[0].levelGrid.n[kZDim]*nR,
                  nEqivCells,  nEqivCells/nCells);
  }
  
#endif SYM3D
  
  
  for ( level = 0; level < gNLevels; level ++){
    
#ifdef SYM1D
    zlsio_FPrintf( f, "<br> <b>Level %2.2d:</b> %1dD %#10.5g cells : %2d zones <br> \n", level,
                  gLevels[level].levelDims, gLevels[level].levelNCells,  gLevels[level].nZones);
#ifdef CARTESIAN
    zlsio_FPrintf( f, " [<i>x = %#10.5g :%#10.5g dx %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_x0, gLevels[level].levelGrid.xr[kXDim]*f_x0, gLevels[level].levelGrid.dx[kXDim]*f_x0, s_Length );
#endif CARTESIAN
#ifdef SPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
#endif SPHERICAL
#ifdef POLARSPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    zlsio_FPrintf( f, " [<i>z = %#10.5g :%#10.5g dz %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_z0, gLevels[level].levelGrid.xr[kXDim]*f_z0, gLevels[level].levelGrid.dx[kXDim]*f_z0, s_Length );
#endif CYLINDRICAL
#endif SYM1D
#ifdef SYM2D
    zlsio_FPrintf( f, "<br> <b>Level %2.2d:</b> %1dD : %#10.5g cells (%3dx%3d) : %2d zones <br> \n",  level,
                  gLevels[level].levelDims,  gLevels[level].levelNCells, gLevels[level].levelGrid.n[kXDim], gLevels[level].levelGrid.n[kYDim],  gLevels[level].nZones);
#ifdef CARTESIAN
    zlsio_FPrintf( f, " [<i>x = %#10.5g :%#10.5g dx %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_x0, gLevels[level].levelGrid.xr[kXDim]*f_x0, gLevels[level].levelGrid.dx[kXDim]*f_x0, s_Length );
    zlsio_FPrintf( f, " [<i>y = %#10.5g :%#10.5g dy %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_x0, gLevels[level].levelGrid.xr[kYDim]*f_x0, gLevels[level].levelGrid.dx[kYDim]*f_x0, s_Length );
#endif CARTESIAN
#ifdef SPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_phi0, gLevels[level].levelGrid.xr[kYDim]*f_phi0, gLevels[level].levelGrid.dx[kYDim]*f_phi0 );
#endif SPHERICAL
#ifdef POLARSPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_phi0, gLevels[level].levelGrid.xr[kYDim]*f_phi0, gLevels[level].levelGrid.dx[kYDim]*f_phi0 );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    zlsio_FPrintf( f, " [<i>z = %#10.5g :%#10.5g dz %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_z0, gLevels[level].levelGrid.xr[kXDim]*f_z0, gLevels[level].levelGrid.dx[kXDim]*f_z0, s_Length );
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_r0, gLevels[level].levelGrid.xr[kYDim]*f_r0, gLevels[level].levelGrid.dx[kYDim]*f_r0, s_Length );
#endif CYLINDRICAL
#ifdef CIRCULAR
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_phi0, gLevels[level].levelGrid.xr[kYDim]*f_phi0, gLevels[level].levelGrid.dx[kYDim]*f_phi0 );
#endif CIRCULAR
#endif SYM2D
#ifdef SYM3D
    zlsio_FPrintf( f, "<br> <b>Level %2.2d:</b> %1dD : %#10.5g cells (%3dx%3dx%3d) : %2d zones <br>\n", level,
                  gLevels[level].levelDims,  gLevels[level].levelNCells, gLevels[level].levelGrid.n[kXDim], gLevels[level].levelGrid.n[kYDim], gLevels[level].levelGrid.n[kZDim],  gLevels[level].nZones);
#ifdef CARTESIAN
    zlsio_FPrintf( f, " [<i>x = %#10.5g :%#10.5g dx %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_x0, gLevels[level].levelGrid.xr[kXDim]*f_x0, gLevels[level].levelGrid.dx[kXDim]*f_x0, s_Length );
    zlsio_FPrintf( f, " [<i>y = %#10.5g :%#10.5g dy %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_x0, gLevels[level].levelGrid.xr[kYDim]*f_x0, gLevels[level].levelGrid.dx[kYDim]*f_x0, s_Length );
    zlsio_FPrintf( f, " [<i>z = %#10.5g :%#10.5g dz %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kZDim]*f_x0, gLevels[level].levelGrid.xr[kZDim]*f_x0, gLevels[level].levelGrid.dx[kZDim]*f_x0, s_Length);
#endif CARTESIAN
#ifdef SPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_phi0, gLevels[level].levelGrid.xr[kYDim]*f_phi0, gLevels[level].levelGrid.dx[kYDim]*f_phi0 );
    zlsio_FPrintf( f, " [<i>theta = %#10.5g :%#10.5g dtheta %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kZDim]*f_theta0, gLevels[level].levelGrid.xr[kZDim]*f_theta0, gLevels[level].levelGrid.dx[kZDim]*f_theta0 );
#endif SPHERICAL
#ifdef POLARSPHERICAL
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_r0, gLevels[level].levelGrid.xr[kXDim]*f_r0, gLevels[level].levelGrid.dx[kXDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>theta = %#10.5g :%#10.5g dtheta %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_theta0, gLevels[level].levelGrid.xr[kYDim]*f_theta0, gLevels[level].levelGrid.dx[kYDim]*f_theta0 );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kZDim]*f_phi0, gLevels[level].levelGrid.xr[kZDim]*f_phi0, gLevels[level].levelGrid.dx[kZDim]*f_phi0 );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    zlsio_FPrintf( f, " [<i>z = %#10.5g :%#10.5g dz %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kXDim]*f_z0, gLevels[level].levelGrid.xr[kXDim]*f_z0, gLevels[level].levelGrid.dx[kXDim]*f_z0, s_Length );
    zlsio_FPrintf( f, " [<i>r = %#10.5g :%#10.5g dr %#10.5g</i> %s]<br> \n",gLevels[level].levelGrid.xl[kYDim]*f_r0, gLevels[level].levelGrid.xr[kYDim]*f_r0, gLevels[level].levelGrid.dx[kYDim]*f_r0, s_Length );
    zlsio_FPrintf( f, " [<i>phi = %#10.5g :%#10.5g dphi %#10.5g</i> &deg;]<br> \n",gLevels[level].levelGrid.xl[kZDim]*f_phi0, gLevels[level].levelGrid.xr[kZDim]*f_phi0, gLevels[level].levelGrid.dx[kZDim]*f_phi0 );
#endif CYLINDRICAL
#endif SYM3D
    
  } // level
  zlsio_FPrintf( f, "</code>\n");
  
  
  
  sprintf( srcfile,"%s%s%s%s",DATADIR, PATHSEP, HTMLTAIL, HTMLEXT);
  hf =  zlsio_openNC ( srcfile, "r");
  if (( f != NULL ) && ( hf != NULL )){
    
    while (!feof(hf)){
      fgets(textline, PATHLEN, hf);
      fputs( textline, f);
    }
    
    zlsio_closeNC( hf );
  }
  
  zlsio_FPrintf( f, "\n");
  
  zlsio_closeNC( f );
  
  
  return( err );
  
}

