#include "zls.h"

int zlsio_FreeIOImages ( IOSettingsPtr io ){
  
  int err = noErr;
  
  if (io->image8 !=NULL)  free(io->image8); io->image8  = NULL; 
  if (io->image16!=NULL) free(io->image16); io->image16 = NULL;
  if (io->imageReal!=NULL) free(io->imageReal); io->imageReal = NULL;
  
  return (err);
}


int zlsio_InitIOImageArrays ( IOSettingsPtr io ){ // private local init routine, no checking. not in .h
  int err = noErr;
  
  io->image8  = NULL;
  io->image16 = NULL;
  io->imageReal = NULL;  
  
  return (err);
}

int zlsio_ScaleAndWritePlot( char * fname, IOSettingsPtr io, Integer ioIndex, Integer nxOut, Integer nyOut, Integer dim, Integer compressOption){
  
  int err = noErr;
  
  if (io->outBitsPerPix == k8bitOutput) {
    
    err = zls_Plot1D8BitData   ( (unsigned char *)io->image8 , io->imageReal, kZLS_IOImage, nxOut,  io,  ioIndex );
    
    if ( err == noErr ) {
      
      if ( compressOption == kUnCompressedOutput ){
        
        if (io->outFormat == kRawOutput){
          zls_Write_raw8_NC(fname, io->image8, nxOut, nyOut);
        }
        
        if (io->outFormat == kTIFFOutput){
          zls_Write_tif8_NC(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outFormat == kJPEGOutput){
          zls_Write_jpeg8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outFormat == kPNGOutput){
          zls_Write_png8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
      } else { // allow compression if ZLIB_COMPRESSED_IO is enabled
        
        if (io->outFormat == kRawOutput){
          zls_Write_raw8(fname, io->image8, nxOut, nyOut);
        }
        
        if (io->outFormat == kTIFFOutput){
          zls_Write_tif8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outFormat == kJPEGOutput){
          zls_Write_jpeg8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outFormat == kPNGOutput){
          zls_Write_png8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
      } // err
      
    } // 8 bit
    
    return( err );
    
  }
  return(err);
}

int zlsio_ScaleAndWriteImage( char * fname, IOSettingsPtr io, Integer ioIndex, Integer nxOut, Integer nyOut, Integer dim, Integer compressOption){
  
  int err = noErr;
  //
  // 2D unsigned scaling calls by setting nk = 1..
  // these routines also reverse the j axis for images when nk = 1;
  // if the symmetry is 1D, then a 256 pixel high plot is drawn instead
  //
  
  if ((io->outBitsPerPix == k8bitOutput)  || (io->outFormat == kJPEGOutput)){
    
    if ( dim == 1) {
      err = zls_Plot1D8BitData   ( (unsigned char *)io->image8 , io->imageReal, kZLS_IOImage, nxOut,  io,  ioIndex );
    } else {
      err = zls_Scale8BitData   ( (unsigned char *)io->image8 , io->imageReal, kZLS_IOImage,
                                 nxOut, nyOut, 1,  io,  ioIndex );
    }
    
  }
  
  if (io->outBitsPerPix == k16bitOutput) {
    
    if ( dim == 1) {
      err = zls_ScaleUnsigned16BitData   ( io->image16 , io->imageReal, kZLS_IOImage,
                                          nxOut, nyOut, 1, 
                                          io,  ioIndex );
    } else {
      err = zls_ScaleUnsigned16BitData   ( io->image16 , io->imageReal, kZLS_IOImage,
                                          nxOut, nyOut, 1, 
                                          io,  ioIndex );
    }
    
  }
  
  if ( err == noErr ) {
    
    if ( compressOption == kUnCompressedOutput ){
      
      if (io->outFormat == kRawOutput){
        
        if (io->outBitsPerPix == k8bitOutput)
          zls_Write_raw8_NC(fname, io->image8, nxOut, nyOut);
        
        if (io->outBitsPerPix == k16bitOutput) 
          zls_Write_raw16_NC(fname, io->image16, nxOut, nyOut);
        
      }
      
      if (io->outFormat == kTIFFOutput){
        
        if (io->outBitsPerPix == k8bitOutput) {
          zls_Write_tif8_NC(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outBitsPerPix == k16bitOutput) 
          zls_Write_tif16_NC(fname,io->image16, nxOut, nyOut);
        
      }
      
      if (io->outFormat == kJPEGOutput){
        //  JPEG output is always 8 bits per channel RGB
        zls_Write_jpeg8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
      }
      
      if (io->outFormat == kPNGOutput){
        //  PNG output is always 8 bits per channel palette
        zls_Write_png8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
      }
      
    } else { // allow compression if ZLIB_COMPRESSED_IO is enabled
      
      if (io->outFormat == kRawOutput){
        
        if (io->outBitsPerPix == k8bitOutput) 
          zls_Write_raw8(fname, io->image8, nxOut, nyOut);
        
        if (io->outBitsPerPix == k16bitOutput) 
          zls_Write_raw16(fname, io->image16, nxOut, nyOut);
        
      }
      
      if (io->outFormat == kTIFFOutput){
        
        if (io->outBitsPerPix == k8bitOutput) {
          zls_Write_tif8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        }
        
        if (io->outBitsPerPix == k16bitOutput) 
          zls_Write_tif16(fname,io->image16, nxOut, nyOut);
        
      }
      
      if (io->outFormat == kJPEGOutput){
        //  JPEG output is always 8 bits per channel RGB
        zls_Write_jpeg8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        
      }
      
      if (io->outFormat == kPNGOutput){
        //  PNG output is always 8 bits per channel palette
        zls_Write_png8(fname,io->image8, io->v[ioIndex].varTIFFCTab, nxOut, nyOut);
        
      }
      
      
      
    }
    
  }
  
  return( err );
  
}



int zlsio_SaveZoneImage ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
                         IOSettingsPtr  io, Integer ioIndex ) {
  //
  // Best if this routine is called zlsio_SaveLevelImages if an image zone is used
  // as that will setup the image zone properly first at the level 0 loop
  // otherwise is it fairly safe to call anytime for normal zones.
  //  
  int err = noErr;
  
  if ( ( z != NULL) && ( io != NULL ) ){
    
    char     outFile[PATHLEN];
    char     vPref[PREFIXLEN];
    
    // get dimensions/info from zone
    
    GridCoords gc      = z->zoneGrid;
    Integer sliceOut   = kSliceXY;
    
    sliceOut   = io->v[ioIndex].imageSlice;
#if (NDIMS > 1)  
    Integer sliceIndex = io->v[ioIndex].imageIndex;
    Integer sliceIndex2 = io->v[ioIndex].imageIndex2;
#endif (NDIMS > 1)
    
    int bIsPLot  = ((io->v[ioIndex].ioTypes & kZLS_IOPlot)  == kZLS_IOPlot);
    
    Counter nx; // these dimensions are not just n[kXDim], n[kYDim]
    Counter ny; // they are computed according to the slice/dimensions settings
    
    // *all* zones return a 2D nx by ny result, even 1D zones and 1D slices of 2D zones :)
    
    err = zls_GetGridImageDims( &nx, &ny, &gc, sliceOut );
    int dim = 2;
    if ( bIsPLot ) {
      dim =   1;
      ny  = 256; // 1D images are expanded to 256 high for plots instead
    }
    
    err = zlsio_GetVarIOVarPrefix ( vPref, io, ioIndex );
    if ( err != noErr ){
      zlsio_ErrPrintf( " IO ERROR: ioIndex out of range in zlsio_SaveZoneImage, %d Err = %d \n", ioIndex, err );
      err = ioErr;
      return (err);
    }
    
    zlsio_AllocateIOImages( io, nx, ny );
    
    if (z->zoneType == kImageZoneType) {
      
      //
      // image zone merged, simplifies file name and get data calls,
      // name doesn't contain zone id, just the level and cycle number
      // image zone already contains the image data, as that is set up in
      // the level 0 loop in zlsio_SaveLevelImages.
      // if not then the images will be odd :)
      //
      
      varIndex = kAnonVar; // image zone data is always in zoneData[0];
      
      if (io->outFormat == kRawOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref, io->outName, level, (int)io->ioState.ioFrameCount, RAWEXT );
      } else if (io->outFormat == kTIFFOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref,io->outName, level, (int)io->ioState.ioFrameCount, TIFFEXT );
      } else if (io->outFormat == kJPEGOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref,io->outName, level, (int)io->ioState.ioFrameCount, JPGEXT );
      } else if (io->outFormat == kPNGOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref,io->outName, level, (int)io->ioState.ioFrameCount, PNGEXT );
      } else { // data file output
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref,io->outName, level, (int)io->ioState.ioFrameCount, DATAFILEEXT );
      }
      
      // for the merged level image zone the calculation of physical quantities has already been
      // done in the level 0 loop.  In that case just copy or slice/sum as needed
      
#ifdef SYM1D          
      err      = zlsp_GetZoneWQuantityCopy ( &io->imageReal, z, varIndex);
#endif SYM1D
#ifdef SYM2D      
      if ( bIsPLot ) {
        err    = zls_GetZoneData1DRow      ( &io->imageReal, z, varIndex, sliceOut, sliceIndex, sliceIndex2);
      } else {
        err    = zlsp_GetZoneWQuantityCopy ( &io->imageReal, z, varIndex);
      }
#endif SYM2D
#ifdef SYM3D          
      if ( bIsPLot ) {
        err    = zls_GetZoneData1DRow      ( &io->imageReal, z, varIndex, sliceOut, sliceIndex, sliceIndex2);
      } else {
        err    = zlsp_GetZoneWQuantityCopy2DSlice  ( &io->imageReal, z, varIndex, sliceOut, sliceIndex );
      }
#endif SYM3D
      
    } else {
      
      if (io->outFormat == kRawOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
                io->v[ioIndex].varCode, PATHSEP,
                vPref, io->outName, level, zone, (int)io->ioState.ioFrameCount, RAWEXT );
      } else if (io->outFormat == kTIFFOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
                io->v[ioIndex].varCode, PATHSEP, 
                vPref,io->outName, level, zone, (int)io->ioState.ioFrameCount, TIFFEXT );
      } else if (io->outFormat == kJPEGOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
                io->v[ioIndex].varCode, PATHSEP, 
                vPref,io->outName, level, zone, (int)io->ioState.ioFrameCount, JPGEXT );
      } else if (io->outFormat == kPNGOutput){
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
                io->v[ioIndex].varCode, PATHSEP, 
                vPref,io->outName, level, zone, (int)io->ioState.ioFrameCount, PNGEXT );
      } else { // data file output
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
                io->v[ioIndex].varCode, PATHSEP, 
                vPref,io->outName, level, zone, (int)io->ioState.ioFrameCount, DATAFILEEXT );
      }
      
#ifdef SYM1D          
      err      = zlsp_GetZoneWQuantityCopy ( &io->imageReal, z, varIndex);
#endif SYM1D
#ifdef SYM2D          
      if ( bIsPLot ) {
        err    = zls_GetZoneData1DRow      ( &io->imageReal, z, varIndex, sliceOut, sliceIndex, sliceIndex2);
      } else {
        err    = zlsp_GetZoneWQuantityCopy ( &io->imageReal, z, varIndex);
      }
#endif SYM2D
#ifdef SYM3D          
      if ( bIsPLot ) {
        err    = zls_GetZoneData1DRow      ( &io->imageReal, z, varIndex, sliceOut, sliceIndex, sliceIndex2);
      } else {
        err    = zlsp_GetZoneWQuantityCopy2DSlice  ( &io->imageReal, z, varIndex, sliceOut, sliceIndex );
      }
#endif SYM3D
      
    }
    
    // if a velocity slice is needed, slice after the variable is made
    
    switch (varIndex) {
      case kWDeltaVx:
     {
       Real vMax =  io->v[ioIndex].scaleMax;
       Real vMin =  io->v[ioIndex].scaleMin;
       zlsp_VelocitySlice( io->imageReal, io->imageReal, vMin, vMax, gc);
     }
        break;
#if (NDIMS >1)        
      case kWDeltaVy:
     {
       Real vMax =  io->v[ioIndex].scaleMax;
       Real vMin =  io->v[ioIndex].scaleMin;
       zlsp_VelocitySlice( io->imageReal, io->imageReal, vMin, vMax, gc);
     }
        break;
#endif (NDIMS >1)        
#if (NDIMS >2)        
      case kWDeltaVz:
     {
       Real vMax =  io->v[ioIndex].scaleMax;
       Real vMin =  io->v[ioIndex].scaleMin;
       zlsp_VelocitySlice( io->imageReal, io->imageReal, vMin, vMax, gc);
     }
        break;
#endif  (NDIMS >2)  
#if (NDIMS >1)        
      case kSymDens: // just like density only self-differences symmetrically about i = j
        zlsp_SymmetricDifference( io->imageReal, gc);
#endif (NDIMS >1)        
      default:
        ;
    }
    
    // Finally write the image...
    // zlsio_LogPrintf  ( " IO: Writing Image:\n       %s, %d x %d pixels.\n", outFile, nx, ny);
    zlsio_FilesPrintf( " IO: Writing Image:\n       %s, %d x %d pixels.\n", outFile, nx, ny);
    
#ifdef  ZLIB_UNCOMPRESSED_IMAGES     
    if (bIsPLot) {
      err    = zlsio_ScaleAndWritePlot     ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput );
    } else {
      err    = zlsio_ScaleAndWriteImage    ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput );
    }
#else 
    if (bIsPLot) {
      err    = zlsio_ScaleAndWritePlot     ( outFile, io, ioIndex, nx, ny, dim, kAllowCompressedOutput );
    } else {
      err    = zlsio_ScaleAndWriteImage    ( outFile, io, ioIndex, nx, ny, dim, kAllowCompressedOutput );
    }
#endif
    
    
#ifdef MPI  
    if ( gMPIrank == io->monitorLevel ){ // only rank io->monitorLevel does monitoring
#endif MPI
      
      if (( io->monitorVar == ioIndex ) &&( io->monitorLevel == level) && ( zone == 0 )) {
        
        // create JPEG/TIFF monitor file
        
        Integer saveIOFormat = io->outFormat;
        io->outFormat = kMonitorFormatOutput;
        
        // new name with framenumber so html can ref new image and brain dead browsers will actually show
        // a new image...
        sprintf(outFile, "%s%s%s%4.4d%s", io->monDir, PATHSEP, io->outMonitorName, io->ioState.ioFrameCount, MONEXT );
        if (bIsPLot) {
          err      = zlsio_ScaleAndWritePlot    ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput);
        } else {
          err      = zlsio_ScaleAndWriteImage    ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput);
        }
        
        // fixed name for last output frame, for image getting scripts..
        sprintf(outFile, "%s%s%s%s", io->monDir, PATHSEP, io->outMonitorName, MONEXT );
        if (bIsPLot) {
          err      = zlsio_ScaleAndWritePlot    ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput);
        } else {
          err      = zlsio_ScaleAndWriteImage    ( outFile, io, ioIndex, nx, ny, dim, kUnCompressedOutput);
        }
        io->outFormat=saveIOFormat;
        
        // named html file for this frame
        sprintf(outFile, "%s%s%s%4.4d%s", io->monDir, PATHSEP, io->outMonitorName, io->ioState.ioFrameCount, HTMLEXT );
        err      = zlsio_WriteHTMLImageDescription ( outFile, io,  io->ioState.ioFrameCount, ioIndex, nx, ny );
        
        // And generic html file for current frame with simulation name
        sprintf(outFile, "%s%s%s%s", io->monDir, PATHSEP, io->outMonitorName, HTMLEXT );
        err      = zlsio_WriteHTMLImageDescription ( outFile, io, -1, ioIndex, nx, ny );
        
        // And generic html file for current frame with constant name
        sprintf(outFile, "%s%s%s%s", io->monDir, PATHSEP, ZLSWEBFILE, HTMLEXT );
        err      = zlsio_WriteHTMLImageDescription ( outFile, io, -1, ioIndex, nx, ny );
        
        
      }
      
#ifdef MPI  
    }
#endif MPI
    
    zlsio_FreeIOImages ( io );
    
    
  } else {
    zlsio_ErrPrintf( " IO ERROR: NULL zone zlsio_SaveZoneImage.\n" );
    err = ioErr;
  }
  
  return (err);
  
}


int zlsio_SaveLevelImages(Integer level, int step){ 
  
  // images are all 2D, even if just barely (ie a single row for 1D input)
  // output for level 0 is merged if possible, defaults to separate zone IO
  // in the event that imageZone == NULL and/or io->outMergedZones != kZLS_IOOutputMergedZones 
  // if nZones = 1 then imageZone is skipped too.
#pragma unused( step)
  
  int       err = noErr;
  
  Integer          zone;
  Integer      varIndex;
  Integer       ioIndex;
  ZonePtr             z;
  
  IOSettingsPtr  io = NULL;
  
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
    
    ZonePtr imageZone = NULL;
    Integer bDoMergedImage = 0;
    
#ifdef MPI
    if ( io->outMergedZoneFrames == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        if ( gMPIrank == level) { // collect merged output on rank == level only
          zls_GetLevelImageZone( &imageZone, level ); 
        }
        bDoMergedImage = 1; // but all ranks participate in merger
      } // other levels do separate zones and imageZone stays NULL;
    }
#else    
    if ( io->outMergedZoneFrames == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        zls_GetLevelImageZone( &imageZone, level );
        if ( imageZone != NULL) bDoMergedImage = 1;
      } // other levels do separate zones and imageZone stays NULL;
    }
#endif !MPI
    
    for ( ioIndex = 0; ioIndex< io->outNVars; ioIndex++){
      
      int bIsImage = ((io->v[ioIndex].ioTypes & kZLS_IOImage) == kZLS_IOImage);
      int bIsPLot  = ((io->v[ioIndex].ioTypes & kZLS_IOPlot)  == kZLS_IOPlot);
      
      if ( bIsImage || bIsPLot ){ // image or plot io type flag
        
        varIndex = kAnonVar;
        
        err = zlsio_GetIOCodeVarIndex (&varIndex, io->v[ioIndex].varCode);
        if ( err != noErr ) {
          zlsio_ErrPrintf( " IO ERROR: varCode unknown in zlsio_SaveLevelImages, %d %s Err = %d \n"
                          , ioIndex, io->v[ioIndex].varCode, err );
          err = ioErr;
          return (err);
        }
        
        if ( varIndex != kNoVar ) { 
          
          if ( bDoMergedImage == 1) {
#ifdef MPI
            err =  zls_MPI_GetLevelMergedZoneData( imageZone, level, varIndex );
#else 
            err =  zls_GetLevelMergedZoneData( imageZone, level, varIndex );
#endif !MPI
          } // imageZone may be null on non rank 0 processes, they send buffers to 0 which collects all
          
          if ((err == noErr) &&( bDoMergedImage == 1)) { 
            
#ifdef MPI  
            if ( gMPIrank == level) { // only rank == level can to write out merged image
#endif MPI
              err = zlsio_SaveZoneImage ( imageZone, kAnonVar, level, 0, io, ioIndex );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO WARNING: zlsio_SaveZoneImage Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
              }
#ifdef MPI  
            }
#endif MPI
            
          } else { 
            
            // do normal separate output, either imageZone failed or merger not requested
            //
            // loop over zones separately
            //
            
            for( zone = 0; zone <gLevels[level].nZones; zone++){
              
              err = zls_GetLevelZoneByIndex( &z, level, zone );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
                err = ioErr;
                return (err);
              }
              
#ifdef MPI  
              if (z->zoneRankID == gMPIrank) {
#endif MPI
                err = zlsio_SaveZoneImage (  z, varIndex, level, zone, io, ioIndex );
                if ( err != noErr ){
                  zlsio_ErrPrintf( " IO WARNING: zlsio_SaveZoneImage Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                  zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
                }
#ifdef MPI  
              }
#endif MPI
              
            } // zone loop
            
          } // else non-merged
          
        } // var index != kNoVar
        
      } // image type io
      
    } // ioIndex
    
    
    if ( level == 0 ) io->ioState.ioFrameCount++;
    
    
  } else { //io
    zlsio_ErrPrintf( " IO ERROR: IOSettings not found zlsio_SaveLevelImages. Err %d\n", err );
    err = ioErr;
  }
  
  return (err);
}


int zlsio_SaveSimpleLevelImages(Integer level, int step){ // images are all 2D, even if just barely (ie a single row for 1D input)
#pragma unused( step)
  
  // ignores var ioType flag = outputs an image no matter what
  
  int        err=noErr;
  
  Integer          zone;
  Integer      varIndex;
  Integer       ioIndex;
  ZonePtr             z;
  
  IOSettingsPtr  io = NULL;
  
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
    
    for ( ioIndex = 0; ioIndex< io->outNVars; ioIndex++){
      
      varIndex = kAnonVar;
      
      err = zlsio_GetIOCodeVarIndex (&varIndex, io->v[ioIndex].varCode);
      if ( err != noErr ) {
        zlsio_ErrPrintf( " IO ERROR: varCode unknown in zlsio_SaveSimpleLevelImages, %d %s Err = %d \n"
                        , ioIndex, io->v[ioIndex].varCode, err );
        err = ioErr;
        return (err);
      }
      
      if ( varIndex!= kNoVar ) { 
        
        //
        // loop over zones separately
        //
        
        for( zone = 0; zone <gLevels[level].nZones; zone++){
          
          err = zls_GetLevelZoneByIndex( &z, level, zone );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
            err = ioErr;
            return (err);
          }
          
#ifdef MPI  
          if (z->zoneRankID == gMPIrank) {
#endif MPI
            
            err = zlsio_SaveZoneImage (  z, varIndex, level, zone,  io, ioIndex );
            if ( err != noErr ){
              zlsio_ErrPrintf( " IO WARNING: zlsio_SaveSimpleLevelImages Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
              zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
            }
#ifdef MPI  
          }
#endif MPI
          
        } // zone loop
        
      }  // kNoVar
      
    } // ioIndex
    
    if ( level == 0 ) io->ioState.ioFrameCount++;
    
  } else { //io
    zlsio_ErrPrintf( " IO ERROR: IOSettings not found zlsio_SaveSimpleLevelImages. Err %d\n", err );
    err = ioErr;
  }
  
  return (err);
}

