#include "zls.h"
#include <jpeglib.h>
//========================================
//
// JPEG File Writing IO Routines
//
//========================================

#define ZLS_JPEG_Q  95

int zls_Write_jpeg8     (char * aFile, UInt8 * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut ){
  int err = noErr;
  
  int    imageSize= 0;
  UInt8  *imgRGB = NULL;
  Counter i, j, jOffset;
  
  int quality;      // -quality parameter */
  int q_scale_factor;   // scaling percentage for -qtables */
  
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  
  JSAMPROW row_pointer[1];  //pointer to JSAMPLE row[s] */
  int row_stride;   // physical row width in image buffer */
  
  
  FILE * fileAccess = NULL;
  
  // allocate rgb chunky pixel array and combine rgb planes
  
  imageSize = 3L*nxOut*nyOut;
  imgRGB = (UInt8 *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = (UInt8)((outTIFFCTab->red[img[i + jOffset]])>>8);
        imgRGB[3*(i + jOffset)+1] = (UInt8)((outTIFFCTab->green[img[i + jOffset]])>>8);
        imgRGB[3*(i + jOffset)+2] = (UInt8)((outTIFFCTab->blue[img[i + jOffset]])>>8);
        
      }//i
    }//j
    
    // Initialize the JPEG compression object with default error handling.
    
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    
    
    fileAccess = zlsio_openNC( aFile, "wb");
    if (fileAccess) {
      
      jpeg_stdio_dest(&cinfo, fileAccess);
      
      // Initialize JPEG parameters.
      
      cinfo.image_width = nxOut;  /* image width and height, in pixels */
      cinfo.image_height = nyOut;
      cinfo.input_components = 3;   /* # of color components per pixel */
      cinfo.in_color_space = JCS_RGB;   /* colorspace of input image */
      
      jpeg_set_defaults(&cinfo);
      
      quality = ZLS_JPEG_Q;   /* default -quality value */
      q_scale_factor = 100;   /* default to no scaling for -qtables */
      
      jpeg_set_quality(&cinfo, quality, TRUE);
      
      jpeg_start_compress(&cinfo, TRUE);
      
      row_stride = nxOut * 3; /* JSAMPLEs per row in image_buffer */
      
      /* Process data */
      while (cinfo.next_scanline < cinfo.image_height) {
        row_pointer[0] = & imgRGB[cinfo.next_scanline * row_stride];
        (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
      }
      
      /* Finish compression and release memory */
      
      jpeg_finish_compress(&cinfo);
      zlsio_closeNC(fileAccess);
      
      jpeg_destroy_compress(&cinfo);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err);
  
}


int zls_Write_jpegRGB     (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  
  int    imageSize= 0;
  UInt8  *imgRGB = NULL;
  Counter i, j, jOffset;
  
  int quality;      // -quality parameter */
  int q_scale_factor;   // scaling percentage for -qtables */
  
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  
  JSAMPROW row_pointer[1];  /* pointer to JSAMPLE row[s] */
  int row_stride;   /* physical row width in image buffer */
  
  
  FILE * fileAccess = NULL;
  
  // allocate rgb chunky pixel array and combine rgb planes
  
  imageSize = 3L*nxOut*nyOut;
  imgRGB = (UInt8 *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    // Initialize the JPEG compression object with default error handling.
    
    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    
    
    fileAccess = zlsio_openNC( aFile, "wb");
    if (fileAccess) {
      
      jpeg_stdio_dest(&cinfo, fileAccess);
      
      // Initialize JPEG parameters.
      
      cinfo.image_width = nxOut;  /* image width and height, in pixels */
      cinfo.image_height = nyOut;
      cinfo.input_components = 3;   /* # of color components per pixel */
      cinfo.in_color_space = JCS_RGB;   /* colorspace of input image */
      
      jpeg_set_defaults(&cinfo);
      
      quality = ZLS_JPEG_Q;     /* default -quality value */
      q_scale_factor = 100;   /* default to no scaling for -qtables */
      
      jpeg_set_quality(&cinfo, quality, TRUE);
      
      jpeg_start_compress(&cinfo, TRUE);
      
      row_stride = nxOut * 3; /* JSAMPLEs per row in image_buffer */
      
      /* Process data */
      while (cinfo.next_scanline < cinfo.image_height) {
        row_pointer[0] = & imgRGB[cinfo.next_scanline * row_stride];
        (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
      }
      
      /* Finish compression and release memory */
      
      jpeg_finish_compress(&cinfo);
      zlsio_closeNC(fileAccess);
      
      jpeg_destroy_compress(&cinfo);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err);
  
}
