#include "zls.h"
#include <png.h>
//========================================
//
// PNG File IO Routines
//
//
// Only write operations are used.
//
//========================================

#define ZLS_PNGINTERLACE  (PNG_INTERLACE_NONE)
#define ZLS_PNGCOMPRESS   (PNG_COMPRESSION_TYPE_BASE)
#define ZLS_PNGFILTER     (PNG_FILTER_TYPE_BASE)
#define ZLS_PNGTRANS      (PNG_TRANSFORM_IDENTITY)
#define ZLS_PNGGAMMA      (1.8)

// here all PNGs, like jpegs, have 8 bits per channel

//================================================
// 8 bit single channel with colour lookup table
//================================================

int zls_Write_png8     (char * aFile, UInt8 * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut ){
  
  int err = noErr;
  
  FILE * fileAccess = NULL;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text text_ptr[4];
  png_colorp palette;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    
    if (png_ptr == NULL)  {
      zlsio_closeNC(fileAccess);
      return (fileErr);
    }
    
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
      zlsio_closeNC(fileAccess);
      png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
      return (fileErr);
    }
    
    png_init_io(png_ptr, fileAccess); // associate file with png_
    
    png_set_IHDR( png_ptr, info_ptr, nxOut, nyOut, 8, 
                 PNG_COLOR_TYPE_PALETTE, ZLS_PNGINTERLACE, ZLS_PNGCOMPRESS, ZLS_PNGFILTER );
    
    palette = ( png_colorp) png_malloc(png_ptr, ( PNG_MAX_PALETTE_LENGTH * sizeof (png_color)) );
    
    int i;
    for (i = 0; i < PNG_MAX_PALETTE_LENGTH; i++){
      palette[i].red   = (UInt8)((outTIFFCTab->red[i])>>8);
      palette[i].green = (UInt8)((outTIFFCTab->green[i])>>8);
      palette[i].blue  = (UInt8)((outTIFFCTab->blue[i])>>8);
    }//i
    
    png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);
    
    text_ptr[0].key = "Title";
    text_ptr[0].text = aFile;
    text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
    text_ptr[1].key = "ZLS Version";
    text_ptr[1].text = "ZLS_VERSION";
    text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
    text_ptr[2].key = "Fyris Problem";
    text_ptr[2].text = FYRIS_PROBLEMNAME;
    text_ptr[2].compression = PNG_TEXT_COMPRESSION_NONE;
    text_ptr[3].key = "Date";
    text_ptr[3].text = zlst_TheCurrentRealTime();
    text_ptr[3].compression = PNG_TEXT_COMPRESSION_NONE;
    png_set_text(png_ptr, info_ptr, text_ptr, 4);
    
    png_write_info(png_ptr, info_ptr);
    
    png_uint_32 row_stride = nxOut; /* per row in image_buffer */
    png_uint_32 k;
    png_bytep row_pointers[nyOut];
    
    for (k = 0; k < nyOut; k++)
      row_pointers[k] = img + k*row_stride;
    
    png_write_image(png_ptr, row_pointers);
    
    png_write_end(png_ptr, info_ptr);
    
    png_destroy_write_struct(&png_ptr, &info_ptr);
    
    zlsio_closeNC(fileAccess);
    
  } else { err = fileErr;}
  
  return (err);
  
}


//================================================
// 8 bit three channel RGB files
//================================================

int zls_Write_pngRGB     (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  
  int err = noErr;
  
  UInt8  *imgRGB = NULL;
  
  FILE * fileAccess = NULL;
  png_structp png_ptr;
  png_infop info_ptr;
  png_text text_ptr[4];
  
  
  // allocate rgb chunky pixel array and combine rgb planes
  
  imgRGB = (UInt8 *) malloc(3L*nxOut*nyOut);
  
  if (imgRGB != NULL) {
    
    int i, j, jOffset;
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    
    fileAccess = zlsio_openNC( aFile, "wb");
    
    if (fileAccess) {
      
      png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
      
      if (png_ptr == NULL)   {
        free(imgRGB);
        zlsio_closeNC(fileAccess);
        return (fileErr);
      }
      
      info_ptr = png_create_info_struct(png_ptr);
      if (info_ptr == NULL)  {
        free(imgRGB);
        zlsio_closeNC(fileAccess);
        png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
        return (fileErr);
      }
      
      png_init_io(png_ptr, fileAccess); // associate file with png_
      
      png_set_IHDR( png_ptr, info_ptr, nxOut, nyOut, 8, 
                   PNG_COLOR_TYPE_RGB, ZLS_PNGINTERLACE, ZLS_PNGCOMPRESS, ZLS_PNGFILTER );
      
      text_ptr[0].key = "Title";
      text_ptr[0].text = aFile;
      text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
      text_ptr[1].key = "ZLS Version";
      text_ptr[1].text = "ZLS_VERSION";
      text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
      text_ptr[2].key = "Fyris Problem";
      text_ptr[2].text = FYRIS_PROBLEMNAME;
      text_ptr[2].compression = PNG_TEXT_COMPRESSION_NONE;
      text_ptr[3].key = "Date";
      text_ptr[3].text = zlst_TheCurrentRealTime();
      text_ptr[3].compression = PNG_TEXT_COMPRESSION_NONE;
      png_set_text(png_ptr, info_ptr, text_ptr, 4);
      
      
      png_write_info(png_ptr, info_ptr);
      
      png_uint_32 row_stride = nxOut * 3; /* per row in image_buffer */
      png_uint_32 k;
      png_bytep row_pointers[nyOut];
      
      for (k = 0; k < nyOut; k++)
        row_pointers[k] = imgRGB + k*row_stride;
      
      png_write_image(png_ptr, row_pointers);
      
      png_write_end(png_ptr, info_ptr);
      
      png_destroy_write_struct(&png_ptr, &info_ptr);
      
      zlsio_closeNC(fileAccess);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err);
  
}
