#include "zls.h"

//========================================
//
// Raw File IO Routines
//
//========================================


int zls_Write_raw8     ( char * aFile, UInt8 * img, int nxOut, int nyOut  ){
  int err = noErr;
  zFile fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut);
    err      =zlsio_write( fileAccess, img, fileSize );
    zlsio_close(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw16    ( char * aFile, UInt16 * img, int nxOut, int nyOut  ){
  int err = noErr;
  
  zFile fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_write( fileAccess, img, fileSize );
    zlsio_close(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw8RGB  (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  
  zFile fileAccess = NULL;
  int     fileSize= 0;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut);
    err      =zlsio_write( fileAccess, imgR, fileSize );
    fileSize = (nxOut*nyOut);
    err      =zlsio_write( fileAccess, imgG, fileSize );
    fileSize = (nxOut*nyOut);
    err      =zlsio_write( fileAccess, imgB, fileSize );
    zlsio_close(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw16RGB (char * aFile, UInt16 * imgR, UInt16 * imgG, UInt16 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  zFile fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_write( fileAccess, imgR, fileSize );
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_write( fileAccess, imgG, fileSize );
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_write( fileAccess, imgB, fileSize );
    zlsio_close(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw8_NC     ( char * aFile, UInt8 * img, int nxOut, int nyOut  ){
  int err = noErr;
  zFileNC fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut);
    err      =zlsio_writeNC( fileAccess, img, fileSize );
    zlsio_closeNC(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw16_NC    ( char * aFile, UInt16 * img, int nxOut, int nyOut  ){
  int err = noErr;
  
  zFileNC fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_writeNC( fileAccess, img, fileSize );
    zlsio_closeNC(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw8RGB_NC (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  
  zFileNC fileAccess = NULL;
  int     fileSize= 0;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut);
    err      =zlsio_writeNC( fileAccess, imgR, fileSize );
    fileSize = (nxOut*nyOut);
    err      =zlsio_writeNC( fileAccess, imgG, fileSize );
    fileSize = (nxOut*nyOut);
    err      =zlsio_writeNC( fileAccess, imgB, fileSize );
    zlsio_closeNC(fileAccess);
  } else { err = fileErr; }
  return (err);
}

int zls_Write_raw16RGB_NC (char * aFile, UInt16 * imgR, UInt16 * imgG, UInt16 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  zFileNC fileAccess = NULL;  int     fileSize= 0;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_writeNC( fileAccess, imgR, fileSize );
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_writeNC( fileAccess, imgG, fileSize );
    fileSize = (nxOut*nyOut*2L);
    err      = zlsio_writeNC( fileAccess, imgB, fileSize );
    zlsio_closeNC(fileAccess);
  } else { err = fileErr; }
  return (err);
}

//========================================
//
// TIFF File IO Routines
//
//========================================


int zls_Write_tif8     (char * aFile, UInt8 * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 8 bit tiff with 256 colour colour table, if outTIFFCTab == NULL then do Grayscale
  //
  zFile fileAccess = NULL;  int     imageSize= 0;
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x00000618; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000006ae; // specific to this header
  const short nTPS              = 0x000C;     // 12 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    
    imageSize = (nxOut*nyOut);
    
    zlsio_write( fileAccess, (void*)&tiffPreHeader0, 4UL );
    zlsio_write( fileAccess, (void*)&tiffPreHeader1, 4UL );
    
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    
    zlsio_write( fileAccess, outTIFFCTab, sizeof(TIFFCTab) );
    
    zlsio_write( fileAccess, (void*)&nTPS, 2 );
    
    tiffTag.tCode = 0x0100; // width code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nxOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0101; // height code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0102; // bits per pixel code
    tiffTag.dataType = 3;
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00080000; // 8 bits per pixel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000008; // 8 bits per pixel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0103; // compression code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; //none compression
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; //none compression
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0106; // Perception Intent code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00030000; // 1 => 0 = black GS,  3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000003; // 1 => 0 = black GS,  3=> colour TABLE
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0115; // samples per pix
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 channel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 channel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block!)
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0117; // StripByteCounts
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = imageSize; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0111; // strip offset from start of whole file
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = tiffImageDataOffset;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000008;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000010;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0140; // colour map/table
    tiffTag.dataType = 3;   // SHORT
    tiffTag.nDataValues = 3*256;
    tiffTag.pDataField = 0x00000018; // location of table n bytes from SOF
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    zlsio_write( fileAccess, (void*)&ifdTermintator, 4UL );
    
    // at last- write the data bytes out....
    
    err      =zlsio_write( fileAccess, img, imageSize );
    
    zlsio_close(fileAccess);
    
  } else { err = fileErr; }
  
  return (err);
  
}


int zls_Write_tif16     (char * aFile, UInt16 * img, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 16 bit tiff  grayscale - no colour table
  //
  zFile fileAccess = NULL;  int     imageSize= 0;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x00000018; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a2; // specific to this header
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  fileAccess = zlsio_open( aFile, "wb");
  if (fileAccess) {
    
    imageSize = (2L*nxOut*nyOut);
    
    zlsio_write( fileAccess, (void*)&tiffPreHeader0, 4UL );
    zlsio_write( fileAccess, (void*)&tiffPreHeader1, 4UL );
    
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
    zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
    
    zlsio_write( fileAccess, (void*)&nTPS, 2 );
    
    tiffTag.tCode = 0x0100; // width code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nxOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0101; // height code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0102; // bits per pixel code
    tiffTag.dataType = 3;
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00100000; // 16 bits per pixel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000010; // 16 bits per pixel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0103; // compression code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // none compression
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0106; // Perception Intent code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 => 0 = black GS,  3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 => 0 = black GS,  3=> colour TABLE
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0115; // samples per pix
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 channel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 channel
#endif
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block!)
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0117; // StripByteCounts
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = imageSize; // 1 channel
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0111; // strip offset from start of whole file
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = tiffImageDataOffset;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000008;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000010;
    
    zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    zlsio_write( fileAccess, (void*)&ifdTermintator, 4UL );
    
    // at last- write the data bytes out....
    imageSize = (2L*nxOut*nyOut);
    
    err      =zlsio_write( fileAccess, img, imageSize );
    
    zlsio_close(fileAccess);
    
  } else { err = fileErr; }
  
  return (err);
  
}

int zls_Write_tif8RGB     (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 16 bit tiff  grayscale - no colour table
  //
  zFile fileAccess = NULL;  int     imageSize= 0;
  char  *imgRGB = NULL;
  int    i, j, jOffset;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x0000001E; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a8; // specific to this header
  const short tiffChannelBits   = 0x0008;     // for 8,8,8 array, bits per rgb 
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  
  // allocate rgb chunky pixel array an combine rgb planes
  imageSize = 3L*nxOut*nyOut;
  imgRGB = (char *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    
    fileAccess = zlsio_open( aFile, "wb");
    if (fileAccess) {
      
      zlsio_write( fileAccess, (void*)&tiffPreHeader0, 4UL );
      zlsio_write( fileAccess, (void*)&tiffPreHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
      zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
      zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per red
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per green
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per blue
      
      zlsio_write( fileAccess, (void*)&nTPS, 2 );
      
      tiffTag.tCode = 0x0100; // width code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nxOut;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0101; // height code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0102; // bits per pixel code
      tiffTag.dataType = 3;
      tiffTag.nDataValues = 3;
      tiffTag.pDataField = 0x00000018; //offset to bbp table
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0103; // compression code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000001; // none compression
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0106; // Perception Intent code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00020000; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000002; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0115; // samples per pix
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00030000; // 3 channels
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000003; // 3 channels
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block! nya nya! 8k sheesh!)
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut; // 1 channel
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0117; // StripByteCounts
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = imageSize; // 1 channel
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0111; // strip offset from start of whole file
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = tiffImageDataOffset;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000008;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000010;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      zlsio_write( fileAccess, (void*)&ifdTermintator, 4UL );
      
      // at last- write the data bytes out....
      
      err      =zlsio_write( fileAccess, imgRGB, imageSize );
      
      zlsio_close(fileAccess);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err);
  
}

int zls_Write_tif16RGB     (char * aFile, UInt16 * imgR, UInt16  * imgG, UInt16  * imgB, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 16 RGB bit tiff - no colour table
  //
  zFile fileAccess = NULL;  
  int     imageSize= 0;
  UInt16   *imgRGB = NULL;
  int    i, j, jOffset;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x0000001E; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a8; // specific to this header
  const short tiffChannelBits   = 0x0010;     // for 16,16,16 array, bits per rgb 
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  // allocate rgb chunky pixel array an combine rgb planes
  imageSize = 6L*nxOut*nyOut;
  imgRGB = (UInt16 *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    
    fileAccess = zlsio_open( aFile, "wb");
    if (fileAccess) {
      
      zlsio_write( fileAccess, (void*)&tiffPreHeader0, 4UL );
      zlsio_write( fileAccess, (void*)&tiffPreHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
      zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
      zlsio_write( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per red
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per green
      zlsio_write( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per blue
      
      zlsio_write( fileAccess, (void*)&nTPS, 2 );
      
      tiffTag.tCode = 0x0100; // width code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nxOut;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0101; // height code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0102; // bits per pixel code
      tiffTag.dataType = 3;
      tiffTag.nDataValues = 3;
      tiffTag.pDataField = 0x00000018; // offset to table of 3 sample sizes
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0103; // compression code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000001; // none compression
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0106; // Perception Intent code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00020000; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000002; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0115; // samples per pix
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00030000; // three samples, rgb
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000003; // three samples, rgb
#endif
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block! nya nya! 8k sheesh!)
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut; // 1 channel
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0117; // StripByteCounts
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = imageSize; // 1 channel
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0111; // strip offset from start of whole file
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = tiffImageDataOffset;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000008;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000010;
      
      zlsio_write( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      zlsio_write( fileAccess, (void*)&ifdTermintator, 4UL );
      
      // at last- write the data bytes out....
      
      err      = zlsio_write( fileAccess, imgRGB, imageSize );
      zlsio_close(fileAccess);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err );
}

int zls_Write_tif8_NC     (char * aFile, UInt8 * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 8 bit tiff with 256 colour colour table, if outTIFFCTab == NULL then do Grayscale
  // Uncompressed even if ZLIB_COMPRESSED_IO is enabled
  //
  zFileNC fileAccess = NULL;  int     imageSize= 0;
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x00000618; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000006ae; // specific to this header
  const short nTPS              = 0x000C;     // 12 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    
    imageSize = (nxOut*nyOut);
    
    zlsio_writeNC( fileAccess, (void*)&tiffPreHeader0, 4UL );
    zlsio_writeNC( fileAccess, (void*)&tiffPreHeader1, 4UL );
    
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
    
    zlsio_writeNC( fileAccess, outTIFFCTab, sizeof(TIFFCTab) );
    
    zlsio_writeNC( fileAccess, (void*)&nTPS, 2 );
    
    tiffTag.tCode = 0x0100; // width code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nxOut;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0101; // height code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0102; // bits per pixel code
    tiffTag.dataType = 3;
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00080000; // 8 bits per pixel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000008; // 8 bits per pixel
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0103; // compression code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; //none compression
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; //none compression
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0106; // Perception Intent code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00030000; // 1 => 0 = black GS,  3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000003; // 1 => 0 = black GS,  3=> colour TABLE
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0115; // samples per pix
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 channel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 channel
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block!)
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut; // 1 channel
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0117; // StripByteCounts
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = imageSize; // 1 channel
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0111; // strip offset from start of whole file
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = tiffImageDataOffset;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000008;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000010;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0140; // colour map/table
    tiffTag.dataType = 3;   // SHORT
    tiffTag.nDataValues = 3*256;
    tiffTag.pDataField = 0x00000018; // location of table n bytes from SOF
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    zlsio_writeNC( fileAccess, (void*)&ifdTermintator, 4UL );
    
    // at last- write the data bytes out....
    
    err      =zlsio_writeNC( fileAccess, img, imageSize );
    
    zlsio_closeNC(fileAccess);
    
  } else { err = fileErr; }
  
  return (err);
  
}

int zls_Write_tif16_NC    (char * aFile, UInt16 * img, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 16 bit tiff  grayscale - no colour table, uncompressed even if ZLIB_COMPRESSED_IO is enabled
  //
  zFileNC fileAccess = NULL;  int     imageSize= 0;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x00000018; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a2; // specific to this header
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  fileAccess = zlsio_openNC( aFile, "wb");
  if (fileAccess) {
    
    imageSize = (2L*nxOut*nyOut);
    
    zlsio_writeNC( fileAccess, (void*)&tiffPreHeader0, 4UL );
    zlsio_writeNC( fileAccess, (void*)&tiffPreHeader1, 4UL );
    
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
    zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
    
    zlsio_writeNC( fileAccess, (void*)&nTPS, 2 );
    
    tiffTag.tCode = 0x0100; // width code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nxOut;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0101; // height code
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0102; // bits per pixel code
    tiffTag.dataType = 3;
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00100000; // 16 bits per pixel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000010; // 16 bits per pixel
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0103; // compression code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // none compression
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0106; // Perception Intent code
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 => 0 = black GS,  3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 => 0 = black GS,  3=> colour TABLE
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x0115; // samples per pix
    tiffTag.dataType = 3; // short
    tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
    tiffTag.pDataField = 0x00010000; // 1 channel
#endif
#ifdef TIFF_LITENDIAN_OS
    tiffTag.pDataField = 0x00000001; // 1 channel
#endif
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block!)
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = nyOut; // 1 channel
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0117; // StripByteCounts
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = imageSize; // 1 channel
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    
    tiffTag.tCode = 0x0111; // strip offset from start of whole file
    tiffTag.dataType = 4; // long
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = tiffImageDataOffset;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000008;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
    tiffTag.dataType = 5;
    tiffTag.nDataValues = 1;
    tiffTag.pDataField = 0x00000010;
    
    zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
    
    zlsio_writeNC( fileAccess, (void*)&ifdTermintator, 4UL );
    
    // at last- write the data bytes out....
    imageSize = (2L*nxOut*nyOut);
    
    err      =zlsio_writeNC( fileAccess, img, imageSize );
    
    zlsio_closeNC(fileAccess);
    
  } else { err = fileErr; }
  
  return (err);
  
}


int zls_Write_tif8RGB_NC    (char * aFile, UInt8 * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write 8 bit tiff  grayscale, uncompressed even if ZLIB_COMPRESSED_IO is in effect
  //
  zFileNC fileAccess = NULL;  int     imageSize= 0;
  char  *imgRGB = NULL;
  int    i, j, jOffset;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x0000001E; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a8; // specific to this header
  const short tiffChannelBits   = 0x0008;     // for 8,8,8 array, bits per rgb 
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  
  // allocate rgb chunky pixel array an combine rgb planes
  imageSize = 3L*nxOut*nyOut;
  imgRGB = (char *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    
    fileAccess = zlsio_openNC( aFile, "wb");
    if (fileAccess) {
      
      zlsio_writeNC( fileAccess, (void*)&tiffPreHeader0, 4UL );
      zlsio_writeNC( fileAccess, (void*)&tiffPreHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per red
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per green
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per blue
      
      zlsio_writeNC( fileAccess, (void*)&nTPS, 2 );
      
      tiffTag.tCode = 0x0100; // width code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nxOut;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0101; // height code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0102; // bits per pixel code
      tiffTag.dataType = 3;
      tiffTag.nDataValues = 3;
      tiffTag.pDataField = 0x00000018; //offset to bbp table
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0103; // compression code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000001; // none compression
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0106; // Perception Intent code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00020000; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000002; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0115; // samples per pix
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00030000; // 3 channels
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000003; // 3 channels
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block! nya nya! 8k sheesh!)
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut; // 1 channel
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0117; // StripByteCounts
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = imageSize; // 1 channel
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0111; // strip offset from start of whole file
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = tiffImageDataOffset;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000008;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000010;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      zlsio_writeNC( fileAccess, (void*)&ifdTermintator, 4UL );
      
      // at last- write the data bytes out....
      
      err      =zlsio_writeNC( fileAccess, imgRGB, imageSize );
      
      zlsio_closeNC(fileAccess);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err);
  
}


int zls_Write_tif16RGB_NC(char * aFile, UInt16 * imgR, UInt16  * imgG, UInt16  * imgB, int nxOut, int nyOut ){
  int err = noErr;
  //
  // write an 16 RGB bit tiff - no colour table, not compressed  even if ZLIB_COMPRESSED_IO is in force
  //
  zFileNC fileAccess = NULL;   // non-compressed file type
  int     imageSize= 0;
  UInt16   *imgRGB = NULL;
  int    i, j, jOffset;
  
  
#ifdef TIFF_BIGENDIAN_OS
  const int tiffPreHeader0= 0x4d4d002a; // basic bigendian TIFF magic number, chose one!
#endif
  
#ifdef TIFF_LITENDIAN_OS
  const int tiffPreHeader0= 0x002a4949; // basic littleendian TIFF magic number
#endif
  
  //
  // fixed header order and offsets.  Preheader, resolution, ctab, tags then data
  //
  const int tiffPreHeader1      = 0x0000001E; // offset to tags
  const int tiffXResHeader0     = 0x00000048; // 72 pix/inch default, numerator
  const int tiffXResHeader1     = 0x00000001; // denominator => 72.0 = 72/1
  const int tiffImageDataOffset = 0x000000a8; // specific to this header
  const short tiffChannelBits   = 0x0010;     // for 16,16,16 array, bits per rgb 
  const short nTPS              = 0x000B;     // 11 tags in IFD
  const int ifdTermintator      = 0x00000000; // all TIFF blocks end in 4 bytes of 0
  
  TagStructure tiffTag;
  
  // allocate rgb chunky pixel array an combine rgb planes
  imageSize = 6L*nxOut*nyOut;
  imgRGB = (UInt16 *) malloc(imageSize);
  
  if (imgRGB != NULL) {
    
    for (j = 0; j<nyOut; j++){
      jOffset = nxOut*j;
      for (i = 0; i<nxOut; i++){
        
        imgRGB[3*(i + jOffset)  ] = imgR[i + jOffset];
        imgRGB[3*(i + jOffset)+1] = imgG[i + jOffset];
        imgRGB[3*(i + jOffset)+2] = imgB[i + jOffset];
        
      }//i
    }//j
    
    
    fileAccess = zlsio_openNC( aFile, "wb");
    if (fileAccess) {
      
      zlsio_writeNC( fileAccess, (void*)&tiffPreHeader0, 4UL );
      zlsio_writeNC( fileAccess, (void*)&tiffPreHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // X-resolution Rational long1/long2
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader0, 4UL ); // Y-resolution Rational long1/long2
      zlsio_writeNC( fileAccess, (void*)&tiffXResHeader1, 4UL );
      
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per red
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per green
      zlsio_writeNC( fileAccess, (void*)&tiffChannelBits, 2 ); // 8 bits per blue
      
      zlsio_writeNC( fileAccess, (void*)&nTPS, 2 );
      
      tiffTag.tCode = 0x0100; // width code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nxOut;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0101; // height code
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0102; // bits per pixel code
      tiffTag.dataType = 3;
      tiffTag.nDataValues = 3;
      tiffTag.pDataField = 0x00000018; // offset to table of 3 sample sizes
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0103; // compression code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00010000; // none compression
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000001; // none compression
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0106; // Perception Intent code
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1; 
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00020000; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000002; // 1 => 0 = black GS, 2=> RGB, 3=> colour TABLE
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x0115; // samples per pix
      tiffTag.dataType = 3; // short
      tiffTag.nDataValues = 1;
#ifdef TIFF_BIGENDIAN_OS
      tiffTag.pDataField = 0x00030000; // three samples, rgb
#endif
#ifdef TIFF_LITENDIAN_OS
      tiffTag.pDataField = 0x00000003; // three samples, rgb
#endif
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0116; // Rows Per Strip (all in one block! nya nya! 8k sheesh!)
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = nyOut; // 1 channel
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0117; // StripByteCounts
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = imageSize; // 1 channel
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      
      tiffTag.tCode = 0x0111; // strip offset from start of whole file
      tiffTag.dataType = 4; // long
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = tiffImageDataOffset;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011A; // X-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000008;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      tiffTag.tCode = 0x011B; // Y-resolution rational offset back to preheader
      tiffTag.dataType = 5;
      tiffTag.nDataValues = 1;
      tiffTag.pDataField = 0x00000010;
      
      zlsio_writeNC( fileAccess, (void*)&tiffTag, sizeof(TagStructure) );
      
      zlsio_writeNC( fileAccess, (void*)&ifdTermintator, 4UL );
      
      // at last- write the data bytes out....
      
      err      = zlsio_writeNC( fileAccess, imgRGB, imageSize );
      zlsio_closeNC(fileAccess);
      
    } else { err = fileErr;}
    
    free(imgRGB);
    
  } // imgRGB allocated OK
  else { err = memErr;}
  
  return (err );
}

//========================================
//
// TIFF Colour Table IO Routines
//
//========================================


int zls_ReadCtab(char * cfname, TIFFCTab * tCtab){
  //
  // Read Text based  Color Tables
  // FORMAT:
  // line 1: fmt  (0 = decimal, 1= hex, 2 = float 0.0 -> 1.0)
  // line 2: componentSize  (8 or 16)
  // line 3: nComponents    (1 [grayscale] or 3 [rgb])
  // line 4: nEntries       (256)
  // line 5-260: index red green blue
  //
  
  int err = noErr;
  
  int colourFmt;
  int colourIndex;
  int componentSize;
  int nComponents;
  int nEntries;
  FILE * cfile;
  
  unsigned int entryIndex, red, green, blue, gray;
  float fred, fgreen, fblue, fgray;
  
  cfile = fopen(cfname, "r");
  
  if (cfile) {
    
    fscanf(cfile,"%d",&colourFmt);
    fscanf(cfile,"%d",&componentSize);
    fscanf(cfile,"%d",&nComponents);
    fscanf(cfile,"%d",&nEntries);
    
    if (colourFmt == 0){ // decimal entries
      
      if (componentSize == 8) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d %d %d",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red   | ( red << 8));
            tCtab->green[entryIndex] = (UInt16)(green | ( green << 8));
            tCtab->blue[entryIndex]  = (UInt16)(blue  | ( blue << 8));
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray   | (gray << 8));
            tCtab->green[entryIndex] = (UInt16)(gray   | (gray << 8));
            tCtab->blue[entryIndex]  = (UInt16)(gray   | (gray << 8));
          }
        }
        
      }
      
      if (componentSize == 16) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d %d %d",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red);
            tCtab->green[entryIndex] = (UInt16)(green);
            tCtab->blue[entryIndex]  = (UInt16)(blue);
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %d",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray);
            tCtab->green[entryIndex] = (UInt16)(gray);
            tCtab->blue[entryIndex]  = (UInt16)(gray);
          }
        }
        
      }
    }
    
    if (colourFmt == 1){ // hex entries
      
      if (componentSize == 8) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x %x %x",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red   |(red << 8));
            tCtab->green[entryIndex] = (UInt16)(green |(green << 8));
            tCtab->blue[entryIndex]  = (UInt16)(blue  |(blue << 8));
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray   | (gray << 8));
            tCtab->green[entryIndex] = (UInt16)(gray   | (gray << 8));
            tCtab->blue[entryIndex]  = (UInt16)(gray   | (gray << 8));
          }
        }
        
      }
      
      if (componentSize == 16) {
        
        if (nComponents == 3) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x %x %x",&entryIndex, &red, &green, &blue);
            tCtab->red[entryIndex]   = (UInt16)(red);
            tCtab->green[entryIndex] = (UInt16)(green);
            tCtab->blue[entryIndex]  = (UInt16)(blue);
          }
        }
        
        if (nComponents == 1) {
          for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
            fscanf(cfile,"%d %x",&entryIndex, &gray);
            tCtab->red[entryIndex]   = (UInt16)(gray);
            tCtab->green[entryIndex] = (UInt16)(gray);
            tCtab->blue[entryIndex]  = (UInt16)(gray);
          }
        }
        
      }
    }
    
    if (colourFmt == 2){ // float entries
      
      // ignore component size and map 0 - 1 to 0-65535  
      
      if (nComponents == 3) {
        for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
          fscanf(cfile,"%d %f %f %f",&entryIndex, &fred, &fgreen, &fblue);
          tCtab->red[entryIndex]   = (UInt16)(fred   * 0x0000ffff);
          tCtab->green[entryIndex] = (UInt16)(fgreen * 0x0000ffff);
          tCtab->blue[entryIndex]  = (UInt16)(fblue  * 0x0000ffff);
        }
      }
      
      if (nComponents == 1) {
        for (colourIndex = 0; colourIndex<nEntries; colourIndex++){
          fscanf(cfile,"%d %f",&entryIndex, &fgray);
          tCtab->red[entryIndex]   = (UInt16)(fgray  * 0x0000ffff);
          tCtab->green[entryIndex] = (UInt16)(fgray  * 0x0000ffff);
          tCtab->blue[entryIndex]  = (UInt16)(fgray  * 0x0000ffff);
        }
      }
      
    }
    
    err      = ferror (cfile); // one error check for whole write sequence
    fclose(cfile);
  } else {err = fileErr;}
  
  return(err);
  
}
