#ifndef ZLS_TIFFIO_H
#define ZLS_TIFFIO_H

#include <zls.h>

// TIFF_BIGENDIAN_OS bigendian - TIFF standard, and native for
// Motorola, PowerPC, SPARC. TIFF_LITENDIAN_OS little endian -
// non-standard TIFFs, easy to write on x86 systems, Alphas etc


#ifdef DATA_BIGENDIAN_OS
#define TIFF_BIGENDIAN_OS
#endif DATA_BIGENDIAN_OS

#ifdef DATA_LITENDIAN_OS
#define TIFF_LITENDIAN_OS
#endif DATA_LITENDIAN_OS

// default TIFF_BIGENDIAN_OS if nothing matches

#ifndef TIFF_BIGENDIAN_OS
#ifndef TIFF_LITENDIAN_OS
#define TIFF_BIGENDIAN_OS
#endif
#endif

//
// TIFF Structures
//

typedef struct {
  UInt16 tCode;
  UInt16 dataType;
  unsigned int   nDataValues;
  unsigned int   pDataField;
} TagStructure;

// TIFF Colour table

typedef struct{
  UInt16 red[256];   // 0x0000 -> 0xFFFF, if input is 0-255 then byte double
  UInt16 green[256]; // 0x0000 -> 0xFFFF, ie 0xCA -> 0xCACA
  UInt16 blue[256];  // 0x0000 -> 0xFFFF
} TIFFCTab;


//========================================
//
// Raw File IO Routines
//
//========================================

// if ZLIB_COMPRESS_IO is enabled these write gz compressed files, otherwise they write uncompressed files

int zls_Write_raw8     (char * aFile, UInt8  * img, int nxOut, int nyOut );
int zls_Write_raw16    (char * aFile, UInt16 * img, int nxOut, int nyOut );

int zls_Write_raw8RGB  (char * aFile, UInt8  * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut );
int zls_Write_raw16RGB (char * aFile, UInt16 * imgR, UInt16 * imgG, UInt16 * imgB, int nxOut, int nyOut );

// Non_Compressed versions, even if ZLIB_COMPRESS_IO is enabled

int zls_Write_raw8_NC     (char * aFile, UInt8  * img, int nxOut, int nyOut );
int zls_Write_raw16_NC    (char * aFile, UInt16 * img, int nxOut, int nyOut );

int zls_Write_raw8RGB_NC  (char * aFile, UInt8  * imgR, UInt8 * imgG, UInt8 * imgB, int nxOut, int nyOut );
int zls_Write_raw16RGB_NC (char * aFile, UInt16 * imgR, UInt16 * imgG, UInt16 * imgB, int nxOut, int nyOut );

//========================================
//
// TIFF File IO Routines
//
//========================================
// if ZLIB_COMPRESS_IO is enabled these write gz compressed files, otherwise they write uncompressed files

int zls_Write_tif8     (char * aFile, UInt8  * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut );   // 8 bit colour table, use gray ctab for grayscale
int zls_Write_tif16    (char * aFile, UInt16 * img, int nxOut, int nyOut );  // 16 bit grayscale

int zls_Write_tif8RGB  (char * aFile, UInt8  * imgR, UInt8   * imgG,  UInt8   * imgB, int nxOut, int nyOut ); // normal 8 bit per channel RGB, no ctab
int zls_Write_tif16RGB (char * aFile, UInt16 * imgR, UInt16  * imgG,  UInt16  * imgB, int nxOut, int nyOut ); // 16 bit per channel RGB, no ctab, for photoshop

// Non_Compressed versions, even if ZLIB_COMPRESS_IO is enabled

int zls_Write_tif8_NC     (char * aFile, UInt8  * img, TIFFCTab * outTIFFCTab, int nxOut, int nyOut );   // 8 bit colour table, use gray ctab for grayscale
int zls_Write_tif16_NC    (char * aFile, UInt16 * img, int nxOut, int nyOut );  // 16 bit grayscale

int zls_Write_tif8RGB_NC  (char * aFile, UInt8  * imgR, UInt8   * imgG,  UInt8   * imgB, int nxOut, int nyOut ); // normal 8 bit per channel RGB, no ctab
int zls_Write_tif16RGB_NC (char * aFile, UInt16 * imgR, UInt16  * imgG,  UInt16  * imgB, int nxOut, int nyOut ); // 16 bit per channel RGB, no ctab, for photoshop

//========================================
//
// TIFF Colour Table IO Routines
//
//========================================


int zls_ReadCtab(char * cfname, TIFFCTab * tCtab);

#endif ZLS_TIFFIO_H
