#ifndef ZLS_HYDRO_H
#define ZLS_HYDRO_H
#include <zls.h>

//========================================
//
int zlsh_PrintW1D    (W1DArr w, Integer cellIndex); // cellIndex = -1 for whole array
int zlsh_ListW1D     (W1DArr w, Integer cellIndex); // cellIndex = -1 for whole array

int zlsh_ClearW1DArr (W1DArr *w);
int zlsh_ClearW2DArr (W2DArr *w);
int zlsh_ClearW3DArr (W3DArr *w);

int zlsh_NewW1DArr  (W1DArr *w, Integer nBC, Counter ni);
int zlsh_NewW2DArr  (W2DArr *w, Integer nBC, Counter nj, Counter ni);
int zlsh_NewW3DArr  (W3DArr *w, Integer nBC, Counter nk, Counter nj, Counter ni);

int zlsh_DisposeW1DArr (W1DArr w);
int zlsh_DisposeW2DArr (W2DArr w);
int zlsh_DisposeW3DArr (W3DArr w);
// set a V1DArr from a W1DArr ready for calc coeffs etc
// fast because it simply copies references, so don't dispose the var
// afterwards.
int zlsh_SetVar1DFromW1D(V1DArr v, W1DArr w, Integer varIndex);
//========================================
//
//========================================
// Vector fields...  field can be 1 2 or 3d structure with 1,2 or 3 D vectors depending on SYMXD
// Used by gravity fields to hold accelleration 
//
int zlsh_NewF1DArr  (F1DArr *f, Integer nBC, Counter ni);
int zlsh_NewF2DArr  (F2DArr *f, Integer nBC, Counter nj, Counter ni);
int zlsh_NewF3DArr  (F3DArr *f, Integer nBC, Counter nk, Counter nj, Counter ni);
//
int zlsh_DisposeF1DArr (F1DArr f);
int zlsh_DisposeF2DArr (F2DArr f);
int zlsh_DisposeF3DArr (F3DArr f);
//
//  One 1 Row getters/setters for Physical/Primitive variable arrays w,
//
//========================================
//
// 1D Zones:
//
int zlsh_FillW1DFrom1DZoneWHD  ( W1DArr  w,  WHD zoneW ); // fills a preexisting w from preallocted WHD
int zlsh_Fill1DZoneWHDFromW1D  ( WHD zoneW,  W1DArr w  ); 
int zlsh_FillW1DDivVFrom1DZone ( W1DArr  w,  ZonePtr z );

//========================================
//
// 2D Zones:
//
int zlsh_FillW1DFrom2DZoneWHDX ( W1DArr w,  WHD zoneW, Integer j ); // fills a preexisting w from preallocted WHD
int zlsh_FillW1DFrom2DZoneWHDY ( W1DArr w,  WHD zoneW, Integer i );

int zlsh_Fill2DZoneWHDFromW1DX ( WHD zoneW,  W1DArr w,  Integer j ); // reverse of the two above
int zlsh_Fill2DZoneWHDFromW1DY ( WHD zoneW,  W1DArr w,  Integer i );

int zlsh_GetZone2DW1DX ( W1DArr *w, ZonePtr z, Integer j ); // creates new w, allocates and releases a WHD
int zlsh_GetZone2DW1DY ( W1DArr *w, ZonePtr z, Integer i );

int zlsh_FillW1DDivVFrom2DZoneX ( W1DArr w,  ZonePtr z, Integer j);
int zlsh_FillW1DDivVFrom2DZoneY ( W1DArr w,  ZonePtr z, Integer i);

//========================================
//
// 3D Zones:
//
int zlsh_FillW1DXFrom3DZoneWHDYZ ( W1DArr w,  WHD zoneW, Integer j, Integer k );
int zlsh_FillW1DYFrom3DZoneWHDZX ( W1DArr w,  WHD zoneW, Integer k, Integer i );
int zlsh_FillW1DZFrom3DZoneWHDXY ( W1DArr w,  WHD zoneW, Integer i, Integer j );

int zlsh_Fill3DZoneWHDYZFromW1DX ( WHD zoneW, W1DArr w, Integer j, Integer k );
int zlsh_Fill3DZoneWHDZXFromW1DY ( WHD zoneW, W1DArr w, Integer k, Integer i );
int zlsh_Fill3DZoneWHDXYFromW1DZ ( WHD zoneW, W1DArr w, Integer i, Integer j );

int zlsh_GetZone3DW1DX ( W1DArr *w, ZonePtr z, Integer j, Integer k ); 
int zlsh_GetZone3DW1DY ( W1DArr *w, ZonePtr z, Integer k, Integer i ); 
int zlsh_GetZone3DW1DZ ( W1DArr *w, ZonePtr z, Integer i, Integer j ); 

int zlsh_FillW1DDivVFrom3DZoneYZ ( W1DArr w,  ZonePtr z, Integer j, Integer k );
int zlsh_FillW1DDivVFrom3DZoneZX ( W1DArr w,  ZonePtr z, Integer k, Integer i ); 
int zlsh_FillW1DDivVFrom3DZoneXY ( W1DArr w,  ZonePtr z, Integer i, Integer j ); 

//========================================

int zlsh_NewWFluxes1DArr     ( Flx1DArr *fx, Integer nBC, Counter ni );
int zlsh_NewWFluxes2DArr     ( Flx2DArr *fx, Integer nBC, Counter nj, Counter ni );
int zlsh_NewWFluxes3DArr     ( Flx3DArr *fx, Integer nBC, Counter nk, Counter nj, Counter ni );

int zlsh_ClearWFluxes1DArr   ( Flx1DArr fx );
int zlsh_ClearWFluxes2DArr   ( Flx2DArr fx );
int zlsh_ClearWFluxes3DArr   ( Flx3DArr fx );

int zlsh_DisposeWFluxes1DArr ( Flx1DArr fx );
int zlsh_DisposeWFluxes2DArr ( Flx2DArr fx );
int zlsh_DisposeWFluxes3DArr ( Flx3DArr fx );

//========================================

int zlsh_CopyU1DfromW1D     ( U1DArr u, W1DArr w );
int zlsh_CopyU1DLRfromW1DLR ( U1DArr u, W1DArr w );
int zlsh_CopyU1DCfromW1DC   ( U1DArr u, W1DArr w );

int zlsh_CopyW1DfromU1D     ( W1DArr w, U1DArr u );
int zlsh_CopyW1DLRfromU1DLR ( W1DArr w, U1DArr u );
int zlsh_CopyW1DCfromU1DC   ( W1DArr w, U1DArr u );

#endif ZLS_HYDRO_H
