#ifndef ZLS_PPM_H
#define ZLS_PPM_H
#include <zls.h>
	

enum {
	kNoFlat            = 0,
	kDoFlat            = 1,
	kNoSkip            = 0,
	kDoSkip            = 1	
};

// flags for LRQArr coeff computations.
enum {
  kZLS_ComputeLR     = 1,
  kZLS_DontComputeLR = 0
};

extern Real kZLS_Omega1      ; // 0.75 decrease for greater sensitivity to +/- 1 P diffs/ +/-2 P Diffs
extern Real kZLS_Omega2      ; // 5.00 increase for stronger flattening/dissipation if threshold are satisfied
extern Real kZLS_CWEpsilon   ; // 0.330 decrease for greater sensitivity to pressure jumps
extern Real kZLS_PPMFlatMin  ; // 0.0 increase for general diffusivity
extern Real kZLS_PPMFlatMax  ; // 1.0 decrease for limiting max diffusivity
extern Real kZLS_PPMFlatDd   ; // 0.05 dd threshold for continuing calculations
extern Real kZLS_PPMFlatDP   ; // 0.05 dP threshold for continuing calculations
extern Real kZLS_PPMFlatDv   ; // 0.05 dv threshold for continuing calculations


// funamanetal PPM interpolation fractions 1/6 and 1/4

static const Real kZLS_quarter = (1.0/4.0);
static const Real kZLS_sixth   = (1.0/6.0);
static const Real kZLS_36th    = (1.0/36.0);

// denominators n = 2 to 8  (n=1 is no interp!)
enum {    
	kAjDeltaN2 =  4, kAjA6N2 =  0,
	kAjDeltaN3 =  3, kAjA6N3 = 27,
	kAjDeltaN4 =  8, kAjA6N4 = 16,
	kAjDeltaN5 =  5, kAjA6N5 = 25,
	kAjDeltaN6 = 12, kAjA6N6 = 54,
	kAjDeltaN7 =  7, kAjA6N7 = 49,
	kAjDeltaN8 = 16, kAjA6N8 = 64
};

// delta coeffs
enum {   
	k0DeltaN2 =  -1, k1DeltaN2 =  +1,
	k0DeltaN3 =  -1, k1DeltaN3 =  +0, k2DeltaN3 =  +1,
	k0DeltaN4 =  -3, k1DeltaN4 =  -1, k2DeltaN4 =  +1, k3DeltaN4 =  +3,
	k0DeltaN5 =  -2, k1DeltaN5 =  -1, k2DeltaN5 =  +0, k3DeltaN5 =  +1,
	k4DeltaN5 =  +2,
	k0DeltaN6 =  -5, k1DeltaN6 =  -3, k2DeltaN6 =  -1, k3DeltaN6 =  +1,
	k4DeltaN6 =  +3, k5DeltaN6 =  +5,
	k0DeltaN7 =  -3, k1DeltaN7 =  -2, k2DeltaN7 =  -1, k3DeltaN7 =  +0,
	k4DeltaN7 =  +1, k5DeltaN7 =  +2, k6DeltaN7 =  +3,
	k0DeltaN8 =  -7, k1DeltaN8 =  -5, k2DeltaN8 =  -3, k3DeltaN8 =  -1,
	k4DeltaN8 =  +1, k5DeltaN8 =  +3, k6DeltaN8 =  +5, k7DeltaN8 =  +7
};


// a6 coeffs
enum {   
	k0A6N2 =   0, k1A6N2 =   0,
	k0A6N3 =  -1, k1A6N3 =  +2, k2A6N3 =  -1,
	k0A6N4 =  -2, k1A6N4 =  +1, k2A6N4 =  +1, k3A6N4 =  -1,
	k0A6N5 =  -2, k1A6N5 =  +1, k2A6N5 =  +2, k3A6N5 =  +1,
	k4A6N5 =  -2,
	k0A6N6 =  -5, k1A6N6 =  +1, k2A6N6 =  +4, k3A6N6 =  +4,
	k4A6N6 =  +1, k5A6N6 =  -5,
	k0A6N7 =  -5, k1A6N7 =  +0, k2A6N7 =  +3, k3A6N7 =  +4,
	k4A6N7 =  +3, k5A6N7 =  +0, k6A6N7 =  -5,
	k0A6N8 =  -7, k1A6N8 =  -1, k2A6N8 =  +3, k3A6N8 =  +5,
	k4A6N8 =  +5, k5A6N8 =  +3, k6A6N8 =  -1, k7A6N8 =  -7
};

extern  Real gZLS_PPM_fxDelta[kNRefineMax][(kNRefineMax+1)];
extern  Real    gZLS_PPM_fxA6[kNRefineMax][(kNRefineMax+1)];

extern const Real kZLS_PPM_Q1; // coefficients for estimating quadratic quantites from ppm coeffs of 
extern const Real kZLS_PPM_Q2; // linear quantities

//====================================================================================
//
//  Cell Boundary Reconstruction, highest level calls.
//
//====================================================================================

int zlsppm_ReconstructPPMWLRStates1DF( W1DArr w, F1DArr phi, Real dt, Real g);
// if f.xi != NULL then phi has row potential. Avge acceleration = f.xi[i].del/dx 

int zlsppm_ReconstructPPMWLRStates1D( W1DArr w, Real dt, Real g); 
// includes geometic coefficient for angular coords. Spatial dx = g*dphi

int zlsppm_InterpolateLRPPM1D( V1DArr v1, V1DArr v0);
//
// compute v1 at left coordinates in x1, interpolating v0, with values at x0 using PPM coeffs in v0
// works without a g factor because interp fractions are ratios of x1/x0 not absolute coordinates.
// left boundary of i = 0 and right boundary of i = ni-1 are assumed to not move and are
// just copied. Assumes courant condition and xl is always in cell i or i-1 only.
// 
// This doens't integrate over domain, but evaluates at the boundary coords of v1 using PPM 
// coeffs in v0.  Not the same as remap or reconstruction routines which integrate (average) 
// over the coordinate differences.
//
// can't do cell centre interp with this routine as the centre may move to i-1, i or i+1,
// use the slower LRC routine for that case. (nyiErr)
//
//====================================================================================
//
// Calculating PPM coefficients - on high level composite structures
//
//====================================================================================
//
// check for places to reduce parabola coeffs
//
int    zlsppm_CalcFlatteningW1DArr      ( W1DArr w ); 

int    zlsppm_CalcAllCoeffsW1DArr       ( W1DArr w,  int iflat ); // general call
int    zlsppm_CalcCoeffsW1DArr          ( W1DArr w); // no flattening by default, used in remap.
int    zlsppm_CalcRemappingCoeffsW1DArr ( W1DArr w ); // same as above, but with P xor E depending on advection scheme
int    zlsppm_CalcFlatHydroCoeffsW1DArr ( W1DArr w ); // just P, d, vx, always flattened


//====================================================================================
// FXDarrs are verctor field arrays which may be based on conserved divergence free
// Field data (see zoneFeield), but not always.  The zoneField data is computed conservatively
// based on an integral form of intra cell interpolation  (not grid interpolation)
// , where the left and righ values are based internal cell values, and the celll\
// (a) value is integrated from the same function, so 
// al and ar may not be free to change. 
//
// Fields always have al, a, ar data only, based on LRQArrs, with a very low level structure.
// They are copied into FXDArrs with full PPM coeffs and coordinates for sweeps of the grid
// so that the force field can be included in the algorithms.
//
// General calls for F1DArrs when f is simply based on average cell data and
// requires grid interpolation, and flattening as usual.
//
// Speciaslised mid-level calls for then FXDArrs are  based on exact l/r values and 
// conservative interpolation coeffs are needed but without changing al or ar
// or for convering non-conservative central value data to conservative 
// cell average data.  This is important for gravtiational potential fields
// where conservtion of energy is particularly important and normal falttening
// and grid interpolation is not needed.
// 
// For everyday fields call the CalcCoeffs calls with just f.xi.a set up, al and 
// ar will be recoputed.  For f.xi with special al and ar values use the 
// zlsppm_CalcDelA6F1DArr and zlsppm_CalcMeanF1DArr calls as needed. ie with gravity
//
int    zlsppm_CalcCoeffsF1DArr ( F1DArr f );
//int    zlsppm_CalcCoeffsF2DArr ( F2DArr f );
//int    zlsppm_CalcCoeffsF3DArr ( F3DArr f );
//====================================================================================
// mid-level calls for F1DArrs F3DArrs F3DArrs, most of which already have l/r values 
// derived from LRQArr data, but are actually VArrs with coords and everything.
// The differences with Var routines is how l and r are treated and if a_c is used
// instead of (a) for intial values.  
//====================================================================================
// IMPORTANT L/R values are fundamental here; they are known, and required to be kept
// unchanged, as they are exact and the basis of the integrals.  This is subtly different to 
// the lr values from grid cell-cell interpolated values of al and ar.
// In this case we don't fiddle the parabolic monotone limits.
// Since al and ar are not free parameters.  This better be used for well behaves fields!
//

int  zlsppm_CalcDelA6F1DArr  ( F1DArr f );

//
// for LR data with central a values to start with and pre-given exact LR values. 
// this routine updates a to contain <a> if PPM is in force  and does the del and a6 coeffs too.
// If PLM is in force then this will replace a with the mean of al and ar
//

int  zlsppm_CalcMeanF1DArr  ( F1DArr f );

// Compute (a) Delta and a6 if the given a is actually the central cell value at
// x_c and not the cell average.  Assuming a PPM fit the equivalent average (a) can 
// be derived from a_c, al, and ar, just as a_c can come from the other three.
//  
//  a6  = 6*[a - 1/2(al + ar)] (standard formula
//  a   = al + del/2 + a6/6  ( standard formula
//
//  a in terms of a_c :
//
//  a  = 1/6*[4a_c + al + ar]
//
//  if we have a_c not a, plus al and ar are pre-determined:
//
//  a6  = 4*[a_c - 1/2(al + ar) ] ( complementary formula
//  a_c = al + del/2 + a6/4 ( complementary formulae
//  
//  a_c in terms of a :
//
//  a_c  = 1/4*[6a - al - ar]
//
// As can be seen, working with a_c or a is similar, however we choose to work with 
// a as it is a conservative quantity based on an integral property of the cell
// so if for some reason we ge a_c values and exact al and ar values then we 
// change a_c to a.   This can happen when a grid is filled with just a central and edge 
// evaluations of a function, especially in 1D, ie a gravity potential without a convenient 
// integral or differntial form.
//
//====================================================================================
//
// Calculating PPM coefficients - on low level structures
//
//====================================================================================
// these calls <a> s to compute al, ar (by interpolation) and then  del and a6, the PPM coeffs

int    zlsppm_CalcLRVar1DArr    ( V1DArr v,  Q1DArr flat );
// just the determination of al and ar, with flattening when needed
int    zlsppm_CalcDelA6Var1DArr ( V1DArr v);
// separated out the PPM delta and a6 calcs

int    zlsppm_CalcCoeffsVar1DArr  ( V1DArr v, Q1DArr flat ); 
// flat may be NULL, v must have a full axis from a grid or elsewhere first, calls the above in order

// these calls will use given al/ar and ac values to compute <a>, del and a6 the PPM coeffs
//int    zlsppm_CalcMeanDelA6Var1DArr (V1DArr v); 
//
//====================================================================================
// low level calls for LRQArrs, most of which already have l/r values 
//====================================================================================
//
// for LR data with a but without pre-given LR values, interps on grid same as PPM, needs coordinates
// as v doesn't contain that info like a full VArr does. Assumes v has average values in place, only!
// al and and ar will change.

int  zlsppm_CalcLRQ1DArr  ( LRQ1DArr va, GCoordPtr grid, Q1DArr flat);

//
//
//int    zlsppm_CalcLRQ2DArr  ( LRQ2DArr va, GCoordPtr grid, Q2DArr flat){
//	int err = nyiErr;
//	return (err);
//}
//
//int    zlsppm_CalcLRQ3DArr  ( LRQ3DArr va, GCoordPtr grid, Q3DArr flat){
//	int err = nyiErr;
//	return(err);
//}
//
//
// for LR data with pre-given LR values, and cental values rather than means.
// If PPM is used the (a) is computed from a_c and al and ar.  If PLM is used 
// (a) is computed form al and ar only.
// Simliar to the F1DArr call, but can be used on the underlying zoneField data
// So when copied to and F1DArr, conversion is not needed every time.  Baiscly an option to 
// do the conversion at a lower level.  nti is needed as va has no size info.  but we are not
// interpolating on a grid so the grid info and flatten are not needed.
//

int  zlsppm_CalcMeanLRQ1DArr  ( LRQ1DArr va, Integer nti);
//int  zlsppm_CalcMeanLRQ2DArr  ( LRQ2DArr va, Integer ntj, Integer nti);
//int  zlsppm_CalcMeanLRQ3DArr  ( LRQ3DArr va, Integer ntk, Integer ntj, Integer nti);

//
//========================================
// PPM specific quadratic weights - ie to determine < d^2 > instead of < d >^2 if PPM coeffs
// are known.
//  < a^2 > = < a >^2 + kZLS_PPM_Q1*( del^2 + kZLS_PPM_Q2*a6^2  ) where del and a6 are the PPM terms
// notice: al is not used, but aj is ie aj = < a > 
//========================================
//========================================
// PPM specific uniform spacing interpolation coefficients
//========================================
//
void   zlsppm_InitVarInterpCoeffs(void);

int    zlsppm_CalcInterpVar1D		(V1DArr  v, InPtr f, int n);
int    zlsppm_CalcInterpVar2D		(V2DArr  v, InPtr f, int n);
int    zlsppm_CalcInterpVar3D		(V3DArr  v, InPtr f, int n);

int    zlsppm_CalcLinInterpVar1D    (V1DArr  v, InPtr f, int n);



#endif ZLS_PPM_H
