#include <zls.h>


Real gInv = 1.0/(ZLS_EOS_GAMMA);
Real g0 = (ZLS_EOS_GAMMA);
Real g1 = (ZLS_EOS_GAMMA-1.0)/(2.0*ZLS_EOS_GAMMA);
Real g2 = (ZLS_EOS_GAMMA+1.0)/(2.0*ZLS_EOS_GAMMA);
Real g3 = (2.0*ZLS_EOS_GAMMA)/(ZLS_EOS_GAMMA-1.0);
Real g4 = 2.0/(ZLS_EOS_GAMMA-1.0);
Real g5 = 2.0/(ZLS_EOS_GAMMA+1.0);
Real g6 = (ZLS_EOS_GAMMA-1.0)/(ZLS_EOS_GAMMA+1.0);
Real g7 = (ZLS_EOS_GAMMA-1.0)/2.0;

int zlsr_InitToroRiemannGammas(Real g);
int zlsr_InitToroRiemannGammas(Real g){
  
  gInv = 1.0/(g);
  g0 = (g);
  g1 = (g-1.0)/(2.0*g);
  g2 = (g+1.0)/(2.0*g);
  g3 = (2.0*g)/(g-1.0);
  g4 = 2.0/(g-1.0);
  g5 = 2.0/(g+1.0);
  g6 = (g-1.0)/(g+1.0);
  g7 = (g-1.0)/2.0;
  
  return(noErr);
}


void zlsr_RiemanSample(Real *p0, Real *u0, Real *r0, Real s, Wstar1DArr wStar, Integer i ){
  //
  //     Riemann solution sampler at speed = s
  //     returns p0, u0 and r0
  //     Use s = 0.0 for Eulerian Fluxes
  //
  Real shl, shr, c, sr, sl, pml, pmr, cml, cmr, stl , str;
	
  // cache arguments:	
	
  Real pm = wStar.P [i].a;
  Real um = wStar.v [i].a; 
  
  Real dl = wStar.d [i].al; 
  Real pl = wStar.P [i].al;
  Real ul = wStar.v [i].al; 
  Real cl = wStar.cs[i].al; 
	
  Real dr = wStar.d [i].ar; 
  Real pr = wStar.P [i].ar;
  Real ur = wStar.v [i].ar; 
  Real cr = wStar.cs[i].ar; 
	
  if (s<=um) {
    //
    //    to the left of the CD
    //      
    if (pm <= pl) {
      //  left rarefaction      
      shl = ul - cl;
      if (s <= shl) {
	//  in left state           
        *u0 = ul;
        *p0 = pl;
        *r0 = dl;
      } else {
        cml = cl*pow((pm/pl), g1);
        stl = um - cml;
        if (s>stl) {
	  //  in left star state           
          *r0 = dl*pow((pm/pl), (gInv));
          *p0 = pm;
          *u0 = um;
        }else{
	  //  in left fan state    
          *u0 = g5*(cl + g7*ul+s);
           c  = g5*(cl + g7*(ul-s));
          *r0 = dl*pow((c/cl), g4);
          *p0 = pl*pow((c/cl), g3);
        }
      }
    } else {
      //  left shock
      pml = pm/pl;
      sl  = ul - cl*sqrt(g2*pml + g1);
      if (s<=sl) {
	//  in left state     
        *r0 = dl;
        *p0 = pl;
        *u0 = ul;  
      } else {
	//  in left star state
        *r0 = dl*(pml+g6)/(pml*g6+1.0);
        *u0 = um;
        *p0 = pm;
      }
    }
  } else{
    //
    //    to the right of the CD
    //      
    if(pm>pr) {
      //   right shock
      pmr = pm/pr;
      sr  = ur + cr*sqrt(g2*pmr + g1);
      if (s>=sr){
	//   in right state
        *r0 = dr;
        *u0 = ur;
        *p0 = pr; 
      } else {
	//   in right star state
        *r0 = dr*(pmr + g6)/(pmr*g6 + 1.0);
        *u0 = um;
        *p0 = pm;
      }
    } else {
      //   right rarefaction
      shr = ur + cr;
      if (s>=shr) {
	//   in right state
        *r0 = dr;
        *u0 = ur;
        *p0 = pr;
      } else {
        cmr =cr*pow((pm/pr),g1);
        str = um + cmr; 
        if (s<=str) {
	  //  in right star state
          *r0 = dr*pow((pm/pr),(gInv));
          *u0 = um;
          *p0 = pm;
        } else {
	  //  in right fan state    
          *u0 = g5*(-cr + g7*ur+s);
           c  = g5*( cr - g7*(ur-s));
          *r0 = dr*pow((c/cr),g4);
          *p0 = pr*pow((c/cr),g3);
        }
      }
    }
  }
	
}

//====================================================================================
//
// Riemann Solvers, first up is Toro 1999 ideal Euler HD solver based on a pressure
// function to get p* and v*.  Pretty good for the weak limit, can take a lot of 
// iterations when the pressure difference becomes large.
//
//====================================================================================


const Real      kZLS_RiemannMinError = 1.0e-10;
const Integer   kZLS_RiemannMaxIter  = 20;


#ifdef TORO
 

void initialP( Real *pinit , 	
               Real pl, Real dl, Real ul, Real cl,
               Real pr, Real dr, Real ur, Real cr );

void fPl(Real *fr, Real *frd, Real ps, Real pl, Real dl, Real cl);
void fPr(Real *fr, Real *frd, Real ps, Real pr, Real dr, Real cr);

//
// ################ BEGIN Private Solver routines ################################
// no public interfaces for these routines, used by the TORO Exact solver
// names are lowercase for this reason.
//
 void initialPW( Real *pinit ,  W Wl, W Wr ){
	
	
  Real Q, cup, ppv, pmin, pmax;
  Real qmax, pq, um, ptl, ptr;
  Real gel, ger;
	
  Q       = 2.0;
  cup     = 0.25*(Wl.d+Wr.d)*(Wl.cs+Wr.cs);
  ppv     = 0.5*((Wl.P+Wr.P) + (Wl.u-Wr.u)*cup);
	
  ppv     = zls_max(0.0, ppv);
  pmin    = zls_min(Wl.P, Wr.P);
  pmax    = zls_max(Wl.P, Wr.P);
  qmax    = pmax/pmin;
	
  *pinit  = 0.50*(Wl.P+Wr.P);
	
  if ((qmax<=Q)&&(pmin<=ppv)&&(ppv<=pmax)) {
    *pinit = ppv;
  } else {
    if (ppv<pmin) {
      pq        = pow((Wl.P/Wr.P),g1);
      um        = (pq*Wl.u/Wl.cs + Wr.u/Wr.cs + g4*(pq-1.0))/(pq/Wl.cs + 1.0/Wr.cs);
      ptl       = 1.0 + g7*(Wl.u-um)/Wl.cs;
      ptr       = 1.0 + g7*(um-Wr.u)/Wr.cs;
      *pinit    = 0.5*(Wl.P*pow(ptl,g3) + Wr.P*pow(ptr,g3));
    } else {
      gel       = sqrt((g5/Wl.d)/(g6*Wl.P+ppv));
      ger       = sqrt((g5/Wr.d)/(g6*Wr.P+ppv));
      *pinit    = (gel*Wl.P+ger*Wr.P - (Wr.u-Wl.u))/(gel+ger);
    }
  }
	
}


 void fPW(Real * f, Real * fd, Real ps, W Wk)  {
  Real prat, ak, bk, qrt;
  //      
  if (ps < Wk.P) {
    //
    //	rarefaction
    //
    prat = ps/Wk.P;
    *f  = g4*Wk.cs*(pow(prat,g1) - 1.0);
    *fd =(1.0/(Wk.d*Wk.cs))* pow(prat, -g2);
  } else {
    //
    //   shock
    //
    ak = g5/Wk.d;
    bk = g6*Wk.P;
    qrt = sqrt(ak/(bk + ps));
    *f  = (ps-Wk.P)*qrt;
    *fd = (1.0 - 0.5*(ps-Wk.P)/(bk+ps))*qrt;
  }
	
}


 void initialP( Real *pinit , 	
                      Real pl, Real dl, Real ul, Real cl,
                      Real pr, Real dr, Real ur, Real cr ){
	
	
  Real Q, cup, ppv, pmin, pmax;
  Real qmax, pq, um, ptl, ptr;
  Real gel, ger;
	
  Q       = 2.0;
  cup     = 0.25*(dl+dr)*(cl+cr);
  ppv     = 0.5*((pl+pr) + (ul-ur)*cup);
	
  ppv     = zls_max(0.0, ppv);
  pmin    = zls_min(pl, pr);
  pmax    = zls_max(pl, pr);
  qmax    = pmax/pmin;
	
  *pinit  = 0.50*(pr+pl);
	
  if ((qmax<=Q)&&(pmin<=ppv)&&(ppv<=pmax)) {
    *pinit = ppv;
  } else {
    if (ppv<pmin) {
      pq        = pow((pl/pr),g1);
      um        = (pq*ul/cl + ur/cr + g4*(pq-1.0))/(pq/cl + 1.0/cr);
      ptl       = 1.0 + g7*(ul-um)/cl;
      ptr       = 1.0 + g7*(um-ur)/cr;
      *pinit    = 0.5*(pl*pow(ptl,g3) + pr*pow(ptr,g3));
    } else {
      gel       = sqrt((g5/dl)/(g6*pl+ppv));
      ger       = sqrt((g5/dr)/(g6*pr+ppv));
      *pinit    = (gel*pl+ger*pr - (ur-ul))/(gel+ger);
    }
  }
	
}

 void fPl(Real * f, Real * fd, Real ps, Real pk, Real rk, Real ck)  {
  Real prat, ak, bk, qrt;
  //      
  if (ps < pk) {
    //
    //	rarefaction
    //
    prat = ps/pk;
    *f  = g4*ck*(pow(prat,g1) - 1.0);
    *fd =(1.0/(rk*ck))* pow(prat, -g2);
  } else {
    //
    //   shock
    //
    ak = g5/rk;
    bk = g6*pk;
    qrt = sqrt(ak/(bk + ps));
    *f  = (ps-pk)*qrt;
    *fd = (1.0 - 0.5*(ps-pk)/(bk+ps))*qrt;
  }
	
}


 void fPr(Real * f, Real * fd, Real ps, Real pk, Real rk, Real ck){
  Real prat, ak, bk, qrt;
  
  //      
  if (ps < pk) {
    //
    //	rarefaction
    //
    prat = ps/pk;
    *f  = g4*ck*(pow(prat,g1) - 1.0);
    *fd = (1.0/(rk*ck))* pow(prat, -g2);
  } else {
    //
    //   shock
    //
    ak = g5/rk;
    bk = g6*pk;
    qrt = sqrt(ak/(bk + ps));
    *f  = (ps-pk)*qrt;
    *fd = (1.0 - 0.5*(ps-pk)/(bk+ps))*qrt;
  }	
}

// ################ END Private Solver routines ####################################

int zlsr_ToroRiemannW1D(Wstar1DArr wStar,  W1DArr wLR){
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = vleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = vright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    v*mid = CD velocity
	
	
  int err = noErr;
	
  Integer i, is, ie, iter;
  Integer nti;
  Real iter_err;
  Real dv;
	
  Real  fl, fr, fld, frd;
	
  is  = 1;
  ie  = wLR.nt-1;
  nti = wLR.nt;
	
  // left-right quantities
  W  Wl;
  W  Wr;
	
  // Riemann vars	
  Real newtonPstar;
	
  Real umid; // velocity
  Real Pmid; // Pressure
	
  wStar.v[0].a  = 0.0;
  wStar.v[nti-1].a  = 0.0;
	
  for (i = is; i < ie; i++) {
    
    Wl.d = wLR.d[i-1].ar;
    Wr.d = wLR.d[i].al;  
		
    Wl.P = wLR.P[i-1].ar;
    Wr.P = wLR.P[i].al;
		
    Wl.u = wLR.vx[i-1].ar;
    Wr.u = wLR.vx[i].al;

    Wl.cs = wLR.cs[i-1].ar; // shuffle left-right while copying to a local cache
    Wr.cs = wLR.cs[i].al;
    
      
      initialPW(&newtonPstar, Wl, Wr);
      newtonPstar = zls_max(kZLS_Epsilon, newtonPstar); 
      dv          = Wr.u - Wl.u;
      
      iter_err = 1.0;
      for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
        
        Pmid = newtonPstar;
        
        fPW(&fl, &fld, newtonPstar, Wl);
        fPW(&fr, &frd, newtonPstar, Wr);
        
        newtonPstar   -= (fl + fr + dv)/(fld  + frd);
        newtonPstar    = zls_max(kZLS_Epsilon, newtonPstar);
        iter_err       = zls_abs(1.0-Pmid/newtonPstar);
        
      } // iter
      
      Pmid  = newtonPstar;
      fPW(&fl, &fld, Pmid, Wl);
      fPW(&fr, &frd, Pmid, Wr);
      umid  = 0.50*(Wl.u + Wr.u + fr - fl) ;
      
    // save solution		
    wStar.P[i].a   = Pmid;
    wStar.P[i].al  = Wl.P;
    wStar.P[i].ar  = Wr.P;
		
    wStar.v[i].a   = umid;
    wStar.v[i].al  = Wl.u;
    wStar.v[i].ar  = Wr.u;
		
//    wStar.d[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region densities if needed.
//    wStar.d[i].al  = Wl.d;
//    wStar.d[i].ar  = Wr.d;
		
//    wStar.cs[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region cs if needed.
//    wStar.cs[i].al  = Wl.cs;
//    wStar.cs[i].ar  = Wr.cs;
		
  } // i
	
  return (err);   
}
//
// Single Riemann Problem interface
//

int zlsr_SingleRiemannW1D(Real *P, Real * v, W Wl, W Wr){
	
  // Input: Reimann left and right cell states , P, d, v
	
  // Output: Reimann star region in  wStar: P, v
	
  int err = noErr;
	
  Integer iter;
  Real iter_err;
  Real p0, dv;
	
  Real  fl, fr, fld, frd;
	
  Real newtonPstar;
  Wl.cs = sqrt(kZLS_eos_gamma * Wl.P/Wl.d);
  Wr.cs = sqrt(kZLS_eos_gamma * Wr.P/Wr.d);
	
  initialPW(&newtonPstar, Wl,  Wr);
  newtonPstar = zls_max(kZLS_Epsilon, newtonPstar); 
  dv          = Wr.u - Wl.u;
  
  iter_err = 1.0;
  for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
		
    p0 = newtonPstar;
		
    fPW(&fl, &fld, newtonPstar, Wl);
    fPW(&fr, &frd, newtonPstar, Wr);
		
    newtonPstar    -= (fl + fr + dv)/(fld  + frd);
    newtonPstar    = zls_max(kZLS_Epsilon, newtonPstar);
    iter_err       = zls_abs(1.0-newtonPstar/p0);
    zlsio_Printf( " Pmid %g NewtP %g iter_err %g \n"
           , p0, newtonPstar, iter_err);
		
  } // iter
	
  *P  = newtonPstar;
	
  fPW(&fl, &fld, newtonPstar, Wl);
  fPW(&fr, &frd, newtonPstar, Wr);
	
  *v = 0.50*(Wl.u + Wr.u + fr - fl) ;			
	
  return (err);   
}

#endif TORO

#ifdef TOROTWOSHOCK
 
int zlsr_InitToroSSRiemannGammas(Real g){

  g0 = (g);
  g1 = (g-1.0)/(2.0*g);
  g2 = (g+1.0)/(2.0*g);
  g3 = (2.0*g)/(g-1.0);
  g4 = 2.0/(g-1.0);
  g5 = 2.0/(g+1.0);
  g6 = (g-1.0)/(g+1.0);
  g7 = (g-1.0)/2.0;
  
  return(noErr);
}

//
 void initialPWS( Real *pinit ,  W Wl, W Wr ){
	
	
  Real Q, cup, ppv, pmin, pmax;
  Real qmax, pq, um, ptl, ptr;
  Real gel, ger;
	
  Q       = 2.0;
  cup     = 0.25*(Wl.d+Wr.d)*(Wl.cs+Wr.cs);
  ppv     = 0.5*((Wl.P+Wr.P) + (Wl.u-Wr.u)*cup);
	
  ppv     = zls_max(0.0, ppv);
  pmin    = zls_min(Wl.P, Wr.P);
  pmax    = zls_max(Wl.P, Wr.P);
  qmax    = pmax/pmin;
	
  *pinit  = 0.50*(Wl.P+Wr.P);
	
  if ((qmax<=Q)&&(pmin<=ppv)&&(ppv<=pmax)) {
    *pinit = ppv;
  } else {
    if (ppv<pmin) {
      pq        = pow((Wl.P/Wr.P),g1);
      um        = (pq*Wl.u/Wl.cs + Wr.u/Wr.cs + g4*(pq-1.0))/(pq/Wl.cs + 1.0/Wr.cs);
      ptl       = 1.0 + g7*(Wl.u-um)/Wl.cs;
      ptr       = 1.0 + g7*(um-Wr.u)/Wr.cs;
      *pinit    = 0.5*(Wl.P*pow(ptl,g3) + Wr.P*pow(ptr,g3));
    } else {
      gel       = sqrt((g5/Wl.d)/(g6*Wl.P+ppv));
      ger       = sqrt((g5/Wr.d)/(g6*Wr.P+ppv));
      *pinit    = (gel*Wl.P+ger*Wr.P - (Wr.u-Wl.u))/(gel+ger);
    }
  }
	
}


 void fPWS(Real * f, Real * fd, Real ps, W Wk)  {
  Real ak, bk, qrt;
  //      
  //
  //   shock
  //
  ak = g5/Wk.d;
  bk = g6*Wk.P;
  qrt = sqrt(ak/(bk + ps));
  *f  = (ps-Wk.P)*qrt;
  *fd = (1.0 - 0.5*(ps-Wk.P)/(bk+ps))*qrt;
	
}

int zlsr_ToroSSRiemannW1D(Wstar1DArr wStar,  W1DArr wLR){
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = vleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = vright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    v*mid = CD velocity
	
	
  int err = noErr;
	
  Integer i, is, ie, iter;
  Integer nti;
  Real iter_err;
  Real dv;
	
  Real  fl, fr, fld, frd;
	
  is  = 1;
  ie  = wLR.nt-1;
  nti = wLR.nt;
	
  // left-right quantities
  W  Wl;
  W  Wr;
	
  // Riemann vars	
  Real newtonPstar;
	
  Real umid; // velocity
  Real Pmid; // Pressure
	
  wStar.v[0].a  = 0.0;
  wStar.v[nti-1].a  = 0.0;
	
  for (i = is; i < ie; i++) {
    
		
    Wl.cs = wLR.cs[i-1].ar; // shuffle left-right while copying to a local cache
    Wr.cs = wLR.cs[i].al;
		
    Wl.d = wLR.d[i-1].ar;
    Wr.d = wLR.d[i].al;  
		
    Wl.P = wLR.P[i-1].ar;
    Wr.P = wLR.P[i].al;
		
    Wl.u = wLR.vx[i-1].ar;
    Wr.u = wLR.vx[i].al;
		
    initialPWS(&newtonPstar, Wl, Wr);
    newtonPstar = zls_max(kZLS_Epsilon, newtonPstar); 
    dv          = Wr.u - Wl.u;
		
    iter_err = 1.0;
    for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
			
      Pmid = newtonPstar;
			
      fPWS(&fl, &fld, newtonPstar, Wl); // two shock pressure function
      fPWS(&fr, &frd, newtonPstar, Wr);
			
      newtonPstar   -= (fl + fr + dv)/(fld  + frd);
      newtonPstar    = zls_max(kZLS_Epsilon, newtonPstar);
      iter_err       = zls_abs(1.0-Pmid/newtonPstar);
			
    } // iter
		
    Pmid  = newtonPstar;
    fPW(&fl, &fld, Pmid, Wl); // full pressure function
    fPW(&fr, &frd, Pmid, Wr); // full pressure function
    umid  = 0.50*(Wl.u + Wr.u + fr - fl) ;
		
    // save solution		
    wStar.P[i].a   = Pmid;
    wStar.P[i].al  = Wl.P;
    wStar.P[i].ar  = Wr.P;
		
    wStar.v[i].a   = umid;
    wStar.v[i].al  = Wl.u;
    wStar.v[i].ar  = Wr.u;
		
    wStar.d[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region densities if needed.
    wStar.d[i].al  = Wl.d;
    wStar.d[i].ar  = Wr.d;
		
    wStar.cs[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region cs if needed.
    wStar.cs[i].al  = Wl.cs;
    wStar.cs[i].ar  = Wr.cs;
		
  } // i
	
  return (err);   
}

#endif TOROTWOSHOCK

#ifdef PVRS
//
// A fast linearised primitive variable non-iterative approximate solver, PVRS
// Toro Chpter 9
//
int zlsr_PVRSRiemannW1D(Wstar1DArr wStar,  W1DArr wLR){
  int err = noErr;
	
  // use only away from strong shocks or rarefactions, intended for use in a
  // hybrid scheme in combination with the exact solver.
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = vleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = vright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    v*mid = CD velocity
	
	
  Real Cl, Cr, f;
  Integer i, is, ie;
	
  is  = 1;
  ie  = wLR.nt-1;
	
  wStar.v[0].a  = 0.0;
  wStar.v[ie].a  = 0.0;
	
  for (i = is; i < ie; i++) {
    
		
    wStar.d[i].al  = wLR.d[i-1].ar;
    wStar.d[i].ar  = wLR.d[i].al;
		
    wStar.P[i].al  = wLR.P[i-1].ar;
    wStar.P[i].ar  = wLR.P[i].al;
		
    wStar.v[i].al  = wLR.vx[i-1].ar;
    wStar.v[i].ar  = wLR.vx[i].al;
		
    wStar.cs[i].al  =  wLR.cs[i-1].ar;
    wStar.cs[i].ar  =  wLR.cs[i].al;
    
    Cl  = wStar.d[i].al*wStar.cs[i].al; // lagrangian sound speed
    Cr  = wStar.d[i].ar*wStar.cs[i].ar; // lagrangian sound speed
		
    f   = 1.0/(Cl + Cr + kZLS_Epsilon);
		
    wStar.P[i].a = f*( Cr*wStar.P[i].al + Cl*wStar.P[i].ar + Cl*Cr*(wStar.v[i].al-wStar.v[i].ar));
    wStar.v[i].a = f*( Cl*wStar.v[i].al + Cr*wStar.v[i].ar + (wStar.P[i].al-wStar.P[i].ar));
    
  }
	
  return(err);
		
}
#endif


#ifdef AIRS
 
int zlsr_InitAIRSRiemannGammas(Real g){

  g0 = (g);
  g1 = (g-1.0)/(2.0*g);
  g2 = (g+1.0)/(2.0*g);
  g3 = (2.0*g)/(g-1.0);
  g4 = 2.0/(g-1.0);
  g5 = 2.0/(g+1.0);
  g6 = (g-1.0)/(g+1.0);
  g7 = (g-1.0)/2.0;
  
  return(noErr);
}

void initialP( Real *pinit , 	
               Real pl, Real dl, Real ul, Real cl,
               Real pr, Real dr, Real ur, Real cr );

void fPl(Real *fr, Real *frd, Real ps, Real pl, Real dl, Real cl);
void fPr(Real *fr, Real *frd, Real ps, Real pr, Real dr, Real cr);

 void initialP( Real *pinit , 	
                      Real pl, Real dl, Real ul, Real cl,
                      Real pr, Real dr, Real ur, Real cr ){
	
	
  Real Q, cup, ppv, pmin, pmax;
  Real qmax, pq, um, ptl, ptr;
  Real gel, ger;
	
  Q       = 2.0;
  cup     = 0.25*(dl+dr)*(cl+cr);
  ppv     = 0.5*((pl+pr) + (ul-ur)*cup);
	
  ppv     = zls_max(0.0, ppv);
  pmin    = zls_min(pl, pr);
  pmax    = zls_max(pl, pr);
  qmax    = pmax/pmin;
	
  *pinit  = 0.50*(pr+pl);
	
  if ((qmax<=Q)&&(pmin<=ppv)&&(ppv<=pmax)) {
    *pinit = ppv;
  } else {
    if (ppv<pmin) {
      pq        = pow((pl/pr),g1);
      um        = (pq*ul/cl + ur/cr + g4*(pq-1.0))/(pq/cl + 1.0/cr);
      ptl       = 1.0 + g7*(ul-um)/cl;
      ptr       = 1.0 + g7*(um-ur)/cr;
      *pinit    = 0.5*(pl*pow(ptl,g3) + pr*pow(ptr,g3));
    } else {
      gel       = sqrt((g5/dl)/(g6*pl+ppv));
      ger       = sqrt((g5/dr)/(g6*pr+ppv));
      *pinit    = (gel*pl+ger*pr - (ur-ul))/(gel+ger);
    }
  }
	
}

 void fPl(Real * f, Real * fd, Real ps, Real pk, Real rk, Real ck)  {
  Real prat, ak, bk, qrt;
  //      
  if (ps < pk) {
    //
    //	rarefaction
    //
    prat = ps/pk;
    *f  = g4*ck*(pow(prat,g1) - 1.0);
    *fd =(1.0/(rk*ck))* pow(prat, -g2);
  } else {
    //
    //   shock
    //
    ak = g5/rk;
    bk = g6*pk;
    qrt = sqrt(ak/(bk + ps));
    *f  = (ps-pk)*qrt;
    *fd = (1.0 - 0.5*(ps-pk)/(bk+ps))*qrt;
  }
	
}


 void fPr(Real * f, Real * fd, Real ps, Real pk, Real rk, Real ck){
  Real prat, ak, bk, qrt;
  
  //      
  if (ps < pk) {
    //
    //	rarefaction
    //
    prat = ps/pk;
    *f  = g4*ck*(pow(prat,g1) - 1.0);
    *fd = (1.0/(rk*ck))* pow(prat, -g2);
  } else {
    //
    //   shock
    //
    ak = g5/rk;
    bk = g6*pk;
    qrt = sqrt(ak/(bk + ps));
    *f  = (ps-pk)*qrt;
    *fd = (1.0 - 0.5*(ps-pk)/(bk+ps))*qrt;
  }	
}

int zlsr_AIRSRiemannW1D (Wstar1DArr wStar,  W1DArr wLR, Real kQ){ // Hybrid Solver, set Q~ 2.0
  int err = noErr;
	
  // use only away from strong shocks or rarefactions, intended for use in a
  // hybrid scheme in combination with the exact solver.
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
  //
  // Switching parameter kQ.  
  // use PVRS if Pmax/Pmin = Q < kQ *and* p* from PVRS pmin<p*<pmax
  // otherwise use Exact solver.  Evaluated cell by cell, so the subroutines are not called
  // above, but are implemented in a mixed form here.
	
  //
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = vleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = vright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    v*mid = CD velocity
	
	
  Real   Cl, Cr,  f;
  Real  Q;
  Integer i, is, ie;
	
  Integer   iter;
  Integer    nti;
  Real   iter_err;
  Real   dv;
  Real   pmax, pmin;
	
  Real  fl, fr, fld, frd;
  // left-right quantities
  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real ul; // velocity
  Real ur;
  Real cl; // sound speeds
  Real cr;
	
  // Riemann vars	
  Real newtonPstar;
	
  Real umid; // velocity
  Real Pmid; // Pressure
	
  is  = 1;
  ie  = wLR.nt-1;
  nti = wLR.nt;
		
  wStar.v[0].a  = 0.0;
  wStar.v[nti-1].a  = 0.0;
	
  for (i = is; i < ie; i++) {
    
		
    cl = wLR.cs[i-1].ar; // 
    cr = wLR.cs[i].al;
		
    dl  = wLR.d[i-1].ar;
    dr = wLR.d[i].al;  
		
    Pl  = wLR.P[i-1].ar;
    Pr = wLR.P[i].al;
		
    ul  = wLR.vx[i-1].ar;
    ur = wLR.vx[i].al;
		
    pmax = zls_max(Pl, Pr);
    pmin = zls_min(Pl, Pr);
    Q    = pmax/(pmin+kZLS_Epsilon);
		
    Pmid = 0.0;
    umid = 0.0;
		
    if (Q < kQ) { // if the pressure ratios are small try PVRS
			
      Cl  = dl*cl; // lagrangian sound speed
      Cr  = dr*cr; // lagrangian sound speed
			
      f   = 1.0/(Cl + Cr + kZLS_Epsilon);
			
      Pmid = f*( Cr*Pl + Cl*Pr + Cl*Cr*(ul-ur));
      umid = f*( Cl*ul + Cr*ur + (Pl-Pr));
			
    }
		
    if (( Pmid < pmin) || ( Pmid > pmax)){
			
      // otherwise do full solution
			
      initialP(&newtonPstar, Pl, dl, ul, cl,  Pr, dr, ur, cr);
      newtonPstar = zls_max(kZLS_Epsilon, newtonPstar); 
      dv             = ur - ul;
      
      iter_err = 1.0;
      for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
				
        Pmid = newtonPstar;
				
        fPl(&fl, &fld, newtonPstar, Pl, dl, cl);
        fPr(&fr, &frd, newtonPstar, Pr, dr, cr);
				
        newtonPstar   -= (fl + fr + dv)/(fld  + frd);
        newtonPstar    = zls_max(kZLS_Epsilon, newtonPstar);
        iter_err       = zls_abs(newtonPstar-Pmid)/newtonPstar;
				
      } // iter
			
      Pmid  = newtonPstar;
      fPl(&fl, &fld, Pmid, Pl, dl, cl);
      fPr(&fr, &frd, Pmid, Pr, dr, cr);
      umid  = 0.50*(ul + ur + fr - fl) ;
      
    } // if
		
    // save solution		
    wStar.P[i].a   = Pmid;
    wStar.P[i].al  = Pl;
    wStar.P[i].ar  = Pr;
		
    wStar.v[i].a   = umid;
    wStar.v[i].al  = ul;
    wStar.v[i].ar  = ur;
		
    wStar.d[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region densities if needed.
    wStar.d[i].al  = dl;
    wStar.d[i].ar  = dr;
		
    wStar.cs[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region cs if needed.
    wStar.cs[i].al  = cl;
    wStar.cs[i].ar  = cr;
    
  } // i
	
  return(err);
}
#endif

// ####-The Colella 1984 interative approximate two shock solver ############################-
#ifdef COLLELLATWOSHOCK

 Real k_eos_gamfac0 = (ZLS_EOS_GAMMA);
 Real k_eos_gamfac1 = (0.5*((ZLS_EOS_GAMMA)+1.0)/(ZLS_EOS_GAMMA));
 Real k_eos_gamfac2 = ((ZLS_EOS_GAMMA) + 1.0);


int zlsr_InitC84RiemannGammas(Real g){

  k_eos_gamfac1 = (0.5*(g+1.0)/g);
  k_eos_gamfac2 = (g + 1.0);
  k_eos_gamfac0 = g;

  return(noErr);
}

  void initialPCW( Real *pinit , 	
                  Real pl, Real dl, Real ul,
                 Real pr, Real dr, Real ur);

  void initialPCW( Real *pinit , 	
                   Real pl, Real dl, Real ul,
                   Real pr, Real dr, Real ur){
    
  Real cl = zls_sqrt(k_eos_gamfac0*pl*dl);
  Real cr = zls_sqrt(k_eos_gamfac0*pr*dr);
  
  Real Ps = pr - pl - cr*(ur-ul);
  Ps = pl + Ps * cl/(cl+cr);
  Ps = zls_max(kZLS_Epsilon,Ps);
  
  *pinit = Ps;
  
}

int zlsr_C84RiemannW1D  (Wstar1DArr wStar,  W1DArr wLR){
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
  
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = uleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = uright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    u*mid = CD velocity
	
	
  int err = noErr;
	
  Integer i, is, ie;
  Integer nti;
  Integer iter;
  Real iter_err;
	
  is  = 1;
  ie  = wLR.nt-1;
  nti = wLR.nt;
  
  // left-right quantities
	
  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real ul; // velocity
  Real ur;
  Real cl; // sound speeds
  Real cr;
  Real vl; // specific volume 1/d
  Real vr;
  
  // Riemann vars	
  Real newtonPstar;
  Real wl; // wave speeds
  Real wr;
	
  Real zl; // u-P slopes
  Real zr;
  
  Real umidl; 
  Real umidr;
  Real umid = 0.0;
  Real Pmid = 0.0;
  
  
  wStar.v[0].a  = 0.0;
  wStar.v[nti-1].a  = 0.0;
	
  for (i = is; i < ie; i++) {
    
    // use the more elaborate Toro initial guess, using Eulerian system
    
    cl  = wLR.cs[i-1].ar; // 
    cr = wLR.cs[i].al;
    
    dl  = wLR.d[i-1].ar;
    dr = wLR.d[i].al;  
		
    Pl  = wLR.P[i-1].ar;
    Pr = wLR.P[i].al;
		
    ul  = wLR.vx[i-1].ar;
    ur = wLR.vx[i].al;
		

    initialPCW(&newtonPstar, Pl, dl, ul,  Pr, dr, ur);
    newtonPstar = zls_max(kZLS_PressureVacuum, newtonPstar); 
    
    cl  *=  dl; // cvt to sqrt(gam P d) Lagrangian sound speed
    cr  *=  dr;
    
    vl  = 1.0/dl;  // Specific Volume
    vr  = 1.0/dr;   // Specific Volume
    
    // the van Leer search
    iter_err = 1.0;
    for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
			
      Pmid   = newtonPstar;
			
      wl = 1.0 + k_eos_gamfac1*(newtonPstar - Pl) / Pl;   
      wr = 1.0 + k_eos_gamfac1*(newtonPstar - Pr) / Pr;
      wl = cl * zls_sqrt(wl);      
      wr = cr * zls_sqrt(wr);   
			
      zl = 4.0 * vl * wl * wl;      
      zr = 4.0 * vr * wr * wr;
			
      zl = -zl * wl/(zl - k_eos_gamfac2*(newtonPstar - Pl));      
      zr =  zr * wr/(zr - k_eos_gamfac2*(newtonPstar - Pr));
			
      umidl = ul - (newtonPstar - Pl) / wl ;
      umidr = ur + (newtonPstar - Pr) / wr ;
			
      newtonPstar += (umidr - umidl) * (zl * zr)/(zr - zl);  
			
      newtonPstar    = zls_max(kZLS_PressureVacuum, newtonPstar);
      iter_err       = zls_abs(1.0-(newtonPstar/Pmid));
			
    } // iter
    
    Pmid  = newtonPstar;
    umidl = ul - (Pmid - Pl) / wl;      
    umidr = ur + (Pmid - Pr) / wr;   

    umid  = 0.5*(umidl + umidr); 
    
    
    wStar.P[i].a   = Pmid;
    wStar.P[i].al  = Pl;
    wStar.P[i].ar  = Pr;
		
    wStar.v[i].a   = umid;
    wStar.v[i].al  = ul;
    wStar.v[i].ar  = ur;
    		
   // wStar.d[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region densities if needed.
   // wStar.d[i].al  = dl;
   // wStar.d[i].ar  = dr;
		
   // wStar.cs[i].a  = 0.0; // use the Toro sample routine to get the internal left right star region cs if needed.
   // wStar.cs[i].al  = cl/dl;
   // wStar.cs[i].ar  = cr/dr;
  } // i
	
  return (err);   
}

//
void initialPWCW( Real *pinit ,  W Wl, W Wr );
  void initialPWCW( Real *pinit ,  W Wl, W Wr ){
  
  Real cl = zls_sqrt(k_eos_gamfac0*Wl.P*Wl.d);
  Real cr = zls_sqrt(k_eos_gamfac0*Wr.P*Wr.d);
  
  Real Ps = Wr.P - Wl.P - cr*(Wr.u-Wl.u);
  Ps = Wl.P + Ps * cl/(cl+cr);
  Ps = zls_max(kZLS_Epsilon,Ps);
  
  *pinit  = Ps;
	
}

int zlsr_TestC84RiemannW1D(Real *P, Real * v, W Wl, W Wr){
  int err = noErr;
	
  Integer iter;
  Real iter_err;
	
  // Riemann vars	
  Real newtonPstar;
  Real cl; // lagrangian sound speeds
  Real cr;
  Real vl; // specific volume
  Real vr;
	
  Real sl; // shock wave lagrangian speeds
  Real sr;
	
  Real zl; // u-P slopes
  Real zr;
	
  Real umidl; 
  Real umidr;
  Real umid = 0.0;
  Real Pmid = 0.0;
  
  Wl.cs = sqrt(k_eos_gamfac0 * Wl.P/Wl.d);
  Wr.cs = sqrt(k_eos_gamfac0 * Wr.P/Wr.d);
	
  initialPWCW(&newtonPstar, Wl,  Wr);
  newtonPstar = zls_max(kZLS_Epsilon, newtonPstar); 
  
  cl = Wl.d* Wl.cs;
  cr = Wr.d* Wr.cs;	
  vl  = 1.0/Wl.d;   // Specific Volume
  vr  = 1.0/Wr.d;   // Specific Volume
	
  zlsio_Printf( " Pl %g Pr %g PN %g \n", Wl.P, Wr.P, newtonPstar);
	
  // the van Leer search
  iter_err = 1.0;
  for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
		
    Pmid   = newtonPstar;
		
    sl = cl * sqrt(1.0 + (k_eos_gamfac1*(newtonPstar - Wl.P) / Wl.P));      
    sr = cr * sqrt(1.0 + (k_eos_gamfac1*(newtonPstar - Wr.P) / Wr.P));   
		
    zl = 4.0 * vl * sl * sl;      
    zr = 4.0 * vr * sr * sr;
		
    zl *= -sl/(zl - k_eos_gamfac2*(newtonPstar - Wl.P));      
    zr *=  sr/(zr - k_eos_gamfac2*(newtonPstar - Wr.P));
		
    umidl = Wl.u - (newtonPstar - Wl.P) / sl ;
    umidr = Wr.u + (newtonPstar - Wr.P) / sr ;
		
    newtonPstar += (umidr - umidl) * (zl * zr)/(zr - zl);  
		
    newtonPstar = zls_max(kZLS_Epsilon, newtonPstar);
    iter_err    = zls_abs(1.0 - Pmid/newtonPstar);
		
    zlsio_Printf( " Pmid %g NewtP %g umidl %g umidr %g iter_err %g \n"
           , Pmid, newtonPstar, umidl, umidr, iter_err);
		
  } // iter
	
  Pmid  = newtonPstar;
  umidl = Wl.u - (Pmid - Wl.P) / sl;      
  umidr = Wr.u + (Pmid - Wr.P) / sr;   
  umid  = 0.5*(umidl + umidr);
	
  *P = Pmid;
  *v = umid;
	
  return (err);   
}
#endif

#ifdef GROTHGOTTLIEB

// ################ The Groth and Gottlieb (1988) Exact Riemann Solver ################
//
// EOS constants for the GG solver, here assumes homogeneous gammas
//
// Not all are used when only a two shock aproximation are used and a_star is not
// reconstructed.
//
//


 Real kZLS_gg_Alpha  = ((ZLS_EOS_GAMMA+1.0)/4.0);
 Real kZLS_gg_Beta   = ((ZLS_EOS_GAMMA-1.0)/2.0);
 Real kZLS_gg_Beta2  = (2.0/(ZLS_EOS_GAMMA-1.0)); // inverse

 Real kZLS_gg_Gamma0 = (ZLS_EOS_GAMMA);
 Real kZLS_gg_Gamma1 = (ZLS_EOS_GAMMA+1.0);
 Real kZLS_gg_Gamma2 = (ZLS_EOS_GAMMA-1.0);

 Real kZLS_gg_Delta  = (2.0*ZLS_EOS_GAMMA)/(ZLS_EOS_GAMMA-1.0);
 Real kZLS_gg_Delta2 = (ZLS_EOS_GAMMA-1.0)/(2.0*ZLS_EOS_GAMMA); // inverse

int zlsr_InitGG88RiemannGammas(Real g){

  kZLS_gg_Alpha = ((g+1.0)/4.0);

  kZLS_gg_Beta  = ((g-1.0)/2.0);
  kZLS_gg_Beta2  = (2.0/(g-1.0)); // inverse

  kZLS_gg_Gamma0 = (g);
  kZLS_gg_Gamma1 = (g+1.0);
  kZLS_gg_Gamma2 = (g-1.0);

  kZLS_gg_Delta = (2.0*g)/(g-1.0);
  kZLS_gg_Delta2 = (g-1.0)/(2.0*g); // inverse
  
  return(noErr);
}

int zlsr_GG88RiemannW1D  (Wstar1DArr wStar,  W1DArr wLR){
  int err = noErr;
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = uleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = uright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    u*mid = CD velocity
	
	
  // a* is not computed for Pstar and Ustar, but can be solved later
	
	
  Integer i, is, ie;
  Integer nti;
	
  Integer iter;
  Real iter_err;
  
	
  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real vl; // velocity
  Real vr;
  
  Real z, ul, ur; // initial speed guess
  
  // Riemann vars	
  Real newtonUstar;
  Real newtonStep;
	
  Real Ustar = 0.0; 
  Real Pstar = 0.0;
	
  Real Pstarl, Pstarr;
  Real dPstarl, dPstarr;
  Real astarl, astarr;
	
  Real al, ar; // Eulerian sound speed
  Real inv_al, inv_ar; // Eulerian sound rates
  Real cl, cr; // Lagrangian sound speed
  Real sl, sr; // intermediate wave slopes
  Real zl, zr; // shock mach number
  
  Real alphal; // (gamma+1)/(4a), predivide and collect constants
  Real alphar; // predivide and collect constants
	
  is  = 1;
  ie  = wLR.nt-1;
  nti = wLR.nt;
	
  wStar.v[0].a  = 0.0;
  wStar.v[nti-1].a  = 0.0;
	
  for (i = is; i < ie; i++) {
				
    dl  = wLR.d[i-1].ar;
    dr = wLR.d[i].al;  
		
    Pl  = wLR.P[i-1].ar;
    Pr = wLR.P[i].al;
		
    vl  = wLR.vx[i-1].ar;
    vr = wLR.vx[i].al;
		
    al = wLR.cs[i-1].ar;  // eulerian speed
    ar = wLR.cs[i].al;
    
    inv_al = 1.0/al;  // eulerian rates
    inv_ar = 1.0/ar;
    
    alphal = kZLS_gg_Alpha*inv_al; // (gamma+1)/(4a), predivide and collect constants
    alphar = kZLS_gg_Alpha*inv_ar; // predivide and collect constants
			
      cl = dl*al; // lagrangian speed
      cr = dr*ar;	
			
      // initial guess
			
      ul = vl + kZLS_gg_Beta2*al;
      ur = vr - kZLS_gg_Beta2*ar;
			
      z = (ar/al)*pow((Pl/Pr), kZLS_gg_Delta2);
      newtonUstar = (ul*z + ur)/(1.0+z);
			
      iter_err = 1.0;
      for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
				
        Ustar  = newtonUstar;
				
        if ( Ustar <= vl ) { // left wave is shock
					
          sl     = (newtonUstar - vl)*alphal;
          zl     = sl - sqrt(1.0+(sl*sl));
          Pstarl = Pl + cl*(newtonUstar - vl)*zl;
          dPstarl = 2*cl*zl*zl*zl/(1.0+zl*zl);
					
        } else { // left wave is rarefaction 
					
          astarl  = al - kZLS_gg_Beta*(newtonUstar - vl);
          Pstarl = Pl * pow(astarl*inv_al, kZLS_gg_Delta);
          dPstarl = -kZLS_gg_Gamma0*Pstarl/astarl;
					
        }
				
        if ( Ustar >= vr ) { // right wave is shock
					
          sr     = (newtonUstar - vr)*alphar;
          zr     = sr + sqrt(1.0+(sr*sr));
          Pstarr = Pr + cr*(newtonUstar - vr)*zr;
          dPstarr = 2*cr*zr*zr*zr/(1.0+zr*zr);
					
        } else { // right wave is rarefaction
					
          astarr  = ar + kZLS_gg_Beta*(newtonUstar - vr); // note +
          Pstarr = Pr * pow(astarr*inv_ar, kZLS_gg_Delta);
          dPstarr = kZLS_gg_Gamma0*Pstarr/astarr;
          
        }
				

        register Real deltaP     = ( Pstarl -  Pstarr);
        register Real deltaDP    = (dPstarl - dPstarr);
        newtonStep = deltaP/deltaDP;
        
        newtonUstar  -=  newtonStep; 
        iter_err      = zls_abs(newtonStep); // absolute error, faster as long as delta V is normalised

//        newtonStep   =  (Pstarl - Pstarr)/zls_tole((dPstarl - dPstarr), kZLS_Small);
//        newtonUstar  -= newtonStep;
//        iter_err    = zls_abs(newtonStep);
//
      } // iter
			
      Ustar  = newtonUstar;
      Pstar  = 0.5*(Pstarl+Pstarr);
			
    wStar.P[i].a   = Pstar; 
    wStar.P[i].al  = Pl;
    wStar.P[i].ar  = Pr;
		
    wStar.v[i].a   = Ustar; 
    wStar.v[i].al  = vl;
    wStar.v[i].ar  = vr;
				
  } // i
	
  return (err);   
}


int zlsr_TestGG88RiemannW1D(Real *P, Real * v, W Wl, W Wr){
  int err = noErr;
	
  // a* is not computed for Pstar and Ustar, but can be solved later
	
  //
  Integer iter;
  Real iter_err;
	
  // Riemann vars	
  Real newtonUstar;
	
  Real Ustar = 0.0; 
  Real Pstar = 0.0;
	
  Real Pstarl, Pstarr;
  Real dPstarl, dPstarr;
  Real astarl, astarr;
	
  Real al, ar; // Eulerian sound speed
  Real inv_al, inv_ar; // Eulerian sound speed
  Real alphal, alphar; // predivide and collect constants
  Real cl, cr; // Lagrangian sound speed
  Real sl, sr; // intermediate wave speeds
  Real zl, zr; // shock mach number
	
  al = sqrt(kZLS_gg_Gamma0 * Wl.P/Wl.d);  // eulerian speed
  ar = sqrt(kZLS_gg_Gamma0 * Wr.P/Wr.d);
	
  inv_al = 1.0/al;  // eulerian rate
  inv_ar = 1.0/ar;
	
  alphal = kZLS_gg_Alpha*inv_al; // (gamma+1)/(4a), predivide and collect constants
  alphar = kZLS_gg_Alpha*inv_ar; // predivide and collect constants
	
  cl = Wl.d*al;
  cr = Wr.d*ar;	
	
  Real z,ul, ur; // initial speed guess
	
  // initial guess
	
  ul = Wl.u + kZLS_gg_Beta2*al;
  ur = Wr.u - kZLS_gg_Beta2*ar;
	
  z = (ar/al)*pow((Wl.P/Wr.P), kZLS_gg_Delta2);
  newtonUstar = (ul*z + ur)/(1.0+z);
		
  iter_err = 1.0;
  for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
		
    Ustar  = newtonUstar;
		
    if ( Ustar <= Wl.u ) { // left wave is shock
			
      sl     = (newtonUstar - Wl.u)*alphal;
      zl     = sl - sqrt(1.0+(sl*sl));
      Pstarl = Wl.P + cl*(newtonUstar - Wl.u)*zl;
      dPstarl = 2*cl*zl*zl*zl/(1.0+zl*zl);
			
    } else { // left wave is rarefaction 
			
      astarl  = al - kZLS_gg_Beta*(newtonUstar - Wl.u);
      Pstarl = Wl.P * pow(astarl*inv_al, kZLS_gg_Delta);
      dPstarl = -kZLS_gg_Gamma0*Pstarl/astarl;
			
    }
		
    if ( Ustar >= Wr.u ) { // right wave is shock
			
      sr     = (newtonUstar - Wr.u)*alphar;
      zr     = sr + sqrt(1.0+(sr*sr));
      Pstarr = Wr.P + cr*(newtonUstar - Wr.u)*zr;
      dPstarr = 2*cr*zr*zr*zr/(1.0+zr*zr);
			
    } else { // right wave is rarefaction
			
      astarr  = ar + kZLS_gg_Beta*(newtonUstar - Wr.u); // note +
      Pstarr = Wr.P * pow(astarr*inv_ar, kZLS_gg_Delta);
      dPstarr = kZLS_gg_Gamma0*Pstarr/astarr;
			
    }
		
		
//    newtonUstar -= (Pstarl - Pstarr)/(dPstarl - dPstarr);
    register Real deltaP     = ( Pstarl -  Pstarr);
    register Real deltaDP    = (dPstarl - dPstarr);
    register Real newtonStep = deltaP/deltaDP;

    newtonUstar  -=  newtonStep; 
    iter_err      = zls_abs(newtonStep); // absolute error, faster as long as delta V is normalised
//
//    //iter_err    = zls_abs(1.0 - Ustar/newtonUstar);
//    iter_err    = zls_abs(1.0 - Pstarl/Pstarr);
//		
//    zlsio_Printf( " Ustar %g NewtU %g Pstarl %g Pstarr %g iter_err %g \n"
//           , Ustar, newtonUstar, Pstarl, Pstarr, iter_err);
  } // iter
	
  Ustar  = newtonUstar;
  Pstar  = 0.5*(Pstarl + Pstarr);
	
  *P = Pstar;
  *v = Ustar;
	
  return (err);   
}
#endif GROTHGOTTLIEB


#ifdef RSSTWOSHOCK

//#define RSS_FASTGUESS
//#define RSS_FASTGUESS2
//#define GG_INITIALGUESS
#define PVRSINITIALGUESS

 Real kZLS_gg_Alpha  = ((ZLS_EOS_GAMMA+1.0)/4.0);
 Real kZLS_gg_Beta1  = ((ZLS_EOS_GAMMA-1.0)/2.0);
 Real kZLS_gg_Beta2  = (2.0/(ZLS_EOS_GAMMA-1.0)); // inverse

#ifdef GG_INITIALGUESS
 Real kZLS_gg_Delta2 = (ZLS_EOS_GAMMA-1.0)/(2.0*ZLS_EOS_GAMMA); // inverse
#endif GG_INITIALGUESS

int zlsr_InitRSSRiemannGammas(Real g){

  kZLS_gg_Alpha = ((g+1.0)/4.0);

  kZLS_gg_Beta1 = ((g-1.0)/2.0);
  kZLS_gg_Beta2 = (2.0/(g-1.0)); // inverse

#ifdef GG_INITIALGUESS
  kZLS_gg_Delta2 = (g-1.0)/(2.0*g); // inverse
#endif GG_INITIALGUESS
  
  return(noErr);
}


int zlsr_RSSRiemannW1D  (Wstar1DArr wStar,  W1DArr wLR){ 
// includes robust Nan prevention and grid noise control
  int err = noErr;
	
  // Input: Reimann left and right cell states are in wLR:
  // wLR.P[i].al  = Pleft;  wLR.P[i].ar = Pright;  wLR.P[i].a ignored
  // wLR.d[i].al  = rleft;  wLR.d[i].ar = rright;  wLR.d[i].a ignored
  // wLR.vx[i].al  = uleft;  wLR.vx[i].ar = uright;  wLR.vx[i].a ignored
	
  // Output: Reimann star region in  wStar:
  // wStar.d[i].al  = dleft   =  wLR.d[i-1].ar
  // wStar.d[i].ar  = dright  =  wLR.d[i].al
	
  // wStar.P[i].al  = Pleft   =  wLR.P[i-1].ar
  // wStar.P[i].ar  = Pright  =  wLR.P[i].al
  // wStar.P[i].a   = solved    p*mid = CD presure
	
  // wStar.v[i].al  = uleft   =  wLR.vx[i-1].ar
  // wStar.v[i].ar  = uright  =  wLR.vx[i].al
  // wStar.v[i].a   = solved    u*mid = CD velocity
	
	
  // a* is not computed for Pstar and Ustar, but can be solved later
	
  //	
  // Modified Groth and Gottlieb solve to use an iterative two shock approximation
  // only - RSS 2005.  much faster as iterator has only multiplies and adds and no logic.  
  // Less divides than C84 solver, same two sqrts
  //
  // Tends to converge in fewer iterations than C84, ie 3 vs 4 for 100, 10, 0 : 1, 1, 0 test
  // and gives exactly the same answer.
  //
	
  Integer i, is, ie;
  Integer nti;
	
  Integer iter;
  Real iter_err = 1.0;

  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real vl; // velocity
  Real vr;
  
  Real ul, ur; // initial speed guess
#ifdef RSS_FASTGUESS2
  Real z;
#endif RSS_FASTGUESS2
#ifdef GG_INITIALGUESS
  Real z;
#endif GG_INITIALGUESS

//  Real f;
  Real Ustar = 0.0; 
  Real Pstar = 0.0;
	
  Real cl;
  Real cr; // Lagrangian sound speed
	
  Real al, ar; // Eulerian sound speed
  
  Real alphal; // (gamma+1)/(4a), predivide and collect constants
  Real alphar; // predivide and collect constants
	
  is  = 1; // zero cell has no left border cell
  ie  = wLR.nt;
  nti = wLR.nt;

  wStar.v[0].a  = wLR.vx[0].al;
  wStar.P[0].a  = wLR.P[0].al; 
  wStar.d[0].a  = wLR.d[0].al; 
	
  for (i = is; i < ie; i++) {
				
    
    dl = wLR.d[i-1].ar;
    dr = wLR.d[i  ].al;  
		
    Pl = wLR.P[i-1].ar;
    Pr = wLR.P[i  ].al;
    
    Real kZLS_Alpha_r = kZLS_gg_Alpha;
    Real kZLS_Alpha_l = kZLS_gg_Alpha;
    
    Real kZLS_Beta_r  = kZLS_gg_Beta2;
    Real kZLS_Beta_l  = kZLS_gg_Beta2;

#ifdef VARIABLE_EOS
#ifdef VARIABLE_GAMMA
    Real eosGamma_r = wLR.g [i-1].ar;
    Real eosGamma_l = wLR.g [i].al;
    
     kZLS_Alpha_r = ((eosGamma_r+1.0)*0.25);
     kZLS_Alpha_l = ((eosGamma_l+1.0)*0.25);
    
     kZLS_Beta_r  = (2.0/(eosGamma_r-1.0));
     kZLS_Beta_l  = (2.0/(eosGamma_l-1.0));
#endif VARIABLE_GAMMA
#endif VARIABLE_EOS

    
#ifdef GG_INITIALGUESS
    Real PRatio = Pl/Pr;
#endif GG_INITIALGUESS
		
    vl = wLR.vx[i-1].ar;
    vr = wLR.vx[i  ].al;
    al = wLR.cs[i-1].ar;  // eulerian adiabatic speed
    ar = wLR.cs[i  ].al;

    if ((zls_abs((Pl-Pr))> kZLS_Small )||(zls_abs((vl-vr))>kZLS_Small)){
      
      cl = dl*al; // lagrangian speed
      cr = dr*ar;	
      

      Pr  = zls_max( kZLS_PressureVacuum, Pr);
      Pl  = zls_max( kZLS_PressureVacuum, Pl);
      
      alphar = kZLS_Alpha_r/ar; // predivide and collect constants
      alphal = kZLS_Alpha_l/al; // (gamma+1)/(4a), predivide and collect constants
      ul     = vl + kZLS_Beta_l*al;
      ur     = vr - kZLS_Beta_r*ar;
      
    // initial guess
      
      register Real newtonUstar  = Ustar;

#ifdef GG_INITIALGUESS
      // G& G suggested initial guess, slow due to fp div and power,
        z = (ar/al)*pow( PRatio, kZLS_gg_Delta2);
        newtonUstar = (ul*z + ur)/(1.0+z);
#endif GG_INITIALGUESS
#ifdef RSS_FASTGUESS
      // RSS v rough  and fast initial guess, iterations are faster than power and divisions!!
      // could break with v large pressure diffs in a static setup.  OK for situations where 
      // supersonic velocity differences dominate throughout (common)
        newtonUstar = 0.5*(vr + vl);
      //  newtonUstar = vr;
#endif RSS_FASTGUESS
#ifdef RSS_FASTGUESS2
      // RSS 2 intermediate initial guess, since kZLS_gg_Delta2 is small, 
      // PRatio^kZLS_gg_Delta2 ~ 1 for a wide range around 0.01-100. slow for divisions
      // but safer for static or standing shocks/rws 
      z = (ar/al);
      newtonUstar = (ul*z + ur)/(1.0+z);
#endif RSS_FASTGUESS2
#ifdef PVRSINITIALGUESS
      //
      // RSS intial guess #3 , using TORO's PVRS solver, 
      // not accurate for shocks, but that is where the iterator
      // is best.  Really seems to cut down the interations needed overall.
      // which makes up for the time spent with a division.
      // If this proves robust it may become the default.
      //
      Real f      = 1.0/(cr + cl);
      newtonUstar = f*( cl*vl + cr*vr + (Pl-Pr)); // PVRS guess for ustar
	 //  Pstar = f*( cr*Pl + cl*Pr + cl*cr*(vl-vr));
#endif PVRSINITIALGUESS
    
      iter_err = 1.0;
      
      register Real Pstarl  = 0.0;
      register Real Pstarr  = 0.0;
      register Real dPstarl = 0.0;
      register Real dPstarr = 0.0;
      register Real newtonStep  = 0.0;
      register Real sl;
      register Real sr; // intermediate wave slopes
      register Real zl;
      register Real zr; // shock mach number

	    register Real deltaDP, deltaP ;
        
      for (iter = 0; ((iter < kZLS_RiemannMaxIter)&&( iter_err > kZLS_RiemannMinError)); iter++) {
        
        Ustar   = newtonUstar;
        
        sl      = (newtonUstar - vl)*alphal;
        zl      = sl - zls_sqrt(1.0+(sl*sl));
        Pstarl  = Pl + cl*(newtonUstar - vl)*zl;
        dPstarl = cl*zl*zl*zl/(1.0+zl*zl);
       
        sr      = (newtonUstar - vr)*alphar;
        zr      = sr + zls_sqrt(1.0+(sr*sr));
        Pstarr  = Pr + cr*(newtonUstar - vr)*zr;
        dPstarr = cr*zr*zr*zr/(1.0+zr*zr);
        
        deltaP     =     ( Pstarl -  Pstarr);
        deltaDP    =   2*(dPstarl - dPstarr);
        newtonStep = deltaP/deltaDP;
                           
        newtonUstar  -=  newtonStep; 
        iter_err      = zls_abs(newtonStep); // absolute error, faster as long as delta V is normalised
        
      } // iter
//
        if ( zls_isnan(newtonUstar)){// if iterations fail be diffusive rather than conservative
#ifdef PVRSINITIALGUESS
            Ustar = f*( cl*vl + cr*vr + (Pl-Pr)); // PVRS guess for ustar
            Ustar = zls_tolx(Ustar, kZLS_Small); 
            Pstar = f*( cr*Pl + cl*Pr + cl*cr*(vl-vr));
            Pstar  = zls_max( kZLS_PressureVacuum, Pstar);
#else
            Ustar = 0.5*(vr + vl);
            Pstar = 0.5*(Pr + Pl);
            Pstar  = zls_max( kZLS_PressureVacuum, Pstar);
#endif !PVRSINITIALGUESS
//          zlsio_Printf(" *Div 0*, iterations failed: Ustar %g Pstar %g Pl %g Pr %g vl %g vr %g its %d Err: %g\n", Ustar, Pstar, Pl, Pr, vl, vr, iter, iter_err);
        } else {
          if ( iter_err < 10.0*kZLS_RiemannMinError) {
            Ustar = zls_tolx(newtonUstar, kZLS_Small); 
            Pstar = (0.5*(Pstarl + Pstarr)); 
            Pstar  = zls_max( kZLS_PressureVacuum, Pstar);
        //zlsio_Printf(" Ustar %g Pstar %g its %d Err: %g\n", Ustar, Pstar, iter, iter_err/Ustar);
          } else { // if iterations fail: be diffusive rather than conservative
            Ustar = 0.5*(vr + vl);
            Pstar = 0.5*(Pr + Pl);
            Pstar  = zls_max( kZLS_PressureVacuum, Pstar);
//           zlsio_Printf("     iterations failed:  Avg Ustar %g Pstar %g Pl %g Pr %g vr %g vl %g dr %g dl %g pm %g px %g\n", Ustar, Pstar, Pl, Pr, vr, vl, dr, dl, 0.0, 0.0);
          }
        }
        
    } else { 
      //
      // uniform constant flow, bypass Riemann and pass values on for
      // the basic advection to handle.
      //
      // Here P* is exactly equal to the current cell(i) pressure 
      // and ditto vel, so evolve won't change values
      // modulo fp noise which evolve will handle.
      //
      Ustar = vr;
      Pstar = Pr;
      Pl    = Pr;
      vl    = vr;
      //
    }
    
    wStar.P[i].a   = Pstar; 
    wStar.P[i].al  = Pl;
    wStar.P[i].ar  = Pr;
    
    wStar.d[i].al  = dl;
    wStar.d[i].ar  = dr;
		
    wStar.v[i].a   = Ustar; 
    wStar.v[i].al  = vl;
    wStar.v[i].ar  = vr;

    wStar.cs[i].al  = al;
    wStar.cs[i].ar  = ar;
		
  } // i
	
  return (err);
  
}

#endif RSSTWOSHOCK

// ################ The Roe/Pike Approximate Generalised Riemann Solver ############################




// ################ Misc WStar structure handling routines ############################


int zlsr_PrintWstar1DArr(Wstar1DArr w, int cellIndex ){ // -1 for all.
	
  int err = noErr;
  SLRQ1   vlr;
  Integer i;
	
  if (cellIndex >= 0) {
    vlr = w.P[cellIndex];
    zlsio_Printf( " Index %4d %s: *%8.4f L%8.4f R%8.4f\n",cellIndex ,"P",
           vlr.a, vlr.al, vlr.ar );
    vlr = w.v[cellIndex];
    zlsio_Printf( " Index %4d %s: *%8.4f L%8.4f R%8.4f\n",cellIndex ,"v",
           vlr.a, vlr.al, vlr.ar );
    vlr = w.d[cellIndex];
    zlsio_Printf( " Index %4d %s: *%8.4f L%8.4f R%8.4f\n",cellIndex ,"d",
           vlr.a, vlr.al, vlr.ar );
    vlr = w.cs[cellIndex];
    zlsio_Printf( " Index %4d %s: *%8.4f L%8.4f R%8.4f\n",cellIndex ,"cs",
                  vlr.a, vlr.al, vlr.ar );
#ifdef VARIABLE_GAMMA
    vlr = w.g [cellIndex];
    zlsio_Printf( " Index %4d %s: *%8.4f L%8.4f R%8.4f\n",cellIndex ,"g ",
                  vlr.a, vlr.al, vlr.ar );
#endif VARIABLE_GAMMA
  } else { // loop whole array , could be big!
    for (i = 0; i< w.nt; i++) {
      zlsr_PrintWstar1DArr(w, i);
    }
  }
  return(err);
	
}

int zlsr_InitWstar1DArr(Wstar1DArr w ){
	
  int err = noErr;
  Integer i;
	
  for (i = 0; i< w.nt; i++){
    w.d[i].a = 0.0;
    w.d[i].al = 0.0;
    w.d[i].ar = 0.0;
		
    w.P[i].a = 0.0;
    w.P[i].al = 0.0;
    w.P[i].ar = 0.0;
		
    w.cs[i].a = 0.0;
    w.cs[i].al = 0.0;
    w.cs[i].ar = 0.0;
		
#ifdef VARIABLE_GAMMA
    w.g [i].a  = kZLS_eos_gamma;
    w.g [i].al = kZLS_eos_gamma;
    w.g [i].ar = kZLS_eos_gamma;
#endif VARIABLE_GAMMA

    w.v[i].a = 0.0;
    w.v[i].al = 0.0;
    w.v[i].ar = 0.0;
  }
  return(err);
	
}

int zlsr_NewWstar1DArr(Wstar1DArr *w, Integer nBC, Counter ni){
	
  int err = noErr;
	
  w->n  = ni;
  w->ns = nBC;
  w->ne = nBC+ni;
  w->nt = nBC+nBC+ni;
	
  w->x = NULL;  // get an axis from an owning grid
	
  w->d  = (SLRQ1*) malloc((nBC+nBC+ni)*sizeof(SLRQ1));
  w->P  = (SLRQ1*) malloc((nBC+nBC+ni)*sizeof(SLRQ1));
  w->cs = (SLRQ1*) malloc((nBC+nBC+ni)*sizeof(SLRQ1));
#ifdef VARIABLE_GAMMA
  w->g  = (SLRQ1*) malloc((nBC+nBC+ni)*sizeof(SLRQ1));
#endif VARIABLE_GAMMA
  w->v  = (SLRQ1*) malloc((nBC+nBC+ni)*sizeof(SLRQ1));
	
  zlsr_InitWstar1DArr( *w);
	
  return(err);
	
}


int zlsr_DisposeWstar1DArr  (Wstar1DArr w){
  int err = noErr;
  //
	// don't dispose x as it is a reference to an external structure
  // never allocated just for w
  //
  if (w.d !=NULL)  free((void *)w.d); w.d = NULL;
  if (w.P !=NULL)  free((void *)w.P); w.P = NULL;
  if (w.cs !=NULL) free((void *)w.cs); w.cs = NULL;
#ifdef VARIABLE_GAMMA
  if (w.g  !=NULL) free((void *)w.g ); w.g  = NULL;
#endif VARIABLE_GAMMA
  
  if (w.v !=NULL)  free((void *)w.v); w.v = NULL;
	
  return(err);
}



int zlsr_ClearWstar1DArr  (Wstar1DArr * w){
  int err = noErr;

  w->n  = 0;
  w->ns = 0;
  w->ne = 0;
  w->nt = 0;  
  w->x  = NULL;
  
  w->d  = NULL;
  w->P  = NULL;
  w->cs = NULL;
  w->v  = NULL;
#ifdef VARIABLE_GAMMA
  w->g = NULL;
#endif VARIABLE_GAMMA
  
  return(err);
}

#ifdef HYDRO

#ifdef UNSPLITSWEEP

#ifdef SYM2D

int zlsr_InitWStar(WStar w ){
	
  int err = noErr;
  Integer i, j;
	Integer ntj = w.nt[kYDim];
	Integer nti = w.nt[kXDim];
  
  for (j = 0; j< ntj; j++){
    for (i = 0; i< nti; i++){
      w.d[j][i].a [kXDim] = 0.0;
      w.d[j][i].a [kYDim] = 0.0;
      w.d[j][i].al[kXDim] = 0.0;
      w.d[j][i].ar[kXDim] = 0.0;
      w.d[j][i].al[kYDim] = 0.0;
      w.d[j][i].ar[kYDim] = 0.0;
      
      w.P[j][i].a [kXDim] = 0.0;
      w.P[j][i].a [kYDim] = 0.0;
      w.P[j][i].al[kXDim] = 0.0;
      w.P[j][i].ar[kXDim] = 0.0;
      w.P[j][i].al[kYDim] = 0.0;
      w.P[j][i].ar[kYDim] = 0.0;
      
      w.cs[j][i].a [kXDim] = 0.0;
      w.cs[j][i].a [kYDim] = 0.0;
      w.cs[j][i].al[kXDim] = 0.0;
      w.cs[j][i].ar[kXDim] = 0.0;
      w.cs[j][i].al[kYDim] = 0.0;
      w.cs[j][i].ar[kYDim] = 0.0;
      
      w.v[j][i].a [kXDim]  = 0.0;
      w.v[j][i].a [kYDim] = 0.0;
      w.v[j][i].al[kXDim] = 0.0;
      w.v[j][i].ar[kXDim] = 0.0;
      w.v[j][i].al[kYDim] = 0.0;
      w.v[j][i].ar[kYDim] = 0.0;
      
#ifdef VARIABLE_GAMMA
      w.g[j][i].a         = kZLS_eos_gamma;
      w.g[j][i].al[kXDim] = kZLS_eos_gamma;
      w.g[j][i].ar[kXDim] = kZLS_eos_gamma;
      w.g[j][i].al[kYDim] = kZLS_eos_gamma;
      w.g[j][i].ar[kYDim] = kZLS_eos_gamma;
#endif VARIABLE_GAMMA

    } // i
  } // j
  return(err);
	
}
#endif SYM2D

#ifdef SYM3D
int zlsr_InitWStar(WStar w ){
	
  int err = noErr;
  Integer i, j, k;
	Integer ntk = w.nt[kZDim];
	Integer ntj = w.nt[kYDim];
	Integer nti = w.nt[kXDim];
  
  for (k = 0; k< ntk; k++){
    for (j = 0; j< ntj; j++){
    for (i = 0; i< nti; i++){
      w.d[k][j][i].a [kXDim] = 0.0;
      w.d[k][j][i].al[kXDim] = 0.0;
      w.d[k][j][i].ar[kXDim] = 0.0;
      w.d[k][j][i].al[kYDim] = 0.0;
      w.d[k][j][i].ar[kYDim] = 0.0;
      w.d[k][j][i].al[kZDim] = 0.0;
      w.d[k][j][i].ar[kZDim] = 0.0;
      
      w.P[k][j][i].a [kXDim] = 0.0;
      w.P[k][j][i].al[kXDim] = 0.0;
      w.P[k][j][i].ar[kXDim] = 0.0;
      w.P[k][j][i].al[kYDim] = 0.0;
      w.P[k][j][i].ar[kYDim] = 0.0;
      w.P[k][j][i].al[kZDim] = 0.0;
      w.P[k][j][i].ar[kZDim] = 0.0;
      
      w.cs[k][j][i].a [kXDim] = 0.0;
      w.cs[k][j][i].a [kYDim] = 0.0;
      w.cs[k][j][i].a [kZDim] = 0.0;

      w.cs[k][j][i].al[kXDim] = 0.0;
      w.cs[k][j][i].ar[kXDim] = 0.0;
      w.cs[k][j][i].al[kYDim] = 0.0;
      w.cs[k][j][i].ar[kYDim] = 0.0;
      w.cs[k][j][i].al[kZDim] = 0.0;
      w.cs[k][j][i].ar[kZDim] = 0.0;
      
      w.v[k][j][i].a [kXDim] = 0.0;
      w.v[k][j][i].a [kYDim] = 0.0;
      w.v[k][j][i].a [kZDim] = 0.0;

      w.v[k][j][i].al[kXDim] = 0.0;
      w.v[k][j][i].ar[kXDim] = 0.0;
      w.v[k][j][i].al[kYDim] = 0.0;
      w.v[k][j][i].ar[kYDim] = 0.0;
      w.v[k][j][i].al[kZDim] = 0.0;
      w.v[k][j][i].ar[kZDim] = 0.0;
      
#ifdef VARIABLE_GAMMA
      w.g[k][j][i].a         = kZLS_eos_gamma;
      w.g[k][j][i].as[kXDim] = kZLS_eos_gamma;
      w.g[k][j][i].al[kXDim] = kZLS_eos_gamma;
      w.g[k][j][i].ar[kXDim] = kZLS_eos_gamma;
      w.g[k][j][i].as[kYDim] = kZLS_eos_gamma;
      w.g[k][j][i].al[kYDim] = kZLS_eos_gamma;
      w.g[k][j][i].ar[kYDim] = kZLS_eos_gamma;
      w.g[k][j][i].as[kZDim] = kZLS_eos_gamma;
      w.g[k][j][i].al[kZDim] = kZLS_eos_gamma;
      w.g[k][j][i].ar[kZDim] = kZLS_eos_gamma;
#endif VARIABLE_GAMMA

    } // i
  } // j
} // k
  return(err);
	
}
#endif SYM3D

int zlsr_ClearWStar  (WStar * w){
  int err = noErr;
#ifdef SYM1D	
  w->n = 0;
  w->ns = 0;
  w->ne = 0;
  w->nt = 0;  
  w->x  = NULL;
#endif SYM1D	
  
#ifdef SYM2D	
  w->n [kXDim] = 0;
  w->ns[kXDim] = 0;
  w->ne[kXDim] = 0;
  w->nt[kXDim] = 0;
  
  w->n [kYDim] = 0;
  w->ns[kYDim] = 0;
  w->ne[kYDim] = 0;
  w->nt[kYDim] = 0;
  
   w->x  = NULL;
   w->y  = NULL;
#endif SYM2D	
#ifdef SYM3D	
   w->n [kXDim] = 0;
   w->ns[kXDim] = 0;
   w->ne[kXDim] = 0;
   w->nt[kXDim] = 0;
   
   w->n [kYDim] = 0;
   w->ns[kYDim] = 0;
   w->ne[kYDim] = 0;
   w->nt[kYDim] = 0;
   
   w->n [kZDim] = 0;
   w->ns[kZDim] = 0;
   w->ne[kZDim] = 0;
   w->nt[kZDim] = 0;
   
   w->x  = NULL;
   w->y  = NULL;
   w->z  = NULL;
#endif SYM3D	
   
   w->d  = NULL;
   w->P  = NULL;
   w->cs = NULL;
   w->v  = NULL;
#ifdef VARIABLE_GAMMA
   w->g = NULL;
#endif VARIABLE_GAMMA
  
  return(err);
}

#ifdef SYM2D

int zlsr_NewWStar(WStar *w, Integer nBC, Counter nj, Counter ni){
	
  int err = noErr;
	
  w->n [kXDim]  = ni;
  w->ns[kXDim] = nBC;
  w->ne[kXDim] = nBC+ni;
  Integer nti = nBC+nBC+ni;
  w->nt[kXDim] = nti;
	
  w->n [kYDim] = nj;
  w->ns[kYDim] = nBC;
  w->ne[kYDim] = nBC+nj;
  Integer ntj = nBC+nBC+nj;
  w->nt[kYDim] = ntj;
	
  w->x = NULL;  // get an axis from an owning grid
  w->y = NULL;  // get an axis from an owning grid
	
  zls_NewLRQ2DArr(&w->d, ntj, nti);
  zls_NewLRQ2DArr(&w->P, ntj, nti);
  zls_NewLRQ2DArr(&w->cs, ntj, nti);
  zls_NewLRQ2DArr(&w->v, ntj, nti);

#ifdef VARIABLE_GAMMA
  zls_NewLRQ2DArr(&w->g, ntj, nti);
#endif VARIABLE_GAMMA
	
  zlsr_InitWStar2D( *w);
	
  return(err);
	
}
#endif SYM2D

#ifdef SYM3D

int zlsr_NewWStar(WStar *w, Integer nBC, Counter nk, Counter nj, Counter ni){
	
  int err = noErr;
	
  w->n [kXDim]  = ni;
  w->ns[kXDim] = nBC;
  w->ne[kXDim] = nBC+ni;
  Integer nti = nBC+nBC+ni;
  w->nt[kXDim] = nti;
	
  w->n [kYDim] = nj;
  w->ns[kYDim] = nBC;
  w->ne[kYDim] = nBC+nj;
  Integer ntj = nBC+nBC+nj;
  w->nt[kYDim] = ntj;

  w->n [kZDim] = nk;
  w->ns[kZDim] = nBC;
  w->ne[kZDim] = nBC+nk;
  Integer ntk = nBC+nBC+nk;
  w->nt[kZDim] = ntk;
  
  w->x = NULL;  // get an axis from an owning grid
  w->y = NULL;  // get an axis from an owning grid
  w->z = NULL;  // get an axis from an owning grid
	
  zls_NewLRQ3DArr(&w->d, ntk, ntj, nti);
  zls_NewLRQ3DArr(&w->P, ntk, ntj, nti);
  zls_NewLRQ3DArr(&w->cs, ntk, ntj, nti);
  zls_NewLRQ3DArr(&w->v, ntk, ntj, nti);
  
#ifdef VARIABLE_GAMMA
  zls_NewLRQ3DArr(&w->g, ntk, ntj, nti);
#endif VARIABLE_GAMMA
	
  zlsr_InitWStar3D( *w);
	
  return(err);
	
}
#endif SYM3D


int zlsr_DisposeWStar  (WStar w){
  int err = noErr;
#ifdef SYM1D	
  if (w.d  !=NULL)  free((void *)w.d); w.d = NULL;
  if (w.P  !=NULL)  free((void *)w.P); w.P = NULL;
  if (w.cs !=NULL)  free((void *)w.cs); w.cs = NULL;
  if (w.v  !=NULL)  free((void *)w.v); w.v = NULL;
#ifdef VARIABLE_GAMMA
  if (w.g  !=NULL) free((void *)w.g ); w.g  = NULL;
#endif VARIABLE_GAMMA
  
#endif SYM1D	

#ifdef SYM2D	
  zls_DisposeLRQ2DArr(w.d) ; w.d = NULL;
  zls_DisposeLRQ2DArr(w.P) ; w.P = NULL;
  zls_DisposeLRQ2DArr(w.cs); w.cs = NULL;
  zls_DisposeLRQ2DArr(w.v); w.v = NULL;
#ifdef VARIABLE_GAMMA
  zls_DisposeLRQ2DArr(w.g); w.g = NULL;
#endif VARIABLE_GAMMA
#endif SYM2D	
  
#ifdef SYM3D	
  zls_DisposeLRQ3DArr(w.d) ; w.d = NULL;
  zls_DisposeLRQ3DArr(w.P) ; w.P = NULL;
  zls_DisposeLRQ3DArr(w.cs); w.cs = NULL;
  zls_DisposeLRQ3DArr(w.v) ; w.v = NULL;
#ifdef VARIABLE_GAMMA
  zls_DisposeLRQ3DArr(w.g) ; w.g = NULL;
#endif VARIABLE_GAMMA
#endif SYM3D	
  
  return(err);
}


#ifdef SYM2D
int zlsr_FillWStar1DFromWStar2DX ( Wstar1DArr w1,  Wstar2DArr w3, Integer j ){
  return(nyiErr);
}
int zlsr_FillWStar1DFromWStar2DY ( Wstar1DArr w1,  Wstar2DArr w3, Integer i ){
  return(nyiErr);
}

int zlsr_FillWStar2DFromWStar1DX ( Wstar2DArr w2,  Wstar1DArr w1, Integer j ){
  return(nyiErr);
}
int zlsr_FillWStar2DFromWStar1DY ( Wstar2DArr w2,  Wstar1DArr w1, Integer i ){
  return(nyiErr);
}
#endif SYM2D

#ifdef SYM3D

int zlsr_FillWStar1DXFromWStar2DYZ ( Wstar1DArr w1,  Wstar3DArr w3, Integer j, Integer k ){
  return(nyiErr);
}
int zlsr_FillWStar1DYFromWStar2DZX ( Wstar1DArr w1,  Wstar3DArr w3, Integer k, Integer i ){
  return(nyiErr);
}
int zlsr_FillWStar1DZFromWStar2DXY ( Wstar1DArr w1,  Wstar3DArr w3, Integer i, Integer j ){
  return(nyiErr);
}

int zlsr_FillWStar3DYZFromWStar1DX ( Wstar3DArr w3,  Wstar1DArr w1, Integer j, Integer k ){
  return(nyiErr);
}
int zlsr_FillWStar3DZXFromWStar1DY ( Wstar3DArr w3,  Wstar1DArr w1, Integer k, Integer i ){
  return(nyiErr);
}
int zlsr_FillWStar3DXYFromWStar1DZ ( Wstar3DArr w3,  Wstar1DArr w1, Integer i, Integer j ){
  return(nyiErr);
}
#endif SYM3D

#endif UNSPLITSWEEP
#endif HYDRO


