#ifndef ZLS_RIEMANN_H
#define ZLS_RIEMANN_H
#include <zls.h>

void zlsr_RiemanSample(Real *p0, Real *u0, Real *r0, Real s, Wstar1DArr wStar, Integer i );

//========================================
int zlsr_PrintWstar1DArr   ( Wstar1DArr w, int cellIndex ); // -1 for all.

int zlsr_InitWstar1DArr    ( Wstar1DArr  w);
int zlsr_ClearWstar1DArr   ( Wstar1DArr *w); // set all to 0 and NULL
int zlsr_NewWstar1DArr     ( Wstar1DArr *w, Integer nBC, Counter ni);
int zlsr_DisposeWstar1DArr ( Wstar1DArr  w);

#ifdef UNSPLITSWEEP

int zlsr_InitWStar          ( WStar  w); 
int zlsr_ClearWStar         ( WStar *w); // set all to 0 and NULL
int zlsr_DisposeWStar       ( WStar  w); 

#ifdef SYM2D
int zlsr_NewWStar           ( WStar      *w, Integer nBC, Counter nj, Counter ni);
#endif SYM2D
#ifdef SYM3D
int zlsr_NewWStar           ( WStar      *w, Integer nBC, Counter nk, Counter nj, Counter ni);
#endif SYM3D

#ifdef SYM2D
int zlsr_FillWStar1DFromWStar2DX ( Wstar1DArr w1,  Wstar2DArr w3, Integer j );
int zlsr_FillWStar1DFromWStar2DY ( Wstar1DArr w1,  Wstar2DArr w3, Integer i );

int zlsr_FillWStar2DFromWStar1DX ( Wstar2DArr w2,  Wstar1DArr w1, Integer j );
int zlsr_FillWStar2DFromWStar1DY ( Wstar2DArr w2,  Wstar1DArr w1, Integer i );
#endif SYM2D

#ifdef SYM3D

int zlsr_FillWStar1DXFromWStar2DYZ ( Wstar1DArr w1,  Wstar3DArr w3, Integer j, Integer k );
int zlsr_FillWStar1DYFromWStar2DZX ( Wstar1DArr w1,  Wstar3DArr w3, Integer k, Integer i );
int zlsr_FillWStar1DZFromWStar2DXY ( Wstar1DArr w1,  Wstar3DArr w3, Integer i, Integer j );

int zlsr_FillWStar3DYZFromWStar1DX ( Wstar3DArr w3,  Wstar1DArr w1, Integer j, Integer k );
int zlsr_FillWStar3DZXFromWStar1DY ( Wstar3DArr w3,  Wstar1DArr w1, Integer k, Integer i );
int zlsr_FillWStar3DXYFromWStar1DZ ( Wstar3DArr w3,  Wstar1DArr w1, Integer i, Integer j );
#endif SYM3D

#endif UNSPLITSWEEP

//====================================================================================
//
//  Primitive Variable Riemann Solvers
//
//====================================================================================

#ifdef TORO
int zlsr_InitToroRiemannGammas (Real g);
int zlsr_ToroRiemannW1D    (Wstar1DArr wStar,  W1DArr wLR);  // Toro 1999 exact solver
int zlsr_SingleRiemannW1D  (Real *P, Real * v, W Wl, W Wr);  // simple interface for a single solve of the Toro
#endif TORO
#ifdef TOROTWOSHOCK
int zlsr_InitToroSSRiemannGammas (Real g);
int zlsr_ToroSSRiemannW1D  (Wstar1DArr wStar,  W1DArr wLR);  // Modified Iterative Two Shock Approx solver derived from Toro exact solver
#endif TOROTWOSHOCK

#ifdef PVRS
int zlsr_InitToroRiemannGammas (Real g);
int zlsr_PVRSRiemannW1D    (Wstar1DArr wStar,  W1DArr wLR);  // Toro 1999 PVRS No. 2 Chapter 9
#endif PVRS
#ifdef AIRS
int zlsr_InitToroRiemannGammas (Real g);
int zlsr_AIRSRiemannW1D    (Wstar1DArr wStar,  W1DArr wLR, Real Q);  // Hybrid Solver, set Q~ 2.0
#endif AIRS

#ifdef COLLELLATWOSHOCK
int zlsr_InitC84RiemannGammas (Real g);
int zlsr_C84RiemannW1D     (Wstar1DArr wStar,  W1DArr wLR);  // Colella 1984 approximate iterative  two shock solver
int zlsr_TestC84RiemannW1D (Real *P, Real * v, W Wl, W Wr);  // single solve version of C84
#endif COLLELLATWOSHOCK

#ifdef GROTHGOTTLIEB
int zlsr_InitGG88RiemannGammas (Real g);
int zlsr_GG88RiemannW1D    (Wstar1DArr wStar,  W1DArr wLR);  // Groth and Gottlieb 1988 solver equiv to Toro but faster
int zlsr_TestGG88RiemannW1D(Real *P, Real * v, W Wl, W Wr);  // single solve interface for full G88 solver
#endif GROTHGOTTLIEB

#ifdef RSSTWOSHOCK
int zlsr_InitRSSRiemannGammas (Real g);
int zlsr_RSSRiemannW1D    (Wstar1DArr wStar,  W1DArr wLR);  // Groth and Gottlieb 1988 solver modified for two shock approximation, equiv to C84 but faster
#endif RSSTWOSHOCK

#endif ZLS_RIEMANN_H
