#include "zls.h"

//
// get event intervals, from an already open config file, called from
// init IO and the problem init as well. Each with a different prefix
// IO uses prefix = "Output"
// init uses prefix = "Sim"
//

int zls_GetEventTiming( TimeCoords * tc, struct ccl_t * config, char * prefix );
int zls_GetEventTiming( TimeCoords * tc, struct ccl_t * config, char * prefix ){
  int err = noErr;
  const char     *str1;
  char keyStr[80];
  
  if ( tc != NULL ) {
    
    time_t realtimeDuration   = zlst_GetRealTimeDuration();
    
    Real timeLimit         = tc->endTime;
    Real ioFrameInterval   = tc->frameTime;
    Real ioUserInterval    = tc->userTime;
    Real ioDataInterval    = tc->dataTime;
    Real ioSaveInterval    = tc->saveTime;
    
    Integer nCycles        = tc->endCycles; 
    Integer nFrameInterval = tc->frameCycles;
    Integer nUserInterval  = tc->userCycles;
    Integer nDataInterval  = tc->dataCycles;
    Integer nSaveInterval  = tc->saveCycles;
    
    Real timingScale = 1.0;
    sprintf( keyStr, "%s_Timing_YearsScale", prefix);
    if ((str1= ccl_get( config, keyStr))){
      timingScale = atof( str1)*f_yrs/f_t0;
    }
    sprintf( keyStr, "%s_Timing_MyrScale", prefix);
    if ((str1= ccl_get( config, keyStr))){
      timingScale = atof( str1)*1.0e6*f_yrs/f_t0;
    }
    
    sprintf( keyStr, "%s_Timing_SecondsScale", prefix);
    if ((str1= ccl_get( config, keyStr))){
      timingScale = atof( str1)/f_t0;
    }
    
    sprintf( keyStr, "%s_RealTime_Duration", prefix); 
    if ((str1= ccl_get( config, keyStr))){
      realtimeDuration = (time_t)atoi( str1);
    }
    
    sprintf( keyStr, "%s_Time_Limit", prefix);
    if ((str1= ccl_get( config, keyStr))){
      timeLimit = atof( str1)*timingScale;
    }
    
    sprintf( keyStr, "%s_n_Limit", prefix);
    if ((str1= ccl_get( config, keyStr))){
      nCycles = strtol( str1, NULL, 0 );
    }
    
    sprintf( keyStr, "%s_Frame_Delta_t", prefix);
    if ((str1= ccl_get( config, keyStr))){
      ioFrameInterval = atof( str1)*timingScale;;
    } 
    
    sprintf( keyStr, "%s_User_Delta_t", prefix);
    if ((str1= ccl_get( config, keyStr))){
      ioUserInterval = atof( str1)*timingScale;;
    } 
    
    sprintf( keyStr, "%s_Data_Delta_t", prefix);
    if ((str1= ccl_get( config, keyStr))){
      ioDataInterval = atof( str1)*timingScale;;
    } 
    
    sprintf( keyStr, "%s_Save_Delta_t", prefix);
    if ((str1= ccl_get( config, keyStr))){
      ioSaveInterval = atof( str1)*timingScale;;
    } 
    
    sprintf( keyStr, "%s_Frame_Delta_n", prefix);
    if ((str1= ccl_get( config, keyStr))){
      nFrameInterval = atoi( str1);
    } 
    
    sprintf( keyStr, "%s_User_Delta_n", prefix);
    if ((str1= ccl_get( config, keyStr))){
      nUserInterval = atoi( str1);
    } 
    
    sprintf( keyStr, "%s_Data_Delta_n", prefix);
    if ((str1= ccl_get( config, keyStr))){
      nDataInterval = atoi( str1);
    } 
    
    sprintf( keyStr, "%s_Save_Delta_n", prefix);
    if ((str1= ccl_get( config, keyStr))){
      nSaveInterval = atoi( str1);
    } 
    
    // -ve values are set  beyond the end time to disable them
    if ( timeLimit <= 0.0 ) timeLimit = kZLS_Huge;
    if ( realtimeDuration <= 0 ) realtimeDuration = 31557600;  // 1 yr :)
    
    if ( ioFrameInterval <= 0.0) ioFrameInterval = timeLimit+kZLS_Huge;
    if ( ioUserInterval <= 0.0 ) ioUserInterval  = timeLimit+kZLS_Huge;
    if ( ioDataInterval <= 0.0 ) ioDataInterval  = timeLimit+kZLS_Huge;
    if ( ioSaveInterval <= 0.0 ) ioSaveInterval  = timeLimit+kZLS_Huge;
    
    zlst_SetAllLevelClocksEndTime      ( timeLimit );
    
    // sets *both* realtime and limit, delta t in whole seconds
    zlst_SetRealTimeDuration           ( realtimeDuration ); 
    
    zlst_SetAllLevelClocksFrameTime    ( ioFrameInterval );
    zlst_SetAllLevelClocksUserTime     ( ioUserInterval );
    zlst_SetAllLevelClocksDataTime     ( ioDataInterval );
    zlst_SetAllLevelClocksSaveTime     ( ioSaveInterval );
    
    if ( nCycles   < 0 ) nCycles = kMaxIntM1;
    if ( nFrameInterval > nCycles) nFrameInterval = nCycles + 1;
    if ( nUserInterval > nCycles ) nUserInterval  = nCycles + 1;
    if ( nDataInterval > nCycles ) nDataInterval  = nCycles + 1;
    if ( nSaveInterval > nCycles ) nSaveInterval  = nCycles + 1;
    
    zlst_SetAllLevelClocksEndCycle     ( nCycles       );
    zlst_SetAllLevelClocksFrameCycle   ( nFrameInterval);
    zlst_SetAllLevelClocksUserCycle    ( nUserInterval );
    zlst_SetAllLevelClocksDataCycle    ( nDataInterval );
    zlst_SetAllLevelClocksSaveCycle    ( nSaveInterval );
    
  } else  { err = memErr; }
  
  return(err);
  
}


int   zlsio_SetUpSimLimitsFromConfigFile(char * confFile){
  
  int err = noErr;
  
  struct ccl_t  config;
  const char     *str1;
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, confFile) == noErr) {
    
#if (ZLS_VERBOSE >0)
    zlsio_DisplaySimLimits( stdout , confFile,  1 );
#endif ZLS_VERBOSE
    
    TimeCoords tc = gLevels[0].levelClock;
    zls_GetEventTiming( &tc, &config, "Sim");
    
#if (ZLS_VERBOSE >0)
    zlsio_DisplaySimLimits( stdout , confFile, 0 );
#endif ZLS_VERBOSE
    
    IOSettingsPtr io =      gLevels[0].levelIO;
    
    if ( io != NULL ) { // allow override of io *if* this config is read *after* io.conf and io exists already
      
      strcpy( io->outRunName , FYRIS_PROBLEMNAME); // default run to name in zls_problem.h
      
      if ((str1= ccl_get( &config,"Simulation_Name"))){
        strcpy(  io->outRunName,  str1);
      }
      zlsio_Printf( " >>>Override:  ZLS Simulation Run Name: %s \n", io->outRunName);
      
      if ((str1= ccl_get( &config,"Results_Directory"))){
        strcpy(  io->outDir,  str1);
        
#ifdef HYDRO
        zlsio_Printf( " >>>Override: FYRIS Hydro Problem Output in: %s \n", str1);
#endif HYDRO
        
      }
      
      if ((str1= ccl_get( &config,"Results_BaseName"))){
        strcpy(  io->outName,  str1);
        
        zlsio_Printf( " >>>Override: Problem Output Base Name in: %s \n", str1);
        
      };
      
      if ((str1= ccl_get( &config,"Results_LogName"))){
        strcpy(  io->outLogName,  str1);
        zlsio_Printf( " >>>Override: Problem Log File: %s \n", str1);
      }
      
      if ((str1= ccl_get( &config,"Results_MonitorDirectory"))){
        strcpy(  io->monDir,  str1);
        zlsio_Printf( " >>>Override: Monitor Location: %s \n", str1);
      }
      
      if ((str1= ccl_get( &config,"Results_MonitorName"))){
        strcpy(  io->outMonitorName,  str1);
        zlsio_Printf( " >>>Override: Monitor File Name: %s \n", str1);
      }
      
    };
    
    ccl_release(&config);
  }; // config
  
  return (err);
  
}


int   zlsio_SetUpIOFromConfigFile(char * confFile){
  
  int err = noErr;
  
  IOSettingsPtr io;
  
  Integer nVars = 0;
  Integer monitorVar = -1;
  Integer monitorLevel = 0;
  Integer         n;
  Integer  ioDataFormat   = kZLS_ZLSData; // read as int, only set outDataFormat for valid values
  Integer  ioDataCellBits = kZLS_Float32Data; // read as int, set outDataBitsPerCell for valid values
  
  
  Integer  ioType    = kZLS_IOImage+kZLS_IOData;
  Integer  transType = kZLS_IOLinearTransform;
  Integer  scaleType = kZLS_IOAutoScale;
  Integer  sliceType = kSliceXY;
  Integer  sliceIndex = 0;
  
  Integer outImageBPP    = k8bitOutput;  // read as int, only set outBitsPerPix for valid values
  Integer outImageFormat = kPNGOutput;  // read as int, only set outFormat for valid values
  
  Integer level;
  
  struct ccl_t  config;
  const char     *str1;
  char     keyStr[256];
  
  err = zlsio_NewIOSettings(&io);
  
  if (io != NULL) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile) == noErr) {
      
#if (ZLS_VERBOSE > 0)
     zlsio_DisplayIOSetup  ( stdout , confFile, 1 );
#endif (ZLS_VERBOSE > 0)
      
      if ((str1= ccl_get( &config,"NIOVars"))){
        nVars = strtol( str1, NULL, 0 );
      }
      
      // override default progress and performance check intervals
      
      if ((str1= ccl_get( &config,"Progress_Cycles"))){
        kZLS_CycleProgressPrint = strtol( str1, NULL, 0 );
      }
      
      if ((str1= ccl_get( &config,"Performance_Cycles"))){
        kZLS_CPUCyclePerformance = strtol( str1, NULL, 0 );
      }
      
      if ((str1= ccl_get( &config,"Monitor_Var"))){
        monitorVar = strtol( str1, NULL, 0 );
      }
      
      if ((str1= ccl_get( &config,"Monitor_Level"))){
        monitorLevel = strtol( str1, NULL, 0 );
      }
      
      monitorLevel = (monitorLevel < 0)? 0: monitorLevel;
      monitorLevel = (monitorLevel >= gNLevels)? 0 : monitorLevel;
      
      err = zlsio_InitIOSettings(io,  nVars, monitorVar, monitorLevel);
      
      if (err == noErr) {
        
        //
        // First go at setting Events, can be overridden in init.conf afterwards.
        // Normally the io.conf sets output intervals, where init.conf sets overall time limits
        // and or restart save intervals.
        //
        TimeCoords tc;
        zlst_DefaultEventTiming ( &tc );
        
        zls_GetEventTiming( &tc, &config, "Output");
        
        strcpy( io->outRunName , FYRIS_PROBLEMNAME); // default run to name in zls_problem.h
        
        if ((str1= ccl_get( &config,"Simulation_Name"))){
          strcpy(  io->outRunName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Simulation Run Name: %s \n", io->outRunName);
#endif (ZLS_VERBOSE>0)
        
        strcpy(  io->outDir,  OUTPUTDIR);
        strcpy(  io->monDir,  MONITORDIR); // defaults to same as Output
        if ((str1= ccl_get( &config,"Output_Directory"))){
          strcpy(  io->outDir,  str1);
          strcpy(  io->monDir,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Directory: %s \n", io->outDir);
#endif (ZLS_VERBOSE>0)
        
        strcpy(  io->outName,  OUTFILE);
        if ((str1= ccl_get( &config,"Output_BaseName"))){
          strcpy(  io->outName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Base Name: %s \n", io->outName);
#endif (ZLS_VERBOSE>0)
        
        if ((str1= ccl_get( &config,"Output_MonitorDirectory"))){
          strcpy(  io->monDir,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Monitor Directory: %s \n", io->monDir);
#endif (ZLS_VERBOSE>0)
        
        strcpy( io->outMonitorName,  MONFILE);
        if ((str1= ccl_get( &config,"Output_MonitorName"))){
          strcpy(  io->outMonitorName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Monitor Name: %s \n", io->outMonitorName);
#endif (ZLS_VERBOSE>0)
        
        strcpy( io->outLogName,  ZLSLOGFILE);
        if ((str1= ccl_get( &config,"Output_Log"))){
          strcpy(  io->outLogName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Log File: %s \n", io->outLogName);
#endif (ZLS_VERBOSE>0)
        
        strcpy( io->outInfoName,  ZLSINFOFILE);
        if ((str1= ccl_get( &config,"Output_ConfigLog"))){
          strcpy(  io->outInfoName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Confguration Log  File: %s \n", io->outInfoName);
#endif (ZLS_VERBOSE>0)
        
        strcpy( io->outFilesName,  ZLSFILESLOG);
        if ((str1= ccl_get( &config,"Output_FilesLog"))){
          strcpy(  io->outFilesName,  str1);
        }
#if (ZLS_VERBOSE>0)
        zlsio_Printf( " ZLS Output Files Log: %s \n", io->outFilesName);
#endif (ZLS_VERBOSE>0)
        
        // support older io.conf files
        if ((str1= ccl_get( &config,"Output_Format"))){
          outImageFormat= strtol( str1, NULL, 0 );
          if (outImageFormat == 0) {
            io->outFormat  = kRawOutput;
          }
          if (outImageFormat == 1) {
            io->outFormat = kTIFFOutput;
          }   
          if (outImageFormat == 2) {
            io->outFormat = kJPEGOutput;
          }   
          if (outImageFormat == 3) {
            io->outFormat = kPNGOutput;
          }   
        }
        
        if ((str1= ccl_get( &config,"Output_BPP"))){
          outImageBPP = strtol( str1, NULL, 0 );
          if (outImageBPP == 8) {
            io->outBitsPerPix = k8bitOutput;
          }
          if (outImageBPP == 16) {
            io->outBitsPerPix = k16bitOutput;
          }   
        }
        
        
        if ((str1= ccl_get( &config,"Output_ImageFormat"))){
          outImageFormat= strtol( str1, NULL, 0 );
          if (outImageFormat == 0) {
            io->outFormat  = kRawOutput;
          }
          if (outImageFormat == 1) {
            io->outFormat = kTIFFOutput;
          }   
          if (outImageFormat == 2) {
            io->outFormat = kJPEGOutput;
          }   
          if (outImageFormat == 3) {
            io->outFormat = kPNGOutput;
          }   
        }
        
        if ((str1= ccl_get( &config,"Output_ImageBPP"))){
          outImageBPP = strtol( str1, NULL, 0 );
          if (outImageBPP == 8) {
            io->outBitsPerPix = k8bitOutput;
          }
          if (outImageBPP == 16) {
            io->outBitsPerPix = k16bitOutput;
          }   
        }
        
        ioDataFormat   = kZLS_ZLSData; // read as int, only set outDataFormat for valid values
        if ((str1= ccl_get( &config,"Output_DataFormat"))){
          outImageFormat= strtol( str1, NULL, 0 );
          if (outImageFormat == 0) {
            ioDataFormat  = kZLS_ZLSData;
          }
          if (outImageFormat == 1) {
            ioDataFormat = kZLS_FITSData;
          }   
          if (outImageFormat == 2) {
            ioDataFormat = kZLS_HDFData;
          }   
        }
        io->outDataFormat = ioDataFormat;
        
        
        ioDataCellBits = kZLS_Float32Data; // read as int, set outDataBitsPerCell for valid values
        if ((str1= ccl_get( &config,"Output_DataBits"))){
          outImageBPP = strtol( str1, NULL, 0 );
          if (outImageBPP == 16) {
            ioDataCellBits = kZLS_SInt16Data;
          }
          if (outImageBPP == 32) {
            ioDataCellBits = kZLS_Float32Data;
          }   
        }
        io->outDataBitsPerCell = ioDataCellBits;
        
        //          // catch older misspelled io.conf files
        //          if ((str1= ccl_get( &config,"Output_BBP"))){
        //            outImageBPP = strtol( str1, NULL, 0 );
        //            if (outImageBPP == 8) {
        //              io->outBitsPerPix = k8bitOutput;
        //            }
        //            if (outImageBPP == 16) {
        //              io->outBitsPerPix = k16bitOutput;
        //            }   
        //          }
        
        if ((str1= ccl_get( &config,"Output_CTAB"))){
          strcpy(  io->outCTABFile, str1);
          if (    (io->outFormat == kTIFFOutput)
              ||(io->outFormat == kJPEGOutput)
              ||(io->outFormat == kPNGOutput)){ // get the ctab
            io->ioTIFFCTab = (TIFFCTab *) malloc (sizeof(TIFFCTab));
            err = zls_ReadCtab(io->outCTABFile, io->ioTIFFCTab);
          }
        }
        
        // record the name of the global CTAB file in a string 
        // so the monitor variable can find a scale bar graphic
        // for the html page...
        
        strcpy( io->monitorCTABFile, io->outCTABFile);
        
        Integer renderMergeData = kZLS_IOOutputSeparateZones;
        
        if ((str1= ccl_get( &config,"Output_MergeData"))){
          
          if (strcmp(str1, "Merged")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Yes")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "YES")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "yes")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Y")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "y")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Separate")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "No")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "NO")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "no")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "N")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "n")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
        };
        
        // alternative key for typos  :) ie the final d is added or left off
        
        if ((str1= ccl_get( &config,"Output_MergedData"))){
          
          
          if (strcmp(str1, "Merged")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Yes")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "YES")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "yes")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Y")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "y")==0) {
            renderMergeData = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Separate")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "No")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "NO")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "no")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "N")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "n")==0) {
            renderMergeData = kZLS_IOOutputSeparateZones;
          }
          
        };
        
        
        Integer renderMergeFrames = kZLS_IOOutputMergedZones;
        
        if ((str1= ccl_get( &config,"Output_MergeFrames"))){
          
          if (strcmp(str1, "Merged")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Yes")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "YES")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "yes")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Y")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "y")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Separate")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "No")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "NO")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "no")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "N")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "n")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
        };
        
        // alternative key for typos  :) ie the final d is added or left off
        
        if ((str1= ccl_get( &config,"Output_MergedFrames"))){
          
          
          if (strcmp(str1, "Merged")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Yes")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "YES")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "yes")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Y")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "y")==0) {
            renderMergeFrames = kZLS_IOOutputMergedZones;
          }
          
          if (strcmp(str1, "Separate")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "No")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "NO")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "no")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "N")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
          if (strcmp(str1, "n")==0) {
            renderMergeFrames = kZLS_IOOutputSeparateZones;
          }
          
        };
        
        // not needed if there is only one zone.  In the future this
        // won't be true if zones become generated on the fly and nZones
        // changes during a simulation...
        
        //          if (gLevels[0].nZones < 2) renderMergeData   = kZLS_IOOutputSeparateZones;
        //          if (gLevels[0].nZones < 2) renderMergeFrames = kZLS_IOOutputSeparateZones;
        
#ifdef SYM3D
#ifdef MPI
        if (gMPIrank == level) { // only allocate on rank == level
#endif MPI
          if ( (renderMergeData == kZLS_IOOutputMergedZones) ||
              (renderMergeFrames == kZLS_IOOutputMergedZones) ) {
            zlsio_ErrPrintf( "WARNING: Merged IO levels selected in 3D.\n" );
            zlsio_ErrPrintf( "       : In three dimensional geometry this may\n" );
            zlsio_ErrPrintf( "       : fail with large multi-zone simulations.\n" );
            zlsio_ErrPrintf( "       : Proceeding, but may fail for lack of memory.\n" );
          }
#ifdef MPI
        } // only allocate on rank == level
#endif MPI
#endif SYM3D
        
        io->outMergedZoneData   = renderMergeData;
        io->outMergedZoneFrames = renderMergeFrames;
        
        if ( (renderMergeData == kZLS_IOOutputMergedZones) ||
            (renderMergeFrames == kZLS_IOOutputMergedZones) ) {
          
          Integer level;
          for (level = 0; level < gNLevels; level++){
            
            // allocate image zone  for level for consolidated output
            
#ifdef MPI
            if (gMPIrank == level) { // only allocate data for level n on rank n
#endif MPI
              // allocate a zoneData to hold an image in floating pt.
              // the image IO routines can treat the level image zone like other zones, but it has just one
              // data array and no fluxes etc.
              
              Range  rng;
              err = zls_GetLevelRange( &rng, level );
              
              ZonePtr  z = NULL;
              err = zls_NewZone( &z );
              err = zls_InitZone( z, kImageZoneType, level, &rng, NULL,  &rng);
              
#if (ZLS_DEBUG == 1)
              zlsio_ErrPrintf( " Merged level %d image data zone created:\n", level );
              zls_PrintGridCoords( stdout, &z->zoneGrid); 
#endif ZLS_DEBUG            
#if (ZLS_DEBUG > 1)
              zls_PrintZoneCoords( stdout, z); 
#endif ZLS_DEBUG      
              err = zls_NewZoneData( z, kAnonVar); 
#if (ZLS_DEBUG > 1)
              zlsio_ErrPrintf( " Allocated ID:%d image zone data on rank %d:\n", z->zoneRankID, gMPIrank );
#endif ZLS_DEBUG      
              z->nQ = 1;
              err = zls_SetLevelImageZone( level, z ); 
              
              
#ifdef MPI
            } else {
              
              // no imageZone needed on other ranks set NULL for safety
              
              err = zls_SetLevelImageZone( level, NULL ); 
              
            }
#endif MPI
          } // level
          
        } else { // !kZLS_IOOutputMergedZones
          
          // no imageZone needed, set NULL for safety
          
          Integer level;
          for (level = 0; level < gNLevels; level++){
            err = zls_SetLevelImageZone( level, NULL ); 
          }
          
        }
        
        //
        // now get variable by variable settings for IO
        //
        if (nVars > 0){
          
          for ( n = 0; n < io->outNVars; n++){
            
            sprintf(keyStr,"Var_%2.2d_Code", n);
            if ((str1= ccl_get( &config, keyStr))){
              strncpy(io->v[n].varCode , str1, CODELEN); // limit copy to max CODELEN chars
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            } else {
              zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
              ccl_release(&config);
              return(ioErr);
            }
            
            sprintf(keyStr,"Var_%2.2d_PrefixName", n);
            if ((str1= ccl_get( &config, keyStr))){
              io->v[n].varPrefix = (char *)malloc(strlen(str1)+1);
              strcpy(io->v[n].varPrefix , str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, io->v[n].varPrefix);
#endif
            } else {
              zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
              ccl_release(&config);
              return(ioErr);
            }
            
            sprintf(keyStr,"Var_%2.2d_CTAB", n);
            if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
              io->v[n].varTIFFCTab = (TIFFCTab *) malloc (sizeof(TIFFCTab));
              if (io->v[n].varTIFFCTab!=NULL){
                err = zls_ReadCtab((char*)str1, io->v[n].varTIFFCTab);
              } else {
                io->v[n].varTIFFCTab = io->ioTIFFCTab;
              }
              
              if (( n == monitorVar) && (io->v[n].varTIFFCTab != NULL)){
                // if this is the monitor variable and it has a specific colour table
                // then we need the name of the ctab file for the monitor 
                // scale bar display
                strcpy( io->monitorCTABFile, str1);
              }
              
              
            } else { // set a ptr to the global default, may be NULL!
              io->v[n].varTIFFCTab = io->ioTIFFCTab;
            }
            
            ioType = kZLS_IOImage+kZLS_IOData; // default images and data
            
            sprintf(keyStr,"Var_%2.2d_IOType", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "Image")==0) {
                ioType = kZLS_IOImage;
              }
              if (strcmp(str1, "ImageOnly")==0) {
                ioType = kZLS_IOImage;
              }
              if (strcmp(str1, "Data")==0) {
                ioType = kZLS_IOData;
              }
              if (strcmp(str1, "DataOnly")==0) {
                ioType = kZLS_IOData;
              }
              if (strcmp(str1, "Image_and_Data")==0) {
                ioType = kZLS_IOImage+kZLS_IOData;
              }
              if (strcmp(str1, "Plot")==0) {
                ioType = kZLS_IOPlot;
              }
              if (strcmp(str1, "PlotOnly")==0) {
                ioType = kZLS_IOPlot;
              }
              
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            
            io->v[n].ioTypes = ioType;
            
#ifdef SYM1D
            sliceType = kSliceX;
#endif SYM1D
#ifdef SYM2D
            sliceType = kSliceXY;
#endif SYM2D
#ifdef SYM3D
            sliceType = kSliceXYSum;
#endif SYM3D
            
            sprintf(keyStr,"Var_%2.2d_ImageSlice", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "X")==0) {
                sliceType = kSliceX;
              }
              
              if (strcmp(str1, "XSum")==0) {
                sliceType = kSliceXSum;
              }
              
#if (NDIMS > 1)
              if (strcmp(str1, "Y")==0) {
                sliceType = kSliceY;
              }
              if (strcmp(str1, "YSum")==0) {
                sliceType = kSliceYSum;
              }
              
              if (strcmp(str1, "XY")==0) {
                sliceType = kSliceXY;
              }
              if (strcmp(str1, "XYSum")==0) {
                sliceType = kSliceXYSum;
              }
              if (strcmp(str1, "YX")==0) {
                sliceType = kSliceYX;
              }
              if (strcmp(str1, "YXSum")==0) {
                sliceType = kSliceYXSum;
              }
              
#if (NDIMS > 2)
              if (strcmp(str1, "Z")==0) {
                sliceType = kSliceZ;
              }
              
              if (strcmp(str1, "XZ")==0) {
                sliceType = kSliceXZ;
              }
              if (strcmp(str1, "YZ")==0) {
                sliceType = kSliceYZ;
              }
              if (strcmp(str1, "ZX")==0) {
                sliceType = kSliceZX;
              }
              if (strcmp(str1, "ZY")==0) {
                sliceType = kSliceZY;
              }
              if (strcmp(str1, "XZSum")==0) {
                sliceType = kSliceXZSum;
              }
              if (strcmp(str1, "YZSum")==0) {
                sliceType = kSliceYZSum;
              }
              if (strcmp(str1, "ZXSum")==0) {
                sliceType = kSliceZXSum;
              }
              if (strcmp(str1, "ZYSum")==0) {
                sliceType = kSliceZYSum;
              }
              
              if (strcmp(str1, "XYZ")==0) {
                sliceType = kSliceXYZ;
              }
              if (strcmp(str1, "XYZSum")==0) {
                sliceType = kSliceXYZ;
              }
              
#endif  (NDIMS > 2)
              
#endif (NDIMS > 1)
              
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            
            io->v[n].imageSlice = sliceType;
            
            sliceIndex = -1;
            
            sprintf(keyStr,"Var_%2.2d_ImageIndex", n);
            if ((str1= ccl_get( &config, keyStr))){
              sliceIndex = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageIndex = sliceIndex;
            
            sprintf(keyStr,"Var_%2.2d_ImageIndex2", n);
            if ((str1= ccl_get( &config, keyStr))){
              sliceIndex = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageIndex2 = sliceIndex;
            
            Integer plotColor = io->v[n].imagePlotColor;
            sprintf(keyStr,"Var_%2.2d_PlotColor", n);
            if ((str1= ccl_get( &config, keyStr))){
              plotColor = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imagePlotColor = plotColor;
            
            plotColor = io->v[n].imagePlotColor;
            sprintf(keyStr,"Var_%2.2d_PlotColour", n);
            if ((str1= ccl_get( &config, keyStr))){
              plotColor = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imagePlotColor = plotColor;
            
            sprintf(keyStr,"Var_%2.2d_PlotIndex", n);
            if ((str1= ccl_get( &config, keyStr))){
              sliceIndex = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageIndex = sliceIndex;
            
            sprintf(keyStr,"Var_%2.2d_PlotIndex2", n);
            if ((str1= ccl_get( &config, keyStr))){
              sliceIndex = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageIndex2 = sliceIndex;
            
            sliceType = io->v[n].imageSlice;
            
            sprintf(keyStr,"Var_%2.2d_PlotSlice", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "X")==0) {
                sliceType = kSliceX;
              }
              
              if (strcmp(str1, "XSum")==0) {
                sliceType = kSliceXSum;
              }
              
#if (NDIMS > 1)
              if (strcmp(str1, "Y")==0) {
                sliceType = kSliceY;
              }
              if (strcmp(str1, "YSum")==0) {
                sliceType = kSliceYSum;
              }
              
              if (strcmp(str1, "XY")==0) {
                sliceType = kSliceX;
              }
              if (strcmp(str1, "XYSum")==0) {
                sliceType = kSliceXSum;
              }
              if (strcmp(str1, "YX")==0) {
                sliceType = kSliceY;
              }
              if (strcmp(str1, "YXSum")==0) {
                sliceType = kSliceYSum;
              }
              
#if (NDIMS > 2)
              if (strcmp(str1, "Z")==0) {
                sliceType = kSliceZ;
              }
              if (strcmp(str1, "XZ")==0) {
                sliceType = kSliceX;
              }
              if (strcmp(str1, "YZ")==0) {
                sliceType = kSliceY;
              }
              if (strcmp(str1, "ZX")==0) {
                sliceType = kSliceZ;
              }
              if (strcmp(str1, "ZY")==0) {
                sliceType = kSliceZ;
              }
              if (strcmp(str1, "XZSum")==0) {
                sliceType = kSliceXSum;
              }
              if (strcmp(str1, "YZSum")==0) {
                sliceType = kSliceYSum;
              }
              if (strcmp(str1, "ZXSum")==0) {
                sliceType = kSliceZSum;
              }
              if (strcmp(str1, "ZYSum")==0) {
                sliceType = kSliceZSum;
              }
#endif  (NDIMS > 2)
              
#endif (NDIMS > 1)
              
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageSlice = sliceType;
            
            scaleType = kZLS_IOAutoScale;
            sprintf(keyStr,"Var_%2.2d_ImageScalingType", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "Auto")==0) {
                scaleType = kZLS_IOAutoScale;
              }
              if (strcmp(str1, "Manual")==0) {
                scaleType = kZLS_IOManualScale;
              }
              if (strcmp(str1, "AutoMin")==0) {
                scaleType = kZLS_IOAutoMinScale;
              }
              if (strcmp(str1, "AutoMax")==0) {
                scaleType = kZLS_IOAutoMaxScale;
              }
              io->v[n].imageScalingType = scaleType;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageScalingType = scaleType;
            
            scaleType = io->v[n].imageScalingType;
            sprintf(keyStr,"Var_%2.2d_PlotScalingType", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "Auto")==0) {
                scaleType = kZLS_IOAutoScale;
              }
              if (strcmp(str1, "Manual")==0) {
                scaleType = kZLS_IOManualScale;
              }
              if (strcmp(str1, "AutoMin")==0) {
                scaleType = kZLS_IOAutoMinScale;
              }
              if (strcmp(str1, "AutoMax")==0) {
                scaleType = kZLS_IOAutoMaxScale;
              }
              io->v[n].imageScalingType = scaleType;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].imageScalingType = scaleType;
            
            scaleType = kZLS_IOAutoScale;
            sprintf(keyStr,"Var_%2.2d_DataScalingType", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              if (strcmp(str1, "Auto")==0) {
                scaleType = kZLS_IOAutoScale;
              }
              if (strcmp(str1, "Manual")==0) {
                scaleType = kZLS_IOManualScale;
              }
              if (strcmp(str1, "AutoMin")==0) {
                scaleType = kZLS_IOAutoMinScale;
              }
              if (strcmp(str1, "AutoMax")==0) {
                scaleType = kZLS_IOAutoMaxScale;
              }
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            io->v[n].dataScalingType = scaleType;
            
            io->v[n].scaleMin = 0.0;
            io->v[n].dataMin = 0.0;
            sprintf(keyStr,"Var_%2.2d_ScaleMin", n);
            if ((str1= ccl_get( &config, keyStr))){
              io->v[n].scaleMin = atof(str1);
              io->v[n].dataMin = atof(str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            } 
            
            io->v[n].scaleMax = 0.0;
            io->v[n].dataMax = 0.0;
            sprintf(keyStr,"Var_%2.2d_ScaleMax", n);
            if ((str1= ccl_get( &config, keyStr))){
              io->v[n].scaleMax = atof(str1);
              io->v[n].dataMax = atof(str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            
            io->v[n].scaleInvRange = 1.0/(io->v[n].scaleMax-io->v[n].scaleMin+kZLS_Epsilon);
            
            
            sprintf(keyStr,"Var_%2.2d_Transform", n);
            if ((str1= ccl_get( &config, keyStr))){
              
              transType = kZLS_IOLinearTransform;
              if (strcmp(str1, "Linear")==0) {
                transType = kZLS_IOLinearTransform;
              }
              if (strcmp(str1, "Log10")==0) {
                transType = kZLS_IOLog10Transform;
              }
              if (strcmp(str1, "Log")==0) {
                transType = kZLS_IOLog10Transform;
              }
              if (strcmp(str1, "LogAbs")==0) {
                transType = kZLS_IOLog10AbsTransform;
              }
              if (strcmp(str1, "Log10Abs")==0) {
                transType = kZLS_IOLog10AbsTransform;
              }
              if (strcmp(str1, "Exp")==0) {
                transType = kZLS_IOExp10Transform;
              }
              if (strcmp(str1, "Power")==0) {
                transType = kZLS_IOPowTransform;
                
                sprintf(keyStr,"Var_%2.2d_TransPower", n);
                if ((str1= ccl_get( &config, keyStr))){
                  io->v[n].transPow = atof(str1);
#if (ZLS_DEBUG>0)
                  zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
                } else {
                  io->v[n].transPow = 0.5;
                }
                
              }
              if (strcmp(str1, "Sqrt")==0) {
                transType = kZLS_IOPowTransform;
                io->v[n].transPow = 0.5;
              }
              if (strcmp(str1, "Square")==0) {
                transType = kZLS_IOPowTransform;
                io->v[n].transPow = 2.0;
              }
              io->v[n].transformType = transType;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            } else {
              io->v[n].transformType = kZLS_IOLinearTransform;
            }
            
            sprintf(keyStr,"Var_%2.2d_TransThreshold", n);
            if ((str1= ccl_get( &config, keyStr))){
              io->v[n].transThreshold = atof(str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            
            sprintf(keyStr,"Var_%2.2d_TransNULL", n);
            if ((str1= ccl_get( &config, keyStr))){
              io->v[n].transNULL = atof(str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            } 
            
          } // n vars
        } // init
        
      } // n vars
      
      ccl_release(&config);
      
    } // config
    
    if (err == noErr) {
      
#if (ZLS_VERBOSE >0)
      zlsio_DisplayIOSetup  ( stdout , confFile, 0 );
#endif
      
      gLevels[0].levelIO = io;
      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelIO = gLevels[0].levelIO;
        }
      }
      
    } else { err = fileErr; }
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 IOSettingsPtr \n" );
    return (err);
  }
  
  return (err);
  
}


int   zlsio_SetUpPPMFromConfigFile(char * confFile){
  
  int err = noErr;
  
  struct ccl_t  config;
  const char    *str1;
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, confFile)==noErr) {
    
    kZLS_CFL     = 0.80; 
    kZLS_FracCFL = 0.80;  // fraction of cfl for first step
    
    if ((str1= ccl_get( &config,"Courant_Number"))){
      kZLS_CFL = atof( str1);
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Courant_Number", kZLS_CFL);
#endif (ZLS_DEBUG > 0)
    }
    
    if ((str1= ccl_get( &config,"Inital_Step_Fraction"))){
      kZLS_FracCFL = atof( str1);
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Inital_Step_Fraction", kZLS_FracCFL);
#endif (ZLS_DEBUG > 0)
    }
    
    kZLS_Omega1       =  0.5; // 0.75 decrease for greater sensitivity to +/- 1 P diffs/ +/-2 P Diffs
    kZLS_Omega2       = 10.0; // 5.00 increase for stronger flattening/dissipation if threshold are satisfied
    kZLS_CWEpsilon    =  0.3;  // 0.330 decrease for greater sensitivity to pressure jumps
    kZLS_PPMFlatMin   =  0.0;  // 0.0 increase for general diffusivity
    kZLS_PPMFlatMax   = 1.00;  // 1.0 decrease for limiting max diffusivity
    kZLS_PPMFlatDv   =  0.05;  // 0.05 - delat v threshold
    kZLS_PPMFlatDP   =  0.05;  // 0.05 - delat v threshold
    kZLS_PPMFlatDd   =  0.05;  // 0.05 - delat v threshold
    
    if ( (str1 = ccl_get( &config,"PPM_Omega1")) ){
      kZLS_Omega1 = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_Omega1", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_Omega2"))){
      kZLS_Omega2 = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_Omega2", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_Epsilon"))){
      kZLS_CWEpsilon = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_Epsilon", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_FlatMin"))){
      kZLS_PPMFlatMin = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_FlatMin", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_FlatMax"))){
      kZLS_PPMFlatMax = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_FlatMax", str1);
#endif
    }
    
    kZLS_RemapAlpha      = 0.01;  
    kZLS_SmoothAxes      = kSmoothAll; // Combination of smooth axes flags
#ifdef DIVVHCORRECTION
    kZLS_DivVScale       = 1.0*zls_sqrt(1.3333333333);  
    kZLS_DivVInvScale    = 1.0/kZLS_DivVScale;  
    kZLS_DivVThreshold   = 1.0*zls_sqrt(1.3333333333);  
    kZLS_RemapMaxAlpha   = 0.1;  
#endif DIVVHCORRECTION
    
    // keep compatible with older conf files..    
    if ((str1= ccl_get( &config,"PPM_RemapSmooth"))){
      kZLS_RemapAlpha = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_RemapSmooth", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_ViscousAlpha"))){
      kZLS_RemapAlpha = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_ViscousAlpha", str1);
#endif
    }
    
    
#ifdef DIVVHCORRECTION
    if ((str1= ccl_get( &config,"PPM_DivV_Scale"))){
      kZLS_DivVScale = atof( str1)*zls_sqrt(1.3333333333);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_DivV_Scale", str1);
#endif
    }
    kZLS_DivVInvScale = 1.0/kZLS_DivVScale;
    
    if ((str1= ccl_get( &config,"PPM_DivV_Threshold"))){
      kZLS_DivVThreshold = atof( str1)*zls_sqrt(1.33333333333333);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_DivV_Threshold", str1);
#endif
    }
    
    // keep compatible with older conf files..    
    if ((str1= ccl_get( &config,"PPM_RemapMaxSmooth"))){
      kZLS_RemapMaxAlpha = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_RemapMaxSmooth", str1);
#endif
    }
    
    if ((str1= ccl_get( &config,"PPM_ViscousMaxAlpha"))){
      kZLS_RemapMaxAlpha = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_ViscousMaxAlpha", str1);
#endif
    }
    
    if ( kZLS_RemapAlpha > kZLS_RemapMaxAlpha ) kZLS_RemapAlpha = kZLS_RemapMaxAlpha;
    
#endif DIVVHCORRECTION
    
    
    kZLS_SmoothAxes = kSmoothNone;
    
    if ((str1= ccl_get( &config,"PPM_ViscousAxes"))){
      
      if (strcmp(str1, "X")==0) {
        kZLS_SmoothAxes = kSmoothXFlag;
      }
      if (strcmp(str1, "Y")==0) {
        kZLS_SmoothAxes = kSmoothYFlag;
      }
      if (strcmp(str1, "Z")==0) {
        kZLS_SmoothAxes = kSmoothZFlag;
      }
      if (strcmp(str1, "XY")==0) {
        kZLS_SmoothAxes = kSmoothXFlag + kSmoothYFlag;
      }
      if (strcmp(str1, "YZ")==0) {
        kZLS_SmoothAxes = kSmoothYFlag + kSmoothZFlag;
      }
      if (strcmp(str1, "ZX")==0) {
        kZLS_SmoothAxes = kSmoothZFlag + kSmoothXFlag;
      }
      
      if (strcmp(str1, "XYZ")==0) {
        kZLS_SmoothAxes = kSmoothXFlag + kSmoothYFlag + kSmoothZFlag;
      }
      if (strcmp(str1, "All")==0) {
        kZLS_SmoothAxes = kSmoothAll;
      }
      
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "PPM_ViscousAxes", str1);
#endif
    }
    
    ccl_release(&config);
    
  } // config
  
  return (err);
  
}

// get the fiducial scales from a config file, usually physics.conf
int   zlsio_InitFiducialScalesConfigFile(char * confFile){
  
  int err = noErr;
  
  Real x0, v0, d0;
  Real muB;
  
  struct ccl_t  config;
  const char    *str1;
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, confFile)==noErr) {
    
    x0 = 1.0;
    
    if ((str1= ccl_get( &config,"Length_Scale_x0"))){
      x0 = atof( str1);
    }
    
    v0 = 1.0;
    if ((str1= ccl_get( &config,"Velocity_Scale_v0"))){
      v0 = atof( str1);
    }
    
    d0 = 1.0;
    if ((str1= ccl_get( &config,"Density_Scale_d0"))){
      d0 = atof( str1);
    }
        
    strcpy(s_Time, "t");
    if ((str1= ccl_get( &config,"Time_Unit_Name"))){
      strcpy(s_Time,  str1);
    }
    
    strcpy(s_Length, "x");
    if ((str1= ccl_get( &config,"Length_Unit_Name"))){
      strcpy(s_Length,  str1);
    }
    
    strcpy(s_Area, "A");
    if ((str1= ccl_get( &config,"Area_Unit_Name"))){
      strcpy(s_Area,  str1);
    }
    
    strcpy(s_Volume, "Vol");
    if ((str1= ccl_get( &config,"Volume_Unit_Name"))){
      strcpy(s_Volume,  str1);
    }
    
    strcpy(s_Velocity, "v");
    if ((str1= ccl_get( &config,"Velocity_Unit_Name"))){
      strcpy(s_Velocity,  str1);
    }
    
    strcpy(s_Accel, "a");
    if ((str1= ccl_get( &config,"Accel_Unit_Name"))){
      strcpy(s_Accel,  str1);
    }
    
    strcpy(s_Mass, "m");
    if ((str1= ccl_get( &config,"Mass_Unit_Name"))){
      strcpy(s_Mass,  str1);
    }
    
    strcpy(s_Density, "d");
    if ((str1= ccl_get( &config,"Density_Unit_Name"))){
      strcpy(s_Density,  str1);
    }
    
    strcpy(s_Pressure, "P");
    if ((str1= ccl_get( &config,"Pressure_Unit_Name"))){
      strcpy(s_Pressure,  str1);
    }
    
    strcpy(s_Energy, "E");
    if ((str1= ccl_get( &config,"Energy_Unit_Name"))){
      strcpy(s_Energy,  str1);
    }
    
    
    strcpy(s_Temperature, "K");
    if ((str1= ccl_get( &config,"Temperature_Unit_Name"))){
      strcpy(s_Temperature,  str1);
    }
    
#ifdef COOLING    
    strcpy(s_Lambda, "L");
    if ((str1= ccl_get( &config,"Lambda_Unit_Name"))){
      strcpy(s_Lambda,  str1);
    }
    
    strcpy(s_Cool, "C");
    if ((str1= ccl_get( &config,"Cooling_Unit_Name"))){
      strcpy(s_Cool,  str1);
    }
#endif COOLING    
#ifdef GRAVITY    
    strcpy(s_Gravity, "G");
    if ((str1= ccl_get( &config,"Gravity_Unit_Name"))){
      strcpy(s_Gravity,  str1);
    }
#endif GRAVITY    
    
    err = zlsp_SetFiducialScales ( x0,  v0, d0, 1.0 );

    ccl_release(&config);
    
  } // config
  
  return (err);
  
}


//
// These are subroutines to make reading the physics config easier
// they take already opened config structures, not new file names
//
//
#ifdef COOLING
int   zlsio_SetUpCoolingFromConfigP(CoolingPtr cp, char * coolFileName, struct ccl_t  * configP);
int   zlsio_SetUpCoolingFromConfigP(CoolingPtr cp, char * coolFileName, struct ccl_t  * configP){
  int err = noErr;
  
  const char  *str1;
  
  Integer     cType;
  Real       cScale;
  Real         minT;
  Real         maxT;
  Real     q1 = 0.0;
  
  if ( cp != NULL ) {
    
    Real cTime = kZLS_CoolTimeFraction;
    if ((str1= ccl_get(configP, "Cool_Timing"))){
      cTime  = atof( str1 );
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Cool_Timing", str1);
#endif
    }
    kZLS_CoolTimeFraction = cTime;
    
    //    cScale =  cp->coolScale;
    cScale = 1.0;
    if ((str1= ccl_get( configP, "Cooling_Scale"))){
      cScale  = atof( str1 );
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Cooling_Scale", str1);
#endif
    }
    cp->coolScale = cScale;
    
    cType = kPhysNoCooling;
    if ((str1= ccl_get( configP, "Cooling_Type"))){
      
      if (strcmp(str1, "0")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "no")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "NO")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "NONE")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "No")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "N")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "None")==0) {
        cType = kPhysNoCooling;
      }
      if (strcmp(str1, "none")==0) {
        cType = kPhysNoCooling;
      }
      
      
      if (strcmp(str1, "1")==0) {
        cType = kPhysFileCooling;
      }
      if (strcmp(str1, "file")==0) {
        cType = kPhysFileCooling;
      }
      if (strcmp(str1, "File")==0) {
        cType = kPhysFileCooling;
      }
      if (strcmp(str1, "FILE")==0) {
        cType = kPhysFileCooling;
      }
      
      
      if (strcmp(str1, "2")==0) {
        cType = kPhysPowerCooling;
      }
      if (strcmp(str1, "Powerlaw")==0) {
        cType = kPhysPowerCooling;
      }
      if (strcmp(str1, "powerlaw")==0) {
        cType = kPhysPowerCooling;
      }
      if (strcmp(str1, "power")==0) {
        cType = kPhysPowerCooling;
      }
      if (strcmp(str1, "Power")==0) {
        cType = kPhysPowerCooling;
      }
      
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Cooling_Type", str1);
#endif
    }
    cp->coolType = cType;
    
    if ( cType == kPhysNoCooling ) {
      err = noErr;
    }
    
    if ( cType == kPhysFileCooling ) {
      
      if ((str1= ccl_get( configP, "Cooling_File"))){
        strcpy( coolFileName,  str1);
      } else {
        strcpy( coolFileName,  "data/DiffuseAtomic.dat" );
      }
      
      // if not just a total function then set individual components
      Integer cCoolComponent = cp->coolFlagAt;
      if ((str1= ccl_get( configP, "Cooling_Atomic"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->coolFlagAt = cCoolComponent;
      
      cCoolComponent = cp->coolFlagH2;
      if ((str1= ccl_get( configP, "Cooling_H2"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->coolFlagH2 = cCoolComponent;
      
      cCoolComponent = cp->coolFlagCO;
      if ((str1= ccl_get( configP, "Cooling_CO"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->coolFlagCO = cCoolComponent;
      
      
      err = zlsp_SetPhysicsCoolingFile( cp, coolFileName );
      
      cCoolComponent = cp->heatFlagCX;
      if ((str1= ccl_get( configP, "Heating_CX"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->heatFlagCX = cCoolComponent;
      
      cCoolComponent = cp->heatFlagHI;
      if ((str1= ccl_get( configP, "Heating_HI"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->heatFlagHI = cCoolComponent;
      
      cCoolComponent = cp->heatFlagGR;
      if ((str1= ccl_get( configP, "Heating_GR"))){
        cCoolComponent = strtol( str1, NULL, 0 );
      }
      cp->heatFlagGR = cCoolComponent;
      
    }
    
    if ( cType == kPhysPowerCooling ) { // get the rest of the parameters
      
      Real cLam0 = cp->coolLambda0; // cgs
      
      if ((str1= ccl_get( configP, "Cooling_Lambda0"))){
        cLam0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Cooling_Lambda0", str1);
#endif
      }
      
      q1   = (k_mu*k_amu);
      q1   = 1.0/(q1*q1);
      cLam0 *= q1;  // cooling per density^2
      // now normalise by f_invlam0 to get model units
      cp->coolLambda0 = cLam0*f_invlam0*cp->coolScale;  // normalised coolng rate
      
      Real cT0 =  cp->coolT0;
      if ((str1= ccl_get( configP, "Cooling_Temperature"))){
        cT0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Cooling_Temperature", str1);
#endif
      }
      cp->coolT0 = 1.0/cT0; // inverse is used (T/T0) == T * invT0;
      
      Real cPower =  cp->coolPower;
      if ((str1= ccl_get( configP, "Cooling_Alpha"))){
        cPower = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Cooling_Alpha", str1);
#endif
      }
      cp->coolPower = cPower;
      
      err = noErr;
    }
    
    minT = cp->coolTTableMin;
    if ((str1= ccl_get( configP, "Cooling_Min_Temp"))){
      minT = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Cooling_Min_Temp", str1);
#endif
    }
    cp->coolMinT = minT;
    cp->coolMinPd = f_invtep0*minT;
    
    maxT = cp->coolTTableMax;
    if (( str1 = ccl_get( configP, "Cooling_Max_Temp"))){
      maxT = atof( str1);
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Cooling_Max_Temp", str1);
#endif
    }
    cp->coolMaxT  = maxT;
    cp->coolMaxPd = f_invtep0*maxT;
    
      
  } else {
    err = memErr; 
  }
  return(err);
}  
#endif COOLING


int   zlsio_SetUpPhysicsFromConfigFile(char * confFile){
  
  int err = noErr;
  
  PhysicsPtr        phys = NULL;
  Integer          level;
#ifdef COOLING
  char  coolFileName[PATHLEN];
#endif COOLING
  
  struct ccl_t  config;
  const char     *str1;
  
  err = zlsp_NewPhysics(&phys);
  
  if ( phys != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0)
     zlsio_DisplayPhysicsSetup  ( stdout , confFile,  NULL, 1 );
#endif (ZLS_VERBOSE > 0)
      
      phys->mu = k_mu;
      if ((str1= ccl_get( &config, "Plasma_mu"))){
        phys->mu = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Plasma_mu", str1);
#endif
      }
      k_mu  = phys->mu ; // mean mol weight per particle, 0.6 ionised, 1.3 neutral, 2.35 molecular
      k_invmu = 1/ k_mu;
      f_tep0    = f_v0*f_v0*k_mu*k_amu/k_boltz; // temperature scaling factor, depends on k_mu:  T = f_tep0 * P/d
      f_invtep0    = 1.0/f_tep0; // inverse temperature scaling factor P = d* T* f_invtep0
      
      
      Real gamma = kZLS_eos_gamma;
      if ((str1= ccl_get( &config, "Plasma_gamma"))){
        gamma = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Plasma_gamma", str1);
#endif
      }
      
      zlsp_InitGlobalEOS_Gamma( phys, gamma);
      
      //      err = zlsio_InitFiducialScalesConfigFile(confFile);
      
#ifdef COOLING  
      err += zlsio_SetUpCoolingFromConfigP(&(phys->coolParameters), coolFileName,  &config);
#endif COOLING    
            
#ifdef GRAVITY
      err += zlsio_SetUpGravityFromConfigP(&(phys->gravPotential),  &config);
#endif GRAVITY
      
      ccl_release(&config);
      
    } // config 
    
    gLevels[0].levelPhysics = phys;
    
    
    if (gNLevels>1){
      for (level = 1; level < gNLevels; level++){
        gLevels[level].levelPhysics = gLevels[0].levelPhysics;
      }
    }
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 PhysicsPtr \n" );
    return (err);
  }
  
#if (ZLS_VERBOSE >0)
#ifdef COOLING
  zlsio_DisplayPhysicsSetup  ( stdout , confFile, coolFileName, 0 );
#else
  zlsio_DisplayPhysicsSetup  ( stdout , confFile,  NULL, 0 );
#endif !COOLING
#endif
  
  return (err);
  
}

int zlsio_SetUpLevelsFromConfigFile(char* confFile ){
  int err = noErr;
  
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  Integer lRefine    = 1;
  Integer lBCells    = 5;
  Integer lInterp    = kInterpNN;
  Integer lNLevels   = 1;
  
  Integer cumulativeRefine;
  
  GridCoords         superLevelGC;
  GridCoords         levelGC;
  Range              rng;
  
  Integer l0Refine = 1;
  
  Integer globalCellsScaling  = 1; // scaling factor >= 1 to scale up all cell index ranges
  Integer globalCellsDivisor  = 1; // scaling factor >= 1 to scale down all cell index ranges
  Real CoordinateScaling = 1.0; // units for coord numbers, if left as 1.0 then use normalised units
  Real level0XL[3];
  Real level0XR[3];
  
  Integer levelAxisType[3];
  Integer levelAxisKind[3];
  
  Integer levelN0[3];
  Integer levelRN[3];
  
  char keyStr[256];
  
  
  TimeCoords cl;  
  zlst_InitTimeCoords  ( &cl );
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, confFile)==noErr) {
    
    
#if (ZLS_VERBOSE >0)
    err = zlsio_PrintConfiguration( stdout , "" );
    zlsio_Printf( "%s\n", DIVIDERLINE );
#endif
    
    lRefine = 4;
    if ((str1= ccl_get( &config,"Refine_Cells"))){
      lRefine = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Refine_Cells", lRefine);
#endif (ZLS_DEBUG > 0)
    }
    
    lBCells = lRefine;
    if ((str1= ccl_get( &config,"Boundary_Cells"))){
      lBCells = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Boundary_Cells", lBCells);
#endif (ZLS_DEBUG > 0)
    }
    
    if ((str1= ccl_get( &config,"Interpolation"))){ // this is ther order of super zone BC interpolation
      lInterp = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Interpolation", lInterp);
#endif
    }
    
    globalCellsScaling = 1;
    if ((str1= ccl_get( &config,"Global_Cell_Scale"))){
      globalCellsScaling = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Global_Cell_Scale", globalCellsScaling);
#endif
    }
    
    globalCellsDivisor = 1;
    if ((str1= ccl_get( &config,"Global_Cell_Div"))){
      globalCellsDivisor = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Global_Cell_Div", globalCellsDivisor);
#endif
    }
    
    CoordinateScaling = 1.0;
    if ((str1= ccl_get( &config,"Coordinate_Scale_pc"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_pc", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )*f_cmpc/f_x0;
    }
    
    if ((str1= ccl_get( &config,"Coordinate_Scale_kpc"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_kpc", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )*f_cmkpc/f_x0;
    }
    
    if ((str1= ccl_get( &config,"Coordinate_Scale_AU"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_AU", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )*f_cmau/f_x0;
    }
    
    if ((str1= ccl_get( &config,"Coordinate_Scale_km"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_km", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )*f_cmkm/f_x0;
    }
    
    if ((str1= ccl_get( &config,"Coordinate_Scale_m"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_m", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )*100.0/f_x0;
    }
    
    if ((str1= ccl_get( &config,"Coordinate_Scale_cm"))){
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %g\n", "Coordinate_Scale_cm", CoordinateScaling);
#endif
      CoordinateScaling = atof( str1 )/f_x0;
    }
    
    //
    if ((str1= ccl_get( &config,"NLevels"))){
      lNLevels = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "NLevels", lNLevels);
#endif
    }
    
    if ((str1= ccl_get( &config,"Geometry"))){
#ifdef SPHERICAL
      if (strcmp(str1, "Spherical")!=0) {
        zlsio_ErrPrintf( " ERROR: Compiled Spherical: Incompatible Geometry %s \n", str1);
        ccl_release(&config);
        return(-1);
      }
#endif SPHERICAL
#ifdef CARTESIAN
      if (strcmp(str1, "Cartesian")!=0) {
        zlsio_ErrPrintf( " ERROR: Compiled Cartesian: Incompatible Geometry %s \n", str1);
        ccl_release(&config);
        return(-1);
      }
#endif CARTESIAN
#ifdef CYLINDRICAL
      if (strcmp(str1, "Cylindrical")!=0) {
        zlsio_ErrPrintf( " ERROR: Compiled Cylindrical: Incompatible Geometry %s \n", str1);
        ccl_release(&config);
        return(-1);
      }
#endif CYLINDRICAL
#ifdef CIRCULAR
      if (strcmp(str1, "Circular")!=0) {
        zlsio_ErrPrintf( " ERROR: Compiled 2D Circular: Incompatible Geometry %s \n", str1);
        ccl_release(&config);
        return(-1);
      }
#endif CIRCULAR
#ifdef POLARSPHERICAL
      if (strcmp(str1, "PolarSpherical")!=0) {
        zlsio_ErrPrintf( " ERROR: Compiled Polar Spherical: Incompatible Geometry %s \n", str1);
        ccl_release(&config);
        return(-1);
      }
#endif POLARSPHERICAL
    }
    
    if ((lNLevels<=0)||(lNLevels>MAXLEVELS)){
      zlsio_ErrPrintf( " ERROR: bad level number - levels %d \n", lNLevels);
      ccl_release(&config);
      return(-1);
    }
    
    cumulativeRefine = 1;
    
    for (level = 0; level <lNLevels; level++){ 
      
      //========================================
      //
      // Setup levels 
      //
      //========================================
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " Creating Level_%2.2d\n", level);
#endif ZLS_DEBUG            
      
      
      // indices, for each level in terms of global or super-level indices
      levelN0[kXDim] = 0;        
#if ( NDIMS > 0 )       
      sprintf(keyStr, "Level_%2.2d_i0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kXDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
#if ( NDIMS > 1 )       
      levelN0[kYDim] = 0;        
      sprintf(keyStr, "Level_%2.2d_j0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kYDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      
#if ( NDIMS > 2 )       
      levelN0[kZDim] = 0;        
      sprintf(keyStr, "Level_%2.2d_k0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kZDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      
#if ( NDIMS > 0 )       
      sprintf(keyStr, "Level_%2.2d_ni",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelRN[kXDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
#if ( NDIMS > 1 )       
      sprintf(keyStr, "Level_%2.2d_nj",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelRN[kYDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      
#if ( NDIMS > 2 )       
      sprintf(keyStr, "Level_%2.2d_nk",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelRN[kZDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      if (level == 0) {
        
        // global coordinates are set for level 0, and computed for all lower levels
#if ( NDIMS > 0 )       
        sprintf(keyStr, "Level_%2.2d_xl",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          level0XL[kXDim] = atof(str1)*CoordinateScaling;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        sprintf(keyStr, "Level_%2.2d_xr",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          level0XR[kXDim] = atof(str1)*CoordinateScaling;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#endif
        
#if ( NDIMS > 1 )       
        sprintf(keyStr, "Level_%2.2d_yl",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
#ifdef CARTESIAN
          level0XL[kYDim] = atof(str1)*CoordinateScaling;
#endif CARTESIAN
#ifdef CYLINDRICAL
          level0XL[kYDim] = atof(str1)*CoordinateScaling;
#endif CYLINDRICAL
#ifdef CIRCULAR
          level0XL[kYDim] = atof(str1);
#endif CIRCULAR
#ifdef SPHERICAL
          level0XL[kYDim] = atof(str1);
#endif SPHERICAL
#ifdef POLARSPHERICAL
          level0XL[kYDim] = atof(str1);
#endif POLARSPHERICAL
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        sprintf(keyStr, "Level_%2.2d_yr",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
#ifdef CARTESIAN
          level0XR[kYDim] = atof(str1)*CoordinateScaling;
#endif CARTESIAN
#ifdef CYLINDRICAL
          level0XR[kYDim] = atof(str1)*CoordinateScaling;
#endif CYLINDRICAL
#ifdef CIRCULAR
          level0XR[kYDim] = atof(str1);
#endif CIRCULAR
#ifdef SPHERICAL
          level0XR[kYDim] = atof(str1);
#endif SPHERICAL
#ifdef POLARSPHERICAL
          level0XR[kYDim] = atof(str1);
#endif POLARSPHERICAL
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#ifdef SPHERICAL
        // Scale angular values by pi if needed
        level0XL[kYDim] *= pi; // phi
        level0XR[kYDim] *= pi;
#endif SPHERICAL
#ifdef POLARSPHERICAL
        level0XL[kYDim] *= pi; // theta
        level0XR[kYDim] *= pi;
#endif POLARSPHERICAL
#ifdef CIRCULAR
        level0XL[kYDim] *= pi; //phi
        level0XR[kYDim] *= pi;
#endif CIRCULAR
#endif
        
        
#if ( NDIMS > 2 )       
        sprintf(keyStr, "Level_%2.2d_zl",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
#ifdef CARTESIAN
          level0XL[kZDim] = atof(str1)*CoordinateScaling;
#endif CARTESIAN
#ifdef CYLINDRICAL
          level0XL[kZDim] = atof(str1);
#endif CYLINDRICAL
#ifdef SPHERICAL
          level0XL[kZDim] = atof(str1);
#endif SPHERICAL
#ifdef POLARSPHERICAL
          level0XL[kZDim] = atof(str1);
#endif POLARSPHERICAL
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        sprintf(keyStr, "Level_%2.2d_zr",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
#ifdef CARTESIAN
          level0XR[kZDim] = atof(str1)*CoordinateScaling;
#endif CARTESIAN
#ifdef CYLINDRICAL
          level0XR[kZDim] = atof(str1);
#endif CYLINDRICAL
#ifdef SPHERICAL
          level0XR[kZDim] = atof(str1);
#endif SPHERICAL
#ifdef POLARSPHERICAL
          level0XR[kZDim] = atof(str1);
#endif POLARSPHERICAL
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#ifdef SPHERICAL
        level0XL[kZDim] *= pi; // angle input in units of pi radians, cvt to radians for internal use
        level0XR[kZDim] *= pi;
#endif SPHERICAL
#ifdef POLARSPHERICAL
        level0XL[kYDim] *= pi; // phi
        level0XR[kYDim] *= pi;
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
        level0XL[kZDim] *= pi; // angle input in units of pi radians, cvt to radians for internal use
        level0XR[kZDim] *= pi;
#endif CYLINDRICAL
#endif
        
        // axis types
        
#if ( NDIMS > 0 )   
        
#ifdef CARTESIAN
        levelAxisType[kXDim] = kAxisTypeSpatial;
        levelAxisKind[kXDim] = kAxisKindLinear;
#endif CARTESIAN
#ifdef SPHERICAL
        levelAxisType[kXDim] = kAxisTypeSRadius;
        levelAxisKind[kXDim] = kAxisKindLinear;
#endif SPHERICAL
#ifdef POLARSPHERICAL
        levelAxisType[kXDim] = kAxisTypeSRadius;
        levelAxisKind[kXDim] = kAxisKindLinear;
#endif POLARSPHERICAL
#ifdef CIRCULAR
        levelAxisType[kXDim] = kAxisTypeCRadius;
        levelAxisKind[kXDim] = kAxisKindLinear;
#endif CIRCULAR
#ifdef CYLINDRICAL
        levelAxisType[kXDim] = kAxisTypeSpatial;
        levelAxisKind[kXDim] = kAxisKindLinear;
#endif CYLINDRICAL
        sprintf(keyStr, "Level_%2.2d_xKind",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG  
          if (levelAxisKind[kXDim] == kAxisTypeSpatial ){
            if (strcmp(str1, "Log")==0) {
              levelAxisKind[kXDim] = kAxisKindLog;
            }
            if (strcmp(str1, "Log10")==0) {
              levelAxisKind[kXDim] = kAxisKindLog;
            }
          }
        }
#endif  ( NDIMS > 0 )
        
#if ( NDIMS > 1 )
#ifdef CARTESIAN
        levelAxisType[kYDim] = kAxisTypeSpatial;
        levelAxisKind[kYDim] = kAxisKindLinear;
        sprintf(keyStr, "Level_%2.2d_yKind",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          if (strcmp(str1, "Log")==0) {
            levelAxisKind[kYDim] = kAxisKindLog;
          }
          if (strcmp(str1, "Log10")==0) {
            levelAxisKind[kYDim] = kAxisKindLog;
          }
        }
#endif CARTESIAN
        
#ifdef POLARSPHERICAL
        levelAxisType[kYDim] = kAxisTypePolarTheta;
        levelAxisKind[kYDim] = kAxisKindLinear;
#endif POLARSPHERICAL
#ifdef CIRCULAR
        levelAxisType[kYDim] = kAxisTypePhi;
        levelAxisKind[kYDim] = kAxisKindLinear;
#endif CIRCULAR
#ifdef SPHERICAL
        levelAxisType[kYDim] = kAxisTypePhi;
        levelAxisKind[kYDim] = kAxisKindLinear;
#endif SPHERICAL
#ifdef CYLINDRICAL
        levelAxisType[kYDim] = kAxisTypeCRadius;
        levelAxisKind[kYDim] = kAxisKindLinear;
#endif CYLINDRICAL
#endif
        
        
#if ( NDIMS > 2 )
#ifdef CARTESIAN
        levelAxisType[kZDim] = kAxisTypeSpatial;
        levelAxisKind[kZDim] = kAxisKindLinear;
        sprintf(keyStr, "Level_%2.2d_zKind",level);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          if (strcmp(str1, "Log")==0) {
            levelAxisKind[kYDim] = kAxisKindLog;
          }
          if (strcmp(str1, "Log10")==0) {
            levelAxisKind[kYDim] = kAxisKindLog;
          }
        }
#endif CARTESIAN
        
#ifdef POLARSPHERICAL
        levelAxisType[kZDim] = kAxisTypePhi;
        levelAxisKind[kZDim] = kAxisKindLinear;
#endif POLARSPHERICAL
#ifdef SPHERICAL
        levelAxisType[kZDim] = kAxisTypeTheta;
        levelAxisKind[kZDim] = kAxisKindLinear;
#endif SPHERICAL
#ifdef CYLINDRICAL
        levelAxisType[kZDim] = kAxisTypePhi;
        levelAxisKind[kZDim] = kAxisKindLinear;
#endif CYLINDRICAL
#endif
        
        
#ifdef SYM1D
        zls_SetRangeDims   (&rng, k1D, lBCells);
        
        zls_SetRangeType   (&rng, kXDim, levelAxisType[kXDim] );
        zls_SetRangeKind   (&rng, kXDim, levelAxisKind[kXDim] );
        
        zls_SetRange       (&rng, kXDim, levelN0[kXDim], levelN0[kXDim]+levelRN[kXDim], level0XL[kXDim], level0XR[kXDim] );
        zls_InitGridCoords  (&levelGC, k1D, lBCells, kAllocateAxes, &rng);
#endif SYM1D
        
#ifdef SYM2D
        zls_SetRangeDims   (&rng, k2D, lBCells);
        
        zls_SetRangeType   (&rng, kXDim, levelAxisType[kXDim] );
        zls_SetRangeKind   (&rng, kXDim, levelAxisKind[kXDim] );
        zls_SetRangeType   (&rng, kYDim, levelAxisType[kYDim] );
        zls_SetRangeKind   (&rng, kYDim, levelAxisKind[kYDim] );
        
        zls_SetRange       (&rng, kXDim, levelN0[kXDim], levelN0[kXDim]+levelRN[kXDim], level0XL[kXDim], level0XR[kXDim] );
        zls_SetRange       (&rng, kYDim, levelN0[kYDim], levelN0[kYDim]+levelRN[kYDim], level0XL[kYDim], level0XR[kYDim] );
        zls_InitGridCoords   (&levelGC, k2D, lBCells, kAllocateAxes, &rng);
#endif SYM2D
        
#ifdef SYM3D
        zls_SetRangeDims   (&rng, k3D, lBCells);
        
        zls_SetRangeType   (&rng, kXDim, levelAxisType[kXDim] );
        zls_SetRangeKind   (&rng, kXDim, levelAxisKind[kXDim] );
        zls_SetRangeType   (&rng, kYDim, levelAxisType[kYDim] );
        zls_SetRangeKind   (&rng, kYDim, levelAxisKind[kYDim] );
        zls_SetRangeType   (&rng, kZDim, levelAxisType[kZDim] );
        zls_SetRangeKind   (&rng, kZDim, levelAxisKind[kZDim] );
        
        zls_SetRange       (&rng, kXDim, levelN0[kXDim], levelN0[kXDim]+levelRN[kXDim], level0XL[kXDim], level0XR[kXDim] );
        zls_SetRange       (&rng, kYDim, levelN0[kYDim], levelN0[kYDim]+levelRN[kYDim], level0XL[kYDim], level0XR[kYDim] );
        zls_SetRange       (&rng, kZDim, levelN0[kZDim], levelN0[kZDim]+levelRN[kZDim], level0XL[kZDim], level0XR[kZDim] );
        
        zls_InitGridCoords   (&levelGC, k3D, lBCells, kAllocateAxes, &rng);
#endif SYM3D
        
#if (ZLS_VERBOSE > 1)
        zlsio_FPrintf( stdout, "\n%s\n",  DIVIDERLINE );
        zls_PrintRange   ( stdout, &rng);
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
        zlsio_FPrintf( stdout, "Level %d Global Grid: \n", level);
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
        zls_PrintGridCoords( stdout, &levelGC); 
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
#endif ZLS_VERBOSE            
        
#ifdef SYM1D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k1D, lBCells, l0Refine, lInterp);
#endif SYM1D
#ifdef SYM2D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k2D, lBCells, l0Refine, lInterp);
#endif SYM2D
#ifdef SYM3D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k3D, lBCells, l0Refine, lInterp);
#endif SYM3D
        
        zls_AllocateLevelZoneArray (level,  MAXLEVELZONES);
        
      } // level = 0
      
      if (level > 0) {
        
        cumulativeRefine *= lRefine;
        zls_GetLevelGrid(&superLevelGC, level-1);
        
#ifdef SYM1D
        zls_SetRangeDims(&rng, k1D, lBCells); 
        zls_GetSubRangeFromGridDelta(&rng, kXDim, &superLevelGC, levelN0[kXDim], levelRN[kXDim]); 
        rng.rn[kXDim]     *= lRefine;
        zls_InitGridCoords   (&levelGC, k1D, lBCells, kAllocateAxes, &rng);
#endif SYM1D
        
#ifdef SYM2D
        zls_SetRangeDims(&rng, k2D, lBCells);
        zls_GetSubRangeFromGridDelta(&rng, kXDim, &superLevelGC, levelN0[kXDim], levelRN[kXDim]);
        zls_GetSubRangeFromGridDelta(&rng, kYDim, &superLevelGC, levelN0[kYDim], levelRN[kYDim]);
        rng.rn[kXDim]     *= lRefine;
        rng.rn[kYDim]     *= lRefine;
        zls_InitGridCoords   (&levelGC, k2D, lBCells, kAllocateAxes, &rng);
#endif SYM2D
        
#ifdef SYM3D
        zls_SetRangeDims(&rng, k3D, lBCells); 
        zls_GetSubRangeFromGridDelta(&rng, kXDim, &superLevelGC, levelN0[kXDim], levelRN[kXDim]);
        zls_GetSubRangeFromGridDelta(&rng, kYDim, &superLevelGC, levelN0[kYDim], levelRN[kYDim]);
        zls_GetSubRangeFromGridDelta(&rng, kZDim, &superLevelGC, levelN0[kZDim], levelRN[kZDim]);
        rng.rn[kXDim]     *= lRefine;
        rng.rn[kYDim]     *= lRefine;
        rng.rn[kZDim]     *= lRefine;
        zls_InitGridCoords   (&levelGC, k3D, lBCells, kAllocateAxes, &rng);
#endif SYM3D
        
#if (ZLS_VERBOSE >1)
        zlsio_FPrintf( stdout, "\n%s\n",  DIVIDERLINE );
        zls_PrintRange   ( stdout, &rng);
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
        zlsio_FPrintf( stdout, "Level %d Global Grid: \n", level);
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
        zls_PrintGridCoords( stdout, &levelGC); 
        zlsio_FPrintf( stdout, "%s\n",  DIVIDERLINE );
#endif ZLS_VERBOSE            
        
#ifdef SYM1D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k1D, lBCells, lRefine, lInterp);
#endif SYM1D
#ifdef SYM2D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k2D, lBCells, lRefine, lInterp);
#endif SYM2D
#ifdef SYM3D
        zls_InitLevelList          (level, cl, kZLS_CFL, levelGC, k3D, lBCells, lRefine, lInterp);
#endif SYM3D
        
        zls_AllocateLevelZoneArray (level, MAXLEVELZONES);
        
      }
      
    } // level loop
    
    ccl_release(&config);
    
  } // parse config file
  
  gNLevels = lNLevels;
  
  
  if (err != noErr ){
    zlsio_ErrPrintf( "%s\n",  DIVIDERLINE );
    zlsio_ErrPrintf( "     ****** ERROR: Level Setup FAILED: %d *******\n", err);
    zlsio_ErrPrintf( "%s\n\n",  DIVIDERLINE );
  }
  
  return(err);
}

int zlsio_SetUpZonesFromConfigFile(char* confFile ){
  
  int err = noErr;
  
  struct ccl_t  config;
  const char    *str1;
  
  ZonePtr z=NULL;
  ZonePtr z1=NULL;
  ZonePtr zNext=NULL;
  
  Integer level;
  Integer nLevelZones;
  Integer nLevelLinks;
  Integer zone;
  Integer zlink;
  Integer direction;
  Integer varIndex;
  
  Integer bdryType;
  Integer calcKind;
  Integer newBCType;
  
  Integer globalCellsScaling  = 1; // scaling factor >= 1 to scale up all cell index ranges
  Integer globalCellsDivisor  = 1; // scaling factor >= 1 to scale down all cell index ranges
  
  GridCoords         levelGC;
  Range               rSuper;
  Range                  rng;
  Range               lvlRng;
  
  Integer zsl; // zone super level from sid;
  Integer zsi; // zone super index
  
  Integer zoneN0[3]= {0, 0, 0};
  Integer zoneRN[3]= {0, 0, 0};
  
  Integer zoneLN0[3] = {0, 0, 0};
  Integer levelN0[3] = {0, 0, 0};
  
  char keyStr[256];
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, confFile)==noErr) {
    
    globalCellsScaling = 1;
    if ((str1= ccl_get( &config,"Global_Cell_Scale"))){
      globalCellsScaling = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Global_Cell_Scale", globalCellsScaling);
#endif
    }
    
    globalCellsDivisor = 1;
    if ((str1= ccl_get( &config,"Global_Cell_Div"))){
      globalCellsDivisor = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG > 0)
      zlsio_Printf( " %s = %d\n", "Global_Cell_Div", globalCellsDivisor);
#endif
    }
    
#ifdef MPI      
    gMPIBlocking = 1;
#endif MPI      
    
    for (level = 0; level <gNLevels; level++){ 
      
      nLevelZones = 0;
      sprintf(keyStr, "Level_%2.2d_Zones",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        nLevelZones = strtol( str1, NULL, 0 );
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
      
#ifdef MPI      
      if ( level == 0 ) {
        gMPIBlocking = nLevelZones/gMPIsize;
      }
      gMPIBlocking = (gMPIBlocking < 1)? 1: gMPIBlocking;
#endif MPI      
      
      // indices, for each level in terms of global or super-level indices
      levelN0[kXDim] = 0;        
#if ( NDIMS > 0 )       
      sprintf(keyStr, "Level_%2.2d_i0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kXDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
#if ( NDIMS > 1 )       
      levelN0[kYDim] = 0;        
      sprintf(keyStr, "Level_%2.2d_j0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kYDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      
#if ( NDIMS > 2 )       
      levelN0[kZDim] = 0;        
      sprintf(keyStr, "Level_%2.2d_k0",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        levelN0[kZDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
      } else {
        zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
        ccl_release(&config);
        return(-1);
      }
#endif
      
      //========================================
      //
      // zones  
      //
      //========================================
      
      for (zone = 0; zone < nLevelZones; zone++){
        
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " Creating Zone_%2.2d:%2.2d\n", level, zone);
        zlsio_Printf( "%s\n", DIVIDERLINE);
#endif ZLS_DEBUG            
        
#if (NDIMS >0)
        sprintf(keyStr, "Zone_%2.2d:%2.2d_i0",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneN0[kXDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        
        sprintf(keyStr, "Zone_%2.2d:%2.2d_ni",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneRN[kXDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#endif
        
#if (NDIMS >1)
        sprintf(keyStr, "Zone_%2.2d:%2.2d_j0",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneN0[kYDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        
        
        sprintf(keyStr, "Zone_%2.2d:%2.2d_nj",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneRN[kYDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#endif
        
        
#if (NDIMS >2)
        sprintf(keyStr, "Zone_%2.2d:%2.2d_k0",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneN0[kZDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
        
        sprintf(keyStr, "Zone_%2.2d:%2.2d_nk",level,zone);
        if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
          zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
          zoneRN[kZDim] = strtol( str1, NULL, 0 )*globalCellsScaling/globalCellsDivisor;
        } else {
          zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
          ccl_release(&config);
          return(-1);
        }
#endif
        
        if (level == 0) {
          
          zls_GetLevelGrid         ( &levelGC, level ); 
          zls_GetRangeDimsFromGrid ( &rng,  &levelGC );
          
          err = zls_NewZone(&z);
          
#ifdef SYM1D
          zls_GetSubRangeFromGridDelta(&rng, kXDim, &levelGC, zoneN0[kXDim], zoneRN[kXDim]);   // specific sub range
#endif SYM1D
          
#ifdef SYM2D
          zls_GetSubRangeFromGridDelta(&rng, kXDim, &levelGC, zoneN0[kXDim], zoneRN[kXDim]); 
          zls_GetSubRangeFromGridDelta(&rng, kYDim, &levelGC, zoneN0[kYDim], zoneRN[kYDim]);
#endif SYM2D
          
#ifdef SYM3D
          zls_GetSubRangeFromGridDelta(&rng, kXDim, &levelGC, zoneN0[kXDim], zoneRN[kXDim]); 
          zls_GetSubRangeFromGridDelta(&rng, kYDim, &levelGC, zoneN0[kYDim], zoneRN[kYDim]);
          zls_GetSubRangeFromGridDelta(&rng, kZDim, &levelGC, zoneN0[kZDim], zoneRN[kZDim]);
#endif SYM3D
          lvlRng =  rng; // zone range on level is same as rng for level 0
          
          err = zls_InitZone(z, kAlgorithmZoneType, level, &lvlRng,  NULL,  &rng);
          
#if (ZLS_DEBUG == 1)
          zls_PrintGridCoords(stdout, &z->zoneGrid); 
#endif ZLS_DEBUG            
#if (ZLS_DEBUG > 1)
          zls_PrintZoneCoords(stdout, z); 
#endif ZLS_DEBUG    
          
#ifdef MPI
          if (z->zoneRankID == gMPIrank){ 
            // only actually allocate data for process' zone, zones for other ranks in this 
            // process are complete except for the zonedata vars whish are empty.
            // All zone structures which are small are allocated to make interzone queries simple.
#endif MPI
            for ( varIndex = 0; varIndex < z->nQ; varIndex++) {
              err = zls_NewZoneData(z, varIndex); // allocate data to work on  
#ifdef MEANSPLITSWEEPS
              err = zls_NewZoneDataAlt(z, varIndex); // allocate data to work on  
#endif MEANSPLITSWEEPS
            }         
            
            err = zls_NewZoneAuxilliaryData (z); // allocate aux arrays
            err = zls_NewZoneField ( z );
            
#ifdef PRIMITIVE_QUAS
            // must be called after all the datas have been allocted successfully to set up zoneW reference
            err = zls_SetZoneWHD( z );
#endif PRIMITIVE_QUAS
#ifdef CONSERVED_QUAS
            // must be called after all the datas have been allocated successfully to set up zoneW reference
            err = zls_SetZoneUHD( z );
#endif CONSERVED_QUAS
            
#ifdef MPI
          }
#endif MPI
          
#if (ZLS_DEBUG > 0)
          zlsio_Printf( "\n Level 0 Zone Created: ID: %d \n", zone);
#endif 
          char keyStr2[512];
          
          // set BCs
          for (direction = 0; direction < kNGeomDirs; direction++){
            
            // default: Computed free out
            
            bdryType   = kBoundaryTypeFlagComputed;
            calcKind   = kBCCKindFree;
            newBCType  =  bdryType | calcKind;
            
            sprintf(keyStr2, "BCType_Zone_%2.2d:%2.2d_Dir_%s", level, zone, zls_dirname(direction));
            if ((str1= ccl_get( &config, keyStr2))){
#if (ZLS_DEBUG > 0)
              zlsio_Printf( "          %s %s \n",keyStr2, str1);
#endif                
              
              if (strcmp(str1, "Computed")==0) {
                
                bdryType = kBoundaryTypeFlagComputed;
                sprintf(keyStr2, "BCKind_Zone_%2.2d:%2.2d_Dir_%s", level, zone, zls_dirname(direction));
                
                if ((str1= ccl_get( &config, keyStr2))){
#if (ZLS_DEBUG > 0)
                  zlsio_Printf( "          %s %s \n",keyStr2, str1);
#endif                
                  if (strcmp(str1, "User")==0) {
                    calcKind = kBCCKindUser;
                  }
                  if (strcmp(str1, "Free")==0) {
                    calcKind = kBCCKindFree;
                  }
                  if (strcmp(str1, "Plus1")==0) {
                    calcKind = kBCCKindFreer;
                  }
                  if (strcmp(str1, "OutOnly")==0) {
                    calcKind = kBCCKindFreeOutOnly;
                  }
                  if (strcmp(str1, "Extrapolated")==0) {
                    calcKind = kBCCKindExtrapol;
                  }
                  if (strcmp(str1, "Reflected")==0) {
                    calcKind = kBCCKindReflected;
                  }
                  if (strcmp(str1, "Reflecting")==0) {
                    calcKind = kBCCKindReflected;
                  }
                  if (strcmp(str1, "Periodic")==0) {
                    calcKind = kBCCKindPeriodic;
                  }
                  if (strcmp(str1, "Constant")==0) {
                    calcKind = kBCCKindConstant;
                  }
                  
                }
              }
              
              if (strcmp(str1, "Super")==0) { // error as super zone has no super, warn and continue
                
                zlsio_ErrPrintf( " Warning: Super zone has no super for BC updates, using computed free BCs\n" );
                
                bdryType = kBoundaryTypeFlagComputed;
                calcKind = kBCCKindFree;
                
              }
              
            } // found a BCType
            
            // set the bdry type + kind, default if no keys found above
            newBCType  =  bdryType | calcKind;
            err = zls_SetZoneBCType (z, NULL, direction, newBCType );
            
            
          } // direction
          
          zsl = 0;
          zsi = zone;
          
        } else { // get super zone id
          
          sprintf(keyStr, "Zone_%2.2d:%2.2d_sid",level,zone);
          if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
            zlsio_Printf( " Current Zone Super ID: %s %s \n", keyStr, str1);
#endif ZLS_DEBUG            
            zsl = level-1;
            zsi = 0;
            err = zls_RangeOptions((char *)str1, &zsl, &zsi);
#if (ZLS_DEBUG > 1)
            zlsio_Printf( "    Range options = %d %d \n",zsl, zsi);
#endif ZLS_DEBUG            
          } else {
            zlsio_ErrPrintf( " Error %s NOT Found in %s \n",keyStr,confFile );
            ccl_release(&config);
            return(-1);
          }
          
          z = gLevels[zsl].levelZones[zsi];
          
          if (z != NULL) {
            
            // get zone range in and coords from super zone.
            // and get zone range in level for level image if used.
            
            zls_GetLevelGrid ( &levelGC, level); 
            
            zls_GetRangeDimsFromGrid( &rng   ,  &levelGC ); // set just ndims and axis types
            zls_GetRangeDimsFromGrid( &lvlRng,  &levelGC );
            
            rSuper = z->zoneLevelRange;
            Integer levelRefine = gLevels[level].levelRefinement;
            
            err = zls_NewZone(&z1);
            
#ifdef SYM1D
            // coords and range in super
            zls_GetSubRangeFromGridDelta(&rng, kXDim,  &(z->zoneGrid), zoneN0[kXDim], zoneRN[kXDim]); 
            
            zoneLN0[kXDim] = rSuper.n0[kXDim]-levelN0[kXDim]+zoneN0[kXDim];
            
            
#if (ZLS_DEBUG > 1)
            zlsio_Printf( "    Level %d SubZone ID: %d  Created \n", level, zone);
            zlsio_Printf( "                     X : %d %d %d %d  final: %d %d \n", 
                         rSuper.n0[kXDim],
                         levelN0[kXDim],
                         zoneN0[kXDim],
                         rSuper.n0[kXDim]-levelN0[kXDim]+zoneN0[kXDim],
                         zoneLN0[kXDim]*levelRefine, zoneRN[kXDim]*levelRefine);
#endif (ZLS_DEBUG > 1)
            
            zoneRN [kXDim] *= levelRefine;
            zoneLN0[kXDim] *= levelRefine;
            
            // range in whole level...
            zls_GetSubRangeFromGridDelta(&lvlRng, kXDim,  &levelGC, zoneLN0[kXDim], zoneRN[kXDim]); 
#endif SYM1D
            
#ifdef SYM2D
            // coords and range in super
            zls_GetSubRangeFromGridDelta(&rng, kXDim,  &(z->zoneGrid), zoneN0[kXDim], zoneRN[kXDim]); 
            zls_GetSubRangeFromGridDelta(&rng, kYDim,  &(z->zoneGrid), zoneN0[kYDim], zoneRN[kYDim]);
            
            // range in whole level...
            zoneLN0[kXDim] = (rSuper.n0[kXDim]-levelN0[kXDim] +zoneN0[kXDim]);
            zoneLN0[kYDim] = (rSuper.n0[kYDim]-levelN0[kYDim] +zoneN0[kYDim]);
            
#if (ZLS_DEBUG > 1)
            zlsio_Printf( "    Level %d SubZone ID: %d  Created \n", level, zone);
            zlsio_Printf( "                     X : %d %d %d %d  final: %d %d \n", 
                         rSuper.n0[kXDim],
                         levelN0[kXDim],
                         zoneN0[kXDim],
                         rSuper.n0[kXDim]-levelN0[kXDim]+zoneN0[kXDim],
                         zoneLN0[kXDim]*levelRefine, zoneRN[kXDim]*levelRefine);
            
            zlsio_Printf( "                     Y : %d %d %d %d  final: %d %d \n", 
                         rSuper.n0[kYDim],
                         levelN0[kYDim],
                         zoneN0[kYDim],
                         rSuper.n0[kYDim]-levelN0[kYDim]+zoneN0[kYDim],
                         zoneLN0[kYDim]*levelRefine, zoneRN[kYDim]*levelRefine);
            
#endif (ZLS_DEBUG > 1)
            
            zoneRN [kXDim] *= levelRefine;
            zoneRN [kYDim] *= levelRefine;
            
            zoneLN0[kXDim] *= levelRefine;
            zoneLN0[kYDim] *= levelRefine;
            
            zls_GetSubRangeFromGridDelta(&lvlRng, kXDim,  &levelGC, zoneLN0[kXDim], zoneRN[kXDim]); 
            zls_GetSubRangeFromGridDelta(&lvlRng, kYDim,  &levelGC, zoneLN0[kYDim], zoneRN[kYDim]); 
#endif SYM2D
            
#ifdef SYM3D
            // coords and range in super
            zls_GetSubRangeFromGridDelta(&rng, kXDim,  &(z->zoneGrid), zoneN0[kXDim], zoneRN[kXDim]); 
            zls_GetSubRangeFromGridDelta(&rng, kYDim,  &(z->zoneGrid), zoneN0[kYDim], zoneRN[kYDim]);
            zls_GetSubRangeFromGridDelta(&rng, kZDim,  &(z->zoneGrid), zoneN0[kZDim], zoneRN[kZDim]);
            
            // range in whole level...
            zoneLN0[kXDim] = rSuper.n0[kXDim]-levelN0[kXDim]+zoneN0[kXDim]; // get zone offset into level grid
            zoneLN0[kYDim] = rSuper.n0[kYDim]-levelN0[kYDim]+zoneN0[kYDim]; // and hence into level image
            zoneLN0[kZDim] = rSuper.n0[kZDim]-levelN0[kZDim]+zoneN0[kZDim]; // zone if used
            
#if (ZLS_DEBUG > 1)
            zlsio_Printf( "    Level %d SubZone ID: %d  Created \n", level, zone);
            zlsio_Printf( "                     X : %d %d %d %d  final: %d %d \n", 
                         rSuper.n0[kXDim],
                         levelN0[kXDim],
                         zoneN0[kXDim],
                         rSuper.n0[kXDim]-levelN0[kXDim]+zoneN0[kXDim],
                         zoneLN0[kXDim]*levelRefine, zoneRN[kXDim]*levelRefine);
            
            zlsio_Printf( "                     Y : %d %d %d %d  final: %d %d \n", 
                         rSuper.n0[kYDim],
                         levelN0[kYDim],
                         zoneN0[kYDim],
                         rSuper.n0[kYDim]-levelN0[kYDim]+zoneN0[kYDim],
                         zoneLN0[kYDim]*levelRefine, zoneRN[kYDim]*levelRefine);
            
            zlsio_Printf( "                     Z : %d %d %d %d final: %d %d \n", 
                         rSuper.n0[kZDim],
                         levelN0[kZDim],
                         zoneN0[kZDim],
                         rSuper.n0[kZDim]-levelN0[kZDim]+zoneN0[kZDim],
                         zoneLN0[kZDim]*levelRefine, zoneRN[kZDim]*levelRefine);
            
#endif (ZLS_DEBUG > 1)
            
            zoneRN[kXDim] *= levelRefine;
            zoneRN[kYDim] *= levelRefine;
            zoneRN[kZDim] *= levelRefine;
            
            zoneLN0[kXDim] *= levelRefine;
            zoneLN0[kYDim] *= levelRefine;
            zoneLN0[kZDim] *= levelRefine;
            
            zls_GetSubRangeFromGridDelta(&lvlRng, kXDim,  &levelGC, zoneLN0[kXDim], zoneRN[kXDim]); 
            zls_GetSubRangeFromGridDelta(&lvlRng, kYDim,  &levelGC, zoneLN0[kYDim], zoneRN[kYDim]); 
            zls_GetSubRangeFromGridDelta(&lvlRng, kZDim,  &levelGC, zoneLN0[kZDim], zoneRN[kZDim]); 
#endif SYM3D
            
            err = zls_InitZone(z1, kAlgorithmZoneType, level, &lvlRng, z,  &rng); 
            
#if (ZLS_DEBUG == 1)
            zls_PrintGridCoords(stdout, &z1->zoneGrid); 
#endif ZLS_DEBUG            
#if (ZLS_DEBUG > 1)
            zls_PrintZoneCoords(z1); 
#endif ZLS_DEBUG            
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "    Level %d SubZone ID: %d  Created \n", level, zone);
            zlsio_Printf( "%s\n", DIVIDERLINE);
#endif                
            
#ifdef MPI
            if (z1->zoneRankID == gMPIrank){ // only actually allocate data for proces' zone, others are empty but the
              // zone structures which are small are allocated to make interzone queries and relationships simple.
#endif MPI
              for ( varIndex = 0; varIndex < z1->nQ; varIndex++){
                err = zls_NewZoneData(z1, varIndex); // allocate data to work on 
#ifdef MEANSPLITSWEEPS
                err = zls_NewZoneDataAlt(z1, varIndex); // allocate data to work on  
#endif MEANSPLITSWEEPS
              }      
              
              err = zls_NewZoneAuxilliaryData  ( z1 );
              err = zls_NewZoneField           ( z1 );
              
#ifdef PRIMITIVE_QUAS
              // must be called after all the datas have been allocated successfully to set up zoneW reference
              err = zls_SetZoneWHD( z1 );
#endif PRIMITIVE_QUAS
#ifdef CONSERVED_QUAS
              // must be called after all the datas have been allocated successfully to set up zoneW reference
              err = zls_SetZoneUHD( z1 );
#endif CONSERVED_QUAS
              
#ifdef MPI
            }
#endif MPI
            
            // set normal subzone  BCs
            for (direction = 0; direction < kNGeomDirs; direction++){
              
              // default: Super Zone Fill
              
              bdryType = kBoundaryTypeFlagSuperZ;
              calcKind = 0;
              newBCType  =  bdryType | calcKind;
              
              sprintf(keyStr, "BCType_Zone_%2.2d:%2.2d_Dir_%s", level, zone, zls_dirname(direction));
              if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
                zlsio_Printf( "          %s %s \n",keyStr, str1);
#endif                
                if (strcmp(str1, "Computed")==0) {
                  
                  bdryType = kBoundaryTypeFlagComputed;
                  sprintf(keyStr, "BCKind_Zone_%2.2d:%2.2d_Dir_%s", level, zone,  zls_dirname(direction));
                  
                  if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
                    zlsio_Printf( "          %s %s \n",keyStr, str1);
#endif
                    
                    if (strcmp(str1, "User")==0) {
                      calcKind = kBCCKindUser;
                    }
                    if (strcmp(str1, "Free")==0) {
                      calcKind = kBCCKindFree;
                    }
                    if (strcmp(str1, "Plus1")==0) {
                      calcKind = kBCCKindFreer;
                    }
                    if (strcmp(str1, "OutOnly")==0) {
                      calcKind = kBCCKindFreeOutOnly;
                    }
                    if (strcmp(str1, "Extrapolated")==0) {
                      calcKind = kBCCKindExtrapol;
                    }
                    if (strcmp(str1, "Reflected")==0) {
                      calcKind = kBCCKindReflected;
                    }
                    if (strcmp(str1, "Reflecting")==0) {
                      calcKind = kBCCKindReflected;
                    }
                    if (strcmp(str1, "Periodic")==0) {
                      calcKind = kBCCKindPeriodic;
                    }
                    if (strcmp(str1, "Constant")==0) {
                      calcKind = kBCCKindConstant;
                    }
                    
                  }
                }
                
                if (strcmp(str1, "Super")==0) {
                  
                  bdryType = kBoundaryTypeFlagSuperZ;
                  calcKind = kBCCKindFree;
                  
                }
                
                
              } // found a BCType
              
              // set the bdry type + kind, default kBoundaryTypeFlagSuperZ if no keys found above
              newBCType  =  bdryType | calcKind;
              err = zls_SetZoneBCType (z1, NULL, direction, newBCType );
              
              
            } // direction
            
            
          } else {zlsio_ErrPrintf( " FATAL ERROR: super not found: %d : %d \n",zsl, zsl);
            err = -1;
            ccl_release(&config);
            return (err);
          } // z != NULL
          
        } // level> 0
        
      } //zones in level
      //
      //========================================
      //
      // Links
      //
      //========================================
      // All zones in the level are done, now check for any linked zones 
      // on the level.
      //
      
      nLevelLinks = 0;
      sprintf(keyStr, "Level_%2.2d_Links",level);
      if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
        zlsio_Printf( " %s = %s\n", keyStr, str1);
#endif ZLS_DEBUG            
        nLevelLinks = strtol( str1, NULL, 0 );
      }
      
      for (zlink = 0; zlink < nLevelLinks; zlink++){
        for (zone = 0; zone < nLevelZones; zone++){
          for (direction = 0; direction < kNGeomDirs; direction++){
            
            sprintf(keyStr, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s",zlink, level, zone, zls_dirname(direction));
            if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
              zlsio_Printf( "    Link %s %s \n",keyStr, str1);
#endif (ZLS_DEBUG > 0)
              zsl = 0;
              zsi = 0;
              err = zls_RangeOptions((char*)str1, &zsl, &zsi);
#if (ZLS_DEBUG > 0)
              zlsio_Printf( "    Link Range options = %d %d \n",zsl, zsi);
#endif (ZLS_DEBUG > 0)
              if (zsl == level) {
                z     = gLevels[zsl].levelZones[zone];
                zNext = gLevels[zsl].levelZones[zsi];
                err = zls_SetZoneBCLink( z, zNext, direction);
              }
            } // found a link
            
          } // direction
        } // zone
      }// zlink
      
      //========================================
      //
      // User BC overrides
      //
      //========================================
      // if any boundary gets a user override then 
      // a UserBC flag is set.  The zlsp_CalcUserZoneBCs routine is called
      // on zones with UserBCs after all other BCs have been set.
      //
      for (zone = 0; zone < nLevelZones; zone++){
        for (direction = 0; direction < kNGeomDirs; direction++){
          
          sprintf(keyStr, "BCUser_Zone_%2.2d:%2.2d_Dir_%s", level, zone, zls_dirname(direction));
          if ((str1= ccl_get( &config, keyStr))){
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "          %s %s \n",keyStr, str1);
#endif                
            // for now the value is ignored, just add the user BC flag to the zoneBCType
            // so the bcs will call the user bc routine after all other bc settings.
            z     = gLevels[level].levelZones[zone];
            z->zoneBCType[direction] |= kBoundaryTypeFlagUserBC;
            
          } // found a BCUser
          
        } // direction
      } // zone
      
      
      
    } // level loop
#if (ZLS_DEBUG > 1)
    zlsbc_PrintLevelBCsAndLinks (0);
#endif (ZLS_DEBUG > 1)
    
    ccl_release(&config);
  } // parse config file
  
#if (ZLS_VERBOSE >0)
  zlsio_DisplayGridSetup( stdout , confFile, 0 );
#endif
  
  if (err != noErr ){
    zlsio_ErrPrintf( "%s\n",  DIVIDERLINE );
    zlsio_ErrPrintf( "     ****** ERROR: Zone Setup FAILED: %d *******\n", err);
    zlsio_ErrPrintf( "%s\n\n",  DIVIDERLINE );
  }
  
  return(err);
}


//
//========================================
//

int zlsio_PrintMethodConfiguration( FILE * f, char * prefix ){
  
  zlsio_FPrintf(f, "%s%s\n", prefix , DIVIDERLINE );
  zlsio_FPrintf(f, "%s %s Simulation Configuration: \n", prefix, FYRIS );
  zlsio_FPrintf(f, "%s Fyris Version: %s \n", prefix, FYRIS_VERSION );
  zlsio_FPrintf(f, "%s%s\n", prefix , DIVIDERLINE );
  
  zlsio_FPrintf( f, "%s ZLS Algorithm Options:\n", prefix );
    
#ifdef HYDRO
#ifdef TORO 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Toro 1998 Full Solver.\n", prefix );
#endif TORO
  
#ifdef TOROTWOSHOCK 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Toro 1998, RSS Modified 2 Shock Solver.\n", prefix );
#endif TOROTWOSHOCK
  
#ifdef PVRS 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Toro Primitive Variable Approximate Solver.\n", prefix );
#endif PVRS
  
#ifdef AIRS 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Toro 1998 AIRS Solver.\n", prefix );
#endif AIRS
  
#ifdef COLLELLATWOSHOCK 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Collella 1984 Two Shock Solver (vh1, ppmlr).\n", prefix );
#endif COLLELLATWOSHOCK
  
#ifdef GROTHGOTTLIEB 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: Groth and Gottlieb 1988 Fast Exact Solver.\n", prefix );
#endif GROTHGOTTLIEB
  
#ifdef RSSTWOSHOCK 
  zlsio_FPrintf( f, "%s   ZLS RIEMANN: RSS Modified GG88 Two Shock Solver.\n", prefix );
#endif RSSTWOSHOCK
#endif HYDRO
  
#ifdef LAGRANGIAN 
  
  
#ifdef SPLITSWEEPS
  zlsio_FPrintf( f, "%s   ZLS METHOD: Split Lagrangian + Remap.\n", prefix );
  
#ifdef MEANSPLITSWEEPS
#if (NDIMS == 2)
  zlsio_FPrintf( f, "%s   ZLS METHOD: Symmetric Split Sweeps.\n", prefix );
#endif (NDIMS == 2)
#endif MEANSPLITSWEEPS
  
  
#else
  zlsio_FPrintf( f, "%s   ZLS METHOD: ERROR in sweep mode. UNsplit Lagrangian.\n", prefix );
  return(-1);
  
#endif !SPLITSWEEPS
  
  
#ifdef VISCOUS_REMAP
  zlsio_FPrintf( f, "%s   ZLS METHOD: Viscous Remap.\n", prefix );
#endif VISCOUS_REMAP
#ifdef DIVVHCORRECTION
  zlsio_FPrintf( f, "%s   ZLS METHOD: Velocity Divergence Dissipation.\n", prefix );
#endif DIVVHCORRECTION
  
#else
  
#ifdef EULERIAN
#ifdef UNSPLITSWEEP
  zlsio_FPrintf( f, "%s   ZLS METHOD: Unsplit Eulerian Sweeps.\n", prefix );
  return(-1);
#else
  zlsio_FPrintf( f, "%s   ZLS METHOD: ERROR in sweep mode. Split Eulerian.\n", prefix );
  return(-1);
#endif !UNSPLITSWEEP
#else
  zlsio_FPrintf( f, "%s   ZLS METHOD: ERROR in sweep mode. Unknown.\n", prefix );
  return(-1);
#endif !EULERIAN
  
#endif !LAGRANGIAN
  
#ifdef UNIFORM_SWEEPS 
  zlsio_FPrintf( f, "%s   ZLS METHOD: Multi-Level Uniform Sweeps.\n", prefix );
#endif UNIFORM_SWEEPS 
#ifdef REFINED_SWEEPS 
  zlsio_FPrintf( f, "%s   ZLS METHOD: Multi-Level Refined Sweeps.\n", prefix );
#endif REFINED_SWEEPS 
  
  
#ifdef HYDRO
  
#ifdef PPM 
  zlsio_FPrintf( f, "%s   ZLS METHOD: 3rd order PPM Cell interpolation.\n", prefix );
#endif PPM
  
#ifdef PLM 
  zlsio_FPrintf( f, "%s   ZLS METHOD: 2nd order PLM Cell interpolation.\n", prefix );
#endif PLM
#ifdef GODUNOV 
  zlsio_FPrintf( f, "%s   ZLS METHOD: 1st order Godunov Cell interpolation.\n", prefix );
#endif GODUNOV
  
#endif HYDRO
    
  
#ifdef COOLING 
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: Radiative Cooling Enabled.\n", prefix );
#endif COOLING

#ifdef VARIABLE_EOS
  
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: Variable EOS Enabled.\n", prefix );
#ifdef VARIABLE_GAMMA 
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: Variable Gamma Enabled.\n", prefix );
#endif VARIABLE_GAMMA
  
#ifdef H_IONISATION
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: H Ionisation Lookup Table Enabled.\n", prefix );
#endif H_IONISATION
    
#endif VARIABLE_EOS
  
#ifdef GRAVITY 
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: Gravitational Potential Field Enabled.\n", prefix );
#ifdef GRAVZFIELD 
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: Non-Radial Z-Field Potential.\n", prefix );
#endif GRAVZFIELD
#if (FIELD_RECON_ORDER == 1)
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: PLM Conserved Potential Reconstruction.\n", prefix );
#endif
#if (FIELD_RECON_ORDER == 2)
  zlsio_FPrintf( f, "%s   ZLS PHYSICS: PPM Conserved Potential Reconstruction.\n", prefix );
#endif
#endif GRAVITY
  
#ifdef PRIMITIVE_QUAS 
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data are physical (W) quantities.\n", prefix );
#endif PRIMITIVE_QUAS
#ifdef PRIMITIVE_VARS 
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data are physical (W) variables+coeffs.\n", prefix );
#endif PRIMITIVE_VARS
#ifdef CONSERVED_QUAS 
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data are conserved (U) quantities.\n", prefix );
#endif PRIMITIVE_QUAS
#ifdef CONSERVED_VARS 
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data are conserved (U) variables+coeffs.\n", prefix );
#endif CONSERVED_VARS
  
#if (NTRACERS > 0) 
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data have %d variables and %d tracers.\n", prefix, kNVars, kNTracers);
#else
  zlsio_FPrintf( f, "%s   ZLS VARS: Zone data have %d variables\n", prefix, kNVars);
#endif NTRACERS
  
#ifdef ACTIVE_FP_NOISE_CONTROL 
  zlsio_FPrintf( f, "%s   ZLS OPTIMIZE: Active Floating Point Noise Control Enabled.\n", prefix );
#ifdef CONTROL_REMAP 
  zlsio_FPrintf( f, "%s   ZLS OPTIMIZE: AFPNC Enabled in Remap.\n", prefix );
#endif CONTROL_REMAP
#ifdef CONTROL_EVOLVE 
  zlsio_FPrintf( f, "%s   ZLS OPTIMIZE: AFPNC Enabled in Evolve.\n", prefix );
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL
  
#ifdef PRETEST_W_VEL 
  zlsio_FPrintf( f, "%s   ZLS OPTIMIZE: W* velocity pre-test Enabled.\n", prefix );
#endif PRETEST_W_VEL
  
  return(noErr);
  
}


int zlsio_PrintGeomConfiguration( FILE * f , char * prefix ){
  
  zlsio_FPrintf( f, "%s ZLS Compiled Geometry Options:\n", prefix );
  
#ifdef SYM1D 
  zlsio_FPrintf( f, "%s   ZLS SYMMETRY: Zone data blocks 1D.\n", prefix );
#ifdef CARTESIAN 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Cartesian X\n", prefix );
#endif CARTESIAN
#ifdef SPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 1D Spherical Radius \n", prefix );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 1D Polar Spherical Radius\n", prefix );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 1D Cylindrical Z \n", prefix );
#endif CYLINDRICAL
#endif SYM1D
#ifdef SYM2D 
  zlsio_FPrintf( f, "%s   ZLS SYMMETRY: Zone data blocks 2D.\n", prefix );
#ifdef CARTESIAN 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Cartesian XY\n", prefix );
#endif CARTESIAN
#ifdef CIRCULAR 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 2D Circular Radius Phi\n", prefix );
#endif CIRCULAR
#ifdef SPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 2D Spherical Radius Phi \n", prefix );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 2D Polar Spherical Radius Phi\n", prefix );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: 2D Cylindrical Z Radius\n", prefix );
#endif CYLINDRICAL
#endif SYM2D
#ifdef SYM3D 
  zlsio_FPrintf( f, "%s   ZLS SYMMETRY: Zone data blocks 3D.\n", prefix );
#ifdef CARTESIAN 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Cartesian XYZ\n", prefix );
#endif CARTESIAN
#ifdef CYLINDRICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Cylindrical Z Radius Phi\n", prefix );
#endif CYLINDRICAL
#ifdef SPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Spherical Radius Phi Theta\n", prefix );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
  zlsio_FPrintf( f, "%s   ZLS GEOMETRY: Polar Spherical Radius Theta Phi\n", prefix );
#endif POLARSPHERICAL
#endif SYM3D
  
  return(noErr);
  
}



int zlsio_PrintOSConfiguration( FILE * f , char * prefix ){
  
  zlsio_FPrintf( f, "%s ZLS Compiled OS/Architecture Options:\n", prefix );
  
#ifdef SINGLE_PRECISION 
  zlsio_FPrintf( f, "%s   ZLS PRECISION: Single Precision Floating Point.\n", prefix );
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION 
  zlsio_FPrintf( f, "%s   ZLS PRECISION: Double Precision Floating Point.\n", prefix );
#endif DOUBLE_PRECISION
  
#if (ZLS_DEBUG > 0)
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  zlsio_FPrintf( f, "%s   ZLS DATA: Zone data blocks are single blocks.\n", prefix );
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
  zlsio_FPrintf( f, "%s   ZLS DATA: Zone data blocks are multi-dimensional arrays.\n", prefix );
#endif ZONEDATA_ARE_MULTIDARRAYS
  
#ifdef FILES_NO_BCS
  zlsio_FPrintf( f, "%s   ZLS IO: Output files do not include boundary cells.\n", prefix );
#endif FILES_NO_BCS
#ifdef FILES_HAVE_BCS
  zlsio_FPrintf( f, "%s   ZLS IO: Output files include boundary cells if present.\n", prefix );
#endif FILES_HAVE_BCS    
#endif (ZLS_DEBUG > 0)
  
#ifdef ZLIB_COMPRESSED_IO
  zlsio_FPrintf( f, "%s   ZLS IO: Binary output files compressed as .gz \n", prefix );
#ifdef ZLIB_UNCOMPRESSED_IMAGES
  zlsio_FPrintf( f, "%s   ZLS IO: Image file output files uncompressed.\n", prefix );
#endif ZLIB_UNCOMPRESSED_IMAGES
#endif ZLIB_COMPRESSED_IO
  zlsio_FPrintf( f, "%s   ZLS IO: Binary data file header: %d bytes \n", prefix, ZLS_DATA_HEADERSIZE );
  
#ifdef MPI 
  zlsio_FPrintf( f, "%s   ZLS OS: Parallel mode using MPI, nodes: %d.\n", prefix, gMPIsize );
#endif MPI
#ifdef PTHREADS 
  zlsio_FPrintf( f, "%s   ZLS OS: Parallel - Shared Memory Preemptive Threading.\n", prefix );
  zlsio_FPrintf( f, "%s   ZLS OS: Parallel - %d cpus.\n", prefix, NCPUs );
  zlsio_FPrintf( f, "%s   ZLS OS: Parallel - %d Max Threads.\n", prefix, ZLS_MAXPTHREADS );
#endif PTHREADS
#ifdef ZLS_LOADSORTING
  zlsio_FPrintf( f, "%s   ZLS OS: Parallel - Zone CPU Load Sorting.\n", prefix );
#endif ZLS_LOADSORTING
  
#ifdef OS_GENERIC 
  zlsio_FPrintf( f, "%s   ZLS OS: Generic GCC/Unix.\n", prefix );
#endif OS_GENERIC
  
#ifdef OS_GENERICLINUX
  zlsio_FPrintf( f, "%s   ZLS OS: Generic Little Endian Intel GCC/LINUX/Unix.\n", prefix );
#endif OS_GENERICLINUX
  
#ifdef OS_CYGWIN
  zlsio_FPrintf( f, "%s   ZLS OS: Generic Cygwin Intel Windows littleendian.\n", prefix );
#endif OS_CYGWIN
  
#ifdef OS_MACOSXCORE
  zlsio_FPrintf( f, "%s   ZLS OS: MacOSX BSD/Darwin intel CORE littleendian.\n", prefix );
#endif OS_MACOSXCORE
  
#ifdef OS_MACOSXG5 
  zlsio_FPrintf( f, "%s   ZLS OS: MacOSX BSD/Darwin ppc970/G5 bigendian.\n", prefix );
#endif OS_MACOSXG5
  
#ifdef OS_MACOSXG4 
  zlsio_FPrintf( f, "%s   ZLS OS: MacOSX BSD/Darwin ppc7450/G4 bigendian.\n", prefix );
#endif OS_MACOSXG4
  
#ifdef OS_SOLARISSPARC 
  zlsio_FPrintf( f, "%s   ZLS OS: Sun Solaris/Sparc bigendian.\n", prefix );
#endif OS_SOLARISSPARC
  
#ifdef OS_SOLARISAMD64 
  zlsio_FPrintf( f, "%s   ZLS OS: Sun Solaris 10 AMD64 Opteron i86 littleendian.\n", prefix );
#endif OS_SOLARISAMD64
  
  
  return(noErr);
  
}

int zlsio_PrintConfiguration( FILE * f , char * prefix ){
  
#ifdef MPI
  if ( gMPIrank == 0){
#endif MPI
    zlsio_PrintMethodConfiguration ( f, prefix );
    zlsio_PrintGeomConfiguration   ( f, prefix );
    zlsio_PrintOSConfiguration     ( f, prefix );
    zlsio_FPrintf( f, "%s%s\n", prefix,  DIVIDERLINE );
#ifdef MPI
  }
#endif MPI
  
  return(noErr);
}



