#include "zls.h"

//
// low level routines for allocating 16 bit data cubes for multi-D zone
// data output (as opposed to images)
//
//========================================
//
//  Data scaling and utilities
//
//========================================
//

//  v low level routine all data is 1D, ptr arith for speed

int zls_DataMaxMinLinear( Real *max, Real *min, QPtr q, Size nelements){
  
  int err = noErr;
  
  register Real valmax = -kZLS_Huge ;
  register Real valmin =  kZLS_Huge ;
  
  if (q != NULL) {
    
    register Counter  i;
    register Real   val;
    
    register QPtr p = q;
    valmin = valmax = *p;
    
    for(i = 0; i<nelements;i++){
      
      val    = *p++;
      valmax = valmax<val?val:valmax;
      valmin = valmin>val?val:valmin;
      
    } //i
    
  } else { err = memErr;}
  
  *max = valmax;
  *min = valmin;
  
  return (err);
}


int zls_DataMaxMinLog10( Real *max, Real *min, QPtr q, Size nelements){
  
  int err = noErr;
  Real v      = 0.0 ; // log and isnan are so slow and not always inline, so no point in suggesting registers
  Real val    = 0.0 ; // log and isnan are so slow and not always inline, so no point in suggesting registers
  Real valmax =  kZLS_LogTiny;
  Real valmin =  kZLS_LogHuge ;
  
  if (q != NULL) {
    Counter  i;
    
    QPtr p = q;
    
    for(i = 0; i<nelements;i++){
      v      = *p++;
      val    =  (v>0.0)?zls_Log10(v):kZLS_LogTiny;
      valmax = valmax<val?val:valmax;
      if (val > kZLS_LogTiny) valmin = valmin>val?val:valmin; // min no equal to log 0 or less
      
    } //i
    
  } else { err = memErr;}
  
  //  zlsio_Printf(" mx mn : %g %g\n", valmax, valmin);
  *max = valmax>kZLS_LogTiny?valmax:0.0;
  *min = valmin<kZLS_LogHuge?valmin:0.0;
  
  return (err);
}

int zls_DataMaxMinLog10Abs( Real *max, Real *min, QPtr q, Size nelements){ // log of abs(value)
  
  int err = noErr;
  Real v      = 0.0 ; // log and isnan are so slow and not always inline, so no point in suggesting registers
  Real val    = 0.0 ; // log and isnan are so slow and not always inline, so no point in suggesting registers
  Real valmax =  kZLS_LogTiny ;
  Real valmin =  kZLS_LogHuge ;
  
  if (q != NULL) {
    Counter  i;
    
    QPtr p = q;
    
    for(i = 0; i<nelements;i++){
      v      = zls_abs(*p);p++;
      val    = (v>0.0)?zls_Log10(v):kZLS_LogTiny;
      valmax = valmax<val?val:valmax;
      if (val > kZLS_LogTiny) valmin = valmin>val?val:valmin; // min no equal to log 0 or less
      
    } //i
    
  } else { err = memErr;}
  
  //  zlsio_Printf(" Abs mx mn : %g %g\n", valmax, valmin);
  *max = valmax>kZLS_LogTiny?valmax:0.0;
  *min = valmin<kZLS_LogHuge?valmin:0.0;
  
  return (err);
}


int zls_DataMaxMinExp10( Real *max, Real *min, QPtr q, Size nelements){
  
  int err = noErr;
  Real val    = 0.0 ;
  Real valmax = -kZLS_Huge ;
  Real valmin =  kZLS_Huge ;
  
  if (q != NULL) {
    Counter  i;
    
    QPtr p = q;
    val    =  zls_Exp10(*p);
    val    = zls_isnan(val)?0.0:val;
    valmin = valmax = val;
    
    for(i = 0; i<nelements;i++){
      
      val    =  zls_Exp10(*p++);
      
      val    = zls_isnan(val)?0.0:val;
      valmax = valmax<val?val:valmax;
      valmin = valmin>val?val:valmin;
      
    } //i
    
  } else { err = memErr;}
  
  *max = valmax;
  *min = valmin;
  
  return (err);
}


int zls_DataMaxMinPow( Real *max, Real *min, QPtr q, Size nelements, Real a){
  
  int err = noErr;
  Real val    = 0.0 ;
  Real valmax = 0.0 ;
  Real valmin = 0.0 ;
  
  if (q != NULL) {
    Counter  i;
    
    QPtr p = q;
    val    = zls_Pow(*p, a);
    val    = zls_isnan(val)?0.0:val;
    valmin = valmax = val;
    
    for(i = 0; i<nelements;i++){
      
      val    =  zls_Pow(*p++, a);
      
      val    = zls_isnan(val)?0.0:val;
      valmax = valmax<val?val:valmax;
      valmin = valmin>val?val:valmin;
      
    } //i
    
  } else { err = memErr;}
  
  *max = valmax;
  *min = valmin;
  
  return (err);
}


int zls_GetLevelMergedZoneData(ZonePtr mergedZone, Integer level, Integer varIndex ) {
  
  int err = noErr;
  
  ZonePtr z;
  Integer zone;
  
  if ( mergedZone != NULL) {
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];    
      if ( z != NULL) {
        
        QPtr subImage = NULL;
        err           = zlsp_GetZoneWQuantityCopy  ( &subImage, z, varIndex); 
        
        if ( ( subImage != NULL )&&(err == noErr) ) {
#ifdef FILES_HAVE_BCS 
          err = zlsim_FillLevelImageFromDataAndZone ( mergedZone,  subImage, kZLS_DataHasBCs, z );
#endif FILES_HAVE_BCS 
#ifdef FILES_NO_BCS 
          err = zlsim_FillLevelImageFromDataAndZone ( mergedZone,  subImage, kZLS_DataNoBCs, z );
#endif FILES_NO_BCS 
          if ( err != noErr ){
            err = ioErr;
            return (err);
          }
          zls_DisposeQuantity ( subImage ); subImage = NULL;  //  clear for next zone to use
        } else {
          err = memErr;
          return (err);
        }
        
      } else { err = zonErr;} // z != NULL
    } // level zone Loop
    
  } else { err = zonErr;} // mergedZone != NULL
  
  return(err);
  
}// zls_GetLevelMergedZoneData

//
// not really scaled, but transformed.  Will clip to 32bit float range from double precision input
// but otherwise only the log/exp/pow etc transforms are done


int zls_Scale32BitFloatData       ( float * fq , QPtr q,  Integer dataType,
                                   Counter ni, Counter nj, Counter nk, 
                                   IOSettingsPtr io, Integer ioIndex ){
  
  
  int err = noErr;
  
  if (( fq != NULL) &&( q != NULL) && ( io != NULL) ) {
    
    Real v, valmax, valmin;
    
    Counter i, j, k;
    Size nelements = (Size) ni * (Size) nj* (Size) nk; // keep it 64 bit if needed
    
    
    Integer transType = io->v[ioIndex].transformType;
    Real   transPower = io->v[ioIndex].transPow;
    
    valmax = valmin = 0.0;
    
    if (transType == kZLS_IOLinearTransform) {
      
      zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);  // just to collect max min info for headers etc, no scaling done
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v = q[i + ni*(j + nj*k)];
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
      
    } // kZLS_IOLinearTransform
    
    if ( transType == kZLS_IOLog10Transform ) {
      
      zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements); // just to collect max min info for headers etc, no scaling done
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v    = zls_Log10(q[i + ni*(j + nj*k)]);
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
    } // kZLS_IOLog10Transform
    
    if ( transType == kZLS_IOLog10AbsTransform ) {
      
      zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v    = zls_Log10(zls_abs(q[i + ni*(j + nj*k)]));
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
    } // kZLS_IOLog10AbsTransform
    
    if ( transType == kZLS_IOExp10Transform ) {
      
      zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v    = zls_Exp10( q[i + ni*(j + nj*k)]);
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
    } // kZLS_IOExp10Transform
    
    if ( transType == kZLS_IOPowTransform ) {
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v    = zls_Pow( q[i + ni*(j + nj*k)], transPower );
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
    } // kZLS_IOPowTransform
    
    
    if ( transType > kZLS_IOLog10AbsTransform ) {
      
      zlsio_ErrPrintf( " WARNING: zls_Scale32BitFloatData, transform type unknown, using Log10. %d \n", transType );
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            
            v    = zls_Log10(q[i + ni*(j + nj*k)]);
            fq[i + ni*(j + nj*k)] = zls_isnan(v)?(float)(kZLS_LogTiny): (float)v;
            
          }//i
        }//j
      }//k
      
    } // ( transType > 1 ) 
    
    io->v[ioIndex].dataMax =  valmax;
    io->v[ioIndex].dataMin =  valmin;
    
  } else { err = memErr;} // q io
  
  return( err );
  
}

// low level scale q into iq according to io and ioIndex, nj and nk can be set to 1 for lower dim data
//
// dataType determines which of the scaling options to use, kZLS_IOImage 
// io->v[ioIndex].imageScalingType, and all others use io->v[ioIndex].dataScalingType
// io->v[ioIndex].ioTypes is a set of flags that the ioIndex is used form but dataType argument
// selects which of the possible flags is actually being prepared for.  ie if called from
// zlsio_ScaleAndWriteImage, and ioIndex record contains the kZLS_IOImage flag then call the 
// scale routine with dataType = kZLS_IOImage, and the imageScalingType is used.  When called from
// a save 16 bit data routine, and the same index record ioTypes flags contains kZLS_IOData, then the 
// dataType = kZLS_IOData, and dataScalingType is used.  This allows for autoscaled images and fixed scaled
// data files to be created from the same io.conf var settings.
//

#define U8INTSCALE 255L
#define U16INTSCALE 65000L
#define S16INTSCALE 32500L

#define U8SCALE 254.0
#define U16SCALE 65000.0
#define S16SCALE 32500.0

int zls_ScaleUnsigned16BitData  ( UInt16 * iq , QPtr q,  Integer dataType,
                                 Counter ni, Counter nj, Counter nk, 
                                 IOSettingsPtr io, Integer ioIndex ){
  
  int err = noErr;
  
  if (( iq != NULL) &&( q != NULL) && ( io != NULL) ) {
    
    Counter i, j, k, jRev, v;
    Size nelements;
    Real val =  0.0;
    Real range = 1.0;
    Real f = 1.0;
    
    Real valmax =  io->v[ioIndex].scaleMax;
    Real valmin =  io->v[ioIndex].scaleMin;
    
    Integer transType = io->v[ioIndex].transformType;
    Real   transPower = io->v[ioIndex].transPow;
    Integer scaleType = io->v[ioIndex].dataScalingType; // default
    
    if (( dataType&kZLS_IOImage) == kZLS_IOImage)
      scaleType = io->v[ioIndex].imageScalingType;
    
    
    nelements = (Size) ni * (Size) nj* (Size) nk; // keep it 64 bit if needed
    
    if (transType == kZLS_IOLinearTransform) {
      
      valmax =  q[0];
      valmin =  valmax;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleUnsigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = 65000.0/range;
      
      // zlsio_LogPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention most unsigned 16 bit 2D data is shown from the bottom up, but is top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      
      if ( nk == 1 ) { // special case for image formats
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            v  =  (q[i + ni*j]);
            v = f*(v-valmin);
            v = v<0.0?0.0:v;
            v = v>U16SCALE?U16SCALE:v;
            iq[i + ni*jRev] = (unsigned short) v;
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              v  =  (q[i + ni*(j + nj*k)]);
              v = f*(v-valmin);
              v = v<0.0?0.0:v;
              v = v>U16SCALE?U16SCALE:v;
              iq[i + ni*(j + nj*k)] = (unsigned short) v;
            }//i
          }//j
        }//k
        
      }
      
      
    } // kZLS_IOLinearTransform
    
    if ( transType == kZLS_IOLog10Transform ) {
      
      val    =  zls_Log10(q[0]);
      val    = zls_isnan(val)?valmin:val;
      valmax =  val;
      valmin =  val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleUnsigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = 65000.0/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      // zlsio_LogPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention most unsigned 16 bit 2D data is shown from the bottom up, but is top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            val  =  zls_Log10(q[i + ni*j]);
            val  = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            v = v>U16SCALE?U16SCALE:v;
            iq[i + ni*jRev] = (unsigned short) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              val  =  zls_Log10(q[i + ni*(j + nj*k)]);
              val  = zls_isnan(val)?valmin:val;
              v = f*(val-valmin);
              v = v<0.0?0.0:v;
              v = v>U16SCALE?U16SCALE:v;
              iq[i + ni*(j + nj*k)] = (unsigned short) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOLog10Transform
    
    if ( transType == kZLS_IOExp10Transform ) {
      
      val    =  zls_Exp10(q[0]);
      val    = zls_isnan(val)?valmin:val;
      valmax =  val;
      valmin =  val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleUnsigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = 65000.0/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      //  zlsio_LogPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention most unsigned 16 bit 2D data is shown from the bottom up, but is top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            val  =  zls_Exp10(q[i + ni*j]);
            val  = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            v = v>U16SCALE?U16SCALE:v;
            iq[i + ni*jRev] = (unsigned short) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              val  =  zls_Exp10(q[i + ni*(j + nj*k)]);
              val  = zls_isnan(val)?valmin:val;
              v = f*(val-valmin);
              v = v<0.0?0.0:v;
              v = v>U16SCALE?U16SCALE:v;
              iq[i + ni*(j + nj*k)] = (unsigned short) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOExp10Transform
    
    
    if ( transType == kZLS_IOPowTransform ) {
      
      val    =  zls_Pow(q[0],transPower);
      val    = zls_isnan(val)?valmin:val;
      valmax =  val;
      valmin =  val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower);
        
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleUnsigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower);
        
      }
      
      range = valmax - valmin;
      
      f     = 65000.0/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      // zlsio_LogPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention most unsigned 16 bit 2D data is shown from the bottom up, but is top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            val  =  zls_Pow(q[i + ni*j], transPower);
            val  = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            v = v>U16SCALE?U16SCALE:v;
            iq[i + ni*jRev] = (unsigned short) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              val  =  zls_Pow(q[i + ni*(j + nj*k)], transPower);
              val  = zls_isnan(val)?valmin:val;
              v = f*(val-valmin);
              v = v<0.0?0.0:v;
              v = v>U16SCALE?U16SCALE:v;
              iq[i + ni*(j + nj*k)] = (unsigned short) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOPowTransform
    
    if ( transType > kZLS_IOLog10AbsTransform ) {
      
      zlsio_ErrPrintf( " WARNING: zls_ScaleUnsigned16BitData, transform type unknown, using Log. %d \n", scaleType );
      
      val = zls_Log10(q[0]);
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmin:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      
      range = valmax - valmin;
      
      f     = U16SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      //  zlsio_LogPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleUnsigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      // Do the scaling
      
      // by convention most unsigned 16 bit 2D data is shown from the bottom up, but is top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            val  =  zls_Log10(q[i + ni*j]);
            val  = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            v = v>U16SCALE?U16SCALE:v;
            iq[i + ni*jRev] = (unsigned short) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              val  =  zls_Log10(q[i + ni*(j + nj*k)]);
              val  = zls_isnan(val)?valmin:val;
              v = f*(val-valmin);
              v = v<0.0?0.0:v;
              v = v>U16SCALE?U16SCALE:v;
              iq[i + ni*(j + nj*k)] = (unsigned short) v;
              
            }//i
          }//j
        }//k
        
      }
      
    }
    
    io->v[ioIndex].dataMax =  valmax;
    io->v[ioIndex].dataMin =  valmin;
    
  } else { err = memErr;} // q io
  
  return( err );
  
}

//
// low level scale q into iq according to io and ioIndex, nj and nk can be set to 1 for lower dim data
//
// dataType determines which of the scaling options to use, kZLS_IOImage 
// io->v[ioIndex].imageScalingType, and all others use io->v[ioIndex].dataScalingType
// io->v[ioIndex].ioTypes is a set of flags that the ioIndex is used form but dataType argument
// selects which of the possible flags is actually being prepared for.  ie if called from
// zlsio_ScaleAndWriteImage, and ioIndex record contains the kZLS_IOImage flag then call the 
// scale routine with dataType = kZLS_IOImage, and the imageScalingType is used.  When called from
// a save 16 bit data routine, and the same index record ioTypes flags contains kZLS_IOData, then the 
// dataType = kZLS_IOData, and dataScalingType is used.  This allows for autoscaled images and fixed scaled
// data files to be created from the same io.conf var settings.
//
int zls_ScaleSigned16BitData  ( SInt16 * iq , QPtr q, Integer dataType,
                               Counter ni, Counter nj, Counter nk, 
                               IOSettingsPtr io, Integer ioIndex ){
  
  int err = noErr;
  
  if (( iq != NULL) &&( q != NULL) && ( io != NULL) ) {
    
    Integer v;
    Counter i, j, k;
    Size nelements;
    Real val   = 0.0;
    Real valmax =  io->v[ioIndex].scaleMax;
    Real valmin =  io->v[ioIndex].scaleMin;
    Real range = 1.0;
    Real f = 1.0;
    
    Integer transType = io->v[ioIndex].transformType;
    Real   transPower = io->v[ioIndex].transPow;
    Integer scaleType = io->v[ioIndex].dataScalingType;
    
    if (( dataType&kZLS_IOImage) == kZLS_IOImage)
      scaleType = io->v[ioIndex].imageScalingType;
    
    nelements = (Size) ni * (Size) nj* (Size) nk; // keep it 64 bit if needed
    
    if (transType == kZLS_IOLinearTransform) {
      
      valmax =  q[0];
      valmin = valmax;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin; // if set max is less even than the min, reset the min
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = S16SCALE/range;  // full auto is 0 - S16SCALE  others may go -ve
      // zlsio_LogPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      zlsio_FilesPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention sigend 16 bit data is not reversed as it is usually for a data file not a
      // standard image file
      
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            val = q[i + ni*(j + nj*k)];
            v = f*(val-valmin);
            v = v < -S16SCALE?-S16SCALE:v;
            v = v >  S16SCALE?S16SCALE:v;
            iq[i + ni*(j + nj*k)] = (short)v;
          }//i
        }//j
      }//k
      
      
    } // kZLS_IOLinearTransform
    
    if ( transType == kZLS_IOLog10Transform ) {
      val = zls_Log10(q[0]);
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmax:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = S16SCALE/range;  // full auto is 0 - S16SCALE  others may go -ve
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention sigend 16 bit data is not reversed as it is usually for a data file not a
      // standard image file
      
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            val = zls_Log10(q[i + ni*(j + nj*k)]);
            val = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v < -S16SCALE?-S16SCALE:v;
            v = v >  S16SCALE?S16SCALE:v;
            iq[i + ni*(j + nj*k)] = (short) v;
          }//i
        }//j
      }//k
      
    } // kZLS_IOLog10Transform
    
    
    
    if ( transType == kZLS_IOExp10Transform ) {
      
      val = zls_Exp10(q[0]);
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmin:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = S16SCALE/range;  // full auto is 0 - S16SCALE  others may go -ve
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention sigend 16 bit data is not reversed as it is usually for a data file not a
      // standard image file
      
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            val =  zls_Exp10(q[i + ni*(j + nj*k)]);
            val = zls_isnan(val)?valmin:val;
            v   = f*(val-valmin);
            v   = v < -S16SCALE?-S16SCALE:v;
            v   = v >  S16SCALE?S16SCALE:v;
            iq[i + ni*(j + nj*k)] = (short) v;
          }//i
        }//j
      }//k
      
    } // kZLS_IOExp10Transform
    
    
    
    if ( transType == kZLS_IOPowTransform ) {
      
      val = zls_Pow(q[0],transPower);
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmin:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements,transPower);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements,transPower);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements,transPower);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements,transPower);
      }
      
      range = valmax - valmin;
      
      f     = S16SCALE/range;  // full auto is 0 - S16SCALE  others may go -ve
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention sigend 16 bit data is not reversed as it is usually for a data file not a
      // standard image file
      
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            val    =  zls_Pow(q[i + ni*(j + nj*k)],transPower);
            val    = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v < -S16SCALE?-S16SCALE:v;
            v = v >  S16SCALE?S16SCALE:v;
            iq[i + ni*(j + nj*k)] = (short) v;
          }//i
        }//j
      }//k
      
    } // kZLS_IOPowTransform
    
    if ( transType > kZLS_IOLog10AbsTransform) {
      
      zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, transform type unknown, using Log10. %d \n", scaleType );
      
      v    =  zls_Log10(q[0]);
      v    = zls_isnan(v)?valmin:v;
      valmax =  v;
      valmin =  v;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_ScaleSigned16BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = S16SCALE/range;  // full auto is 0 - S16SCALE  others may go -ve
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_ScaleSigned16BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      // by convention sigend 16 bit data is not reversed as it is usually for a data file not a
      // standard image file
      
      for(k = 0;k<nk;k++){
        for(j = 0;j<nj;j++){
          for(i = 0;i<ni;i++){
            val    =  zls_Log10(q[i + ni*(j + nj*k)]);
            val    = zls_isnan(val)?valmin:val;
            v = f*(val-valmin);
            v = v < -S16SCALE?-S16SCALE:v;
            v = v >  S16SCALE?S16SCALE:v;
            iq[i + ni*(j + nj*k)] = (short) v;
          }//i
        }//j
      }//k
      
    }
    
    io->v[ioIndex].dataMax =  valmax;
    io->v[ioIndex].dataMin =  valmin;
    
    
  } else { err = memErr;} // q io
  
  return( err );
  
}

// low level scale q into iq according to io and ioIndex, nj and nk can be set to 1 for lower dim data
//
// dataType determines which of the scaling options to use, kZLS_IOImage 
// io->v[ioIndex].imageScalingType, and all others use io->v[ioIndex].dataScalingType
// io->v[ioIndex].ioTypes is a set of flags that the ioIndex is used form but dataType argument
// selects which of the possible flags is actually being prepared for.  ie if called from
// zlsio_ScaleAndWriteImage, and ioIndex record contains the kZLS_IOImage flag then call the 
// scale routine with dataType = kZLS_IOImage, and the imageScalingType is used.  When called from
// a save 16 bit data routine, and the same index record ioTypes flags contains kZLS_IOData, then the 
// dataType = kZLS_IOData, and dataScalingType is used.  This allows for autoscaled images and fixed scaled
// data files to be created from the same io.conf var settings.
//

int zls_Scale8BitData       ( UInt8 * iq , QPtr q,  Integer dataType,
                             Counter ni, Counter nj, Counter nk, 
                             IOSettingsPtr io, Integer ioIndex ){
  
  
  int err = noErr;
  
  if (( iq != NULL) &&( q != NULL) && ( io != NULL) ) {
    
    Integer v;
    Counter i, j, k, jRev;
    Counter nelements;
    Real val    = 0.0;
    Real valmax =  io->v[ioIndex].scaleMax;
    Real valmin =  io->v[ioIndex].scaleMin;
    Real range  = 1.0;
    Real f      = 1.0;
    
    Integer transType = io->v[ioIndex].transformType;
    Real   transPower = io->v[ioIndex].transPow;
    Integer scaleType = io->v[ioIndex].dataScalingType;
    if (( dataType&kZLS_IOImage) == kZLS_IOImage)
      scaleType = io->v[ioIndex].imageScalingType;
    
    nelements = (Size) ni * (Size) nj* (Size) nk; // keep it 64 bit if needed
    
    if (transType == kZLS_IOLinearTransform) {
      
      val = q[0];
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmax:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        val = q[0];
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = q[0];
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
      }
      
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
      
#if (ZLS_DEBUG > 1)
      zlsio_Printf(" zls_Scale8BitData:Linear Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf(" zls_Scale8BitData:Linear Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling into the integer buffer
      
      // by convention most 8 bit data is shown from the bottom up, but top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            v = f*(q[i + ni*j]-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE: (unsigned char)v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              v = f*(q[i + ni*(j + nj*k)]-valmin);
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE: (unsigned char)v;
              
            }//i
          }//j
        }//k
        
      }
      
      
    } // kZLS_IOLinearTransform
    
    if ( transType == kZLS_IOLog10Transform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Log10(q[0]);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        val = zls_Log10(q[0]);
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Log10(q[0]);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            val    = zls_Log10( q[i + ni*j] );
            val    = zls_isnan(val)?valmin:val;
            
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              val    = zls_Log10(q[i + ni*(j + nj*k)]);
              val    = zls_isnan(val)?valmin:val;
              
              v = f*(val-valmin);
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOLog10Transform
    
    if ( transType == kZLS_IOLog10AbsTransform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Log10(zls_abs(q[0]));
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        val = zls_Log10(zls_abs(q[0]));
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Log10(zls_abs(q[0]));
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Scale8BitData: Log10Abs Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Scale8BitData: Log10Abs Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            val    = zls_Log10( zls_abs(q[i + ni*j]) );
            val    = zls_isnan(val)?valmin:val;
            
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              val    = zls_Log10(zls_abs(q[i + ni*(j + nj*k)]));
              val    = zls_isnan(val)?valmin:val;
              
              v = f*(val-valmin);
              
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOLog10AbsTransform
    
    if ( transType == kZLS_IOExp10Transform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Exp10( q[0]);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        val = zls_Exp10( q[0]);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Exp10( q[0]);
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            val    = zls_Exp10( q[i + ni*j]);
            val    = zls_isnan(val)?valmin:val;
            
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              val    = zls_Exp10( q[i + ni*(j + nj*k)]);
              val    = zls_isnan(val)?valmin:val;
              
              v = f*(val-valmin);
              
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOExp10Transform
    
    if ( transType == kZLS_IOPowTransform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Pow( q[0] , transPower);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        val = zls_Pow( q[0] , transPower);
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Pow( q[0] , transPower);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
      }
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Scale8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            val    = zls_Pow( q[i + ni*j], transPower );
            val    = zls_isnan(val)?valmin:val;
            
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              val    = zls_Pow( q[i + ni*(j + nj*k)], transPower );
              val    = zls_isnan(val)?valmin:val;
              
              v = f*(val-valmin);
              
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // kZLS_IOPowTransform
    
    
    
    if ( transType > kZLS_IOLog10AbsTransform ) {
      
      zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, transform type unknown, using Log10. %d \n", scaleType );
      
      v    =  zls_Log10(q[0]);
      v    = zls_isnan(v)?valmin:v;
      valmax =  v;
      valmin =  v;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Scale8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      
      range = valmax - valmin;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Scale8BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Scale8BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // by convention most 8 bit data is show from the bottom up, but top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      
      if ( nk == 1 ) { // special case for images
        for(j = 0;j<nj;j++){
          jRev = nj-j-1;
          for(i = 0;i<ni;i++){
            
            val    =  zls_Log10(q[i + ni*j]);
            val    = zls_isnan(val)?valmin:val;
            
            v = f*(val-valmin);
            v = v<0.0?0.0:v;
            iq[i + ni*jRev] = v >U8SCALE?(unsigned char)U8INTSCALE: (unsigned char)v;
            
          }//i
        }//j
        
      } else {
        
        for(k = 0;k<nk;k++){
          for(j = 0;j<nj;j++){
            for(i = 0;i<ni;i++){
              
              val    = zls_Log10(q[i + ni*(j + nj*k)]);
              val    = zls_isnan(val)?valmin:val;
              
              v = f*(val-valmin);
              
              v = v<0.0?0.0:v;
              iq[i + ni*(j + nj*k)] = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
              
            }//i
          }//j
        }//k
        
      }
      
    } // ( transType > 1 ) 
    
    io->v[ioIndex].dataMax =  valmax;
    io->v[ioIndex].dataMin =  valmin;
    
  } else { err = memErr;} // q io
  
  return( err );
  
}


int zls_Plot1D8BitData     ( UInt8 * iq , QPtr q,  Integer dataType,
                            Counter ni,  
                            IOSettingsPtr io, Integer ioIndex ){
  
  int err = noErr;
  
  if (( iq != NULL) &&( q != NULL) && ( io != NULL) ) {
    
    Integer v;
    Counter i = 0, j=0;
    Counter nelements;
    
    Real val    = 0.0;
    Real valmax =  io->v[ioIndex].scaleMax;
    Real valmin =  io->v[ioIndex].scaleMin;
    Real valmax2 =  io->v[ioIndex].scaleMax;
    Real valmin2 =  io->v[ioIndex].scaleMin;
    Integer plotColor = io->v[ioIndex].imagePlotColor;
    Real range  = 1.0;
    Real f      = 1.0;
    
    Integer transType = io->v[ioIndex].transformType;
    Real   transPower = io->v[ioIndex].transPow;
    Integer scaleType = io->v[ioIndex].imageScalingType;
    
    nelements = (Size) ni ; // 1D only
    
    zlsg_SetImage8( iq, ni, 256); // 256 pixels high
    zlsg_FrameRect (0,0, ni-1, 255);
    
    if (transType == kZLS_IOLinearTransform) {
      
      val = q[0];
      valmax = zls_isnan(val)?valmin:val;
      valmin = zls_isnan(val)?valmax:val;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        val = q[0];
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = q[0];
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
        
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLinear( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range    = valmax2 - valmin2;
      
      f        = U8SCALE/range;
      
#if (ZLS_DEBUG > 1)
      zlsio_Printf(" zls_Plot1D8BitData:Linear Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf(" zls_Plot1D8BitData:Linear Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling and plot
      
      for(i = 0;i<ni;i++){
        
        v = f*(q[i]-valmin2);
        v = v<0.0?0.0:v;
        j = v > U8SCALE?(unsigned char)U8INTSCALE: (unsigned char)v;
        SetPixelInteger(i, j, plotColor);
        
      }//i
      
    } // kZLS_IOLinearTransform
    
    if ( transType == kZLS_IOLog10Transform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Log10(q[0]);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        val = zls_Log10(q[0]);
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Log10(q[0]);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range = valmax2 - valmin2;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      for(i = 0;i<ni;i++){
        
        val    = zls_Log10( q[i] );
        val    = zls_isnan(val)?valmin2:val;
        
        v = f*(val-valmin2);
        v = v<0.0?0.0:v;
        j = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
        SetPixelInteger(i, j, plotColor);
        
      }//i
      
    } // kZLS_IOLog10Transform
    
    if ( transType == kZLS_IOLog10AbsTransform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Log10(zls_abs(q[0]));
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      }  else if ( scaleType == kZLS_IOAutoMaxScale){ 
        
        val = zls_Log10(zls_abs(q[0]));
        valmin = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){ 
        
        val = zls_Log10(zls_abs(q[0]));
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;      
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10Abs( &valmax, &valmin,  q, nelements);
        
      }
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range = valmax2 - valmin2;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Plot1D8BitData: Log10Abs Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Plot1D8BitData: Log10Abs Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling and plotting
      
      for(i = 0;i<ni;i++){
        
        val    = zls_Log10( zls_abs(q[i]) );
        val    = zls_isnan(val)?valmin2:val;
        
        v = f*(val-valmin2);
        v = v<0.0?0.0:v;
        j = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
        SetPixelInteger(i, j, plotColor);
        
      }//i
      
    } // kZLS_IOLog10AbsTransform
    
    if ( transType == kZLS_IOExp10Transform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Exp10( q[0]);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        val = zls_Exp10( q[0]);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Exp10( q[0]);
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinExp10( &valmax, &valmin,  q, nelements);
      }
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range = valmax2 - valmin2;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the plotting
      
      for(i = 0;i<ni;i++){
        
        val    = zls_Exp10( q[i + ni*j]);
        val    = zls_isnan(val)?valmin2:val;
        
        v = f*(val-valmin2);
        v = v<0.0?0.0:v;
        j = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
        SetPixelInteger(i, j, plotColor);
        
      }//i
      
    } // kZLS_IOExp10Transform
    
    if ( transType == kZLS_IOPowTransform ) {
      
      
      if ( scaleType == kZLS_IOAutoScale){
        val = zls_Pow( q[0] , transPower);
        valmax = zls_isnan(val)?valmin:val;
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        val = zls_Pow( q[0] , transPower);
        valmin = zls_isnan(val)?valmax:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        val = zls_Pow( q[0] , transPower);
        valmax = zls_isnan(val)?valmin:val;
        
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinPow( &valmax, &valmin,  q, nelements, transPower );
      }
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range = valmax2 - valmin2;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Plot1D8BitData: Log10 Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // Do the scaling
      
      
      for(i = 0;i<ni;i++){
        
        val    = zls_Pow( q[i + ni*j], transPower );
        val    = zls_isnan(val)?valmin2:val;
        
        v = f*(val-valmin2);
        v = v<0.0?0.0:v;
        j = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
        SetPixelInteger(i, j, plotColor);
        
      }//i
      
    } // kZLS_IOPowTransform
    
    
    
    if ( transType > kZLS_IOLog10AbsTransform ) {
      
      zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, transform type unknown, using Log10. %d \n", scaleType );
      
      v    =  zls_Log10(q[0]);
      v    = zls_isnan(v)?valmin:v;
      valmax =  v;
      valmin =  v;
      
      if ( scaleType == kZLS_IOAutoScale){
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        
      } else if ( scaleType == kZLS_IOManualScale){ 
        
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  io->v[ioIndex].scaleMin;
        
      } else if ( scaleType == kZLS_IOAutoMaxScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmax =  io->v[ioIndex].scaleMax;
        valmin =  valmin>valmax?valmax:valmin;
        
      } else if ( scaleType == kZLS_IOAutoMinScale){  
        
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
        valmin =  io->v[ioIndex].scaleMin;
        valmax =  valmax<valmin?valmin:valmax;
        
      } else {
        zlsio_ErrPrintf( " WARNING: zls_Plot1D8BitData, scaling type unknown, using auto. %d \n", scaleType );
        zls_DataMaxMinLog10( &valmax, &valmin,  q, nelements);
      }
      
      
      range = valmax - valmin;
      
      Real delta = zls_abs(range)*0.05; // make room so plot doens't go into edges
      
      valmax2 = valmax + delta;
      valmin2 = valmin - delta;
      range = valmax2 - valmin2;
      
      f     = U8SCALE/range;
#if (ZLS_DEBUG > 1)
      zlsio_Printf( " zls_Plot1D8BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
#endif (ZLS_DEBUG > 1)
      zlsio_FilesPrintf( " zls_Plot1D8BitData: Max Min Range and Scale %g %g %g %g\n",valmax, valmin, range, f);
      
      // by convention most 8 bit data is show from the bottom up, but top down in image files.
      // if nk == 1 then assume the buffer is destined for an image file like tiff and needs reversing
      
      for(i = 0;i<ni;i++){
        
        val    =  zls_Log10(q[i + ni*j]);
        val    = zls_isnan(val)?valmin2:val;
        
        v = f*(val-valmin2);
        v = v<0.0?0.0:v;
        j = v >U8SCALE?(unsigned char)U8INTSCALE:(unsigned char) v;
        SetPixelValue(i, j+1, v/U8SCALE);
        
      }//i
      
    } // ( transType > 1 ) 
    
    io->v[ioIndex].dataMax =  valmax;
    io->v[ioIndex].dataMin =  valmin;
    
  } else { err = memErr;} // q io
  
  return( err );
  
}

//
//========================================
//
//  Write 16 bit Data Files
//
//========================================
//

int zls_Allocate16BitZoneData ( short ** iq , ZonePtr z ){
  int err = noErr;
  short * sdata = NULL;
  
  if ( z != NULL) {
    
    Counter ni ;
    Counter nj ;
    Counter nk ;
    
    GridCoords gc = z->zoneGrid;
    
    zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc );
    
    sdata = (short *) malloc(sizeof(short)*ni*nj*nk);
    if ( sdata == NULL ) { err = memErr; }
    
  } else {err = zonErr;}
  
  *iq = sdata;
  
  return (err);
  
} 

int zls_Free16BitZoneData ( short * iq ){
  int err = noErr;
  
  if ( iq != NULL) free ((void *) iq);
  
  return (err);
}   


// low level write

int zlsio_Write16BitDataBinaryHeader ( zFile f, ZonePtr z, IOSettingsPtr io, Integer ioIndex  ){
  
  int err = noErr;
  
  binaryDataHeader vdh;
  
  GridCoords   gc  = z->zoneGrid;
  Range        rng = z->zoneLevelRange;
  Integer varIndex = kAnonVar;
  
  vdh.fMagicTag   = ZLS_MAGICTAG;
  vdh.fDataOffset = (unsigned int)ZLS_DATA_HEADERSIZE; // offset to data from start of file
  vdh.fVersion    = ZLS_VERSION;
  vdh.fFormatTag  = ZLS_FORMATTAG;              
  vdh.fFormatRecSize = (Counter)((Size) (&vdh.fTimeTag) - (Size)(&vdh.fFileType));
  vdh.fFileType   = kZLS_SInt16Data;
  
  vdh.fTimeTag    = ZLS_TIMETAG;
  vdh.fTimeRecSize= (Counter)((Size) (&vdh.fVarTag) - (Size)(&vdh.fTime));
  vdh.fTime       = zlst_GlobalTime;
  vdh.fTimeScale  = f_t0;
  vdh.fCycles     = zlst_GlobalCycles;
  
  vdh.fVarTag    = ZLS_VARTAG;
  vdh.fVarRecSize = (Counter)((Size) (&vdh.fTransTag) - (Size)(&vdh.fVarType));
  vdh.fVarType    = *((unsigned int *) io->v[ioIndex].varCode);
  
  vdh.fUnitsTag    = ZLS_UNITSTAG;
  vdh.fUnitsRecSize = (Counter)((Size) (&vdh.fTransTag) - (Size)(&vdh.fScalingUnit));
  zlsio_GetIOCodeVarIndex( &varIndex,io->v[ioIndex].varCode);
  zlsio_GetVarFiducialValue( &vdh.fScalingUnit, varIndex);  // Fiducial scaling factor to convert normalised data value to actual units, 1.0 for adiabatic case
  
  vdh.fTransTag    = ZLS_TRANSFORMTAG;
  vdh.fTransRecSize= (Counter)((Size)(&vdh.fDataTag) - (Size)(&vdh.fTransType));
  vdh.fTransType   = io->v[ioIndex].transformType;
  vdh.fTransPow   = io->v[ioIndex].transPow;
  
  vdh.fDataTag      = ZLS_DATATAG;    //  Data maximum index
  vdh.fDataRecSize  = (Counter)((Size)(&vdh.fDimsTag) - (Size)(&vdh.fDataMaxTag));
  vdh.fDataMaxTag   = ZLS_DMAXTAG;    //  Data maximum index
  vdh.fDataMax      = S16INTSCALE;          // Max value used in data (= S16INTSCALE usually) in signed 32 int for alignment
  vdh.fDataMinTag   = ZLS_DMINTAG;    // Data minimum index
  vdh.fDataMin      = 0L;              // Min value used in data (= 0 usually) in signed 32 int for alignment
  vdh.fMaxTag       = ZLS_MAXRTAG;    // transform tag
  vdh.fMax          = io->v[ioIndex].dataMax;   // Value to associate with datamax, 8 byte field = 2x4 byte
  vdh.fMinTag       = ZLS_MINRTAG;    // Min value for data min index
  vdh.fMin          = io->v[ioIndex].dataMin;     // Value to associate with datamin, 8 byte field = 2x4 byte
  
  vdh.fDimsTag  = ZLS_DIMSTAG;
  vdh.fDimsRecSize  = (Counter)((Size)(&vdh.fRngeTag) - (Size)(&vdh.fDims));
  vdh.fDims     = gc.nD;
  
#ifdef FILES_HAVE_BCS 
  vdh.fXDimTag  = ZLS_XDIMTAG;
  vdh.fXDim     = gc.nt[kXDim];
  vdh.fYDimTag  = ZLS_YDIMTAG;
  vdh.fYDim     = gc.nt[kYDim];
  vdh.fZDimTag  = ZLS_ZDIMTAG;
  vdh.fZDim     = gc.nt[kZDim];
  vdh.fNGSTTag  = ZLS_NGSTTAG;
  vdh.fNGST     = gc.nG;
#endif FILES_HAVE_BCS 
#ifdef FILES_NO_BCS 
  vdh.fXDimTag  = ZLS_XDIMTAG;
  vdh.fXDim     = gc.n[kXDim];
  vdh.fYDimTag  = ZLS_YDIMTAG;
  vdh.fYDim     = gc.n[kYDim];
  vdh.fZDimTag  = ZLS_ZDIMTAG;
  vdh.fZDim     = gc.n[kZDim];
  vdh.fNGSTTag  = ZLS_NGSTTAG;
  vdh.fNGST     = 0L;
#endif FILES_NO_BCS 
  
  vdh.fRngeTag     = ZLS_RNGETAG;     // 'RNGE' tag , location of data in super grid, to allow reconstruction of relative location of data
  vdh.fRngeRecSize = (Counter)((Size)(&vdh.fSampleTag) - (Size)(&vdh.fRngeDims)); // Offset to next record, (fSampleTag currently) after reading this value       
  vdh.fRngeDims    = rng.nD;   // 1, 2 or 3, any other value is an error, should == fDims
  // use this to determine which of the following three records to use.  All three are always
  // present, but some may not be used, and just contains unit values
  
  vdh.fXrngTag     = ZLS_XRNGTAG;  // 'XRNG' tag
  vdh.fXrngRecSize = (Counter)((Size)(&vdh.fYrngTag) - (Size)(&vdh.fXTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fXTypeTag    = ZLS_XTYPTAG;  // 'XTYP' tag
  vdh.fXType       = rng.type[kXDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fXKindTag    = ZLS_XKNDTAG;  // 'XKND' tag
  vdh.fXKind       = rng.kind[kXDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fXn0Tag      = ZLS_XCRDTAG;  // 'XCRD' coordinates tag
  vdh.fXn0         = rng.n0[kXDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fXrn         = rng.rn[kXDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fXx0         = rng.x0[kXDim];    // coordinate value of left edge, fXn0, 8 byte field
  vdh.fXrx         = rng.rx[kXDim];    // coordinate range of fXrn, 8 byte field
  vdh.fXscale      = zls_AxisFiducialScale(kXDim);        // scaling factor for X-coords to recover real units
  
  vdh.fYrngTag     = ZLS_YRNGTAG;  // 'YRNG' tag
  vdh.fYrngRecSize = (Counter)((Size)(&vdh.fZrngTag) - (Size)(&vdh.fYTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fYTypeTag    = ZLS_YTYPTAG;  // 'YTYP' tag
  vdh.fYType       = rng.type[kYDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fYKindTag    = ZLS_YKNDTAG;  // 'YKND' tag
  vdh.fYKind       = rng.kind[kYDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fYn0Tag      = ZLS_YCRDTAG;  // 'YCRD' coordinates tag
  vdh.fYn0         = rng.n0[kYDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fYrn         = rng.rn[kYDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fYx0         = rng.x0[kYDim];    // coordinate value of left edge, fXn0, 8 byte field
  vdh.fYrx         = rng.rx[kYDim];    // coordinate range of fXrn, 8 byte field
  vdh.fYscale      = zls_AxisFiducialScale(kYDim);        // scaling factor for X-coords to recover real units
  
  vdh.fZrngTag     = ZLS_ZRNGTAG;  // 'ZRNG' tag
  vdh.fZrngRecSize = (Counter)((Size)(&vdh.fSampleTag) - (Size)(&vdh.fZTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fZTypeTag    = ZLS_ZTYPTAG;  // 'ZTYP' tag
  vdh.fZType       = rng.type[kZDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fZKindTag    = ZLS_ZKNDTAG;  // 'ZKND' tag
  vdh.fZKind       = rng.kind[kZDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fZn0Tag      = ZLS_ZCRDTAG;  // 'ZCRD' coordinates tag
  vdh.fZn0         = rng.n0[kZDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fZrn         = rng.rn[kZDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fZx0         = rng.x0[kZDim];    // coordinate value of left edge, fXn0, 8 byte field
  vdh.fZrx         = rng.rx[kZDim];    // coordinate range of fXrn, 8 byte field
  vdh.fZscale      = zls_AxisFiducialScale(kZDim);        // scaling factor for X-coords to recover real units
  
  vdh.fSampleTag      = ZLS_SAMPLETAG;
  vdh.fSampleRecSize  = 4UL;
  vdh.fSampleSize     = ((unsigned int) sizeof(SInt16));
  
  Size chunkSize = ZLS_DATA_HEADERSIZE;
  err = zlsio_write( f, &vdh, chunkSize );
  
  return(err);
  
}


int zlsio_Write16BitDataFile  ( zFile fileAccess, short * iq, Counter ni, Counter nj, Counter nk ){
  int err = noErr;
  
  
  if (fileAccess) {
    
    Size chunkSize = (Size)ni*nj*nk*sizeof(short);
    
    if ((err == noErr) && (iq != NULL)){
      err    =  zlsio_write( fileAccess, iq, chunkSize );
      if ( err != noErr ){
        zlsio_ErrPrintf( " ERROR: zls_Write16BitData: IO StateWrite Error: %d \n", err );
        return (err);
      } 
      
    } else {
      zlsio_ErrPrintf( " ERROR: zls_Write16BitData: data not found\n" );
      err = memErr;
      return(err);
    }
    
  } else { return (-1);}
  
  return( err ); 
}


int zlsio_Save16BitZoneData  ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
                              IOSettingsPtr  io, Integer ioIndex ) {
  int err = noErr; 
  
  short *    iq = NULL;
  QPtr        q = NULL;
  
  if ( ( z != NULL) && ( io != NULL ) ){
    
    char     outFile[PATHLEN];
    char     vPref[PREFIXLEN];
    
    
    err = zlsio_GetVarIOVarPrefix ( vPref, io, ioIndex );
    if ( err != noErr ){
      zlsio_ErrPrintf( " IO ERROR: ioIndex out of range in zlsio_Save16BitZoneData, %d Err = %d \n", ioIndex, err );
      err = ioErr;
      return (err);
    }
    
    err = zls_Allocate16BitZoneData ( &iq ,  z ); // returns a array the same dims a z in 16 bit signed cells
    
    if (err == noErr ) {
      
      if (z->zoneType == kImageZoneType) {
        //
        // data zone was merged, simplifies file name and get data calls, will make get data calls uniform later
        // name doesn't contain zone id, just the level and cycle number
        //
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref, io->outName, level, (int)io->ioState.ioDataCount, BINFILEEXT );        
        
        err      = zls_GetZoneDataCopy        ( &q, z, kAnonVar);
        
      } else {
        
        if ( gLevels[level].nZones == 1){ // only 1 zone, save without zone id like a merged zone
          
          sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, io->v[ioIndex].varCode, PATHSEP,
                  vPref, io->outName, level, (int)io->ioState.ioDataCount, BINFILEEXT );        
          
        } else { // multiple zones in level, unmerged.  Make file into a zone id name inside a cycle count directory
          
          // first create the cycle count directory
          sprintf(outFile, "%s%s%s%sC%4.4d", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount );  
          
          err = zls_mkdir(outFile); // must work!
          if ( err != noErr) { return (fileErr);}
          
          // then the pathname of the file
          sprintf(outFile, "%s%s%s%sC%4.4d%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount,  PATHSEP,
                  vPref, io->outName, level, zone, (int)io->ioState.ioDataCount, BINFILEEXT );  
          
        }
        
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
        
      }
      
      if ( ( q != NULL ) &&  ( iq != NULL ) ) {
        GridCoords gc = z->zoneGrid;
        
        Counter ni ;
        Counter nj ;
        Counter nk ;
        
        zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc );
        
        // apply velocity slice limits if needed
        switch (varIndex) {
          case kWDeltaVx:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#if (NDIMS >1)        
          case kWDeltaVy:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif (NDIMS >1)        
#if (NDIMS >2)        
          case kWDeltaVz:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif  (NDIMS >2)   
          default:
            ;
        }
        
#if (ZLS_DEBUG > 0)       
        zlsio_Printf( " IO: Writing Data 16bit:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
#endif     
        //zlsio_LogPrintf  ( " IO: Writing Data 16bit:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        zlsio_FilesPrintf( " IO: Writing Data 16bit:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        
        err = zls_ScaleSigned16BitData  ( iq ,  q, kZLS_IOData,
                                         ni, nj, nk, 
                                         io, ioIndex );
        zFile fileAccess = NULL;
        
        fileAccess = zlsio_open(outFile, "wb");
        
        if (fileAccess) {
          
          err = zlsio_Write16BitDataBinaryHeader (fileAccess, z, io,  ioIndex );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save16BitZoneData: zlsio_Write16BitDataBinaryHeader Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          err = zlsio_Write16BitDataFile  ( fileAccess, iq, ni, nj, nk );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save16BitZoneData: zlsio_Write16BitDataFile Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          
          zlsio_close(fileAccess);
          
        }
        
        
      }
      
      err = zls_Free16BitZoneData ( iq );
      if (q!=NULL)  free(q) ; q = NULL; 
      
      
    } else { err = memErr;  return (err);} // iq
    
    
  } else { err = zonErr;} // z io
  
  return ( err );
  
}


int zlsio_SaveLevel16BitZoneData(Integer level, int step){ 
#pragma unused(step)  
  
  int        err=noErr;
  Integer          zone;
  Integer      varIndex;
  Integer       ioIndex;
  ZonePtr             z;
  
  IOSettingsPtr  io = NULL;
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
    
    ZonePtr mergedZone    = NULL;
    Integer bDoMergedData = 0;
    
    //#ifndef MPI        
    ////
    //// only merge full data files for non-MPI jobs, for IO performance reasons
    //// in massively parallel MPI jobs
    //// data headers contain all the information needed to reconstruct 
    //// multiple zones into levels later
    ////
    //    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
    //      if ( gLevels[level].nZones >1) { // then merge output
    //        zls_GetLevelImageZone( &mergedZone, level ); 
    //      } 
    //    }
    //    
    //#endif !MPI
#ifdef MPI
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        if ( gMPIrank == level) { // collect merged output on rank == level only
          zls_GetLevelImageZone( &mergedZone, level ); 
        }
        bDoMergedData = 1; // but all ranks participate in merger
      } // other levels do separate zones and imageZone stays NULL;
    }
#else    
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        zls_GetLevelImageZone( &mergedZone, level );
        if ( mergedZone != NULL) bDoMergedData = 1;
      } // other levels do separate zones and imageZone stays NULL;
    }
#endif !MPI
    
    
    for ( ioIndex = 0; ioIndex< io->outNVars; ioIndex++){
      
      if ((io->v[ioIndex].ioTypes & kZLS_IOData) == kZLS_IOData) { // data io type flag
        
        varIndex = kAnonVar;
        
        err = zlsio_GetIOCodeVarIndex (&varIndex, io->v[ioIndex].varCode);
        if ( err != noErr ) {
          zlsio_ErrPrintf( " IO ERROR: varCode unknown in zlsio_SaveLevel32BitFloatZoneData, %d %s Err = %d \n"
                          , ioIndex, io->v[ioIndex].varCode, err );
          err = ioErr;
          return (err);
        }
        
        if ( varIndex != kNoVar ) { 
          
          
          if ( bDoMergedData == 1) {
#ifdef MPI
            err =  zls_MPI_GetLevelMergedZoneData( mergedZone, level, varIndex );
#else 
            err =  zls_GetLevelMergedZoneData( mergedZone, level, varIndex );
#endif !MPI
          } // imageZone may be null on non rank 0 processes, they send buffers to 0 which collects all
          
          //#ifndef MPI        
          //        if ( mergedZone != NULL ) {
          //          err =  zls_GetLevelMergedZoneData( mergedZone, level, varIndex );
          //        }
          //#endif !MPI
          
          if ((err == noErr) &&( bDoMergedData == 1 )) { // if the merge succeeded
            
#ifdef MPI  
            if ( gMPIrank == level) { // only rank == level can to write out merged image
#endif MPI
              // imageZone always has data in zoneData[0], so kAnonVar is used here
              
              err = zlsio_Save16BitZoneData ( mergedZone, varIndex, level, 0, io, ioIndex );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO WARNING: zlsio_Save16BitZoneData Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
              }
#ifdef MPI  
            }
#endif MPI
            
          } else { 
            
            
            //
            // loop over zones separately
            //
            
            
            for( zone = 0; zone <gLevels[level].nZones; zone++){
              
              err = zls_GetLevelZoneByIndex( &z, level, zone );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
                err = ioErr;
                return (err);
              }
              
#ifdef MPI  
              if (z->zoneRankID == gMPIrank) {
#endif MPI
                
                err = zlsio_Save16BitZoneData (  z, varIndex, level, zone, io, ioIndex );
                if ( err != noErr ){
                  zlsio_ErrPrintf( " IO WARNING: zlsio_Save16BitZoneData Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                  zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
                }
#ifdef MPI  
              }
#endif MPI
              
            } // zone loop
            
          }
          
        }// !kNoVar
        
      } // data io type flag
      
    } // ioIndex
    
    if ( level == 0 ) io->ioState.ioDataCount++;
    
  } else { //io
    zlsio_ErrPrintf( " IO ERROR: IOSettings not found zlsio_SaveLevel16BitZoneData. Err %d\n", err );
    err = ioErr;
  }
  
  return (err);
}

//
//========================================
//
//  Write 32 bit float Data Files
//
//========================================
//

int zls_Allocate32BitFloatZoneData ( float ** fq , ZonePtr z ){
  int err = noErr;
  float * fdata = NULL;
  
  if ( z != NULL) {
    
    Counter ni ;
    Counter nj ;
    Counter nk ;
    
    GridCoords gc = z->zoneGrid;
    
    zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc );
    
    fdata = (float *) malloc(sizeof(float)*ni*nj*nk);
    if ( fdata == NULL ) { err = memErr; }
    
  } else {err = zonErr;}
  
  *fq = fdata;
  
  return (err);
  
} 

int zls_Free32BitFloatZoneData ( float * fq ){
  int err = noErr;
  
  if ( fq != NULL) free ((void *) fq);
  
  return (err);
}   


// low level write

int zlsio_Write32BitFloatDataBinaryHeader ( zFile f, ZonePtr z, IOSettingsPtr io, Integer ioIndex  ){
  
  int err = noErr;
  
  binaryDataHeader vdh;
  
  GridCoords     gc  = z->zoneGrid;
  Range          rng = z->zoneLevelRange;
  Integer   varIndex = kAnonVar;
  
  vdh.fMagicTag   = ZLS_MAGICTAG;
  vdh.fDataOffset = (unsigned int)ZLS_DATA_HEADERSIZE; // offset to data from start of file
  vdh.fVersion    = ZLS_VERSION;
  vdh.fFormatTag  = ZLS_FORMATTAG;              
  vdh.fFormatRecSize = (Counter)((Size) (&vdh.fTimeTag) - (Size)(&vdh.fFileType));
  vdh.fFileType   = kZLS_Float32Data;
  
  vdh.fTimeTag    = ZLS_TIMETAG;
  vdh.fTimeRecSize= (Counter)((Size) (&vdh.fVarTag) - (Size)(&vdh.fTime));
  vdh.fTime       = zlst_GlobalTime;
  vdh.fTimeScale  = f_t0;
  vdh.fCycles     = zlst_GlobalCycles;
  
  vdh.fVarTag    = ZLS_VARTAG;
  vdh.fVarRecSize = (Counter)((Size) (&vdh.fUnitsTag) - (Size)(&vdh.fVarType));
  vdh.fVarType    = *((unsigned int *) io->v[ioIndex].varCode);
  
  
  vdh.fUnitsTag    = ZLS_UNITSTAG;
  vdh.fUnitsRecSize = (Counter)((Size) (&vdh.fTransTag) - (Size)(&vdh.fScalingUnit));
  zlsio_GetIOCodeVarIndex( &varIndex,io->v[ioIndex].varCode);
  zlsio_GetVarFiducialValue( &vdh.fScalingUnit, varIndex);  // Fiducial scaling factor to convert normalised data value to actual units, 1.0 for adiabatic case
  
  vdh.fTransTag    = ZLS_TRANSFORMTAG;
  vdh.fTransRecSize= (Counter)((Size)(&vdh.fDataTag) - (Size)(&vdh.fTransType));
  vdh.fTransType   = io->v[ioIndex].transformType;
  vdh.fTransPow   = io->v[ioIndex].transPow;
  
  vdh.fDataTag      = ZLS_DATATAG;    //  Data maximum index
  vdh.fDataRecSize  = (Counter)((Size)(&vdh.fDimsTag) - (Size)(&vdh.fDataMaxTag));
  vdh.fDataMaxTag   = ZLS_DMAXTAG;    //  Data maximum index
  vdh.fDataMax      = 1.0F;          // not used in 32 bit float data, set to 1.0 and ignored as scaling is not used
  vdh.fDataMinTag   = ZLS_DMINTAG;    // Data minimum index
  vdh.fDataMin      = 0.0F;              //  not used in 32 bit float data, set to 0.0 and ignored
  vdh.fMaxTag       = ZLS_MAXRTAG;    // transform tag
  vdh.fMax          = io->v[ioIndex].dataMax;   // Value to associate with datamax, 8 byte field = 2x4 byte  may align on 8byte bdry
  vdh.fMinTag       = ZLS_MINRTAG;    // Min value for data min index
  vdh.fMin          = io->v[ioIndex].dataMin;     // Value to associate with datamin, 8 byte field = 2x4 byte  may align on 8byte bdry
  
  vdh.fDimsTag  = ZLS_DIMSTAG;
  vdh.fDimsRecSize  = (Counter)((Size)(&vdh.fRngeTag) - (Size)(&vdh.fDims));
  vdh.fDims     = gc.nD;
  
#ifdef FILES_HAVE_BCS 
  vdh.fXDimTag  = ZLS_XDIMTAG;
  vdh.fXDim     = gc.nt[kXDim];
  vdh.fYDimTag  = ZLS_YDIMTAG;
  vdh.fYDim     = gc.nt[kYDim];
  vdh.fZDimTag  = ZLS_ZDIMTAG;
  vdh.fZDim     = gc.nt[kZDim];
  vdh.fNGSTTag  = ZLS_NGSTTAG;
  vdh.fNGST     = gc.nG;
#endif FILES_HAVE_BCS 
#ifdef FILES_NO_BCS 
  vdh.fXDimTag  = ZLS_XDIMTAG;
  vdh.fXDim     = gc.n[kXDim];
  vdh.fYDimTag  = ZLS_YDIMTAG;
  vdh.fYDim     = gc.n[kYDim];
  vdh.fZDimTag  = ZLS_ZDIMTAG;
  vdh.fZDim     = gc.n[kZDim];
  vdh.fNGSTTag  = ZLS_NGSTTAG;
  vdh.fNGST     = 0L;
#endif FILES_NO_BCS 
  
  vdh.fRngeTag     = ZLS_RNGETAG;     // 'RNGE' tag , location of data in super grid, to allow reconstruction of relative location of data
  vdh.fRngeRecSize = (Counter)((Size)(&vdh.fSampleTag) - (Size)(&vdh.fRngeDims)); // Offset to next record, (fSampleTag currently) after reading this value       
  vdh.fRngeDims    = rng.nD;   // 1, 2 or 3, any other value is an error, should == fDims
  // use this to determine which of the following three records to use.  All three are always
  // present, but some may not be used, and just contains unit values
  
  vdh.fXrngTag     = ZLS_XRNGTAG;  // 'XRNG' tag
  vdh.fXrngRecSize = (Counter)((Size)(&vdh.fYrngTag) - (Size)(&vdh.fXTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fXTypeTag    = ZLS_XTYPTAG;  // 'XTYP' tag
  vdh.fXType       = rng.type[kXDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fXKindTag    = ZLS_XKNDTAG;  // 'XKND' tag
  vdh.fXKind       = rng.kind[kXDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fXn0Tag      = ZLS_XCRDTAG;  // 'XCRD' coordinates tag
  vdh.fXn0         = rng.n0[kXDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fXrn         = rng.rn[kXDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fXx0         = rng.x0[kXDim];    // coordinate value of left edge, fXn0, 8 byte field may align on 8byte bdry
  vdh.fXrx         = rng.rx[kXDim];    // coordinate range of fXrn, 8 byte field may align on 8byte bdry
  
  vdh.fXscale      = zls_AxisFiducialScale(kXDim);        // scaling factor for X-coords to recover real units
  
  vdh.fYrngTag     = ZLS_YRNGTAG;  // 'YRNG' tag
  vdh.fYrngRecSize = (Counter)((Size)(&vdh.fZrngTag) - (Size)(&vdh.fYTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fYTypeTag    = ZLS_YTYPTAG;  // 'YTYP' tag
  vdh.fYType       = rng.type[kYDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fYKindTag    = ZLS_YKNDTAG;  // 'YKND' tag
  vdh.fYKind       = rng.kind[kYDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fYn0Tag      = ZLS_YCRDTAG;  // 'YCRD' coordinates tag
  vdh.fYn0         = rng.n0[kYDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fYrn         = rng.rn[kYDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fYx0         = rng.x0[kYDim];    // coordinate value of left edge, fXn0, 8 byte field, may align on 8byte bdry
  vdh.fYrx         = rng.rx[kYDim];    // coordinate range of fXrn, 8 byte field may align on 8byte bdry
  vdh.fYscale      = zls_AxisFiducialScale(kYDim);        // scaling factor for Y-coords to recover real units
  
  vdh.fZrngTag     = ZLS_ZRNGTAG;  // 'ZRNG' tag
  vdh.fZrngRecSize = (Counter)((Size)(&vdh.fSampleTag) - (Size)(&vdh.fZTypeTag)); ;  // Offset to next record within the RNGE record, after reading this value       
  vdh.fZTypeTag    = ZLS_ZTYPTAG;  // 'ZTYP' tag
  vdh.fZType       = rng.type[kZDim];  // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  vdh.fZKindTag    = ZLS_ZKNDTAG;  // 'ZKND' tag
  vdh.fZKind       = rng.kind[kZDim];  // int, >= 0,  0 = linear, 1 = log 
  vdh.fZn0Tag      = ZLS_ZCRDTAG;  // 'ZCRD' coordinates tag
  vdh.fZn0         = rng.n0[kZDim];    // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
  vdh.fZrn         = rng.rn[kZDim];    // int, index range on Super grid, if nR >1 then this will be < XDIMS above
  vdh.fZx0         = rng.x0[kZDim];    // coordinate value of left edge, fXn0, 8 byte field may align on 8byte bdry
  vdh.fZrx         = rng.rx[kZDim];    // coordinate range of fXrn, 8 byte field may align on 8byte bdry
  
  vdh.fZscale      = zls_AxisFiducialScale(kZDim);        // scaling factor for Z-coords to recover real units
  
  vdh.fSampleTag      = ZLS_SAMPLETAG;
  vdh.fSampleRecSize  = 4UL;
  vdh.fSampleSize     = ((unsigned int) sizeof(float));
  
  Size chunkSize = ZLS_DATA_HEADERSIZE;
  err = zlsio_write( f, &vdh, chunkSize );
  
  return(err);
  
}


int zlsio_Write32BitFloatDataFile  ( zFile fileAccess, float * fq, Counter ni, Counter nj, Counter nk ){
  int err = noErr;
  
  
  if (fileAccess) {
    
    Size chunkSize = (Size)ni*nj*nk*sizeof(float);
    
    if ((err == noErr) && (fq != NULL)){
      err    =  zlsio_write( fileAccess, fq, chunkSize );
      if ( err != noErr ){
        zlsio_ErrPrintf( " ERROR: zlsio_Write32BitFloatDataFile: IO zlsio_write Error: %d \n", err );
        return (err);
      } 
      
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_Write32BitFloatDataFile: data not found\n" );
      err = memErr;
      return(err);
    }
    
  } else { return (-1);}
  
  return( err ); 
}


int zlsio_Save32BitFloatZoneData  ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
                                   IOSettingsPtr  io, Integer ioIndex ) {
  int err = noErr; 
  
  float * fq = NULL;
  QPtr q = NULL;
  
  if ( ( z != NULL) && ( io != NULL ) ){
    
    char     outFile[PATHLEN];
    char     vPref[PREFIXLEN];
    
    
    err = zlsio_GetVarIOVarPrefix ( vPref, io, ioIndex );
    if ( err != noErr ){
      zlsio_ErrPrintf( " IO ERROR: ioIndex out of range in zlsio_Save32BitFloatZoneData, %d Err = %d \n", ioIndex, err );
      err = ioErr;
      return (err);
    }
    
    err = zls_Allocate32BitFloatZoneData ( &fq ,  z ); // returns a array the same dims a z in 32 bit float cells
    
    if (err == noErr ) {
      
      if (z->zoneType == kImageZoneType) {
        //
        // data zone was merged, simplifies file name and get data calls, will make get data calls uniform later
        // name doesn't contain zone id, just the level and cycle number
        //
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref, io->outName, level, (int)io->ioState.ioDataCount, BINFILEEXT );        
        
        err      = zls_GetZoneDataCopy        ( &q, z, kAnonVar);
        
      } else {
        
        if ( gLevels[level].nZones == 1){ // only 1 zone, save without zone id like a merged zone
          
          sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, io->v[ioIndex].varCode, PATHSEP,
                  vPref, io->outName, level, (int)io->ioState.ioDataCount, BINFILEEXT );        
          
        } else { // multiple zones in level, unmerged.  Make file into a zone id name inside a cycle count directory
          
          // first create the cycle count directory
          sprintf(outFile, "%s%s%s%sC%4.4d", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount );  
          
          err = zls_mkdir(outFile); // must work!
          if ( err != noErr) { return (fileErr);}
          
          // then the pathname of the file
          sprintf(outFile, "%s%s%s%sC%4.4d%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount,  PATHSEP,
                  vPref, io->outName, level, zone, (int)io->ioState.ioDataCount, BINFILEEXT );  
          
        }
        
#ifdef SYM1D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM1D
#ifdef SYM2D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM2D
#ifdef SYM3D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM3D
        
      }
      
      
      if ( ( q != NULL ) &&  ( fq != NULL ) ) {
        
        Counter ni ;
        Counter nj ;
        Counter nk ;
        
        GridCoords gc = z->zoneGrid;
        zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc );
        
        // apply velocity slice limits if needed
        switch (varIndex) {
          case kWDeltaVx:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#if (NDIMS >1)        
          case kWDeltaVy:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif (NDIMS >1)        
#if (NDIMS >2)        
          case kWDeltaVz:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif  (NDIMS >2)   
          default:
            ;
        }
        
        
#if (ZLS_DEBUG > 0)       
        zlsio_Printf( " IO: Writing Data 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
#endif     
        //zlsio_LogPrintf  ( " IO: Writing Data 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        zlsio_FilesPrintf( " IO: Writing Data 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        
        err = zls_Scale32BitFloatData  ( fq ,  q, kZLS_IOData,
                                        ni, nj, nk, 
                                        io, ioIndex );
        zFile fileAccess = NULL;
        
        fileAccess = zlsio_open(outFile, "wb");
        
        if (fileAccess) {
          
          err = zlsio_Write32BitFloatDataBinaryHeader (fileAccess, z, io,  ioIndex );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save32BitFloatZoneData: zlsio_Write32BitFloatDataBinaryHeader Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          err = zlsio_Write32BitFloatDataFile  ( fileAccess, fq, ni, nj, nk );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save32BitFloatZoneData: zlsio_Write32BitFloatDataFile Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          
          zlsio_close(fileAccess);
          
        }
        
        
      }
      
      err = zls_Free32BitFloatZoneData ( fq );
      if (q!=NULL)  free(q) ; q = NULL; 
      
      
    } else { err = memErr;  return (err);} // iq
    
    
  } else { err = zonErr;} // z io
  
  return ( err );
  
}


int zlsio_SaveLevel32BitFloatZoneData(Integer level, int step){ 
#pragma unused(step)  
  
  int        err=noErr;
  Integer          zone;
  Integer      varIndex;
  Integer       ioIndex;
  ZonePtr             z;
  
  IOSettingsPtr  io = NULL;
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
    
    ZonePtr mergedZone    = NULL;
    Integer bDoMergedData = 0;
    
    //#ifndef MPI        
    ////
    //// only merge full data files for non-MPI jobs, for IO performance reasons
    //// in massively parallel MPI jobs
    //// data headers contain all the information needed to reconstruct 
    //// multiple zones into levels later
    ////
    //    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
    //      if ( gLevels[level].nZones >1) { // then merge output
    //        zls_GetLevelImageZone( &mergedZone, level ); 
    //      } 
    //    }
    //    
    //#endif !MPI
#ifdef MPI
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        if ( gMPIrank == level) { // collect merged output on rank == level only
          zls_GetLevelImageZone( &mergedZone, level ); 
        }
        bDoMergedData = 1; // but all ranks participate in merger
      } // other levels do separate zones and imageZone stays NULL;
    }
#else    
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        zls_GetLevelImageZone( &mergedZone, level );
        if ( mergedZone != NULL) bDoMergedData = 1;
      } // other levels do separate zones and imageZone stays NULL;
    }
#endif !MPI
    
    
    for ( ioIndex = 0; ioIndex< io->outNVars; ioIndex++){
      
      if ((io->v[ioIndex].ioTypes & kZLS_IOData) == kZLS_IOData) { // data io type flag
        
        varIndex = kAnonVar;
        
        err = zlsio_GetIOCodeVarIndex (&varIndex, io->v[ioIndex].varCode);
        if ( err != noErr ) {
          zlsio_ErrPrintf( " IO ERROR: varCode unknown in zlsio_SaveLevel32BitFloatZoneData, %d %s Err = %d \n"
                          , ioIndex, io->v[ioIndex].varCode, err );
          err = ioErr;
          return (err);
        }
        
        if ( varIndex != kNoVar ) { 
          
          
          if ( bDoMergedData == 1) {
#ifdef MPI
            err =  zls_MPI_GetLevelMergedZoneData( mergedZone, level, varIndex );
#else 
            err =  zls_GetLevelMergedZoneData( mergedZone, level, varIndex );
#endif !MPI
          } // imageZone may be null on non rank 0 processes, they send buffers to 0 which collects all
          
          if ((err == noErr) &&( bDoMergedData == 1 )) { // if the merge succeeded
            
#ifdef MPI  
            if ( gMPIrank == level) { // only rank == level can to write out merged image
#endif MPI
              // imageZone always has data in zoneData[0], so kAnonVar is used here
              
              err = zlsio_Save32BitFloatZoneData ( mergedZone, varIndex, level, 0, io, ioIndex );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO WARNING: zlsio_Save32BitFloatZoneData Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
              }
#ifdef MPI  
            }
#endif MPI
            
          } else { 
            
            // do normal separate output, either mergeZone failed or merger not requested
            
            
            //
            // loop over zones separately
            //
            
            
            for( zone = 0; zone <gLevels[level].nZones; zone++){
              
              err = zls_GetLevelZoneByIndex( &z, level, zone );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
                err = ioErr;
                return (err);
              }
              
#ifdef MPI  
              if (z->zoneRankID == gMPIrank) {
#endif MPI
                
                err = zlsio_Save32BitFloatZoneData (  z, varIndex, level, zone, io, ioIndex );
                if ( err != noErr ){
                  zlsio_ErrPrintf( " IO WARNING: zlsio_Save32BitFloatZoneData Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                  zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
                }
#ifdef MPI  
              }
#endif MPI
              
            } // zone loop
            
          }
          
        } // !kNoVar
        
      } // data io type flag
      
    } // ioIndex
    
    if ( level == 0 ) io->ioState.ioDataCount++;
    
  } else { //io
    zlsio_ErrPrintf( " IO ERROR: IOSettings not found zlsio_SaveLevel32BitFloatZoneData. Err %d\n", err );
    err = ioErr;
  }
  
  return (err);
}


//
//========================================
//
//  Read (External) Data Files
//
//========================================
//

//========================================
//
int zlsio_ReadRealData      ( Real    *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = nyiErr;
  
  zFile fileAccess;
  Size chunkSize;
  Counter nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) q; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( q != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      if ( skip > 0) {
        err  =  zlsio_seek(fileAccess, skip);
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadRealData: Skip Bytes Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      chunkSize = ni*nj*sizeof(Real);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nChunks; k++ ) {
        err  =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadRealData: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
#ifdef SINGLE_PRECISION 
        zls_ByteSwapFloatBlock ( q, chunkSize*nChunks);
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION
        zls_ByteSwapDoubleBlock( q, chunkSize*nChunks);
#endif DOUBLE_PRECISION
      }
      
      zlsio_close( fileAccess );
      
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_ReadRealData: File Not Opened Error: %s \n", inFile );
    return ( fileErr ); } // access
  } else {
    zlsio_ErrPrintf( " ERROR: zlsio_ReadRealData: Buffer Not Allocted: %d \n", memErr );
  return ( memErr ); } // buffer
  
  return (err);
}
//
//  As above, but Integer is always signed 32 bit.
//
int zlsio_ReadIntegerData ( Integer    *  iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = noErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) iq; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( iq != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(Integer);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadIntegerData: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapInt32Block ( iq, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

//
// type specific reads - unsigned and signed, and specific floats - usually for externally created data.
//
int zlsio_ReadUInt32Data    ( UInt32  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = noErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) iq; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( iq != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(UInt32);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadUInt32Data: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapInt32Block ( iq, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

int zlsio_ReadUInt16Data    ( UInt16  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = noErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) iq; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( iq != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(UInt16);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadUInt16Data: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapInt16Block ( iq, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

int zlsio_ReadUInt8Data     ( UInt8   * iq, Counter ni, Counter nj, Counter nk, Size skip, char * inFile){
#pragma unused(skip)
  int err = noErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) iq; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( iq != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(UInt8);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadUInt8Data: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // no swap data for 8 bit - just a byte stream
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

//
int zlsio_ReadSInt16Data    ( SInt16  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = noErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) iq; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( iq != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(SInt16);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadSInt16Data: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapInt16Block ( iq, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

//
int zlsio_ReadDoubleData    ( double  *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = nyiErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) q; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 2D and 3D nk = 1 and a single read is done;
  if ( q != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(double);
      nChunks   = nk;
      
      zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err       =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadDoubleData: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapDoubleBlock( q, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
  } else { return ( memErr ); } // buffer
  
  return (err);
}

int zlsio_ReadFloatData     ( float   *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile){
#pragma unused(skip)
  int err = nyiErr;
  
  zFile fileAccess;
  Size chunkSize;
  Size nChunks;
  Counter k;
  UInt8 * p = (UInt8 *) q; // treat buffer as a byte stream
  
  // default to reading 3D sets in nk slices, assumes ni and nj are big enough to make 
  // the system's buffered IO efficient. in 1D and 2D nk = 1 and a single read is done;
  if ( q != NULL ){
    
    fileAccess = zlsio_open( inFile, "rb");
    
    if (fileAccess) {
      
      chunkSize = ni*nj*sizeof(float);
      nChunks   = nk;
      
      zlsio_Printf( " Reading float Variable of size: %u bytes, X %u Y %u Z %u\n", chunkSize*nChunks, ni, nj, nk);
      
      for ( k = 0; k < nk; k++ ) {
        err  =  zlsio_read( fileAccess, p, chunkSize );
        p   += chunkSize;
        if ( err != noErr ){
          zlsio_ErrPrintf( " ERROR: zlsio_ReadFloatData: Data Read Error: %d \n", err );
          zlsio_close(fileAccess);
          return (err);
        } //header
      }
      
      // swap data if needed
      if ( swapBytes == kZLS_SWAPFormat ) {
        zls_ByteSwapFloatBlock ( q, chunkSize*nChunks);
      }
      
      zlsio_close( fileAccess );
      
    } else { return ( fileErr ); } // access
    
  } else { return ( memErr ); } // buffer
  
  return (err);
}


//
//========================================
//
// mid level call, allocates QPtr, uses SYM and BCs info plus grid (usually from a zone) to determine
// dimensions etc.  Returns plain block with *no* BCs, can be used in a SetZoneDataCopy call after
//
//
int zlsio_GetFileRealData ( QPtr *q, GridCoords g, char * inFile, Size skip, Integer swapBytes){
  int err = noErr;
  QPtr  d = NULL;
  
#ifdef SYM1D
  Counter ni = g.n[kXDim];
  
  err = zls_NewQuantity1D(&d, ni);
  if ( err != noErr ){
    zls_DisposeQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadRealData ( d, ni, 1L, 1L, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
  
#endif SYM1D
  
#ifdef SYM2D
  Counter ni = g.n[kXDim];
  Counter nj = g.n[kYDim];
  
  err = zls_NewQuantity2D(&d, ni, nj);
  if ( err != noErr ){
    zls_DisposeQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadRealData ( d, ni, nj, 1L, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
#endif SYM2D
  
#ifdef SYM3D
  Counter ni = g.n[kXDim];
  Counter nj = g.n[kYDim];
  Counter nk = g.n[kZDim];
  
  err = zls_NewQuantity3D(&d, ni, nj, nk);
  if ( err != noErr ){
    zls_DisposeQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadRealData ( d, ni, nj, nk, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
  
#endif SYM3D
  *q = d;
  return(err);
}


int zlsio_ReadZoneData       ( ZonePtr z, Integer varIndex, char * inFile, Integer swapBytes){
  int err = noErr;
  QPtr q=NULL;
  
  // int   zls_SetZoneDataCopy(ZonePtr z, QPtr d, Integer varIndex);
  // and assumes file does not include BCs. May use temporary buffers if
  // the zone data does have BCs, but copies straight if the zone doesn't
  
  if ( z != NULL ){
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
      err =  zlsio_GetFileRealData ( &q, z->zoneGrid,  inFile, kZLS_NoSkip, swapBytes );
      
      if ( err != noErr ){
        zls_DisposeQuantity(q);
        return( err ); 
      }
      
      if ( q != NULL ){
        
        err = zls_SetZoneDataCopy( z, q, varIndex);
        if ( err != noErr ){
          zls_DisposeQuantity(q);
          return( err ); 
        }
        zls_DisposeQuantity(q);
      }
      
    } else { err = dimErr; }
  } else { err = memErr; }
  
  return(err);
  
}

int zlsio_ReadZoneDataDirect ( ZonePtr z, Integer varIndex, char * inFile, Integer swapBytes){
  int err = noErr;
  QPtr q;
  
  // Uses zls_SetZoneDataDirect(ZonePtr z, QPtr d, Integer varIndex) to read straight into
  // the zone data.  Minimum memory use, but a little more complex to handle cases.
  
  
  if ( z != NULL ){
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
#ifdef ZONEDATA_ARE_PLAINBLOCKS
      Counter nti = z->zoneGrid.nt[kXDim];
      q = z->zoneData[varIndex]; // simple isn't it!!!
      err = zlsio_ReadRealData ( q, nti, 1L, 1L, kZLS_NoSkip, swapBytes, inFile);
#endif ZONEDATA_ARE_PLAINBLOCKS
      
#ifdef ZONEDATA_ARE_MULTIDARRAYS
      
#ifdef SYM1D
      Counter nti = z->zoneGrid.nt[kXDim];
      q   = z->zoneData[varIndex]; // same as a plain block
      err = zlsio_ReadRealData ( q, nti, 1L, 1L, kZLS_NoSkip, swapBytes, inFile);
#endif SYM1D
      
#ifdef SYM2D
      Counter nti = z->zoneGrid.nt[kXDim];
      Counter ntj = z->zoneGrid.nt[kYDim];
      err = zls_GetZoneDataDirect( &q, z, varIndex); // copy zone data reference, BCs n all if present
      if ( err != noErr ) {
        return(err);
      }
      err = zlsio_ReadRealData ( q, nti, ntj, 1L, kZLS_NoSkip, swapBytes, inFile);
#endif SYM2D
      
#ifdef SYM3D
      Counter nti = z->zoneGrid.nt[kXDim];
      Counter ntj = z->zoneGrid.nt[kYDim];
      Counter ntk = z->zoneGrid.nt[kZDim];
      err =  zls_GetZoneDataDirect( &q, z, varIndex); // copy zone data reference, BCs n all if present
      if ( err != noErr ) {
        return(err);
      }
      err = zlsio_ReadRealData ( q, nti, ntj, ntk, kZLS_NoSkip, swapBytes, inFile);
#endif SYM3D
      
#endif ZONEDATA_ARE_MULTIDARRAYS
      
    } else { err = dimErr; }
  } else { err = memErr; }
  
  return(err);
  
}

int zlsio_SaveLevelZoneData(Integer level, int step){ 
#pragma unused(step)  
  
  int        err=noErr;
  
  IOSettingsPtr  io = NULL;
  err = zls_GetLevelIOPrefs( &io, level );
  if (io != NULL) {
    
    if  (io->outDataFormat == kZLS_FITSData ){
      
      zlsio_SaveLevel32BitFloatZoneFITS( level, gLevels[0].levelClock.cycles);
      
    } else {
      
      
      switch (io->outDataBitsPerCell){
        default:
        case kZLS_Float32Data:
          zlsio_SaveLevel32BitFloatZoneData( level, gLevels[0].levelClock.cycles);
          break;
        case kZLS_SInt16Data:
          zlsio_SaveLevel16BitZoneData( level, gLevels[0].levelClock.cycles);
      }
      
    }
    
  }
  return(err);
  
}

int zlsio_GetFileFloatData ( FQPtr *q, GridCoords g, char * inFile, Size skip, Integer swapBytes){
  int  err = noErr;
  FQPtr  d = NULL;
  
#ifdef SYM1D
  Counter ni = g.n[kXDim];
  
  err = zls_NewFQuantity1D(&d, ni);
  if ( err != noErr ){
    zls_DisposeFQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadFloatData ( d, ni, 1L, 1L, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
  
#endif SYM1D
  
#ifdef SYM2D
  Counter ni = g.n[kXDim];
  Counter nj = g.n[kYDim];
  
  err = zls_NewFQuantity2D(&d, ni, nj);
  if ( err != noErr ){
    zls_DisposeFQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadFloatData ( d, ni, nj, 1L, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
#endif SYM2D
  
#ifdef SYM3D
  Counter ni = g.n[kXDim];
  Counter nj = g.n[kYDim];
  Counter nk = g.n[kZDim];
  
  err = zls_NewFQuantity3D(&d, ni, nj, nk);
  if ( err != noErr ){
    zls_DisposeFQuantity(d);
    *q = NULL;
    return( err ); 
  }
  
  if ( d != NULL ){
    err = zlsio_ReadFloatData ( d, ni, nj, nk, skip, swapBytes, inFile);
    if ( err != noErr ){
      return( err ); 
    }
  }
  
#endif SYM3D
  *q = d;
  return(err);
}


