#ifndef ZLS_IODATA_H
#define ZLS_IODATA_H
#include "zls.h"

// align fields on 4 byte boundaries, even if naturally 8 byte aligned.
// requires gcc to work, if not then you will generate slightly larger headers.

#pragma pack(4)

typedef struct {
  //------- ZLS BINARY FILE HEADER IDENTIFER AND GLOBAL OFFSET
  //
  UInt32 fMagicTag;      // magic number 'ZLS!' to idenify zls  bin files
  UInt32 fDataOffset;    // offset to beginning of data from start of file, header limited to 4GB ! :)
  UInt32 fVersion;       // 4 byte int version 
} binaryid;
#define ZLS_ID_HEADERSIZE (sizeof(binaryid))

typedef struct {
  //------- ZLS BINARY FILE HEADER Tag
  UInt32 fTag;                   
  UInt32 fRecSize;    // Offset to next record, after reading this value = 4 bytes
} binarytagstream;
#define ZLS_TAGSIZE (sizeof(binarytagstream))


typedef struct {
//
// a little like TIFF, only with +ve only offsets in a contiguous series at the start of the file
// a bit more restrictive than TIFF, which does offsets from the start of file and allows 
// multiple references to tagged data at any offset.  Here all offsets are relative to current location
// and go forward, compatible with the stream metaphor, and not requiring rewinds to read or write.
//
// Only exception is the fDataOffset which is from the start of file, but this is part of the 12 byte 
// meta header (magic, size and version) and not part of the tag stream as such.
//
// This stucture is extendible as new fields can be added and old readers will skip unknown/unneeded tags etc.
//
// Depends on reader already knowing the internal format of each tagged record type, internal record format
// is not encoded in the record.
//
//------- ZLS BINARY FILE HEADER IDENTIFER AND GLOBAL OFFSET
//
UInt32 fMagicTag;      // magic number 'ZLS!' to idenify zls  bin files
UInt32 fDataOffset;    // offset to beginning of data from start of file, header limited to 4GB ! :)
UInt32 fVersion;       // 4 byte int version 
//
//------- TAGGED RECORD STREAM FOLLOWS
//
UInt32 fFormatTag;     // File Format Tag                
UInt32 fFormatRecSize; // Offset to next record, after reading this value = 4 bytes      
UInt32 fFileType;      // File Format = kZLS_SInt16Data

UInt32   fTimeTag;     // Model Time tag:                
UInt32 fTimeRecSize;   // Offset to next record, after reading this value = 12 bytes      
double      fTime;     // global model (scaled) time, t, extended to double if single, 2x4 bytes 
double fTimeScale;     // Fiducial Scaling factor to recover real time, ie time = t * scale 
UInt32    fCycles;     // Model steps (4 bytes) 

UInt32 fVarTag;        // Variable tag:                
UInt32 fVarRecSize;    // Offset to next record, after reading this value = 4 bytes          
UInt32 fVarType;       // Variable type code    

UInt32 fUnitsTag;      // units tag
UInt32 fUnitsRecSize;  // Offset to next record, after reading this value = 12 bytes          
double fScalingUnit;   // Fiducial scaling factor to convert normalised data value to actual units, 1.0 for adiabatic case

UInt32 fTransTag;      // transform tag
UInt32 fTransRecSize;  // Offset to next record, after reading this value = 12 bytes          
UInt32 fTransType;     // transform used on data
double fTransPow;      // transform power index if used, 8 byte field

UInt32 fDataTag;       // Data scaling information tag, this record has internal tags to to make hacking a file easier
UInt32 fDataRecSize;   // Offset to next record, after reading this value = 40 bytes          
UInt32 fDataMaxTag;    // Data maximum index
SInt32 fDataMax;       // Max value used in data (= 32000 usually) in signed 32 int for alignment
UInt32 fDataMinTag;    // Data minimum index
SInt32 fDataMin;       // Min value used in data (= 0 usually) in signed 32 int for alignment
UInt32 fMaxTag;        // transform tag
double fMax;           // Value to associate with datamax, 8 byte field
UInt32 fMinTag;        // Min value for data min index
double fMin;           // Value to associate with datamin, 8 byte field

UInt32 fDimsTag;     // 'DIMS' tag , whole cube adjusted to include BCs if IO settings require
UInt32 fDimsRecSize;   // Offset to next record, after reading this value = 36 bytes          
UInt32 fDims;        // 1, 2 or 3, any other value is an error
UInt32 fXDimTag;     // 'XDIM' tag
UInt32 fXDim;        // int, >= 1
UInt32 fYDimTag;     // 'YDIM' tag
UInt32 fYDim;        // int, >= 1, 1 if 1D,  > 1 if 2D or 3D
UInt32 fZDimTag;     // 'ZDIM' tag
UInt32 fZDim;        // int, >= 1, 1 if 1D or 2D,  > 1 if 3D
UInt32 fNGSTTag;     // 'NGST' tag
UInt32 fNGST;        // int, 0, 2, 4 etc, number of BC cells 
           // in data,to aid interpretation

UInt32 fRngeTag;     // 'RNGE' tag , location of data in super grid, to allow reconstruction of relative location of data
UInt32 fRngeRecSize; // Offset to next record, (fSampleTag currently) after reading this value       
UInt32 fRngeDims;    // 1, 2 or 3, any other value is an error, should == fDims
// use this to determine which of the following three records to use.  All three are always
// present, but some may not be used, and just contains unit values

UInt32 fXrngTag;       // 'XRNG' tag
UInt32 fXrngRecSize;  // Offset to next record within the RNGE record, after reading this value       
UInt32 fXTypeTag;     // 'XTYP' tag
UInt32 fXType;        // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
UInt32 fXKindTag;     // 'XKND' tag
UInt32 fXKind;        // int, >= 0,  0 = linear, 1 = log 
UInt32 fXn0Tag;       // 'XCRD' coordinates tag
SInt32 fXn0;          // int,= n0 index of first active cell on level, not counting NGST if NGST>0
SInt32 fXrn;          // int, index ranges index ne =  n0+rn :  i = n0; i<ne; i++
      // index of first active cell on super grid, not counting NGST if NGST>0
double fXx0;           // coordinate value of left edge, fXn0, 8 byte field
double fXrx;           // coordinate range of fXrn, 8 byte field
double fXscale;        // scaling factor for X-coords to recover real units

UInt32 fYrngTag;       // 'YRNG' tag
UInt32 fYrngRecSize;  // Offset to next record within the RNGE record, after reading this value        
UInt32 fYTypeTag;     // 'YTYP' tag
UInt32 fYType;        // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
UInt32 fYKindTag;     // 'YKND' tag
UInt32 fYKind;        // int, >= 0,  0 = linear, 1 = log 
UInt32 fYn0Tag;       // 'YCRD' coordinates tag
SInt32 fYn0;          // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
SInt32 fYrn;          // int, index ranges index ne =  n0+rn :  i = n0; i<ne; i++
      // index of first active cell on super grid, not counting NGST if NGST>0
double fYx0;           // coordinate value of left edge, fYn0, 8 byte field
double fYrx;           // coordinate range of fYrn, 8 byte field
double fYscale;        // scaling factor for Y-coords to recover real units


UInt32 fZrngTag;       // 'ZRNG' tag
UInt32 fZrngRecSize;  // Offset to nex record within the RNGE record, after reading this value        
UInt32 fZTypeTag;     // 'ZTYP' tag
UInt32 fZType;        // int, >= 0,  0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
UInt32 fZKindTag;     // 'ZKND' tag
UInt32 fZKind;        // int, >= 0,  0 = linear, 1 = log 
UInt32 fZn0Tag;       // 'ZCRD' coordinates tag
SInt32 fZn0;          // int,= n0 index of first active cell on super grid, not counting NGST if NGST>0
SInt32 fZrn;          // int, index ranges index ne =  n0+rn :  i = n0; i<ne; i++
      // index of first active cell on super grid, not counting NGST if NGST>0
double fZx0;           // coordinate value of left edge, fYn0, 8 byte field
double fZrx;           // coordinate range of fYrn, 8 byte field
double fZscale;        // scaling factor for Z-coords to recover real units

UInt32 fSampleTag;   // 'SMPL'
UInt32 fSampleRecSize;   // Offset to next record, after reading this value = 4 bytes          
UInt32 fSampleSize;  //  zone data element size in bytes, == sizeof(SInt16) etc

} binaryDataHeader;

#define ZLS_DATA_HEADERSIZE (sizeof(binaryDataHeader))

#pragma pack()

//========================================
//
//  gateway call to all binary data saves, at the moment it selects 16 or 32 bit IO
// will do formats in the future too.
//
int zlsio_SaveLevelZoneData(Integer level, int step); 
//
//========================================
//
// Low level data scaling routines, used by image routines as well
//

// v. low level, 64 bit ready

int zls_DataMaxMinLinear   ( Real *max, Real *min, QPtr q, Size nelements);
int zls_DataMaxMinLog10    ( Real *max, Real *min, QPtr q, Size nelements);
int zls_DataMaxMinLog10Abs ( Real *max, Real *min, QPtr q, Size nelements);
int zls_DataMaxMinExp10    ( Real *max, Real *min, QPtr q, Size nelements);
int zls_DataMaxMinPow      ( Real *max, Real *min, QPtr q, Size nelements, Real a);

//
// scale q into iq according to io and ioIndex
// low level, treats qs as 1D, with n elements

// Reverses Y axis when nk = 1 (2D image data)
int zls_ScaleUnsigned16BitData  ( UInt16 * iq , QPtr q,  Integer dataType,
          Counter ni, Counter nj, Counter nk, 
          IOSettingsPtr io, Integer ioIndex );

// Never Reverses Y axis when nk = 1 (multi-D data)
int zls_ScaleSigned16BitData  ( SInt16 * iq , QPtr q,  Integer dataType, 
        Counter ni, Counter nj, Counter nk, 
        IOSettingsPtr io, Integer ioIndex );

// Reverses Y axis when nk = 1 (2D image data)
int zls_Scale8BitData       ( UInt8* iq , QPtr q,  Integer dataType,
        Counter ni, Counter nj, Counter nk, 
        IOSettingsPtr io, Integer ioIndex );

// like zls_Scale8BitData, but produces a ni x 256 8 bit plot of a 1D row

int zls_Plot1D8BitData       ( UInt8* iq , QPtr q,  Integer dataType,
                             Counter ni, IOSettingsPtr io, Integer ioIndex );
//
// not really scaled, but transformed.  Will clip to 32bit float range
// from double precision input but otherwise only the log/exp/pow etc
// transforms are done
//
int zls_Scale32BitFloatData       ( float * fq , QPtr q,  Integer dataType,
          Counter ni, Counter nj, Counter nk, 
          IOSettingsPtr io, Integer ioIndex );
//
//========================================
//
//
int zls_Allocate16BitZoneData ( short ** iq , ZonePtr z );
int zls_Free16BitZoneData     ( short * iq );

int zls_Allocate32BitFloatZoneData ( float ** fq , ZonePtr z );
int zls_Free32BitFloatZoneData ( float * fq );

//
//
//========================================
// When merged IO for a level has been selected, get GetWQuantity result
// is mapped into a level grid sized zone.
// used by image and data io loops.
//
int zls_GetLevelMergedZoneData(ZonePtr mergedZone, Integer level, Integer varIndex );
//
// intended primarily for 3D data, but works for all.  save images
// is better for 2D and some 3D as well.  1D is best as text
// low level write a 16 bit buffer to disk, file speficied by name
//
//
int zlsio_Write16BitDataBinaryHeader ( zFile f, ZonePtr z, IOSettingsPtr io, Integer ioIndex );
int zlsio_Write16BitDataFile  ( zFile fileAccess, short * iq, Counter ni, Counter nj, Counter nk );

// higher level save a zone varaiable as a 16 bit data file, file speficied by io
int zlsio_Save16BitZoneData     ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
             IOSettingsPtr  io, Integer ioIndex );

// high level, save a level of zones as separate 16bit data files, same dims as zones
int zlsio_SaveLevel16BitZoneData ( Integer level, int step );
//
//========================================
// 32 bit single precision float files, better range than 16, but bigger
//intended primarily for 3D data, but works for all.  save images
// is better for 2D and some 3D as well.  1D is best as text
// low level write a 16 bit buffer to disk, file speficied by name
//

int zlsio_Write32BitFloatDataBinaryHeader ( zFile f, ZonePtr z, IOSettingsPtr io, Integer ioIndex  );
int zlsio_Write32BitFloatDataFile  ( zFile fileAccess, float * fq, Counter ni, Counter nj, Counter nk );

int zlsio_Save32BitFloatZoneData  ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
          IOSettingsPtr  io, Integer ioIndex );

// high level, save a level of zones as separate 32 bit float data
// files, same dims as zones
int zlsio_SaveLevel32BitFloatZoneData(Integer level, int step);

//========================================
//
//
// NOTE: these calls read raw data - without the format header
// information used in zls_restart.c/h .  These calls cannot determine
// byteswapping as a result and will simply read data as a block.  It
// is up to the user to ensure the data file is the correct format.  To
// save a zone data set in binary format for read in independently of
// the architecture of the cpu, use the restart routines, which could
// be used for purposes other than restarts. If the data is known to be
// swapped relative to the progam, then pass kZLS_SWAPFormat for the
// swapBytes argument and it will be swapped manually.  pass 0 or
// kZLS_NOSWAPFormat if swapping is not needed.
//
// Data files may be compressed as .gz files if ZLIB_COMPRESSED_IO is
// on, plain data files can be read in all cases.
//
// Low level call to load a 1D 2D or 3D data file into a buffer of type
// Real. The size of Real depends on the Single (32 bit) or Double (64
// bit) precision settings in force.  Any n can be 1 for data less than
// 3D. skip is the number of elements to skip before reading.  Use skip
// to pass headers or even read parts with sequential calls. These
// calls open and close the file each time, user needs to keep track of
// successive skips if desired.  For one step reads just set skip to 0.
//  Not implemented yet and ignored in current routines.
//
// To read into a multi-dimensional array, use the base block pointer
// to refer to the data: a is Q1DPtr use a a is Q1DArr use a a is
// Q2DArr use a[0] a is Q3DArr use a[0][0]
//
// To read into a QPtr for a zone data array use 
//
// int   zls_GetZoneDataDirect(QPtr *d, ZonePtr z, Integer varIndex);
//
// To get the base block pointer in d, for 1, 2 or 3D data. multi-d or
// plain blocks and use d as the destination in these calls. Get zone
// dimensions and sizes from z->zoneGrid.  Otherwise you can allocate 
// a buffer manually if needed with malloc, this would be sensible for
// temporary buffers that are being rescaled before use.
//
int zlsio_ReadRealData      ( Real    *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);
//
//  As above, but Integer is always signed 32 bit.
//
int zlsio_ReadIntegerData   ( Integer * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes,  char * inFile);
//
// type specific reads - unsigned and signed, and specific floats - usually for externally created data.
//
int zlsio_ReadUInt32Data    ( UInt32  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);
int zlsio_ReadUInt16Data    ( UInt16  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);

// no byte swap option for single byte data, so no swapBytes arg.
int zlsio_ReadUInt8Data     ( UInt8   * iq, Counter ni, Counter nj, Counter nk, Size skip, char * inFile);
//
int zlsio_ReadSInt16Data    ( SInt16  * iq, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);
//
int zlsio_ReadDoubleData    ( double  *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);
int zlsio_ReadFloatData     ( float   *  q, Counter ni, Counter nj, Counter nk, Size skip, Integer swapBytes, char * inFile);
//
//========================================
//
// mid level call, allocates QPtr, uses SYM and info from grid (usually
// from a zone) to determine dimensions etc.  Returns plain block with
// *no* BCs, can be used in a SetZoneDataCopy call after for zones with
// BCs or used for other purposes.  A multi-d array can be formed
// around the resulting q , using
// zls_SetQuantity2DArr/zls_SetQuantity3DArr as needed
//

int zlsio_GetFileRealData ( QPtr *q, GridCoords g, char * inFile, Size skip, Integer swapBytes);

//
// higher level read a zone varaiable from a Real data file, calls
// zlsio_ReadRealData and so respects the SINGLE/ DOUBLE precision
// settings. Assumes file does *NOT* include BCs and will offset data
// into zonedata by allocating a temporary buffers if needed.
//

int zlsio_ReadZoneData     ( ZonePtr z, Integer varIndex, char * inFile, Integer swapBytes);

//
// higher level read a zone varaiable from a Real data file, calls
// zlsio_ReadRealData and so respects the SINGLE/ DOUBLE precision
// settings. Assumes file matches the zone size exactly, including BCs
// if present.  Uses less memory by reading diretly into the zone data
// arrays.
//

int zlsio_ReadZoneDataDirect ( ZonePtr z, Integer varIndex, char * inFile, Integer swapBytes);
//
//========================================
//
int zlsio_GetFileFloatData ( FQPtr *q, GridCoords g, char * inFile, Size skip, Integer swapBytes);
//
//========================================
#endif ZLS_IODATA_H
