#include "zls.h"

#define FITSCARDSIZE 2880

Counter g_nLine = 0;

// low level fits write

void SetCharLine(  zFile f, charline *chl, char* k, char * v);
void SetCharLine(  zFile f, charline *chl, char* k, char * v){
  Size len = ZLS_FITS_LINESIZE;
  int i = strlen(k); int j = strlen(v);
  int l = 0;
  for ( l = 0; l< 8; l++) chl->key[l] = ' '; // clear key
  for ( l = 0; l< i; l++) chl->key[l] = k[l]; // set key chars
  for ( l = 0; l< 71; l++) chl->value[l] = ' '; // clear value str
  for ( l = 0; l< j; l++) chl->value[l] = v[l]; // set key chars
  zlsio_write( f, chl, len ); // ignoring errors here :)
  g_nLine++;
}

int zlsio_Write32BitFloatFITSHeader ( zFile f, ZonePtr z, IOSettingsPtr io, Integer ioIndex  ){
  
  int err = nyiErr;
  
  if (f) {
    
    if ( z != NULL) {
      
      time_t writeTime;
      writeTime = time(NULL);
      
      charline chl;
      chl.sep = '=';
      
      Size  remainder = 0;
      char keyString[9];
      char valString[72];
      
      Integer   varIndex = kAnonVar;
      zlsio_GetIOCodeVarIndex( &varIndex,io->v[ioIndex].varCode);
      
      Real scaleFactor = 1.0;
      zlsio_GetVarFiducialValue( &scaleFactor, varIndex);
      
      Real   transPower = 1.0;
      transPower = io->v[ioIndex].transPow;
      
      Counter ni ;
      Counter nj ;
      Counter nk ;
      
      GridCoords gc = z->zoneGrid;
      zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc ); // included BCs if present and requested
      
      // STANDARD FITS KEYWORDS:    
      
      g_nLine = 0;
      sprintf(keyString,"SIMPLE");  sprintf(valString," T ");  SetCharLine(f, &chl, keyString, valString);
      sprintf(keyString,"BITPIX");  sprintf(valString," -32");  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"COMMENT");  sprintf(valString," Variable  = %s", io->v[ioIndex].varCode);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"BZERO");  sprintf(valString," 0.0");  SetCharLine( f, &chl, keyString, valString );
      
      switch (io->v[ioIndex].transformType){
          
        case kZLS_IOLinearTransform:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Linear Scaling");  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"BSCALE");  sprintf(valString," %#16.9g", scaleFactor );  SetCharLine( f, &chl, keyString, valString );
          break;
        case kZLS_IOLog10Transform:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Log10 Scaling");  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"BSCALE");  sprintf(valString," %#16.9g", zls_Log10(scaleFactor) );  SetCharLine( f, &chl, keyString, valString );
          break;
        case kZLS_IOExp10Transform:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Exp10 Scaling");  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"BSCALE");  sprintf(valString," %#16.9g", zls_Exp10(scaleFactor) );  SetCharLine( f, &chl, keyString, valString );
          break;
        case kZLS_IOPowTransform:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Powerlaw Scaling");  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Powerlaw alpha %g",transPower );  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"BSCALE");  sprintf(valString," %#16.9g", zls_Pow( scaleFactor, transPower ) );  SetCharLine( f, &chl, keyString, valString );
          
          break;
        case kZLS_IOLog10AbsTransform:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Log10|x| Scaling");  SetCharLine( f, &chl, keyString, valString );
          sprintf(keyString,"BSCALE");  sprintf(valString," %#16.9g", zls_Log10(zls_abs(scaleFactor)) );  SetCharLine( f, &chl, keyString, valString );
          break;
        default:
          sprintf(keyString,"COMMENT");  sprintf(valString," Transform: Unknown!");  SetCharLine( f, &chl, keyString, valString );
      }
      
      
      sprintf(keyString,"NAXIS");  sprintf(valString,"  %d", kSymDims);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"NAXIS1");  sprintf(valString," %d", ni);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>1)
      sprintf(keyString,"NAXIS2");  sprintf(valString," %d", nj);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>2)
      sprintf(keyString,"NAXIS3");  sprintf(valString," %d", nk);  SetCharLine( f, &chl, keyString, valString );
#endif
#endif    
      
      
      sprintf(keyString,"CRPIX1");  sprintf(valString," %d", 1);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>1)
      sprintf(keyString,"CRPIX2");  sprintf(valString," %d", 1);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>2)
      sprintf(keyString,"CRPIX3");  sprintf(valString," %d", 1);  SetCharLine( f, &chl, keyString, valString );
#endif
#endif    
      
      sprintf(keyString,"CRVAL1");  sprintf(valString," %#16.9g", gc.xl[kXDim]);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>1)
      sprintf(keyString,"CRVAL2");  sprintf(valString," %#16.9g", gc.xl[kYDim]);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>2)
      sprintf(keyString,"CRVAL3");  sprintf(valString," %#16.9g", gc.xl[kZDim]);  SetCharLine( f, &chl, keyString, valString );
#endif
#endif    
      
      sprintf(keyString,"CDELT1");  sprintf(valString," %#16.9g", gc.dx[kXDim]);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>1)
      sprintf(keyString,"CDELT2");  sprintf(valString," %#16.9g", gc.dx[kYDim]);  SetCharLine( f, &chl, keyString, valString );
#if (NDIMS>2)
      sprintf(keyString,"CDELT3");  sprintf(valString," %#16.9g", gc.dx[kZDim]);  SetCharLine( f, &chl, keyString, valString );
#endif
#endif    
      
#ifdef CARTESIAN 
      sprintf(keyString,"CTYPE1");  sprintf(valString," X-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef CIRCULAR 
      sprintf(keyString,"CTYPE1");  sprintf(valString," R-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CIRCULAR
#ifdef SPHERICAL 
      sprintf(keyString,"CTYPE1");  sprintf(valString," R-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
      sprintf(keyString,"CTYPE1");  sprintf(valString," R-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
      sprintf(keyString,"CTYPE1");  sprintf(valString," Z-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
      
      
#if (NDIMS>1)
      
#ifdef CARTESIAN 
      sprintf(keyString,"CTYPE2");  sprintf(valString," Y-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef CIRCULAR 
      sprintf(keyString,"CTYPE2");  sprintf(valString," Phi-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CIRCULAR
#ifdef SPHERICAL 
      sprintf(keyString,"CTYPE2");  sprintf(valString," Phi-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
      
#ifdef SYM3D
#ifdef POLARSPHERICAL 
      sprintf(keyString,"CTYPE2");  sprintf(valString," Theta-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#else
#ifdef POLARSPHERICAL 
      sprintf(keyString,"CTYPE2");  sprintf(valString," Phi-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#endif
      
#ifdef CYLINDRICAL 
      sprintf(keyString,"CTYPE2");  sprintf(valString," R-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
      
#if (NDIMS>2)
#ifdef CARTESIAN 
      sprintf(keyString,"CTYPE3");  sprintf(valString," Z-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef SPHERICAL 
      sprintf(keyString,"CTYPE3");  sprintf(valString," Theta-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
      sprintf(keyString,"CTYPE3");  sprintf(valString," Phi-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
      sprintf(keyString,"CTYPE3");  sprintf(valString," R-Axis");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
#endif
#endif
      
      
      sprintf(keyString,"OBJECT");  sprintf(valString," %s", io->outRunName);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"ORIGIN");  sprintf(valString," %s v 0x%8.8x", FYRIS, ZLS_VERSION);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"DATE");  sprintf(valString," %s", ctime(&writeTime));  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"DATE-OBS");  sprintf(valString," %12.5g", zlst_GlobalTime *f_t0);  SetCharLine( f, &chl, keyString, valString );
      
      // NON-STANDARD KEYWORDS
      
      sprintf(keyString,"ZONELVL");  sprintf(valString," %d", z->zoneLevel);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"ZONEID");  sprintf(valString," %d", z->zoneID);  SetCharLine( f, &chl, keyString, valString );
#ifdef MPI    
      sprintf(keyString,"ZONERANK");  sprintf(valString," %d", z->zoneRankID);  SetCharLine( f, &chl, keyString, valString );
#endif MPI    
      sprintf(keyString,"CYCLE");  sprintf(valString," %d", zlst_GlobalCycles);  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"FRAME");  sprintf(valString," %d", io->ioState.ioFrameCount);  SetCharLine( f, &chl, keyString, valString );
      
      // BULK COMMENTS:    
      sprintf(keyString,"COMMENT");  sprintf(valString,"ZLS Algorithm Options:");  SetCharLine( f, &chl, keyString, valString );

#ifdef HYDRO      
#ifdef TORO 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Toro 1998 Full Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif TORO
#ifdef TOROTWOSHOCK 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Toro 1998, RSS Modified 2 Shock Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif TOROTWOSHOCK
#ifdef PVRS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Toro Primitive Variable Approximate Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif PVRS
#ifdef AIRS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Toro 1998 AIRS Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif AIRS
#ifdef COLLELLATWOSHOCK 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Collella 1984 Two Shock Solver (vh1, ppmlr).");  SetCharLine( f, &chl, keyString, valString );
#endif COLLELLATWOSHOCK
#ifdef GROTHGOTTLIEB 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: Groth and Gottlieb 1988 Fast Exact Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif GROTHGOTTLIEB
#ifdef RSSTWOSHOCK 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS RIEMANN: RSS Modified GG88 Two Shock Solver.");  SetCharLine( f, &chl, keyString, valString );
#endif RSSTWOSHOCK
      
#ifdef LAGRANGIAN 
#ifdef SPLITSWEEPS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: Split Lagrangian + Remap.");  SetCharLine( f, &chl, keyString, valString );
#else
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: ERROR in sweep mode.");  SetCharLine( f, &chl, keyString, valString );
#endif SPLITSWEEPS
      
#ifdef VISCOUS_REMAP
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: Viscous Remap.");  SetCharLine( f, &chl, keyString, valString );
#endif VISCOUS_REMAP
      
#else
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: Unsplit Eulerian.");  SetCharLine( f, &chl, keyString, valString );
#endif !LAGRANGIAN

#endif HYDRO
            
      
#ifdef UNIFORM_SWEEPS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: Multi-Level Uniform Sweeps.");  SetCharLine( f, &chl, keyString, valString );
#endif UNIFORM_SWEEPS 
#ifdef REFINED_SWEEPS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: Multi-Level Refined Sweeps.");  SetCharLine( f, &chl, keyString, valString );
#endif REFINED_SWEEPS 
      

#ifdef HYDRO
#ifdef PPM 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: 3rd order PPM Cell interpolation.");  SetCharLine( f, &chl, keyString, valString );
#endif PPM
#ifdef PLM 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: 2nd order PLM Cell interpolation.");  SetCharLine( f, &chl, keyString, valString );
#endif PLM
#ifdef GODUNOV 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS METHOD: 1st order Godunov Cell interpolation.");  SetCharLine( f, &chl, keyString, valString );
#endif GODUNOV
#endif HYDRO     
            
#ifdef COOLING 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: Radiative Cooling Enabled.");  SetCharLine( f, &chl, keyString, valString );
#endif COOLING
#ifdef VARIABLE_EOS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: Variable EOS Enabled.");  SetCharLine( f, &chl, keyString, valString );
#ifdef VARIABLE_GAMMA 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: Variable Gamma Enabled.");  SetCharLine( f, &chl, keyString, valString );
#endif VARIABLE_GAMMA
#ifdef H_IONISATION
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: H Ionisation Lookup Table Enabled.");  SetCharLine( f, &chl, keyString, valString );
#endif H_IONISATION
#endif VARIABLE_EOS
#ifdef GRAVITY 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: Gravitational Potential Field Enabled.");  SetCharLine( f, &chl, keyString, valString );
#ifdef GRAVZFIELD 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: Non-Radial Z-Field Potential.");  SetCharLine( f, &chl, keyString, valString );
#endif GRAVZFIELD
#if (FIELD_RECON_ORDER == 1)
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: PLM Conserved Potential Reconstruction.");  SetCharLine( f, &chl, keyString, valString );
#endif
#if (FIELD_RECON_ORDER == 2)
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PHYSICS: PPM Conserved Potential Reconstruction.");  SetCharLine( f, &chl, keyString, valString );
#endif
#endif GRAVITY
#ifdef PRIMITIVE_QUAS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data are physical (W) quantities.");  SetCharLine( f, &chl, keyString, valString );
#endif PRIMITIVE_QUAS
#ifdef PRIMITIVE_VARS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data are physical (W) variables+coeffs.");  SetCharLine( f, &chl, keyString, valString );
#endif PRIMITIVE_VARS
#ifdef CONSERVED_QUAS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data are conserved (U) quantities.");  SetCharLine( f, &chl, keyString, valString );
#endif PRIMITIVE_QUAS
#ifdef CONSERVED_VARS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data are conserved (U) variables+coeffs.");  SetCharLine( f, &chl, keyString, valString );
#endif CONSERVED_VARS
#if (NTRACERS > 0) 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data have %d variables and %d tracers.", kNVars, kNTracers);  SetCharLine( f, &chl, keyString, valString );
#else
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS VARS: Zone data have %d variables", kNVars);  SetCharLine( f, &chl, keyString, valString );
#endif NTRACERS
      
#ifdef ACTIVE_FP_NOISE_CONTROL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OPTIMIZE: Active Floating Point Noise Control Enabled.");  SetCharLine( f, &chl, keyString, valString );
#ifdef CONTROL_REMAP 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OPTIMIZE: AFPNC Enabled in Remap.");  SetCharLine( f, &chl, keyString, valString );
#endif CONTROL_REMAP
#ifdef CONTROL_EVOLVE 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OPTIMIZE: AFPNC Enabled in Evolve.");  SetCharLine( f, &chl, keyString, valString );
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL
#ifdef PRETEST_W_VEL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OPTIMIZE: W* velocity pre-test Enabled.");  SetCharLine( f, &chl, keyString, valString );
#endif PRETEST_W_VEL
      sprintf(keyString,"COMMENT");  sprintf(valString,"ZLS Compiled Geometry Options:");  SetCharLine( f, &chl, keyString, valString );
#ifdef SYM1D 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS SYMMETRY: Zone data blocks 1D.");  SetCharLine( f, &chl, keyString, valString );
#ifdef CARTESIAN 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Cartesian X");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef SPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 1D Spherical Radius ");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 1D Polar Spherical Radius");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 1D Cylindrical Z ");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
#endif SYM1D
#ifdef SYM2D 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS SYMMETRY: Zone data blocks 2D.");  SetCharLine( f, &chl, keyString, valString );
#ifdef CARTESIAN 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Cartesian XY");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef CIRCULAR 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 2D Circular Radius Phi");  SetCharLine( f, &chl, keyString, valString );
#endif CIRCULAR
#ifdef SPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 2D Spherical Radius Phi ");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 2D Polar Spherical Radius Phi");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: 2D Cylindrical Z Radius");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
#endif SYM2D
#ifdef SYM3D 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS SYMMETRY: Zone data blocks 3D.");  SetCharLine( f, &chl, keyString, valString );
#ifdef CARTESIAN 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Cartesian XYZ");  SetCharLine( f, &chl, keyString, valString );
#endif CARTESIAN
#ifdef CYLINDRICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Cylindrical Z Radius Phi");  SetCharLine( f, &chl, keyString, valString );
#endif CYLINDRICAL
#ifdef SPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Spherical Radius Phi Theta");  SetCharLine( f, &chl, keyString, valString );
#endif SPHERICAL
#ifdef POLARSPHERICAL 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS GEOMETRY: Polar Spherical Radius Theta Phi");  SetCharLine( f, &chl, keyString, valString );
#endif POLARSPHERICAL
#endif SYM3D
      sprintf(keyString,"COMMENT");  sprintf(valString,"ZLS Compiled OS/Architecture Options:");  SetCharLine( f, &chl, keyString, valString );
#ifdef SINGLE_PRECISION 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PRECISION: Single Precision Floating Point.");  SetCharLine( f, &chl, keyString, valString );
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS PRECISION: Double Precision Floating Point.");  SetCharLine( f, &chl, keyString, valString );
#endif DOUBLE_PRECISION
      
#if (ZLS_DEBUG > 0)
#ifdef ZONEDATA_ARE_PLAINBLOCKS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS DATA: Zone data blocks are single blocks.");  SetCharLine( f, &chl, keyString, valString );
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS DATA: Zone data blocks are multi-dimensional arrays.");  SetCharLine( f, &chl, keyString, valString );
#endif ZONEDATA_ARE_MULTIDARRAYS
#ifdef FILES_NO_BCS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS IO: Output files do not include boundary cells.");  SetCharLine( f, &chl, keyString, valString );
#endif FILES_NO_BCS
#ifdef FILES_HAVE_BCS
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS IO: Output files include boundary cells if present.");  SetCharLine( f, &chl, keyString, valString );
#endif FILES_HAVE_BCS    
      
#endif (ZLS_DEBUG > 0)
      
#ifdef ZLIB_COMPRESSED_IO
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS IO: Binary output files compressed as .gz ");  SetCharLine( f, &chl, keyString, valString );
#ifdef ZLIB_UNCOMPRESSED_IMAGES
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS IO: Image file output files uncompressed.");  SetCharLine( f, &chl, keyString, valString );
#endif ZLIB_UNCOMPRESSED_IMAGES
#endif ZLIB_COMPRESSED_IO
      
#ifdef MPI 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Parallel mode using MPI, nodes: %d.", gMPIsize );  SetCharLine( f, &chl, keyString, valString );
#endif MPI
#ifdef PTHREADS 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Parallel - Shared Memory Preemptive Threading.");  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Parallel - %d cpus.", NCPUs );  SetCharLine( f, &chl, keyString, valString );
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Parallel - %d Max Threads.", ZLS_MAXPTHREADS );  SetCharLine( f, &chl, keyString, valString );
#endif PTHREADS
#ifdef ZLS_LOADSORTING
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Parallel - Zone CPU Load Sorting.");  SetCharLine( f, &chl, keyString, valString );
#endif ZLS_LOADSORTING
      
#ifdef OS_GENERIC 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Generic GCC/Unix.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_GENERIC
#ifdef OS_GENERICLINUX
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Generic Little Endian Intel GCC/LINUX/Unix.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_GENERICLINUX
#ifdef OS_MACOSXCORE
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: MacOSX BSD/Darwin intel CORE littleendian.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_MACOSXCORE
#ifdef OS_MACOSXG5 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: MacOSX BSD/Darwin ppc970/G5 bigendian.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_MACOSXG5
#ifdef OS_MACOSXG4 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: MacOSX BSD/Darwin ppc7450/G4 bigendian.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_MACOSXG4
#ifdef OS_SOLARISSPARC 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Sun Solaris/Sparc bigendian.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_SOLARISSPARC
#ifdef OS_SOLARISAMD64 
      sprintf(keyString,"COMMENT");  sprintf(valString,"  ZLS OS: Sun Solaris 10 AMD64 Opteron i86 littleendian.");  SetCharLine( f, &chl, keyString, valString );
#endif OS_SOLARISAMD64
      
      sprintf(keyString,"END");  chl.sep = ' '; sprintf(valString,"");  SetCharLine( f, &chl, keyString, valString );
      
      remainder =  FITSCARDSIZE-((g_nLine*80) % FITSCARDSIZE);
      remainder = (remainder == FITSCARDSIZE)? 0: remainder;
      
      if ( remainder > 0) {
        // pad out to next 2880 bytes
        QPtr p = (QPtr) malloc(remainder);
        if ( p != NULL) {
          memset((void *) p, ' ', remainder); // pad with spaces.
          err    =  zlsio_write( f, p, remainder );
          free(p);
        }
        
      }
      
    } else { err = memErr; }
    
  }
  
  return(err);
  
}


int zlsio_Write32BitFloatFITSData  ( zFile f, float * fq, Counter ni, Counter nj, Counter nk ){
  int err = noErr;
  
  if (f) {
    
    Size chunkSize = (Size)ni*nj*nk*sizeof(float);
    Size remainder = FITSCARDSIZE - (chunkSize % FITSCARDSIZE);  // bytes to round up to next nearest 2880 byte boundary
    remainder = (remainder == FITSCARDSIZE)? 0: remainder;
    
#ifdef DATA_LITENDIAN_OS
    zls_ByteSwapFloatBlock ( fq, chunkSize);
#endif DATA_LITENDIAN_OS
    
    if ((err == noErr) && (fq != NULL)){
      err    =  zlsio_write( f, fq, chunkSize );
      if ( err != noErr ){
        zlsio_ErrPrintf( " ERROR: zlsio_Write32BitFloatFITSFile: IO zlsio_write Error: %d \n", err );
        return (err);
      } 
      
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_Write32BitFloatFITSFile: data not found\n" );
      err = memErr;
      return(err);
    }
    
    if ( remainder > 0) {
      // pad out to next FITSCARDSIZE bytes
      QPtr p = (QPtr) malloc(remainder);
      if ( p != NULL) {
        bzero((void *) p, remainder); // pad with 0s
        err    =  zlsio_write( f, p, remainder );
        free(p);
      }
      
    }
    
  } else { return (-1);}
  
  return( err ); 
}


int zlsio_Save32BitFloatZoneFITS  ( ZonePtr z, Integer varIndex, Integer level, Integer zone,
                                   IOSettingsPtr  io, Integer ioIndex ) {
  int err = noErr; 
  
  float * fq = NULL;
  QPtr q = NULL;
  
  if ( ( z != NULL) && ( io != NULL ) ){
    
    char     outFile[PATHLEN];
    char     vPref[PREFIXLEN];
    
    
    err = zlsio_GetVarIOVarPrefix ( vPref, io, ioIndex );
    if ( err != noErr ){
      zlsio_ErrPrintf( " IO ERROR: ioIndex out of range in zlsio_Save32BitFloatZoneFITS, %d Err = %d \n", ioIndex, err );
      err = ioErr;
      return (err);
    }
    
    err = zls_Allocate32BitFloatZoneData ( &fq ,  z ); // returns a array the same dims a z in 32 bit float cells
    
    if (err == noErr ) {
      
      if (z->zoneType == kImageZoneType) {
        //
        // data zone was merged, simplifies file name and get data calls, will make get data calls uniform later
        // name doesn't contain zone id, just the level and cycle number
        //
        sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                io->v[ioIndex].varCode, PATHSEP,
                vPref, io->outName, level, (int)io->ioState.ioDataCount, FITSFILEEXT );        
        
        err      = zls_GetZoneDataCopy        ( &q, z, kAnonVar);
        
      } else {
        
        if ( gLevels[level].nZones == 1){ // only 1 zone, save without zone id like a merged zone
          
          sprintf(outFile, "%s%s%s%s%s_%s_L%2.2dC%4.4d%s", io->outDir, PATHSEP, io->v[ioIndex].varCode, PATHSEP,
                  vPref, io->outName, level, (int)io->ioState.ioDataCount, FITSFILEEXT );        
          
        } else { // multiple zones in level, unmerged.  Make file into a zone id name inside a cycle count directory
          
          // first create the cycle count directory
          sprintf(outFile, "%s%s%s%sC%4.4d", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount );  
          
          err = zls_mkdir(outFile); // must work!
          if ( err != noErr) { return (fileErr);}
          
          // then the pathname of the file
          sprintf(outFile, "%s%s%s%sC%4.4d%s%s_%s_L%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP, 
                  io->v[ioIndex].varCode,  PATHSEP, (int)io->ioState.ioDataCount,  PATHSEP,
                  vPref, io->outName, level, zone, (int)io->ioState.ioDataCount, FITSFILEEXT );  
          
        }
        
#ifdef SYM1D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM1D
#ifdef SYM2D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM2D
#ifdef SYM3D          
        err      = zlsp_GetZoneWQuantityCopy  ( &q, z, varIndex);
#endif SYM3D
        
      }
      
      
      if ( ( q != NULL ) &&  ( fq != NULL ) ) {
        
        Counter ni ;
        Counter nj ;
        Counter nk ;
        
        GridCoords gc = z->zoneGrid;
        zls_GetGridCubeDims  ( &ni, &nj, &nk, &gc );
        
        // apply velocity slice limits if needed
        switch (varIndex) {
          case kWDeltaVx:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#if (NDIMS >1)        
          case kWDeltaVy:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif (NDIMS >1)        
#if (NDIMS >2)        
          case kWDeltaVz:
         {
           Real vMax =  io->v[ioIndex].scaleMax;
           Real vMin =  io->v[ioIndex].scaleMin;
           zlsp_VelocitySlice( q, q, vMin, vMax, gc);
         }
            break;
#endif  (NDIMS >2)   
          default:
            ;
        }
        
        
#if (ZLS_DEBUG > 0)       
        zlsio_Printf( " IO: Writing FITS 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
#endif     
        //zlsio_LogPrintf  ( " IO: Writing FITS 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        zlsio_FilesPrintf( " IO: Writing FITS 32bit float:\n       %s, %d x %dx %d cells.\n", outFile, ni, nj, nk);
        
        err = zls_Scale32BitFloatData  ( fq ,  q, kZLS_IOData,
                                        ni, nj, nk, 
                                        io, ioIndex );
        zFile fileAccess = NULL;
        
        fileAccess = zlsio_open(outFile, "wb");
        
        if (fileAccess) {
          
          err = zlsio_Write32BitFloatFITSHeader (fileAccess, z, io,  ioIndex );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save32BitFloatZoneFITS: zlsio_Write32BitFloatFITSHeader Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          err = zlsio_Write32BitFloatFITSData  ( fileAccess, fq, ni, nj, nk );
          if ( err != noErr ){
            zlsio_ErrPrintf( " IOERROR: zlsio_Save32BitFloatZoneFITS: zlsio_Write32BitFloatFITSData Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } 
          
          zlsio_close(fileAccess);
          
        }
        
        
      }
      
      err = zls_Free32BitFloatZoneData ( fq );
      if (q!=NULL)  free(q) ; q = NULL; 
      
      
    } else { err = memErr;  return (err);} // iq
    
    
  } else { err = zonErr;} // z io
  
  return ( err );
  
}


int zlsio_SaveLevel32BitFloatZoneFITS(Integer level, int step){ 
#pragma unused(step)  
  
  int        err=noErr;
  Integer          zone;
  Integer      varIndex;
  Integer       ioIndex;
  ZonePtr             z;
  
  IOSettingsPtr  io = NULL;
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
    
    ZonePtr mergedZone    = NULL;
    Integer bDoMergedData = 0;
    
#ifdef MPI
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        if ( gMPIrank == level) { // collect merged output on rank == level only
          zls_GetLevelImageZone( &mergedZone, level ); 
        }
        bDoMergedData = 1; // but all ranks participate in merger
      } // other levels do separate zones and imageZone stays NULL;
    }
#else    
    if ( io->outMergedZoneData == kZLS_IOOutputMergedZones) {
      if (gLevels[level].nZones >1) { // then merge output
        zls_GetLevelImageZone( &mergedZone, level );
        if ( mergedZone != NULL) bDoMergedData = 1;
      } // other levels do separate zones and imageZone stays NULL;
    }
#endif !MPI
    
    for ( ioIndex = 0; ioIndex< io->outNVars; ioIndex++){
      
      if ((io->v[ioIndex].ioTypes & kZLS_IOData) == kZLS_IOData) { // data io type flag
        
        varIndex = kAnonVar;
        
        err = zlsio_GetIOCodeVarIndex (&varIndex, io->v[ioIndex].varCode);
        if ( err != noErr ) {
          zlsio_ErrPrintf( " IO ERROR: varCode unknown in zlsio_SaveLevel32BitFloatZoneFITS, %d %s Err = %d \n"
                          , ioIndex, io->v[ioIndex].varCode, err );
          err = ioErr;
          return (err);
        }
        
        if ( varIndex != kNoVar ) { 
          
          
          if ( bDoMergedData == 1) {
#ifdef MPI
            err =  zls_MPI_GetLevelMergedZoneData( mergedZone, level, varIndex );
#else 
            err =  zls_GetLevelMergedZoneData( mergedZone, level, varIndex );
#endif !MPI
          } // imageZone may be null on non rank 0 processes, they send buffers to 0 which collects all
          
          if ((err == noErr) &&( bDoMergedData == 1 )) { // if the merge succeeded
            
#ifdef MPI  
            if ( gMPIrank == level) { // only rank == level can to write out merged image
#endif MPI
              
              err = zlsio_Save32BitFloatZoneFITS ( mergedZone, varIndex, level, 0, io, ioIndex );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO WARNING: zlsio_Save32BitFloatZoneFITS Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
              }
#ifdef MPI  
            }
#endif MPI
            
          } else { 
            
            // do normal separate output, either mergeZone failed or merger not requested
            
            
            //
            // loop over zones separately
            //
            
            
            for( zone = 0; zone <gLevels[level].nZones; zone++){
              
              err = zls_GetLevelZoneByIndex( &z, level, zone );
              if ( err != noErr ){
                zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
                err = ioErr;
                return (err);
              }
              
#ifdef MPI  
              if (z->zoneRankID == gMPIrank) {
#endif MPI
                
                err = zlsio_Save32BitFloatZoneFITS (  z, varIndex, level, zone, io, ioIndex );
                if ( err != noErr ){
                  zlsio_ErrPrintf( " IO WARNING: zlsio_Save32BitFloatZoneFITS Failed, IOIndex = %d , Err = %d \n", ioIndex, err );
                  zlsio_ErrPrintf( "           : will skip and try the next var. \n" );
                }
#ifdef MPI  
              }
#endif MPI
              
            } // zone loop
            
          }
          
        } // !kNoVar
        
      } // data io type flag
      
    } // ioIndex
    
    if ( level == 0 ) io->ioState.ioDataCount++;
    
  } else { //io
    zlsio_ErrPrintf( " IO ERROR: IOSettings not found zlsio_SaveLevel32BitFloatZoneFITS. Err %d\n", err );
    err = ioErr;
  }
  
  return (err);
}
