#include "zls.h"

void zls_Usage(void){
  
  zlsio_ErrPrintf(  "%s\n", "usage : fyris [-z zls.conf ] [-i init.conf ] [-o io.conf ] [-g grid.conf ] [-p phys.conf ]\n" );
  zlsio_ErrPrintf(  "%s\n", " fyris uses zls.conf by default, and finds other conf files there." );
  zlsio_ErrPrintf(  "%s\n", " user may optionally override any/all conf files including zls.conf" );
  zlsio_ErrPrintf(  "%s\n", " with -z -i -o -g and -p arguments." );
  exit(-1);
  
}

int zls_Finalize( void ){
  zlsio_EndLog           ( );
  zlsio_CloseLogFile     ( ); // log is kept open, so call this at end of sim
  zlsio_CloseUserLogFile ( ); // log is kept open, so call this at end of sim
  zlsio_CloseFilesFile   ( ); // log is kept open, so call this at end of sim
  
#ifdef MPI
//  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
#endif MPI

  return ( noErr ); // too late for any errors now !
}

int zls_Initialize(int argc, char *argv[]){
  
  int err = noErr;

#ifdef MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &gMPIrank);
  MPI_Comm_size(MPI_COMM_WORLD, &gMPIsize);
  gMPIBlocking = 1; // allocate zones to ranks 1 at a time by default;
#endif MPI

  // command line args
  
  int c;
  // init conf vars
  extern char* optarg;
  struct ccl_t config;
  const char *str1;
  
  // start the global clock
  
  zlst_InitGlobalTiming(&gZLS_GlobalCPUTimer);
  
  char zlsLogName [PATHLEN];
  sprintf( zlsLogName,    "%s%s", ZLSLOGFILE   , LOGEXT );
  
  char zlsUserLogName [PATHLEN];
  sprintf( zlsUserLogName,    "%s%s", ZLSUSERLOGFILE   , LOGEXT );
  
  char restartConfName[PATHLEN];
  sprintf( restartConfName, "%s%s", RESTARTFILE, CONFEXT );
  
  char restartSaveName[PATHLEN];
  sprintf( restartSaveName, "%s"  , RESTARTFILE );
  
  char restartDirName[PATHLEN];
  sprintf( restartDirName, "%s"  , RESTARTDIR );
  
  char gridConfName[PATHLEN];
  sprintf( gridConfName,  "%s%s", GRIDCONFFILE, CONFEXT );
  
  char physConfName[PATHLEN];
  sprintf( physConfName,  "%s%s", PHYSCONFFILE, CONFEXT );
  char ioConfName  [PATHLEN];
  sprintf( ioConfName,    "%s%s", IOCONFFILE  , CONFEXT );
  char initConfName[PATHLEN];
  sprintf( initConfName,  "%s%s", INITCONFFILE, CONFEXT );
  
  char ppmConfName[PATHLEN];
  sprintf( ppmConfName,  "%s%s", PPMCONFFILE, CONFEXT );
  
  // default conf file name, the rest follow according to intial file and or comm line args
  
  char zlsConfName [PATHLEN];
  sprintf( zlsConfName,  "%s%s", ZLSCONFFILE, CONFEXT );
  
  int restartDirConf    = 0;
  int restartConf       = 0;
  int restartSaveChange = 0;
  int gridConf = 0;
  int initConf = 0;
  int physConf = 0;
  int ioConf   = 0;
  int zlsConf  = 0;
  
  // get comm args second to overrride conf file specifications if any
  
  while ((c = getopt(argc, argv, "z:p:g:i:r:o:h")) != -1){
#ifdef DEBUG
    zlsio_Printf( " Arg %d %c\n", optind ,c );
#endif
    switch (c) {
      case 'p':// physics init config file
        strcpy(physConfName, optarg);
        physConf = 1;
        zlsio_Printf( " Physics Config file set = %s \n",physConfName);
        break;
      case 'g':// grid config file
        strcpy(gridConfName, optarg);
        gridConf = 1;
        zlsio_Printf( " Grid Config file set = %s \n",gridConfName);
        break;
      case 'o':// io config file
        strcpy(ioConfName, optarg);
        ioConf = 1;
        zlsio_Printf( " I/O Config file set = %s \n",ioConfName);
        break;
      case 'i':// user init config file
        strcpy(initConfName, optarg);
        initConf = 1;
        zlsio_Printf( " User Init Config file set = %s \n",initConfName);
        break;
      case 'r':// restart config file - can't change SaveBaseName is comm line arg used instead
        strcpy(restartConfName, optarg);
        restartConf = 1;
        zlsio_Printf( " Restart Config file set = %s \n",restartConfName);
        break;
      case 'z':// redo config from alternative to zls.conf
        strcpy(zlsConfName, optarg);
        zlsConf = 1;
        zlsio_Printf( " Main zls Config file set = %s \n",zlsConfName);
        break;
      case 'h':
      case '?':
        zls_Usage();
        /*NOTREACHED*/
    }
  }
  
  //
  // read conf files
  //
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, zlsConfName )==noErr) {
    
    if ((str1= ccl_get( &config,"PPM_File"))){
      strcpy(  ppmConfName,  str1);
    }
    
    if ( gridConf == 0 ) {
      if ((str1= ccl_get( &config,"Grid_File"))){
        strcpy(  gridConfName,  str1);
        gridConf = 1;
      }
    }
    
    if ( physConf == 0 ) {
      if ((str1= ccl_get( &config,"Physics_File"))){
        strcpy(  physConfName,  str1);
        physConf = 1;
      }
    }
    
    if ( ioConf == 0 ) {
      if ((str1= ccl_get( &config,"IO_File"))){
        strcpy(  ioConfName,  str1);
        ioConf = 1;
      }
    }
    
    if ( initConf == 0 ) {
      if ((str1= ccl_get( &config,"Init_File"))){
        strcpy(  initConfName,  str1);
        initConf = 1;
      }
    }
    
    
    if ( restartConf == 0 ) {
      
      if ((str1= ccl_get( &config,"Restart_Directory"))){
        //
        // Only happens if the key: Restart_Dir =  is in zls.conf
        // and has not been commented out
        //
        strcpy(  restartDirName,  str1);
        restartDirConf = 1;
      }
      
      
      if ((str1= ccl_get( &config,"Restart_File"))){
        //
        // Only happens if the key: Restart_File =  is in zls.conf
        // and has not been commented out
        //
        strcpy(  restartConfName,  str1);
        restartConf = 1;
      }
      
      if ((str1= ccl_get( &config,"Restart_SaveBaseName"))){
        //
        // Only happens if the key: Restart_SaveBaseName = is in zls.conf
        // and has not been commented out
        //
        strcpy(  restartSaveName,  str1);
        restartSaveChange = 1;
      }
    }
    
    ccl_release(&config);
    
  }
  
  //========================================
  //
  // General ZLS Initialisation - based on  config files, grid, physics, io
  //
  //========================================
  
  //
  //
  // Note: log file doesn't exist yet, errors go just to stderr in the meantime
  //
  //
  
  err = zlsio_SetUpPPMFromConfigFile  ( ppmConfName );
  if (err != noErr) {
    zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file.\n", ppmConfName);
    return (err);
  }
  
  if ( physConf != 0) {
    err = zlsio_InitFiducialScalesConfigFile( physConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading Fiducial Scales from %s config file.\n", physConfName);
      return (err);
    }
  } else {
    zlsio_ErrPrintf( " INIT ERROR:: must specify a physics init config file.\n");
    return(fileErr);
  }
  
  if ( gridConf != 0) {
    err = zlsio_SetUpLevelsFromConfigFile    ( gridConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file.\n", gridConfName);
      return (err);
    }
    
    err = zlsio_SetUpZonesFromConfigFile  ( gridConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file.\n", gridConfName);
      return (err);
    }
  } else {
    zlsio_ErrPrintf( " INIT ERROR:: must specify a grid config file.\n");
    return(-1);
  }
  
  if ( physConf != 0) {
    err = zlsio_SetUpPhysicsFromConfigFile( physConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file.\n", physConfName);
      return (err);
    }
  } else {
    zlsio_ErrPrintf( " INIT ERROR:: must specify a physics init config file.\n");
    return(fileErr);
  }
  
  if ( ioConf != 0) {
    err = zlsio_SetUpIOFromConfigFile    ( ioConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file.\n", ioConfName);
      return (err);
    }
  } else {
    zlsio_ErrPrintf( " INIT ERROR: must specify an I/O settings config file.\n");
    return(ioErr);
  }
  
  //========================================
  //
  // User Initialisation
  //
  //========================================
  
  if ( initConf != 0) {
    
    err = zlsio_SetUpSimLimitsFromConfigFile ( initConfName );
    if (err != noErr) {
      zlsio_ErrPrintf( " INIT ERROR: Error Reading %s config file setting sim limits.\n", initConfName);
      return (err);
    }
    
  } else {
    
    zlsio_ErrPrintf( " INIT ERROR: must specify a problem init config file.\n");
    return(fileErr);
    
  }
  
  IOSettingsPtr  io = NULL;
  zls_GetLevelIOPrefs( &io, 0 );
  
  if ( io != NULL ){
    
    if ( restartDirConf == 0 ) {
      zlsio_SetIOSettingsRestartDir( io, io->outDir );
    } else {
      zlsio_SetIOSettingsRestartDir( io, restartDirName );
    }
    
    //
    // create output directories after being set in io.conf and possibly
    // overridden in init.conf, before restarting
    //
    err =  zls_mkdir(io->outDir);
    if ( err != noErr ) {
      zlsio_ErrPrintf( " INIT ERROR: Error creating %s directory.\n", io->outDir);
      return (err);
    }
    // create output var dirs
    
    Integer n;
    char varDir[PATHLEN];
    Integer varIndex;
    
    if ( io->outNVars > 0) {
      for ( n = 0; n < io->outNVars; n++){
        zlsio_GetIOCodeVarIndex (&varIndex,  io->v[n].varCode);
        if ( varIndex != kNoVar ){
          
          sprintf(varDir,"%s%s%s",io->outDir, PATHSEP, io->v[n].varCode);
          err =  zls_mkdir(varDir);
          if ( err != noErr )  {
            zlsio_ErrPrintf( " INIT ERROR: Error creating %s directory.\n", varDir);
            return (err);
          }
          
        }
      }
    }
    
    // create user output dir
    
    sprintf(varDir,"%s%s%s",io->outDir, PATHSEP, USERDIR);
    err =  zls_mkdir(varDir);
    if ( err != noErr )  {
      zlsio_ErrPrintf( " INIT ERROR: Error creating %s directory.\n", varDir);
      return (err);
    }
    
    // create user monitor variable dir
    
    if ( io->monitorVar > -1) {
      sprintf(varDir,"%s",io->monDir);
      err =  zls_mkdir(varDir);
      if ( err != noErr )  {
        zlsio_ErrPrintf( " INIT ERROR: Error creating %s directory.\n", varDir);
        return (err);
      }
    }
    //
    // create output logfile after being set in io.conf and possibly
    // overridden in init.conf, before restarting
    //
    err = zlsio_CreateLogFile ( kZLS_ContinueLogFile );
    if ( err != noErr )  {
      zlsio_ErrPrintf( " INIT ERROR: Error creating log file.\n");
      return (err);
    }
    err = zlsio_CreateUserLogFile ( kZLS_ContinueLogFile );
    if ( err != noErr )  {
      zlsio_ErrPrintf( " INIT ERROR: Error creating user log file.\n");
      return (err);
    }
    err = zlsio_CreateFilesFile ( kZLS_ContinueFilesFile );
    if ( err != noErr )  {
      zlsio_ErrPrintf( " INIT ERROR: Error creating IO log file.\n");
      return (err);
    }
    
    zlsio_StartLog ( );
    
  } else { err = memErr; return(err); }
  
  // Now log file exists we can do user setup and record print statements
  
  err = zlsp_SetUpProblemFromConfigFile ( initConfName );
  if ( err != noErr )  {
    zlsio_ErrPrintf( " INIT ERROR: Error in zlsp_SetUpProblemFromConfigFile: %d \n", err);
    return (err);
  }
  
  //========================================
  //
  // Restarting  Initialisation, optional, but must come after all other initialisation
  //
  //========================================
  
  if ( restartSaveChange != 0 ) { 
    // Only happens if the key: Restart_SaveBaseName = is in zls.conf
    // and has not been commented out
    // not common, but used to change the default name for future save states
    // otherwise "Restart" is used as usual.
    zlsio_SetIOSettingsRestartName (  io, restartSaveName );
  }
  
  // record startup in log file, has to be called late as log file doesn't exist for 
  // much of the startup sequence.
  
  zlsio_LogSetup ( zlsConfName, ppmConfName, gridConfName, physConfName, initConfName, ioConfName, restartConfName );
  
#ifdef MPI  
  if (gMPIrank == 0) {
#endif MPI
    zlsio_CreateInfoFile ( kZLS_CreateNewInfoFile ); // records the simulations config files
    zlsio_RecordConfiguration( zlsConfName, ppmConfName, gridConfName, physConfName, initConfName, ioConfName, restartConfName );
    zlsio_CloseInfoFile( ); 
#ifdef MPI  
  }
#endif MPI
  
  if ( restartConf != 0) {
    
    // Only happens if the key: Restart_File = is in zls.conf
    // and has not been commented out
    
    sprintf( zlsConfName,"%s%s%s", io->restartDir, PATHSEP, restartConfName );
    strcpy ( restartConfName, zlsConfName );
    
    err = zlsio_RestoreState ( restartConfName );
    
    if (err != noErr) { 
      zlsio_ErrPrintf( "%s\n",  DIVIDERLINE );
#ifdef HYDRO
      zlsio_ErrPrintf( " %s PPM-LR Hydro FAILED Restart From %s. \n", FYRIS, restartConfName );
#endif HYDRO
      zlsio_ErrPrintf( "%s\n\n",  DIVIDERLINE );
      return (err);
    }
    
  } else {
    zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef HYDRO
    zlsio_Printf( " %s PPM-LR Hydro Initialization Completed. \n", FYRIS );
#endif HYDRO
    zlsio_Printf( "%s\n\n",  DIVIDERLINE );
  }
  
#ifdef MPI
  MPI_Barrier(MPI_COMM_WORLD); // wait for everyone before proceeding to rest of code
#endif MPI
  
  return(err);
} 
