#include "zls.h"
#ifdef PTHREADS
pthread_mutex_t gIOMutex    = PTHREAD_MUTEX_INITIALIZER; // mutex for gIOCount vars and general File IO
#endif PTHREADS

int zlsio_InitVarIORec(VarIORecPtr vio){
  
  Integer err = noErr;
  
  vio->varCode      = (char *)malloc(CODELEN+1); // up to 8 letter code for specifying output var + NULL, initially just straight vars, padded to 9 bytes
  strncpy(vio->varCode,"none", CODELEN);
  vio->varPrefix    = NULL; // var string for building file names
  
  vio->varTIFFCTab  = NULL;       // for 8 bit TIFFs, NULL if not used
  
  vio->ioTypes      = kZLS_IOImage+kZLS_IOData;  // default images and datafiles (possibly with different timing)
  vio->imageSlice   = kSliceX;       // plane if sliced, ie 2D slice of 3D vol, or just a 2D plane as is.
  vio->imageIndex   = 0;       // indices along perp axis where sliced/rowed
  vio->imageIndex2  = 0;       // 2nd index along perp axis where sliced/rowed if needed
  vio->imagePlotColor = 128;   // colour index in CTAB for 1D row plots
  
  vio->imageScalingType = kZLS_IOAutoScale; // kZLS_IOAutoScale, kZLS_IOManualScale
  vio->dataScalingType  = kZLS_IOAutoScale; // kZLS_IOAutoScale, kZLS_IOManualScale
  vio->transformType    = kZLS_IOLinearTransform; // kZLS_IOLinearTransform, kZLS_IOLog10Transform, kZLS_IOExp10Transform, kZLS_IOPowTransform
  
  vio->dataMin       = 0.0; // Min if manualScale
  vio->dataMax       = 1.0; // Max if manualScale
  
  vio->scaleMin       = 0.0;  // Min if manualScale
  vio->scaleMax       = 1.0;  // Max if manualScale
  vio->scaleInvRange  = 1.0;  // 1.0/(Max-Min) if manualScale
  
  vio->transPow       = 1.0;  // index if kZLS_IOPowTransform
  vio->transThreshold = 0.0;  // threshold value
  vio->transNULL      = 0.0;  // value for invalid cells
  
  return(err);
  
}


int zlsio_SetIOSettingsRestartName (  IOSettingsPtr io, char * restartSaveName ){
  int err = noErr;
  if (io != NULL) {
    strcpy( io->restartName, restartSaveName);
  } else { err = ioErr;}
  return (err);
}

int zlsio_SetIOSettingsRestartDir (  IOSettingsPtr io, char * restartDirName ){
  int err = noErr;
  if (io != NULL) {
    strcpy( io->restartDir, restartDirName);
  } else { err = ioErr;}
  return (err);
}


int zlsio_AllocateIOImages ( IOSettingsPtr io, int nx, int ny){
  
  int err = noErr;
  
  if (io->image8!=NULL)  free(io->image8) ; io->image8 = NULL; 
  if (io->image16!=NULL) free(io->image16); io->image16 = NULL;
  if (io->imageReal!=NULL) free(io->imageReal); io->imageReal = NULL;
  
  io->image8 = (UInt8 *) malloc(nx*ny);
  if ( io->image8==NULL) { 
    return (memErr);
  }
  
  io->image16 = (UInt16 *) malloc(nx*ny*sizeof(UInt16));
  if ( io->image16==NULL) { 
    return (memErr);
  }
  
  // imageReal is not allocated here, it is set by calls to get data functions;
  
  return (err);
  
}



int zlsio_InitIOSettings( IOSettingsPtr io, Integer nVars, Integer monitorVar, Integer monitorLevel ){
  int err = noErr;
  
  Integer i;
  
  if (io != NULL) {
    
    strcpy( io->outRunName , FYRIS_PROBLEMNAME); // default to name in zls_problem.h
    
    strcpy( io->outDir,  OUTPUTDIR);
    strcpy( io->monDir,  MONITORDIR);
    strcpy( io->restartDir, RESTARTDIR);
    
    strcpy( io->fileName, DATAFILE);
    strcpy( io->outName , OUTFILE);
    strcpy( io->outMonitorName,  MONFILE);
    strcpy( io->restartName, RESTARTFILE);
    
    strcpy( io->outLogName,  ZLSLOGFILE);
    io->outLogFILE = NULL;
    
    strcpy( io->outUserLogName,  ZLSUSERLOGFILE);
    io->outUserLogFILE = NULL;
    
    strcpy( io->outInfoName,  ZLSINFOFILE);
    io->outInfoFILE = NULL;
    
    strcpy( io->outFilesName,  ZLSFILESLOG);
    io->outFilesFILE = NULL;
    
    strcpy( io->outCTABFile, OUTCTAB);
    strcpy( io->monitorCTABFile, OUTCTAB);
    
    io->outDataFormat       = kZLS_ZLSData;  
    io->outDataBitsPerCell  = kZLS_Float32Data;
    io->outMergedZoneData   = kZLS_IOOutputSeparateZones;
    
    io->outFormat           = kTIFFOutput;
    io->outBitsPerPix       = k8bitOutput;
    io->outMergedZoneFrames = kZLS_IOOutputMergedZones;
    
    io->ioState.ioFrameCount     = 0; // output step count to keep all Image IO in sync.x
    io->ioState.ioUserCount      = 0; // output step count to keep all User IO in sync
    io->ioState.ioDataCount      = 0; // output step count to keep all 16bit data IO in sync
    io->ioState.ioRealDataCount  = 0; // output step count to keep all flaoting point (other than restarts) data IO in sync
    io->ioState.ioSaveCount      = 0; // output step count to keep all restart IO in sync
    
    io->outNVars     = nVars;
    io->monitorVar   = monitorVar;
    
    io->monitorLevel = monitorLevel;
    if ( nVars > 0) {
      io->v        = (VarIORecPtr) malloc( nVars*sizeof(VarIORec) );
    } else {
      io->v = NULL;
    }
    
    if (( nVars > 0)&&( io->v == NULL )) {
      err = memErr;
      zlsio_ErrPrintf( " ERROR: IOSettings variable settings array failed to allocate. \n" );
      return(err);
    }
    
    for (i = 0; i< nVars; i++) {
      zlsio_InitVarIORec(&(io->v[i]));
      if ( i == monitorVar) { // make sure the monitor var has image output format
        io->v[i].ioTypes  |= kZLS_IOImage;
      }
    }
    
    io->image8  = NULL;
    io->image16 = NULL;
    io->imageReal = NULL;  
    
  }
  
  return (err);
}

int zlsio_NewIOSettings(IOSettingsPtr * io){
  int err = noErr;
  
  *io = (IOSettingsPtr)malloc(sizeof(IOSettings));
  
  if (*io == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: IOSettings allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsio_DisposeIOSettings(IOSettingsPtr io){
  int err = nyiErr;
  
  if ( io != NULL ) {
    if (io->image8!=NULL)  free(io->image8) ; io->image8 = NULL; 
    if (io->image16!=NULL) free(io->image16); io->image16 = NULL;
    free(io);
  }
  
  return (err);
}

int zlsio_GetVarFiducialValue( Real *f0, Integer varIndex) {
  int err = noErr;
  
  Real f = 1.0;
  
  switch (varIndex){
      
#ifdef MHD
    case kBx:
    case kBy:
    case kBz:
      f = f_B0;
      break;
    case kBPres:
    case kBE:
      f = f_E0;
      break;
#endif MHD
#ifdef PRIMITIVE_QUAS
    case kWPres:
      f = f_P0;
      break;
    case kWE:
    case kWTotE:
    case kWKE:
    case kWEE:
    case kWTotEE:
    case kWKEE:
      f = f_E0;
      break;
    case kWEL:
    case kWTotEL:
    case kWKEL:
      f = f_EL0;
      break;
    case kSymDens:
    case kWDens:
      f = f_d0;
      break;
    case kWVx:
    case kWVy:
    case kWVz:
    case kWVelm:
      f = f_v0;
      break;
      
    case kWM:
    case kWMx:
    case kWMy:
    case kWMz:
      f = f_d0*f_v0;
      break;
      
#endif PRIMITIVE_QUAS
#ifdef CONSERVED_QUAS
    case kSymDens:
    case kUDens:
      f = f_d0;
      break;
    case kUEneg:
      f = f_E0;
      break;
    case kUMx:
    case kUMy:
    case kUMz:
      f = f_d0*f_v0;
      break;
#endif CONSERVED_QUAS
      
    case kWVelmx:
    case kWVelmy:
    case kWVelmz:
    case kWVelmxy:
    case kWVelmyz:
    case kWVelmzx:
    case kWDeltaVx:
    case kWDeltaVy:
    case kWDeltaVz:
      f = f_v0;
      break;
      
#ifdef COOLING              
    case kWCoolAt:
    case kWCoolCO:
    case kWCoolH2:
    case kWCool:
      f = f_cool0;
      break;
    case kWHeat:
    case kWHTgr:
    case kWEmGr:
    case kWHCRX:
    case kWHIPE:
      f = f_cool0;
      break;
    case kWLambAt:
    case kWLambCO:
    case kWLambH2:
    case kWLamb:
      f = f_lam0;
      break;
#endif COOLING              
    case kWdHI:
    case kWdHII:
    case kWdH2:
      f = f_d0;
      break;
      
    case kWCs:
    case kWAs:
      f = f_v0;
#ifdef GRAVITY              
    case kGPhi:
      f = k_g;
      break;
    case kdPhi:
    case kdPhiX:
    case kdPhiY:
      f = f_a0;
      break;
#endif GRAVITY              
    default:
      f = 1.0;
  } // switch derived vars
  
  *f0 = f;
  
  return (err);
}

int zlsio_GetIOCodeVarIndex (Integer *varIndex, char * varCode){
  int err = noErr;
  
  Integer vindex = kNoVar;
  
  if (strcmp(varCode, "Eint")==0) {
    // lagrangian internal energy: P/((g-1) d)
    // Eulerian internal energy  : P/((g-1))
    vindex = kWE;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Eint_E")==0) {
    // Eulerian internal energy  : P/((g-1))
    vindex = kWEE;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Eint_L")==0) {
    // lagrangian internal energy: P/((g-1) d)
    vindex = kWEL;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Etot")==0) {
    // Lagrangian total energy : P/((g-1)d) + 0.5*v^2 [+ 0.5*B^2/d]
    // Eulerian total energy   : P/((g-1)) + 0.5d*v^2 [+ 0.5*B^2]
    vindex = kWTotE;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Etot_E")==0) {
    // Lagrangian total energy : P/((g-1)d) + 0.5*v^2 [+ 0.5*B^2/d]
    vindex = kWTotEE;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Etot_L")==0) {
    // Eulerian total energy   : P/((g-1)) + 0.5d*v^2 [+ 0.5*B^2]
    vindex = kWTotEL;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "entr")==0) {
    vindex = kWEntr;
    *varIndex = vindex;  return(err);
  } //  A, P/d^g : isoentropic quantity
  
  if (strcmp(varCode, "enth")==0) {
    // P/((g-1) d) + P/d, Lagrangian specific enthalpy
    // P/((g-1)) + P, Eulerian specific enthalpy
    vindex = kWEnth;
    *varIndex = vindex;  return(err);
  } // h, 
  
  if (strcmp(varCode, "enth_E")==0) {
    // P/((g-1)) + P, Eulerian specific enthalpy
    vindex = kWEnthE;
    *varIndex = vindex;  return(err);
  } // h, 
  
  if (strcmp(varCode, "enth_L")==0) {
    // P/((g-1) d) + P/d, Lagrangian specific enthalpy
    vindex = kWEnthL;
    *varIndex = vindex;  return(err);
  } // h, 
  
  if (strcmp(varCode, "Mome")==0) {
    vindex = kWM;
    *varIndex = vindex;  return(err);
  } // d* sqrt( vx^2 + vy^2 + vz^2)
  
  if (strcmp(varCode, "momx")==0) {
    vindex = kWMx;
    *varIndex = vindex;  return(err);
  } // d* velx
  
  if (strcmp(varCode, "momy")==0) {
    vindex = kWMy;
    *varIndex = vindex;  return(err);
  } // d* vely
  
  if (strcmp(varCode, "momz")==0) {
    vindex = kWMz;
    *varIndex = vindex;  return(err);
  } // d* velz
  
    
#if (NDIMS > 1)
  if (strcmp(varCode, "vort")==0) {
    vindex = kVortex2D;
    *varIndex = vindex;  return(err);
  } // dens
#endif (NDIMS > 1)
  
  if (strcmp(varCode, "schl")==0) {
    vindex = kSclieren;
    *varIndex = vindex;  return(err);
  } // shlieren image of dens, dd/dx
  
#if (NDIMS > 1)
  if (strcmp(varCode, "shdw")==0) {
    vindex = kShadow;
    *varIndex = vindex;  return(err);
  } // shadogram of density
#endif (NDIMS > 1)
  
  if (strcmp(varCode, "dens")==0) {
    vindex = kWDens;
    *varIndex = vindex;  return(err);
  } // dens
  
#if (NDIMS >1)        
  if (strcmp(varCode, "symd")==0) {
    vindex = kSymDens;
    *varIndex = vindex;  return(err);
  } // dens
#endif (NDIMS >1)        
  
  if (strcmp(varCode, "pres")==0) {
    vindex = kWPres;
    *varIndex = vindex;  return(err);
  } // pres
  
  if (strcmp(varCode, "velx")==0) {
    vindex = kWVx;
    *varIndex = vindex;  return(err);
  } // velx
  if (strcmp(varCode, "vlmx")==0) {
    vindex = kWVelmx;
    *varIndex = vindex;  return(err);
  } // velx abs
  if (strcmp(varCode, "dvx_")==0) {
    vindex = kWDeltaVx;
    *varIndex = vindex;  return(err);
  } // velx in an interval, 0.0 otherwise
  
#if (NDIMS > 1 )
  if (strcmp(varCode, "vely")==0) {
    vindex = kWVy;
    *varIndex = vindex;  return(err);
  } // vely
  if (strcmp(varCode, "vlmy")==0) {
    vindex = kWVelmy;
    *varIndex = vindex;  return(err);
  } // vely
  if (strcmp(varCode, "dvy_")==0) {
    vindex = kWDeltaVy;
    *varIndex = vindex;  return(err);
  } // vely in an interval, 0.0 otherwise
#endif
  
#if (NDIMS > 2 )
  if (strcmp(varCode, "velz")==0) {
    vindex = kWVz;
    *varIndex = vindex;  return(err);
  } // velz
  if (strcmp(varCode, "vmxy")==0) {
    vindex = kWVelmxy;
    *varIndex = vindex;  return(err);
  } // velxy
  if (strcmp(varCode, "vmyz")==0) {
    vindex = kWVelmyz;
    *varIndex = vindex;  return(err);
  } // velyz
  if (strcmp(varCode, "vmzx")==0) {
    vindex = kWVelmzx;
    *varIndex = vindex;  return(err);
  } // velzx
  if (strcmp(varCode, "vlmz")==0) {
    vindex = kWVelmz;
    *varIndex = vindex;  return(err);
  } // vely
  if (strcmp(varCode, "dvz_")==0) {
    vindex = kWDeltaVz;
    *varIndex = vindex;  return(err);
  } // velz in an interval, 0.0 otherwise
#endif
  
  if (strcmp(varCode, "DivV")==0) {
    vindex = kWDivV;
    *varIndex = vindex;  return(err);
  } 
  if (strcmp(varCode, "divx")==0) {
    vindex = kWDvDx;
    *varIndex = vindex;  return(err);
  } 
#if (NDIMS > 1 )
  if (strcmp(varCode, "divy")==0) {
    vindex = kWDvDy;
    *varIndex = vindex;  return(err);
  } 
#endif
#if (NDIMS > 2 )
  if (strcmp(varCode, "divz")==0) {
    vindex = kWDvDz;
    *varIndex = vindex;  return(err);
  } 
#endif
  
  if (strcmp(varCode, "temp")==0) {
    vindex = kWTemp;
    *varIndex = vindex;  return(err);
  } // temp
  
  if (strcmp(varCode, "velm")==0) {
    vindex = kWVelm;
    *varIndex = vindex;  return(err);
  } // velm
  
  if (strcmp(varCode, "spd.")==0) {
    vindex = kWSpeed;
    *varIndex = vindex;  return(err);
  } // spd.
  
  if (strcmp(varCode, "cspd")==0) {
    vindex = kWCs;
    *varIndex = vindex;  return(err);
  } // adiabatic sound speed
  
  if (strcmp(varCode, "aspd")==0) {
    vindex = kWAs;
    *varIndex = vindex;  return(err);
  } // isothermal sound speed
  
  if (strcmp(varCode, "mach")==0) {
    vindex = kWMach;
    *varIndex = vindex;  return(err);
  } // mach
  
#ifdef COOLING
  if (strcmp(varCode, "clCO")==0) {
    vindex = kWCoolCO;
    *varIndex = vindex;  return(err);
  } // C0 cool
  
  if (strcmp(varCode, "clH2")==0) {
    vindex = kWCoolH2;
    *varIndex = vindex;  return(err);
  } // H2 cool
  
  if (strcmp(varCode, "clAt")==0) {
    vindex = kWCoolAt;
    *varIndex = vindex;  return(err);
  } // atomic cool
  
  if (strcmp(varCode, "cool")==0) {
    vindex = kWCool;
    *varIndex = vindex;  return(err);
  } // total cool
  
  if (strcmp(varCode, "lamb")==0) {
    vindex = kWLamb;
    *varIndex = vindex;  return(err);
  } // lamb
  
  if (strcmp(varCode, "heat")==0) {
    vindex = kWHeat;
    *varIndex = vindex;  return(err);
  } // heat
      
#ifdef VARIABLE_EOS  
  
  if (strcmp(varCode, "mumu")==0) {
    vindex = kWMu;
    *varIndex = vindex;  return(err);
  } // EOS mol weight
  
  if (strcmp(varCode, "xhi_")==0) {
    vindex = kWXHI;
    *varIndex = vindex;  return(err);
  } // EOS neutral H fraction
  if (strcmp(varCode, "xhii")==0) {
    vindex = kWXHII;
    *varIndex = vindex;  return(err);
  } // EOS Ionised H fraction
  if (strcmp(varCode, "xh2_")==0) {
    vindex = kWXH2;
    *varIndex = vindex;  return(err);
  } // EOS mol H2 fraction
  
  if (strcmp(varCode, "dHi_")==0) {
    vindex = kWdHI;
    *varIndex = vindex;  return(err);
  } // EOS neutral H fraction * density, column when summed
  if (strcmp(varCode, "dHii")==0) {
    vindex = kWdHII;
    *varIndex = vindex;  return(err);
  } // EOS Ionised H fraction * density, column when summed
  if (strcmp(varCode, "dH2_")==0) {
    vindex = kWdH2;
    *varIndex = vindex;  return(err);
  } // EOS mol H2 fraction * density, column when summed
  
  
  if (strcmp(varCode, "nhi_")==0) {
    vindex = kWnHI;
    *varIndex = vindex;  return(err);
  } // EOS neutral H fraction * nHtot, column when summed
  if (strcmp(varCode, "nhii")==0) {
    vindex = kWnHII;
    *varIndex = vindex;  return(err);
  } // EOS Ionised H fraction * nHtot, column when summed
  if (strcmp(varCode, "nh2_")==0) {
    vindex = kWnH2;
    *varIndex = vindex;  return(err);
  } // EOS mol H2 fraction * nHtot, column when summed
  
  if (strcmp(varCode, "rcom")==0) {
    vindex = kRecomb;
    *varIndex = vindex;  return(err);
  } 
  
  if (strcmp(varCode, "Rayl")==0) {
    vindex = kRayleigh;
    *varIndex = vindex;  return(err);
  } 
  
  
  if (strcmp(varCode, "gama")==0) {
    vindex = kWGamma;
    *varIndex = vindex;  return(err);
  } // EOS gamma
  
#endif VARIABLE_EOS  
  
  if (strcmp(varCode, "lmCO")==0) {
    vindex = kWLambCO;
    *varIndex = vindex;  return(err);
  } // lamb
  
  if (strcmp(varCode, "lmH2")==0) {
    vindex = kWLambH2;
    *varIndex = vindex;  return(err);
  } // lamb
#endif COOLING
  
  
  if (strcmp(varCode, "gphi")==0) {
    vindex = kGPhi;
    *varIndex = vindex;  return(err);
  } // grav potential
  
  if (strcmp(varCode, "dphi")==0) {
    vindex = kdPhi;
    *varIndex = vindex;  return(err);
  } // grav acceleration
  
  if (strcmp(varCode, "dphx")==0) {
    vindex = kdPhiX;
    *varIndex = vindex;  return(err);
  } // grav acceleration x comp
  
  if (strcmp(varCode, "dphy")==0) {
    vindex = kdPhiY;
    *varIndex = vindex;  return(err);
  } // grav acceleration y comp
  
  if (strcmp(varCode, "dphz")==0) {
    vindex = kdPhiZ;
    *varIndex = vindex;  return(err);
  } // grav acceleration z comp
  
#if (NTRACERS > 0)    
  Integer i;
  char keyStr[16];
  
  if (strcmp(varCode, "trac")==0) {
    vindex = kTracer0;
    *varIndex = vindex;  return(err);
  } // trac stands for tracer 0 by default
  
  for ( i = 0; i < kNTracers; i ++ ){
    sprintf(keyStr,"trc%1.1d", i);
    if (strcmp(varCode, keyStr)==0) {
      vindex = kTracer0+i;
      *varIndex = vindex;  return(err);
    } 
  }
  
  if (strcmp(varCode, "scal")==0) {
    vindex = kScalar0;
    *varIndex = vindex;  return(err);
  } // scal stands for tracer0/d by default
  
  for ( i = 0; i < kNTracers; i ++ ){
    sprintf(keyStr,"scl%1.1d", i);
    if (strcmp(varCode, keyStr)==0) {
      vindex = kScalar0+i;
      *varIndex = vindex;  return(err);
    } 
  }
  
  if (strcmp(varCode, "rado")==0) {
    vindex = kRadioJnu0;
    *varIndex = vindex;  return(err);
  } // scal stands for tracer0/d by default
  
  for ( i = 0; i < 3; i ++ ){
    sprintf(keyStr,"rad%1.1d", i);
    if (strcmp(varCode, keyStr)==0) {
      vindex = kRadioJnu0+i;
      *varIndex = vindex;  return(err);
    } 
  }
  
#endif
  
  *varIndex = vindex;
  
  return(err);  
}


int zlsio_GetVarIOVarPrefix    ( char * vPref, IOSettingsPtr io,  Integer varIndex ){
  // caller must make sure vPref has space for result!
  int err = noErr;
  
  if (io != NULL) {
    if (( varIndex >= 0 ) && (varIndex < io->outNVars) ) {
      
      VarIORec vio = io->v[varIndex];
      strcpy( vPref , vio.varPrefix);
      
    }  else { err = dimErr; }
  } else { err = ioErr; }
  
  return(err);  
  
}


int zlsio_WriteInitialOutput(void) {
  
  int err = noErr;    // write out initial conditions before sim starts
  
  Integer level;
  
  for (level = (gNLevels-1); level >=0; level--){
    
#if (ZLS_DEBUG>0)    
    zlsio_Printf( " Save Initial State:\n" );
#endif
    
    if ((gLevels[level].levelClock.userTime <= gLevels[level].levelClock.endTime)){
#if (ZLS_DEBUG>0)    
      zlsio_Printf( "  Initial User Ouput, level: %d:\n", level);
#endif
      err = zlsio_SaveLevelUserIO        ( level , zlst_GlobalCycles);
      if (err != noErr ) return ( err );
    }
    
    if ((gLevels[level].levelClock.frameTime <= gLevels[level].levelClock.endTime)){
#if (ZLS_DEBUG>0)    
      zlsio_Printf( "  Initial Frame Ouput, level: %d:\n", level);
#endif
      err = zlsio_SaveLevelImages        ( level , zlst_GlobalCycles);
      if (err != noErr ) return ( err );
    }
    
    if ((gLevels[level].levelClock.dataTime <= gLevels[level].levelClock.endTime)){
#if (ZLS_DEBUG>0)    
      zlsio_Printf( "  Initial Data Ouput, level: %d:\n", level);
#endif
      err = zlsio_SaveLevelZoneData ( level , zlst_GlobalCycles);
      if (err != noErr ) return ( err );
    }
    
  }
  
  return ( err );
}

