#ifndef ZLS_IO_H
#define ZLS_IO_H
#include "zls.h"

//
// big and little endian tags, to reflect writers architecture
// so that the tags appear in the same byte order in the file 
// no matter what endian writes the tags.  Data values are written
// in OS endian order, only the tags are independent.
//
// Only the main MAGIC TAG is written in OS order so it can be used to 
// check the file format and endianness.

#ifdef DATA_BIGENDIAN_OS

#define ZLS_VERSTAG     0x56455253
#define ZLS_IDTAG       0x5a4e4944

#define ZLS_FILETAG      0x46494c45
#define ZLS_FORMATTAG    0x46524d54

#define ZLS_TIMETAG      0x54494d45
#define ZLS_CYCLESTAG    0x4359434c

#define ZLS_TRANSFORMTAG 0x5452414e

#define ZLS_DMAXTAG      0x444d4158
#define ZLS_DMINTAG      0x444d494e
#define ZLS_MAXRTAG      0x4d415852
#define ZLS_MINRTAG      0x4d494e52

#define ZLS_DIMSTAG      0x44494d53

#define ZLS_XDIMTAG      0x5844494d
#define ZLS_YDIMTAG      0x5944494d
#define ZLS_ZDIMTAG      0x5a44494d
#define ZLS_NGSTTAG      0x4e475354

#define ZLS_VARTAG       0x56415253
#define ZLS_NVARTAG      0x4e564152

#define ZLS_SAMPLETAG    0x534d504c
#define ZLS_DATATAG      0x44415441
#define ZLS_UNITSTAG     0x554e4954

#define ZLS_RNGETAG      0x524e4745

#define ZLS_XRNGTAG      0x58524e47
#define ZLS_XTYPTAG      0x58545950
#define ZLS_XKNDTAG      0x584b4e44
#define ZLS_XCRDTAG      0x58435244

#define ZLS_YRNGTAG      0x59524e47
#define ZLS_YTYPTAG      0x59545950
#define ZLS_YKNDTAG      0x594b4e44
#define ZLS_YCRDTAG      0x59435244

#define ZLS_ZRNGTAG      0x5a524e47
#define ZLS_ZTYPTAG      0x5a545950
#define ZLS_ZKNDTAG      0x5a4b4e44
#define ZLS_ZCRDTAG      0x5a435244

#endif DATA_BIGENDIAN_OS

#ifdef DATA_LITENDIAN_OS

#define ZLS_VERSTAG      0x53524556
#define ZLS_IDTAG        0x44494e5a

#define ZLS_FILETAG      0x454c4946
#define ZLS_FORMATTAG    0x544d5246

#define ZLS_TIMETAG      0x454d4954
#define ZLS_CYCLESTAG    0x4c435943

#define ZLS_TRANSFORMTAG 0x4e415254

#define ZLS_DMAXTAG      0x58414d44
#define ZLS_DMINTAG      0x4e494d44
#define ZLS_MAXRTAG      0x5258414d
#define ZLS_MINRTAG      0x524e494d

#define ZLS_DIMSTAG      0x534d4944
#define ZLS_XDIMTAG      0x4d494458
#define ZLS_YDIMTAG      0x4d494459
#define ZLS_ZDIMTAG      0x4d49445a
#define ZLS_NGSTTAG      0x5453474e

#define ZLS_VARTAG       0x53524156
#define ZLS_NVARTAG      0x5241564e

#define ZLS_SAMPLETAG    0x4c504d53
#define ZLS_DATATAG      0x41544144
#define ZLS_UNITSTAG     0x54494e55

#define ZLS_RNGETAG      0x45474e52

#define ZLS_XRNGTAG      0x474e5258
#define ZLS_XTYPTAG      0x50595458
#define ZLS_XKNDTAG      0x444e4b58
#define ZLS_XCRDTAG      0x44524358

#define ZLS_YRNGTAG      0x474e5259
#define ZLS_YTYPTAG      0x50595459
#define ZLS_YKNDTAG      0x444e4b59
#define ZLS_YCRDTAG      0x44524359

#define ZLS_ZRNGTAG      0x474e525a
#define ZLS_ZTYPTAG      0x5059545a
#define ZLS_ZKNDTAG      0x444e4b5a
#define ZLS_ZCRDTAG      0x4452435a

#endif DATA_LITENDIAN_OS


// binary data file file formats
// 0, ZLS, 1, FITS, 2 HDF (nyi, will need to link to HDF library)
enum {  kZLS_ZLSData   = 0x00000000,
  kZLS_FITSData   = 0x00000001,
  kZLS_HDFData   = 0x00000002
};

// binary data file file cell values
enum {  kZLS_UInt8Data   = 0x00000000,
  kZLS_SInt8Data   = 0x00000001,
  kZLS_UInt16Data  = 0x00000010,
  kZLS_SInt16Data  = 0x00000011,
  kZLS_UInt32Data  = 0x00000020,
  kZLS_SInt32Data  = 0x00000021,
  kZLS_Float32Data = 0x00000022,
  kZLS_UInt64Data  = 0x00000040,
  kZLS_SInt64Data  = 0x00000041,
  kZLS_RealData    = 0x00000042
};

// output resolutions - bits per pixel, image formats

enum {  kRawOutput   = 0x00,
  kTIFFOutput  = 0x01,
  kJPEGOutput  = 0x02,
  kPNGOutput   = 0x03,
  k8bitOutput  = 0x08,
  k16bitOutput = 0x10,
  k24bitOutput = 0x18,  // 3x8 rgb not used yet
  k32bitOutput = 0x20   // alpha rgb 4x8 bit  see also kZLS_Float32Data for 32 bit float data files in zls_data.h
};

// there is one of these records for each varIndex in the zones

enum {
  kZLS_IOAutoScale   = 0,  // Full Auto 
  kZLS_IOManualScale = 1,  //  specified max and min
  kZLS_IOAutoMaxScale = 2, // Auto with a specified max
  kZLS_IOAutoMinScale = 3  // Auto with a specified min
};

enum{
  kZLS_IOLinearTransform = 0,
  kZLS_IOLog10Transform = 1,
  kZLS_IOExp10Transform = 2,
  kZLS_IOPowTransform = 3,
  kZLS_IOLog10AbsTransform = 4
};

enum{
  kZLS_IOTEXT   = 0x00,
  kZLS_IOImage  = 0x01,
  kZLS_IOData   = 0x02,
  kZLS_IORender = 0x04,
  kZLS_IOPlot   = 0x08
};

enum{
  kZLS_IOOutputSeparateZones = 0,
  kZLS_IOOutputMergedZones   = 1
};

#define PATHLEN   512
#define DIRLEN    512
#define PREFIXLEN  16
#define CODELEN     8

//==============================================================================
// Various Global and User Interface Strings
//==============================================================================
#ifdef HYDRO
#define FYRIS  "Fyris Alpha"
#endif HYDRO

#ifdef MHD
#define FYRIS  "Fyris Beta"
#endif MHD

#ifdef SRHD
#define FYRIS  "Fyris Gamma"
#endif SRHD

#ifndef FYRIS
#define FYRIS "Fyris"
#endif

#define DIVIDERLINE  "================================================================"


//==============================================================================
// default file name strings, not all used yet :)
//==============================================================================

#define PATHSEP         "/"
#define ZLSCONFFILE     "zls"
#define RESTARTFILE     "Restart"
#define RESTARTDIR      "Output"
#define CONFEXT         ".conf"
#define OUTPUTDIR       "Output"
#define MONITORDIR      "monitor"
#define DATADIR         "data"
#define USERDIR         "user"


#define CTABEXT         ".ctab"

// jpeg and png are always compressesd

#define JPGEXT          ".jpg"  
#define PNGEXT          ".png"

// binary io files may be optionally saved and read from .gz format
#ifdef ZLIB_COMPRESSED_IO
#ifdef  ZLIB_UNCOMPRESSED_IMAGES     
// io is compressed and but 2D images are not
#define TIFFEXT         ".tif"
#define RAWEXT          ".raw"
#else
// io is compressed and so are 2D images
#define TIFFEXT         ".tif.gz"
#define RAWEXT          ".raw.gz"
#endif ! ZLIB_UNCOMPRESSED_IMAGES     
#else
// normal not compressed anyway
#define TIFFEXT         ".tif"
#define RAWEXT          ".raw"
#endif !ZLIB_COMPRESSED_IO

#define USEREXT         ".txt"

#define GRIDCONFFILE    "grid"
#define PHYSCONFFILE    "physics"
#define IOCONFFILE      "io"
#define INITCONFFILE    "init"
#define PPMCONFFILE     "ppm"

#define LEVELNAME       "level"
#define LEVELFILEEXT    ".conf"
#define ZONENAME        "zone"
#define ZONEFILEEXT     ".conf"
#define STATENAME       "state"
#define DATANAME        "data"

// binary io files may be optionally saved and read from .gz format
#ifdef ZLIB_COMPRESSED_IO
#define BINFILE         "zls.bdat.gz"
#define BINFILEEXT      ".bdat.gz"
#define STATEFILEEXT    ".sdat.gz"
#define DATAFILEEXT     ".zdat.gz"
#define DATAFILE        "zls.zdat.gz"
#define FITSFILE        "zls.fits.gz"
#define FITSFILEEXT     ".fits.gz"
#else
#define BINFILE         "zls.bdat"
#define BINFILEEXT      ".bdat"
#define STATEFILEEXT    ".sdat"
#define DATAFILEEXT     ".zdat"
#define DATAFILE        "zls.zdat"
#define FITSFILE        "zls.fits"
#define FITSFILEEXT     ".fits"
#endif !ZLIB_COMPRESSED_IO

#define HTMLEXT      ".html"
#define HTMLHEAD     "header"
#define HTMLTAIL     "tail"
#define ZLSLOGFILE   "zlsLog"
#define ZLSWEBFILE   "zlsWeb"
#define ZLSUSERLOGFILE   "zlsUserLog"
#define LOGEXT       ".txt"
#define ZLSINFOFILE  "zlsConfLog"
#define INFOEXT      ".txt"
#define ZLSFILESLOG  "zlsIOLog"
#define FILESEXT     ".txt"

#define ZLS_MONITOR_FILES_PNG

#define MONFILE      "fyris"

#ifdef ZLS_MONITOR_FILES_JPEG
enum {  kMonitorFormatOutput = kJPEGOutput};
#define MONEXT        ".jpg"
#endif ZLS_MONITOR_FILES_JPEG

#ifdef ZLS_MONITOR_FILES_TIFF
enum {  kMonitorFormatOutput = kTIFFOutput};
#define MONEXT        ".tif"
#endif ZLS_MONITOR_FILES_TIFF

#ifdef ZLS_MONITOR_FILES_PNG
enum {  kMonitorFormatOutput = kPNGOutput};
#define MONEXT        ".png"
#endif ZLS_MONITOR_FILES_PNG

#define OUTFILE      "zls"
#define OUTCTAB      "./ctabs/01.rss.ctab"


typedef struct iostate {
  
  // these look like io specifc items, but since they are time dependent
  // and we don't want to lose track of file numbers after restarts they are
  // separate and are saved in the levelclock files.
  
  Counter   ioFrameCount;    // output step count to keep all Image IO in sync.x
  Counter   ioUserCount;     // output step count to keep all User IO in sync
  Counter   ioDataCount;     // output step count to keep all 16bit data IO in sync
  Counter   ioRealDataCount; // output step count to keep all flaoting point (other than restarts) data IO in sync
  Counter   ioSaveCount;     // output step count to keep all restart IO in sync
  
  
} IOStateRec, * IOStatePtr;

typedef struct vario {
  char     * varCode   ; // >= 4 letter code for specifying output var, initially just straight vars
  char     * varPrefix ;  // var string for building file names
  
  TIFFCTab * varTIFFCTab; // for 8 bit TIFFs, NULL if not used
  
  Integer ioTypes;         // TEXT, image, data etc, Which is used depends on caller selecting
  // sum of flags: kZLS_IOTEXT   = 0x00,
  // kZLS_IOImage  = 0x01,
  // kZLS_IOData   = 0x02,
  // kZLS_IORender = 0x04
  
  Integer imageSlice;    // (nD-1) slices planes, kSliceXY, kSliceYZ, kSliceXZ ie 2D slice of 3D vol.
  Integer imageIndex;    // index along axis where plane is sliced from 3D, or row is taken from 2D
  Integer imageIndex2;   // 2nd index along axis where row is extracted from 3D
  Integer imagePlotColor; // colour index in CTAB used for plotting
  
  // will include 3D volumetric rendering engine eventually ??
  
  Integer  imageScalingType;  // kZLS_IOAutoScale, kZLS_IOManualScale
  Integer  dataScalingType;   // kZLS_IOAutoScale, kZLS_IOManualScale
  
  Real dataMin;     // Min if manualScale, data min if auto or semiauto
  Real dataMax;     // Max if manualScale, data max if auto or semiauto
  
  Real scaleMin;      // Min if manualScale, data min if auto or semiauto
  Real scaleMax;      // Max if manualScale, data max if auto or semiauto
  
  Real scaleInvRange;   // 1.0/(Max-Min) if manualScale
  
  Integer  transformType; // kZLS_IOLinearTransform, kZLS_IOLog10Transform, kZLS_IOExp10Transform, kZLS_IOPowTransform
  
  Real transPow;      // index if kZLS_IOPowTransform
  Real transThreshold;  // threshold value
  Real transNULL;     // value for invalid cells
  
} VarIORec, * VarIORecPtr;

typedef struct io { 
  //
  // Structure to contain common IO data
  // Zones and Levels can contain references
  // to some of this data (rather than copies)
  //
  // 
  // Data Associated with Variable Input/Output 
  //
  //
  // base strings for filenames
  //
  char fileName        [PATHLEN]; // scratch for any file use
  
  char outDir          [PATHLEN]; // base dir for output
  char outName         [PATHLEN]; // base name for output
  char outLogName      [PATHLEN]; // base name for output log file
  char outUserLogName  [PATHLEN]; // base name for output user log file
  char outInfoName     [PATHLEN]; // base name for configuration log file
  char outFilesName    [PATHLEN]; // base name for file creation log file
  
  char outRunName      [PATHLEN]; // string to idenify particular run in IO and monitor
  
  char outMonitorName  [PATHLEN]; // base name for monitor var html file
  char monDir          [PATHLEN]; // base dir for monitor var  file
  
  char restartName     [PATHLEN]; // base name for restart file, set by zls.conf and defaults to "Restart"
  char restartDir      [PATHLEN]; // location for restart file, set by zls.conf and defaults to "./"
  
  char outCTABFile     [PATHLEN]; // string for reading TIFF CTABS
  char monitorCTABFile [PATHLEN]; // string for scale bar on monitor page
  //
  //
  FILE * outUserLogFILE;  // accessor for log file already open with append permissions
  FILE * outLogFILE;      // accessor for log file already open with append permissions
  FILE * outInfoFILE;     // accessor for Conf log file already open with append permissions
  FILE * outFilesFILE;    // accessor for Files log file already open with append permissions
  //
  //
  TIFFCTab *     ioTIFFCTab; //  default CTAB if var CTABS are not defined
  //
  IOStateRec ioState;
  
  // data  file format common data, all data files from a run will have the
  // same format
  Integer  outDataBitsPerCell;
  Integer  outDataFormat;  // 0, ZLS, 1, FITS, 2 HDF (nyi, will need to link to HDF library)
  Integer  outMergedZoneData    ;     
  
  // image format common data, all image files from a run will have the 
  // same format
  //
  Integer  outBitsPerPix   ;
  Integer  outFormat       ;
  Integer  outMergedZoneFrames  ;     
  //
  // whether to merge zones in levels and write levels, or zones independently
  // Not recommended for large 3D simulations when the imagezone will be too large
  // leading to sub zone processing in the first place :)
  // Good for 2D sims split into lots of small tiles, or small 3D sims.
  // Large 3D sims should put out separate zones and joining is part of post processing
  //
  
  // scratch images, allocated and destroyed dyamically
  // image sizes are determined zone by zone, and are not
  // uniform in general.
  //
  
  unsigned char  *   image8;
  UInt16         *   image16;
  Real           *   imageReal;
  
  // info for each io variable:
  Integer outNVars;
  Integer monitorVar; // index of monitoring variable html output, -1 for none
  Integer monitorLevel; // sim level for monitoring variable html output, 0 by default or if out of range
  
  VarIORecPtr v;  // array dynamically allocated at startup, outNVars elements
  
} IOSettings,  *IOSettingsPtr;

typedef IOSettingsPtr IOSettingsArr;

//
//========================================
//

int zlsio_AllocateIOImages       ( IOSettingsPtr io, int nx, int ny );
int zlsio_FreeIOImages           ( IOSettingsPtr io );
int zlsio_InitIOImageArrays      ( IOSettingsPtr io );

int zlsio_InitVarIORec           ( VarIORecPtr vio );

int zlsio_InitIOSettings         ( IOSettingsPtr io, Integer nVars, Integer monitorVar, Integer monitorLevel );
int zlsio_NewIOSettings          ( IOSettingsPtr * io );
int zlsio_DisposeIOSettings      ( IOSettingsPtr io );

//
// Only the save name for future restarts is set, the restart being used is, if any, set
// in zls.conf - always.
//

int zlsio_SetIOSettingsRestartName ( IOSettingsPtr io, char * restartSaveName );
int zlsio_SetIOSettingsRestartDir  ( IOSettingsPtr io, char * restartDirName  );

//
//========================================
//

int zlsio_GetVarFiducialValue    ( Real *f0, Integer varIndex);
int zlsio_GetIOCodeVarIndex      ( Integer *varIndex, char * varCode );
int zlsio_GetVarIOVarPrefix      ( char * vPref, IOSettingsPtr io,  Integer varIndex );

//
//========================================
//
int zlsio_WriteInitialOutput(void) ;  // writes out a User, Frame and 16/32bit data dump of the initial state

//
// Basic IO calls that allow compession if enabled by ZLIB_COMPRESSED_IO
// or simply pass through the stdio if ZLIB_COMPRESSED_IO is disabled.
//

static inline zFile zlsio_open ( char * path, char * mode );
static inline zFile zlsio_open ( char * path, char * mode ){
#ifdef ZLIB_COMPRESSED_IO
  return( gzopen( path, mode ) );
#else
  return(  fopen( path, mode ) );
#endif !ZLIB_COMPRESSED_IO
}

static inline int zlsio_write ( zFile f, void * p, Size s );
static inline int zlsio_write ( zFile f, void * p, Size s ){
#ifdef ZLIB_COMPRESSED_IO
  int err;
  gzwrite( f, p, s );
  gzerror ( f, &err );
  if ( err >= 0 ) err = noErr;
  return( err );
#else
  fwrite( p, s, 1, f);
  return( ferror ( f ) );
#endif !ZLIB_COMPRESSED_IO
}


static inline Size zlsio_read ( zFile f, void * p, Size s );
static inline Size zlsio_read ( zFile f, void * p, Size s ){
#ifdef ZLIB_COMPRESSED_IO
  int err;
  gzread( f, p, s );
  gzerror ( f, &err );
  if ( err >= 0 ) err = noErr;
  return( err );
#else
  fread( p, s, 1, f);
  return( ferror ( f ) );
#endif !ZLIB_COMPRESSED_IO
}

static inline int zlsio_seek ( zFile f, Size offset ); // always from start of file, ie SEEK_SET
static inline int zlsio_seek ( zFile f, Size offset ){
#ifdef ZLIB_COMPRESSED_IO
  int err;
  gzseek( f, offset, SEEK_SET );
  gzerror ( f, &err );
  if ( err >= 0 ) err = noErr;
  return( err );
#else
  fseek( f, offset, SEEK_SET );
  return( ferror ( f ) );
#endif !ZLIB_COMPRESSED_IO
}


static inline int zlsio_close ( zFile f );
static inline int zlsio_close ( zFile f ){
#ifdef ZLIB_COMPRESSED_IO
  return( gzclose( f ) );
#else
  return( fclose ( f ) );
#endif !ZLIB_COMPRESSED_IO
}
//
//  IO calls that don't compress ever
//

static inline zFileNC zlsio_openNC ( char * path, char * mode );
static inline zFileNC zlsio_openNC ( char * path, char * mode ){
  return(  fopen( path, mode ) );
}

static inline int zlsio_writeNC ( zFileNC f, void * p, Size s );
static inline int zlsio_writeNC ( zFileNC f, void * p, Size s ){
  fwrite( p, s, 1, f);
  return( ferror ( f ) );
}


static inline Size zlsio_readNC ( zFileNC f, void * p, Size s );
static inline Size zlsio_readNC ( zFileNC f, void * p, Size s ){
  fread( p, s, 1, f);
  return( ferror ( f ) );
}

static inline int zlsio_closeNC ( zFileNC f );
static inline int zlsio_closeNC ( zFileNC f ){
  return( fclose ( f ) );
}



//
//========================================

#endif
