#include "zls.h"

int zlsio_WriteZoneDataBinaryHeader ( zFile f, ZonePtr z);
int zlsio_WriteZoneDataBinaryHeader ( zFile f, ZonePtr z){
  
  int err = noErr;
  zoneDataHeader zdh;
  
  zdh.ZLS_MagicTag   = ZLS_MAGICTAG;
  zdh.ZLS_DataOffset = (unsigned int)ZLS_ZONEDATA_HEADERSIZE; // offset to data from start of file
  zdh.ZLS_Version    = ZLS_VERSION;
  zdh.ZLS_IDTag      = ZLS_IDTAG;
  zdh.ZLS_ZoneID     = z->zoneID;
  
  zdh.ZLS_VarsTag    = ZLS_VARTAG;
  zdh.ZLS_VarsType   = z->zoneType;
  zdh.ZLS_nVarTag    = ZLS_NVARTAG;
  zdh.ZLS_nVars      = z->nQ;
  
  zdh.ZLS_DimsTag  = ZLS_DIMSTAG;
  zdh.ZLS_Dims     = z->zoneGrid.nD;
  zdh.ZLS_XDimTag  = ZLS_XDIMTAG;
  zdh.ZLS_Xdim     = z->zoneGrid.nt[kXDim];
  zdh.ZLS_YDimTag  = ZLS_YDIMTAG;
  zdh.ZLS_Ydim     = z->zoneGrid.nt[kYDim];
  zdh.ZLS_ZDimTag  = ZLS_ZDIMTAG;
  zdh.ZLS_Zdim     = z->zoneGrid.nt[kZDim];
  zdh.ZLS_NGSTTag  = ZLS_NGSTTAG;
  zdh.ZLS_NGST     = z->zoneGrid.nG;
  
  zdh.ZLS_SampleTag  = ZLS_SAMPLETAG;
  zdh.ZLS_SampleSize = ((unsigned int) sizeof(Real));
  
  Size chunkSize = ZLS_ZONEDATA_HEADERSIZE;
  err = zlsio_write( f, &zdh, chunkSize );
  
  return(err);
  
}

//
// Attempt to byte swap data fields of zone data header in memory..
// just swap values not tags
//
int zlsio_ByteSwapZoneDataBinaryHeader ( zoneDataHeader *zdh);
int zlsio_ByteSwapZoneDataBinaryHeader ( zoneDataHeader *zdh){
  
  int err = noErr;
  
  zls_Int32Swap(&zdh->ZLS_DataOffset);
  zls_Int32Swap(&zdh->ZLS_Version);
  zls_Int32Swap(&zdh->ZLS_ZoneID);
  zls_Int32Swap(&zdh->ZLS_VarsType);
  zls_Int32Swap(&zdh->ZLS_nVars);
  zls_Int32Swap(&zdh->ZLS_Dims);
  zls_Int32Swap(&zdh->ZLS_Xdim);
  zls_Int32Swap(&zdh->ZLS_Ydim);
  zls_Int32Swap(&zdh->ZLS_Zdim);
  zls_Int32Swap(&zdh->ZLS_NGST);
  zls_Int32Swap(&zdh->ZLS_SampleSize);
  
  return(err);
  
}


int zlsio_DumpReal_ZoneData  (char * aFile, ZonePtr z );
int zlsio_DumpReal_ZoneData  (char * aFile, ZonePtr z ){
  int err = noErr; 
  
  zFile fileAccess;
  Size chunkSize;
  
  fileAccess = zlsio_open( aFile, "wb");
  
  if (fileAccess) {
    
    err = zlsio_WriteZoneDataBinaryHeader ( fileAccess, z );
    
    if (err == noErr) { 
      
      Integer ni = z->zoneGrid.nt[kXDim];
      Integer nj = z->zoneGrid.nt[kYDim];
      Integer nk = z->zoneGrid.nt[kZDim];
      
      ni = ni > 0? ni : 1L;
      nj = nj > 0? nj : 1L;
      nk = nk > 0? nk : 1L;
      
      chunkSize = (Size)ni*nj*nk*sizeof(Real); // in general write in big chunks
      
      Integer n = z->nQ;
      Integer varIndex;
      QPtr q;
      
      for (varIndex = 0; varIndex<n; varIndex++) {
        
        err = zls_GetZoneDataDirect(&q, z, varIndex); // returns a direct pointer whatever format (plain or multi-D) the zone is;
        
        if ((err == noErr) && (q != NULL)){
          err = zlsio_write( fileAccess, q, chunkSize );
          if ( err != noErr ) { 
            zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_ZoneData: file write error: %d \n", err );
            zlsio_close(fileAccess);
            return(err);
          }
          
        } else {
          zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_ZoneData: zone data not found, index %d \n", varIndex );
          err = zonErr;
          zlsio_close(fileAccess);
          return(err);
        }
        
      } // varIndex loop
      
      
    } else { 
      zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_ZoneData: Header Read Error: %d \n", err );
      zlsio_close(fileAccess);
      return (err);
    } //header
    
    zlsio_close(fileAccess);
    
  } else { return (fileErr);} // access
  
  return ( err );
  
}


int zlsio_ReadReal_ZoneData  ( ZonePtr z, char * aFile );
int zlsio_ReadReal_ZoneData  ( ZonePtr z, char * aFile ){
  int err = noErr; 
  
  zFile fileAccess;
  Size chunkSize;
  zoneDataHeader zdh;
  
  fileAccess = zlsio_open( aFile, "rb");
  
  if (fileAccess) {
    
    chunkSize = ZLS_ZONEDATA_HEADERSIZE; 
    err       =  zlsio_read( fileAccess, &zdh, chunkSize );
    if ( err != noErr ){
      zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_ZoneData: Header Read Error: %d \n", err );
      zlsio_close(fileAccess);
      return (err);
    } //header
    
    Integer dataEndian = kZLS_BadFormat;
    
    if ( zdh.ZLS_MagicTag == CURRENT_ENDIAN_MAGIC ) {
      dataEndian = kZLS_OKFormat;
    } else if ( zdh.ZLS_MagicTag == OPPOSITE_ENDIAN_MAGIC ) {
      zlsio_ErrPrintf( "WARNING: zlsio_ReadReal_ZoneData, %s.\n", aFile );
      zlsio_ErrPrintf( "WARNING: Data has a valid format,\n" );
      zlsio_ErrPrintf( "WARNING: Data has a valid format, but is opposite byteswap format to current system.\n" );
      zlsio_ByteSwapZoneDataBinaryHeader(&zdh);
      dataEndian = kZLS_SWAPFormat;
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_ZoneData, %s, zone data invalid format.\n", aFile );
      err = fileErr;
      zlsio_close(fileAccess);
      return(err);
    }
    
    if (( dataEndian == kZLS_OKFormat ) || ( dataEndian == kZLS_SWAPFormat )  ){
      
      Integer dataFormat = kZLS_OKFormat;
      // check version and version tag, but ignore for now.
      if ( zdh.ZLS_Version != ZLS_VERSION ) dataFormat = kZLS_BadFormat;
      
      dataFormat = kZLS_OKFormat;
      
      // now test and fail on any ...
      
      if (( zdh.ZLS_VarsTag  != ZLS_VARTAG )    || (zdh.ZLS_VarsType != (unsigned int) z->zoneType) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_nVarTag  != ZLS_NVARTAG )   || (zdh.ZLS_nVars    != (unsigned int) z->nQ) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_DimsTag  != ZLS_DIMSTAG )   || (zdh.ZLS_Dims     != (unsigned int) z->zoneGrid.nD) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_XDimTag  != ZLS_XDIMTAG )   || (zdh.ZLS_Xdim     != (unsigned int) z->zoneGrid.nt[kXDim]) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_YDimTag  != ZLS_YDIMTAG )   || (zdh.ZLS_Ydim     != (unsigned int) z->zoneGrid.nt[kYDim]) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_ZDimTag  != ZLS_ZDIMTAG )   || (zdh.ZLS_Zdim     != (unsigned int) z->zoneGrid.nt[kZDim]) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_NGSTTag  != ZLS_NGSTTAG )   || (zdh.ZLS_NGST     != (unsigned int) z->zoneGrid.nG ) ) dataFormat = kZLS_BadFormat;
      if (( zdh.ZLS_SampleTag!= ZLS_SAMPLETAG ) || (zdh.ZLS_SampleSize   != ((unsigned int) sizeof(Real))) ) dataFormat = kZLS_BadFormat;
      
      
      chunkSize = ZLS_ZONEDATA_HEADERSIZE; 
      if ( zdh.ZLS_DataOffset !=  chunkSize ) dataFormat = kZLS_BadFormat ; // offset to data from start of file
      
      if (  dataFormat == kZLS_OKFormat ) {
        // just carry on reading... header sizes match.  Not dealing with non matching headers at this point in time
        QPtr q;
        Integer varIndex;
        
        for ( varIndex = 0; varIndex < z->nQ; varIndex++ ) {
          chunkSize = (Size)zdh.ZLS_SampleSize * zdh.ZLS_Xdim * zdh.ZLS_Ydim * zdh.ZLS_Zdim; // set each time to be safe
          
          zlsio_LogPrintf( " Reading Variable of size: %u bytes, X %u, Y %u Z %u\n", chunkSize, zdh.ZLS_Xdim, zdh.ZLS_Ydim, zdh.ZLS_Zdim);
          if (  zls_GetZoneDataDirect(&q, z, varIndex) == noErr ) {
            err       =  zlsio_read( fileAccess, q, chunkSize );
            if ( err != noErr ){
              zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_ZoneData: Data Read Error: %d \n", err );
              zlsio_close(fileAccess);
              return (err);
            } //header
            
            // swap data if needed
            if ( dataEndian == kZLS_SWAPFormat ) {
#ifdef SINGLE_PRECISION 
              zls_ByteSwapFloatBlock ( q, chunkSize);
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION
              zls_ByteSwapDoubleBlock( q, chunkSize);
#endif DOUBLE_PRECISION
            }
            
          }
        }
        
      } else { zlsio_close( fileAccess ); return (fileErr); }
      
    } else { zlsio_close( fileAccess ); return (fileErr); } //header
    
    zlsio_close( fileAccess );
    
  } else { return ( fileErr ); } // access
  
  return ( err );
  
}


int zlsio_SaveZoneRealData(ZonePtr z, IOSettingsPtr io){
  //
  int        err = noErr;
  if ( ( z != NULL) && ( io != NULL) ){
    
    char outFile[PATHLEN];
    
    sprintf(outFile, "%s%s%sL%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP,
            io->outName, z->zoneLevel, z->zoneID, (int)io->ioState.ioRealDataCount, DATAFILEEXT );
    
    err =  zlsio_DumpReal_ZoneData  ( outFile, z );
    
    if (err != noErr ){
      zlsio_ErrPrintf( " IO ERROR: zlsio_SaveZoneRealData, %s Err = %d \n", outFile, err );
    };
    
  };
  
  return(err);
}

int zlsio_SaveLevelZoneRealData(Integer level, int step){
#pragma unused(step)
  int     err = noErr;
  
  Integer              zone;  
  ZonePtr                 z;
  
  IOSettingsPtr  io = NULL;
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL){
    
    for( zone = 0; zone <gLevels[level].nZones; zone++){
#ifdef MPI  
      if (z->zoneRankID == gMPIrank) {
#endif MPI
        
        err = zls_GetLevelZoneByIndex( &z, level, zone );
        if ( z != NULL )  err = zlsio_SaveZoneRealData( z, io );
#ifdef MPI  
      }
#endif MPI
      
    } // zone
    if ( level == 0 ) io->ioState.ioRealDataCount++;
  } // io
  
  return(err);
}


int zlsio_WriteStateBinaryHeader ( zFile f );
int zlsio_WriteStateBinaryHeader ( zFile f ){
  int err = noErr;
  stateDataHeader cdh;
  
  cdh.ZLS_MagicTag   = ZLS_MAGICTAG;
  cdh.ZLS_DataOffset = (unsigned int) ZLS_STATEDATA_HEADERSIZE; // offset to data from start of file
  cdh.ZLS_VersTag    = ZLS_VERSTAG;
  cdh.ZLS_Version    = ZLS_VERSION;
  
  Size chunkSize = ZLS_STATEDATA_HEADERSIZE;
  err       = zlsio_write( f, &cdh, chunkSize );
  
  return(err);
}
//
// Attempt to byte swap data fields of state data header in memory..
// just swap values not tags
//
int zlsio_ByteSwapStateBinaryHeader ( stateDataHeader *cdh);
int zlsio_ByteSwapStateBinaryHeader ( stateDataHeader *cdh){
  
  int err = noErr;
  zls_Int32Swap(&cdh->ZLS_DataOffset);
  zls_Int32Swap(&cdh->ZLS_Version);
  return(err);  
}

int zlsio_ByteSwapTimeCoords ( TimeCoords * tc );
int zlsio_ByteSwapTimeCoords ( TimeCoords * tc ){
  
  int err = noErr;
#ifdef SINGLE_PRECISION 
  zls_FloatSwap(&tc->t);   // simulation time in Zone
  zls_FloatSwap(&tc->dt);   // timestep in Zone    
  zls_FloatSwap(&tc->endTime);   // time to end sim, in scaled model units
  zls_FloatSwap(&tc->tFrame);   // increment by dt, reset to 0 when triggered
  zls_FloatSwap(&tc->tUser);   // timer for user output routine
  zls_FloatSwap(&tc->tData);   // timer for 16 bit signed data  output routine
  zls_FloatSwap(&tc->tSave);   // timer for complete state save
  zls_FloatSwap(&tc->frameTime);   // when to trigger the next frame, an interval rather
  zls_FloatSwap(&tc->userTime);   // than an absolute time
  zls_FloatSwap(&tc->dataTime);   // 
  zls_FloatSwap(&tc->saveTime);   // 
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION
  zls_DoubleSwap(&tc->t);   // simulation time in Zone
  zls_DoubleSwap(&tc->dt);   // timestep in Zone    
  zls_DoubleSwap(&tc->endTime);   // time to end sim, in scaled model units
  zls_DoubleSwap(&tc->tFrame);   // increment by dt, reset to 0 when triggered
  zls_DoubleSwap(&tc->tUser);   // timer for user output routine
  zls_DoubleSwap(&tc->tData);   // timer for 16 bit signed data  output routine
  zls_DoubleSwap(&tc->tSave);   // timer for complete state save
  zls_DoubleSwap(&tc->frameTime);   // when to trigger the next frame, an interval rather
  zls_DoubleSwap(&tc->userTime);   // than an absolute time
  zls_DoubleSwap(&tc->dataTime);   // 
  zls_DoubleSwap(&tc->saveTime);   // 
#endif DOUBLE_PRECISION
  
  
  zls_Int32Swap(&tc->cycles);   // computation cycles in Zone
  zls_Int32Swap(&tc->endCycles);   // cycle number to end sim
  zls_Int32Swap(&tc->timingCycles);   // computation cycles , start at 0 even after a restart
  
  zls_Int32Swap(&tc->nCyclesFrame);  // cycles between frame output
  zls_Int32Swap(&tc->nCyclesUser);   // cycles between user output
  zls_Int32Swap(&tc->nCyclesData);   // cycles between 16 bit signed data output
  zls_Int32Swap(&tc->nCyclesSave);   // cycles between restart output
  
  zls_Int32Swap(&tc->frameCycles);  // cycles for frame output
  zls_Int32Swap(&tc->userCycles);   // cycles for user output
  zls_Int32Swap(&tc->dataCycles);   // cycles for 16 bit signed data output
  zls_Int32Swap(&tc->saveCycles);   // cycles for state save  output
  
  err = noErr;
  return(err);  
}



int zlsio_ByteSwapIOStateRec ( IOStateRec * ios );
int zlsio_ByteSwapIOStateRec ( IOStateRec * ios ){
  int err = noErr;
  zls_Int32Swap(&ios->ioFrameCount);    // output step count to keep all Image IO in sync.x
  zls_Int32Swap(&ios->ioUserCount);     // output step count to keep all User IO in sync
  zls_Int32Swap(&ios->ioDataCount);     // output step count to keep all 16/32bit data IO in sync
  zls_Int32Swap(&ios->ioRealDataCount); // output step count to keep all flaoting point (other than restarts) data IO in sync
  zls_Int32Swap(&ios->ioSaveCount);     // output step count to keep all restart IO in sync
  zlsio_LogPrintf( " Swapped ioStateRec: F%u U%u D%u R%u S%u \n", ios->ioFrameCount,
                  ios->ioUserCount,
                  ios->ioDataCount,
                  ios->ioRealDataCount,
                  ios->ioSaveCount);
  return(err);  
}


int zlsio_ByteSwapCPUCoords ( CPUCoords * ct );
int zlsio_ByteSwapCPUCoords ( CPUCoords * ct ){
  int err = noErr;
  
  zls_Int32Swap(&ct->cputot[0]);    
  zls_Int32Swap(&ct->cputot[1]);  
  
  zls_Int32Swap(&ct->cpuusr[0]);    
  zls_Int32Swap(&ct->cpuusr[1]);   
  
  zls_Int32Swap(&ct->cpusys[0]);    
  zls_Int32Swap(&ct->cpusys[1]);    
  
  zls_Int32Swap(&ct->realtime[0]);    
  zls_Int32Swap(&ct->realtime[1]);    
  
#ifdef DOUBLE_PRECISION
  zls_DoubleSwap(&ct->modeltime[0]);    
  zls_DoubleSwap(&ct->modeltime[1]);    
#endif DOUBLE_PRECISION
#ifdef SINGLE_PRECISION
  zls_FloatSwap(&ct->modeltime[0]);    
  zls_FloatSwap(&ct->modeltime[1]);    
#endif SINGLE_PRECISION
  
  zls_Int32Swap(&ct->modelcycles[0]);    
  zls_Int32Swap(&ct->modelcycles[1]);    
  
  zlsio_LogPrintf( " Swapped CPUCoords: T%u U%u S%u R%u \n", 
                  ct->cputot[0],
                  ct->cpuusr[0],
                  ct->cpusys[0],
                  ct->realtime[0]);
  return(err);  
}


int zlsio_DumpReal_StateData  (char * aFile, TimeCoords * tc , IOStateRec * ios, CPUCoords * ct );
int zlsio_DumpReal_StateData  (char * aFile, TimeCoords * tc , IOStateRec * ios, CPUCoords * ct ){
  int err = noErr; 
  
  zFile fileAccess;
  Size chunkSize;
  
  fileAccess = zlsio_open( aFile, "wb");
  
  if (fileAccess) {
    
    err = zlsio_WriteStateBinaryHeader ( fileAccess );
    
    if (err == noErr) { 
      
      chunkSize = sizeof(TimeCoords); 
      
      if ( tc != NULL){
        err = zlsio_write( fileAccess, tc, chunkSize );
        if ( err != noErr ) { 
          zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData: file write error %d \n", err );
          zlsio_close(fileAccess);
          return(err);
        }
        
      } else {
        zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData, %s, clock data not found.\n", aFile );
        err = fileErr;
        zlsio_close(fileAccess);
        return(err);
      };
      
      chunkSize = sizeof(IOStateRec); 
      
      if ( ios != NULL){
        err    = zlsio_write( fileAccess, ios, chunkSize );
        if ( err != noErr ) { 
          zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData: file write error %d \n", err );
          zlsio_close(fileAccess);
          return(err);
        }
        
      } else {
        zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData, %s, iostate data not found.\n", aFile );
        err = fileErr;
        zlsio_close(fileAccess);
        return(err);
      };
      
      chunkSize = sizeof(CPUCoords); 
      
      if ( ct != NULL){
        err    = zlsio_write( fileAccess, ct, chunkSize );
        if ( err != noErr ) { 
          zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData: file write error %d \n", err );
          zlsio_close(fileAccess);
          return(err);
        }
        
      } else {
        zlsio_ErrPrintf( " ERROR: zlsio_DumpReal_StateData, %s, CPUCoords data not found.\n", aFile );
        err = fileErr;
        zlsio_close(fileAccess);
        return(err);
      };
      
    } else { return (fileErr); } //header
    
    zlsio_close( fileAccess );
    
  } else { return ( fileErr ); } // access
  
  return ( err );
  
}

int zlsio_ReadReal_StateData  ( TimeCoords * tc, IOStateRec * ios, CPUCoords * ct, char * aFile );
int zlsio_ReadReal_StateData  ( TimeCoords * tc, IOStateRec * ios, CPUCoords * ct, char * aFile ){
  int err = noErr; 
  
  zFile fileAccess;
  Size chunkSize;
  stateDataHeader sdh;
  
  fileAccess = zlsio_open( aFile, "rb");
  
  if (fileAccess) {
    
    chunkSize = ZLS_STATEDATA_HEADERSIZE; 
    err       =  zlsio_read( fileAccess, &sdh, chunkSize );
    if ( err != noErr ){
      zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_StateData: Header Read Error: %d \n", err );
      zlsio_close(fileAccess);
      return (err);
    } //header
    
    Integer dataEndian = kZLS_BadFormat;
    
    if ( sdh.ZLS_MagicTag == CURRENT_ENDIAN_MAGIC ) {
      dataEndian = kZLS_OKFormat;
    } else if ( sdh.ZLS_MagicTag == OPPOSITE_ENDIAN_MAGIC ) {
      zlsio_ErrPrintf( "WARNING: zlsio_ReadReal_StateData, %s.\n", aFile );
      zlsio_ErrPrintf( "WARNING: Data has a valid format, but is opposite byteswap format to current system.\n" );
      zlsio_ErrPrintf( "WARNING: Will try byteswapping, but may fail.\n" );
      zlsio_ByteSwapStateBinaryHeader(&sdh);
      dataEndian = kZLS_SWAPFormat;
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_StateData, %s, clock data invalid format.\n", aFile );
      err = fileErr;
      zlsio_close(fileAccess);
      return(err);
    }
    
    if ((dataEndian == kZLS_OKFormat) ||(dataEndian == kZLS_SWAPFormat)) {
      
      Integer dataFormat = kZLS_OKFormat;
      // check version and version tag, but ignore for now.
      if (( sdh.ZLS_VersTag != ZLS_VERSTAG ) || ( sdh.ZLS_Version != ZLS_VERSION )) dataFormat = kZLS_BadFormat;
      
      dataFormat = kZLS_OKFormat;
      
      chunkSize = ZLS_STATEDATA_HEADERSIZE ;
      
      if ( sdh.ZLS_DataOffset != chunkSize ) dataFormat = kZLS_BadFormat ; // offset to data from start of file
      
      if (  dataFormat == kZLS_OKFormat ) {
        // just carry on reading... header sizes match.  Not dealing with non matching headers at this point in time
        
        chunkSize = sizeof(TimeCoords); 
        
        if (tc != NULL) {  
          err       =  zlsio_read( fileAccess, tc, chunkSize );
          if ( err != noErr ){
            zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_StateData: Clock Data Read Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } // tc
          
          // swap data if needed
          if ( dataEndian == kZLS_SWAPFormat ) {
            zlsio_ByteSwapTimeCoords( tc );
          }
          
          zlst_SetClockTimingCycles( tc, 0); // reset timing cycles if any
        }
        
        chunkSize = sizeof(IOStateRec); 
        if (ios != NULL)  {
          err       =  zlsio_read( fileAccess, ios, chunkSize );
          if ( err != noErr ){
            zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_StateData: IO State Data Read Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } // ios
          // swap data if needed
          if ( dataEndian == kZLS_SWAPFormat ) {
            zlsio_ByteSwapIOStateRec(ios);
          }
          
        }
        
        chunkSize = sizeof(CPUCoords); 
        if (ct != NULL)  {
          err       =  zlsio_read( fileAccess, ct, chunkSize );
          if ( err != noErr ){
            zlsio_ErrPrintf( " ERROR: zlsio_ReadReal_StateData: IO State CPUCoords Data Read Error: %d \n", err );
            zlsio_close(fileAccess);
            return (err);
          } // ios
          // swap data if needed
          if ( dataEndian == kZLS_SWAPFormat ) {
            zlsio_ByteSwapCPUCoords(ct);
          }
          
        }
        
        
      } else { zlsio_close( fileAccess ); return (fileErr); }
      
    } else { zlsio_close( fileAccess ); return (fileErr); } //header
    
    zlsio_close( fileAccess );
    
  } else { return ( fileErr ); } // access
  
  return ( err );
  
}


int zlsio_RestartPhysCheck( ccl_t * config, char * conFile, char * prefix);
int zlsio_RestartPhysCheck( ccl_t * config, char * conFile, char * prefix){
  
  int err = noErr;
  const char     *str1;
  
  char keyStr[PATHLEN];
  
  sprintf( keyStr, "%s_Cooling", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef COOLING 
    if (strcmp(str1, "ON") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Cooling is %s, not ON . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#else
    if (strcmp(str1, "OFF") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Cooling is %s, not OFF . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif COOLING
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Cooling not specified.\n", conFile);
    return(dimErr);
  }
  
  sprintf( keyStr, "%s_Hydrodynamics", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef HYDRO 
    if (strcmp(str1, "ON") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Hydrodynamics is %s, not ON . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#else
    if (strcmp(str1, "OFF") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Hydrodynamics is %s, not OFF . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif HYDRO
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Hydrodynamics not specified.\n", conFile);
    return(dimErr);
  }
    
  sprintf( keyStr, "%s_Primitive_Vars", prefix );
  if ((str1= ccl_get( config, keyStr))){
    if (strcmp(str1, "OFF") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Primitive_Vars is %s, not OFF . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
  } else {
    zlsio_ErrPrintf( "WARNING  : zlsio_Restart, %s, Primitive_Vars not specified.\n", conFile);
   // return(dimErr);
  }

  sprintf( keyStr, "%s_Conserved_Vars", prefix );
  if ((str1= ccl_get( config, keyStr))){
    if (strcmp(str1, "OFF") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Conserved_Vars is %s, not OFF . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
  } else {
    zlsio_ErrPrintf( "WARNING  : zlsio_Restart, %s, Conserved_Vars not specified.\n", conFile);
   // return(dimErr);
  }
  
  sprintf( keyStr, "%s_Variable_EOS", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef VARIABLE_EOS 
    if (strcmp(str1, "ON") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Variable EOS is %s, not ON . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#else
    if (strcmp(str1, "OFF") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Variable EOS is %s, not OFF . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif VARIABLE_EOS
  } else {
    zlsio_ErrPrintf( "WARNING  : zlsio_Restart, %s, EOS not specified.\n", conFile);
   // return(dimErr);
  }
    

  Integer var = 0;
  sprintf( keyStr, "%s_VarType", prefix );
  if ((str1= ccl_get( config, keyStr))){
    var = strtol( str1, NULL, 0 );
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Zone Variables Type not specified.\n", conFile);
    return(dimErr);
  }
  
  if ( var != kAlgorithmZoneType) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, ZoneType is %d, not %d. Incompatible.\n", conFile, var, kAlgorithmZoneType);
    return(dimErr);
  }
  
  sprintf( keyStr, "%s_nVars", prefix );
  if ((str1= ccl_get( config, keyStr))){
    var = strtol( str1, NULL, 0 );
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Number of variables not specified.\n", conFile);
    return(dimErr);
  }
  
  if ( var != kNVars) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Number of variables is %d, not %d. Incompatible.\n", conFile, var, kNVars);
    return(dimErr);
  }
  
  sprintf( keyStr, "%s_nTracers", prefix );
  if ((str1= ccl_get( config, keyStr))){
    var = strtol( str1, NULL, 0 );
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Number of tracers not specified.\n", conFile);
    return(dimErr);
  }
  
  if ( var != kNTracers) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Number of tracers is %d, not %d. Incompatible.\n", conFile, var, kNTracers);
    return(dimErr);
  }
  
  sprintf( keyStr, "%s_TotalVars", prefix );
  if ((str1= ccl_get( config, keyStr))){
    var = strtol( str1, NULL, 0 );
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Total number of variables+tracers not specified.\n", conFile);
    return(dimErr);
  }
  
  if ( var != kNAllVars) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Number of variables+tracers is %d, not %d. Incompatible.\n", conFile, var, kNAllVars);
    zlsio_ErrPrintf( "       : Restart file, %s, may be corrupted or have been edited incorrectly.\n", conFile);
    return(dimErr);
  }
  
  return(err);
  
}


int zlsio_RestartGeomCheck ( ccl_t * config, char * conFile, char * prefix);
int zlsio_RestartGeomCheck ( ccl_t * config, char * conFile, char * prefix) {
  
  int err = noErr;
  const char     *str1;
  
  char keyStr[PATHLEN];
  
  Integer dims = 0;
  sprintf( keyStr, "%s_Dimensions", prefix );
  if ((str1= ccl_get( config, keyStr))){
    dims = strtol( str1, NULL, 0 );
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Dimensions not specified.\n", conFile);
    return(dimErr);
  }
  
#ifdef SYM1D 
  if ( dims != 1) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %1.1dD, not 1D. Incompatible.\n", conFile, dims);
    return(dimErr);
  }
#endif SYM1D
  
#ifdef SYM2D 
  if ( dims != 2) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %1.1dD, not 2D. Incompatible.\n", conFile, dims);
    return(dimErr);
  }
#endif SYM2D
  
#ifdef SYM3D 
  if ( dims != 3) { 
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %1.1dD, not 3D. Incompatible.\n", conFile, dims);
    return(dimErr);
  }
#endif SYM3D
  
  
  sprintf( keyStr, "%s_Geometry", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef CARTESIAN 
    if (strcmp(str1, "Cartesian") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %s, not Cartesian. Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif CARTESIAN
#ifdef CIRCULAR 
    if (strcmp(str1, "Circular") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %s, not Circular. Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif CIRCULAR
#ifdef SPHERICAL 
    if (strcmp(str1, "Spherical") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %s, not Spherical. Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif SPHERICAL
#ifdef POLARSPHERICAL 
    if (strcmp(str1, "PolarSpherical") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %s, not PolarSpherical. Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
    if (strcmp(str1, "Cylindrical") !=0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, is %s, not Cylindrical. Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif CYLINDRICAL
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Geometry not specified.\n", conFile);
    return(dimErr);
  }
  
  return(err);
  
}


int zlsio_RestartOSCheck ( ccl_t * config, char * conFile, char * prefix);
int zlsio_RestartOSCheck ( ccl_t * config, char * conFile, char * prefix){
  
  int err = noErr;
  
  const char     *str1;
  char   keyStr[PATHLEN];
  
  sprintf( keyStr, "%s_Precision", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef SINGLE_PRECISION 
    if (strcmp(str1, "Single") != 0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, Precision is %s, not Single . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION
    if (strcmp(str1, "Double") != 0 ) {
      zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Precision is %s, not Double . Incompatible.\n", conFile, str1);
      return(dimErr);
    }
#endif DOUBLE_PRECISION
  } else {
    zlsio_ErrPrintf( "ERROR  : zlsio_Restart, %s, Precision not specified.\n", conFile);
    return(dimErr);
  }
  
  sprintf( keyStr, "%s_DataEndian", prefix );
  if ((str1= ccl_get( config, keyStr))){
#ifdef DATA_LITENDIAN_OS 
    if (strcmp(str1, "LittleEndian") != 0 ) {
      zlsio_ErrPrintf( "WARNING: zlsio_Restart, %s.\n", conFile);
      zlsio_ErrPrintf( "WARNING: DataEndian is %s, not LittleEndian.\n", str1);
    }
#endif DATA_LITENDIAN_OS
#ifdef DATA_BIGENDIAN_OS
    if (strcmp(str1, "BigEndian") != 0 ) {
      zlsio_ErrPrintf( "WARNING: zlsio_Restart, %s.\n", conFile);
      zlsio_ErrPrintf( "WARNING: DataEndian is %s, not BigEndian.\n", str1);
    }
#endif DATA_BIGENDIAN_OS
  } else {
    zlsio_ErrPrintf( "WARNING: zlsio_Restart, %s.\n", conFile);
    zlsio_ErrPrintf( "WARNING: DataEndian is not specified or unknown, using current system but may fail.\n" );
  }
  
  return(err);
}


int zlsio_ReadZoneConf(char* zoneFile, char* zoneDir, Integer level, Integer zone, Integer extend ); 
int zlsio_ReadZoneConf(char* zoneFile, char* zoneDir, Integer level, Integer zone, Integer extend ){ 
#pragma unused(extend)
  int err = noErr;
  
  //  
  // if extend == kZLS_RestartExtendSim, then only set main times, leaving counters and events to current setup
  // this allows us to extend a simulation past the end of the old one, but the old one was saved when the end 
  // conditions were satisfied, and that was recorded in the level clocks.  
  // if extend == kZLS_RestartContinueSim, we assume the save was done at an intermediate point in a simulation where then end point is 
  // still wanted so we can just load the whole clock into the level clock in one go without a local temp.
  //
  
  struct ccl_t  config;
  const char     *str1;
  char   keyStr[PATHLEN];
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, zoneFile) == noErr) {
    
    // check version, and warn.  If fail then try, but it may not work.
    Integer version = 0;
    if ((str1= ccl_get( &config,"Zone_Version"))){
      version = strtol( str1, NULL, 0 );
    }
    
    if ( version != ZLS_VERSION ) {
      zlsio_ErrPrintf( "WARNING: zlsio_ReadZoneConf, %s, different version, %d\n", zoneFile, err);
      zlsio_ErrPrintf( "WARNING: proceeding, but may fail. Please find code with correct version.\n" );
    }
    
    // check basic geometry, if fails then stop
    err = zlsio_RestartGeomCheck ( &config, zoneFile, "Zone");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    }
    
    // check basic OS Compatibility, if fails then stop
    err = zlsio_RestartOSCheck ( &config, zoneFile, "Zone");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    };
    
    //
    // OK to read the conf to find the  zone info
    //
    // Basicly so far dimensions, symmetry and precision + Endian are OK.
    // now we just check consistency of nCells then go ahead..
    //
    
    
    char stateFile[PATHLEN];
    char datFile  [PATHLEN];
    
    sprintf( keyStr, "Zone_Directory" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( zoneDir, str1 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadZoneConf, %s\n", zoneFile);
      zlsio_ErrPrintf( " WARNING: Directory that contains the zones not found.\n" );
      zlsio_ErrPrintf( " WARNING: Using default from level conf %s, may fail.\n", zoneDir);
    };
    
    // clock info
    sprintf( stateFile, "%s%s%s%s%s", zoneDir, PATHSEP, ZONENAME, STATENAME, STATEFILEEXT);
    
    sprintf( keyStr, "Zone_State" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( stateFile, str1 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadZoneConf, %s\n", zoneFile);
      zlsio_ErrPrintf( " WARNING: Zone State File not Specified.\n" );
      zlsio_ErrPrintf( " WARNING: Using default %s, may fail.\n", stateFile);
    };
    
    // zone info
    sprintf( datFile,   "%s%s%s%s%s", zoneDir, PATHSEP, ZONENAME, DATANAME, DATAFILEEXT );
    
    sprintf( keyStr, "Zone_Data" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( datFile, str1 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadZoneConf, %s\n", zoneFile);
      zlsio_ErrPrintf( " WARNING: Zone Data File not Specified.\n" );
      zlsio_ErrPrintf( " WARNING: Using default %s, may fail.\n", datFile);
    };
    
    //
    // only check integer cells numbers, floats can't be checked for equality from a text file.
    // we are assuming the coordinates are set up already and we are just swapping in cell data
    // so only the number of cells matter. The rest of the zone conf entries are for general info, not
    // the actual restart process.
    // We know the dimensions, symmetry and endianness + precision are OK, and that the zone/level type is
    // OK.  Get nVars anyway as it is more convenient than passing it down from level.conf, but really if 
    // it is wrong by now we are hosed to there is no point in being too fussy.  Restart or bust.
    //
    Integer zBCs = 0;
    Counter nti= 1;
    Counter ntj= 1;
    Counter ntk= 1;
    Integer nVars = kNAllVars;  // total number of vars+tracers
    
    if ((str1= ccl_get( &config, "Zone_BCs"))){
      zBCs = strtol( str1, NULL, 0 );
    };
    
    if ((str1= ccl_get( &config, "Zone_nti"))){
      nti = strtol( str1, NULL, 0 );
    };
    
    if ((str1= ccl_get( &config, "Zone_ntj"))){
      ntj = strtol( str1, NULL, 0 );
    };
    
    if ((str1= ccl_get( &config, "Zone_ntk"))){
      ntk = strtol( str1, NULL, 0 );
    };
    
    if ((str1= ccl_get( &config, "Zone_AllVars"))){
      nVars = strtol( str1, NULL, 0 );
    };
    
    
    ccl_release(&config);
    
    ZonePtr z = NULL;
    err  = zls_GetLevelZoneByIndex( &z, level, zone );
    
    if (( err == noErr ) && ( z != NULL )) {
      
      TimeCoords tc;
      zlst_InitTimeCoords  ( &tc );
      
      err = zlsio_ReadReal_StateData( &tc, NULL, NULL,  stateFile ); // don't bother with ios or cpu, it is read by owning level, hence NULL
      
      if ( err != noErr )  {
        zlsio_ErrPrintf( " ERROR  : zlsio_ReadZoneConf, %s, Zone State Data, %s,  Failed.\n", zoneFile, stateFile );
        return(fileErr);
      }
      
      // set the clock and then the data..
      
      z->zoneClock.t      = tc.t;
      z->zoneClock.dt     = tc.dt;
      z->zoneClock.cycles = tc.cycles;
      
      // get the data if the right size
      if (  ( zBCs == z->zoneBCells ) 
          && ( nti == z->zoneGrid.nt[kXDim] ) 
          && ( ntj == z->zoneGrid.nt[kYDim] ) 
          && ( ntk == z->zoneGrid.nt[kZDim] )
          && ( nVars == z->nQ ) ) {
#ifdef MPI
        if ( z->zoneRankID == gMPIrank) { // only read data for zones with data held in current rank
#endif MPI
          
          err = zlsio_ReadReal_ZoneData(  z , datFile );
          
          if ( err != noErr ) {
            zlsio_ErrPrintf( " ERROR  : zlsio_ReadZoneConf: %s. Err = %d\n", datFile, err );
            zlsio_ErrPrintf( " ERROR  : Level %d Zone %d  Reading Zone Data Failed.\n", level, zone );
            err = fileErr;
            return(err);
          }
#ifdef MPI
        }
#endif MPI
        
      } else {
        zlsio_ErrPrintf( " ERROR  : zlsio_ReadZoneConf, %s.\n", zoneFile );
        zlsio_ErrPrintf( " ERROR  : Level %d Zone %d \n", level, zone);
        zlsio_ErrPrintf( " ERROR  : Cell number mismatch %d %d %d %d %d in file.\n", zBCs, nti, ntj, ntk, nVars );
        err = zonErr;
        return(err);
      }
      
    } else { 
      zlsio_ErrPrintf( " ERROR  : zlsio_ReadZoneConf, %s.\n", zoneFile );
      zlsio_ErrPrintf( " ERROR  : Failed to get zone to fill L%2.2d, Z%4.4d .\n", level, zone );
      err = zonErr;
      return(err);
    }
    
  } else  {  err = fileErr; }
  
  return (err );
  
}


int zlsio_ReadLevelConf(char* levelFile, char* resDir, Integer level, Integer extend ); // this also then calls the zone conf files for that level.
int zlsio_ReadLevelConf(char* levelFile, char* resDir, Integer level, Integer extend ){ // this also then calls the zone conf files for that level.
  int err = noErr;
  
  //  
  // if extend == kZLS_RestartExtendSim, then only set main times, leaving counters and events to current setup
  // this allows us to extend a simulation past the end of the old one, but the old one was saved when the end 
  // conditions were satisfied, and that was recorded in the level clocks.  
  // if extend == kZLS_RestartContinueSim, we assume the save was done at an intermediate point in a simulation where then end point is 
  // still wanted so we can just load the whole clock into the level clock in one go without a local temp.
  //
  
  struct ccl_t  config;
  const char     *str1;
  char   keyStr[PATHLEN];
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, levelFile) == noErr) {
    
    // check version, and warn.  If fail then try, but it may not work.
    Integer version = 0;
    if ((str1= ccl_get( &config,"Level_Version"))){
      version = strtol( str1, NULL, 0 );
    }
    if ( version != ZLS_VERSION ) {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadLevelConf, %s, different version, %d\n", levelFile, err);
      zlsio_ErrPrintf( "        : proceeding, but may fail. Please find code with correct version.\n" );
    }
    
    // check basic geometry, if fails then stop
    err = zlsio_RestartGeomCheck ( &config, levelFile, "Level");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    }
    
    // check basic OS Compatibility, if fails then stop
    err = zlsio_RestartOSCheck ( &config, levelFile, "Level");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    };
    
    char probName[80];
    strcpy( probName, "Problem Name Not Found." );
    
    sprintf( keyStr, "Level_ProblemName" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( probName, str1 );
      
    }
    
    if( strcmp(probName, FYRIS_PROBLEMNAME ) != 0){
      // the problem names don't match, likely the data is not consistent with the
      // current simulation, but  it is still possible to try to load a data set into
      // a different problem, so just warn and continue.
      zlsio_ErrPrintf( " WARNING: zlsio_ReadLevelConf, %s,\n", levelFile);
      zlsio_ErrPrintf( " WARNING: Level Problem Name: \"%s\", does not match.\n", probName );
      zlsio_ErrPrintf( " WARNING: current problem: \"%s\"\n", FYRIS_PROBLEMNAME );
      zlsio_ErrPrintf( " WARNING: Restart may fail, or produce invalid results.\n" );
    }
    
    // OK to read the level conf to find the level and zone info
    //
    // Skipping # ZLS Level Configuration Check  as it is too fiddly, from here on pray :)
    // The info in the .conf is for information as much as anything and so a user can read it
    // by hand.  It may come in useful one day so we'll just leave it.  ie, if the level dimensions
    // are not the same as the global dimensions already tested above then we are all screwed anyway
    // and there is no point continuing.  (Will check zone sizes when they are read though)
    
    Integer nZones = 0;
    
    
    char levelDir  [PATHLEN];
    strcpy( levelDir, resDir );
    
    char zoneFile [PATHLEN];
    char zoneDir  [PATHLEN];
    
    char stateFile[PATHLEN];
    
    sprintf( keyStr, "Level_Directory" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( levelDir, str1 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadLevelConf, %s,\n", levelFile);
      zlsio_ErrPrintf( " WARNING: Level Directory that contains the zones not found.\n" );
      zlsio_ErrPrintf( " WARNING: Using default %s, may fail.\n", levelDir);
      ccl_release(&config);
      return(fileErr);
    };
    
    sprintf(stateFile, "%s%s%s%s%s", levelDir, PATHSEP, LEVELNAME, STATENAME, DATAFILEEXT );
    
    sprintf( keyStr, "Level_StateFile" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( stateFile, str1 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_ReadLevelConf, %s\n", levelFile);
      zlsio_ErrPrintf( " WARNING: Level State File not Specified.\n" );
      zlsio_ErrPrintf( " WARNING: Using default %s, may fail.\n", stateFile);
    };
    
    IOStateRec ios;
    TimeCoords tc;
    zlst_InitTimeCoords  ( &tc );
    
    CPUCoords ct;
    zlst_SaveCPUTimes ( &ct, &gZLS_GlobalCPUTimer);
    
    IOSettingsPtr  io = NULL;
    zls_GetLevelIOPrefs( &io, level );
    
    if ( io != NULL) {
      ios = io->ioState; // set default sensible values
      err = zlsio_ReadReal_StateData( &tc,  &ios, &ct, stateFile );
    } else {
      err = zlsio_ReadReal_StateData( &tc,  NULL, NULL, stateFile ); // skip ios if it doesn't work out, minor error
    }
    
    if ( err != noErr ) {
      zlsio_ErrPrintf( " ERROR  : zlsio_ReadLevelConf: %s\n", levelFile);
      zlsio_ErrPrintf( " ERROR  : Level State Data: %s,  Failed.\n", stateFile );
      ccl_release(&config);
      return(fileErr);
    }
    
    if ( level == 0 ) zlst_ResumeCPUTimes(&gZLS_GlobalCPUTimer, &ct);
    
    if ( io != NULL)  io->ioState = ios;
    
    if ( extend == kZLS_RestartContinueSim ) {
      gLevels[level].levelClock.t = tc.t;
      gLevels[level].levelClock.dt = tc.dt;
      gLevels[level].levelClock.cycles = tc.cycles;
      gLevels[level].levelClock.timingCycles = 0;
    };
    
    if ( extend == kZLS_RestartExtendSim ) {
      //
      // need to make sure we are not reading in simulation times that will just terminate the simulation 
      // immediately with the limits already set in init.  Will happen often when people forget to 
      // change the limits in init.conf before restarting from a restart state that was the end of 
      // a simultions, ie not continuing.  If the time in the clock file is later than the current 
      // end time then there is not a lot of point in continuing....
      //
      if ( (gLevels[level].levelClock.endTime <= tc.t ) || ( gLevels[level].levelClock.endCycles <= tc.cycles )) {
        zlsio_ErrPrintf( " WARNING  : zlsio_ReadLevelConf: %s\n", levelFile);
        zlsio_ErrPrintf( " WARNING  : Simulation Extension failed.\n\n" );
        zlsio_ErrPrintf( " WARNING  : The current level clock will\n" );
        zlsio_ErrPrintf( " WARNING  : terminate the simulation immediately.\n" );
        zlsio_ErrPrintf( " WARNING  : The current end time is <= the state time in:  \n" );
        zlsio_ErrPrintf( " WARNING  : %s.\n", stateFile );
        zlsio_ErrPrintf( " WARNING  : Change the init.conf settings and try again.\n\n" );
        
        zlsio_ErrPrintf( "   Current:\n t    %10g  cycles 0x%8.8x\n endTime %10g nCycles 0x%8.8x\n\n",
                        gLevels[level].levelClock.t, (unsigned int)gLevels[level].levelClock.cycles, 
                        gLevels[level].levelClock.endTime, (unsigned int)gLevels[level].levelClock.endCycles );    
        
        zlsio_ErrPrintf( "      File:\n t    %10g  cycles 0x%8.8x\n endTime %10g nCycles 0x%8.8x\n\n",
                        tc.t, (unsigned int)tc.cycles , tc.endTime, (unsigned int)tc.endCycles ); 
        
        return(levErr);
      };
      
      gLevels[level].levelClock.t = tc.t;
      gLevels[level].levelClock.dt = tc.dt;
      gLevels[level].levelClock.cycles = tc.cycles;
      gLevels[level].levelClock.timingCycles = 0;
      
    }; //kZLS_RestartExtendSim
    
    
    
    sprintf( keyStr, "Level_nZones" );
    if ((str1= ccl_get( &config, keyStr))){
      nZones = strtol( str1, NULL, 0 );
    } else {
      zlsio_ErrPrintf( " ERROR  : zlsio_ReadLevelConf, %s.\n", levelFile);
      zlsio_ErrPrintf( " ERROR  : Level Zone Number not Specified.\n" );
      ccl_release(&config);
      return(fileErr);
    };
    
    ccl_release(&config); // release so ccl is free to read zones
    
    if ( nZones  == gLevels[level].nZones ) {
      
      Integer zone = 0;
      
      for( zone = 0; zone < nZones; zone++ ){
        
        sprintf( zoneDir, "%s%s%sL%2.2dZ%4.4d", levelDir, PATHSEP, ZONENAME, level, zone);
        sprintf( zoneFile, "%s%s", zoneDir, CONFEXT );
        err = zlsio_ReadZoneConf( zoneFile,  zoneDir, level,  zone , extend);
        if ( err != noErr ) {
          return(err);
        } 
        
      } // zone
      
    } else { 
      zlsio_ErrPrintf( " ERROR  : zlsio_ReadLevelConf, %s.\n", levelFile);
      zlsio_ErrPrintf( " ERROR  : Level: %d, Zone Number: %d.\n",level, nZones );
      zlsio_ErrPrintf( " ERROR  : Not compatible with current Zone Number: %d\n", gLevels[level].nZones );
      err = zonErr; 
    }
    
    
  } else  {  err = fileErr; }
  
  return (err );
  
}

int zlsio_RestoreState   ( char * restartFile ){
  // file name is full .conf file name, the directory name is read form the file
  // This is different to Save state where the file name is without the  .conf.
  int err = noErr;
  
  struct ccl_t  config;
  const char     *str1;
  char   keyStr[PATHLEN];
  
  zlsio_Printf ( "\n%s\n", DIVIDERLINE );
  zlsio_Printf ( " ZLS Restarting from %s \n", restartFile);
  zlsio_Printf ( "%s\n",  DIVIDERLINE );
  
  config.comment_char = '#';
  config.sep_char     = '=';
  config.str_char     = '"';
  
  if (ccl_parse(&config, restartFile) == noErr) {
    zlsio_Printf ( "\n%s\n", DIVIDERLINE );
    zlsio_Printf ( " ZLS Restarting from %s \n", restartFile);
    zlsio_Printf ( "%s\n",  DIVIDERLINE );
    // 1st. Check the restart conf to confirm the basic format is OK and compatible with 
    // the currently compiled options.
    
    // check version, and warn.  If fail then try, but it may not work.
    Integer version = 0;
    if ((str1= ccl_get( &config,"Restart_Version"))){
      version = strtol( str1, NULL, 0 );
    }
    if ( version != ZLS_VERSION ) {
      zlsio_ErrPrintf( " WARNING: zlsio_Restart, %s, different version, %d\n", restartFile, err);
      zlsio_ErrPrintf( "        : proceeding, but may fail. Please find code with correct version.\n" );
    }
    
    // check basic geometry, if fails then stop
    err = zlsio_RestartGeomCheck ( &config, restartFile, "Restart");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    }
    
    // check basic OS Compatibility, if fails then stop
    err = zlsio_RestartOSCheck ( &config, restartFile, "Restart");
    if ( err != noErr ){
      ccl_release(&config);
      return( err );
    }
    
    
    Integer extend =   kZLS_RestartContinueSim;
    
    sprintf( keyStr, "Restart_Continuing" );
    if ((str1= ccl_get( &config, keyStr))){
      
      if (strcmp(str1, "NO")==0) {
        extend = kZLS_RestartExtendSim;
      }
      
      if (strcmp(str1, "No")==0) {
        extend = kZLS_RestartExtendSim;
      }
      
      if (strcmp(str1, "no")==0) {
        extend = kZLS_RestartExtendSim;
      }
      
      if (strcmp(str1, "n")==0) {
        extend = kZLS_RestartExtendSim;
      }
      
      if (strcmp(str1, "N")==0) {
        extend = kZLS_RestartExtendSim;
      }
      
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_Restart, %s.\n", restartFile);
      zlsio_ErrPrintf( " Continuing State not specified, assuming continuing = NO.\n" );
      extend = kZLS_RestartExtendSim;
    }
    
    char resDir[PATHLEN];
    char levelFile[PATHLEN];
    
    sprintf( keyStr, "Restart_Directory" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( resDir,  str1 );
    } else {
      zlsio_ErrPrintf( " ERROR: zlsio_Restart, %s.\n", restartFile);
      zlsio_ErrPrintf( " Restart Directory that contains the levels and zones not found.\n" );
      ccl_release(&config);
      return(fileErr);
    }
    
    Integer nLevels = gNLevels;
    sprintf( keyStr, "Restart_NLevels" );
    if ((str1= ccl_get( &config, keyStr))){
      nLevels = strtol( str1, NULL, 0 );
    } else {
      zlsio_ErrPrintf( " WARNING: zlsio_Restart, %s, NLevels not found using current: %d - may fail Horribly!.\n", restartFile, nLevels);
    }
    
    char probName[80];
    strcpy( probName, "Problem Name Not Found." );
    
    sprintf( keyStr, "Restart_ProblemName" );
    if ((str1= ccl_get( &config, keyStr))){
      strcpy( probName, str1 );
    }
    
    if( strcmp(probName, FYRIS_PROBLEMNAME ) != 0){
      // the problem names don't match, likely the data is not consistent with the
      // current simulation, but  it is still possible to try to load a data set into
      // a different problem, so just warn and continue.
      zlsio_ErrPrintf( " WARNING: zlsio_Restart, %s,\n", levelFile);
      zlsio_ErrPrintf( " WARNING: Restart Problem Name: \"%s\", does not match.\n", probName );
      zlsio_ErrPrintf( " WARNING: current problem: \"%s\".\n", FYRIS_PROBLEMNAME );
      zlsio_ErrPrintf( " WARNING: Restart may fail, or produce invalid results.\n" );
    }
    
    // release restart conf as soon as done with.  Don't know that ccl is reentrant or not.
    // so we are free to open new conf files as we go.
    
    ccl_release(&config);
    
    if ( nLevels == gNLevels ) { // OK to proceed, structures are allocated ready to receive new data
      
      Integer level = 0;
      
      for ( level = 0; level < nLevels; level ++ ) {
        
        sprintf( levelFile, "%s%s%s%2.2d%s", resDir, PATHSEP, LEVELNAME, level, LEVELFILEEXT );
        
        err = zlsio_ReadLevelConf( levelFile, resDir, level, extend ); // this also then calls the zone conf files for that level.
        
        if ( err != noErr ) {
          zlsio_ErrPrintf( " IO ERROR: zlsio_Restart FAILED.\n" );
          zlsio_ErrPrintf( " Reading level config file, %s,  Err = %d \n\n", levelFile, err);
          return ( err );
        }         
        
      } // level
      
    } else  { // nLevels != gNLevels, structures are not present to receive data, no way to proceed
      
      err = zonErr;
      zlsio_ErrPrintf( " ERROR: zlsio_Restart, %s, NLevels, %d.\n", restartFile, nLevels );
      zlsio_ErrPrintf( " ERROR: Not equal to Current %d, Cannot continue.\n", gNLevels );
      
    }
    
    
  } else { 
    err = fileErr;
    zlsio_ErrPrintf( "ERROR: zlsio_Restart, %s failed to open properly, %d\n", restartFile, err);
  }
  
  if ( err == noErr ) {
    zlsio_Printf ("\n%s\n", DIVIDERLINE );
    zlsio_Printf (" ZLS Restarting from %s OK, proceeding from: \n", restartFile );
    zlsio_Printf (" Cycle : %6d Time : %10.4g = %10.4g %s\n",
                  zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
    zlsio_Printf ("%s\n", DIVIDERLINE );
    zlsio_DisplaySimLimits( stdout , restartFile, 0 );
  };
  
  return (err);
}

int zlsio_WriteRestartGeomCheck ( FILE * f , char * prefix);
int zlsio_WriteRestartGeomCheck ( FILE * f , char * prefix){
  
  int err = noErr;
  
#ifdef SYM1D 
  zlsio_FPrintf( f, "%s_Dimensions = %d\n", prefix, 1 );
#endif SYM1D
  
#ifdef SYM2D 
  zlsio_FPrintf( f, "%s_Dimensions = %d\n" , prefix, 2 );
#endif SYM2D
  
#ifdef SYM3D 
  zlsio_FPrintf( f, "%s_Dimensions = %d\n" , prefix, 3 );
#endif SYM3D
  
#ifdef CARTESIAN 
  zlsio_FPrintf( f, "%s_Geometry = %s \n" , prefix, "Cartesian");
#endif CARTESIAN
#ifdef CIRCULAR 
  zlsio_FPrintf( f, "%s_Geometry = %s \n" , prefix, "Circular");
#endif CIRCULAR
#ifdef SPHERICAL 
  zlsio_FPrintf( f, "%s_Geometry = %s \n" , prefix, "Spherical");
#endif SPHERICAL
#ifdef POLARSPHERICAL 
  zlsio_FPrintf( f, "%s_Geometry = %s \n" , prefix, "PolarSpherical");
#endif POLARSPHERICAL
#ifdef CYLINDRICAL 
  zlsio_FPrintf( f, "%s_Geometry = %s \n" , prefix, "Cylindrical");
#endif CYLINDRICAL
  
  return(err);
  
}


int zlsio_WriteRestartPhysCheck( FILE * f , char * prefix);
int zlsio_WriteRestartPhysCheck( FILE * f , char * prefix){
  
  int err = noErr;
  
#ifdef HYDRO 
  zlsio_FPrintf( f, "%s_Hydrodynamics = ON\n" , prefix );
#else
  zlsio_FPrintf( f, "%s_Hydrodynamics = OFF\n" , prefix );
#endif HYDRO
    
#ifdef PRIMITIVE_QUAS 
  zlsio_FPrintf( f, "%s_Primitive_Vars= ON\n" , prefix );
#else
  zlsio_FPrintf( f, "%s_Primitive_Vars= OFF\n" , prefix );
#endif PRIMITIVE_QUAS
  
#ifdef CONSERVED_QUAS 
  zlsio_FPrintf( f, "%s_Conserved_Vars= ON\n" , prefix );
#else
  zlsio_FPrintf( f, "%s_Conserved_Vars= OFF\n" , prefix );
#endif CONSERVED_QUAS
  
#ifdef GRAVITY 
  zlsio_FPrintf( f, "%s_Gravity       = ON\n" , prefix );
#else
  zlsio_FPrintf( f, "%s_Gravity       = OFF\n" , prefix );
#endif GRAVITY
  
#ifdef VARIABLE_EOS 
  zlsio_FPrintf( f, "%s_Variable_EOS  = ON\n" , prefix );
#else
  zlsio_FPrintf( f, "%s_Variable_EOS  = OFF\n" , prefix );
#endif VARIABLE_EOS
    
  zlsio_FPrintf( f, "%s_VarType     =  %d \n" , prefix, kAlgorithmZoneType );
  zlsio_FPrintf( f, "%s_nVars       =  %d \n" , prefix, kNVars    );
  zlsio_FPrintf( f, "%s_nTracers    =  %d \n" , prefix, kNTracers );
  zlsio_FPrintf( f, "%s_TotalVars   =  %d \n" , prefix, kNAllVars );
  
  return(err);
  
}


int zlsio_WriteRestartOSCheck( FILE * f , char * prefix);
int zlsio_WriteRestartOSCheck( FILE * f , char * prefix){
  
  int err = noErr;
  
#ifdef SINGLE_PRECISION 
  zlsio_FPrintf( f, "%s_Precision =  Single \n" , prefix);
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION 
  zlsio_FPrintf( f, "%s_Precision =  Double \n" , prefix);
#endif DOUBLE_PRECISION
  
#ifdef DATA_BIGENDIAN_OS 
  zlsio_FPrintf( f, "%s_DataEndian =  BigEndian \n", prefix);
#endif DATA_BIGENDIAN_OS
#ifdef DATA_LITENDIAN_OS 
  zlsio_FPrintf( f, "%s_DataEndian =  LittleEndian \n" , prefix);
#endif DATA_LITENDIAN_OS
  
#ifndef DATA_BIGENDIAN_OS 
#ifndef DATA_LITENDIAN_OS 
  zlsio_FPrintf( f, "%s_DataEndian =  Unknown \n" , prefix);
#endif DATA_LITENDIAN_OS
#endif DATA_BIGENDIAN_OS
  
  return(err);
  
}

int  zlsio_WriteRestartConf( FILE * f , char * resDir, Integer extend ) ;
int  zlsio_WriteRestartConf( FILE * f , char * resDir, Integer extend ) {
  int err = noErr;
  
  Integer nLevels = gNLevels;
  
  time_t writeTime;
  writeTime = time(NULL);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS PPM-LR Hydro Simulation Restart Configuration \n" );
  err = zlsio_PrintConfiguration( f , "#" );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# Written: %s", ctime(&writeTime));
  zlsio_FPrintf( f, "# Cycle : %6d Time : %12.5g = %12.5g %s\n",
                zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Configuration Check \n" );
  zlsio_FPrintf( f, "Restart_ZLSMAGIC    = 0x%8.8x \n", ZLS_MAGICTAG );
  zlsio_FPrintf( f, "Restart_Version     = 0x%8.8x \n", ZLS_VERSION );
  if ( extend == kZLS_RestartContinueSim) {
    zlsio_FPrintf( f, "Restart_Continuing  = YES \n" );
  } else {
    zlsio_FPrintf( f, "Restart_Continuing  = NO  \n" );
  }
  zlsio_FPrintf( f, "Restart_Directory   = \"%s\" \n", resDir );
  zlsio_FPrintf( f, "Restart_NLevels     = %d     \n", nLevels );
  zlsio_FPrintf( f, "Restart_ProblemName  = \"%s\" \n",  FYRIS_PROBLEMNAME );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Geometry Check \n" );
  err = zlsio_WriteRestartGeomCheck ( f , "Restart");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Physics/Variables Check \n" );
  err = zlsio_WriteRestartPhysCheck ( f , "Restart");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS OS Check \n" );
  err = zlsio_WriteRestartOSCheck   ( f , "Restart");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  
  return(err);
}


int  zlsio_WriteLevelConf( FILE * f , char * levelDir, char * stateFile , Integer level);
int  zlsio_WriteLevelConf( FILE * f , char * levelDir, char * stateFile , Integer level) {
  int err = noErr;
  time_t writeTime;
  
  Integer nZones = gLevels[level].nZones;
  
  writeTime = time(NULL);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS PPM-LR Hydro Simulation Level Restart Configuration \n" );
  err = zlsio_PrintConfiguration( f , "#" );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# Written: %s", ctime(&writeTime));
  zlsio_FPrintf( f, "# Cycle : %6d Time : %12.5g = %12.5g %s\n",
                zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "Level_Version      = 0x%8.8x \n", ZLS_VERSION );
  zlsio_FPrintf( f, "Level_Directory    = \"%s\" \n", levelDir );
  zlsio_FPrintf( f, "Level_StateFile    = \"%s\" \n", stateFile );
  zlsio_FPrintf( f, "Level_nZones       = %d \n", nZones );
  zlsio_FPrintf( f, "Level_ZoneConfBaseName  = %s \n", ZONENAME );
  zlsio_FPrintf( f, "Level_ProblemName  = \"%s\" \n",  FYRIS_PROBLEMNAME );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Level Configuration Check \n" );
  zlsio_FPrintf( f, "Level_Level          = %d \n", gLevels[level].level );
  zlsio_FPrintf( f, "Level_maxZones       = %d \n", gLevels[level].maxZones );
  zlsio_FPrintf( f, "Level_levelNCells    = %g \n", gLevels[level].levelNCells );
  zlsio_FPrintf( f, "Level_levelZoneKind  = %d \n", gLevels[level].levelZoneKind );
  zlsio_FPrintf( f, "Level_levelDims      = %d \n", gLevels[level].levelDims );
  zlsio_FPrintf( f, "Level_levelRefine    = %d \n", gLevels[level].levelRefinement );
  zlsio_FPrintf( f, "Level_TotalRefine    = %d \n", gLevels[level].levelTotalRefinement );
  zlsio_FPrintf( f, "Level_levelBCs       = %d \n", gLevels[level].levelBCells );
  zlsio_FPrintf( f, "Level_SuperBCs       = %d \n", gLevels[level].levelSuperBCells );
  zlsio_FPrintf( f, "Level_Interpolation  = %d \n", gLevels[level].levelInterpolation );
  
  zlsio_FPrintf( f, "Level_Time           = %g \n", gLevels[level].levelClock.t );
  zlsio_FPrintf( f, "Level_dt             = %g \n", gLevels[level].levelClock.dt );
  zlsio_FPrintf( f, "Level_Cycles         = %d \n", (int)gLevels[level].levelClock.cycles );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Further Configuration Checks \n" );
  err = zlsio_WriteRestartOSCheck   ( f ,"Level");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  err = zlsio_WriteRestartGeomCheck ( f ,"Level");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Physics/Variables Check \n" );
  err = zlsio_WriteRestartPhysCheck ( f ,"Level");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  
  return(err);
}


int  zlsio_WriteZoneConf( FILE * f , char * zonDir , char * zonFile , char * stateFile , ZonePtr z);
int  zlsio_WriteZoneConf( FILE * f , char * zonDir , char * zonFile , char * stateFile , ZonePtr z) {
  int err = noErr;
  time_t writeTime;
  
  writeTime = time(NULL);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS PPM-LR Hydro Simulation Zone Restart Configuration \n" );
  err = zlsio_PrintConfiguration( f , "#" );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# Written: %s", ctime(&writeTime));
  zlsio_FPrintf( f, "# Cycle : %6d Time : %12.5g = %12.5g %s\n",
                zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "Zone_Version   = 0x%8.8x \n", ZLS_VERSION );
  zlsio_FPrintf( f, "Zone_Directory = \"%s\" \n", zonDir );
  zlsio_FPrintf( f, "Zone_Data      = \"%s\" \n", zonFile );
  zlsio_FPrintf( f, "Zone_State     = \"%s\" \n", stateFile );
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Configuration Check \n" );
  zlsio_FPrintf( f, "Zone_Level  = %d \n", z->zoneLevel );
  zlsio_FPrintf( f, "Zone_ID     = %d \n", z->zoneID );
  zlsio_FPrintf( f, "Zone_Dims   = %d \n", z->zoneGrid.nD );
  zlsio_FPrintf( f, "Zone_Refine = %d \n", z->zoneRefine);
  zlsio_FPrintf( f, "Zone_BCs    = %d \n", z->zoneBCells);
  
  zlsio_FPrintf( f, "Zone_nti    = %d \n", z->zoneGrid.nt[kXDim] );
  zlsio_FPrintf( f, "Zone_ntj    = %d \n", z->zoneGrid.nt[kYDim] );
  zlsio_FPrintf( f, "Zone_ntk    = %d \n", z->zoneGrid.nt[kZDim] );
  zlsio_FPrintf( f, "Zone_X0     = %g \n", z->zoneGrid.xl[kXDim] );
  zlsio_FPrintf( f, "Zone_X1     = %g \n", z->zoneGrid.xr[kXDim] );
  zlsio_FPrintf( f, "Zone_Y0     = %g \n", z->zoneGrid.xl[kYDim] );
  zlsio_FPrintf( f, "Zone_Y1     = %g \n", z->zoneGrid.xr[kYDim] );
  zlsio_FPrintf( f, "Zone_Z0     = %g \n", z->zoneGrid.xl[kZDim] );
  zlsio_FPrintf( f, "Zone_Z1     = %g \n", z->zoneGrid.xr[kZDim] );
  
  zlsio_FPrintf( f, "Zone_Type   = %d \n", z->zoneType);
  zlsio_FPrintf( f, "Zone_AllVars= %d \n", z->nQ);
  zlsio_FPrintf( f, "Zone_Cells  = %g \n", z->zoneNCells);
  
  zlsio_FPrintf( f, "Zone_CFL    = %g \n", z->zoneCFL);
  zlsio_FPrintf( f, "Zone_Time   = %g \n", z->zoneClock.t);
  zlsio_FPrintf( f, "Zone_dt     = %g \n", z->zoneClock.dt);
  zlsio_FPrintf( f, "Zone_cycles = %d \n", (int)z->zoneClock.cycles);
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Further Configuration Checks \n" );
  err = zlsio_WriteRestartOSCheck   ( f ,"Zone");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  err = zlsio_WriteRestartGeomCheck ( f ,"Zone");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  zlsio_FPrintf( f, "# ZLS Physics/Variables Check \n" );
  err = zlsio_WriteRestartPhysCheck ( f ,"Zone");
  zlsio_FPrintf( f, "#%s\n", DIVIDERLINE );
  
  return(err);
}


int zlsio_SaveState ( char * restartFile ){ 
  // file name is just the base name, the .conf is added and the directory name is derived as well
  // This is different to Restore state where the file name is the conf file.
  
  int err = noErr;
  
  FILE * restartConfig  = NULL;
  FILE * levelConfig  = NULL;
  FILE * zoneConfig  = NULL;
  char resDir[PATHLEN];
  char resFile[PATHLEN];
  
  IOStateRec ioS0;
  IOSettingsPtr  io = NULL;
  
  zls_GetLevelIOPrefs( &io, 0 ); // level 0 ios
  if ( io != NULL) { ioS0 = io->ioState;} else { return(memErr); }
  
  if (strcmp(io->restartDir, ".")==0){
    sprintf(resDir , "%s%3.3d", restartFile, (int)ioS0.ioSaveCount );
  } else {
    sprintf(resDir , "%s%s%s%3.3d", io->restartDir, PATHSEP, restartFile, (int)ioS0.ioSaveCount );
  }
  sprintf(resFile, "%s%s"     , resDir, CONFEXT );
  io->ioState.ioSaveCount++;
  
#ifdef MPI
  if ( gMPIrank == 0) { // rank 0 process creates restart conf and restart dir
#endif MPI    
    restartConfig = fopen(resFile, "w");
    if (restartConfig) {
      
      // configure default save states to be continuing sims that have not finished
      // saved sate continues to have valid intervals and ending times, which are loaded
      
      Integer extend  = kZLS_RestartContinueSim; 
      
      // if the sim has actually ended then it will need to be a restart that extends the saved 
      // simulation limits, the saved limits are not reloaded.
      //
      if (  gLevels[0].levelClock.cycles >=  gLevels[0].levelClock.endCycles )
        extend  = kZLS_RestartExtendSim; 
      
      if (  gLevels[0].levelClock.t >=  gLevels[0].levelClock.endTime )
        extend  = kZLS_RestartExtendSim; 
      
      err = zlsio_WriteRestartConf( restartConfig , resDir, extend );
      
      fclose(restartConfig);
    } else { 
      err = fileErr; 
      zlsio_ErrPrintf( " IO ERROR: zlsio_SaveState Failed, Err = %d \n", err);
      return (err);
    }
    
    zls_mkdir( resDir );
#ifdef MPI
  }
  
  MPI_Barrier(MPI_COMM_WORLD); // wait for everyone , so output dir etc exist for everyone
#endif MPI    
  
  Integer nLevels = gNLevels;
  Integer level = 0;
  
  char levelFile[PATHLEN];
  char levelDir[PATHLEN];
  char zoneFile[PATHLEN];
  char zoneDir[PATHLEN];
  
  char stateFile[PATHLEN];
  char datFile[PATHLEN];
  
  CPUCoords ct;
  zlst_SaveCPUTimes ( &ct, &gZLS_GlobalCPUTimer);
  
  for ( level = 0; level < nLevels; level++) {
    
    sprintf(levelDir,  "%s%s%s%2.2d", resDir, PATHSEP , LEVELNAME, level );
    sprintf(levelFile, "%s%s"       , levelDir, CONFEXT);
    sprintf(stateFile, "%s%s%s%s%s" , levelDir, PATHSEP , LEVELNAME, STATENAME, DATAFILEEXT );
    
#ifdef MPI
    if ( gMPIrank == 0) { // rank 0 process creates level conf and level dir
#endif MPI    
      
      levelConfig = fopen(levelFile, "w");
      if (levelConfig) {
        
        err = zlsio_WriteLevelConf( levelConfig , levelDir , stateFile, level);
        
        fclose(levelConfig);
      } else { 
        err = fileErr; 
        zlsio_ErrPrintf(" IO ERROR: zlsio_SaveState Failed, %s, Err = %d \n", levelFile, err);
        return (err);
      }
      
      zls_mkdir(levelDir);
#ifdef MPI
    }
    MPI_Barrier(MPI_COMM_WORLD); // wait for everyone , so level dir and conf exist for everyone
#endif MPI    
    
    // save time and io state vars in level clock file
    TimeCoords tc = gLevels[level].levelClock;
    
    
    IOStateRec ios;
    io = NULL;
    zls_GetLevelIOPrefs( &io, level );
    if ( io != NULL) { ios = io->ioState; } else { return(memErr);}
    
    err = zlsio_DumpReal_StateData  ( stateFile, &tc, &ios, &ct );
    
    Integer nZones = gLevels[level].nZones;
    Integer zone = 0;
    ZonePtr z = NULL;
    
    for( zone = 0; zone < nZones; zone++ ){
      
      sprintf( zoneDir,   "%s%s%sL%2.2dZ%4.4d", levelDir, PATHSEP, ZONENAME, level, zone);
      sprintf( zoneFile,  "%s%s"            , zoneDir, ZONEFILEEXT );
      sprintf( datFile,   "%s%s%s%s%s"      , zoneDir, PATHSEP, ZONENAME, DATANAME, DATAFILEEXT );
      sprintf( stateFile, "%s%s%s%s%s"      , zoneDir, PATHSEP, ZONENAME, STATENAME, STATEFILEEXT);
      
      err = zls_GetLevelZoneByIndex ( &z, level, zone );
      if (( err == noErr ) && ( z!= NULL)) {
        
#ifdef MPI
        if ( z->zoneRankID == gMPIrank) { //each rank process writes it's own zones that have data
#endif MPI    
          
          zoneConfig = fopen( zoneFile, "w");
          if (zoneConfig) {
            
            err = zlsio_WriteZoneConf( zoneConfig , zoneDir, datFile, stateFile, z );
            
            fclose(zoneConfig);
          } else { 
            err = fileErr; 
            zlsio_ErrPrintf( " IO ERROR: zlsio_SaveState Failed, %s, Err = %d \n", zoneFile, err);
            return (err);
          }
          
          err = zls_mkdir( zoneDir );
          
          err = zlsio_DumpReal_ZoneData  ( datFile, z );
          
          // zone clocks don't need iostate vars, but can add them anyway so we make them more uniform.
          err = zlsio_DumpReal_StateData  ( stateFile, &z->zoneClock,  &ios,  &ct);
#ifdef MPI
        } //each rank process writes it's own zones that have data, so no barrier needed
#endif MPI    
        
      } else {err = zonErr;}
      
    } // zone
    
  } // level
  
  
  return( err );
}
