#ifndef ZLS_RESTART_H
#define ZLS_RESTART_H
#include "zls.h"

enum {  kZLS_RestartContinueSim = 0,  // continue a simulation that has not finished yet - read all of clock data in
  kZLS_RestartExtendSim   = 1 // continue a simulation that has ended, but wish to extend with new
  // setup time limits - read just current times and cycles, leaving event timers set by the startup.
  // so that new end times and so on can be set.
};

enum{  
  kZLS_BadFormat = 0,
  kZLS_OKFormat  = 1,
  kZLS_SWAPFormat  = 2,
  kZLS_NOSWAPFormat  = 3
};

enum{  
  kZLS_NoSkip = 0
};


//
// All header fields are 4 bytes, and stay 4 byte alligned
// for 64 bit file size compatibility, the dimensions and sample size
// are documented, so the reader can compute the data size, in 64 bit if
// the OS supports it.  In practise readers and writers should buffer 
// IO as needed.
//
// zonedata.zdat format is header followed by nVars QPtrs dumped in order 0.. kNAllVars-1
// intended for restart dumps, 16 bit scaled output data for analysis are created by other
// routines.
// The header is 32 format for compatibility, and has one small 32 bit offset to start of data
// which can then be read as a stream.
// try to pad the header to 8 byte boundaries to make reading easier
//
typedef struct {
  
  unsigned int ZLS_MagicTag;  // magic number 'ZLS!' to idenify zls zonedata bin files
  
  unsigned int ZLS_DataOffset;   // offset to beginning of data from start of file
  unsigned int ZLS_Version;   // 4 byte int version 
  unsigned int ZLS_IDTag;     // 'ZNID' 
  unsigned int ZLS_ZoneID;    // 4 byte int zoneID field 
  
  unsigned int ZLS_VarsTag;   
  unsigned int ZLS_VarsType;  // Variable type:                
  unsigned int ZLS_nVarTag;   
  unsigned int ZLS_nVars;     // total number of vars, incl tracers
  
  unsigned int ZLS_DimsTag;   // 'DIMS' tag
  unsigned int ZLS_Dims;      // 1, 2 or 3, any other value is an error
  unsigned int ZLS_XDimTag;      // 'XDIM' tag
  unsigned int ZLS_Xdim;      // int, >= 1, all dims include BCs, ie = nt not n
  unsigned int ZLS_YDimTag;      // 'YDIM' tag
  unsigned int ZLS_Ydim;      // int, >= 1, 1 if 1D,  > 1 if 2D or 3D
  unsigned int ZLS_ZDimTag;      // 'ZDIM' tag
  unsigned int ZLS_Zdim;      // int, >= 1, 1 if 1D or 2D,  > 1 if 3D
  unsigned int ZLS_NGSTTag;   // 'NGST' tag
  unsigned int ZLS_NGST;      // int, 0, 2, 4 etc, number of BC cells 
  // in data,to aid interpretation
  
  unsigned int ZLS_SampleTag;    // 'SMPL'
  unsigned int ZLS_SampleSize;   //  zone data element size in bytes
  
} zoneDataHeader;

//
// Header size is constant for a given version, ie all dims are written even if some are not
// used in lower dimensional data
//
#define ZLS_ZONEDATA_HEADERSIZE  (sizeof(zoneDataHeader))
#define ZLS_ZONEDATA_IOCHUNKSIZE (0x00100000)


typedef struct {
  
  unsigned int ZLS_MagicTag;    // magic number 'ZLS!' to idenify zls zonedata bin files
  unsigned int ZLS_DataOffset;  //  offset to data from start of file
  unsigned int ZLS_VersTag;     // 'VERS' tag
  unsigned int ZLS_Version;     // 4 byte int version 
  
} stateDataHeader;

#define ZLS_STATEDATA_HEADERSIZE (sizeof(stateDataHeader))


int zlsio_SaveZoneRealData      ( ZonePtr z, IOSettingsPtr io);
int zlsio_SaveLevelZoneRealData ( Integer level, int step);

int zlsio_RestoreState   ( char * restartFile );
int zlsio_SaveState      ( char * restartFile );

#endif ZLS_RESTART_H
