#include "zls.h"

#ifndef TEXTOUTPUT
#define TEXTOUTPUT(f, z) zlsio_TextZoneOutput(f, z)
#endif


int zlsio_SaveLevelUserIO(Integer level, Counter step){
#pragma unused( step )
  //
  int       err = noErr;
  
  char outFile[PATHLEN];
  zFile      f   = NULL;
  time_t      writeTime;
  
  Integer          zone;
  ZonePtr             z;
  IOSettingsPtr  io = NULL;
  
  err = zls_GetLevelIOPrefs( &io, level );
  
  if (io != NULL) {
  
//
// loop over zones separately
//

    for( zone = 0; zone <gLevels[level].nZones; zone++){
  
            err = zls_GetLevelZoneByIndex( &z, level, zone );
            
            if ( err != noErr ){
              zlsio_ErrPrintf( " IO ERROR: zls_GetLevelZoneByIndex Failed, Index = %d Err = %d \n", zone , err);
              err = ioErr;
              return (err);
            }

#ifdef MPI  
            if (z->zoneRankID == gMPIrank) {
#endif MPI
              
#ifdef ZLIB_COMPRESSED_IO
              sprintf(outFile, "%s%s%s%s%sL%2.2dZ%2.2dC%4.4d%s.gz", io->outDir, PATHSEP, USERDIR, PATHSEP,
                      io->outName, level, zone, (int)io->ioState.ioUserCount, USEREXT);
#else
              sprintf(outFile, "%s%s%s%s%sL%2.2dZ%2.2dC%4.4d%s", io->outDir, PATHSEP, USERDIR, PATHSEP,
                      io->outName, level, zone, (int)io->ioState.ioUserCount, USEREXT);
#endif !ZLIB_COMPRESSED_IO
              
              f = zlsio_open(outFile, "wa");
              if (f) {
                
                writeTime = time(NULL);
                zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
                zlsio_ZFPrintf( f, " %s Simulation TEXT Output\n", FYRIS );
                zlsio_ZFPrintf( f, " Level = %d, Zone = %d\n", level, zone);
                
#if (ZLS_VERBOSE > 1)
                zlsp_ZPrintPhysicsConfiguration ( f, NULL );
#endif 
                zlsio_ZFPrintf( f, " Written: %s", ctime(&writeTime));
                zlsio_ZFPrintf( f, " Cycle : %11d Time : %#12.6g = %#12.6g %s\n",
                                zlst_GlobalCycles,
                                zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
                zlsio_ZFPrintf( f, "%s\n\n", DIVIDERLINE );
                
                
#if (ZLS_DEBUG >0)
#ifdef HYDRO 
#ifdef COOLING
                zlsio_ZFPrintf( f, " Scale Factors:\n" );
                zlsio_ZFPrintf( f, "      t_0            X_0            V_0             d_0            P_0          Cool_0 \n" );
                zlsio_ZFPrintf( f, " %#12.6g    %#12.6g   %#12.6g    %#12.6g    %#12.6g    %#12.6g  \n\n", 
                                f_t0, f_x0, f_v0, f_d0, f_P0, f_cool0);
#else 
                zlsio_ZFPrintf( f, " Scale Factors:\n" );
                zlsio_ZFPrintf( f, "      t_0            X_0            V_0             d_0            P_0 \n" );
                zlsio_ZFPrintf( f, " %#12.6g    %#12.6g   %#12.6g    %#12.6g    %#12.6g\n\n", 
                                f_t0, f_x0, f_v0, f_d0, f_P0);
#endif !COOLING
#endif HYDRO 

#endif (ZLS_DEBUG >0)
                
                zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
                
                err = TEXTOUTPUT(f, z);
                if ( err != noErr ){
                  zlsio_ErrPrintf( " IO WARNING: zlsio_TextZoneOutput Failed, File: %s Err = %d \n",outFile, err );
                  err = ioErr;
                  zlsio_close(f);
                  return (err);
                }
            
            
                zlsio_ZFPrintf( f, "%s\n\n", DIVIDERLINE );
               
              } // f
              zlsio_close(f);
              
#ifdef MPI
            } // MPI Rank
#endif MPI
            
          } // zone loop
                    
#ifdef ASTROBLAST
 // entire level analysis output
    zlsp_IntegrateBlast   ( level );
    
#ifdef MPI
    if ( gMPIrank == level){
#endif MPI
      
#ifdef ZLIB_COMPRESSED_IO
    sprintf(outFile, "%s%s%s%sLevel%sL%2.2dC%4.4d%s.gz", io->outDir, PATHSEP, USERDIR, PATHSEP,
            io->outName, level, (int)io->ioState.ioUserCount, USEREXT);
#else
    sprintf(outFile, "%s%s%s%sLevel%sL%2.2dC%4.4d%s", io->outDir, PATHSEP, USERDIR, PATHSEP,
            io->outName, level, (int)io->ioState.ioUserCount, USEREXT);
#endif !ZLIB_COMPRESSED_IO
    
    f = zlsio_open(outFile, "wa");
    if (f) {
      zlsio_Printf( " zlsio_SaveLevelUserIO, Level = %d Step = %d File = %s\n", level , step, outFile);

      writeTime = time(NULL);
      zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
#ifdef HYDRO
      zlsio_ZFPrintf( f, " FYRIS/ZLS PPM-LR Hydro Simulation TEXT Output\n" );
#endif HYDRO
      zlsio_ZFPrintf( f, " Level = %d, nZones = %d\n", level, gLevels[level].nZones);
      zlsio_Printf( " Level = %d, nZones = %d\n", level, gLevels[level].nZones);
#if (ZLS_VERBOSE > 1)
      zlsp_ZPrintPhysicsConfiguration ( f, NULL );
#endif 
      zlsio_ZFPrintf( f, " Written: %s", ctime(&writeTime));
      zlsio_ZFPrintf( f, " Cycle : %11d Time : %#12.6g = %#12.6g %s\n",
                     zlst_GlobalCycles,
                     zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
      zlsio_ZFPrintf( f, "%s\n\n", DIVIDERLINE );
      
      zlsp_TextLevelBlastAnalysis( f, level);
      
    } // f
    zlsio_close(f);
    
    
#ifdef MPI
    }
#endif MPI
#endif ASTROBLAST

#ifdef DWARFGALAXY_ANALYSIS
    // entire level analysis output
    zlsp_SumLevelEnergy ( level);
    zlsp_SumLevelColdGas( level);
    
#ifdef MPI
    if ( gMPIrank == level){
#endif MPI
      
#ifdef ZLIB_COMPRESSED_IO
      sprintf(outFile, "%s%s%s%sLevel%sL%2.2dC%4.4d%s.gz", io->outDir, PATHSEP, USERDIR, PATHSEP,
              io->outName, level, (int)io->ioState.ioUserCount, USEREXT);
#else
      sprintf(outFile, "%s%s%s%sLevel%sL%2.2dC%4.4d%s", io->outDir, PATHSEP, USERDIR, PATHSEP,
              io->outName, level, (int)io->ioState.ioUserCount, USEREXT);
#endif !ZLIB_COMPRESSED_IO
      
      f = zlsio_open(outFile, "wa");
      if (f) {
        zlsio_Printf( " zlsio_SaveLevelUserIO, Level = %d Step = %d File = %s\n", level , step, outFile);
        
        writeTime = time(NULL);
        zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
#ifdef HYDRO
        zlsio_ZFPrintf( f, " FYRIS/ZLS PPM-LR Hydro Simulation TEXT Output\n" );
#endif HYDRO

        zlsio_ZFPrintf( f, " Level = %d, nZones = %d\n", level, gLevels[level].nZones);
        zlsio_Printf( " Level = %d, nZones = %d\n", level, gLevels[level].nZones);
#if (ZLS_VERBOSE > 1)
        zlsp_ZPrintPhysicsConfiguration ( f, NULL );
#endif 
        zlsio_ZFPrintf( f, " Written: %s", ctime(&writeTime));
        zlsio_ZFPrintf( f, " Cycle : %11d Time : %#12.6g = %#12.6g %s\n",
                       zlst_GlobalCycles,
                       zlst_GlobalTime, zlst_GlobalTime *f_t0, s_Time);
        zlsio_ZFPrintf( f, "%s\n\n", DIVIDERLINE );
        
        zlsp_TextLevelColdGas( f, level);
        
      } // f
      zlsio_close(f);
      
      
#ifdef MPI
    }
#endif MPI
#endif DWARFGALAXY_ANALYSIS
    
    if ( level == 0 ) io->ioState.ioUserCount++;
    
  } // io OK
  
  return(err);
}

int zlsio_TextZoneOutput( zFile f, ZonePtr z){ 
  //
  // Write out a row from a zone.  
  // in 1D k, and j are ignored
  // in 2D the jth row is written
  // in 3D the kth,jth row is written
  //
  int err = noErr;
  //
  WHD               zoneW;
  Field             zoneF;
  
  PhysicsPtr phys  = NULL;
  Real       gamma = kZLS_eos_gamma;  // global plasma adiabtic index
  Real       gamm  = kZLS_eos_gamma_1;  // global plasma adiabtic index-1.0
  Real       mu    = k_mu;  // global plasma mu

  CoolingPtr cp    = NULL;
  
#ifdef COOLING
  
  Real      fH[3] = {1.0, 0.0, 0.0};  // atom fractions of all H atoms, H2, HI, HII (= 0.0 here)
  
  Real        C  = 0.0;  // cooling rate
  Real      Ctot = 0.0;  // atomic/ion cooling rate
  Real       CAt = 0.0;  // atomic/ion cooling rate
  Real       CH2 = 0.0;  // H2 cooling rate
  Real       CCO = 0.0;  // CO cooling rate
  Real      Cnet = 0.0;  // net cooling - heating rate
#endif COOLING
 
#ifdef GRAVITY
  Real            gPot; // Cell Average Gravity Potential
#ifdef SYM1D
  Real          dPotdx;  // cell average accelleration
#endif SYM1D
#ifdef SYM2D
  Real          dPotdx;  // cell average accelleration
  Real          dPotdy;  // cell average accelleration
#endif SYM2D
#ifdef SYM3D
  Real          dPotdx;  // cell average accelleration
  Real          dPotdy;  // cell average accelleration
  Real          dPotdz;  // cell average accelleration
#endif SYM3D
#endif GRAVITY
  Real            d0;
  Real            P0;
  Real           Pd0; // P0/d0
  Real     e, El, Ee; // internal, total lagrangian and total eulerian energy density
  Real             T;  // Temperature
 
  Integer col  = 1; // column counter 
  
#ifdef SYM1D
  ACoordPtr          xx;
#endif SYM1D
#ifdef SYM2D
  ACoordPtr          xx, yx;
#endif SYM2D
#ifdef SYM3D
  ACoordPtr          xx, yx, zx;
#endif SYM3D
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  phys  = gLevels[ z->zoneLevel].levelPhysics;
  if (phys != NULL) {
      gamma  = phys->gamma;  // global plasma adiabtic index
      gamm   = phys->gamm;  // global plasma adiabtic index -1.0
      mu     = phys->mu;     // global plasma mu
#ifdef COOLING
      cp    = &(phys->coolParameters);
#endif COOLING  
  }
  
  if ((io != NULL) && (z != NULL)){
    
    err     = zls_GetZoneWHD   (&zoneW, z);
    err     = zls_GetZoneField (&zoneF, z);
   
    if (err == noErr) {
      
      
#ifdef SYM1D
      Integer is  =  z->zoneGrid.ns[kXDim];
      Integer ie  =  z->zoneGrid.ne[kXDim];
      Integer  i;
#ifdef FILES_HAVE_BCS     
      Integer nti  =  z->zoneGrid.nt[kXDim];
#endif FILES_HAVE_BCS
#endif SYM1D
#ifdef SYM2D
      Integer is  =  z->zoneGrid.ns[kXDim];
      Integer ie  =  z->zoneGrid.ne[kXDim];
      Integer  i  = (ie + is)/2;
#ifdef FILES_HAVE_BCS     
      Integer nti  =  z->zoneGrid.nt[kXDim];
#endif FILES_HAVE_BCS
      
      Integer js  =  z->zoneGrid.ns[kYDim];
      Integer je  =  z->zoneGrid.ne[kYDim];
      Integer j  = (je + js)/2;
#ifdef FILES_HAVE_BCS     
      Integer ntj  =  z->zoneGrid.nt[kYDim];
#endif FILES_HAVE_BCS
#endif SYM2D
#ifdef SYM3D
      Integer is  =  z->zoneGrid.ns[kXDim];
      Integer ie  =  z->zoneGrid.ne[kXDim];
      Integer  i  = (ie + is)/2;
#ifdef FILES_HAVE_BCS     
      Integer nti  =  z->zoneGrid.nt[kXDim];
#endif FILES_HAVE_BCS
      
      Integer js  =  z->zoneGrid.ns[kYDim];
      Integer je  =  z->zoneGrid.ne[kYDim];
      Integer j   = (je + js)/2;
#ifdef FILES_HAVE_BCS     
      Integer ntj  =  z->zoneGrid.nt[kYDim];
#endif FILES_HAVE_BCS
      
      Integer ks  =  z->zoneGrid.ns[kZDim];
      Integer ke  =  z->zoneGrid.ne[kZDim];
      Integer k;
#ifdef FILES_HAVE_BCS     
      Integer ntk  =  z->zoneGrid.nt[kZDim];
#endif FILES_HAVE_BCS
#endif SYM3D
      
#ifdef SYM1D
      xx = z->zoneGrid.x;
#endif SYM1D
#ifdef SYM2D
      xx = z->zoneGrid.x;
      yx = z->zoneGrid.y;
#endif SYM2D
#ifdef SYM3D
      xx = z->zoneGrid.x;
      yx = z->zoneGrid.y;
      zx = z->zoneGrid.z;
#endif SYM3D
      
      if (f) {
        
        zlsio_ZFPrintf( f, " Zone ID = %d\n", z->zoneID);
#if (ZLS_VERBOSE > 1)
        zls_ZPrintZoneCoords( f,  z);
        i = is;
#endif (ZLS_VERBOSE > 1)
        
#ifdef SYM1D
        zlsio_ZFPrintf( f, " Zone Quantities along row: \n" );
#endif SYM1D
#ifdef SYM2D
        zlsio_ZFPrintf( f, " Zone Quantities along row: \n" );
        i = is;
        
        j = js;
        Integer jc;
        j = j % 2;
#ifdef ORSZAGTANG        
        for ( jc = 0; jc < je; jc++){
        if ( (yx[jc].l <= 0.3125) &&
             (yx[jc].r >  0.3125) ) {
               j = jc;
             }
        }
#endif ORSZAGTANG        
        
#ifdef CARTESIAN
        zlsio_ZFPrintf( f, "      j = %d,   y = %#5g \n", j, yx[j].c);
#endif CARTESIAN
#ifdef SPHERICAL
        zlsio_ZFPrintf( f, "      j = %d, phi = %#5g \n", j, yx[j].c);
#endif SPHERICAL
#ifdef POLARSPHERICAL
        zlsio_ZFPrintf( f, "      j = %d, phi = %#5g \n", j, yx[j].c);
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
        zlsio_ZFPrintf( f, "      j = %d,   r = %#5g \n", j, yx[j].c);
#endif CYLINDRICAL
#ifdef CIRCULAR
        zlsio_ZFPrintf( f, "      j = %d, phi = %#5g \n", j, yx[j].c);
#endif CIRCULAR
#endif SYM2D
        
#ifdef SYM3D
        zlsio_ZFPrintf( f, " Zone Quantities along row: \n" );
        i = is;
        j = js;
        k = ks;
#ifdef CARTESIAN
        zlsio_ZFPrintf( f, "      j = %d,     y = %#5g \n", j, yx[j].c);
        zlsio_ZFPrintf( f, "      k = %d,     z = %#5g \n", k, zx[k].c);
#endif CARTESIAN
#ifdef SPHERICAL
        zlsio_ZFPrintf( f, "      j = %d,   phi = %#5g (%#5g) \n", j, invpi*yx[j].c, yx[j].c);
        zlsio_ZFPrintf( f, "      k = %d, theta = %#5g (%#5g) \n", k, invpi*zx[k].c, zx[k].c);
#endif SPHERICAL
#ifdef POLARSPHERICAL
        zlsio_ZFPrintf( f, "      j = %d, theta = %#5g (%#5g) \n", j, invpi*yx[j].c, yx[j].c);
        zlsio_ZFPrintf( f, "      k = %d,   phi = %#5g (%#5g) \n", k, invpi*zx[k].c, zx[k].c);
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
        zlsio_ZFPrintf( f, "      j = %d,     r = %#5g \n", j, yx[j].c );
        zlsio_ZFPrintf( f, "      k = %d,   phi = %#5g (%#5g) \n", k, invpi*zx[k].c, zx[k].c);
#endif CYLINDRICAL
#endif SYM3D
        zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
        
#ifdef HYDRO

#ifdef CARTESIAN
        zlsio_ZFPrintf( f, " Cartesian Coordinates\n" );
#ifdef SYM1D
        zlsio_ZFPrintf( f, "    #         [2]X_l         [3]X_c         [4]X_r         [5]<d>         [6]<P>       [7]<v_x>");
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "    #         [2]X_l         [3]X_c         [4]X_r         [5]<d>         [6]<P>       [7]<v_x>       [8]<v_y>");
#endif SYM2D        
#ifdef SYM3D
        zlsio_ZFPrintf( f, "    #         [2]X_l         [3]X_c         [4]X_r         [5]<d>         [6]<P>       [7]<v_x>       [8]<v_y>       [9]<v_z>");
#endif SYM3D        
#endif CARTESIAN
#ifdef CYLINDRICAL
        zlsio_ZFPrintf( f, " Cylindrical Coordinates \n" );
#ifdef SYM1D
        zlsio_ZFPrintf( f, "    #         [2]Z_l         [3]Z_c         [4]Z_r         [5]<d>         [6]<P>       [7]<v_z>       [8]<v_r>");
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "    #         [2]Z_l         [3]Z_c         [4]Z_r         [5]<d>         [6]<P>       [7]<v_z>       [8]<v_r>");
#endif SYM2D        
#ifdef SYM3D
        zlsio_ZFPrintf( f, "    #         [2]Z_l         [3]Z_c         [4]Z_r         [5]<d>         [6]<P>       [7]<v_z>       [8]<v_r>      [9]<v_phi>");
#endif SYM3D        
#endif CYLINDRICAL
#ifdef CIRCULAR
        zlsio_ZFPrintf( f, " Circular Coordinates\n" );
#ifdef SYM1D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>");
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>      [8]<v_phi>");
#endif SYM2D        
#ifdef SYM3D
        { zlsio_ErrPrintf( " ERROR User IO: CIRCULAR geometry not allowed in 3D\n" );
          err = dimErr;
          return(err);}
#endif SYM3D        
#endif CIRCULAR
#ifdef SPHERICAL
        zlsio_ZFPrintf( f, " Spherical Coordinates\n" );
#ifdef SYM1D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>");
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>     [7]<v_phi>");
#endif SYM2D        
#ifdef SYM3D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>      [8]<v_phi>   [9]<v_theta>");
#endif SYM3D        
#endif SPHERICAL
#ifdef POLARSPHERICAL
        zlsio_ZFPrintf( f, " Polar Spherical Coordinates\n" );
#ifdef SYM1D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>");
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>     [8]<v_phi>");
#endif SYM2D        
#ifdef SYM3D
        zlsio_ZFPrintf( f, "    #         [2]R_l         [3]R_c         [4]R_r         [5]<d>         [6]<P>       [7]<v_r>    [8]<v_theta>     [9]<v_phi>");
#endif SYM3D        
#endif POLARSPHERICAL

#ifdef SYM1D
        zlsio_ZFPrintf( f, "          [8]<e>       [9]<E_l>     [10]<E_e>        [11]<T>");
        col = 11;
#endif SYM1D
#ifdef SYM2D
        zlsio_ZFPrintf( f, "          [9]<e>      [10]<E_l>     [11]<E_e>        [12]<T>");
        col = 12;
#endif SYM2D
#ifdef SYM3D
        zlsio_ZFPrintf( f, "         [10]<e>      [11]<E_l>     [12]<E_e>        [13]<T>");
        col = 13;
#endif SYM3D


#endif HYDRO
   

#ifdef COOLING
          if ( cp->coolType != kPhysNoCooling) {
            
            if ( cp->coolType != kPhysFileCooling) {
              zlsio_ZFPrintf( f, "   [%d]<Net Cool>      [%d]<Cool>", ++col, ++col);
            } else {
              
              if ( k_mu < 1.0) {
                fH[0] = 0.0;
                fH[1] = 0.0;
                fH[2] = 1.0; // initially fully ionised guess
              } else  if ( k_mu < 2.0) {
                fH[0] = 0.0;
                fH[1] = 1.0; // initially fully neutral guess
                fH[2] = 0.0; 
              } else {
                fH[0] = 1.0; // initially fully molecular guess
                fH[1] = 0.0;
                fH[2] = 0.0; 
              }
              
              
#ifdef VARIABLE_EOS
              
#ifdef VARIABLE_GAMMA
              
#ifdef H_IONISATION
              zlsio_ZFPrintf( f, "      [%d]< mu>   [%d]< gamma>    [%d]< fH_2>     [%d]< fHI>    [%d]< fHII> [%d]<Net Cool>     [%d]<Cool>",
                                            ++col,      ++col,          ++col,            ++col,          ++col,       ++col,          ++col);
#else
              zlsio_ZFPrintf( f, "      [%d]< mu>   [%d]< gamma> [%d]<Net Cool>     [%d]<Cool>",
                                            ++col,      ++col,          ++col,           ++col);
#endif H_IONISATION
              
#else 
              
#ifdef H_IONISATION
              zlsio_ZFPrintf( f, "      [%d]< mu>    [%d]< fH_2>     [%d]< fHI>    [%d]< fHII> [%d]<Net Cool>     [%d]<Cool>",
                                            ++col,         ++col,         ++col,        ++col,         ++col,          ++col);
#else
              zlsio_ZFPrintf( f, "      [%d]< mu> [%d]<Net Cool>     [%d]<Cool>",
                                            ++col,         ++col,         ++col);
#endif H_IONISATION
              
#endif !VARIABLE_GAMMA
              
#else
              zlsio_ZFPrintf( f, " [%d]<Net Cool>     [%d]<Cool>",
                                            ++col,         ++col);
#endif !VARIABLE_EOS
              
            }
            
            if ( cp->coolType == kPhysFileCooling) {
              zlsio_ZFPrintf( f, "     [%d]<Atom>       [%d]<H2>       [%d]<CO>",
                                            ++col,         ++col,         ++col);
            }
            
          } // not no cooling  
#endif COOLING
        
#ifdef GRAVITY
#ifdef SYM1D
        zlsio_ZFPrintf( f, "      [%d]<Pot>       [%d]<a1>",
                                      ++col,         ++col);
#endif SYM1D        
#ifdef SYM2D
        zlsio_ZFPrintf( f, "      [%d]<Pot>       [%d]<a1>       [%d]<a2>",
                                      ++col,         ++col,         ++col);
#endif SYM2D        
#ifdef SYM3D
        zlsio_ZFPrintf( f, "      [%d]<Pot>       [%d]<a1>       [%d]<a2>       [%d]<a3>",
                                      ++col,         ++col,         ++col,         ++col);
#endif SYM3D        
#endif GRAVITY
        
        zlsio_ZFPrintf( f, "\n" ); // end columns line
#ifdef FILES_NO_BCS            
        for(i = is; i< ie; i++){
          Integer idx = i;
#endif FILES_NO_BCS            
          
#ifdef FILES_HAVE_BCS            
          Integer nti =  z->zoneGrid.nt[kXDim];
          for( i= 0; i< nti; i++){
            
            if ( i == is ) {
              zlsio_ZFPrintf( f, " ^ Boundary Cells ^\n");
            }
            
            if ( i == ie) {
              zlsio_ZFPrintf( f, " v Boundary Cells v\n");
            }
            Integer idx = i;
#endif FILES_HAVE_BCS  

#ifdef SYM1D
#ifdef HYDRO
            zlsio_ZFPrintf( f, " %4d   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g",  
                            idx, xx[i].l, xx[i].c, xx[i].r,
                            zoneW.d[i], zoneW.P[i], 
                            zoneW.vx[i]);
#endif HYDRO
            
            d0   = zoneW.d[i];
            P0   = zoneW.P[i];
            Pd0  = P0/d0;

            mu   = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
#ifdef VARIABLE_GAMMA
            gamma = zlsp_Gamma_Pd(Pd0, cp);
            gamm  = gamma -1.0;
#endif VARIABLE_GAMMA
            e    = P0/(d0*gamm);
            El   = e 
#ifdef HYDRO
            + 0.5*(zoneW.vx[i] *zoneW.vx[i]);
#endif HYDRO
            +0.5*(zoneW.vx[i] *zoneW.vx[i] );
            Ee   = d0*El;
            T    = zlsp_T_Pd(Pd0,cp); // works also for variable Mu and gamma, cp may be null if not needed

#ifdef H_IONISATION
            fH[1] = zlsp_XHI_T ( T, cp );  
            fH[2] = zlsp_XHII_T( T, cp );
            fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
            
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g",e, El, Ee, T);
            
#ifdef COOLING
            if ( cp->coolType != kPhysNoCooling) {
              
              if ( cp->coolType != kPhysFileCooling) {
                C  = d0*d0*zlsp_Lambda(T, cp);
              } else {
                C  = zlsp_LambdaComponents( &Ctot, &CAt, &CH2, &CCO, T, cp);
                Ctot *= d0*d0;
                CAt  *= d0*d0;
                CH2  *= d0*d0;
                CCO  *= d0*d0;
                C    *= d0*d0;
              }

              Cnet = C;
              
              if ( cp->coolType != kPhysFileCooling) {
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", Cnet, C);
              } else {
                
                mu     = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
                gamma  = zlsp_Gamma_Pd(Pd0, cp);
                
#ifdef VARIABLE_EOS
                
#ifdef VARIABLE_GAMMA
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, gamma, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", mu, gamma);
#endif H_IONISATION
                
#else 
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g", mu);
#endif H_IONISATION
                
#endif !VARIABLE_GAMMA
                
#endif VARIABLE_EOS
                
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", Cnet, C, CAt, CH2, CCO);
                
                
              }
              
            }  
#endif COOLING
            
#ifdef GRAVITY
            gPot   = zoneF.p [i].a; // Cell Average Gravity Potential
            dPotdx = -(zoneF.p[i].ar- zoneF.p[i].al)*xx[i].idx;  // cell average accelleration in axis 1
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", gPot, dPotdx);
#endif GRAVITY
            
#endif SYM1D
            
#ifdef SYM2D
#ifdef HYDRO
            zlsio_ZFPrintf( f, " %4d   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g",  
                            idx, xx[idx].l, xx[idx].c, xx[idx].r,
                            zoneW.d[j][i], zoneW.P[j][i], 
                            zoneW.vx[j][i], zoneW.vy[j][i] );   
            
#endif HYDRO
            d0   = zoneW.d [j][i];
            P0   = zoneW.P [j][i];
            Pd0  = P0/d0;
            
            
            mu  = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
#ifdef VARIABLE_GAMMA
            gamma = zlsp_Gamma_Pd(Pd0, cp);
            gamm = gamma -1.0;
#endif VARIABLE_GAMMA
            e    = P0/(d0*gamm);
            El   = e 
#ifdef HYDRO
            + 0.5*(zoneW.vx[j][i] *zoneW.vx[j][i] 
                 + zoneW.vy[j][i] *zoneW.vy[j][i]);
#endif HYDRO
            Ee   = d0*El;
            T = zlsp_T_Pd(Pd0,cp); // works also for variable Mu and gamma, cp may be null if not needed
#ifdef H_IONISATION
            fH[1] = zlsp_XHI_T ( T, cp );  
            fH[2] = zlsp_XHII_T( T, cp );
            fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
            
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g", e, El, Ee, T);
#ifdef COOLING
            if ( cp->coolType != kPhysNoCooling) {
              
              if ( cp->coolType != kPhysFileCooling) {
                C  = d0*d0*zlsp_Lambda(T, cp);
              } else {
                
                C  = zlsp_LambdaComponents( &Ctot, &CAt, &CH2, &CCO, T, cp);
                
                Ctot *= d0*d0;
                CAt  *= d0*d0;
                CH2  *= d0*d0;
                CCO  *= d0*d0;
                C    *= d0*d0;
              }
              Cnet = C;
                            
              if ( cp->coolType != kPhysFileCooling) {
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g",  Cnet, C);
              } else {
                
                mu     = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
                gamma  = zlsp_Gamma_Pd(Pd0, cp);
                
#ifdef VARIABLE_EOS
                
#ifdef VARIABLE_GAMMA
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, gamma, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", mu, gamma);
#endif H_IONISATION
                
#else 
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g", mu);
#endif H_IONISATION
                
#endif !VARIABLE_GAMMA

#endif !VARIABLE_EOS
                
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", Cnet, C, CAt, CH2, CCO);
                
              }
              
            }  
#endif COOLING
            
#ifdef GRAVITY
#ifdef SCALAR_LRQS
              gPot   = zoneF.p[j][i].a; // Cell Average Gravity Potential
#endif SCALAR_LRQS
              
#ifdef VECTOR_LRQS
            gPot   = zoneF.p[j][i].a[k0Dim]; // Cell Average Gravity Potential
#endif VECTOR_LRQS
            dPotdx = -(zoneF.p[j][i].ar[kXDim]- zoneF.p[j][i].al[kXDim])*xx[i].idx;  // cell average accelleration in axis 1
            dPotdy = -(zoneF.p[j][i].ar[kYDim]- zoneF.p[j][i].al[kYDim])*yx[j].idx;  // cell average accelleration in axis 2
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g", gPot, dPotdx, dPotdy);
#endif GRAVITY
            
#endif SYM2D
            
#ifdef SYM3D
            zlsio_ZFPrintf( f, " %4d   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g",  
                            idx, xx[idx].l, xx[idx].c, xx[idx].r,
                            zoneW.d [k][j][i], zoneW.P [k][j][i], 
                            zoneW.vx[k][j][i], zoneW.vy[k][j][i], zoneW.vz[k][j][i]);       
            
            d0 = zoneW.d[k][j][i];
            P0 = zoneW.P[k][j][i];
            Pd0  = P0/d0;
            mu  = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
#ifdef VARIABLE_GAMMA
            gamma = zlsp_Gamma_Pd(Pd0, cp);
            gamm = gamma -1.0;
#endif VARIABLE_GAMMA
            e    = P0/(d0*gamm);
            El   = e+0.5*(zoneW.vx[k][j][i] *zoneW.vx[k][j][i] 
                    + zoneW.vy[k][j][i] *zoneW.vy[k][j][i] 
                    + zoneW.vz[k][j][i] *zoneW.vz[k][j][i]);
            Ee   = El*d0;
            T = zlsp_T_Pd(Pd0,cp); // works also for variable Mu and gamma, cp may be null if not needed
#ifdef H_IONISATION
            fH[1] = zlsp_XHI_T ( T, cp );  
            fH[2] = zlsp_XHII_T( T, cp );
            fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
            
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g", e, El, Ee, T);
            
#ifdef COOLING
            if ( cp->coolType != kPhysNoCooling) {
              
              if ( cp->coolType != kPhysFileCooling) {
                C  = d0*d0*zlsp_Lambda(T,cp);
              } else {
                C  = zlsp_LambdaComponents(&Ctot, &CAt, &CH2, &CCO, T, cp);
                Ctot *= d0*d0;
                CAt  *= d0*d0;
                CH2  *= d0*d0;
                CCO  *= d0*d0;
                C    *= d0*d0;
              }
              Cnet = C;
              
              if ( cp->coolType != kPhysFileCooling) {
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", Cnet, C);
              } else {
                
                mu     = zlsp_Mu_Pd(Pd0,cp);  // local plasma mu
                gamma  = zlsp_Gamma_Pd(Pd0, cp);
                
#ifdef VARIABLE_EOS
                
#ifdef VARIABLE_GAMMA
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, gamma, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g", mu, gamma);
#endif H_IONISATION
                
#else 
                
#ifdef H_IONISATION
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g", mu, fH[0], fH[1], fH[2]);
#else
                zlsio_ZFPrintf( f, "   %#12.6g", mu);
#endif H_IONISATION
                
#endif !VARIABLE_GAMMA

#endif VARIABLE_EOS
                
                zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g   %#12.6g", Cnet, C, CAt, CH2, CCO);
              }
              
            }
#endif COOLING
            
#ifdef GRAVITY
#ifdef SCALAR_POTENTIAL
              gPot   = zoneF.p[k][j][i].a ; // Cell Average Gravity Potential, k0Dim = default central value
#endif SCALAR_POTENTIAL
#ifdef VECTOR_FIELD
            gPot   = zoneF.p[k][j][i].a [k0Dim]; // Cell Average Gravity Potential, k0Dim = default central value
#endif VECTOR_FIELD
            dPotdx = -(zoneF.p[k][j][i].ar[kXDim]- zoneF.p[k][j][i].al[kXDim])*xx[i].idx;  // cell average accelleration in axis 1
            dPotdy = -(zoneF.p[k][j][i].ar[kYDim]- zoneF.p[k][j][i].al[kYDim])*yx[j].idx;  // cell average accelleration in axis 2
            dPotdz = -(zoneF.p[k][j][i].ar[kZDim]- zoneF.p[k][j][i].al[kZDim])*zx[k].idx;  // cell average accelleration in axis 2
            zlsio_ZFPrintf( f, "   %#12.6g   %#12.6g   %#12.6g   %#12.6g ", gPot, dPotdx, dPotdy, dPotdz);
#endif GRAVITY
            
#endif SYM3D
            
            zlsio_ZFPrintf( f, "\n" );
#ifdef FILES_NO_BCS            
          } // 
#endif FILES_NO_BCS            
          
#ifdef FILES_HAVE_BCS            
        } //
#endif FILES_HAVE_BCS  
        
      } else { err = fileErr; }
      
      //zlsp_ReleaseZoneWHD(zoneW);
      
    } // zoneW
    
  }  else { //z, io
    err = memErr;
  }
  
  return(err);
}

