#include <zls.h>

int zlsp_InitCooling(CoolingPtr cp){
  
  int err = noErr;
  
  if ( cp != NULL) {
    
		cp->coolType    = 0;  // no cooling by default
    cp->collAuxType = 0;  // info on presence of additional info such a HI fractions and Xrays
		cp->coolScale   = 1.0;  // no scaling by default
		cp->coolMinT    = 0.0;        // minimum cooling threshold, used for all forms of cooling
		cp->coolMaxT    = 0.0;        // maximum cooling limit, used for all forms of cooling
		cp->coolMinPd   = 0.0;        // minimum cooling threshold, used for all forms of cooling
		cp->coolMaxPd   = 0.0;        // maximum cooling limit, used for all forms of cooling
		cp->coolFlagTot = kZLS_SingleCoolingFunction; // total cooling, sum of components atomic, and molecular, default allowed :)
		cp->coolFlagAt  = kZLS_NoCoolingComponent; // Atomic Component component
		cp->coolFlagH2  = kZLS_NoCoolingComponent; // molecular cooling is off by default, as it requires special
		cp->coolFlagCO  = kZLS_NoCoolingComponent; // preparation on a case by case basis. Sets what is read in, and has no
		// performace impact on the inline Lambda function
		cp->heatFlagCX   =  kZLS_NoCoolingComponent;
		cp->heatFlagHI   =  kZLS_NoCoolingComponent;
		cp->heatFlagGR   =  kZLS_NoCoolingComponent;
		
		// parameters for lookup table
		cp->coolNTable    = 0;     // number of entries in table
		cp->coolDeltaT    = 0.0;  
    
		cp->coolTTableMin = 0.0;   // minimum T in cooling table
		cp->coolTTableMax = 0.0;   // maximum T in cooling table
    
		cp->coolPdTableMin = 0.0;   // minimum P/d in cooling table
		cp->coolPdTableMax = 0.0;   // maximum P/d in cooling table
		
		cp->coolUdValues    = NULL;  // table of internal thermal energy density U(Pd)
		cp->coolPdValues    = NULL;  // table of P/d, to map to T or mu or gamma, ie basic EOS
		cp->coolTValues     = NULL;  // table of temperatures, to map an index to a temp
    
    cp->coolFunction    = NULL;  // table of cooling function
		cp->coolFunctionAt  = NULL;  // table of At cooling function
		cp->coolFunctionH2  = NULL;  // table of H2 cooling function
		cp->coolFunctionCO  = NULL;  // table of CO cooling function
		cp->coolDFunction   = NULL;  // table of cooling derivative
		cp->coolDAtFunction = NULL;  // table of At cooling derivative
		cp->coolDH2Function = NULL;  // table of H2 cooling derivative
		cp->coolDCOFunction = NULL;  // table of CO cooling derivative
    
    cp->coolmuValues    = NULL;  // table of mol Weights(Pd)
		cp->coolGammaValues = NULL;  // table of effective gammas(Pd)
    
		cp->coolfh2Values   = NULL;      // table of H2 fractions
		cp->coolfhiValues   = NULL;      // table of HI fractions
		cp->coolfhiiValues  = NULL;      // table of HII fractions
    
		cp->coolfXtotal     = NULL;      // table of Xray cooling fractions
		cp->coolfXsoft      = NULL;      // table of soft 0.1- 1.5 keV Cooling fractions
		cp->coolfXhard      = NULL;      // table of hard 1.5-10.0 keV Cooling fractions
    
		// parameters for powerlaw cooling
		cp->coolLambda0  = 1.0e-27; // cgs default
		cp->coolT0       = 1.0e4; // default K
		cp->coolPower    = 3.0; // default power index
		
		// parameters for phot equilubrium heating if used (nyi)
		cp->coolPHeat0   = 0.0;
		cp->coolPHSlope  = 0.0;
		cp->coolPHFactor = 0.0;
    
    
  } else { 
    err = memErr;
  }
  return(err);
}

int zlsp_NewCooling(CoolingPtr cp){
  int err = noErr;
  err = zlsp_InitCooling(cp);
	return (err);
}


int zlsp_DisposeCooling(CoolingPtr c){
  int err = noErr;
  if (c->coolTValues!=NULL)     free(c->coolTValues)     ; c->coolTValues     = NULL; 
  if (c->coolUdValues   !=NULL) free(c->coolUdValues   ) ; c->coolUdValues = NULL;
  if (c->coolPdValues   !=NULL) free(c->coolPdValues)    ; c->coolPdValues     = NULL; 
  
  if (c->coolFunction!=NULL)    free(c->coolFunction)    ; c->coolFunction    = NULL;
  if (c->coolFunctionAt!=NULL)  free(c->coolFunctionAt)  ; c->coolFunctionAt  = NULL;
  if (c->coolFunctionH2!=NULL)  free(c->coolFunctionH2)  ; c->coolFunctionH2  = NULL;
  if (c->coolFunctionCO!=NULL)  free(c->coolFunctionCO)  ; c->coolFunctionCO  = NULL;
  
  if (c->coolDFunction!=NULL)   free(c->coolDFunction)   ; c->coolDFunction   = NULL;
  if (c->coolDAtFunction!=NULL) free(c->coolDAtFunction) ; c->coolDAtFunction = NULL;
  if (c->coolDH2Function!=NULL) free(c->coolDH2Function) ; c->coolDH2Function = NULL;
  if (c->coolDCOFunction!=NULL) free(c->coolDCOFunction) ; c->coolDCOFunction = NULL;
  
  if (c->coolmuValues   !=NULL) free(c->coolmuValues)    ; c->coolmuValues    = NULL;
  if (c->coolGammaValues!=NULL) free(c->coolGammaValues) ; c->coolGammaValues = NULL;
  
  if (c->coolfhiiValues!=NULL) free(c->coolfhiiValues)   ; c->coolfhiiValues = NULL;
  if (c->coolfhiValues !=NULL) free(c->coolfhiValues)    ; c->coolfhiValues = NULL;
  if (c->coolfh2Values !=NULL) free(c->coolfh2Values)    ; c->coolfh2Values = NULL;
  
  if (c->coolfXtotal!=NULL) free(c->coolfXtotal)   ; c->coolfXtotal = NULL;
  if (c->coolfXsoft !=NULL) free(c->coolfXsoft)    ; c->coolfXsoft  = NULL;
  if (c->coolfXhard !=NULL) free(c->coolfXhard)    ; c->coolfXhard  = NULL;
  
  return (err);
}


int zlsp_SetPhysicsCoolingFile(CoolingPtr cp, char * cFile) {
#define BUFSIZE 1024
  
  int err   = noErr;
	FILE * f  = NULL;
  
	char buffer[BUFSIZE];
	
	Integer nEntries, nFormat, coolIndex;
	
	Real logT, logL, logLat, logLH2, logLCO, dlogL;
  
  Real muL, gammaL, PdL;
  
  Real fh2L, fhiL, fhiiL;
  
  Real fx, fxs, fxh;
  
	Real T, L, LAt, LH2, LCO , dL, dT;
	
	Real q1, x;
	
	if ( cp != NULL ) {
    
    x =  cp->coolScale;
		
		f = fopen(cFile, "r");
		if (f) {
			
			// skip header
			strcpy(buffer, "#");
			while(*buffer == '#'){
        fgets(buffer, BUFSIZE, f);
#if (ZLS_VERBOSE >0)
        if (*buffer != '#') {
#ifdef MPI
          if ( gMPIrank == 0){
#endif MPI
            puts(buffer);
#ifdef MPI
          }
#endif MPI
        }
#endif ZLS_VERBOSE
      }
      
      fscanf( f, "%d", &nFormat );
      //  nFormat == 0, just basic cooling functions
      //  nFormat == 1, add cooling components
      //  nFormat == 2, include mu and gamma info
      //  nFormat == 3, 2+ include H2, HI, and HII info
      //  nFormat == 5, 3+ include xray, soft xr, and haerd xr fractions
      //  nFormat == 4, raw atomic, molecular components, 
      //                requires,local composition to compute
      //                mu, gamma, and total cooling
      
      fscanf( f, "%d",&nEntries);
      if ( nEntries > 1 ) { // read n entries that follow
        
        cp->coolNTable = nEntries;
        //
        // all tables are allocated, even if disabled on a component by component basis
        // in a given run
        //
        zls_NewQuantity1D(&cp->coolTValues,     nEntries);
        zls_NewQuantity1D(&cp->coolPdValues,    nEntries);
        zls_NewQuantity1D(&cp->coolUdValues,    nEntries);
        
        zls_NewQuantity1D(&cp->coolFunction,    nEntries);
        zls_NewQuantity1D(&cp->coolFunctionAt,  nEntries);
        zls_NewQuantity1D(&cp->coolFunctionH2,  nEntries);
        zls_NewQuantity1D(&cp->coolFunctionCO,  nEntries);
        
        zls_NewQuantity1D(&cp->coolDFunction,   nEntries);
        zls_NewQuantity1D(&cp->coolDAtFunction,   nEntries);
        zls_NewQuantity1D(&cp->coolDH2Function,   nEntries);
        zls_NewQuantity1D(&cp->coolDCOFunction,   nEntries);
        
        zls_NewQuantity1D(&cp->coolmuValues,    nEntries);
        zls_NewQuantity1D(&cp->coolGammaValues, nEntries);
        
        zls_NewQuantity1D(&cp->coolfh2Values,   nEntries);
        zls_NewQuantity1D(&cp->coolfhiValues,   nEntries);
        zls_NewQuantity1D(&cp->coolfhiiValues,  nEntries);
        
        zls_NewQuantity1D(&cp->coolfXtotal,   nEntries);
        zls_NewQuantity1D(&cp->coolfXsoft,    nEntries);
        zls_NewQuantity1D(&cp->coolfXhard,    nEntries);
        
        if (( cp->coolTValues != NULL ) &&( cp->coolFunction != NULL ) && ( cp->coolDFunction != NULL ) ) {
          
          q1   = f_invlam0 /(k_amu*k_amu); // cooling per density^2 normalised, less mu factor
          
          for (coolIndex = 0; coolIndex < nEntries ; coolIndex++){
            
            // temperature
            fscanf( f, "%lf",&logT); // first column for all types is log_10(T)
            
            T      = zls_Exp10(logT);
            
            fh2L   = fhiL = fhiiL = 0.0;
            fx     = fxs  = fxh   = 0.0;
            
            // Cooling
            if ( nFormat == 0) { // Single function
              
              fscanf( f, "%lf", &logL);
              L     = zls_Exp10(logL);
              dL    = dlogL*( L/T );
              
              logLat = logL; 
              logLH2 = logLCO = -100.0;
              
              muL    = k_mu;
              gammaL = kZLS_eos_gamma;
              
              fh2L   = fhiL = fhiiL = 0.0;
              fx     = fxs  = fxh   = 0.0;
              
              cp->coolFlagTot = kZLS_SingleCoolingFunction;
              cp->coolFlagAt  = kZLS_AllowCoolingComponent;
              cp->coolFlagH2  = kZLS_NoCoolingComponent;
              cp->coolFlagCO  = kZLS_NoCoolingComponent;
              
            } else   if ( nFormat == 4) { // Raw Cooling Components
              
              fscanf( f, "%lf %lf %lf", &logLat, &logLH2, &logLCO);

              logL   = logLat; 
              muL    = k_mu;
              gammaL = kZLS_eos_gamma;
              
              fh2L   = fhiL = fhiiL = 0.0;
              fx     = fxs  = fxh   = 0.0;
              
              dL    = 0.0;
              dlogL = 0.0;
              
              cp->coolFlagTot = kZLS_ComponentCoolingFunction;
              
              cp->coolFlagTot = kZLS_MultipleCoolingFunction;
              cp->coolFlagAt  = kZLS_AllowCoolingComponent;
              cp->coolFlagH2  = kZLS_NoCoolingComponent;
              cp->coolFlagCO  = kZLS_NoCoolingComponent;
              
            } else if ( nFormat > 0) {  // Cooling Components
              
              fscanf( f, "%lf %lf %lf %lf",&logL, &logLat, &logLH2,  &logLCO);
              
              muL    = k_mu;
              gammaL = kZLS_eos_gamma;
              
              fh2L   = fhiL = fhiiL = 0.0;
              fx     = fxs  = fxh   = 0.0;
              
              // EOS and ionisation
              
              if ( nFormat > 1) {
                fscanf( f, "%lf %lf"    , &muL, &gammaL);
                cp->collAuxType |= kZLS_HasGammaMu;
              }
              
              if ( nFormat == 3)  {
                fscanf( f, "%lf %lf %lf", &fh2L, &fhiL, &fhiiL);
                cp->collAuxType |= kZLS_HasHFractions;
              }
              
              if ( nFormat == 5)  {
                fscanf( f, "%lf %lf %lf", &fx, &fxs, &fxh);
                cp->collAuxType |= kZLS_HasXFractions;
              }
              
              cp->coolFlagTot = kZLS_MultipleCoolingFunction;
              
            }
            
            if (coolIndex == 0)          cp->coolTTableMin = logT;
            if (coolIndex == nEntries-1) cp->coolTTableMax = logT;
            
            
            L     =  0.0;
            LH2    = 0.0;
            LCO    = 0.0;
            LAt    = 0.0;
            
            if ( cp->coolFlagTot == kZLS_SingleCoolingFunction ){
              
              if (logL > -60.0) L    = zls_Exp10(logL);
              
              LAt    =   L;
              LH2    = 0.0;
              LCO    = 0.0;
              
            } else { // sum by separate components
              
              LAt    = 0.0;
              if (cp->coolFlagAt == kZLS_AllowCoolingComponent){
                if (logLat > -60.0) LAt    = zls_Exp10(logLat);
              }
              
              LH2    = 0.0;
              if (cp->coolFlagH2 == kZLS_AllowCoolingComponent){
                if (logLH2 > -60.0) LH2    = zls_Exp10(logLH2);
              }
              
              LCO    = 0.0;
              if (cp->coolFlagCO == kZLS_AllowCoolingComponent){
                if (logLCO > -60.0)  LCO    = zls_Exp10(logLCO);
              }
              
              L = LAt+LH2+LCO;
              logL = zls_Log10(L); 
                            
            }
            
            
            cp->coolTValues[coolIndex]       = T; 
            
            PdL = (k_mu/muL)*T/f_tep0; 
            if (coolIndex == 0)          cp->coolPdTableMin = PdL; // already linear
            if (coolIndex == nEntries-1) cp->coolPdTableMax = PdL; // already linear
            
            cp->coolPdValues [coolIndex]   = PdL; 
            cp->coolUdValues [coolIndex]   = PdL/(gammaL-1.0); 
            
            cp->coolmuValues   [coolIndex] = muL; 
            cp->coolGammaValues[coolIndex] = gammaL; 
            
            cp->coolfh2Values [coolIndex]  = fh2L; 
            cp->coolfhiValues [coolIndex]  = fhiL; 
            cp->coolfhiiValues[coolIndex]  = fhiiL; 
            
            cp->coolfXtotal[coolIndex]     = fx; 
            cp->coolfXsoft [coolIndex]     = fxs; 
            cp->coolfXhard [coolIndex]     = fxh; 
            
            muL    = k_mu;
            gammaL = kZLS_eos_gamma;
            
            cp->coolFunction[coolIndex]      = x*q1 * L/(muL*muL);  // normalised cooling rate with f_invlam0
            cp->coolFunctionAt[coolIndex]    = x*q1 * LAt/(muL*muL); 
            cp->coolFunctionH2[coolIndex]    = x*q1 * LH2/(muL*muL); 
            cp->coolFunctionCO[coolIndex]    = x*q1 * LCO/(muL*muL);  
            
          }
          // set up linear interp values
          for (coolIndex = 0; coolIndex < nEntries-1 ; coolIndex++){
            
            T = cp->coolTValues[coolIndex    ]; 
            dT = cp->coolTValues[coolIndex + 1] - T; 
            
            L  = cp->coolFunction[coolIndex    ]; 
            dL = cp->coolFunction[coolIndex + 1] - L ; 
            cp->coolDFunction[coolIndex] = dL/dT;
            
            L  = cp->coolFunctionAt[coolIndex    ]; 
            dL = cp->coolFunctionAt[coolIndex + 1] - L ; 
            cp->coolDAtFunction[coolIndex] = dL/dT;
            
            L  = cp->coolFunctionH2[coolIndex    ]; 
            dL = cp->coolFunctionH2[coolIndex + 1] - L ; 
            cp->coolDH2Function[coolIndex] = dL/dT;
            
            L  = cp->coolFunctionCO[coolIndex    ]; 
            dL = cp->coolFunctionCO[coolIndex + 1] - L ; 
            cp->coolDCOFunction[coolIndex] = dL/dT;
            
            //               zlsio_Printf(" T %g L %g dT %g dL %g dL/dT %g \n", T, L, dT, dL, dL/dT);
            
          }
          
          cp->coolDeltaT = 0.0;
          if (nEntries > 1) cp->coolDeltaT = (Real)(nEntries-1)/(cp->coolTTableMax-cp->coolTTableMin);
          
          cp->coolTTableMin   = zls_Exp10( cp->coolTTableMin ); // make linear
          cp->coolTTableMax   = zls_Exp10( cp->coolTTableMax ); // make linear
          
          cp->coolPdTableMin   = cp->coolPdValues[0];
          cp->coolPdTableMax   = cp->coolPdValues[nEntries-1];
          
          
        } else {
          zlsio_ErrPrintf( " ERROR: SetPhysicsCoolingFile, FAILED to allocate tables %d \n", nEntries);
          err = memErr;
        }
        
      } else { 
        zlsio_ErrPrintf( " ERROR: SetPhysicsCoolingFile, bad number of entries: %d \n", nEntries);
        err = fileErr;
      }
      fclose(f);
      
		} else { err = fileErr; }
    
	}
  
  return (err);
  
}

#ifdef COOLING

#define RUNGEKUTTA_COOLING
//4th order RK single step integration


Real RK4(Real dt, Real U0, Real T, Real d, CoolingPtr cp);
Real RK4(Real dt, Real U0, Real T, Real d, CoolingPtr cp){
  
  Real  delta = 0.0;
  
  register Real U, Ud;
  register Real d2    = d*d;
  register Real dtd2  = dt*d2;
  register Real invd  = 1.0/d;
  
  register Real k1, k2, k3, k4;
    
  k1     = -dtd2*zlsp_Lambda( T, cp );
    
  U     = U0 + k1*0.5;
  Ud    = U*invd;
  
  T     = zlsp_T_Ud ( Ud, cp);
  k2    = -dtd2*zlsp_Lambda( T, cp );
    
  U     = U0 + k2*0.5;
  Ud    = U*invd;
  
  T     = zlsp_T_Ud ( Ud, cp);
  k3    = -dtd2*zlsp_Lambda( T, cp );
  
  U     = U0 + k3;
  Ud    = U*invd;

  T     = zlsp_T_Ud ( Ud, cp);
  k4    = -dtd2*zlsp_Lambda( T, cp );
    
  delta = (k1+k2+k2+k3+k3+k4)*sixth;
  
  return(delta);
}
#endif COOLING


int zlsp_CoolZoneWHD( ZonePtr z, Real dt ){
  //
  // cools zone cells and keeps track of global smallest cooling timescale
  // for use later in timestep determinations.
  // cp->coolMinTime contains the current global minimum cooling time
  // or kZLS_Huge if no global minimum is yet known.  Set in init routine for physics
  // (zlsp_InitPhysics) and reset after each global timestep is computed in zlst_CalcAllTimeSteps.
  //
#ifdef COOLING
  int err     = noErr;
  register Real U0, U, DeltaU;
  register Real T;
  register Real dtd2;
  
  PhysicsPtr phys = NULL;
  WHD          zoneW ;
  GridCoords       gc;
  Integer       level;
  
  Real          d, d2, invd;
  Real          P;
  Real   invgamm   = kZLS_eos_invgamm;
  Real         Pd;
  Real         Ud;
  
  Real minUd;
    
  Real avgCoolingTime = kZLS_Huge; // average cooling integrated over dt
  Real minCoolingTime = kZLS_Huge; // global minimum cached
  	
#ifdef SYM1D
  Integer        i;
#endif SYM1D
	
#ifdef SYM2D
  Integer     i, j;
#endif SYM2D
	
#ifdef SYM3D
  Integer  i, j, k;
#endif SYM3D
	
  if ( z != NULL ) {
    
	  zoneW =  z->zoneW;
	  
	  if ( zoneW.P != NULL ) {
		  
		  level  = z->zoneLevel;
		  phys = gLevels[level].levelPhysics;
		  gc = z->zoneGrid;
		  
		  if ( phys != NULL ) {
        
        CoolingPtr cp = &(phys->coolParameters);
        
        if (cp->coolType > kPhysNoCooling ){
                    
          minUd = zlsp_Ud_Pd(cp->coolMinPd, cp);
          
#ifdef SYM1D
				  
				  Integer is = kZLS_LimitEvolveLimits;
				  Integer ie = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
				  
				  for(i = is; i< ie; i++){
            
            P     = zoneW.P[i];
					  d     = zoneW.d[i];
            
            invd  = 1.0/d;
            
            Pd    = P*invd;

					  T     = zlsp_T_Pd(Pd, cp);
            
#ifdef H_IONISATION
            fH[1] = zlsp_XHI_T ( T, cp );  
            fH[2] = zlsp_XHII_T( T, cp );
            fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
            
#ifdef VARIABLE_GAMMA
            gamm  = zlsp_Gamma_Pd(Pd, cp)-1.0;
            invgamm = 1.0/gamm;
#endif VARIABLE_GAMMA
            
					  U0    =  P*invgamm;
            
					  if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
						  
						  d2    = d*d;
						  dtd2  = dt*d2;
              
#ifdef RUNGEKUTTA_COOLING
              // explicit 4th order RK integration, refined stepping if error large
              
              DeltaU = RK4( dt, U0, T, Tgr, d, fH, cp);
              Real  U1   = U0 + DeltaU;
              U1      = U1 < (minUd*d) ? (minUd*d): U1;
              DeltaU = U1 - U0;
              
              
              if (( zls_abs(DeltaU)/U0)>0.05) {
                
                Real Ui       = U0;
                Real DeltaUi0 = 0.0;
                Real DeltaUi  = DeltaU;
                
                Integer iter;
                
                for (iter = 1; iter < 5; iter++) { // up to 5 levels of power of 2 refinement
                  
                  if ((zls_abs((DeltaUi-DeltaUi0))/U0)>0.01) { // successive cooling worse than 1% different
                    
                    
                    Integer steps = (1 << iter); // 2, 4, 8, 16
                    Integer step = 0;
                    
                    
                    Real dt_iter = dt/(1.0*steps);
                    
                    DeltaUi0  = DeltaUi; // save previous cooling
                    
                    Ui = U0;

                    T     = zlsp_T_Ud ( Ui*d, cp);
                    
                    DeltaUi = 0.0;
                    
                    for ( step = 0 ; step < steps; step++){
                      DeltaUi = RK4( dt_iter, Ui, T, Tgr, d, fH, cp);
                      Ui      = Ui + DeltaUi;
                      Ui      = Ui < (minUd*d) ? (minUd*d): Ui;

                      T     = zlsp_T_Ud ( Ui*d, cp);

                    }
                    
                    DeltaUi = Ui - U0; // compute new cooling
                    
                    if ( iter > 4){ zlsio_Printf("Cooling Iteration: %d steps %d %g %g %g \n", iter,steps,  DeltaUi0, DeltaUi, U0 );
                      zlsio_Printf("test %g \n", (zls_abs((DeltaUi-DeltaUi0))/U0));
                    }
                    
                    
                  } 
                  
                }
                
                DeltaU = DeltaUi;
                
              } // check for refine step if total cooling is large
              
#endif RUNGEKUTTA_COOLING
              
              U      = U0 + DeltaU;
              U      = U < (minUd*d) ? (minUd*d): U;
						  
						  avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
						  minCoolingTime= avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime; // keep time if a minimum
						  
              Ud    = U*invd;

              Pd    = zlsp_Pd_Ud( Ud, cp);
              
              if ( Pd < cp->coolMinPd ) {
                zoneW.P[i] = zls_max( kZLS_PressureVacuum, d*cp->coolMinPd);
              } else {
                zoneW.P[i] = zls_max( kZLS_PressureVacuum,  d*Pd );
              }  
              
              z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
              
					  } // T
					  
            
				  } //i
				  
				  
#endif SYM1D
				  
				  
#ifdef SYM2D
          
				  Integer is = kZLS_LimitEvolveLimits;
				  Integer ie = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
				  Integer js = kZLS_LimitEvolveLimits;
				  Integer je = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
          
          for(i = is; i< ie; i++){
            
            for(j = js; j < je; j++){
              
              d     = zoneW.d[j][i];
              P     = zoneW.P[j][i];
              invd  = 1.0/d;
              
              Pd    = P*invd;

              T     = zlsp_T_Pd(Pd, cp);
#ifdef H_IONISATION
              fH[1] = zlsp_XHI_T ( T, cp );  
              fH[2] = zlsp_XHII_T( T, cp );
              fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
#ifdef VARIABLE_GAMMA
              gamm  = zlsp_Gamma_Pd(Pd, cp)-1.0;
              invgamm = 1.0/gamm;
#endif VARIABLE_GAMMA
              
              U0    =  P*invgamm;
              
              if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
                
                d2    =   d*d;
                dtd2  = dt*d2;
                                
#ifdef RUNGEKUTTA_COOLING
                // explicit 4th order RK integration, refined stepping if error large
                
                DeltaU = RK4( dt, U0, T, Tgr, d, fH, cp);
                Real  U1   = U0 + DeltaU;
                U1      = U1 < (minUd*d) ? (minUd*d): U1;
                
                DeltaU = U1 - U0;
                
                if (( zls_abs(DeltaU)/U0)>0.05) {
                  
                  Real Ui       = U0;
                  Real DeltaUi0 = 0.0;
                  Real DeltaUi  = DeltaU;
                  
                  Integer iter;
                  
                  for (iter = 1; iter < 5; iter++) { // up to 5 levels of power of 2 refinement
                    
                    if ((zls_abs((DeltaUi-DeltaUi0))/U0)>0.01) { // successive cooling worse than 1% different
                      
                      
                      Integer steps = (1 << iter); // 2, 4, 8, 16
                      Integer step = 0;
                      
                      
                      Real dt_iter = dt/(1.0*steps);
                      
                      DeltaUi0  = DeltaUi; // save previous cooling
                      
                      Ui = U0;

                      T     = zlsp_T_Ud ( Ui*d, cp);
                      
                      DeltaUi = 0.0;
                      
                      for ( step = 0 ; step < steps; step++){
                        DeltaUi = RK4( dt_iter, Ui, T, Tgr, d, fH, cp);
                        Ui      = Ui + DeltaUi;
                        Ui      = Ui < (minUd*d) ? (minUd*d): Ui;

                        T     = zlsp_T_Ud ( Ui*d, cp);

                      }
                      
                      DeltaUi = Ui - U0; // compute new cooling
                      
                      if ( iter > 4){ zlsio_Printf("Cooling Iteration: %d steps %d %g %g %g \n", iter,steps,  DeltaUi0, DeltaUi, U0 );
                        zlsio_Printf("test %g \n", (zls_abs((DeltaUi-DeltaUi0))/U0));
                      }
                      
                      
                    } 
                    
                  }
                  
                  DeltaU = DeltaUi;
                  
                } // check for refine step if total cooling is large
                
#endif RUNGEKUTTA_COOLING
                U      = U0 + DeltaU;
                U      = U < (minUd*d) ? (minUd*d): U;
                
                avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
                minCoolingTime = avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime;
                
                Ud    = U*invd;
                Pd    = zlsp_Pd_Ud( Ud, cp);
                
                if ( Pd < cp->coolMinPd ) {
                  zoneW.P[j][i] = zls_max( kZLS_PressureVacuum, d*cp->coolMinPd);
                } else {
                  zoneW.P[j][i] = zls_max( kZLS_PressureVacuum, d*Pd );
                }   
                
                z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
                
              } // T
              
						  
					  } //i
					  
				  }//j
				  
#endif SYM2D
				  
				  
#ifdef SYM3D
				  
				  Integer is = kZLS_LimitEvolveLimits;
				  Integer ie = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
				  Integer js = kZLS_LimitEvolveLimits;
				  Integer je = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
				  Integer ks = kZLS_LimitEvolveLimits;
				  Integer ke = gc.nt[kZDim]-kZLS_LimitEvolveLimits;
				  
				  
          for(i = is; i< ie; i++){

            for(k = ks; k < ke; k++){
              for(j = js; j < je; j++){
                
                
                d     = zoneW.d[k][j][i];
                P     = zoneW.P[k][j][i];
                invd  = 1.0/d;
                
                Pd    = P*invd;

                T     = zlsp_T_Pd(Pd, cp);
#ifdef H_IONISATION
                fH[1] = zlsp_XHI_T ( T, cp );  
                fH[2] = zlsp_XHII_T( T, cp );
                fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
#ifdef VARIABLE_GAMMA
                gamm  = zlsp_Gamma_Pd(Pd, cp)-1.0;
                invgamm = 1.0/gamm;
#endif VARIABLE_GAMMA
                
                // rk    =  f_tep0*gamm*invd;
                U0    =  P*invgamm;
                
                if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
                  
                  d2    = d*d;
                  dtd2  = dt*d2;
                                    
#ifdef RUNGEKUTTA_COOLING
                  // explicit 4th order RK integration, refined stepping if error large
                  
                  DeltaU = RK4( dt, U0, T, d, cp);
                  
                  Real  U1   = U0 + DeltaU;
                  U1      = U1 < (minUd*d) ? (minUd*d): U1;
                  
                  DeltaU = U1 - U0;
                  
                  if (( zls_abs(DeltaU)/U0)>0.05) {
                    
                    Real Ui       = U0;
                    Real DeltaUi0 = 0.0;
                    Real DeltaUi  = DeltaU;
                    
                    Integer iter;
                    
                    for (iter = 1; iter < 5; iter++) { // up to 5 levels of power of 2 refinement
                      
                      if ((zls_abs((DeltaUi-DeltaUi0))/U0)>0.01) { // successive cooling worse than 1% different
                        
                        
                        Integer steps = (1 << iter); // 2, 4, 8, 16
                        Integer step = 0;
                        
                        
                        Real dt_iter = dt/(1.0*steps);
                        
                        DeltaUi0  = DeltaUi; // save previous cooling
                        
                        Ui = U0;

                        T     = zlsp_T_Ud ( Ui*d, cp);
                        
                        DeltaUi = 0.0;
                        
                        for ( step = 0 ; step < steps; step++){
                          DeltaUi = RK4( dt_iter, Ui, T, d, cp);
                          Ui      = Ui + DeltaUi;
                          Ui      = Ui < (minUd*d) ? (minUd*d): Ui;

                          T     = zlsp_T_Ud ( Ui*d, cp);
                        }
                        
                        DeltaUi = Ui - U0; // compute new cooling
                        
                        if ( iter > 4){ zlsio_Printf("Cooling Iteration: %d steps %d %g %g %g \n", iter,steps,  DeltaUi0, DeltaUi, U0 );
                          zlsio_Printf("test %g \n", (zls_abs((DeltaUi-DeltaUi0))/U0));
                        }
                        
                        
                      } 
                      
                    }
                    
                    DeltaU = DeltaUi;
                    
                  } // check for refine step if total cooling is large
                  
#endif RUNGEKUTTA_COOLING
                  
                  U      = U0 + DeltaU;
                  U      = U < (minUd*d) ? (minUd*d): U;
                  
                  avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
                  minCoolingTime= avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime;
                  
                  Ud    = U*invd;
                  Ud    = U*invd;

                  Pd    = zlsp_Pd_Ud( Ud, cp);
                  
                  if ( Pd < cp->coolMinPd ) {
                    zoneW.P[k][j][i] = zls_max( kZLS_PressureVacuum, d*cp->coolMinPd);
                  } else {
                    zoneW.P[k][j][i] = zls_max( kZLS_PressureVacuum, d*Pd );
                  }      
                  
                  z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
                  
                } // T
                
                
              } //i
						  
					  }//j
					  
				  }//k
          
#endif SYM3D
				  z->zoneMinCoolingTime  =  minCoolingTime;
				  
			  } // not noCooling
			  
        
		  } // p
		  
	  } else  {
		  zlsio_ErrPrintf( " WARNING: zlsp_CoolZoneWHD , zoneW not defined. \n" );
		  err = memErr;
	  }
	  
  } else  {
    zlsio_ErrPrintf( " WARNING: zlsp_CoolZoneWHD , z not defined. \n" );
    err = memErr;
  }
  return(err);
	
#else 
  int err = nyiErr;
  zlsio_ErrPrintf( " WARNING: zlsp_CoolZoneWHD , cooling not enabled. \n" );
  return( err );
#endif COOLING
	
}


int zlsp_InitialCoolingTimeWHD( Real *minDt,  ZonePtr z,  Real dt ){
  //
	// evaluates the cooling timescales without updating the 
	// grid variables.  Mainly used to help estimate initial timesteps
	// for strongly cooling initial conditions.
	// pass the initial guess in dt, and dt holds the min at the end,
	// and the  cp->coolMinTime is set as well
  // NOTE: assumes a constant molecular weight and gamma, unlike the full routine
  //
#ifdef COOLING
  int err     = noErr;
  register Real U0, U, DeltaU;
  register Real T;
  register Real dtd2, rk;
  register Real k1, k2, k3, k4;
  
  PhysicsPtr phys = NULL;
  GridCoords       gc;
  Integer       level;
  
  Real          d, d2;
  
  Real avgCoolingTime = kZLS_Huge;
  Real minCoolingTime = kZLS_Huge;
  *minDt = kZLS_Huge;
  
  Real         fH[3] = {0.0, 0.0, 1.0}; // default fully ionised (affect crx and hipe)
  	
#ifdef SYM1D
  Integer        i;
#endif SYM1D
	
#ifdef SYM2D
  Integer     i, j;
#endif SYM2D
	
#ifdef SYM3D
  Integer  i, j, k;
#endif SYM3D
	
  if ( z != NULL) {
	  
	  WHD zoneW =  z->zoneW;
	  if ( zoneW.P != NULL ) {
		  
		  level  = z->zoneLevel;
		  phys = gLevels[level].levelPhysics;
		  gc = z->zoneGrid;
		  
		  if ( phys != NULL ) {
        
        CoolingPtr cp = &(phys->coolParameters);
        
        if (cp->coolType > kPhysNoCooling ){	
          
          if ( k_mu < 1.0) {
            fH[0] = 0.0;
            fH[1] = 0.0;
            fH[2] = 1.0; // initially fully ionised guess
          } else  if ( k_mu < 2.0) {
            fH[0] = 0.0;
            fH[1] = 1.0; // initially fully neutral guess
            fH[2] = 0.0; 
          } else {
            fH[0] = 1.0; // initially fully molecular guess
            fH[1] = 0.0;
            fH[2] = 0.0; 
          }
				  
#ifdef SYM1D
				  
				  Integer is = gc.ns[kXDim];
				  Integer ie = gc.ne[kXDim];
				  
				  for(i = is; i< ie; i++){
					  
					  d     = zoneW.d[i];
					  rk    =  f_tep0*kZLS_eos_gamma_1/d;
					  
					  U0    = zoneW.P[i]*kZLS_eos_invgamm;
					  T     = U0*rk;
					  
					  if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
						  
						  d2    = d*d;
						  dtd2  = dt*d2;
						  
						  

              k1     = -dtd2*zlsp_Lambda( T, cp );
						  U     = U0 + k1*0.5;
						  T     = U*rk;

              k2     = -dtd2*zlsp_Lambda( T, cp );
						  U     = U0 + k2*0.5;
						  T     = U*rk;

              k3     = -dtd2*zlsp_Lambda( T, cp );
						  U     = U0 + k3;
						  T     = U*rk;

              k4     = -dtd2*zlsp_Lambda( T, cp );
						  DeltaU = (k1+k2+k2+k3+k3+k4)*sixth;
						  U      = U0 + DeltaU;
						  
						  avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
						  minCoolingTime= avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime;
						  
					  } // T
					  
				  } //i
          
#endif SYM1D
				  
				  
#ifdef SYM2D
				  
				  Integer is = gc.ns[kXDim];
				  Integer ie = gc.ne[kXDim];
				  Integer js = gc.ns[kYDim];
				  Integer je = gc.ne[kYDim];
				  
				  for(j = js; j < je; j++){
					  for(i = is; i< ie; i++){
						  
						  d     = zoneW.d[j][i];
						  rk    = f_tep0*kZLS_eos_gamma_1/d;
						  
						  U0    = zoneW.P[j][i]*kZLS_eos_invgamm;
						  T     = U0*rk;
						  
						  if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
							  d2    = d*d;
							  dtd2  = dt*d2;
							  
                k1     = -dtd2*zlsp_Lambda( T, cp );
							  U     = U0 + k1*0.5;
							  T     = U*rk;

                k2     = -dtd2*zlsp_Lambda( T, cp );
							  U     = U0 + k2*0.5;
							  T     = U*rk;

                k3     = -dtd2*zlsp_Lambda( T, cp );
							  U     = U0 + k3;
							  T     = U*rk;

                k4     = -dtd2*zlsp_Lambda( T, cp );
							  DeltaU = (k1+k2+k2+k3+k3+k4)*sixth;
							  U      = U0 + DeltaU;
							  
                avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
							  minCoolingTime= avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime;
							  
						  } // T
						  
					  } //i
					  
				  }//j
				  
				  
#endif SYM2D
				  
				  
#ifdef SYM3D
				  
				  Integer is = gc.ns[kXDim];
				  Integer ie = gc.ne[kXDim];
				  Integer js = gc.ns[kYDim];
				  Integer je = gc.ne[kYDim];
				  Integer ks = gc.ns[kZDim];
				  Integer ke = gc.ne[kZDim];
				  
				  
				  for(k = ks; k < ke; k++){
					  for(j = js; j < je; j++){
						  for(i = is; i< ie; i++){
							  
							  d     = zoneW.d[k][j][i];
							  rk    = f_tep0*kZLS_eos_gamma_1/d;
							  U0    = zoneW.P[k][j][i]*kZLS_eos_invgamm;
							  T     = U0*rk;
							  
							  if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
								  
								  d2    = d*d;
								  dtd2  = dt*d2;
								  
                  k1     = -dtd2*zlsp_Lambda( T, cp );
								  U     = U0 + k1*0.5;
								  T     = U*rk;

                  k2     = -dtd2*zlsp_Lambda( T, cp );
								  U     = U0 + k2*0.5;
								  T     = U*rk;

                  k3     = -dtd2*zlsp_Lambda( T, cp );
								  U     = U0 + k3;
								  T     = U*rk;

                  k4     = -dtd2*zlsp_Lambda( T, cp );

								  DeltaU = (k1+k2+k2+k3+k3+k4)*sixth;
								  U      = U0 + DeltaU;
								  
                  avgCoolingTime = zls_abs( U/DeltaU )*dt; // note this is fraction of final U, not initial U
								  minCoolingTime = avgCoolingTime < minCoolingTime? avgCoolingTime : minCoolingTime;
								  
							  } // T
							  
						  } //i
						  
					  }//j
					  
				  }//k
          
#endif SYM3D
				  z->zoneMinCoolingTime  =  minCoolingTime;
				  *minDt = minCoolingTime;
			  } // not noCooling
		  } // p
		  
	  } else  {
		  zlsio_ErrPrintf( " WARNING: zlsp_InitialCoolingTimeWHD , zoneW not defined. \n" );
		  err = memErr;
	  }
	  
  } else  {
    zlsio_ErrPrintf( " WARNING: zlsp_InitialCoolingTimeWHD , z not defined. \n" );
    err = memErr;
  }
  return(err);
	
#else 
  int err = nyiErr;
  zlsio_ErrPrintf( " WARNING: zlsp_InitialCoolingTimeWHD , cooling not enabled. \n" );
  return( err );
#endif COOLING
	
}


