#include <zls.h>

int zls_InitField( FieldPtr f){
  int err = noErr;
	
	(f)->p    = NULL; 
  
  return (err);
}

int zls_NewFieldPtr( FieldPtr * f){
  int err = noErr;
	
  *f = (FieldPtr)malloc(sizeof(Field));
	
  if (*f == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
		zlsio_ErrPrintf( " ERROR: zls_NewField allocation failed %d \n",err);
#endif
    return (err);
  }	
  
	(*f)->p   = NULL; 
  
  return (err);
}


int zls_DisposeField(Field f){
  int err = noErr;
  
  
#ifdef SYM1D
  zls_DisposeLRQ1DArr(f.p); 
  f.p = NULL;
#endif SYM1D
  
#ifdef SYM2D
  zls_DisposeLRQ2DArr(f.p); 
  f.p = NULL;
#endif SYM2D
  
#ifdef SYM3D
  zls_DisposeLRQ3DArr(f.p); 
  f.p = NULL;
#endif SYM3D
  
	return (err);
}

int zls_DisposeFieldPtr(FieldPtr f){
  int err = noErr;
	
	if ( f != NULL ) {
    
#ifdef SYM1D
    zls_DisposeLRQ1DArr(f->p); 
    f->p = NULL;
#endif SYM1D
    
#ifdef SYM2D
    zls_DisposeLRQ2DArr(f->p); 
    f->p = NULL;
#endif SYM2D
    
#ifdef SYM3D
    zls_DisposeLRQ3DArr(f->p); 
    f->p = NULL;
#endif SYM3D
    
		free(f);
    
	}
	
	return (err);
}


#ifdef GRAVITY

#ifdef GRAVITY_POTENTIAL
// version where field represents the potential field

int zlsg_InitZoneGravityPotentialField( ZonePtr z){
  int err = noErr;
  Real r;
  
#ifdef SYM1D
  ACoordPtr xx  = z->zoneGrid.x;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 2 end points in 1D
  Real x0, x1;
  Real phi0, phi1;
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 3 points in 1D
  Real x0, x1, x2;
  Real phi0, phi1, phi2;
#endif  
  
	Integer nti   = z->zoneGrid.nt[kXDim];
  Integer i = 0;
#endif SYM1D
  
#ifdef SYM2D
  ACoordPtr xx  = z->zoneGrid.x;
  ACoordPtr yy  = z->zoneGrid.y;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 4 vertex points in 2D
  // 
  Integer c; // counter for points to sum
  Real xc[4]; // only 2 values but duped to make loop simple:, xl, xr, xl, xr
  Real yc[4]; // only 2 values but duped to make loop simple, yl, yr, yl, yr
  Real phi[4]; // phi at each corner (x,y)
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 9 points in 2D!
  Integer c; // counter for points to sum
  Real xc[9]; //(xl, xc, xr)*3 planes
  Real yc[9]; //(yl, yc, yr)*3 planes
  Real phi[9]; //eval in single sum loop
#endif  
  
	Integer nti   = z->zoneGrid.nt[kXDim];
	Integer ntj   = z->zoneGrid.nt[kYDim];
  Integer i = 0;
  Integer j = 0;
  
  // these are the interpolated 2D phi values already combined
  // with the component fractions and conservatively averaged
   Real PhiC ; // cell centre phi
 
  Real phiXl; // left   phi in X direction
#if (FIELD_RECON_ORDER == 2 )
  Real phiXc; // centre phi in X direction, used in order 2
  Real phiYc; // centre phi in Y direction, used in order 2
#endif
  Real phiXr; // right  phi in X direction
  
  Real phiYl; // left   phi in Y direction
  Real phiYr; // right  phi in Y direction
  
#endif SYM2D
  
#ifdef SYM3D
  ACoordPtr xx  = z->zoneGrid.x;
  ACoordPtr yy  = z->zoneGrid.y;
  ACoordPtr zz  = z->zoneGrid.z;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 8 vertex points in 3D!
  Integer c; // counter for points to sum
  Real xc[8]; // two values, xl and xr, but copied for each pair of vetexes, xl, xr, xl ,xr
  Real yc[8];
  Real zc[8];
  Real phi[8];
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 27 points in 3D!
  // common points between cells could be shared, but it is messy in 3D
  // and initially we are just going to get it working with lots of calls.
  Integer c; // counter for points to sum
  Real xc[27];
  Real yc[27];
  Real zc[27];
  Real phi[27];
#endif  
  
  Integer nti   = z->zoneGrid.nt[kXDim];
  Integer ntj   = z->zoneGrid.nt[kYDim];
  Integer ntk   = z->zoneGrid.nt[kZDim];
  
  Integer i = 0;
  Integer j = 0;
  Integer k = 0;
  

#if (FIELD_RECON_ORDER == 2 )
  Real phiXc; // centre phi used in order 2
//  Real phiYc; // centre phi used in order 2
//  Real phiZc; // centre phi used in order 2
#endif 
  
  Real phiXl; // left   phi in X direction
  Real phiXr; // right  phi in X direction
  
  Real phiYl; // left   phi in Y direction
  Real phiYr; // right  phi in Y direction
  
  Real phiZl; // left   phi in Z direction
  Real phiZr; // right  phi in Z direction
  
#endif SYM3D
  
  PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global gravity parameters
  if ((phys != NULL)&&( z != NULL)) {
    
#if (ZLS_DEBUG > 0)
    zlsio_Printf( " Zone Field Initialization: %2.2d. \n", z->zoneID );
#endif
    
    PotentialPtr gp  = &(phys->gravPotential);
    
    Field  f;
    zls_GetZoneField(&f, z);
    
    if ( f.p != NULL ) {
#ifdef SYM1D
      for(i = 0; i< nti; i++){
        
#if (FIELD_RECON_ORDER == 1 )
        // two calls
        x0    = xx[i].l;
#ifdef GRAVZFIELD
        r     = x0;
#else
        r     = zls_CvtCoordsToRadius1D( x0);
#endif !GRAVZFIELD
        
        phi0  = zlsg_phi ( r , gp );
        
        x1    = xx[i].r;
#ifdef GRAVZFIELD
        r     = x1;
#else
        r     = zls_CvtCoordsToRadius1D( x1);
#endif !GRAVZFIELD
        phi1  = zlsg_phi ( r , gp );
        
        f.p[i].a   = 0.5*(phi0 + phi1);
        f.p[i].al  = phi0;
        f.p[i].ar  = phi1;
#endif
#if (FIELD_RECON_ORDER == 2 )
        // three calls
        x0    = xx[i].l;
#ifdef GRAVZFIELD
        r     = x0;
#else
        r     = zls_CvtCoordsToRadius1D( x0);
#endif !GRAVZFIELD
        phi0  = zlsg_phi ( r , gp );
        
        x1    = xx[i].c;
#ifdef GRAVZFIELD
        r     = x1;
#else
        r     = zls_CvtCoordsToRadius1D( x1);
#endif !GRAVZFIELD
        phi1  = zlsg_phi ( r , gp );
        
        x2    = xx[i].r;
#ifdef GRAVZFIELD
        r     = x2;
#else
        r     = zls_CvtCoordsToRadius1D( x2);
#endif !GRAVZFIELD
        phi2  = zlsg_phi ( r , gp );
        
        f.p[i].a   = kZLS_sixth*(4.0*phi1 + phi0 + phi2);
        f.p[i].al  = phi0;
        f.p[i].ar  = phi2;
#endif
      }
#endif SYM1D
      
#ifdef SYM2D
      
      for(j = 0; j< ntj; j++){
        
#if (FIELD_RECON_ORDER == 1 )
        yc[0]    = yy[j].l; // min y
        yc[1]    = yc[0];
        yc[2]    = yy[j].r; // max y
        yc[3]    = yc[2];
#endif
#if (FIELD_RECON_ORDER == 2 )
        yc[0]    = yy[j].l; // min
        yc[1]    =    yc[0];
        yc[2]    =    yc[0];
        yc[3]    = yy[j].c; // mid
        yc[4]    =    yc[3];
        yc[5]    =    yc[3];
        yc[6]    = yy[j].r; // max
        yc[7]    =    yc[6];
        yc[8]    =    yc[6];
#endif
        
        for(i = 0; i< nti; i++){
          
#if (FIELD_RECON_ORDER == 1 )
          xc[0]    = xx[i].l; // one per left and right
          xc[1]    = xx[i].r;
          for ( c = 2; c<4; c++){
            xc[c] = xc[c-2];
          }
#endif
#if (FIELD_RECON_ORDER == 2 )
          xc[0]    = xx[i].l; // one per left,centre and right
          xc[1]    = xx[i].c;
          xc[2]    = xx[i].r;
          for ( c = 3; c<9; c++){
            xc[c] = xc[c-3];
          }
#endif
          
#if (FIELD_RECON_ORDER == 1 )
          
          // collect phi and component fractions
          
          for ( c = 0; c< 4; c++){
#ifdef GRAVZFIELD
            r     = yc[c] ;
#else
            r     = zls_CvtCoordsToRadius2D( xc[c], yc[c] );
#endif !GRAVZFIELD
            phi[c]  = zlsg_phi ( r , gp );
          }
          
          phiXl = 0.5*(phi[0] + phi[2]);
          phiXr = 0.5*(phi[1] + phi[3]);
          
          PhiC = 0.5*( phiXl + phiXr);
          
          f.p[j][i].a [kXDim]  = PhiC;
          f.p[j][i].al[kXDim]  = phiXl;
          f.p[j][i].ar[kXDim]  = phiXr;
          
          phiYl = 0.5*(phi[0] + phi[1]);
          phiYr = 0.5*(phi[2] + phi[3]);
          
          // f.p[j][i].a [kYDim]  = 0.5*( phiYl + phiYr);
          f.p[j][i].a [kYDim]  = PhiC;
          f.p[j][i].al[kYDim]  = phiYl;
          f.p[j][i].ar[kYDim]  = phiYr;
          
#endif
#if (FIELD_RECON_ORDER == 2 )
          
          for ( c = 0; c< 9; c++){
#ifdef GRAVZFIELD
            r     = yc[c] ;
#else
            r     = zls_CvtCoordsToRadius2D(  xc[c], yc[c] );
#endif !GRAVZFIELD
            phi[c]  = zlsg_phi ( r , gp );
          }
          
          // total phi magnitude conserved integral average:
          // interpolation is done on the projected values so the
          // projection is averaged properly as well. esp if cell is large
          // compared to r
          
          // <X>  interpolation
          
          phiXl = kZLS_sixth*(4.0*phi[3] + phi[0] + phi[6]);
          phiXc = kZLS_sixth*(4.0*phi[4] + phi[1] + phi[7]);
          phiXr = kZLS_sixth*(4.0*phi[5] + phi[2] + phi[8]);
          
          PhiC = kZLS_sixth*(4.0*phiXc + phiXl + phiXr);
          
          f.p[j][i].a          = PhiC;
          
          f.p[j][i].al[kXDim]  = phiXl;
          f.p[j][i].ar[kXDim]  = phiXr;
          
          phiYl = kZLS_sixth*(4.0*phi[1] + phi[0] + phi[2]);
         // phiYc = kZLS_sixth*(4.0*phi[4] + phi[3] + phi[5]);
          phiYr = kZLS_sixth*(4.0*phi[7] + phi[6] + phi[8]);
          
          //f.p[j][i].a [kYDim]  = kZLS_sixth*(4.0*phiYc + phiYl + phiYr);

          f.p[j][i].al[kYDim]  = phiYl;
          f.p[j][i].ar[kYDim]  = phiYr;
          
#endif
          
        }
      }
#endif SYM2D
      
#ifdef SYM3D
      
      for( k = 0; k< ntk; k++){
        
#if (FIELD_RECON_ORDER == 1 )
        zc[0]    = zz[k].l; // lower z-level 
        for ( c = 1; c<4; c++){
          zc[c] = zc[c-1];
        }
        zc[4]    = zz[k].r; // upper z-level 
        for ( c = 5; c<8; c++){
          zc[c] = zc[c-1];
        }
#endif
#if (FIELD_RECON_ORDER == 2 )
        zc[0]    = zz[k].l; // lower z-level 
        for ( c = 1; c<9; c++){
          zc[c] = zc[c-1];
        }
        zc[9]    = zz[k].c; // middle z-level 
        for ( c = 10; c<18; c++){
          zc[c] = zc[c-1];
        }
        zc[18]    = zz[k].r; // middle z-level 
        for ( c = 19; c<27; c++){
          zc[c] = zc[c-1];
        }
#endif
        
        for(j = 0; j< ntj; j++){
          
#if (FIELD_RECON_ORDER == 1 )
          yc[0]    = yy[j].l; //min
          yc[1]    = yc[0];
          yc[4]    = yc[0];
          yc[5]    = yc[0];
          
          yc[2]    = yy[j].r; //max
          yc[3]    = yc[2];
          yc[6]    = yc[2];
          yc[7]    = yc[2];
#endif
#if (FIELD_RECON_ORDER == 2 )
          yc[0]    = yy[j].l; // min 
          yc[1]    = yc[0]; 
          yc[2]    = yc[0]; 
          
          yc[9]    = yc[0]; 
          yc[10]   = yc[0]; 
          yc[11]   = yc[0]; 
          
          yc[18]   = yc[0]; 
          yc[19]   = yc[0]; 
          yc[20]   = yc[0]; 
          
          yc[3]    = yy[j].c; // mid y plane 
          yc[4]    = yc[3]; 
          yc[5]    = yc[3]; 
          
          yc[12]   = yc[3]; 
          yc[13]   = yc[3]; 
          yc[14]   = yc[3]; 
          
          yc[21]   = yc[3]; 
          yc[22]   = yc[3]; 
          yc[23]   = yc[3]; 
          
          yc[6]    = yy[j].r; // max 
          yc[7]    = yc[6]; 
          yc[8]    = yc[6]; 
          
          yc[15]   = yc[6]; 
          yc[16]   = yc[6]; 
          yc[17]   = yc[6]; 
          
          yc[24]   = yc[6]; 
          yc[25]   = yc[6]; 
          yc[26]   = yc[6]; 
          
#endif
          
          for(i = 0; i< nti; i++){
            
#if (FIELD_RECON_ORDER == 1 )
            xc[0]    = xx[i].l; // one per left,centre and right
            xc[1]    = xx[i].r;
            for ( c = 2; c<8; c++){
              xc[c] = xc[c-2];
            }
#endif
#if (FIELD_RECON_ORDER == 2 )
            xc[0]    = xx[i].l; // one per left,centre and right
            xc[1]    = xx[i].c;
            xc[2]    = xx[i].r;
            for ( c = 3; c<27; c++){
              xc[c] = xc[c-3];
            }
#endif
            
#if (FIELD_RECON_ORDER == 1 )
            
            // collect phi and component fractions
            for ( c = 0; c< 8; c++){
#ifdef GRAVZFIELD
              r     = zc[c] ;
#else
              r     = zls_CvtCoordsToRadius3D(  xc[c], yc[c], zc[c] );
#endif !GRAVZFIELD
              phi[c]  = zlsg_phi ( r , gp );
            }
            
            // magnitude, conserved integral average:
            
            phiXl = 0.25*(phi[0] + phi[2] + phi[4] + phi[6] );
            phiXr = 0.25*(phi[1] + phi[3] + phi[5] + phi[7] );
            
            f.p[k][j][i].a          =  0.5*( phiXl + phiXr);
            f.p[k][j][i].al[kXDim]  = phiXl;
            f.p[k][j][i].ar[kXDim]  = phiXr;
            
            // <Y> component interpolation in Y Direction
            
            phiYl = 0.25*(phi[0] + phi[1] + phi[4] + phi[5] );
            phiYr = 0.25*(phi[2] + phi[3] + phi[6] + phi[7] );
            
            //f.p[k][j][i].a [kYDim]  =  0.5*( phiYl + phiYr);

            f.p[k][j][i].al[kYDim]  = phiYl;
            f.p[k][j][i].ar[kYDim]  = phiYr;
            
            // <Z> component interpolation in Z Direction
            
            phiZl = 0.25*(phi[0] + phi[1] + phi[2] + phi[3] );
            phiZr = 0.25*(phi[4] + phi[5] + phi[6] + phi[7] );
            
            //f.p[k][j][i].a [kZDim]  =  0.5*( phiZl + phiZr);

            f.p[k][j][i].al[kZDim]  = phiZl;
            f.p[k][j][i].ar[kZDim]  = phiZr;
            
#endif
#if (FIELD_RECON_ORDER == 2 )
            
            for ( c = 0; c< 27; c++){
#ifdef GRAVZFIELD
              r     = zc[c] ;
#else
              r     = zls_CvtCoordsToRadius3D(xc[c], yc[c], zc[c]);
#endif !GRAVZFIELD
              phi[c]  = zlsg_phi ( r , gp );
            }
            
            // magnitude, conserved integral average:
            
            phiXl = kZLS_36th*( 16.0* phi[12]
                               + 4.0*(phi[3] + phi[9] + phi[15] + phi[21])
                               +     (phi[0] + phi[6] + phi[18] + phi[24])
                               );
            
            phiXc = kZLS_36th*( 16.0* phi[13]
                               + 4.0*(phi[4] + phi[10] + phi[16] + phi[22])
                               +     (phi[1] + phi[7 ] + phi[19] + phi[25])
                               );
            
            phiXr = kZLS_36th*( 16.0* phi[14] 
                               + 4.0*(phi[5] + phi[11] + phi[17] + phi[23])
                               +     (phi[2] + phi[8 ] + phi[20] + phi[26])
                               );
                               
            
            f.p[k][j][i].a  = kZLS_sixth*(4.0*phiXc+ phiXl + phiXr);
            
            // interpolation is done on the projected values so the
            // projection is averaged properly as well. esp if cell is large
            // compared to r
            
            // in X directions:
            
            f.p[k][j][i].al[kXDim]  = phiXl;
            f.p[k][j][i].ar[kXDim]  = phiXr;
            
            // <Y> component interpolation
            // in Y directions:
            
            phiYl = kZLS_36th*( 16.0* phi[10]  
                               + 4.0*(phi[1]  + phi[9]  + phi[11]  + phi[19] )
                               +     (phi[0]  + phi[2]  + phi[18]  + phi[20] )
                               );
//            phiYc = kZLS_36th*( 16.0* phi[13]  
//                               + 4.0*(phi[4]  + phi[12]  + phi[14]  + phi[22] )
//                               +     (phi[3]  + phi[5 ]  + phi[11]  + phi[23] )
//                               );
            
            phiYr = kZLS_36th*( 16.0* phi[16]  
                               + 4.0*(phi[7]  + phi[15]  + phi[17]  + phi[25] )
                               +     (phi[6]  + phi[8 ]  + phi[24]  + phi[26] )
                               );
            
           //f.p[k][j][i].a [kYDim]  = kZLS_sixth*(4.0*phiYc+ phiYl + phiYr); // should equal kXDim value
            
            f.p[k][j][i].al[kYDim]  = phiYl;
            f.p[k][j][i].ar[kYDim]  = phiYr;
            
            // <Z> component interpolation
            // in Z directions:
            
            phiZl = kZLS_36th*( 16.0* phi[4 ]  
                               + 4.0*(phi[1]  + phi[3]  + phi[5 ]  + phi[7 ] )
                               +     (phi[0]  + phi[2]  + phi[6 ]  + phi[8 ] )
                               );
            
//            phiZc = kZLS_36th*( 16.0* phi[13]  
//                               + 4.0*(phi[10]  + phi[12]  + phi[14]  + phi[16] )
//                               +     (phi[9 ]  + phi[11]  + phi[15]  + phi[17] )
//                               );
            
            phiZr = kZLS_36th*( 16.0* phi[22]  
                               + 4.0*(phi[19]  + phi[21]  + phi[23]  + phi[25] )
                               +     (phi[18]  + phi[20]  + phi[24]  + phi[26] )
                               );
            
           // f.p[k][j][i].a [kZDim]  = kZLS_sixth*(4.0*phiZc+ phiZl + phiZr); // should equal kXDim value

            f.p[k][j][i].al[kZDim]  = phiZl;
            f.p[k][j][i].ar[kZDim]  = phiZr;
            
#endif
            
          } //i
        } // j
      } // k
#endif SYM3D
      
    } // f
		
  }
	
	return (err);
}

#endif GRAVITY_POTENTIAL

// version where field represents the derivative of the potential fiel: ie
// the accelleration

#ifdef GRAVITY_ACCELL
int zlsg_InitZoneGravityAccellField( ZonePtr z){
  
  int err = noErr;
  
  Real r;
  
#ifdef SYM1D
  ACoordPtr xx  = z->zoneGrid.x;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 2 end points in 1D
  Real x0, x1;
  Real dPhi0, dPhi1;
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 3 points in 1D
  Real x0, x1, x2;
  Real dPhi0, dPhi1, dPhi2;
#endif  
  
	Integer nti   = z->zoneGrid.nt[kXDim];
  Integer i = 0;
#endif SYM1D
  
#ifdef SYM2D
  ACoordPtr xx  = z->zoneGrid.x;
  ACoordPtr yy  = z->zoneGrid.y;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 4 vertex points in 2D
  // 
  Integer c; // counter for points to sum
  Real xc[4]; // only 2 values but duped to make loop simple:, xl, xr, xl, xr
  Real yc[4]; // only 2 values but duped to make loop simple, yl, yr, yl, yr
  Real dPhi[4]; // total dphi at each corner (x,y)
  Real dPhi_i[4]; // dphi at each corner (x,y) in i direction
  Real dPhi_j[4]; // dphi at each corner (x,y) in j direction
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 9 points in 2D!
  Integer c; // counter for points to sum
  Real xc[9]; //(xl, xc, xr)*3 planes
  Real yc[9]; //(yl, yc, yr)*3 planes
  Real dPhi[9]; // total dphi at each corner (x,y)
  Real dPhi_i[9]; // dphi at each corner (x,y) in i direction
  Real dPhi_j[9]; // dphi at each corner (x,y) in j direction
#endif  
  
	Integer nti   = z->zoneGrid.nt[kXDim];
	Integer ntj   = z->zoneGrid.nt[kYDim];
  Integer i = 0;
  Integer j = 0;
  
  Real fi, fj; // projected components of dPhi_r in coordinate system
  
  // these are the interpolated 2D phi values already combined
  // with the component fractions and conservatively averaged
  
  Real dPhiXl; // left   phi in X direction
#if (FIELD_RECON_ORDER == 2 )
  Real dPhiXc; // centre phi in X direction, used in order 2
  Real dPhiYc; // centre phi in Y direction, used in order 2
#endif
  Real dPhiXr; // right  phi in X direction
  
  Real dPhiYl; // left   phi in Y direction
  Real dPhiYr; // right  phi in Y direction
  
#endif SYM2D
  
#ifdef SYM3D
  ACoordPtr xx  = z->zoneGrid.x;
  ACoordPtr yy  = z->zoneGrid.y;
  ACoordPtr zz  = z->zoneGrid.z;
  
#if (FIELD_RECON_ORDER == 1 )
  // linear conservative cell reconstruction requires 8 vertex points in 3D!
  Integer c; // counter for points to sum
  Real xc[8]; // two values, xl and xr, but copied for each pair of vetexes, xl, xr, xl ,xr
  Real yc[8];
  Real zc[8];
  Real dPhi[8];
  Real dPhi_i[8];
  Real dPhi_j[8];
  Real dPhi_k[8];
#endif  
#if (FIELD_RECON_ORDER == 2 )
  // ppm conservative cell reconstruction requires 27 points in 3D!
  // common points between cells could be shared, but it is messy in 3D
  // and initially we are just going to get it working with lots of calls.
  Integer c; // counter for points to sum
  Real xc[27];
  Real yc[27];
  Real zc[27];
  Real dPhi[27];
  Real dPhi_i[27];
  Real dPhi_j[27];
  Real dPhi_k[27];
#endif  
  
  Integer nti   = z->zoneGrid.nt[kXDim];
  Integer ntj   = z->zoneGrid.nt[kYDim];
  Integer ntk   = z->zoneGrid.nt[kZDim];
  
  Integer i = 0;
  Integer j = 0;
  Integer k = 0;
  Real fi, fj, fk; // projected components of dPhi_r in coordinate system
  
#if (FIELD_RECON_ORDER == 2 )
  Real dPhiXc; // centre dPhi used in order 2
  Real dPhiYc; // centre dPhi used in order 2
  Real dPhiZc; // centre dPhi used in order 2
#endif 
  
  Real dPhiXl; // left   dPhi in X direction
  Real dPhiXr; // right  dPhi in X direction
  
  Real dPhiYl; // left   dPhi in Y direction
  Real dPhiYr; // right  dPhi in Y direction
  
  Real dPhiZl; // left   dPhi in Z direction
  Real dPhiZr; // right  dPhi in Z direction
  
#endif SYM3D
  
  PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global gravity parameters
  if ((phys != NULL)&&( z != NULL)) {
    
#if (ZLS_DEBUG > 0)
    zlsio_Printf( " Zone Field Initialization: %2.2d. \n", z->zoneID );
#endif
    
    PotentialPtr gp  = &(phys->gravPotential);
    
    Field  f;
    zls_GetZoneField(&f, z);
    
    if ( f.p != NULL ) {
#ifdef SYM1D
      for(i = 0; i< nti; i++){
        
#if (FIELD_RECON_ORDER == 1 )
        // two calls
        x0    = xx[i].l;
#ifdef GRAVZFIELD
        r     = x0;
#else
        r     = zls_CvtCoordsToRadius1D( x0);
#endif !GRAVZFIELD
        
        dPhi0  = zlsg_dphi ( r , gp );
        
        x1    = xx[i].r;
#ifdef GRAVZFIELD
        r     = x1;
#else
        r     = zls_CvtCoordsToRadius1D( x1);
#endif !GRAVZFIELD
        dPhi1  = zlsg_dphi ( r , gp );
        
        f.p[i].a   = 0.5*(dPhi0 + dPhi1);
        f.p[i].al  = dPhi0;
        f.p[i].ar  = dPhi1;
#endif
#if (FIELD_RECON_ORDER == 2 )
        // three calls
        x0    = xx[i].l;
#ifdef GRAVZFIELD
        r     = x0;
#else
        r     = zls_CvtCoordsToRadius1D( x0);
#endif !GRAVZFIELD
        dPhi0  = zlsg_dphi ( r , gp );
        
        x1    = xx[i].c;
#ifdef GRAVZFIELD
        r     = x1;
#else
        r     = zls_CvtCoordsToRadius1D( x1);
#endif !GRAVZFIELD
        dPhi1  = zlsg_dphi ( r , gp );
        
        x2    = xx[i].r;
#ifdef GRAVZFIELD
        r     = x2;
#else
        r     = zls_CvtCoordsToRadius1D( x2);
#endif !GRAVZFIELD
        dPhi2  = zlsg_dphi ( r , gp );
        
        f.p[i].a   = kZLS_sixth*(4.0*dPhi1 + dPhi0 + dPhi2);
        f.p[i].al  = dPhi0;
        f.p[i].ar  = dPhi2;
#endif
      }
#endif SYM1D
      
#ifdef SYM2D
      
      for(j = 0; j< ntj; j++){
        
#if (FIELD_RECON_ORDER == 1 )
        yc[0]    = yy[j].l; // min y
        yc[1]    = yc[0];
        yc[2]    = yy[j].r; // max y
        yc[3]    = yc[2];
#endif
#if (FIELD_RECON_ORDER == 2 )
        yc[0]    = yy[j].l; // min
        yc[1]    =    yc[0];
        yc[2]    =    yc[0];
        yc[3]    = yy[j].c; // mid
        yc[4]    =    yc[3];
        yc[5]    =    yc[3];
        yc[6]    = yy[j].r; // max
        yc[7]    =    yc[6];
        yc[8]    =    yc[6];
#endif
        
        for(i = 0; i< nti; i++){
          
#if (FIELD_RECON_ORDER == 1 )
          xc[0]    = xx[i].l; // one per left and right
          xc[1]    = xx[i].r;
          for ( c = 2; c<4; c++){
            xc[c] = xc[c-2];
          }
#endif
#if (FIELD_RECON_ORDER == 2 )
          xc[0]    = xx[i].l; // one per left,centre and right
          xc[1]    = xx[i].c;
          xc[2]    = xx[i].r;
          for ( c = 3; c<9; c++){
            xc[c] = xc[c-3];
          }
#endif
          
#if (FIELD_RECON_ORDER == 1 )
          
          // collect phi and component fractions
          
          for ( c = 0; c< 4; c++){
#ifdef GRAVZFIELD
            r     = zls_CvtCoordsToProjRadius2D(&fi, &fj,   0.0, yc[c] );
#else
            r     = zls_CvtCoordsToProjRadius2D(&fi, &fj, xc[c], yc[c] );
#endif !GRAVZFIELD
            dPhi[c]  = zlsg_dphi ( r , gp );
            dPhi_i[c]  = dPhi[c]*fi;
            dPhi_j[c]  = dPhi[c]*fj;
          }
          
          dPhiXl = 0.5*(dPhi_i[0] + dPhi_i[2]);
          dPhiXr = 0.5*(dPhi_i[1] + dPhi_i[3]);
          
          f.p[j][i].a    = 0.5*( dPhiXl + dPhiXr);
          f.p[j][i].al[kXDim]  = dPhiXl;
          f.p[j][i].ar[kXDim]  = dPhiXr;
          
          dPhiYl = 0.5*(dPhi_j[0] + dPhi_j[1]);
          dPhiYr = 0.5*(dPhi_j[2] + dPhi_j[3]);
          
          f.p[j][i].a [kYDim]  = 0.5*( dPhiYl + dPhiYr);
          f.p[j][i].al[kYDim]  = dPhiYl;
          f.p[j][i].ar[kYDim]  = dPhiYr;
          
#endif
#if (FIELD_RECON_ORDER == 2 )
          
          for ( c = 0; c< 9; c++){
#ifdef GRAVZFIELD
            r     = zls_CvtCoordsToProjRadius2D(&fi, &fj,   0.0, yc[c] );
#else
            r     = zls_CvtCoordsToProjRadius2D(&fi, &fj, xc[c], yc[c] );
#endif !GRAVZFIELD
            dPhi[c]    = zlsg_dphi ( r , gp );
            dPhi_i[c]  = dPhi[c]*fi;
            dPhi_j[c]  = dPhi[c]*fj;
          }
          
          // total phi magnitude conserved integral average:
          // interpolation is done on the projected values so the
          // projection is averaged properly as well. esp if cell is large
          // compared to r
          
          // <X>  interpolation
          
          dPhiXl = kZLS_sixth*(4.0*dPhi_i[3] + dPhi_i[0] + dPhi_i[6]);
          dPhiXc = kZLS_sixth*(4.0*dPhi_i[4] + dPhi_i[1] + dPhi_i[7]);
          dPhiXr = kZLS_sixth*(4.0*dPhi_i[5] + dPhi_i[2] + dPhi_i[8]);
          
          f.p[j][i].a [kXDim]  = kZLS_sixth*(4.0*dPhiXc + dPhiXl + dPhiXr);
          f.p[j][i].al[kXDim]  = dPhiXl;
          f.p[j][i].ar[kXDim]  = dPhiXr;
          
          dPhiYl = kZLS_sixth*(4.0*dPhi_j[1] + dPhi_j[0] + dPhi_j[2]);
          dPhiYc = kZLS_sixth*(4.0*dPhi_j[4] + dPhi_j[3] + dPhi_j[5]);
          dPhiYr = kZLS_sixth*(4.0*dPhi_j[7] + dPhi_j[6] + dPhi_j[8]);
          
          f.p[j][i].a [kYDim]  = kZLS_sixth*(4.0*dPhiYc + dPhiYl + dPhiYr);
          f.p[j][i].al[kYDim]  = dPhiYl;
          f.p[j][i].ar[kYDim]  = dPhiYr;
          
#endif
          
        }
      }
#endif SYM2D
      
#ifdef SYM3D
      
      for( k = 0; k< ntk; k++){
        
#if (FIELD_RECON_ORDER == 1 )
        zc[0]    = zz[k].l; // lower z-level 
        for ( c = 1; c<4; c++){
          zc[c] = zc[c-1];
        }
        zc[4]    = zz[k].r; // upper z-level 
        for ( c = 5; c<8; c++){
          zc[c] = zc[c-1];
        }
#endif
#if (FIELD_RECON_ORDER == 2 )
        zc[0]    = zz[k].l; // lower z-level 
        for ( c = 1; c<9; c++){
          zc[c] = zc[c-1];
        }
        zc[9]    = zz[k].c; // middle z-level 
        for ( c = 10; c<18; c++){
          zc[c] = zc[c-1];
        }
        zc[18]    = zz[k].r; // middle z-level 
        for ( c = 19; c<27; c++){
          zc[c] = zc[c-1];
        }
#endif
        
        for(j = 0; j< ntj; j++){
          
#if (FIELD_RECON_ORDER == 1 )
          yc[0]    = yy[j].l; //min
          yc[1]    = yc[0];
          yc[4]    = yc[0];
          yc[5]    = yc[0];
          
          yc[2]    = yy[j].r; //max
          yc[3]    = yc[2];
          yc[6]    = yc[2];
          yc[7]    = yc[2];
#endif
#if (FIELD_RECON_ORDER == 2 )
          yc[0]    = yy[j].l; // min 
          yc[1]    = yc[0]; 
          yc[2]    = yc[0]; 
          
          yc[9]    = yc[0]; 
          yc[10]   = yc[0]; 
          yc[11]   = yc[0]; 
          
          yc[18]   = yc[0]; 
          yc[19]   = yc[0]; 
          yc[20]   = yc[0]; 
          
          yc[3]    = yy[j].c; // mid y plane 
          yc[4]    = yc[3]; 
          yc[5]    = yc[3]; 
          
          yc[12]   = yc[3]; 
          yc[13]   = yc[3]; 
          yc[14]   = yc[3]; 
          
          yc[21]   = yc[3]; 
          yc[22]   = yc[3]; 
          yc[23]   = yc[3]; 
          
          yc[6]    = yy[j].r; // max 
          yc[7]    = yc[6]; 
          yc[8]    = yc[6]; 
          
          yc[15]   = yc[6]; 
          yc[16]   = yc[6]; 
          yc[17]   = yc[6]; 
          
          yc[24]   = yc[6]; 
          yc[25]   = yc[6]; 
          yc[26]   = yc[6]; 
          
#endif
          
          for(i = 0; i< nti; i++){
            
#if (FIELD_RECON_ORDER == 1 )
            xc[0]    = xx[i].l; // one per left,centre and right
            xc[1]    = xx[i].r;
            for ( c = 2; c<8; c++){
              xc[c] = xc[c-2];
            }
#endif
#if (FIELD_RECON_ORDER == 2 )
            xc[0]    = xx[i].l; // one per left,centre and right
            xc[1]    = xx[i].c;
            xc[2]    = xx[i].r;
            for ( c = 3; c<27; c++){
              xc[c] = xc[c-3];
            }
#endif
            
#if (FIELD_RECON_ORDER == 1 )
            
            // collect phi and component fractions
            for ( c = 0; c< 8; c++){
#ifdef GRAVZFIELD
              r     = zls_CvtCoordsToProjRadius2D(&fi, &fj, &fk,   0.0,   0.0, zc[c] );
#else
              r     = zls_CvtCoordsToProjRadius2D(&fi, &fj, &fk, xc[c], yc[c], zc[c] );
#endif !GRAVZFIELD
              dPhi[c]  = zlsg_dphi ( r , gp );
              dPhi_i[c]  = dPhi[c]*fi;
              dPhi_j[c]  = dPhi[c]*fj;
              dPhi_k[c]  = dPhi[c]*fk;
            }
            
            // magnitude, conserved integral average:
            
            dPhiXl = 0.25*(dPhi_i[0] + dPhi_i[2] + dPhi_i[4] + dPhi_i[6] );
            dPhiXr = 0.25*(dPhi_i[1] + dPhi_i[3] + dPhi_i[5] + dPhi_i[7] );
            
            f.p[k][j][i].a [kXDim]  =  0.5*( dPhiXl + dPhiXr);
            f.p[k][j][i].al[kXDim]  = dPhiXl;
            f.p[k][j][i].ar[kXDim]  = dPhiXr;
            
            // <Y> component interpolation in Y Direction
            
            dPhiYl = 0.25*(dPhi_j[0] + dPhi_j[1] + dPhi_j[4] + dPhi_j[5] );
            dPhiYr = 0.25*(dPhi_j[2] + dPhi_j[3] + dPhi_j[6] + dPhi_j[7] );
            
            f.p[k][j][i].a [kYDim]  = 0.5*( dPhiYl + dPhiYr);
            f.p[k][j][i].al[kYDim]  = dPhiYl;
            f.p[k][j][i].ar[kYDim]  = dPhiYr;
            
            // <Z> component interpolation in Z Direction
            
            dPhiZl = 0.25*(dPhi_k[0] + dPhi_k[1] + dPhi_k[2] + dPhi_k[3] );
            dPhiZr = 0.25*(dPhi_k[4] + dPhi_k[5] + dPhi_k[6] + dPhi_k[7] );
            
            f.p[k][j][i].a [kZDim]  = 0.5*( dPhiZl + dPhiZr);
            f.p[k][j][i].al[kZDim]  = dPhiZl;
            f.p[k][j][i].ar[kZDim]  = dPhiZr;
            
#endif
#if (FIELD_RECON_ORDER == 2 )
            
            for ( c = 0; c< 27; c++){
#ifdef GRAVZFIELD
              r     = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk,   0.0,   0.0, zc[c] );
#else
              r     = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, xc[c], yc[c], zc[c] );
#endif !GRAVZFIELD
              dPhi[c]  = zlsg_dphi ( r , gp );
              dPhi_i[c]  = dPhi[c]*fi;
              dPhi_j[c]  = dPhi[c]*fj;
              dPhi_k[c]  = dPhi[c]*fk;
            }
            
            // magnitude, conserved integral average:
            
            dPhiXl = kZLS_36th*( 16.0* dPhi_i[12]
                                + 4.0*(dPhi_i[3] + dPhi_i[9] + dPhi_i[15] + dPhi_i[21])
                                +     (dPhi_i[0] + dPhi_i[6] + dPhi_i[18] + dPhi_i[24])
                                );
            
            dPhiXc = kZLS_36th*( 16.0* dPhi_i[13]
                                + 4.0*(dPhi_i[4] + dPhi_i[10] + dPhi_i[16] + dPhi_i[22])
                                +     (dPhi_i[1] + dPhi_i[7 ] + dPhi_i[19] + dPhi_i[25])
                                );
            
            dPhiXr = kZLS_36th*( 16.0* dPhi_i[14] 
                                + 4.0*(dPhi_i[5] + dPhi_i[11] + dPhi_i[17] + dPhi_i[23])
                                +     (dPhi_i[2] + dPhi_i[8 ] + dPhi_i[20] + dPhi_i[26])
                                );
            
            f.p[k][j][i].a [kXDim]  = kZLS_sixth*(4.0*dPhiXc+ dPhiXl + dPhiXr);
            f.p[k][j][i].al[kXDim]  = dPhiXl;
            f.p[k][j][i].ar[kXDim]  = dPhiXr;
            
            // <Y> component interpolation
            // in Y directions:
            
            dPhiYl = kZLS_36th*( 16.0* dPhi_j[10]  
                                + 4.0*(dPhi_j[1]  + dPhi_j[9]  + dPhi_j[11]  + dPhi_j[19] )
                                +     (dPhi_j[0]  + dPhi_j[2]  + dPhi_j[18]  + dPhi_j[20] )
                                );
            
            dPhiYc = kZLS_36th*( 16.0* dPhi_j[13]  
                                + 4.0*(dPhi_j[4]  + dPhi_j[12]  + dPhi_j[14]  + dPhi_j[22] )
                                +     (dPhi_j[3]  + dPhi_j[5 ]  + dPhi_j[11]  + dPhi_j[23] )
                                );
            
            dPhiYr = kZLS_36th*( 16.0* dPhi_j[16]  
                                + 4.0*(dPhi_j[7]  + dPhi_j[15]  + dPhi_j[17]  + dPhi_j[25] )
                                +     (dPhi_j[6]  + dPhi_j[8 ]  + dPhi_j[24]  + dPhi_j[26] )
                                );
            
            f.p[k][j][i].a [kYDim]  = kZLS_sixth*(4.0*dPhiYc+ dPhiYl + dPhiYr);
            f.p[k][j][i].al[kYDim]  = dPhiYl;
            f.p[k][j][i].ar[kYDim]  = dPhiYr;
            
            // <Z> component interpolation
            // in Z directions:
            
            dPhiZl = kZLS_36th*( 16.0* dPhi_k[4 ]  
                                + 4.0*(dPhi_k[1]  + dPhi_k[3]  + dPhi_k[5 ]  + dPhi_k[7 ] )
                                +     (dPhi_k[0]  + dPhi_k[2]  + dPhi_k[6 ]  + dPhi_k[8 ] )
                                );
            
            dPhiZc = kZLS_36th*( 16.0* dPhi_k[13]  
                                + 4.0*(dPhi_k[10]  + dPhi_k[12]  + dPhi_k[14]  + dPhi_k[16] )
                                +     (dPhi_k[9 ]  + dPhi_k[11]  + dPhi_k[15]  + dPhi_k[17] )
                                );
            
            dPhiZr = kZLS_36th*( 16.0* dPhi_k[22]  
                                + 4.0*(dPhi_k[19]  + dPhi_k[21]  + dPhi_k[23]  + dPhi_k[25] )
                                +     (dPhi_k[18]  + dPhi_k[20]  + dPhi_k[24]  + dPhi_k[26] )
                                );
            
            f.p[k][j][i].a [kZDim]  = kZLS_sixth*(4.0*dPhiZc+ dPhiZl + dPhiZr);
            f.p[k][j][i].al[kZDim]  = dPhiZl;
            f.p[k][j][i].ar[kZDim]  = dPhiZr;
            
#endif
            
          } //i
        } // j
      } // k
#endif SYM3D
      
    } // f
		
  }
	
	return (err);
}
#endif GRAVITY_ACCELL

int zlsg_InitGravityPotential(PotentialPtr gp){
  
  int err = noErr;
  if ( gp != NULL) {
    
    gp->gravTypeFlags = 0;
    
		gp->M          = 1.0;
		gp->rN0        = 0.001;
		
		gp->r_ss       = 1.0;
		gp->invr_ss    = 1.0;
		gp->m_ss       = 1.0;
		
		gp->r_iso      = 1.0;
		gp->invr_iso   = 1.0;
		gp->s_iso      = 1.0;
		gp->s2_iso     = 1.0;
		gp->invs_iso   = 1.0;
		gp->invs2_iso  = 1.0;
		
		gp->r_s        = 1.0;
		gp->v_s        = 1.0/zls_sqrt(2.0);  // circular rather than dispersion, smaller by sqrt(2), ie v^2 = GM/r
		gp->rho_s      = 1.0;  // derived if needed
    
    //
// test parameters:
// M_vir = 1.2e9Msol, 2.386920E+42 g
// x_vir = 26.60904 from comsological sims
// f_m(x_vir) = 6.870434E+01 for data file: einlogwlogdw.txt
// Delta = 340
// rho_u = 2.76e-30 g/cm^3
//
// from this is derived:
// r_s = 1031.345 pc , 3.182422E+21 cm, independent of potential form
    //
    // Einasto:
    // f_m(x_vir) = 42.46255 for data file: einlogwlogdw.txt
    // v_s = 1.085686E+06 cm/s 10.85686 km/s
    // NFW:
    // f_m(x_vir) = 43.53603  for data file: nfwlogwlogdw.txt
    // v_s = 1.07221698E+06 cm/s 10.7221698 km/s
    // HIsothermal:
    // f_m(x_vir) = 68.70434 for data file: hisologwlogdw.txt
    // v_s = 8.535226E+05 cm/s 8.535226 km/s
    // 
    // But all scale their potentials to give the same physical quantities!
    //

		gp->M_vir      = 1.2e9;
		gp->x_vir      = 26.60904;
    gp->f_mvir     = 42.46255;
    gp->f_m1       = 3.75;
    
		gp->delta_rho    = 340;     // usually 200, or 340 for small halos at z=0
		gp->rho_universe = 2.76e-30;  // 2.76e-30 gm/cm^3 at z= 0
    
		gp->v2_s       = 1.0;  // GM/r
		gp->invv_s     = 1.0;
		gp->invv2_s    = 1.0;
		
    gp->gravRWspl   = NULL;
		gp->gravWspl    = NULL;
		gp->gravWspl2   = NULL;
		gp->gravdWspl   = NULL;
		gp->gravdWspl2  = NULL;
		gp->gravNWpline = 0;
    
		gp->d_diskZ       = 1.0; // D
		gp->s0_diskZ      = 1.0; // Surface density
		gp->re_diskZ      = 0.0; // Volume Density of Halo/Thick Component in midplane
    
  } else { 
    err = memErr;
  }
  return(err);
}


#ifdef PTHREADS
#include  <pthread.h>
int gSweepGravityLock = 0;
pthread_mutex_t gSweepGravityMutex    = PTHREAD_MUTEX_INITIALIZER; // mutex for gSweepLock and comms
pthread_cond_t  gSweepGravityCond     = PTHREAD_COND_INITIALIZER; // clear to sweep, each independent so no mutexes needed on sweeps themselves
pthread_cond_t  gNextSweepGravityCond = PTHREAD_COND_INITIALIZER; // request next sweep trigger or end


int zlsg_QueuePThreadGravityInit ( ZonePtr z );
void *  zlsg_PThreadGravityInit  ( void * vz );
int zlsg_WaitPThreadGravityInit  ( int nThreads ); // conduct sweeper threads

int zlsg_QueuePThreadGravityInit  ( ZonePtr z ) {
  int err = noErr;
  pthread_t thread;
  pthread_attr_t attr;
  
  err = pthread_attr_init(&attr);
  err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  err = pthread_create(&thread, &attr, zlsg_PThreadGravityInit, (void *)z);
  err = pthread_attr_destroy(&attr);
  
  return(err);
}

void *  zlsg_PThreadGravityInit  ( void * vz ) {
  
  int err = noErr;
  ZonePtr z = (ZonePtr)vz;
  
  pthread_mutex_lock            ( &gSweepGravityMutex);
  if (gSweepGravityLock >0) pthread_cond_broadcast  ( &gNextSweepGravityCond );
  pthread_cond_wait             ( &gSweepGravityCond, &gSweepGravityMutex); // first time in, sit and wait for gSweepGravityCond launches
  pthread_mutex_unlock          ( &gSweepGravityMutex);
  
  // init zone of thread in parallel, OK as zones are independent here while zlsg_WaitPThreadGravityInit is mastering
  
#ifdef GRAVITY_POTENTIAL 
  err = zlsg_InitZoneGravityPotentialField( z );
#endif GRAVITY_POTENTIAL
#ifdef GRAVITY_ACCELL 
  err = zlsg_InitZoneGravityAccellField( z );
#endif GRAVITY_ACCELL
  
  pthread_mutex_lock             ( &gSweepGravityMutex);
  
  gSweepGravityLock--;     // if gSweepGravityLock hits 0 then it is all over
  pthread_cond_broadcast  ( &gNextSweepGravityCond ); // request next or end of all sweeps (gSweepLock hits 0)
  pthread_mutex_unlock    ( &gSweepGravityMutex);
  pthread_exit(NULL); // thread done
  
}

int zlsg_WaitPThreadGravityInit( int nThreads ){ // conduct sweeper threads
  int err = noErr;
  err = pthread_mutex_lock       ( &gSweepGravityMutex );  
  gSweepGravityLock = nThreads;
  while ( gSweepGravityLock > 0 ) {
    err = pthread_cond_broadcast ( &gSweepGravityCond );
    err = pthread_cond_wait      ( &gNextSweepGravityCond, &gSweepGravityMutex );
  }
  gSweepGravityLock = 0;
  err = pthread_mutex_unlock     ( &gSweepGravityMutex );
  return(err);
}

#endif PTHREADS


int zlsg_InitLevelGravityField( Integer level ){
  int err = noErr;
  
  Integer zone;
  ZonePtr z;
  
  if ((level >= 0) && (level < gNLevels)){
    
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d Gravity Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
    
#ifdef PTHREADS
    Integer nZones   = gLevels[level].nZones; // number of zones to be dealt with, fixed at startup for now, but could change in runttime.
    
    Integer nThreads     = 0;  // number of zones actually threaded
    Integer queueThreads = 0;  // index in queue for batch = 0 .. ZLS_MAXPTHREADS-1
    Integer batchNumber  = 0;  // the index of the batches going through
    
    gSweepGravityLock       = 0;	// global thread counting mutex variable.
    
    Integer nZoneBatches     = nZones / ZLS_MAXPTHREADS;
    Integer nZoneRemainder   = nZones % ZLS_MAXPTHREADS;
    //
    // if nZones > 	ZLS_MAXPTHREADS then run in batches of ZLS_MAXPTHREADS and then
    // finish off any remainder in a < ZLS_MAXPTHREADS lot.
    // as only algorithm zones are inited, if there are a lot of non-algorithm zones
    // present then some batches may not be full.  Keep non algorithm zones to a minimum
    //
    // if nZones == ZLS_MAXPTHREADS then just one batch will run
    //
    // if nZones < ZLS_MAXPTHREADS then just the remainder will run
    //
    //
    // init actual zone counter, increments even if not queued so the level is 
    // fully checked and zones are not missed.  Starts at zone == 0
    zone = 0; 
    
    if ( nZoneBatches > 0 ) {
      
      for( batchNumber = 0; batchNumber <nZoneBatches; batchNumber++){
        
				nThreads = 0;
				
				for ( queueThreads = 0 ; queueThreads < ZLS_MAXPTHREADS; queueThreads++ ) {
					z =  gLevels[level].levelZones[zone++]; // access zone and increment counter
					if (z != NULL){
						if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
							zlsg_QueuePThreadGravityInit( z );
							nThreads++;
						} // z->zoneType == kAlgorithmZoneType
					} // z!= NULL
				}// queue
				
        zlsg_WaitPThreadGravityInit( nThreads ); // wait for each batch
        
				
			} // batchNumber
			
    } // nZoneBatches > 0
    
    if ( nZoneRemainder > 0 ) {
      
      // finish off any remainder, no batch loop needed
      nThreads = 0;
      
      for ( queueThreads = 0 ; queueThreads < nZoneRemainder; queueThreads++ ) {
        z =  gLevels[level].levelZones[zone++]; // access zone and increment counter
        if (z != NULL){
#ifdef MPI	
          if (z->zoneRankID == gMPIrank) {
#endif MPI
            if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
              zlsg_QueuePThreadGravityInit( z );
              nThreads++;
            } // z->zoneType == kAlgorithmZoneType
#ifdef MPI	
          }
#endif MPI
        } // z!= NULL
      }  // queue
      
			zlsg_WaitPThreadGravityInit( nThreads );
		} // nZoneRemainder > 0
#endif PTHREADS
    
#ifndef PTHREADS
    
    
    for ( zone = 0; zone <gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
        // precompute the potential of to the gravity field over the zone
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
#ifdef GRAVITY_POTENTIAL 
            err = zlsg_InitZoneGravityPotentialField( z );
#endif GRAVITY_POTENTIAL
#ifdef GRAVITY_ACCELL 
            err = zlsg_InitZoneGravityAccellField( z );
#endif GRAVITY_ACCELL
            if (err != noErr ) {
              zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Zone Field Failed: %d,  %d \n",zone, err );
              return(err);
            }
          }
#ifdef MPI			
        }
#endif MPI
        
      }
      
    } // zone
    
#endif !PTHREADS
  } else { 
    err = levErr; 
    zlsio_ErrPrintf( " LEVEL ERROR: zlsg_InitLevelGravityField level number %d out of range %d \n", level, gNLevels); 
    return (err); 
  } // (level < gNLevels )
  
  return (err);
}


int zlsg_DisposeGravityPotential(PotentialPtr gp){
  
  int err = noErr;
  
  gp->gravNWpline = 0;
  if (gp->gravRWspl !=NULL)     free(gp->gravRWspl)    ; gp->gravRWspl   = NULL; 
  if (gp->gravWspl  !=NULL)     free(gp->gravWspl)     ; gp->gravWspl    = NULL; 
  if (gp->gravWspl2 !=NULL)     free(gp->gravWspl2)    ; gp->gravWspl2   = NULL; 
  if (gp->gravdWspl !=NULL)     free(gp->gravdWspl)    ; gp->gravdWspl   = NULL; 
  if (gp->gravdWspl2!=NULL)     free(gp->gravdWspl2)   ; gp->gravdWspl2  = NULL; 
  
  return(err);
}

Real  zlsg_phi ( Real r,  PotentialPtr gp){
  
	Real phi    = 0.0;
  
  if (( gp->gravTypeFlags & kPhysKingSpheroidGravityFlag ) == kPhysKingSpheroidGravityFlag ){
    r = zls_abs(r);
    phi += zlsg_phi_sphere( r, gp);
  }
  
  if (( gp->gravTypeFlags & kPhysIsothermalGravityFlag ) == kPhysIsothermalGravityFlag ){
    if ( r > 0.0 ) {
      phi += zlsg_phi_iso    ( r, gp );
    }
  }
  
  if (( gp->gravTypeFlags & kPhysHaloGravityFlag ) == kPhysHaloGravityFlag ){
    if ( r > 0.0 ) {
      phi += zlsg_phi_halo    ( r, gp );
    }
  }
  
  if (( gp->gravTypeFlags & kPhysCosmicHaloGravityFlag ) == kPhysCosmicHaloGravityFlag ){
    if ( r > 0.0 ) {   // exactly same as above, but intialised differently
      phi += zlsg_phi_halo    ( r, gp );
    }
  }
  
  if (( gp->gravTypeFlags & kPhysNewtonianGravityFlag ) == kPhysNewtonianGravityFlag ){
    if ( r > 0.0 ) {
      phi += zlsg_phi_newton ( r, gp);
    }
  }
  
  if (( gp->gravTypeFlags & kPhysUniformGravityFlag ) == kPhysUniformGravityFlag ){
    // pass cartesian coord for r to get a plane || field
    phi += zlsg_phi_Uniform ( r, gp );
  }
  
  if (( gp->gravTypeFlags & kPhysDiskZGravityFlag ) == kPhysDiskZGravityFlag ){
    // use r = y in 2D cartesian, or r = z in 3D cartesian
    phi += zlsg_phi_DiskZ ( r, gp );
  }
  
  if (( gp->gravTypeFlags & kPhysTanhZGravityFlag ) == kPhysTanhZGravityFlag ){
    // use r = y in 2D cartesian, or r = z in 3D cartesian
    phi += zlsg_phi_TanhZ ( r, gp );
  }
  
	return (phi);
}


Real  zlsg_dphi ( Real r,  PotentialPtr gp){
  
	/*
	 c phi - total potential
	 c phi_ss - potential due to stellar spheroid component
	 c phi_iso - potential due to isothermal spheres
	 c r_a - radial axisymmetric coordinate
	 c z - z axisymmetric coordinate
	 c r_sph - spherical r coordinate
	 */
	Real dphi = 0.0;
	
  if (( gp->gravTypeFlags & kPhysKingSpheroidGravityFlag ) == kPhysKingSpheroidGravityFlag ){
    dphi += zlsg_dphi_sphere( r, gp );
  }
  
  if (( gp->gravTypeFlags & kPhysIsothermalGravityFlag ) == kPhysIsothermalGravityFlag ){
    dphi += zlsg_dphi_iso    ( r, gp);
  }
  
  if (( gp->gravTypeFlags & kPhysHaloGravityFlag ) == kPhysHaloGravityFlag ){
    dphi += zlsg_dphi_halo    ( r, gp);
  }
  
  if (( gp->gravTypeFlags & kPhysNewtonianGravityFlag ) == kPhysNewtonianGravityFlag ){
    dphi += zlsg_dphi_newton ( r, gp );
  }
  
  if (( gp->gravTypeFlags & kPhysUniformGravityFlag ) == kPhysUniformGravityFlag ){
    // pass cartesian axis for r for a plane || field
    dphi += zlsg_dphi_Uniform ( r, gp);
  }
  
  if (( gp->gravTypeFlags & kPhysDiskZGravityFlag ) == kPhysDiskZGravityFlag ){
    // pass cartesian axis for r for a plane || field
    dphi += zlsg_dphi_DiskZ ( r, gp);
  }
  
  if (( gp->gravTypeFlags & kPhysTanhZGravityFlag ) == kPhysTanhZGravityFlag ){
    // pass cartesian axis for r for a plane || field
    dphi += zlsg_dphi_DiskZ ( r, gp);
  }
  
	return (dphi);
  
}

Real  zlsg_phi_newton  ( Real r ,  PotentialPtr gp  ){ // M is solar masses
	Real phi = -(k_gmsol*gp->M/gp->rN0);
	if (r > gp->rN0) {
		phi = -(k_gmsol*gp->M/r);
  } 
	return ( phi );
}

Real  zlsg_dphi_newton  ( Real r , PotentialPtr gp  ){ // M is solar masses
	Real dphi = 0.0;
	if (r > gp->rN0) {
		dphi = (k_gmsol*gp->M/(r*r)); 
  }
	return ( dphi );
}

Real  zlsg_phi_sphere  ( Real r , PotentialPtr gp ){
	// approximate fit to King spheroid, valid < 3*rss, m in solar masses
	Real phi = 0.0;
  
	Real  u1, u2;
  
  r = ( r < 0.0 ) ? 0.0: r; // sanity check
	
	u1 = r/gp->r_ss;
	u2 = zls_sqrt(1.00 +u1*u1);
	
  phi = -k_gmsol * gp->m_ss * log(u1+u2)/ (r) ;
  
	return ( phi );
}

Real  zlsg_dphi_sphere  ( Real r , PotentialPtr gp ){
  // m_ss in solar masses, r normalised
	Real dphi = 0.0;
	
	Real  u1, u2, u3;
	dphi = 0.0;
  r = ( r < 0.0 ) ? 0.0: r; // sanity check
	u1 = r/gp->r_ss;
	u2 = zls_sqrt(1.0 +u1*u1);
	u3 = log( u1 + u2);
	
  dphi = -(k_gmsol*gp->m_ss/(r*r))*(u1/u2 - u3);
	
	return ( dphi );
}

Real  zlsg_phi_iso (Real r, PotentialPtr gp ){
	/*c
   c phi_iso - potential due to pure isothermal sphere - exact numerical integration
	 c evaluated by spline fit to numerical integration of iso equations
	 c RMS errors wrt to integration: 
	 c RMS mean relative  error: < 2e-10 form log(r/r0) -4 to +4.0
	 c max  relative  error: 1e-9
	 c Integration specified epsilon 1e-8
	 c
	 c potential specified in terms of dimensionless W, sigma, and r_iso
	 c
	 c transformed to physical units
	 c Note: in ppmlr, Phi is -ve, +ve in notes.
	 c
	 c r - radial  coordinate, normalised
	 c     Include files for common variables
	 include 'global.h'
	 include 'potential.h'
	 */
	Real phi, x, y, w;
	
	
	phi = 0.0;
	w   = 0.0;
	
  
	if ((r*gp->invr_iso) >= 1.0e-4) {
    x = zls_Log10(r*gp->invr_iso);
    zls_splint(&y, x, gp->gravNWpline,  gp->gravRWspl, gp->gravWspl, gp->gravWspl2);
    w = pow(10.0,y); // y is log(-W)
	}
	
	phi = w*(gp->s_iso*gp->s_iso); // Phi = -s^2 W, W is negative, w = -W, Phi = s^2 w, Phi is +ve
  
  return (phi);
}

Real  zlsg_dphi_iso (Real r, PotentialPtr gp ){
	
	Real dphi, x, yd, dw;
	
	dphi = 0.0;
	dw = 0.0;
	
	if ((r/gp->r_iso) >= 1.0e-4) {
    x = zls_Log10(r*gp->invr_iso);
    zls_splint(&yd, x, gp->gravNWpline,  gp->gravRWspl, gp->gravdWspl, gp->gravdWspl2);
    dw = pow(10.0,yd); // zls_Log10(-dW/dx)
	}
	
	dphi = gp->s2_iso*dw*gp->invr_iso; // dPhi/dr = -(s^2/r_c) dW/dx
	
	return (dphi);
}


Real  zlsg_phi_halo (Real r, PotentialPtr gp ){
	/*
   c
   c phi_halo - potential due to spherical numerical integration
	 c evaluated by spline fit to numerical integration of dimensionless density equations
	 c RMS errors wrt to integration: 
   c
	 c potential specified in terms of dimensionless W, v_s, and r_s
	 c NOTE: v_s = s_iso/Sqrt(2);
   c
	 */
  
	Real phi, x, y, w;
	
	phi = 0.0;
	w   = 0.0;
  
	if ((r*gp->invr_s) >= 1.0e-4) {
    x = zls_Log10(r*gp->invr_s);
    zls_splint(&y, x, gp->gravNWpline,  gp->gravRWspl, gp->gravWspl, gp->gravWspl2);
    w = pow(10.0,y); // y is log(-W)
	}
	
	phi = w*(gp->v_s*gp->v_s); // Phi = -v_s^2 W, W is negative, w = -W, Phi = v_s^2 w, Phi is +ve
//  zlsio_Printf(" %g %g \n", r*gp->invr_s, exp(-phi/(0.11*0.11)));    

  return (phi);
}

Real  zlsg_dphi_halo (Real r, PotentialPtr gp ){
	
	Real dphi, x, yd, dw;
	
	dphi = 0.0;
	dw   = 0.0;
	
	if ((r/gp->invr_s) >= 1.0e-4) {
    x = zls_Log10(r*gp->invr_s);
    zls_splint(&yd, x, gp->gravNWpline,  gp->gravRWspl, gp->gravdWspl, gp->gravdWspl2);
    dw = pow(10.0,yd); // zls_Log10(-dW/dx)
	}
	
	dphi = gp->v2_s*dw*gp->invr_s; // dPhi/dr = -(v_s^2/r_s) dW/dx
	
	return (dphi);
}

Real  zlsg_phi_Uniform ( Real z, PotentialPtr gp ){ // r coordinate = z, depending on useage
	//
  // Dimensionless potential: phi = (s^2/r_0)*z
  // Sharing the isothermal sphere parameters
  // 
	Real phi, w;
	w   = zls_abs(z*gp->invr_iso);
	phi = gp->s2_iso*w;
  
  return (phi);
}

Real  zlsg_dphi_Uniform ( Real z, PotentialPtr gp ){
  //
  // Accelleration is -dPhi/dz = -(s^2/r_0), independent of z
  //
  
	Real dphi = -gp->s2_iso*gp->invr_iso; // dPhi/dz = -(s^2/r_0) 
	
	return (dphi);
}


Real  zlsg_phi_DiskZ ( Real z, PotentialPtr gp ){ 
  //
  // r coordinate = z out of Disk
  // fit to Kuijken and Gilmore potential form
  // using Holmberg and Flynn parameters:
  // s0 = 56 M/pc^2
  // re_eff = 0.007
  //
  // p0 = 0.102 M/pc^2 is current limit for total density, about 0.03 is gas limit
  //
  // and assuming a D, consistent with Siebert 2006 analysis of -D correlations
  // of D = 300 pc
  //
	Real phi, G; 
  Real s, d, a;
  
	d = 	gp->d_diskZ ; // D
	s = 	gp->s0_diskZ; // Surface density in plane
	a = 	gp->re_diskZ; // Volume Density of Halo/Thick Component in midplane
  G =   twopi*k_g;
  
  phi = G*(s*(zls_sqrt(z*z+d*d)-d)+a*z*z);
  
  return (phi);
}

Real  zlsg_dphi_DiskZ ( Real z, PotentialPtr gp ){
	// numerical derivative of above polynomial.
	Real dphi, G; 
  Real s, d, a;
	
	dphi = 0.0;
	d = 	gp->d_diskZ       ; // D
	s = 	gp->s0_diskZ      ; // Surface density
	a = 	gp->re_diskZ      ; // Volume Density of Halo/Thick Component in midplane
  G = twopi*k_g;
  
	dphi = -G*(2.0*a*z + (s*z)/zls_sqrt(z*z+d*d)); 
	
	return (dphi);
}

Real  zlsg_phi_TanhZ ( Real z, PotentialPtr gp ){
  //
  // Dimensionless potential where the accelleration is -s^2 tanh(r/r_0)
  // Sharing the isothermal sphere parameters
  // 
	Real phi, w;
  Real x = z*gp->invr_iso;
	w   = log(0.5*(exp(-x)+exp(x))); // Log[Cosh(x)]
	phi = w*gp->s2_iso;
  
  return (phi);
}

Real  zlsg_dphi_TanhZ ( Real z, PotentialPtr gp ){
  //
  // Accelleration is -s^2 tanh(r/r_0)
  // Sharing the isothermal sphere parameters
  // 
	Real dphi, dw;
	dw   = tanh(z*gp->invr_iso);
	dphi = -dw*gp->s2_iso;
  
  return (dphi);
}



int zlsg_SetIsothermalGravityFile ( PotentialPtr gp, char * gFile){
	
#define BUFSIZE 1024
	
	int err   = noErr;
	FILE * f  = NULL;
	
	char buffer[BUFSIZE];
	
	Integer nEntries, gravIndex;
	
	Real logR, logmWspl, logmdWspl, logmWspl2, logmdWspl2;
	
	if ( gp != NULL ) {
        
    if (( gp->gravTypeFlags & kPhysHaloGravityFlag ) == kPhysHaloGravityFlag ){
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Halo Potential: %s. \n", gFile );
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
    } else if (( gp->gravTypeFlags & kPhysCosmicHaloGravityFlag ) == kPhysCosmicHaloGravityFlag ){
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Cosmic Halo Potential: %s. \n", gFile );
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
    } else {
   
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Isothermal Potential: %s. \n", gFile );
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
      
		}
    
		f = fopen(gFile, "r");
		if (f) {
			
			// skip header
			strcpy(buffer, "#");
			while(*buffer == '#'){
				fgets(buffer, BUFSIZE, f);
				if (*buffer != '#')
					puts(buffer);
			}
			
			fscanf( f, "%d",&nEntries);
			if ( nEntries > 1 ) { // read n entries that follow
				
				gp->gravNWpline = nEntries;
				zls_NewQuantity1D(&gp->gravRWspl, nEntries);
				zls_NewQuantity1D(&gp->gravWspl, nEntries);
				zls_NewQuantity1D(&gp->gravWspl2, nEntries);
				zls_NewQuantity1D(&gp->gravdWspl, nEntries);
				zls_NewQuantity1D(&gp->gravdWspl2, nEntries);
				
				if (( gp->gravRWspl != NULL ) &&( gp->gravWspl != NULL ) && ( gp->gravWspl2 != NULL ) ) {
					
					for (gravIndex = 0; gravIndex < nEntries ; gravIndex++){
						
						fscanf( f, "%lf %lf %lf %lf %lf",&logR, &logmWspl, &logmdWspl, &logmWspl2,  &logmdWspl2);
						
						//Real R      = pow(10.0,logR);
						
						gp->gravRWspl [gravIndex]  = logR; 
						gp->gravWspl  [gravIndex]  = logmWspl; 
						gp->gravWspl2 [gravIndex]  = logmWspl2; 
						gp->gravdWspl [gravIndex]  = logmdWspl; 
						gp->gravdWspl2[gravIndex]  = logmdWspl2; 
						
					}
					
				} else {
					zlsio_ErrPrintf( " ERROR: zlsp_SetPhysicsIsothermalGravityFile, FAILED to allocate tables %d \n", nEntries);
					err = memErr;
				}
				
			} else { 
				zlsio_ErrPrintf( " ERROR: zlsp_SetPhysicsIsothermalGravityFile, bad number of entries: %d \n", nEntries);
				err = fileErr;
			}
			fclose(f);
			
		} else { err = fileErr; }
		
	}
	
	return (err);
}

int   zlsio_SetUpGravityFromConfigP(PotentialPtr gp, struct ccl_t * configP){
  
  int err = noErr;
  const char     *str1;
  
  char gravFileName[PATHLEN];
  Integer          gType;
  
  if ( gp != NULL) {
    
    gType = kPhysNoGravityFlag;
    if ((str1= ccl_get( configP, "Gravity_Types"))){
      gType  = strtol( str1, NULL, 0 );
#if (ZLS_DEBUG>0)
      zlsio_Printf( "   %s = %s\n", "Gravity_Types", str1);
#endif
    }
    
    gp->gravTypeFlags = gType;
    
    if (( gType & kPhysIsoGravityTypes ) != 0 ){
      
      if (( gType & kPhysIsothermalGravityFlag ) == kPhysIsothermalGravityFlag ){
        if ((str1= ccl_get( configP, "Gravity_IsothermalFile"))){
          strcpy( gravFileName,  str1);
        } else {
          strcpy( gravFileName,  "data/potentials/isologwlogdw.txt" );
        }
        
        err = zlsg_SetIsothermalGravityFile( gp, gravFileName );
      }
      
      if (( gType & kPhysHaloGravityFlag ) == kPhysHaloGravityFlag ){
        if ((str1= ccl_get( configP, "Gravity_PotentialFile"))){
          strcpy( gravFileName,  str1);
        } else {
          strcpy( gravFileName,  "data/potentials/hisologwlogdw.txt" ); // v_s^2 scaled version of the s^2 isothermal file.
        }
        
        err = zlsg_SetIsothermalGravityFile( gp, gravFileName ); // read the same
      }
      
      if (( gType & kPhysCosmicHaloGravityFlag ) == kPhysCosmicHaloGravityFlag ){

        if ((str1= ccl_get( configP, "Gravity_PotentialFile"))){
          strcpy( gravFileName,  str1);
        } else {
          strcpy( gravFileName,  "data/potentials/hisologwlogdw.txt" ); // v_s^2 scaled version of the s^2 isothermal file.
        }
        
        err = zlsg_SetIsothermalGravityFile( gp, gravFileName );
      }
 
      // get the core radius, r_iso = r_s, set one and the other is set the same
      
      Real R_iso = gp->r_iso;
      if ((str1= ccl_get( configP, "Gravity_R_Iso"))){
        R_iso = atof( str1)*f_cmpc/f_x0;  // r in pc
      }
      
      gp->r_iso    = R_iso; 
      gp->invr_iso = 1.0/gp->r_iso;
      
      gp->r_s      = R_iso;  // halos the same
      gp->invr_s   = gp->invr_iso;
      
      Real R_s = gp->r_s;
      if ((str1= ccl_get( configP, "Gravity_R_s"))){
        R_s = atof( str1)*f_cmpc/f_x0;  // r in pc
      }

      gp->r_s      = R_s;  // halos the same
      gp->invr_s   = 1.0/gp->r_s;

      gp->r_iso    = R_s; 
      gp->invr_iso = gp->invr_s;

      // for input , s is sqrt(2) v_s, and can set either
      
      Real S_iso = gp->s_iso;
      if ((str1= ccl_get( configP, "Gravity_S_Iso"))){
        S_iso = atof( str1)*f_cmkm/f_v0; // in km/s
      } 
      
      gp->s_iso     = S_iso;
      gp->s2_iso    = gp->s_iso*gp->s_iso;
      
      gp->invs_iso  = 1.0/gp->s_iso;
      gp->invs2_iso = 1.0/gp->s2_iso;
      
      gp->v_s       = S_iso/zls_sqrt(2.0);
      gp->v2_s      = gp->v_s*gp->v_s;
      
      gp->invv_s    = 1.0/gp->v_s;
      gp->invv2_s   = 1.0/gp->v2_s;
      
      Real V_s = gp->v_s;
      if ((str1= ccl_get( configP, "Gravity_V_s"))){
        V_s = atof( str1)*f_cmkm/f_v0; // a circular velocity in km/s
      } 
      
      gp->v_s       = V_s;
      gp->v2_s      = gp->v_s*gp->v_s;
      
      gp->invv_s    = 1.0/gp->v_s;
      gp->invv2_s   = 1.0/gp->v2_s;
      
      gp->s_iso     = V_s*zls_sqrt(2.0);
      gp->s2_iso    = gp->s_iso*gp->s_iso;
      
      gp->invs_iso  = 1.0/gp->s_iso;
      gp->invs2_iso = 1.0/gp->s2_iso;
      
      
    }
    
    if (( gType & kPhysCosmicHaloGravityFlag ) == kPhysCosmicHaloGravityFlag ){
      //
      // specify Delta rho_u and M_vir, specify x_vir, and provide f_m at x_vir 
      // and then derive r_s and v_s, setting r_iso and s_iso for consistency
      //
      
      Real Mvir = gp->M_vir;
      if ((str1= ccl_get( configP, "Gravity_HaloVirialMass"))){
        Mvir = atof( str1); // special assumed m is in solar masses
      } 
      Mvir *= k_msol/f_m0;
      gp->M_vir  = Mvir; // convert to cgs then normalise
      
      Real delrho = gp->delta_rho;
      if ((str1= ccl_get( configP, "Gravity_HaloDelta"))){
        delrho = atof( str1); // ratio, no units
      } 
      gp->delta_rho  = delrho;
      
      Real rho_u = gp->rho_universe;
      if ((str1= ccl_get( configP, "Gravity_UniverseRho"))){
        rho_u = atof( str1); // 
      } 
      rho_u /= f_d0;
      gp->delta_rho  = rho_u;
      
      Real Xvir = gp->x_vir;
      if ((str1= ccl_get( configP, "Gravity_HaloXVirial"))){
        Xvir = atof( str1); // ratio, no units
      } 
      gp->x_vir  = Xvir;
      
      Real FMvir = gp->f_mvir;
      if ((str1= ccl_get( configP, "Gravity_HaloFMVirial"))){
        FMvir = atof( str1); // ratio, no units, must be looked up from elsewhere....
      } 
      gp->f_mvir  = FMvir;
      
      Real FM1 = gp->f_m1;
      if ((str1= ccl_get( configP, "Gravity_HaloFM1"))){
        FM1 = atof( str1); // ratio, no units, constant for each potential file
      } 
      gp->f_m1  = FM1;
      
      // Now derive r_s and v_s
      
      Real r_vir   = zls_Pow( ((3.0*Mvir)/(fourpi*delrho*rho_u)) , 1.0/3.0 );
      
      gp->r_s      = r_vir/Xvir; // already in normalised units
      gp->invr_s   = 1.0/gp->r_s;
      
      gp->r_iso    = gp->r_s; 
      gp->invr_iso = 1.0/gp->r_s;

      gp->M_s     = Mvir/FMvir;
      
      gp->v2_s    = k_g*gp->M_s/gp->r_s; 
      gp->v_s     = zls_sqrt( gp->v2_s);
      
      gp->M_1     = gp->f_m1*gp->M_s;
      gp->v2_1    = gp->f_m1*gp->v2_s; 
      gp->v_1     = zls_sqrt( gp->v2_1);

      gp->invv_s  = 1.0/gp->v_s;
      gp->invv2_s = 1.0/gp->v2_s;
      
      gp->s_iso   = gp->v_s*zls_sqrt(2.0);
      gp->s2_iso  = gp->s_iso*gp->s_iso;
      
      gp->invs_iso  = 1.0/gp->s_iso;
      gp->invs2_iso = 1.0/gp->s2_iso;
      
      // don't need rho_s here.
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Cosmological Halo Potential Derived Parameters\n" );
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Halo Scale Radius            : %#12.5g (pc)\n"  , gp->r_s*f_x0/f_cmpc );
      zlsio_Printf( " Halo Scale Mass              : %#12.5g (Msol)\n", gp->M_s*f_m0/k_msol );
      zlsio_Printf( " Halo Scale Circular Velocity : %#12.5g (km/s)\n", gp->v_s*f_v0/f_cmkm );
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Core, x = r/r_s = 1.0 \n" );
      zlsio_Printf( " Halo Core Mass              : %#12.5g (Msol)\n", gp->M_1*f_m0/k_msol );
      zlsio_Printf( " Halo Core Circular Velocity : %#12.5g (km/s)\n", gp->v_1*f_v0/f_cmkm );
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
      
    }    
    
    if (( gType & kPhysNewtonianGravityFlag ) == kPhysNewtonianGravityFlag ){
      
      Real M0 = gp->M;
      if ((str1= ccl_get( configP, "Gravity_Mass"))){
        M0 = atof( str1); // special assumed m is in solar masses
      } 
      gp->M  = M0; // Note: leave in Msol
      
      Real R0 = gp->rN0;
      if ((str1= ccl_get( configP, "Gravity_R0"))){
        R0 = atof( str1);
      } 
      gp->rN0  = R0/f_x0;
      
    }
    
    if (( gType & kPhysUniformGravityFlag ) == kPhysUniformGravityFlag ){
      
      Real R_iso = gp->r_iso;
      if ((str1= ccl_get( configP, "Gravity_R"))){
        R_iso = atof( str1); 
      }
      gp->r_iso    = R_iso/f_x0;
      gp->invr_iso = 1.0/gp->r_iso;
      
      Real S_iso = gp->s_iso;
      if ((str1= ccl_get( configP, "Gravity_S"))){
        S_iso = atof( str1);
      } 
      gp->s_iso  = S_iso/f_v0;
      gp->s2_iso = gp->s_iso*gp->s_iso;
      
      gp->invs_iso  = 1.0/gp->s_iso;
      gp->invs2_iso = 1.0/gp->s2_iso;
      
    }
    
    if (( gType & kPhysKingSpheroidGravityFlag ) == kPhysKingSpheroidGravityFlag ){
      
      Real R_ss = gp->r_ss;
      if ((str1= ccl_get( configP, "Gravity_R_ss"))){
        R_ss = atof( str1)*f_cmpc/f_x0;  // r in pc
      }
      gp->r_ss = R_ss;
      gp->invr_ss = 1.0/gp->r_ss;
      
      Real Mss = gp->m_ss;
      if ((str1= ccl_get( configP, "Gravity_M_ss"))){
        Mss = atof( str1); // special assumed m is in solar masses
      } 
      gp->m_ss  = Mss; // Note: leave in Msol
    }
    
    
    if (( gType & kPhysDiskZGravityFlag ) == kPhysDiskZGravityFlag ){
      
      Real D_diskZ = 300.0;          
      if ((str1= ccl_get( configP, "Gravity_D_DiskZ"))){
        D_diskZ  = atof( str1);  // r in pc
      }
      gp->d_diskZ  = D_diskZ *f_cmpc/f_x0; // pc->cm then normalised
      
      Real SurfaceDensity_diskZ = 56.0; // M_0/pc^2
      if ((str1= ccl_get( configP, "Gravity_S0_DiskZ"))){
        SurfaceDensity_diskZ  = atof( str1);
      } 
      gp->s0_diskZ = SurfaceDensity_diskZ*k_msol/(f_cmpc*f_cmpc); // cgs
      gp->s0_diskZ *= (f_x0*f_x0)/f_m0; // normalised
      
      
      Real HaloDensity_diskZ = 0.007; // M_0/pc^3
      if ((str1= ccl_get( configP, "Gravity_RhoEff_DiskZ"))){
        HaloDensity_diskZ  = atof( str1);
      } 
      
      gp->re_diskZ   = HaloDensity_diskZ *k_msol/(f_cmpc*f_cmpc*f_cmpc); // cgs
      gp->re_diskZ  *= (f_x0*f_x0*f_x0)/f_m0; // normalised
      
    }
    
    
  } else { 
    err = memErr;
  }
  return(err);
}  

#endif GRAVITY

#ifdef GRAVITY
// only used for gravity right now, probably for other fields later.
int zlsh_FillF1DFrom1DField ( F1DArr f,  Field zoneField ){
#ifndef SYM1D
#pragma unused( f, zoneField)
  return(dimErr);
#else
  //
  // Copy a field row from zoneField into f. This routine handles
  // dimensions and BCs, and f is already created to match zoneField.
  //
  //
  // Assumes everything is set up properly first, v. little error checking here
  // as it is a v. low level and hopefully quick routine
  //
	
  int err = noErr;
	
  Integer i;
  LRQ1    a;
  Integer nti;
  
  nti = f.nt; // assumes f is set up correctly from the owning zone.
  for(i = 0; i<nti; i++){// copy whole row including bcs
    
    a  = zoneField.p[i];
    f.xi[i].a     = a.a;
    f.xi[i].al    = a.al;
    f.xi[i].ar    = a.ar;
    
#ifdef PPM      
    f.xi[i].del   = a.ar- a.al;
    f.xi[i].a6    = 6.0*(a.a - 0.5*(a.al + a.ar));
#endif PPM      
    
  }
  
  return(err);
#endif
}

int zlsh_FillF1DFrom2DFieldX ( F1DArr f,  Field zoneField, Integer j ){
#ifndef SYM2D
  return(dimErr);
#else
  //
  // Copy a row from zoneW into w.This routine handles
  // dimensions and BCs.
  //
  //
  // Assumes everything is set up properly first, v. little error checking here
  // as it is a v. low level and hopefully quick routine
  //
	
  int err = noErr;
	
  Integer i;
  LRQ2    a;
  
  Integer nti;
  
  nti = f.nt; // assumes w is set up correctly from the owning zone.
  for(i = 0; i<nti; i++){// copy whole row including bcs
    
    a             = zoneField.p[j][i];
    f.xi[i].al    = zls_tolx(a.al[kXDim], kZLS_Small);
    f.xi[i].ar    = zls_tolx(a.ar[kXDim], kZLS_Small);

#ifdef SCALAR_LRQS
    f.xi[i].a     = zls_tolx(a.a       , kZLS_Small);
    
#ifdef PPM
    f.xi[i].del   = a.ar[kXDim]- a.al[kXDim];
    f.xi[i].a6    = 6.0*(a.a - 0.5*(a.al[kXDim] + a.ar[kXDim]));
#endif PPM
#endif SCALAR_LRQS
    
#ifdef VECTOR_LRQS
    f.xi[i].a     = zls_tolx(a.a [kXDim], kZLS_Small);
    
#ifdef PPM
    f.xi[i].del   = a.ar[kXDim]- a.al[kXDim];
    f.xi[i].a6    = 6.0*(a.a[kXDim] - 0.5*(a.al[kXDim] + a.ar[kXDim]));
#endif PPM
#endif VECTOR_LRQS

    
  }
  
  return(err);
#endif
}

int zlsh_FillF1DFrom2DFieldY ( F1DArr f,  Field zoneField, Integer i ){
#ifndef SYM2D
  return(dimErr);
#else
  //
  // Copy a row from zoneW into w.This routine handles
  // dimensions and BCs.
  //
  //
  // Assumes everything is set up properly first, v. little error checking here
  // as it is a v. low level and hopefully quick routine
  //
  int err = noErr;
  Integer j;
  LRQ2    a;
	
  Integer ntj;
  ntj = f.nt; // assumes w is set up correctly from the owning zone.
	
  for(j = 0; j<ntj; j++){// copy whole row including bcs
    
    a             = zoneField.p[j][i];
    
    f.xi[j].al    = zls_tolx(a.al[kYDim], kZLS_Small);
    f.xi[j].ar    = zls_tolx(a.ar[kYDim], kZLS_Small);
    
    
#ifdef SCALAR_LRQS
    f.xi[j].a     = zls_tolx(a.a        , kZLS_Small);
#ifdef PPM
    f.xi[j].del   = a.ar[kYDim]- a.al[kYDim];
    f.xi[j].a6    = 6.0*(a.a - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif SCALAR_LRQS
    
#ifdef VECTOR_LRQS
    f.xi[j].a     = zls_tolx(a.a [kYDim], kZLS_Small);
#ifdef PPM
    f.xi[j].del   = a.ar[kYDim]- a.al[kYDim];
    f.xi[j].a6    = 6.0*(a.a[kYDim] - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif VECTOR_LRQS
		
  }
  
  return(err);
#endif
}

int zlsh_FillF2DFrom2DField  ( F2DArr f,  Field zoneField ){
#ifndef SYM2D
  return(dimErr);
#else
  //
  // Copy a row from zoneW into w.This routine handles
  // dimensions and BCs.
  //
  //
  // Assumes everything is set up properly first, v. little error checking here
  // as it is a v. low level and hopefully quick routine
  //
	
  int err = noErr;
	
  Integer i;
  Integer j;
  LRQ2    a;
  
  Integer nti;
  Integer ntj;
  
  nti = f.nt[kXDim]; // assumes w is set up correctly from the owning zone.
  ntj = f.nt[kYDim]; // assumes w is set up correctly from the owning zone.
  for(j = 0; j<ntj; j++){// copy whole row including bcs
    for(i = 0; i<nti; i++){// copy whole row including bcs
      
      a                 = zoneField.p[j][i];
      
      f.xi[j][i].al[kXDim]    = a.al[kXDim];
      f.xi[j][i].ar[kXDim]    = a.ar[kXDim];
      
      
      f.xi[j][i].al[kYDim]    = a.al[kYDim];
      f.xi[j][i].ar[kYDim]    = a.ar[kYDim];

#ifdef SCALAR_LRQS
      f.xi[j][i].a [kXDim]    = a.a;
      f.xi[j][i].a [kYDim]    = a.a;
#ifdef PPM
      f.xi[j][i].del[kXDim]   = a.ar[kXDim]- a.al[kXDim];
      f.xi[j][i].a6[kXDim]    = 6.0*(a.a - 0.5*(a.al[kXDim] + a.ar[kXDim]));
      f.xi[j][i].del[kYDim]   = a.ar[kYDim]- a.al[kYDim];
      f.xi[j][i].a6[kYDim]    = 6.0*(a.a - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif SCALAR_LRQS
      
#ifdef VECTOR_LRQS
      f.xi[j][i].a [kXDim]    = a.a[kXDim];
      f.xi[j][i].a [kYDim]    = a.a[kYDim];
#ifdef PPM
      f.xi[j][i].del[kXDim]   = a.ar[kXDim]- a.al[kXDim];
      f.xi[j][i].a6[kXDim]    = 6.0*(a.a[kXDim] - 0.5*(a.al[kXDim] + a.ar[kXDim]));
      f.xi[j][i].del[kYDim]   = a.ar[kYDim]- a.al[kYDim];
      f.xi[j][i].a6[kYDim]    = 6.0*(a.a[kYDim] - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif VECTOR_LRQS

      
    }
  }
  
  return(err);
#endif
}

int zlsh_FillF1DXFrom3DFieldYZ ( F1DArr f, Field zoneField, Integer j, Integer k ){
#ifndef SYM3D
#pragma unused( f, zoneField, j, k)
  return(dimErr);
#else
  //
  //
  int err = noErr;
  Integer i;
  LRQ3    a;
  
  Integer nti = f.nt; // assumes f is set up correctly from the owning zone.
	
  for(i = 0; i<nti; i++){// copy whole row including bcs
		
    a =  zoneField.p[k][j][i];
    f.xi[i].al    = zls_tolx(a.al[kXDim], kZLS_Small);
    f.xi[i].ar    = zls_tolx(a.ar[kXDim], kZLS_Small);

#ifdef SCALAR_LRQS
    f.xi[i].a     = zls_tolx(a.a , kZLS_Small);
#ifdef PPM
    f.xi[i].del   = a.ar[kXDim] - a.al[kXDim];
    f.xi[i].a6    = 6.0*(a.a - 0.5*(a.al[kXDim] + a.ar[kXDim]));
#endif PPM
#endif SCALAR_LRQS
    
#ifdef VECTOR_LRQS
    f.xi[i].a     = zls_tolx(a.a [kXDim], kZLS_Small);
#ifdef PPM
    f.xi[i].del   = a.ar[kXDim] - a.al[kXDim];
    f.xi[i].a6    = 6.0*(a.a[kXDim] - 0.5*(a.al[kXDim] + a.ar[kXDim]));
#endif PPM
#endif VECTOR_LRQS
		
  }
  
  return(err);
#endif
}

int zlsh_FillF1DYFrom3DFieldZX ( F1DArr f, Field zoneField, Integer k, Integer i ){
#ifndef SYM3D
#pragma unused( f, zoneField, i, k)
  return(dimErr);
#else
  //
  //
  int err = noErr;
  Integer j;
  LRQ3    a;
  
  Integer ntj = f.nt; // assumes f is set up correctly from the owning zone.
	
  for(j = 0; j<ntj; j++){// copy whole row including bcs
		
    a =  zoneField.p[k][j][i];
    f.xi[j].al    = zls_tolx(a.al[kYDim], kZLS_Small);
    f.xi[j].ar    = zls_tolx(a.ar[kYDim], kZLS_Small);

#ifdef SCALAR_LRQS
    f.xi[j].a     = zls_tolx(a.a , kZLS_Small);
#ifdef PPM
    f.xi[j].del   = a.ar[kYDim] - a.al[kYDim];
    f.xi[j].a6    = 6.0*(a.a - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif SCALAR_LRQS
    
#ifdef VECTOR_LRQS
    f.xi[j].a     = zls_tolx(a.a [kYDim], kZLS_Small);
#ifdef PPM
    f.xi[j].del   = a.ar[kYDim] - a.al[kYDim];
    f.xi[j].a6    = 6.0*(a.a[kYDim] - 0.5*(a.al[kYDim] + a.ar[kYDim]));
#endif PPM
#endif VECTOR_LRQS
    
		
  }
  
  return(err);
#endif
}
int zlsh_FillF1DZFrom3DFieldXY ( F1DArr f, Field zoneField, Integer i, Integer j ){
#ifndef SYM3D
#pragma unused( f, zoneField, i, j)
  return(dimErr);
#else
  //
  //
  int err = noErr;
  Integer k;
  LRQ3    a;
  
  Integer ntk = f.nt; // assumes f is set up correctly from the owning zone.
	
  for(k = 0; k<ntk; k++){// copy whole row including bcs
		
    a =  zoneField.p[k][j][i];
    f.xi[k].al    = zls_tolx(a.al[kZDim], kZLS_Small);
    f.xi[k].ar    = zls_tolx(a.ar[kZDim], kZLS_Small);

#ifdef SCALAR_LRQS
    f.xi[k].a     = zls_tolx(a.a , kZLS_Small);
#ifdef PPM
    f.xi[k].del   = a.ar[kZDim] - a.al[kZDim];
    f.xi[k].a6    = 6.0*(a.a - 0.5*(a.al[kZDim] + a.ar[kZDim]));
#endif PPM
#endif SCALAR_LRQS
    
#ifdef VECTOR_LRQS
    f.xi[k].a     = zls_tolx(a.a [kZDim], kZLS_Small);
#ifdef PPM
    f.xi[k].del   = a.ar[kZDim] - a.al[kZDim];
    f.xi[k].a6    = 6.0*(a.a[kZDim] - 0.5*(a.al[kZDim] + a.ar[kZDim]));
#endif PPM
#endif VECTOR_LRQS
    
		
  }
  
  return(err);
#endif
}

int zlsh_FillF3DFrom3DField ( F3DArr f,  Field zoneField ){
#ifndef SYM3D
#pragma unused( f, zoneField)
	return(dimErr);
#else
	
	int err = noErr;
	
	Integer i;
	Integer j;
	Integer k;
	LRQ3    a;
	
	Integer nti;
	Integer ntj;
	Integer ntk;
	
	nti = f.nt[kXDim]; // assumes w is set up correctly from the owning zone.
	ntj = f.nt[kYDim]; // assumes w is set up correctly from the owning zone.
	ntk = f.nt[kZDim]; // assumes w is set up correctly from the owning zone.
	for(k = 0; k<ntk; k++){ // copy whole row including bcs
		for(j = 0; j<ntj; j++){ // copy whole row including bcs
			for(i = 0; i<nti; i++){ // copy whole row including bcs
				
				a                 = zoneField.p[k][j][i];
				f.xi[k][j][i].al[kXDim]    = a.al[kXDim];
				f.xi[k][j][i].ar[kXDim]    = a.ar[kXDim];
				f.xi[k][j][i].al[kYDim]    = a.al[kYDim];
				f.xi[k][j][i].ar[kYDim]    = a.ar[kYDim];
				f.xi[k][j][i].al[kZDim]    = a.al[kZDim];
				f.xi[k][j][i].ar[kZDim]    = a.ar[kZDim];
        
#ifdef SCALAR_LRQS
				f.xi[k][j][i].a[kXDim]     = a.a;
				f.xi[k][j][i].a[kYDim]     = a.a;
				f.xi[k][j][i].a[kZDim]     = a.a;
        
#ifdef PPM
				f.xi[k][j][i].del[kXDim]   = a.ar[kXDim]- a.al[kXDim];
				f.xi[k][j][i].a6[kXDim]    = 6.0*(a.a - 0.5*(a.al[kXDim] + a.ar[kXDim]));
				f.xi[k][j][i].del[kYDim]   = a.ar[kYDim]- a.al[kYDim];
				f.xi[k][j][i].a6[kYDim]    = 6.0*(a.a - 0.5*(a.al[kYDim] + a.ar[kYDim]));
				f.xi[k][j][i].del[kZDim]   = a.ar[kZDim]- a.al[kZDim];
				f.xi[k][j][i].a6[kZDim]    = 6.0*(a.a - 0.5*(a.al[kZDim] + a.ar[kZDim]));
#endif PPM
#endif SCALAR_LRQS
        
#ifdef VECTOR_LRQS
				f.xi[k][j][i].a[kXDim]     = a.a[kXDim];
				f.xi[k][j][i].a[kYDim]     = a.a[kYDim];
				f.xi[k][j][i].a[kZDim]     = a.a[kZDim];
        
#ifdef PPM
				f.xi[k][j][i].del[kXDim]   = a.ar[kXDim]- a.al[kXDim];
				f.xi[k][j][i].a6[kXDim]    = 6.0*(a.a[kXDim] - 0.5*(a.al[kXDim] + a.ar[kXDim]));
				f.xi[k][j][i].del[kYDim]   = a.ar[kYDim]- a.al[kYDim];
				f.xi[k][j][i].a6[kYDim]    = 6.0*(a.a[kYDim] - 0.5*(a.al[kYDim] + a.ar[kYDim]));
				f.xi[k][j][i].del[kZDim]   = a.ar[kZDim]- a.al[kZDim];
				f.xi[k][j][i].a6[kZDim]    = 6.0*(a.a[kZDim] - 0.5*(a.al[kZDim] + a.ar[kZDim]));
#endif PPM
#endif VECTOR_LRQS
        
				
			}
		}
	}
	
	return(err);
#endif
}

#endif GRAVITY

