#ifndef ZLS_PHYSINLINES_H
#define ZLS_PHYSINLINES_H
#include <zls.h>

static inline Real zlsp_XH2_T( Real T , CoolingPtr cp);
static inline Real zlsp_XH2_T( Real T , CoolingPtr cp){
  
	Real fh2 = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType == kPhysFileCooling)  {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        
        Integer n = cp->coolNTable;
        
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        
        do{
          k=(khi+klo)>>1;
          if(cp->coolTValues[k]>T){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        
        fh2   = cp->coolfh2Values[klo]; 
        fh2  =  fh2 < 0.0 ? 0.0: fh2;
      }
    }
  }
#endif COOLING
  return(fh2);
  
}

static inline Real zlsp_XH2_Pd( Real Pd , CoolingPtr cp);
static inline Real zlsp_XH2_Pd( Real Pd , CoolingPtr cp){
	Real fh2 = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolPdValues[k]>Pd){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fh2   = cp->coolfh2Values[klo]; 
        fh2  =  fh2 < 0.0 ? 0.0: fh2;
      }
    }
  }
#endif COOLING
  return(fh2);
  
}

static inline Real zlsp_XH2_Ud( Real Ud , CoolingPtr cp);
static inline Real zlsp_XH2_Ud( Real Ud , CoolingPtr cp){
	Real fh2 = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolUdValues[k]>Ud){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fh2   = cp->coolfh2Values[klo]; 
        fh2  =  fh2 < 0.0 ? 0.0: fh2;
      }
    }
  }
#endif COOLING
  return(fh2);
  
}

static inline Real zlsp_XHI_T( Real T , CoolingPtr cp);
static inline Real zlsp_XHI_T( Real T , CoolingPtr cp){
	Real fhi = 1.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if(cp->coolTValues[k]>T){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhi   = cp->coolfhiValues[klo]; 
        fhi  =  fhi < 0.0 ? 0.0: fhi;
      }
    }
  }
#endif COOLING
  return(fhi);
}

static inline Real zlsp_XHI_Pd( Real Pd , CoolingPtr cp);
static inline Real zlsp_XHI_Pd( Real Pd , CoolingPtr cp){
	Real fhi = 1.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolPdValues[k]>Pd){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhi   = cp->coolfhiValues[klo]; 
        fhi  =  fhi < 0.0 ? 0.0: fhi;
      }
    }
  }
#endif COOLING
  return(fhi);
}

static inline Real zlsp_XHI_Ud( Real Ud , CoolingPtr cp);
static inline Real zlsp_XHI_Ud( Real Ud , CoolingPtr cp){
	Real fhi = 1.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolUdValues[k]>Ud){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhi   = cp->coolfhiValues[klo]; 
        fhi  =  fhi < 0.0 ? 0.0: fhi;
      }
    }
  }
#endif COOLING
  return(fhi);
}

static inline Real zlsp_XHII_T( Real T , CoolingPtr cp);
static inline Real zlsp_XHII_T( Real T , CoolingPtr cp){
	Real fhii = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if(cp->coolTValues[k]>T){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhii   = cp->coolfhiiValues[klo]; 
        fhii   =  fhii < 0.0 ? 0.0: fhii;
      }
    }
  }
#endif COOLING
	return(fhii);
}

static inline Real zlsp_XHII_Pd( Real Pd , CoolingPtr cp);
static inline Real zlsp_XHII_Pd( Real Pd , CoolingPtr cp){
	Real fhii = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolPdValues[k]>Pd){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhii   = cp->coolfhiiValues[klo]; 
        fhii   =  fhii < 0.0 ? 0.0: fhii;
      }
    }
  }
#endif COOLING
  return(fhii);
}

static inline Real zlsp_XHII_Ud( Real Ud , CoolingPtr cp);
static inline Real zlsp_XHII_Ud( Real Ud , CoolingPtr cp){
	Real fhii = 0.0;
#ifdef COOLING
  if ( cp != NULL) {
    if( cp->coolType== kPhysFileCooling) {
      if( (cp->collAuxType & kZLS_HasHFractions) == kZLS_HasHFractions)  {
        Integer n = cp->coolNTable;
        register Integer k;
        register Integer klo=0;
        register Integer khi=n-1;
        do{
          k=(khi+klo)>>1;
          if( cp->coolUdValues[k]>Ud){
            khi=k;
          } else {
            klo=k;
          }
        } while ( khi-klo > 1) ;
        fhii   = cp->coolfhiiValues[klo]; 
        fhii   =  fhii < 0.0 ? 0.0: fhii;
      }
    }
  }
#endif COOLING
  return(fhii);
}


static inline Real zlsp_Gamma_Eff( Real g0, Real P0, Real g1, Real P1); 
static inline Real zlsp_Gamma_Eff( Real g0, Real P0, Real g1, Real P1){
	Real gam = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  Real Beta = P1/P0;
  Real gm0  = (g0 - 1.0);
  Real gm1  = (g1 - 1.0);
	Real gamm = ((1.0+Beta)*gm0*gm1)/(Beta*gm0 + gm1);
  gam  = 1.0+gamm;
#endif VARIABLE_GAMMA
	return(gam);
}

static inline Real zlsp_Gamma_T( Real T , CoolingPtr cp);
static inline Real zlsp_Gamma_T( Real T , CoolingPtr cp){
  
	Real gam = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  
	if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if(cp->coolTValues[k]>T){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      gam   = cp->coolGammaValues[klo]; 
    }
  }

#endif VARIABLE_GAMMA
  
	return(gam);
  
}

static inline Real zlsp_Gamma_Pd( Real Pd , CoolingPtr cp); // look up gamma for P/d
static inline Real zlsp_Gamma_Pd( Real Pd , CoolingPtr cp){ // look up gamma for P/d
  
	Real gam = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  
	if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if(cp->coolPdValues[k]>Pd){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      gam   = cp->coolGammaValues[klo]; 
    }
  }
#endif VARIABLE_GAMMA
  
	return(gam);
  
}

static inline Real zlsp_Gamma_Ud ( Real Ud , CoolingPtr cp); // look up gamma for U/d
static inline Real zlsp_Gamma_Ud ( Real Ud , CoolingPtr cp){ // look up gamma for U/d
  
	Real gam = kZLS_eos_gamma;
  
#ifdef VARIABLE_GAMMA
  if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if(cp->coolUdValues[k]>Ud){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      gam   = cp->coolGammaValues[klo]; 
    }
  }
#endif VARIABLE_GAMMA
  
	return(gam);
  
}


static inline Real zlsp_Mu_fH( Real *fH );  // fH[0] = H2,  fH[1] = Neutral H Atoms, fH[2] = HII
static inline Real zlsp_Mu_fH( Real *fH ){
	Real mu = k_mu;
#ifdef VARIABLE_MU
  
  // masses in amu
  const Real mH = 1.008;  // H atom
  const Real nH = 1.0;
  const Real mZ = 4.03;   // average heavy ion
  const Real nZ = 0.109;  // average heavy ion abundances
  const Real me = 5.4466e-4; // electron
  
  Real ne = fH[2]*(1.0+2.0*nZ); // aprox electron density - fully ionised with Hydrogen
  
  Real m = (nH*mH + nZ*mZ + ne*me);
  Real n = (ne + 0.5*fH[0] + fH[1] + fH[2] + nZ); 
  mu     = m/n;
#endif VARIABLE_MU
  
  return(mu);
}


static inline Real zlsp_Mu_T( Real T , CoolingPtr cp);
static inline Real zlsp_Mu_T( Real T , CoolingPtr cp){
	Real mu = k_mu;
#ifdef COOLING 
#ifdef VARIABLE_MU
  
	if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if(cp->coolTValues[k]>T){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      mu   = cp->coolmuValues[klo]; 
    }
  }
#endif VARIABLE_MU
#endif COOLING
  
	return(mu);
  
}

static inline Real zlsp_Mu_Pd( Real Pd , CoolingPtr cp);
static inline Real zlsp_Mu_Pd( Real Pd , CoolingPtr cp){
  
	Real mu = k_mu;
#ifdef VARIABLE_MU
  
	if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if( cp->coolPdValues[k]>Pd){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      mu   = cp->coolmuValues[klo]; 
    }
  }

#endif VARIABLE_MU
  
	return(mu);
}

static inline Real zlsp_Mu_Ud( Real Ud , CoolingPtr cp);
static inline Real zlsp_Mu_Ud( Real Ud , CoolingPtr cp){
	Real mu = k_mu;
#ifdef VARIABLE_MU
  
	if( cp->coolType== kPhysFileCooling){
    if( (cp->collAuxType & kZLS_HasGammaMu) == kZLS_HasGammaMu)  {
      Integer n = cp->coolNTable;
      register Integer k;
      register Integer klo=0;
      register Integer khi=n-1;
      do{
        k=(khi+klo)>>1;
        if( cp->coolUdValues[k]>Ud){
          khi=k;
        } else {
          klo=k;
        }
      } while ( khi-klo > 1) ;
      mu   = cp->coolmuValues[klo]; 
    }
  }

#endif VARIABLE_MU
  
	return(mu);
}

static inline Real zlsp_Pd_T( Real T, CoolingPtr cp);
static inline Real zlsp_Pd_T( Real T, CoolingPtr cp){
  Real Pd = f_invtep0*T;
#ifdef VARIABLE_MU  
  Real  mu  =  zlsp_Mu_T( T, cp );
  Pd *= (k_mu/mu);
#endif VARIABLE_MU
	return(Pd);
}

static inline Real zlsp_Pd_T_table( Real T, CoolingPtr cp);
static inline Real zlsp_Pd_T_table( Real T, CoolingPtr cp){
  Real  mu  =  zlsp_Mu_T( T,cp );
	return((k_mu/mu)*f_invtep0*T);
}


static inline Real zlsp_T_Pd( Real Pd, CoolingPtr cp);
static inline Real zlsp_T_Pd( Real Pd, CoolingPtr cp){
  Real T   = f_tep0*Pd;
#ifdef VARIABLE_MU  
  Real  mu  =  zlsp_Mu_Pd(Pd, cp);
  T        *= (mu*k_invmu);
#endif VARIABLE_MU
	return(T);
}

static inline Real zlsp_T_Ud ( Real Ud , CoolingPtr cp); // look up T for U/d
static inline Real zlsp_T_Ud ( Real Ud , CoolingPtr cp){ // look up T for U/d
  
	Real gamm  = kZLS_eos_gamma_1;
	Real T     = f_tep0*gamm*Ud; // P/d = U/d * (gamma-1), default mu/k_mu = 1.0;
  //   T     = (mu/k_mu)*f_tep0*Pd;
#ifdef VARIABLE_MU  
  Real  mu   = zlsp_Mu_Ud(Ud, cp);
#ifdef VARIABLE_GAMMA
  gamm       = zlsp_Gamma_Ud(Ud, cp) - 1.0;
#endif VARIABLE_GAMMA
  T         *= (mu*k_invmu);
#endif VARIABLE_MU
  
	return(T);
  
}


static inline Real zlsp_Pd_Ud( Real Ud, CoolingPtr cp);
static inline Real zlsp_Pd_Ud( Real Ud, CoolingPtr cp){
  
	Real gamm  = kZLS_eos_gamma_1;
  
#ifdef VARIABLE_GAMMA
  gamm       =  zlsp_Gamma_Ud(Ud, cp) - 1.0;
#endif VARIABLE_GAMMA
  
	Real Pd    = Ud*gamm;
  
	return(Pd);
}

static inline Real zlsp_Ud_Pd( Real Pd, CoolingPtr cp);
static inline Real zlsp_Ud_Pd( Real Pd, CoolingPtr cp){
  
  Real invgamm = kZLS_eos_invgamm;
#ifdef VARIABLE_GAMMA
  invgamm    = 1.0/( zlsp_Gamma_Pd(Pd, cp) - 1.0 );
#endif VARIABLE_GAMMA
	Real Ud    = Pd*invgamm;
  
	return( Ud );
}

static inline Real zlsp_En_Pd( Real P, Real d, CoolingPtr cp);
static inline Real zlsp_En_Pd( Real P, Real d, CoolingPtr cp){
  
  Real g = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  g    = zlsp_Gamma_Pd(P/d, cp);
#endif VARIABLE_GAMMA
	Real En   = P*(zls_Pow(d, (-g)));
  
	return( En );
}



static inline Real zlsp_Cs_T( Real T , CoolingPtr cp);
static inline Real zlsp_Cs_T( Real T , CoolingPtr cp){
  
	Real gamma = kZLS_eos_gamma;
  Real Pd     = f_invtep0*T;
#ifdef VARIABLE_GAMMA
  gamma   = zlsp_Gamma_Pd(Pd, cp);
#endif VARIABLE_GAMMA
  Real cs   = zls_sqrt(gamma*Pd);
	return( cs);
  
}

static inline Real zlsp_Cs_Pd( Real Pd , CoolingPtr cp); // look up gamma for P/d
static inline Real zlsp_Cs_Pd( Real Pd , CoolingPtr cp){ // look up gamma for P/d
	Real gamma = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  gamma   = zlsp_Gamma_Pd(Pd, cp);
#endif VARIABLE_GAMMA
  Real cs   = zls_sqrt(gamma*Pd);
	return( cs);
  
}

static inline Real zlsp_Cs_Ud ( Real Ud , CoolingPtr cp); // sqrt(gamma*p/d) for U/d
static inline Real zlsp_Cs_Ud ( Real Ud , CoolingPtr cp){ // sqrt(gamma*p/d) for U/d
  // U = P/(gamma-1.0), gamma P = gamma(gamma-1.0) U
	Real gamma = kZLS_eos_gamma;
	Real gamm = kZLS_eos_gamma_1;
#ifdef VARIABLE_GAMMA
  gamma     = zlsp_Gamma_Ud(Ud, cp);
#endif VARIABLE_GAMMA
  Real cs   = zls_sqrt(gamma*Ud*gamm);
	return( cs);
  
}


#ifdef SYM3D
static inline Real zlsp_KE_L_WHD( WHD zoneW, Integer k, Integer j, Integer i){
  return(0.5*(zoneW.vx[k][j][i] *zoneW.vx[k][j][i] 
              + zoneW.vy[k][j][i] *zoneW.vy[k][j][i] 
              + zoneW.vz[k][j][i] *zoneW.vz[k][j][i]));
}
static inline Real zlsp_KE_WHD( WHD zoneW, Integer k, Integer j, Integer i){
  return(0.5*zoneW.d[k][j][i]*(zoneW.vx[k][j][i] *zoneW.vx[k][j][i] 
                               + zoneW.vy[k][j][i] *zoneW.vy[k][j][i] 
                               + zoneW.vz[k][j][i] *zoneW.vz[k][j][i]));
}
static inline Real zlsp_Mom_L_WHD( WHD zoneW, Integer k, Integer j, Integer i){
  return(zls_sqrt(zoneW.vx[k][j][i] *zoneW.vx[k][j][i] 
                  + zoneW.vy[k][j][i] *zoneW.vy[k][j][i] 
                  + zoneW.vz[k][j][i] *zoneW.vz[k][j][i]));
}
static inline Real zlsp_Mom_WHD( WHD zoneW, Integer k, Integer j, Integer i){
  return(zoneW.d[k][j][i]*zls_sqrt(zoneW.vx[k][j][i] *zoneW.vx[k][j][i] 
                                   + zoneW.vy[k][j][i] *zoneW.vy[k][j][i] 
                                   + zoneW.vz[k][j][i] *zoneW.vz[k][j][i]));
}
#endif SYM3D

#ifdef SYM2D
static inline Real zlsp_KE_L_WHD( WHD zoneW, Integer j, Integer i){
  return(0.5*(zoneW.vx[j][i] *zoneW.vx[j][i] 
              + zoneW.vy[j][i] *zoneW.vy[j][i]));
}
static inline Real zlsp_KE_WHD( WHD zoneW, Integer j, Integer i){
  return(0.5*zoneW.d[j][i]*(zoneW.vx[j][i] *zoneW.vx[j][i] 
                            + zoneW.vy[j][i] *zoneW.vy[j][i]));
}
static inline Real zlsp_Mom_L_WHD( WHD zoneW, Integer j, Integer i){
  return(zls_sqrt(zoneW.vx[j][i] *zoneW.vx[j][i] 
                  + zoneW.vy[j][i] *zoneW.vy[j][i]));
}
static inline Real zlsp_Mom_WHD( WHD zoneW, Integer j, Integer i){
  return(zoneW.d[j][i]*zls_sqrt(zoneW.vx[j][i] *zoneW.vx[j][i] 
                                + zoneW.vy[j][i] *zoneW.vy[j][i]));
}
#endif SYM2D

#ifdef SYM1D
static inline Real zlsp_KE_L_WHD( WHD zoneW, Integer i){
  return(0.5*(zoneW.vx[i] *zoneW.vx[i]));
}
static inline Real zlsp_KE_WHD( WHD zoneW, Integer i){
  return(0.5*zoneW.d[i]*(zoneW.vx[i] *zoneW.vx[i]));
}
static inline Real zlsp_Mom_L_WHD( WHD zoneW, Integer i){
  return(zls_abs(zoneW.vx[i]));
}
static inline Real zlsp_Mom_WHD( WHD zoneW, Integer i){
  return(zoneW.d[i]*zls_abs(zoneW.vx[i]));
}
#endif SYM1D

static inline Real zlsp_PFromE ( Real E, Real kE, Real d){
  
  //
  // get gamma-1
  //
  Real gamm = kZLS_eos_gamma_1;
  
#ifdef LAGRANGIAN
  Real Ud   = (E - kE); // U/d
#endif LAGRANGIAN
  
#ifdef EULERIAN
  Real Ud   = (E - kE)/d; // U/d
#endif EULERIAN
  
#ifdef VARIABLE_EOS
  
#ifdef VARIABLE_GAMMA  
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
  gamm = zlsp_Gamma_Ud(Ud, cp)-1.0;
#endif VARIABLE_GAMMA
  
#endif VARIABLE_EOS
  
  Real P =  zls_max( kZLS_PressureVacuum, Ud*d*gamm);
  
  return(P);
  
}

static inline Real zlsp_EFromP ( Real P, Real kE, Real d){
  //
  // get gamma-1
  //
  Real gamm = kZLS_eos_gamma_1;
  
#ifdef VARIABLE_EOS
  
#ifdef VARIABLE_GAMMA  
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
  gamm = zlsp_Gamma_Pd((P/d), cp)-1.0;
#endif VARIABLE_GAMMA
  
#endif VARIABLE_EOS
  
#ifdef LAGRANGIAN
  Real E  = (P/(d*gamm) + kE);  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  Real E  = (P/(gamm) + kE);  
#endif EULERIAN
  
  return(E);
  
}


static inline Real zlsp_PFromEB ( Real E, Real kE, Real BE, Real d){
  
  //
  // get gamma-1
  //
  Real gamm = kZLS_eos_gamma_1;
  
#ifdef LAGRANGIAN
  Real Ud   = (E - kE - BE); // U/d
#endif LAGRANGIAN
  
#ifdef EULERIAN
  Real Ud   = (E - kE - BE)/d; // U/d
#endif EULERIAN
  
#ifdef VARIABLE_EOS
  
#ifdef VARIABLE_GAMMA  
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
  gamm = zlsp_Gamma_Ud(Ud, cp)-1.0;
#endif VARIABLE_GAMMA
  
#endif VARIABLE_EOS
  
  Real P = zls_max( kZLS_PressureVacuum, Ud*d*gamm);
  
  return(P);
  
}

static inline Real zlsp_EFromPB ( Real P, Real kE, Real BE, Real d){
  //
  // get gamma-1
  //
  Real gamm = kZLS_eos_gamma_1;
  
#ifdef VARIABLE_EOS
  
#ifdef VARIABLE_GAMMA  
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
  gamm = zlsp_Gamma_Pd((P/d), cp)-1.0;
#endif VARIABLE_GAMMA
  
#endif VARIABLE_EOS
  
#ifdef LAGRANGIAN
  Real E  = (P/(d*gamm) + kE + BE);  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  Real E  = (P/(gamm) + kE + BE);  
#endif EULERIAN
  
  return(E);
  
}

static inline Real zlsp_PFromE_W1D ( W1DArr w , Integer i){
  
  Real E = w.E[i].a;  // lagrangian total energy density per mass
  Real d = w.d[i].a;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].a;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real vz = w.vz[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].a;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real vz = w.vz[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);
  
#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_W1D ( W1DArr w , Integer i){
  
  Real P     = w.P[i].a;
  Real d     = w.d[i].a;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].a;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real vz = w.vz[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].a;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].a;
  Real vy = w.vy[i].a;
  Real vz = w.vz[i].a;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

static inline Real zlsp_PFromE_U1D ( U1DArr u , Integer i){
  
  Real E = u.E[i].a;  // lagrangian total energy density per mass
  Real d = u.d[i].a;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real mx = u.mx[i].a;
  Real kE = 0.5*(mx * mx )/(d*d);
#endif
#ifdef SYM2D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ))/(d*d);
#endif
#ifdef SYM3D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real mz = u.mz[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/(d*d);
#endif
#endif LAGRANGIAN
  
#ifdef EULERIAN
#ifdef SYM1D
  Real mx = u.mx[i].a;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real mz = u.mz[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);

#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_U1D ( U1DArr u , Integer i){
  
  Real P     = u.P[i].a;
  Real d     = u.d[i].a;
  
#ifdef HYDRO  
#ifdef SYM1D
  Real mx = u.mx[i].a;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].a;
  Real my = u.my[i].a;
  Real mz = u.mz[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
  
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

static inline Real zlsp_PFromE_W1DL ( W1DArr w , Integer i){
  
  Real E = w.E[i].al;  // lagrangian total energy density per mass
  Real d = w.d[i].al;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].al;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real vz = w.vz[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].al;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real vz = w.vz[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);
  
#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_W1DL ( W1DArr w , Integer i){
  
  Real P     = w.P[i].al;
  Real d     = w.d[i].al;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].al;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real vz = w.vz[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].al;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].al;
  Real vy = w.vy[i].al;
  Real vz = w.vz[i].al;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

static inline Real zlsp_PFromE_U1DL ( U1DArr u , Integer i){
  
  Real E = u.E[i].al;  // lagrangian total energy density per mass
  Real d = u.d[i].al;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real mx = u.mx[i].al;
  Real kE = 0.5*(mx * mx )/(d*d);
#endif
#ifdef SYM2D
  Real mx = u.mx[i].al;
  Real my = u.my[i].al;
  Real kE = 0.5*((mx * mx ) + (my *my ))/(d*d);
#endif
#ifdef SYM3D
  Real mx = u.mx[i].al;
  Real my = u.my[i].al;
  Real mz = u.mz[i].al;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/(d*d);
#endif
#endif LAGRANGIAN
  
#ifdef EULERIAN
#ifdef SYM1D
  Real mx = u.mx[i].al;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].al;
  Real my = u.my[i].al;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].al;
  Real my = u.my[i].al;
  Real mz = u.mz[i].al;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);

#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_U1DL ( U1DArr u , Integer i){
  
  Real P     = u.P[i].al;
  Real d     = u.d[i].al;
  
#ifdef HYDRO  
#ifdef SYM1D
  Real mx = u.mx[i].al;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].al;
  Real my = u.my[i].a;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].al;
  Real my = u.my[i].al;
  Real mz = u.mz[i].al;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
  
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

static inline Real zlsp_PFromE_W1DR ( W1DArr w , Integer i){
  
  // derive pressure from total energy for right side of cells
  
  Real E = w.E[i].ar;  // lagrangian total energy density per mass
  Real d = w.d[i].ar;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].ar;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real vz = w.vz[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].ar;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real vz = w.vz[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);
  
#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_W1DR ( W1DArr w , Integer i){
  
  Real P     = w.P[i].ar;
  Real d     = w.d[i].ar;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].ar;
  Real kE = 0.5*(vx * vx );
#endif
#ifdef SYM2D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real vz = w.vz[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
  
#endif LAGRANGIAN
  
#ifdef EULERIAN
  // Eulerian kinetic energy component
  
#ifdef SYM1D
  Real vx = w.vx[i].ar;
  Real kE = 0.5*(vx * vx )*d;
#endif
#ifdef SYM2D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ))*d;
#endif
#ifdef SYM3D
  Real vx = w.vx[i].ar;
  Real vy = w.vy[i].ar;
  Real vz = w.vz[i].ar;
  Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ))*d;
#endif
  
#endif EULERIAN
  
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

static inline Real zlsp_PFromE_U1DR ( U1DArr u , Integer i){
  
  Real E = u.E[i].ar;  // lagrangian total energy density per mass
  Real d = u.d[i].ar;
  
#ifdef HYDRO
  
#ifdef LAGRANGIAN
  // lagrangian kinetic energy component
  
#ifdef SYM1D
  Real mx = u.mx[i].ar;
  Real kE = 0.5*(mx * mx )/(d*d);
#endif
#ifdef SYM2D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ))/(d*d);
#endif
#ifdef SYM3D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real mz = u.mz[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/(d*d);
#endif
#endif LAGRANGIAN
  
#ifdef EULERIAN
#ifdef SYM1D
  Real mx = u.mx[i].ar;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real mz = u.mz[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
#endif EULERIAN
  
  Real P = zlsp_PFromE( E, kE, d);

#endif HYDRO
    
  return(P);
  
}

static inline Real zlsp_EFromP_U1DR ( U1DArr u , Integer i){
  
  Real P     = u.P[i].ar;
  Real d     = u.d[i].ar;
  
#ifdef HYDRO  
#ifdef SYM1D
  Real mx = u.mx[i].ar;
  Real kE = 0.5*(mx * mx )/d;
#endif
#ifdef SYM2D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
  Real mx = u.mx[i].ar;
  Real my = u.my[i].ar;
  Real mz = u.mz[i].ar;
  Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
  
  Real E = zlsp_EFromP( P, kE, d);
  
#endif HYDRO
    
  return(E);
  
}

#endif ZLS_PHYSINLINES_H
