#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

typedef int Integer;
typedef double Real;

typedef struct varlr {
  Real       a;  // middle
  Real      al;  // left 
  Real      ar;  // right
} VarLR;

typedef struct w1{
  VarLR      d; 
  VarLR      P;
  VarLR      v;
} W1D; 

char    theFile[512];
FILE *  theFileRef;

Real  gam;
Real  alpha, beta, delta;
Real  g0, g1, g2, g3, g4, g5, g6, g7;

void InitGammas(Real g);
void Usage(void);
int Initialize(int argc, char *argv[]);
int Finalize(void);

void PrintW1Df( FILE * f, W1D w);
void WriteSamples(Integer n, Integer subsample, Real xl, Real xr, Real x0, Real t, W1D *wStar);

void RiemanSample(Real * d0, Real *p0, Real * u0, Real s,  W1D * wStar);
int WaveLocations(Real t, Real x0, W1D * wStar);
int GG88RiemannW (W1D * wStar);

int main( int argc, char *argv[] ) {
  
  int err = 0;
  
  err = Initialize(argc, argv);
  
  if ( err  == 0 ){
    
    W1D Win;   
    
    Win.d.a = 0.0;
    Win.P.a = 0.0;
    Win.v.a = 0.0;
    
    Real x0, t;
    fprintf(stdout," x0, time t : ");
    fscanf(stdin,"%lf %lf",&x0, &t);
    
    fprintf(stdout,"\n Riemann Variables: \n");   
    fprintf(stdout," Left  W: d, P, v : ");   
    fscanf(stdin,"%lf %lf %lf",&Win.d.al, &Win.P.al, &Win.v.al);
    
    fprintf(stdout," Right W: d, P, v : ");
    fscanf(stdin,"%lf %lf %lf",&Win.d.ar, &Win.P.ar, &Win.v.ar);
    
    // PrintW1Df(stdout, Win);
    
    Real xl = 0.0;
    Real xr = 1.0;
    
    fprintf(stdout," Sample Range, xLeft, xRight  : ");   
    fscanf(stdin,"%lf %lf",&xl, &xr);
    
    Integer n;
    fprintf(stdout," Sample Resolution (ncells>0) : ");   
    fscanf(stdin,"%ld",&n);
    
    Integer os;
    fprintf(stdout," Oversample (n subcells>0) : ");   
    fscanf(stdin,"%ld",&os);
    
    fprintf(stdout," Solving and saving results to %s : \n\n", theFile);   
    
    GG88RiemannW(&Win);
    
    WaveLocations( t, x0, &Win);
    
    WriteSamples ( n, os, xl, xr, x0, t, &Win);
    
  } else {
    fprintf( stdout, " ERROR: Initialize failed %d \n", err);
  }
  
  Finalize();
  
  return (err);
}


void PrintW1Df( FILE * f, W1D w){
  fprintf(f, "#--------------------------------------------------------------\n");
  fprintf(f," Left  W: d %#11.6g  P %#11.6g v  %#11.6g \n", w.d.al, w.P.al, w.v.al);   
  fprintf(f," Right W: d %#11.6g  P %#11.6g v  %#11.6g \n", w.d.ar, w.P.ar, w.v.ar);   
  fprintf(f," Mid   W: d %#11.6g  P %#11.6g v  %#11.6g \n", w.d.a , w.P.a , w.v.a );   
  fprintf(f, "#--------------------------------------------------------------\n");
}

void WriteSamples(Integer n, Integer subsample, Real xl, Real xr, Real x0, Real t, W1D *wStar){
  
  Real dx = (xr - xl)/(Real)n;

  if ( subsample <= 1){
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(stdout,     "#       N Cells: %d\n", n);
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#       N Cells: %d\n", n);
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  
  } else {
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(stdout,     "#       N Cells: %d  Oversample and Integrate: %d \n", n, subsample);
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#       N Cells: %d  Oversample and Integrate: %d \n", n, subsample);
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  }
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(stdout,     "#       <x>          <d>          <P>          <v>          <e>\n");
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#       <x>          <d>          <P>          <v>          <e>\n");
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  
  if ( subsample <= 1){
  
	  Integer i;
	  for ( i = 0; i < n; i++){
	    
	    Real d0, p0, v0, e0;
	    
	    Real sp  = (xl+((i+0.5)*dx)-x0)/t;
	    
	    RiemanSample( &d0, &p0, &v0, sp, wStar);
	    e0 = p0/(d0*(gam-1.0));
	    
	    fprintf ( stdout, "%#11.6g  %#11.6g  %#11.6g  %#11.6g  %#11.6g \n", 
			  (xl+((i+0.5)*dx)), d0, p0, v0, e0);
	    fprintf ( theFileRef, "%#11.6g  %#11.6g  %#11.6g  %#11.6g  %#11.6g \n", 
			  (xl+((i+0.5)*dx)), d0, p0, v0, e0);
	    
	  }

	} else {
	  
	    
	  Real inv = 1.0/(Real)subsample;
	  Real ddx = dx*inv;
  
	   
	   Integer i;
	   
	   for ( i = 0; i < n; i++){
		   
		    Real d0 = 0.0;
		    Real p0 = 0.0;
		    Real v0 = 0.0;
		    Real e0 = 0.0;
		
		    Integer j;
		    for ( j = 1; j <= subsample; j++){
		
		    Real ps, vs, ds, es;
		
		    Real sp  = (xl+((i+0.0)*dx + (j-0.5)*ddx)-x0)/t;
		    
		    RiemanSample( &ds, &ps, &vs, sp, wStar);
		    es = ps/(ds*(gam-1.0));
		    
		    d0 += ds;
		    p0 += ps;
		    v0 += vs;
		    e0 += es;
		    
		    }
		    
		    d0 *= inv;
		    p0 *= inv;
		    v0 *= inv;
		    e0 *= inv;
		  
		    fprintf ( stdout, "%#11.6g  %#11.6g  %#11.6g  %#11.6g  %#11.6g \n", 
				  (xl+((i+0.5)*dx)), d0, p0, v0, e0);
		    fprintf ( theFileRef, "%#11.6g  %#11.6g  %#11.6g  %#11.6g  %#11.6g \n", 
				  (xl+((i+0.5)*dx)), d0, p0, v0, e0);
		    
		  }
	  }
  
  fprintf(stdout,     "#--------------------------------------------------------------\n");
  fprintf(theFileRef, "#--------------------------------------------------------------\n");
  
}

void InitGammas(Real g){
  
  gam = g;
  
// Groth and Gottlieb solver constants
  
  alpha = ((gam+1.0)/4.0);
  beta  = ((gam-1.0)/2.0);
  delta = (2.0*gam)/(gam-1.0);
  
//      
// Toro sampler constants
//
  g0 = (gam);
  g1 = (gam-1.0)/(2.0*gam);
  g2 = (gam+1.0)/(2.0*gam);
  g3 = (2.0*gam)/(gam-1.0);
  g4 = 2.0/(gam-1.0);
  g5 = 2.0/(gam+1.0);
  g6 = (gam-1.0)/(gam+1.0);
  g7 = (gam-1.0)/2.0;
}


void Usage(void){
  
  fprintf(stderr,"%s\n", "Usage: riem [options]");
  exit(-1);
  
}


int Initialize(int argc, char *argv[]){
	int c;
  char  optarg[256];
  
  strcpy( theFile , "riem_out.txt");
  
  while ((c = getopt(argc, argv, "o:h")) != -1){
    switch (c) {
      case 'o':
        fprintf( stdout, " outfile = %s\n",optarg);
        strcpy( theFile, optarg);
        break;
      case 'h':
      case '?':
        Usage();
        
    }
  }
  
  Real g = 1.4;
  fprintf(stdout," Polytropic gamma : ");
  fscanf(stdin,"%lf",&g);
  
  if (( g >= 1.0) && ( g <= 2.0)) {
    InitGammas(g);
  } else {
    return(-1);
  }
  
  theFileRef = fopen(theFile, "w");
  
  return(0);
  
}

int Finalize(void){
  
  fclose(theFileRef);
  
  return(0);
  
}


void RiemanSample(Real * d0, Real *p0, Real * u0, Real s,  W1D * wStar){
//
//     Riemann solution sampler at speed = s
//     returns d0, p0, u0 
//     Use s = 0.0 for Eulerian Fluxes
//
  Real shl, shr, c, sr, sl, pml, pmr, cml, cmr, stl , str;
  
// cache arguments:	
  
  Real pm = wStar->P.a;
  Real um = wStar->v.a; 
  
  Real dl = wStar->d.al; 
  Real pl = wStar->P.al;
  Real ul = wStar->v.al; 
  
  Real cl = sqrt(gam*pl/dl);
  
  Real dr = wStar->d.ar; 
  Real pr = wStar->P.ar;
  Real ur = wStar->v.ar; 
  
  Real cr = sqrt(gam*pr/dr);
  
  if (s<=um) {
  //
  //    to the left of the CD
  //      
    if (pm <= pl) {
    //  left rarefaction      
      shl = ul - cl;
      if (s <= shl) {
//  in left state           
        *u0 = ul;
        *p0 = pl;
        *d0 = dl;
      } else {
        cml = cl*pow((pm/pl), g1);
        stl = um - cml;
        if (s>stl) {
  //  in left star state           
          *d0 = dl*pow((pm/pl), (1.0/gam));
          *p0 = pm;
          *u0 = um;
        }else{
  //  in left fan state    
          *u0 = g5*(cl + g7*ul+s);
          c   = g5*(cl + g7*(ul-s));
          *d0 = dl*pow((c/cl), g4);
          *p0 = pl*pow((c/cl), g3);
        }
      }
    } else {
    //  left shock
      pml = pm/pl;
      sl  = ul - cl*sqrt(g2*pml + g1);
      if (s<=sl) {
//  in left state     
        *d0 = dl;
        *p0 = pl;
        *u0 = ul;  
      } else {
//  in left star state
        *d0 = dl*(pml+g6)/(pml*g6+1.0);
        *u0 = um;
        *p0 = pm;
      }
    }
  } else{
  //
  //    to the right of the CD
  //      
    if(pm>pr) {
    //   right shock
      pmr = pm/pr;
      sr  = ur + cr*sqrt(g2*pmr + g1);
      if (s>=sr){
//   in right state
        *d0 = dr;
        *u0 = ur;
        *p0 = pr; 
      } else {
//   in right star state
        *d0 = dr*(pmr + g6)/(pmr*g6 + 1.0);
        *u0 = um;
        *p0 = pm;
      }
    } else {
    //   right rarefaction
      shr = ur + cr;
      if (s>=shr) {
//   in right state
        *d0 = dr;
        *u0 = ur;
        *p0 = pr;
      } else {
        cmr =cr*pow((pm/pr),g1);
        str = um + cmr; 
        if (s<=str) {
  //  in right star state
          *d0 = dr*pow((pm/pr),(1.0/gam));
          *u0 = um;
          *p0 = pm;
        } else {
  //  in right fan state    
          *u0 = g5*(-cr + g7*ur+s);
          c  = g5*( cr - g7*(ur-s));
          *d0 = dr*pow((c/cr),g4);
          *p0 = pr*pow((c/cr),g3);
        }
      }
    }
  }
  
}

int GG88RiemannW (W1D * wStar){
  int err = 0;
  
  Integer i, is, ie;
  Integer nti;
  
  Integer iter;
  Real iter_err;
  
  
  Real Pl; // Pressure
  Real Pr;
  Real dl; // density
  Real dr;
  Real vl; // velocity
  Real vr;
  
  Real z, ul, ur; // initial speed guess
  
// Riemann vars	
  Real newtonUstar;
  Real newtonStep;
  
  Real Ustar = 0.0; 
  Real Pstar = 0.0;
  
  Real Pstarl, Pstarr;
  Real dPstarl, dPstarr;
  Real astarl, astarr;
  
  Real al, ar; // Eulerian sound speed

  Real cl, cr; // Lagrangian sound speed
  Real sl, sr; // intermediate wave slopes
  Real zl, zr; // shock mach number
  
  Real alphal; // (gamma+1)/(4a), predivide and collect constants
  Real alphar; // predivide and collect constants
  
  dl  = wStar->d.al;
  dr  = wStar->d.ar;  
  
  Pl  = wStar->P.al;
  Pr  = wStar->P.ar;
  
  vl  = wStar->v.al;
  vr  = wStar->v.ar;
  
  al = sqrt(gam*Pl/dl);  // eulerian speed
  ar = sqrt(gam*Pr/dr);
  
  alphal = alpha/al; // (gamma+1)/(4a), predivide and collect constants
  alphar = alpha/ar; // predivide and collect constants
  
  cl = dl*al; // lagrangian speed
  cr = dr*ar;	
  
// initial guess
  
  ul = vl + al/beta;
  ur = vr - ar/beta;
  
  z = (ar/al)*pow((Pl/Pr), 1.0/delta);
  newtonUstar = (ul*z + ur)/(1.0+z);
  
  iter_err = 1.0;
  for (iter = 0; ((iter < 20)&&( iter_err > 1.0e-12)); iter++) {
    
    Ustar  = newtonUstar;
    
    if ( Ustar <= vl ) { // left wave is shock
      
      sl      = (newtonUstar - vl)*alphal;
      zl      = sl - sqrt(1.0+(sl*sl));
      Pstarl  = Pl + cl*(newtonUstar - vl)*zl;
      dPstarl = 2*cl*zl*zl*zl/(1.0+zl*zl);
      
    } else { // left wave is rarefaction 
      
      astarl  = al - beta*(newtonUstar - vl);
      Pstarl  = Pl * pow(astarl/al, delta);
      dPstarl = -gam*Pstarl/astarl;
      
    }
    
    if ( Ustar >= vr ) { // right wave is shock
      
      sr      = (newtonUstar - vr)*alphar;
      zr      = sr + sqrt(1.0+(sr*sr));
      Pstarr  = Pr + cr*(newtonUstar - vr)*zr;
      dPstarr = 2*cr*zr*zr*zr/(1.0+zr*zr);
      
    } else { // right wave is rarefaction
      
      astarr  = ar + beta*(newtonUstar - vr); // note +
      Pstarr = Pr * pow(astarr/ar, delta);
      dPstarr = gam*Pstarr/astarr;
      
    }
    
    
    register Real deltaP     = ( Pstarl -  Pstarr);
    register Real deltaDP    = (dPstarl - dPstarr);
    newtonStep = deltaP/deltaDP;
    
    newtonUstar  -= newtonStep; 
    iter_err      = fabs(newtonStep); // absolute error, faster as long as delta V is normalised
//
  } // iter
  
  Ustar  = newtonUstar;
  Pstar  = 0.5*(Pstarl+Pstarr);
  
  wStar->P.a   = Pstar; 
  wStar->v.a   = Ustar; 
  
  return (err);   
}



int WaveLocations(Real t, Real x0, W1D * wStar){
  int err = 0;
  
  
  Real pm = wStar->P.a;
  Real um = wStar->v.a;
  
  Real pl = wStar->P.al;
  Real ul = wStar->v.al;
  Real dl = wStar->d.al;
  Real cl = sqrt(gam*pl/dl);
  
  Real pr = wStar->P.ar;
  Real ur = wStar->v.ar;
  Real dr = wStar->d.ar;
  Real cr = sqrt(gam*pr/dr);
  
  
  Real rmr, rml;
  
  Real shl, shr, sr, sl, pml, pmr, cml, cmr, stl , str;
  
  
  fprintf ( theFileRef, "#--------------------------------------------------------------\n");
  fprintf ( theFileRef, "# Exact Rieman Solution\n");
  fprintf ( theFileRef, "#--------------------------------------------------------------\n");
  fprintf ( theFileRef, "# Gamma = %#11.6g\n",gam);
  fprintf ( theFileRef, "# x0    = %#11.6g\n",x0);
  fprintf ( theFileRef, "# t     = %#11.6g\n",t);
  fprintf ( theFileRef, "#--------------------------------------------------------------\n");
  
  fprintf ( stdout, "#--------------------------------------------------------------\n");
  fprintf ( stdout, "# Exact Rieman Solution\n");
  fprintf ( stdout, "#--------------------------------------------------------------\n");
  fprintf ( stdout, "# Gamma = %#11.6g\n",gam);
  fprintf ( stdout, "# x0    = %#11.6g\n",x0);
  fprintf ( stdout, "# t     = %#11.6g\n",t);
  fprintf ( stdout, "#--------------------------------------------------------------\n");
  
  if (pm<=pl) {
    shl = ul - cl;
    cml = pow(cl*(pm/pl),g1);
    stl = um - cml;
    fprintf(stdout,     "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shl*t, x0+stl*t);
    fprintf(theFileRef, "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shl*t, x0+stl*t);
    rml = pow(dl*(pm/pl),(1.0/g0));
  } else { 
    pml = pm/pl;
    sl  = ul - cl*sqrt(g2*pml + g1);
    fprintf(stdout,     "#           S-Wave x = %#11.6g \n",x0+sl*t);
    fprintf(theFileRef, "#           S-Wave x = %#11.6g \n",x0+sl*t);
    rml = dl*(pml+g6)/(pml*g6+1.0);
  }
  
  fprintf(theFileRef,   "# Contact Discont. x = %#11.6g \n",x0+um*t);
  fprintf(stdout,       "# Contact Discont. x = %#11.6g \n",x0+um*t);
  
  if(pm>pr) {
    pmr = pm/pr;
    sr  = ur + cr*sqrt(g2*pmr + g1);
    fprintf(stdout,     "#           S-Wave x = %#11.6g \n",x0+sr*t);
    fprintf(theFileRef, "#           S-Wave x = %#11.6g \n",x0+sr*t);
    rmr = dr*(pmr+g6)/(pmr*g6+1.0);
  } else {
    shr = ur + cr;
    cmr = pow(cr*(pm/pr),g1);
    str = um + cmr;
    fprintf(    stdout, "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shr*t, x0+str*t);
    fprintf(theFileRef, "#           R-Wave x = %#11.6g  %#11.6g\n",x0+shr*t, x0+str*t);
    rmr = pow(dr*(pm/pr),(1.0/g0));
  }
  
  
  fprintf ( theFileRef, "#--------------------------------------------------------------\n");     
  fprintf ( theFileRef, "# d = %#11.6g  %#11.6g  %#11.6g %#11.6g\n",dl, rml, rmr, dr);
  fprintf ( theFileRef, "# P = %#11.6g       %#11.6g       %#11.6g\n",pl, pm, pr);
  fprintf ( theFileRef, "# v = %#11.6g       %#11.6g       %#11.6g\n",ul, um, ur);
  fprintf ( theFileRef, "#--------------------------------------------------------------\n");
  
  fprintf ( stdout, "#--------------------------------------------------------------\n");
  fprintf ( stdout, "# d = %#11.6g %#11.6g %#11.6g %#11.6g\n",dl, rml, rmr, dr);
  fprintf ( stdout, "# P = %#11.6g       %#11.6g       %#11.6g\n",pl, pm, pr);
  fprintf ( stdout, "# v = %#11.6g       %#11.6g       %#11.6g\n",ul, um, ur);
  fprintf ( stdout, "#--------------------------------------------------------------\n");
  
  
  
  return(err);
}

