#ifndef ZLS_MPI
#define ZLS_MPI

#ifdef MPI

#include "zls.h"

//=============================================================================
// Highlevel calls for general fyris/zls code to call
//=============================================================================
//
// zls_MPI_ImposeLevelLinks - the mpi version of links, handles all links
// remote links to other MPI ranks and local links within the current rank.
//
// scans a level of zones, sets all current process zone link send buffers,
// looks for remote zones and queues up a series of 
// non-blocking receive requests, then issues sends for all current process
// zone buffers, before waiting for all pending recieves to be completed
// and then the BCs are set from the receive buffers, and a final MPI_Barrier 
// ensures all MPI links and in-rank links are set before proceeding.
//

int zls_MPI_ImposeLevelLinks( Integer level );

//
int zls_MPI_GetLevelMergedZoneData(ZonePtr mergedZone, Integer level, Integer varIndex );
//
int zls_MPI_GetRemoteZoneData(  ZonePtr  z, ZonePtr zR, Integer varIndex );
int zls_MPI_SetRemoteZoneData(  ZonePtr zR,  ZonePtr z, Integer varIndex );

//=============================================================================
// lowlevel calls for internal MPI implementation calls
//=============================================================================
//
// MPI transaction request list routines for non-blocking receives.
//

int zls_MPI_CountLevelRecieveRequests     ( int * nReqs, Integer level );
int zls_MPI_CountLevelRecieveRequestsDim  ( int * nReqs, Integer level, Integer dim );
int zls_MPI_CountLevelRecieveRequestsDir  ( int * nReqs, Integer level, Integer dir );

int zls_MPI_ClearRequests( MPI_Request * reqs, int n);
int zls_MPI_WaitRequests ( MPI_Request * reqs, int n);
//
// requests a non-blocking receive, and returns request object in 
// reqs[n] for later use in a wait_all
//
int zls_MPI_NonBlockingReceiveBuffers( ZonePtr z, ZonePtr zR, Integer linkDir, MPI_Request * reqs, int n);
//
// standard MPI recv, wait all later not used, so request list not needed.
//
int zls_MPI_ReceiveBuffers( ZonePtr z, ZonePtr zR, Integer linkDir);
//
// standard MPI send , may or may not return immediately, depending on receives and or 
// MPI buffering system.
//
int zls_MPI_SendBuffers( ZonePtr zR, ZonePtr z, Integer linkDir);
//
//  MPI ready send , requires a pending recieve already in place, more efficient
// in some cases than standard send
//
int zls_MPI_ReadySendBuffers( ZonePtr zR, ZonePtr z, Integer linkDir);
//
// filling and getting buffers to and from grid and BCs
// sets the recieve buffers in zR to the send buffer in z in the link direction
// from z to zR
int zls_MPI_SetBuffersDirect(  ZonePtr zR, ZonePtr z, Integer linkDir);
//
int zls_SetSendBufferFromGridAll      ( ZonePtr z );
int zls_SetSendBufferFromGridAllLinks ( ZonePtr z );
int zls_SetSendBufferFromGridDir      ( ZonePtr z, Integer linkDir );

//
// Get calls get data from buffers back into the zone data
//

int zls_GetBCsfromReceiveBufferAll     ( ZonePtr z );
int zls_GetBCsfromReceiveBufferAllLinks( ZonePtr z );  
int zls_GetBCsfromReceiveBufferDir     ( ZonePtr z, Integer linkDir );

#endif MPI

#endif ZLS_MPI
