#ifndef ZLS_ARRS_H
#define ZLS_ARRS_H
#include <stdlib.h>
//
// private array creater/freers/setters/releasers, public interface
// should use New/Dispose functions
//
// uncomment if needed when compiling into external projects:
// typedef size_t Size;
//
// private array constructors, public interface should use New functions
//
//
// these primitive routines return pointers, not error values, public
// interfaces use errors
//
// 
// a[k][j][i] is the ith element of the jth row, of the kth layer, i
// sequential in memory,
// 
// use k,j,i loops like FORTRAN and reverse the indices:   C a[k][j][i]
// == FORTAN a(i,j,k) ie   loop k{j{i{ a[k][j][i]}}} , and x(i) is
// sequential (typical backwards C)
//
// This is equivalent to a loop: do k do j do i a(i,j,k) enddo enddo
// enddo, i sequential in memory in Fortran.
//
// The actual 3D data type is pointed to by a[0][0], and the rest of
// the 2D plane cross ref the 3D block rows and layers. a[][][] thus
// has 3 malloc'd blocks, a[0][0], 3d of elements. a[0] 2d of (void*)
// and a 1d of (void *) freeing this is in reverse order - trash
// a[0][0], then a[0] then a, after checkign that a, a[0] and
// a[0][0][0] exist first.  zlsFree3DArr calles zlsFree2DArr after
// disposing of a[0][0] to do this.
//
// All elementsizes in bytes.  The type Size allows for 64 bit memory
// systems.
//

void **  zlsAlloc2DArr(Size elementsize, Size nj, Size ni);
void *** zlsAlloc3DArr(Size elementsize, Size nk, Size nj, Size ni);

void zlsFree2DArr(void **a);
void zlsFree3DArr(void ***a);

// Set up arrays around an existing  data block

void **  zlsSet2DArr(void * data, Size elementsize, Size nj, Size ni);
void *** zlsSet3DArr(void * data, Size elementsize, Size nk, Size nj, Size ni);

//
// delete offset allocations, but leave data alone, assuming there is a reference to the 
// data elsewhere, goes with Set
//
void zlsRelease2DArr(void ** a);
void zlsRelease3DArr(void ***a);
  
// copy all elements from src to dst quickly,inline for speed

static inline void zlsCopy1DArr( void * dst, void * src, Size elementsize, Size ni){ 
  if ((dst)&&(src)){
    bcopy((void *) src,(void *) dst, ni*elementsize); 
  }                       
}

// copy all elements from src to dst quickly,inline for speed

static inline void zlsCopy2DArr( void ** dst, void ** src, Size elementsize, Size nj, Size ni){ 
  if ((dst[0])&&(src[0])){
    bcopy((void *) src[0],(void *) dst[0], ni*nj*elementsize); 
  }                       
}


static inline void zlsCopy3DArr( void *** dst, void *** src, Size elementsize, Size nk, Size nj, Size ni){ 
  if ((dst[0][0])&&(src[0][0])){
    bcopy((void *) src[0][0],(void *) dst[0][0], ni*nj*nk*elementsize); 
  }                       
}

// set all elements to 0 quickly

static inline void zlsZero1DArr(void * a, Size elementsize, Size ni){ 
  if (a){
    bzero((void *) a, ni*elementsize); 
  }                       
}

static inline void zlsZero2DArr(void ** a, Size elementsize, Size nj, Size ni){ 
  if (a[0]){
    bzero((void *) a[0], ni*nj*elementsize); 
  }                       
}

static inline void zlsZero3DArr( void *** a, Size elementsize, Size nk, Size nj, Size ni){ 
  if (a[0][0]){
    bzero((void *) a[0][0], ni*nj*nk*elementsize); 
  }                       
}

#endif
