#ifndef ZLS_BOUNDARY_H
#define ZLS_BOUNDARY_H

#include "zls.h"

//
//========================================
//
//  Routines for filling/linking entire levels and their BCs if present.,,
//
//========================================
//
int zlsbc_Fill2DSuperLevelFromLevel       ( Integer level );
int zlsbc_FillSuperLevelFromLevel         ( Integer level );

int zlsbc_ImposeLevelSuperBCs             ( Integer level, int iteration );
int zlsbc_ImposeLevelLinks                ( Integer level );
int zlsbc_ImposeLevelBCs			             ( Integer level );

#ifdef ZLS_USER_BOUNDARIES
int zlsbc_ImposeLevelUserBCs              ( Integer level );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsbc_ImposeLevelUserUpdates          ( Integer level );
#endif ZLS_USER_UPDATE_ROUTINE

int zlsbc_ImposeAllLevelBCs               ( Integer level, int iteration );

// for debugging problematic grids...

int  zlsbc_PrintLevelBCsAndLinks          ( Integer level);
//
//========================================
//
//  Routines for filling/linking zones and their BCs if present.,,
//
//========================================
//
int zlsbc_FillUpperZoneFromZone         ( ZonePtr zDest, ZonePtr z, int boundaryExpand );
int zlsbc_FillSuperZoneFromZone         ( ZonePtr z, int boundaryExpand );
int zlsbc_FillZoneFromSuperZone         ( ZonePtr z, int inclBCs );

int zlsbc_FillZoneBCsFromSuperZone1D    ( ZonePtr z, int iteration );
int zlsbc_FillZoneBCsFromSuperZone2D    ( ZonePtr z, int iteration );
int zlsbc_FillZoneBCsFromSuperZone3D    ( ZonePtr z, int iteration );
int zlsbc_FillZoneBCsFromSuperZone      ( ZonePtr z , int iteration); // Generic call

int  zlsbc_ImposeZoneSuperBCs         ( ZonePtr z, int iteration ); // check boundary types then call fill if needed

int zlsbc_FillZoneBCsFromNextZone     ( ZonePtr z, int direction );
int zlsbc_ImposeZoneLinks             ( ZonePtr z, int direction); 

int zlsbc_CalcZoneBCs                 ( ZonePtr z, int direction, int bcType );
int zlsbc_ImposeZoneBCs               ( ZonePtr z, int direction );

#ifdef ZLS_USER_BOUNDARIES
int zlsbc_ImposeZoneUserBCs           ( ZonePtr z, int direction );
#endif ZLS_USER_BOUNDARIES


//========================================
//
// Similar types of routines for copying zone data to the level's ( or
// any other ) image zone (type  ==  kImageZoneType )
//
// Image zones are inteded to be the same dims and coordinates as the
// zone's super zone.  All that is needed is the z->zoneSuperRange even
// if zoneSuper is NULL (ie level 0 zones)
// 
// Initialy ZLS will use just one level 0 image zone with one zone data
// to save memory. data from sub level can be propogaed up to level 0
// and then the level image formed when contiguous output from multiple
// levels is needed.  An alternative would be to create an image zone
// for each level separately, and of course individual zone output will
// stilll be possible.
//
// Will be able to copy data from zones according to varIndex, or
// compute derived quantities and copy those in two different versions
// of the routines, that may also use io output keys (not decided yet).
// First up is only the var copying, composite variables will be
// separate for now. (ie 'dens' is OK, 'temp' is not
//
// Special note: image zones have the same dimensionality as the source
// zone, ie 1D 2D and 3D, so they are more than just images.  in 3D
// they can be sliced and summed like normal zones with similar
// (simpler) routines.
//
//
int  zlsim_FillImageZoneFromZoneVarIndex ( ZonePtr iz,  ZonePtr z, Integer varIndex );
//
//========================================
//
// like above, but zlsim_FillImageZoneFromData  uses provided grid to put
// plain block data (a simple QPtr), often from a source outside a
// zone, into an image zone.  Most useful if the data only represents a
// sub area of the image zone otherwise a set zone data could be used,
// and be faster.  
//
// The image zone is assumed to have just one anon var, and the data
// input is a simple block and the grid provided is consistent withe
// the image zone being a valid super grid of the grid, ie boundary
// cells if any are satisfied.
// 
// Only works on anonymous image zone types.
//
// Finally, data doens't have BCs unless FILES_HAVE_BCS is set
//
// Useage: when a derived quantity is obtained from a zone and the
// results are being merged into a single image zone, the data returned
// from a GetWQuantity call can be passed here, whether ic came from a
// literal var or was computed on the fly.  Use the grid from the
// original zone to provide the coordinates, and assume image zone
// coords are already set up properly in initialization
//
// this is not as elegant as a zone to zone transfer, but does allow
// outside data to be copied through the scaling FillSuper process into
// part of a zone.  If the data exactly matches the destination one
// dimensions then it is much quicker to use the zls_SetZoneDataCopy,
// or zls_SetZoneDataDirect routines.
//
enum { 
  kZLS_DataNoBCs  = 0,
  kZLS_DataHasBCs = 1
};

int  zlsim_FillImageZoneFromData       ( ZonePtr iz,  QPtr data, GridCoords grid, Range destRng, Integer nR, Integer dataHasBCs );

//

int  zlsim_FillLevelImageFromDataAndZone ( ZonePtr iz,  QPtr data, Integer dataHasBCs, ZonePtr coordZone );

#endif
