#ifndef ZLS_QS_H
#define ZLS_QS_H
#include "zls.h"
#include <zls_arrs.h>
//==============================================================================
//
// Plain, uniterpreted quantity arrays
//
// QPtr, D1DArr, Q2DArr, Q3DArr = Real 1, 2 and 3D
// blocks or multi-D arrays.
//
//==============================================================================
//
// Declared as Q * 1DArr, Q ** 2DArr and Q *** 3DArr, may accessed as
// Q[i], Q[j][i] and Q[k][j][i] or at single ptr from Q, Q[0] or
// Q[0][0] in 1D , 2D and 3D.
//
// zls_qs contains utility routines for allocating and deacllocating
// these arrays dynamically
//==============================================================================
//
// Requires typdefs for Real, Integer, Counter and Complex
//
// Usually provided by zls_OSdefs.h, but if zls_qs is needed independently then
// uncomment the following, and remove the include zls.h above.
//
//  #include "zls_arrs.h"
//
//  typedef   double          Real; // change to float for 32 bit data
//  typedef   int          Integer;
//  typedef   unsigned int Counter;
//#pragma pack(0)
//  typedef struct {
//     Real re, im;
//  } Complex;
//#pragma pack()
//  #define c_re(c)  ((c).re)
//  #define c_im(c)  ((c).im)
//
//  enum {
//    noErr   = 0, 
//    fileErr = -1, 
//    memErr  = -2
//  };
//==============================================================================

typedef Real    * QPtr;  // plainest block , 1 2 or 3D

typedef Real    * Q1DArr;  // 1D array of Qs
typedef Real   ** Q2DArr;  // 2D array of Qs q   [j][i]  q   [0] is QPtr
typedef Real  *** Q3DArr;  // 3D array of Qs q[k][j][i]  q[0][0] is QPtr

#ifdef SYM1D
typedef Q1DArr     QArr;
typedef Real     QBCArr;
#endif

#ifdef SYM2D
typedef Q2DArr     QArr;
typedef Q1DArr   QBCArr;
#endif

#ifdef SYM3D
typedef Q3DArr     QArr;
typedef Q2DArr   QBCArr;
#endif

//

int zls_NewQuantity   ( QPtr *q, RangePtr r ); // just use the dimension and integer parts of r
int zls_NewQuantity1D ( QPtr *q, Counter ni ); // same as zls_NewQuantity1DArr but for QPtr syntax
int zls_NewQuantity2D ( QPtr *q, Counter ni, Counter nj );
int zls_NewQuantity3D ( QPtr *q, Counter ni, Counter nj, Counter nk );

int zls_DisposeQuantity ( QPtr q );

//
// multi-dimensional array versions...
//

//
// Note reversed idices, reference arrays by  [j][i] or  [k][j][i]
//

int zls_NewQuantity1DArr ( Q1DArr *q, Counter ni); // same as zls_NewQuantity1D but for Q1DArr syntax
int zls_NewQuantity2DArr ( Q2DArr *q, Counter nj, Counter ni );
int zls_NewQuantity3DArr ( Q3DArr *q, Counter nk, Counter nj, Counter ni );

int zls_DisposeQuantity1DArr ( Q1DArr q );
int zls_DisposeQuantity2DArr ( Q2DArr q );
int zls_DisposeQuantity3DArr ( Q3DArr q );

int zls_DisposeQuantityArr ( QArr q );

//
// Wrap an array around and unstructured block, alternative to New, use
// Release to free without deleting data, no nBC option as the ns must
// equal the input data dims exactly Not needed for 1D arrays.
//

int zls_SetQuantity2DArr ( Q2DArr *q, QPtr data, Counter nj, Counter ni );
int zls_SetQuantity3DArr ( Q3DArr *q, QPtr data, Counter nk, Counter nj, Counter ni );

int zls_ReleaseQuantity2DArr ( Q2DArr q );
int zls_ReleaseQuantity3DArr ( Q3DArr q );

//
// Extract an unstructured block, useful for I/O
//

int zls_GetQuantity2DArr ( QPtr *data, Q2DArr q );
int zls_GetQuantity3DArr ( QPtr *data, Q3DArr q );

// Extract lower dim Qs from higher dim qs, all as multi-d arrays

int zls_SetQ1DArrQ2DArrY( Q1DArr q1, Q2DArr q, Integer nj, Integer i );
int zls_SetQ1DArrQ2DArrX( Q1DArr q1, Q2DArr q, Integer ni, Integer j );

int zls_SetQ1DArrQ3DArrX( Q1DArr q1, Q3DArr q, Integer ni, Integer k, Integer j ); 
int zls_SetQ1DArrQ3DArrY( Q1DArr q1, Q3DArr q, Integer nj, Integer k, Integer i ); 
int zls_SetQ1DArrQ3DArrZ( Q1DArr q1, Q3DArr q, Integer nk, Integer j, Integer i ); 

int zls_SetQ2DArrQ3DArrYZ( Q2DArr q2, Q3DArr q, Integer nj, Integer nk, Integer i ); 
int zls_SetQ2DArrQ3DArrZX( Q2DArr q2, Q3DArr q, Integer nk, Integer ni, Integer j ); 
int zls_SetQ2DArrQ3DArrXY( Q2DArr q2, Q3DArr q, Integer ni, Integer nj, Integer k ); 

//==============================================================================
// 
//  IQPtr, IQ1DArr, IQ2DArr, IQ3DArr - Integer blocks/Arrays
// 
//==============================================================================

typedef Integer    * IQPtr  ;   // plainest block , 1 2 or 3D

typedef Integer   * IQ1DArr;  // 1D array of Is
typedef Integer  ** IQ2DArr;  // 2D array of Is i[j][i]  i[0] is 
typedef Integer *** IQ3DArr;  // 3D array of Is i[k][j][i]  i[0][0] is 

#ifdef SYM1D
typedef IQ1DArr      IQArr;
typedef Integer    IQBCArr; // dimension -1 array for boundaries
#endif

#ifdef SYM2D
typedef IQ2DArr      IQArr;
typedef IQ1DArr    IQBCArr;
#endif

#ifdef SYM3D
typedef IQ3DArr      IQArr;
typedef IQ2DArr    IQBCArr;
#endif

//

int zls_NewIQuantity   ( IQPtr *i, RangePtr r ); // just use the dimension and integer parts of r
int zls_NewIQuantity1D ( IQPtr *i, Counter ni );
int zls_NewIQuantity2D ( IQPtr *i, Counter ni, Counter nj );
int zls_NewIQuantity3D ( IQPtr *i, Counter ni, Counter nj, Counter nk );

int zls_DisposeIQuantity ( IQPtr i );

//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by  [j][i] or  [k][j][i]
//

int zls_NewIQuantity2DArr ( IQ2DArr *i, Counter nj, Counter ni );
int zls_NewIQuantity3DArr ( IQ3DArr *i, Counter nk, Counter nj, Counter ni );

int zls_DisposeIQuantity2DArr ( IQ2DArr i );
int zls_DisposeIQuantity3DArr ( IQ3DArr i );

//
// Wrap an array around and unstructured block, alternative to New, 
// use Release to free without deleting data, no nBC option as the ns
// must equal the input data dims exactly
// Not needed for 1D arrays.
//

int zls_SetIQuantity2DArr ( IQ2DArr *i, IQPtr data, Counter nj, Counter ni );
int zls_SetIQuantity3DArr ( IQ3DArr *i, IQPtr data, Counter nk, Counter nj, Counter ni );

int zls_ReleaseIQuantity2DArr ( IQ2DArr i );
int zls_ReleaseIQuantity3DArr ( IQ3DArr i );

//
// Extract an unstructured block, useful for I/O
//

int zls_GetIQuantity2DArr( IQPtr *data, IQ2DArr i );
int zls_GetIQuantity3DArr( IQPtr *data, IQ3DArr i );

//
//==============================================================================
//
//   CQPtr, CQ1DArr, CQ2DArr, CQ3DArr : blocks/arrays of
//   Complex quantities. 
//
//==============================================================================


typedef Complex    * CQPtr;  // plainest block , 1 2 or 3D

typedef Complex    * CQ1DArr;  // 1D array of Qs
typedef Complex   ** CQ2DArr;  // 2D array of Qs q   [j][i]  q   [0] is QPtr
typedef Complex  *** CQ3DArr;  // 3D array of Qs q[k][j][i]  q[0][0] is QPtr

//

int zls_NewCQuantity1D ( CQPtr *cq, Counter ni ); // same as zls_NewQuantity1DArr but for QPtr syntax
int zls_NewCQuantity2D ( CQPtr *cq, Counter ni, Counter nj );
int zls_NewCQuantity3D ( CQPtr *cq, Counter ni, Counter nj, Counter nk );

int zls_DisposeCQuantity ( CQPtr cq );

//
// multi-dimensional array versions...
//

//
// Note reversed idices, reference arrays by  [j][i] or  [k][j][i]
//

int zls_NewCQuantity1DArr ( CQ1DArr *cq, Counter ni); // same as zls_NewQuantity1D but for Q1DArr syntax
int zls_NewCQuantity2DArr ( CQ2DArr *cq, Counter nj, Counter ni );
int zls_NewCQuantity3DArr ( CQ3DArr *cq, Counter nk, Counter nj, Counter ni );

int zls_DisposeCQuantity1DArr ( CQ1DArr cq );
int zls_DisposeCQuantity2DArr ( CQ2DArr cq );
int zls_DisposeCQuantity3DArr ( CQ3DArr cq );

//
// Wrap an offset array around and unstructured block, alternative to New, use
// Release to free without deleting data, no nBC option as the ns must
// equal the input data dims exactly: Not needed for 1D arrays.
//

int zls_SetCQuantity2DArr ( CQ2DArr *cq, CQPtr cdata, Counter nj, Counter ni );
int zls_SetCQuantity3DArr ( CQ3DArr *cq, CQPtr cdata, Counter nk, Counter nj, Counter ni );

int zls_ReleaseCQuantity2DArr ( CQ2DArr cq );
int zls_ReleaseCQuantity3DArr ( CQ3DArr cq );

//
// Extract an unstructured block, useful for I/O
//

int zls_GetCQuantity2DArr ( CQPtr *cdata, CQ2DArr cq );
int zls_GetCQuantity3DArr ( CQPtr *cdata, CQ3DArr cq );

// Extract lower dim CQs from higher dim cqs, all as multi-d arrays

int zls_SetCQ1DArrCQ2DArrY( CQ1DArr cq1, CQ2DArr cq, Integer nj, Integer i );
int zls_SetCQ1DArrCQ2DArrX( CQ1DArr cq1, CQ2DArr cq, Integer ni, Integer j );

int zls_SetCQ1DArrCQ3DArrX( CQ1DArr cq1, CQ3DArr cq, Integer ni, Integer k, Integer j ); 
int zls_SetCQ1DArrCQ3DArrY( CQ1DArr cq1, CQ3DArr cq, Integer nj, Integer k, Integer i ); 
int zls_SetCQ1DArrCQ3DArrZ( CQ1DArr cq1, CQ3DArr cq, Integer nk, Integer j, Integer i ); 

int zls_SetCQ2DArrCQ3DArrYZ( CQ2DArr cq2, CQ3DArr cq, Integer nj, Integer nk, Integer i ); 
int zls_SetCQ2DArrCQ3DArrZX( CQ2DArr cq2, CQ3DArr cq, Integer nk, Integer ni, Integer j ); 
int zls_SetCQ2DArrCQ3DArrXY( CQ2DArr cq2, CQ3DArr cq, Integer ni, Integer nj, Integer k ); 
//
//========================================
//
// Dimension limited Vector Quantities - stored in dimensioned zones,
// vector size/elements = the dimensions in force from SYM1D,2D, 3D
// ie 3D vectors in 3D arrays, 2D vectors in 2D arrays
// 1D vectors in 1D arrays are redundance and the same as QPtrs
// but are provided for symmetry in coding.
// 

typedef struct qv1{
  Real       a ;
} QVec1;

typedef struct qv2{
  Real       a[2];
} QVec2;

typedef struct qv3{
  Real       a[3];
} QVec3;


typedef QVec1   * VQ1DArr;  // 1D array of 1D values  vq[i].a
typedef QVec2  ** VQ2DArr;  // 2D array of 2D vectors vq[j][i].a[2]
typedef QVec3 *** VQ3DArr;  // 3D array of 3D vectors vq[k][j][i].a[3] 

#ifdef SYM1D
typedef QVec1         QVec;
typedef VQ1DArr      VQArr;
#endif
#ifdef SYM2D
typedef QVec2         QVec;
typedef VQ2DArr      VQArr;
#endif
#ifdef SYM3D
typedef QVec3         QVec;
typedef VQ3DArr      VQArr;
#endif

typedef QVec        * VQPtr;  // plainest block , 1 2 or 3D

int zls_NewVQuantity1D ( VQPtr *vq, Counter ni );
int zls_NewVQuantity2D ( VQPtr *vq, Counter ni, Counter nj );
int zls_NewVQuantity3D ( VQPtr *vq, Counter ni, Counter nj, Counter nk );

int zls_DisposeVQuantity ( VQPtr vq );

int zls_NewVQuantity1DArr ( VQ1DArr *vq, Counter ni);
int zls_NewVQuantity2DArr ( VQ2DArr *vq, Counter nj, Counter ni );
int zls_NewVQuantity3DArr ( VQ3DArr *vq, Counter nk, Counter nj, Counter ni );

int zls_DisposeVQuantity1DArr ( VQ1DArr vq );
int zls_DisposeVQuantity2DArr ( VQ2DArr vq );
int zls_DisposeVQuantity3DArr ( VQ3DArr vq );


//==============================================================================
//
// Some specific format quantities, intended for IO on specific file formats
// 
//==============================================================================
// 32 bit float quantity/arrays, always 32 bit float independent of whether
// Real is single or double precision
//==============================================================================
//
typedef float    * FQPtr;  // plainest block , 1 2 or 3D
//
typedef float    * FQ1DArr;  // 1D array of Qs
typedef float   ** FQ2DArr;  // 2D array of Qs q   [j][i]  q   [0] is QPtr
typedef float  *** FQ3DArr;  // 3D array of Qs q[k][j][i]  q[0][0] is QPtr
//
#ifdef SYM1D
typedef FQ1DArr     FQArr;
#endif
#ifdef SYM2D
typedef FQ2DArr     FQArr;
#endif
#ifdef SYM3D
typedef FQ3DArr     FQArr;
#endif
//
int zls_NewFQuantity   ( FQPtr *fq, RangePtr r ); // just use the dimension and integer parts of r
int zls_NewFQuantity1D ( FQPtr *fq, Counter ni ); // same as zls_NewQuantity1DArr but for QPtr syntax
int zls_NewFQuantity2D ( FQPtr *fq, Counter ni, Counter nj );
int zls_NewFQuantity3D ( FQPtr *fq, Counter ni, Counter nj, Counter nk );
//
int zls_DisposeFQuantity ( FQPtr fq );
//
// Note reversed idices, reference arrays by  [j][i] or  [k][j][i]
//
int zls_NewFQuantity1DArr ( FQ1DArr *fq, Counter ni); // same as zls_NewQuantity1D but for Q1DArr syntax
int zls_NewFQuantity2DArr ( FQ2DArr *fq, Counter nj, Counter ni );
int zls_NewFQuantity3DArr ( FQ3DArr *fq, Counter nk, Counter nj, Counter ni );
//
int zls_DisposeFQuantity1DArr ( FQ1DArr fq );
int zls_DisposeFQuantity2DArr ( FQ2DArr fq );
int zls_DisposeFQuantity3DArr ( FQ3DArr fq );
//
int zls_DisposeFQuantityArr ( FQArr fq );
//
// Wrap an array around and unstructured block, alternative to New, use
// Release to free without deleting data, no nBC option as the ns must
// equal the input data dims exactly Not needed for 1D arrays.
//
int zls_SetFQuantity2DArr ( FQ2DArr *fq, FQPtr data, Counter nj, Counter ni );
int zls_SetFQuantity3DArr ( FQ3DArr *fq, FQPtr data, Counter nk, Counter nj, Counter ni );
//
int zls_ReleaseFQuantity2DArr ( FQ2DArr fq );
int zls_ReleaseFQuantity3DArr ( FQ3DArr fq );
//
// Extract an unstructured block, useful for I/O
//
int zls_GetFQuantity2DArr ( FQPtr *data, FQ2DArr fq );
int zls_GetFQuantity3DArr ( FQPtr *data, FQ3DArr fq );

//==============================================================================
// 64 bit double quantity/arrays, always 64 bit double independent of whether
// Real is single or double precision
//==============================================================================
//
typedef double    * DQPtr;  // plainest block , 1 2 or 3D
//
typedef double    * DQ1DArr;  // 1D array of Qs
typedef double   ** DQ2DArr;  // 2D array of Qs q   [j][i]  q   [0] is QPtr
typedef double  *** DQ3DArr;  // 3D array of Qs q[k][j][i]  q[0][0] is QPtr
//
#ifdef SYM1D
typedef DQ1DArr     DQArr;
#endif
#ifdef SYM2D
typedef DQ2DArr     DQArr;
#endif
#ifdef SYM3D
typedef DQ3DArr     DQArr;
#endif
//
int zls_NewDQuantity   ( DQPtr *dq, RangePtr r ); // just use the dimension and integer parts of r
int zls_NewDQuantity1D ( DQPtr *dq, Counter ni ); // same as zls_NewQuantity1DArr but for QPtr syntax
int zls_NewDQuantity2D ( DQPtr *dq, Counter ni, Counter nj );
int zls_NewDQuantity3D ( DQPtr *dq, Counter ni, Counter nj, Counter nk );
//
int zls_DisposeDQuantity ( DQPtr dq );
//
// Note reversed idices, reference arrays by  [j][i] or  [k][j][i]
//
int zls_NewDQuantity1DArr ( DQ1DArr *dq, Counter ni); // same as zls_NewQuantity1D but for Q1DArr syntax
int zls_NewDQuantity2DArr ( DQ2DArr *dq, Counter nj, Counter ni );
int zls_NewDQuantity3DArr ( DQ3DArr *dq, Counter nk, Counter nj, Counter ni );
//
int zls_DisposeDQuantity1DArr ( DQ1DArr dq );
int zls_DisposeDQuantity2DArr ( DQ2DArr dq );
int zls_DisposeDQuantity3DArr ( DQ3DArr dq );
//
int zls_DisposeDQuantityArr ( DQArr dq );
//
// Wrap an array around and unstructured block, alternative to New, use
// Release to free without deleting data, no nBC option as the ns must
// equal the input data dims exactly Not needed for 1D arrays.
//
int zls_SetDQuantity2DArr ( DQ2DArr *dq, DQPtr data, Counter nj, Counter ni );
int zls_SetDQuantity3DArr ( DQ3DArr *dq, DQPtr data, Counter nk, Counter nj, Counter ni );
//
int zls_ReleaseDQuantity2DArr ( DQ2DArr dq );
int zls_ReleaseDQuantity3DArr ( DQ3DArr dq );
//
// Extract an unstructured block, useful for I/O
//
int zls_GetDQuantity2DArr ( DQPtr *data, DQ2DArr dq );
int zls_GetDQuantity3DArr ( DQPtr *data, DQ3DArr dq );

#endif ZLS_QS_H
