#include <zls.h>

const Real kZLS_TimeEventSlop = 1.15; 

CPUCoords gZLS_GlobalCPUTimer;

// these are not in a timecoords structure as they are not strictly part of a
// simulation state, being reinstated each run, and gZLS_GlobalSimRealTime != 0
// gnerally.  
static time_t    gZLS_GlobalSimRealTime         =  0;
static time_t    gZLS_GlobalSimRealTimeLimit    = -1;
static time_t    gZLS_GlobalSimRealTimeDuration = -1;
//
//
// compute cycles between printing cycle progress reports > 16, even number
//
Counter kZLS_CycleProgressPrint  = ZLS_PROGRESS_INTERVAL; 
//
// compute cycles between cpu/cell/cycles performance checks 
// must be an integer multiple of kZLS_CycleProgressPrint
//
Counter kZLS_CPUCyclePerformance = ZLS_PERFORM_INTERVAL; 

#ifndef NCPUs
#define NCPUs 1
#endif

int zlst_ClearCPUTimes( CPUCoordPtr ct ){
  int err = noErr;
  
  ct->cpuusr[0]      = 0;
  ct->cpusys[0]      = 0;
  ct->cputot[0]      = 0;
  ct->realtime[0]    = 0;
  
  ct->modeltime[0]   = 0.0;
  ct->modelcycles[0] = 0;
  
  ct->cpuusr[1]      = 0;
  ct->cpusys[1]      = 0;
  ct->cputot[1]      = 0;
  ct->realtime[1]    = 0;
  
  ct->modeltime[1]   = 0.0;
  ct->modelcycles[1] = 0;
  
  return(err);
}


int zlst_InitGlobalTiming      ( CPUCoordPtr gctp ){
  int err = noErr;
  struct tms  cb;
  
  clock_t c =  times(&cb);
  
  gctp->cpuusr[0]      = cb.tms_utime;
  gctp->cpusys[0]      = cb.tms_stime;
  gctp->cputot[0]      = gctp->cpuusr[0]+gctp->cpusys[0];
  gctp->realtime[0]    = c;
  
  gctp->modeltime[0]   = zlst_GlobalSeconds;
  gctp->modelcycles[0] = zlst_GlobalTimingCycles;
  
  return(err);
} 

int zlst_StartCPUTimes( CPUCoordPtr ct ){
  int err = noErr;
  struct tms  cb;
  
  clock_t c =  times(&cb);
  
  ct->cpuusr[0]      = cb.tms_utime;
  ct->cpusys[0]      = cb.tms_stime;
  ct->cputot[0]      = ct->cpuusr[0]+ct->cpusys[0];
  ct->realtime[0]    = c;
  
  ct->modeltime[0]   = zlst_GlobalSeconds;
  ct->modelcycles[0] = zlst_GlobalTimingCycles;
  
  return(err);
}


int zlst_SaveCPUTimes( CPUCoordPtr ct_old, CPUCoordPtr ct){
  int err = noErr;
  struct tms  cb;
  
  ct_old->cpuusr  [0]    = ct->cpuusr  [0];
  ct_old->cpusys  [0]    = ct->cpusys  [0];
  ct_old->cputot  [0]    = ct->cputot  [0];
  ct_old->realtime[0]    = ct->realtime[0];
  
  ct_old->modeltime  [0] = ct->modeltime  [0];
  ct_old->modelcycles[0] = ct->modelcycles[0];
  
  clock_t c =  times(&cb);
  
  ct_old->cpuusr  [1]    = cb.tms_utime;
  ct_old->cpusys  [1]    = cb.tms_stime;
  ct_old->cputot  [1]    = ct->cpuusr[1]+ct->cpusys[1];
  ct_old->realtime[1]    = c;
  
  ct_old->modeltime  [1] = zlst_GlobalSeconds;
  ct_old->modelcycles[1] = zlst_GlobalTimingCycles;
  
  
  return(err);
}

int zlst_ResumeCPUTimes( CPUCoordPtr ct, CPUCoordPtr ct_old ){
  int err = noErr;
  struct tms  cb;
  clock_t c =  times(&cb);
  
  Counter delta = ct_old->cpuusr[1] - ct_old->cpuusr[0];
  ct->cpuusr[0]      = cb.tms_utime - delta;
  
  delta = ct_old->cpusys[1] - ct_old->cpusys[0];
  ct->cpusys[0]      = cb.tms_stime - delta;
  
  delta              = ct_old->cputot[1] - ct_old->cputot[0];
  Counter ttot       = cb.tms_stime + cb.tms_utime;
  ct->cputot[0]      = ttot - delta;
  
  ct->realtime[0]    = c - (ct_old->realtime[1]-ct_old->realtime[0]);
  
  delta              = ct_old->modeltime[1] - ct_old->modeltime[0];
  ct->modeltime  [0] = zlst_GlobalSeconds - delta;
  
  delta              = ct_old->modelcycles[1] - ct_old->modelcycles[0];
  ct->modelcycles[0] = zlst_GlobalTimingCycles - delta ;
  
  return(err);
}


int zlst_EndCPUTimes( CPUCoordPtr ct ){
  int err = noErr;
  struct tms  cb;
  
  clock_t c       =  times(&cb);
  
  ct->cpuusr[1]   = cb.tms_utime;
  ct->cpusys[1]   = cb.tms_stime;
  ct->cputot[1]   = ct->cpuusr[1]+ct->cpusys[1];
  ct->realtime[1] = c;
  
  ct->modeltime[1]   = zlst_GlobalSeconds;
  ct->modelcycles[1] = zlst_GlobalTimingCycles;
  
  return(err);
}


int zlst_ElapsedCPUTimes( CPUCoordPtr ct, char * title ){
  
  int err = noErr;
#ifdef OS_GENERIC
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_GENERICLINUX
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_CYGWIN
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#ifdef OS_MACOSXG4
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXG5
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXCORE
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_SOLARISSPARC
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_SOLARISAMD64
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
  
  double usr, sys, tot, rea, f0, f1, f2;
  double invNcpus = 1.0/(NCPUs); // NCPUs is set in zls_OSdefs.h
  
  usr = ((double)(ct->cpuusr[1]-ct->cpuusr[0]))*spc*invNcpus; 
  sys = ((double)(ct->cpusys[1]-ct->cpusys[0]))*spc*invNcpus;
  tot = ((double)(ct->cputot[1]-ct->cputot[0]))*spc*invNcpus;
  rea = ((double)(ct->realtime[1]-ct->realtime[0]))*spc;
  f0 = (usr/rea); f1 =  (sys/rea);  f2 =  (tot/rea); 
  
  zlsio_Printf( "\n Timing for %s.\n", title);
  zlsio_Printf( "%s\n", DIVIDERLINE );
  zlsio_Printf( " ZLS /CPU: %8.4fu %8.4fs %8.4ft %8.4fw\n", usr, sys, tot, rea);
  zlsio_Printf( "    Loads: %6.2f%s %6.2f%s %6.2f%s %6.2f%s \n", f0,  "u/w", f1, "s/w", sys/usr, "s/u", f2, "t/w");
  zlsio_Printf( "%s\n", DIVIDERLINE );
  
  return(err);
}

int zlst_ElapsedCellRates( CPUCoordPtr ct, double nCells, char * title ){
#pragma unused( title )
  
  int err = noErr;
#ifdef OS_GENERIC
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_GENERICLINUX
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_CYGWIN
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_MACOSXG4
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXG5
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_MACOSXCORE
#ifdef OS_MACOSX105TIMING  
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#else
  const double spc = 1.0/((double)CLOCKS_PER_SEC);
#endif
#endif  
#ifdef OS_SOLARISSPARC
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
#ifdef OS_SOLARISAMD64
  const double spc = 1.0e4/((double)CLOCKS_PER_SEC);
#endif  
  double usr, tot, rea, model, f0,  f2, f3;
  Counter cyc;
  
  usr = ((double)(ct->cpuusr[1]-ct->cpuusr[0]))*spc;
  tot = ((double)(ct->cputot[1]-ct->cputot[0]))*spc;
  rea = ((double)(ct->realtime[1]-ct->realtime[0]))*spc; // real time is real, so don't scale by NCPUs
  model = ((ct->modeltime[1]-ct->modeltime[0])); // model time is real, so don't scale by NCPUs
  cyc  = ((ct->modelcycles[1]-ct->modelcycles[0])); // model in cycles is real, so don't scale by NCPUs
  
#ifdef PTHREADS    
  zlsio_LogPrintf( "\n %s\n", zlst_TheCurrentRealTime());
  zlsio_Printf( "\n Cycle/Cell Rates, CPUs: %d, Threads %d,  Eff.: %3.1f (CPUs).\n", NCPUs, ZLS_MAXPTHREADS, (usr/rea));
  zlsio_Printf( "%s\n", DIVIDERLINE );
  f0 = (usr/(NCPUs*cyc));  f2 =  (tot/(NCPUs*cyc));
  zlsio_Printf( " s/Cycle     : %10.4fu %10.4ft \n", f0,  f2 );
  f0 = (nCells*cyc/usr);  f2 =  (nCells*cyc/tot);  
  zlsio_Printf( " Cells/s/CPU : %10.1fu %10.1ft %10.1fAll \n\n", f0,  f2,  f2*NCPUs );
  f3 =  (rea /cyc); 
  zlsio_Printf( " Real Time Total s/Cycle : %10.4fw\n", f3);
  f3 =  (nCells*cyc/rea); 
  zlsio_Printf( " Real Time Total Cells/s : %10.1fw \n", f3 );
  f3 =  ( model /rea ); 
  zlsio_Printf( " Reality Overdrive       : %10.4g  (model/real time) \n",  f3 );
  zlsio_Printf( "%s\n\n", DIVIDERLINE );
  
#else     
  
#ifdef MPI
  if (gMPIrank == 0) {
#endif MPI

    zlsio_LogPrintf( "\n %s\n", zlst_TheCurrentRealTime());
#ifdef MPI
    zlsio_Printf( "\n Cycle/Cell Rates, MPI nodes: %d,  Eff.: %3.1f (CPUs).\n", gMPIsize, (usr/rea));
#else
    zlsio_Printf( "\n Cycle/Cell Rates, CPUS: %d,  Eff.: %3.1f (CPUs).\n", NCPUs, (usr/rea));
#endif MPI
    zlsio_Printf( "%s\n", DIVIDERLINE );
    f0 = (usr/cyc);  f2 =  (tot/cyc);
    zlsio_Printf( " s/Cycle : %10.4fu %10.4ft \n"  , f0,  f2 );
    f0 = (nCells*cyc/usr);  f2 =  (nCells*cyc/tot);  
    zlsio_Printf( " Cells/s : %10.1fu %10.1ft \n\n", f0,  f2 );
    f3 =  (rea /cyc); 
    zlsio_Printf( " Real Time Total s/Cycle : %10.4fw \n", f3);
    f3 =  (nCells*cyc/rea); 
    zlsio_Printf( " Real Time Total Cells/s : %10.1fw  \n",  f3 );
    f3 =  ( model /rea ); 
    zlsio_Printf( " Reality Overdrive       : %10.4g (model/real time) \n",  f3 );
    zlsio_Printf( "%s\n\n", DIVIDERLINE );
    
#ifdef MPI
  }
#endif MPI
  
#endif !PTHREADS    
  
  return(err);
}


//====================================================================================
//
// this is the common public routine for checking progress and performance of the sim.
//

int zlst_ProgressCheck( CPUCoordPtr ctp, double ncells ) {
  
  int err = noErr;
#ifdef MPI
  if (gMPIrank == 0) {
#endif MPI
    if ( (zlst_GlobalCycles==0) || (zlst_GlobalCycles % kZLS_CycleProgressPrint) == 0) {
      
      if ( kZLS_CycleProgressPrint > 0) {
        if ( f_t0 != 1.0 ) {
          zlsio_Printf( " Cycle: %10d\n", zlst_GlobalCycles);
          zlsio_Printf( " Time : %#10.5g = %#10.5g %s = %#10.5g %s\n", 
                       zlst_GlobalTime    , zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr");
          zlsio_Printf( " Delta: %#10.5g = %#10.5g %s = %#10.5g %s\n",
                       zlst_GlobalTimeStep, zlst_GlobalSecondsStep, s_Time, zlst_GlobalYearsStep, "yr");
        } else {
          zlsio_Printf( " Cycle: %10d   Time: %#10.5g   Delta: %#10.5g\n",
                       zlst_GlobalCycles, zlst_GlobalTime, zlst_GlobalTimeStep);      
        }
      }
      
      if ( kZLS_CPUCyclePerformance > 0) {
        if (((zlst_GlobalCycles % kZLS_CPUCyclePerformance) == 0)&&(zlst_GlobalCycles > 0)) {
          zlst_EndCPUTimes( ctp );
          zlst_ElapsedCellRates( ctp, ncells , " " );
          zlst_StartCPUTimes( ctp );
        }
      }
      
    }
#ifdef MPI
  }
#endif MPI
  
  return(err);
}


char gZLS_RealTimeString[80];

char * zlst_TheCurrentRealTime(void){
  time_t  startTime = time(NULL);
  sprintf(gZLS_RealTimeString, "%s", ctime(&startTime));
  return(gZLS_RealTimeString);
}

//====================================================================================
// these are the only routines to advance a clock time, *do not* bypass and access a 
// zone clock directly or you get what you deserve!
//

int zlst_AdvanceCounters( TimeCoords * zoneclock ){
  int err = noErr;
  
  zoneclock->cycles++;  // computation cycles in Zone
  
  zoneclock->timingCycles++; // computation cycles in Zone, 
  // start from 0 always even after restart
  // *only* used for performance timing.
  
  zoneclock->nCyclesFrame++; // cycles for frame output
  zoneclock->nCyclesUser++; // cycles for user output
  zoneclock->nCyclesData++; // cycles for 16 bit signed data output
  zoneclock->nCyclesSave++; // cycles for restart output
  
  return(err);
}


int zlst_AdvanceClock( TimeCoords * zoneclock){
  int err = noErr;
  
  Real dt = zoneclock->dt;
  
  zoneclock->t      += dt; // overall sim time
  zoneclock->tFrame += dt; // time for image frame output
  zoneclock->tUser  += dt; // time for user output
  zoneclock->tData  += dt; // time for data output
  zoneclock->tSave  += dt; // time for restart output
  
  zlst_AdvanceCounters(zoneclock);
  
  return(err);
}

//
//====================================================================================
// These are the routines to determine simulation step sizes and 
// set the actual zone clocks prior to a simulation step
//====================================================================================
//====================================================================================
//
// Fundamental Courant-Friedrichs-Level timestep number (0 - 1)
// Not used directly, but used to set level cfl values in zls_conf.c
//
Real  kZLS_CFL     = 0.80; 
Real  kZLS_FracCFL = 0.80;  // fraction of cfl for first step
//
//====================================================================================
//====================================================================================
//
const Real kZLS_RateLimiter = 1.15;
Real kZLS_CoolTimeFraction  = 1.25;  // not const as it can be set in physics.conf
//
//====================================================================================

int zlst_CalcZoneNewTimeScale   ( ZonePtr z);
int zlst_CalcZoneNewTimeScale   ( ZonePtr z){
  
  // estimate minimum timescale for a zone, using primitive variables
  
  int err = noErr;
  
  if (z != NULL){
    
    WHD        w;
    
    Real            ztime;
    Real              cfl;
    Real    tau_min = kZLS_Huge;
    Real    min_dt  =  kZLS_Huge; 
    
    Real         cs = 0.0;
    
    PhysicsPtr   phys = NULL;
    Real        gamma = kZLS_eos_gamma;
    Real         mu   = k_mu;
#ifdef COOLING        
    CoolingPtr   cp   = NULL;
    Real   tau_cool = kZLS_Huge;
#endif COOLING        
    
    Integer     inc = 2;
    
#ifdef SYM1D    
    Real             dx;
    Real   tau_vx = 0.0;
    Integer           i;
    Integer          is;
    Integer          ie;
    ACoordPtr        xx;
#endif SYM2D    
    
#ifdef SYM2D    
    Real     dx, dy, ds;
    Real   tau_vx = 0.0;
    Real   tau_vy = 0.0; 
    Real        g = 1.0;
    Integer        i, j;
    Integer      is, js;
    Integer      ie, je;
    ACoordPtr       xx,yx;
#endif SYM2D    
    
#ifdef SYM3D    
    Real     dx, dy, dz, ds;
    Real       tau_vx = 0.0;
    Real       tau_vy = 0.0; 
    Real       tau_vz = 0.0; 
    Real        g = 1.0;
    Integer         i, j, k;
    Integer      is, js, ks;
    Integer      ie, je, ke;
    ACoordPtr       xx, yx, zx;
#endif SYM3D  
    
#ifdef COURANT_VC_SEPARATE
    Real  tau_cs = 0.0;
#endif COURANT_VC_SEPARATE
    
    tau_min = kZLS_Huge;
    cfl     = z->zoneCFL;
    ztime   = z->zoneClock.t;
    
    w       = z->zoneW;
    
    GridCoords gc = z->zoneGrid;
        
#ifdef YSO_TIMING
    Real v0        = gLevels[0].levelProblem->gJetVelm;
    Real tau_v_jet = kZLS_Huge;
#endif YSO_TIMING    
    
#ifdef ISOTHERMAL_TIMING
    phys            = gLevels[0].levelPhysics;  
    PotentialPtr gp   = &(phys->gravPotential);
    Integer  gType  =  gp->gravTypeFlags;
    Real vGravIso   = 0.0;
    Real tau_v_iso  = kZLS_Huge;
    
    if (( gType & kPhysIsoGravityTypes ) != 0 ){
      vGravIso = gp->s_iso*sqrt(2.0); // circular velocity for the potential
    }
#endif ISOTHERMAL_TIMING    
    
#ifdef RADIOJET_TIMING
    //  Real v0       =  (ztime < 1.0)?gLevels[0].levelProblem->gJetVelm:kZLS_Small;
    Real v0       =  gLevels[0].levelProblem->gJetVelm;
    Real tau_v_jet = kZLS_Huge;
#endif RADIOJET_TIMING    
    
#ifdef STELLAR_WIND_TIMING
    
    Integer nS     = gLevels[0].levelProblem->gNstars;
    
    Real v0        = kZLS_Tiny;
    Real v1        = kZLS_Tiny;
    
    
    for ( i = 0; i < nS; i++){
      v1 = gLevels[0].levelProblem->gVWind[i];
      v0 = v1>v0?v1:v0;
    }
    
    
    Real tau_v_wind = kZLS_Huge;
#endif STELLAR_WIND_TIMING  
    
    phys  = gLevels[ z->zoneLevel].levelPhysics;
    if (phys != NULL) {
      gamma  = phys->gamma;  // global plasma adiabtic index
      mu     = phys->mu;     // global plasma mu
#ifdef COOLING
      cp    = &(phys->coolParameters);
#endif COOLING  
    }
    
#ifdef SYM1D    
    // all 1D axes are spatial, so no conversion from radians to spatial units needed
    
    if (err == noErr) {   
      
      is =  gc.ns[kXDim];
      ie =  gc.ne[kXDim];
      xx =  gc.x;
      
#ifdef COOLING        
      tau_cool = kZLS_CoolTimeFraction*z->zoneMinCoolingTime;
#endif COOLING  
      
#ifdef YSO_TIMING
      dx        = xx[0].dx;
      tau_v_jet = dx/v0;
#endif YSO_TIMING  
      
#ifdef RADIOJET_TIMING
      dx        = xx[0].dx;
      tau_v_jet = dx/v0;
#endif RADIOJET_TIMING  
      
#ifdef STELLAR_WIND_TIMING
      dx         = xx[0].dx;
      tau_v_wind = dx/v0;
#endif STELLAR_WIND_TIMING  
      
#ifdef ISOTHERMAL_TIMING
      dx        = xx[0].dx;
      tau_v_iso = dx/vGravIso;
#endif ISOTHERMAL_TIMING    
      
      
      for(i = is; i<ie; i++){
        dx  = xx[i].dx;
        
#ifdef OS_MACOSXG4
        cs  = zls_sqrtfaster(gamma*w.P[i]/w.d[i]); // approximate sound speed OK here
#else
        cs  = zls_sqrt(gamma*w.P[i]/w.d[i]); 
#endif OS_MACOSXG4

#ifdef COURANT_VC_SEPARATE
        tau_cs = dx/(cs);
        tau_min = zls_min( tau_min, tau_cs );
        tau_vx  = dx/(zls_abs(w.vx[i]));
#endif COURANT_VC_SEPARATE
        
#ifdef COURANT_VC_COMBINED
        tau_vx = dx/(zls_abs(w.vx[i])+cs);
#endif COURANT_VC_COMBINED
        
        tau_min = zls_min( tau_min, tau_vx );
        
      } //i
      
#ifdef RADIOJET_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif RADIOJET_TIMING  
#ifdef YSO_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif YSO_TIMING  
#ifdef STELLAR_WIND_TIMING
      tau_min  = zls_min( tau_v_wind, tau_min );
#endif STELLAR_WIND_TIMING  
      
#ifdef ISOTHERMAL_TIMING
      tau_min  = zls_min( tau_v_iso, tau_min );
#endif ISOTHERMAL_TIMING 
      
#ifdef COOLING      
      if (kZLS_CoolTimeFraction <= 20.0 ) 
        tau_min  = zls_min( tau_cool, tau_min );
#endif COOLING        
#ifdef CSTAR_SHELL
      tau_min  = zls_min( 0.1*gLevels[0].levelProblem->gWburst, tau_min );
#endif CSTAR_SHELL   
      
      min_dt = cfl * tau_min;
      
    }
#endif SYM1D    
    
#ifdef SYM2D    
    
    if (err == noErr) {   
      
      xx  =  gc.x;
      yx  =  gc.y;
      
      is =  gc.ns[kXDim]; //
      ie =  gc.ne[kXDim];
      
      js =  gc.ns[kYDim]; //
      je =  gc.ne[kYDim];
      
#ifdef COOLING        
      tau_cool = kZLS_CoolTimeFraction*z->zoneMinCoolingTime;
#endif COOLING  
      
#ifdef RADIOJET_TIMING
      dx  = xx[0].dx;
      tau_v_jet = dx/v0; // OK as first axis is always spatial
#endif RADIOJET_TIMING  
      
#ifdef YSO_TIMING
      dx  = xx[0].dx;
      tau_v_jet = dx/v0; // OK as first axis is always spatial
#endif YSO_TIMING  
      
#ifdef STELLAR_WIND_TIMING
      dx  = xx[0].dx;
      tau_v_wind = dx/v0;
#endif STELLAR_WIND_TIMING  
      
#ifdef ISOTHERMAL_TIMING
      dx  = xx[0].dx;
      tau_v_iso = dx/vGravIso;
#endif ISOTHERMAL_TIMING    
      
      g  =  1.0; // default geometry projection
      
      for(i = is; i < ie; i+=inc){
        dx  = xx[i].dx;
#ifdef SPHERICAL
        g =   xx[i].c; // radius to scale dPhi to spatial units
#endif SPHERICAL
#ifdef CIRCULAR
        g =   xx[i].c; // radius to scale dPhi to spatial units
#endif CIRCULAR
        
        for(j = js; j < je; j+=inc){
#ifdef POLARSPHERICAL
          g = xx[i].c * yx[j].sc; // radius sin(theta) to scale dphi to spatial units
#endif POLARSPHERICAL
          dy = g*yx[j].dx;
          
#ifdef OS_MACOSXG4
          cs    = zls_sqrtfaster(gamma*w.P[j][i]/(w.d[j][i]+kZLS_Epsilon)); // approx OK for timescales
#else
          cs    = zls_sqrt(gamma*w.P[j][i]/(w.d[j][i]+kZLS_Epsilon));
#endif OS_MACOSXG4
          
          ds     =  zls_min( dx, dy);
#ifdef COURANT_VC_SEPARATE
          tau_cs = ds/(cs);
          tau_min = zls_min( tau_min, tau_cs );
          tau_vx = dx/(zls_abs(w.vx[j][i])+kZLS_Epsilon);
          tau_vy = dy/(zls_abs(w.vy[j][i])+kZLS_Epsilon); 
#endif COURANT_VC_SEPARATE
          
#ifdef COURANT_VC_COMBINED
          tau_vx = dx/(zls_abs(w.vx[j][i])+cs);
          tau_vy = dy/(zls_abs(w.vy[j][i])+cs);
#endif COURANT_VC_COMBINED
          
          tau_min = zls_min( tau_min, tau_vx );
          tau_min = zls_min( tau_min, tau_vy );
          
        } //i
      } //j
      
#ifdef RADIOJET_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif RADIOJET_TIMING  
#ifdef YSO_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif YSO_TIMING  
      
#ifdef STELLAR_WIND_TIMING
      tau_min  = zls_min( tau_v_wind, tau_min );
#endif STELLAR_WIND_TIMING  
#ifdef ISOTHERMAL_TIMING
      tau_min  = zls_min( tau_v_iso, tau_min );
#endif ISOTHERMAL_TIMING 
      
#ifdef COOLING         
      if (kZLS_CoolTimeFraction <= 20.0 ) 
        tau_min  = zls_min( tau_cool, tau_min );
#endif COOLING        
      
#ifdef CSTAR_SHELL
      tau_min  = zls_min( 0.1*gLevels[0].levelProblem->gWburst, tau_min );
#endif CSTAR_SHELL   
      
      min_dt = cfl * tau_min;
      
    }
#endif SYM2D    
    
    
#ifdef SYM3D    
    
    if (err == noErr) {   
      
      xx  =  gc.x;
      yx  =  gc.y;
      zx  =  gc.z;
      
      g  =  1.0; // default geometry projection
      
      is =  gc.ns[kXDim]; //
      ie =  gc.ne[kXDim];
      
      js =  gc.ns[kYDim]; //
      je =  gc.ne[kYDim];
      
      ks =  gc.ns[kZDim]; //
      ke =  gc.ne[kZDim];
      
#ifdef COOLING        
      tau_cool = kZLS_CoolTimeFraction*z->zoneMinCoolingTime;
#endif COOLING
      
#ifdef RADIOJET_TIMING
      dx  = xx[0].dx;
      tau_v_jet = dx/v0; // OK as first axis is always spatial
#endif RADIOJET_TIMING  
      
#ifdef YSO_TIMING
      dx  = xx[0].dx;
      tau_v_jet = dx/v0; // OK as first axis is always spatial
#endif YSO_TIMING  
      
#ifdef STELLAR_WIND_TIMING
      dx  = xx[0].dx;
      tau_v_wind = dx/v0;
#endif STELLAR_WIND_TIMING  
      
#ifdef ISOTHERMAL_TIMING
      dx  = xx[0].dx;
      tau_v_iso = dx/vGravIso;
#endif ISOTHERMAL_TIMING    
      
      for(i = is; i<ie; i+=inc){
        dx  = xx[i].dx;
        for(j = js; j<je; j+=inc){
#ifdef POLARSPHERICAL
          g = xx[i].c; // radius to scale dtheta to spatial units r theta phi => r sin theta dr dtheta dphi
          dy = g*yx[j].dx;
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
          dy  = yx[j].dx;
#endif CYLINDRICAL
#ifdef CARTESIAN
          dy  = yx[j].dx;
#endif CARTESIAN
          for(k = ks; k<ke; k+=inc){
#ifdef CYLINDRICAL
            g =    yx[j].c; // radius to scale dPhi to spatial units
            dz = g*zx[k].dx;
#endif CYLINDRICAL
#ifdef SPHERICAL
            g =    xx[i].c*zx[k].cc; // cos theta radius to scale dPhi to spatial units r phi theta => r cos theta dr dphi dtheta
            dy = g*yx[j].dx;
            g =    xx[i].c; // just radius to scale dtheta to spatial units
            dz = g*zx[k].dx;
#endif SPHERICAL
#ifdef POLARSPHERICAL
            g =    xx[i].c* zx[k].sc; // sin theta radius to scale dPhi to spatial units r phi theta => r sin theta dr dtheta dphi
            dz = g*zx[k].dx;
#endif POLARSPHERICAL
#ifdef CARTESIAN
            dz  = zx[k].dx;
#endif CARTESIAN
            
#ifdef OS_MACOSXG4
            cs  = zls_sqrtfaster(gamma*w.P[k][j][i]/w.d[k][j][i]); // approx OK
#else
            cs  = zls_sqrt(gamma*w.P[k][j][i]/w.d[k][j][i]);
#endif
            ds     =  zls_min3( dx, dy, dz);
#ifdef COURANT_VC_SEPARATE
            tau_cs = ds/(cs);
            tau_min = zls_min( tau_min, tau_cs );
            tau_vx = dx/(zls_abs(w.vx[k][j][i])+kZLS_Epsilon);
            tau_vy = dy/(zls_abs(w.vy[k][j][i])+kZLS_Epsilon); 
            tau_vz = dz/(zls_abs(w.vz[k][j][i])+kZLS_Epsilon); 
#endif COURANT_VC_SEPARATE
            
#ifdef COURANT_VC_COMBINED
            tau_vx = dx/(zls_abs(w.vx[k][j][i])+cs);
            tau_vy = dy/(zls_abs(w.vy[k][j][i])+cs);
            tau_vz = dz/(zls_abs(w.vz[k][j][i])+cs);
#endif COURANT_VC_COMBINED
            
            tau_min = zls_min( tau_min, tau_vx );
            tau_min = zls_min( tau_min, tau_vy );
            tau_min = zls_min( tau_min, tau_vz );
            
          } //i
        } //j
      } //k
      
#ifdef RADIOJET_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif RADIOJET_TIMING  
      
#ifdef YSO_TIMING
      tau_min  = zls_min( tau_v_jet, tau_min );
#endif YSO_TIMING  
      
#ifdef STELLAR_WIND_TIMING
      tau_min  = zls_min( tau_v_wind, tau_min );
#endif STELLAR_WIND_TIMING  
      
#ifdef ISOTHERMAL_TIMING
      tau_min  = zls_min( tau_v_iso, tau_min );
#endif ISOTHERMAL_TIMING 
      
#ifdef COOLING         
      if (kZLS_CoolTimeFraction <= 20.0 ) 
        tau_min  = zls_min( tau_cool, tau_min );
#endif COOLING    
      
#ifdef CSTAR_SHELL
      tau_min  = zls_min( 0.1*gLevels[0].levelProblem->gWburst, tau_min );
#endif CSTAR_SHELL   
      
      min_dt = cfl * tau_min;
      
    }
#endif SYM3D
    
#ifdef COOLING        
    z->zoneMinCoolingTime = kZLS_Huge; 
#endif COOLING 

    z->zoneTimeScaleStatus = kZLS_Valid; 
//    z->zoneOldDt           = z->zoneNewDt;
    z->zoneNewDt           = min_dt;
    
    
  } //z
  
  return(err);
}

//====================================================================================

int zlst_CalcZoneTimeStep   ( ZonePtr z ){
  
  int err = noErr;
  
  if (z != NULL){
    if (z->zoneType == kAlgorithmZoneType) {
    
    Real   ztime = z->zoneClock.t;
    Real  prevdt = z->zoneOldDt;

    if ( z->zoneTimeScaleStatus != kZLS_Valid){
        err = zlst_CalcZoneNewTimeScale ( z);
    }
    
    Real   zdt   = z->zoneNewDt;
    
    Real dta   = (prevdt)*kZLS_RateLimiter;
    Real dtmin = (prevdt)*0.001;
    
    Real  min_dt  = zls_max( (zls_min( dta, zdt )), (dtmin));
        
    if (ztime/min_dt > 1.e8) {
      
      zlsio_ErrPrintf(  "\n %s\n", zlst_TheCurrentRealTime());
      zlsio_ErrPrintf(  " WARNING: CalcZoneTimeStep, Zone %d, Level %d \n", z->zoneID, z->zoneLevel);
      zlsio_ErrPrintf(  " Timestep too small: prev min lim new = %g %g %g %g\n",prevdt, min_dt, dta, zdt);
      zlsio_ErrPrintf(  " Time: %g\n", ztime);
      
      min_dt  = prevdt;
      err     = timErr;
    }
    
    z->zoneNewDt = min_dt;
    
    }
    
  } //z
  
  return(err);
}


int zlst_CalcLevelTimeStep  (Real * dt, Integer level){
  
  //
  // collect a level wide constrained time step from all level zones
  //
  
  int err = noErr;
  
  Integer         zone;
  Integer       nZones;
  ZonePtr            z;
  
  Real leveldt = kZLS_Huge; // set large to capture overall min
  Real zonedt  = *dt;      // set to prev value to allow accellerator limiter to work
  nZones = gLevels[level].nZones;
  
  if (nZones > 0){
  
    for( zone = 0; zone < nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
    if (z->zoneType == kAlgorithmZoneType) {
#ifdef MPI  
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          z->zoneOldDt = *dt;
          err     = zlst_CalcZoneTimeStep(z);
          zonedt  = z->zoneNewDt;
          leveldt = leveldt > zonedt ? zonedt : leveldt;
#ifdef MPI      
        }
#endif MPI
        }
      } else { err = zonErr; }
    } // zone loop
    
  } else { err = zonErr; }
  
  
#ifdef MPI  
  Real globaldt;    
  MPI_Allreduce( &leveldt, &globaldt, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  //  if ( leveldt != globaldt) zlsio_Printf(" Rank: %d  globaldt: %g\n", gMPIrank, leveldt, globaldt);
  *dt = globaldt;
#else 
  *dt = leveldt;
#endif !MPI
  
  return(err);
  
}

//
// Notes: In any given step, the timestep dt cannot increase by more than
// the kZLS_RateLimiter ratio, to stop runaway timsteps and overshooting.
// By default this is 1.15.  So, timesteps can start small and grow gradually
// towards a maximum safe stride, but can be reduces as quickly as needed if 
// they physical conditions change, ie the gas gets hot.
// In that case the steps may be small for a while for physical reasons.
// 
// However, sometimes the timestep is shortened a lot to allow the IO to land
// on a definite time, shortening (or even slightly lengthening) the physical 
// timestep.  
//
// If the last global time step was foreshortened to allow an IO operation
// land on an exact interval, we want to restore the pre-truncated timescale
// in the next step, to save time accelerating up to speed after a very short
// IO step.  ie if the step was shortened to only 0.2 of the optimum step
// then it could take up to 11 or more steps to regain the full timestep stride
// at the maximum 1.15 acceleration limit.  In this case the previous timestep
// is restored as a starting point for a full timestep calculation, which is OK 
// because the previous timestep was physical (ie courant/cooling limited)
// and should still be close to useable if the truncated step was shorter, meaning
// not a lot should have changed since the previous step.
//
// This uses two globals, that may need to be managed better if MPI is implemented.
// Only the global (level 0) timestep is considered, all lower levels are
// derived as normal.  With PTHREADS this is used by the global thread, not the sweepers,
// and so all it well for pre-emptive parallel operation.
//

enum{
  kZLS_StepNormal = 0,
  kZLS_PrevStepWasShortenedForIO = 1
};

static Real timingPrevTimeStep     = 0.0; // inited to 0.0 if and set to 0.0 if not in use.
static Integer timingShortenedStep = kZLS_StepNormal; // start normally

int zlst_CalcAllTimeSteps  (Real * dt){
  
  //
  // collect a level wide constrained time step from all level zones
  //
  
  int err = noErr;
  
  Integer       level;
  Real        refined;
  Real    minTimeStep;
  Real       timeStep;
  Real            dt0;
  
  
  if (timingShortenedStep == kZLS_PrevStepWasShortenedForIO ){
    *dt = timingPrevTimeStep;
  }
  
  timeStep = *dt;
  dt0 = *dt;
  minTimeStep = kZLS_RateLimiter*dt0;
  
  for (level = 0; level < gNLevels; level ++){
    refined = (Real)gLevels[level].levelTotalRefinement;
    err = zlst_CalcLevelTimeStep  (&timeStep, level);
    timeStep *= refined;
    minTimeStep = minTimeStep > timeStep? timeStep: minTimeStep;
  } // level
  
  *dt = minTimeStep;
  
  timingShortenedStep    = kZLS_StepNormal;
  timingPrevTimeStep     =     minTimeStep;
  
  return(err);
}


int zlst_SynchLevelZoneClocks  ( Integer level );
int zlst_SynchLevelZoneClocks  ( Integer level ){
  //
  // this is a simpler synch time and dt  in a level and all zones in the level
  // Do not call directly, used internally by the calls below.
  //
  int err = noErr;
  
  Integer   zone;
  ZonePtr    z;
  if (gLevels[level].nZones>0){
    for( zone = 0; zone < gLevels[level].nZones; zone++){
      z      = gLevels[level].levelZones[zone];
      if (z != NULL){
        z->zoneClock.dt = gLevels[level].levelClock.dt;
        z->zoneClock.t  = gLevels[level].levelClock.t;
      } else { err = zonErr; }
    } // zone
  } else { err = zonErr; }
  
  return(err);
}

int zlst_GetLevelClocksTimeStep  ( Real *dt , Integer level);
int zlst_GetLevelClocksTimeStep  ( Real *dt , Integer level){ // should be a macro, see zls_timing.h
  //
  int err = noErr;
  *dt = gLevels[level].levelClock.dt ;
  return(err);
}

int zlst_SetAllClocksUniformTimeStep  (Real * dt){
  //
  // Sets *all* level clocks *and* each zone clock step sizes to the finest step in the heirachy
  // input dt is the global courant limited step requested for the top level as used in 
  // zlst_SetAllClocksTimeStep above.
  //
  int err = noErr;
  
  Integer          level;
  Real           refined;
  Real       absoluteTime;  
  // the time in level 0 master clock, set to prevent
  // roundoff errors in all other clocks
  Real            ldt, rdt;
  
  rdt = *dt;
  refined = (Real) gLevels[gNLevels-1].levelTotalRefinement;
  ldt     = rdt/refined;
  absoluteTime = zlst_GlobalTime;
  
  for (level = (gNLevels-1); level >= 0; level--) {
    
    if ((gLevels[level].levelClock.tFrame+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.frameTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.frameTime - gLevels[level].levelClock.tFrame; 
    }
    
    if ((gLevels[level].levelClock.tUser+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.userTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.userTime - gLevels[level].levelClock.tUser; 
    } 
    
    if ((gLevels[level].levelClock.tData+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.dataTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.dataTime - gLevels[level].levelClock.tData; 
    } 
    
    if ((gLevels[level].levelClock.tSave+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.saveTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.saveTime - gLevels[level].levelClock.tSave; 
    } 
    
    if ((gLevels[level].levelClock.t+(ldt*kZLS_TimeEventSlop))>gLevels[level].levelClock.endTime){
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.endTime -gLevels[level].levelClock.t; 
    }
    
    gLevels[level].levelClock.dt = ldt;
    gLevels[level].levelClock.t  = absoluteTime;
    
    err = zlst_SynchLevelZoneClocks( level );
    
  } // level
  rdt     = ldt*refined;
  
  *dt = rdt;
  
  return(err);
}


int zlst_SetAllClocksTimeStep  (Real * dt){
  // this is critical, it sets all level clock *and* each zone clock step size
  int err = noErr;
#ifdef UNIFORM_SWEEPS
  // special case for multi-level uniform sweeps.
  err = zlst_SetAllClocksUniformTimeStep  (dt);
#else
  Integer           level;
  Real           refined;
  Real       absoluteTime;  
  // the time in level 0 master clock, set to prevent
  // roundoff errors in all other clocks
  Real            ldt, rdt;
  
  rdt = *dt;
  absoluteTime = zlst_GlobalTime;
  
  for (level = 0; level < gNLevels; level++){
    
    refined = (Real) gLevels[level].levelTotalRefinement;
    
    ldt     = rdt/refined;
    
    if ((gLevels[level].levelClock.tFrame+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.frameTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.frameTime - gLevels[level].levelClock.tFrame; 
    }
    
    if ((gLevels[level].levelClock.tUser+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.userTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.userTime - gLevels[level].levelClock.tUser; 
    } 
    
    if ((gLevels[level].levelClock.tData+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.dataTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.dataTime - gLevels[level].levelClock.tData; 
    } 
    
    if ((gLevels[level].levelClock.tSave+(ldt*kZLS_TimeEventSlop)) > gLevels[level].levelClock.saveTime) {
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.saveTime - gLevels[level].levelClock.tSave; 
    } 
    
    if ((gLevels[level].levelClock.t+(ldt*kZLS_TimeEventSlop))>gLevels[level].levelClock.endTime){
      timingPrevTimeStep = rdt;
      timingShortenedStep = kZLS_PrevStepWasShortenedForIO;
      ldt = gLevels[level].levelClock.endTime -gLevels[level].levelClock.t; 
    }
    
    gLevels[level].levelClock.dt = ldt;
    gLevels[level].levelClock.t  = absoluteTime;
    
    rdt = ldt * refined;
    
    err = zlst_SynchLevelZoneClocks( level );
    
  } // level
  
  *dt = rdt;
#endif !UNIFORM_SWEEPS
  
  return(err);
}

//
//====================================================================================
//
int    zlst_InitTimeCoords  ( TimeCoords * tc ){
  int err = noErr;
  if ( tc != NULL ) {
    Size chunkSize = sizeof(TimeCoords);
    bzero((void*)tc, chunkSize);
  } else { err = memErr; }
  return(err);                 
}

int zlst_DefaultEventTiming( TimeCoords *tc ){
  
  zlst_InitTimeCoords  ( tc );
  
  tc->endTime   = 1.0;
  tc->frameTime = 0.1;
  tc->userTime  = 0.1;
  tc->dataTime  = 2.0;
  tc->saveTime  = 1.0;
  
  tc->endCycles   =   (kMaxInt >> MAXLEVELS); 
  tc->frameCycles = 1+(kMaxInt >> MAXLEVELS); 
  tc->userCycles  = 1+(kMaxInt >> MAXLEVELS); 
  tc->dataCycles  = 1+(kMaxInt >> MAXLEVELS); 
  tc->saveCycles  = 1+(kMaxInt >> MAXLEVELS); 
  
  zlst_SetRealTimeDuration  ( (time_t) 31557600 ); // 1 Julian Year real time limit!
  
  return(noErr);
}

//
int zlst_SetClockTime      ( TimeCoords * tc,  Real t){
  int err = noErr;
  
  tc->t = t;
  
  return(err);
}

//
//====================================================================================
//
int zlst_SetRealTime      ( void ){
  int err = noErr;
  gZLS_GlobalSimRealTime = time(NULL);
  return(err);
}

int zlst_SetRealTimeLimit  ( time_t  rt){
  int err = noErr;
  gZLS_GlobalSimRealTimeLimit = rt;
  gZLS_GlobalSimRealTimeDuration = rt - gZLS_GlobalSimRealTime;
  return(err);
}

int zlst_SetRealTimeDuration  ( time_t Delta_rt){ // sets *both* realtime and limit ,durtation is from now.
  int err = noErr;
  gZLS_GlobalSimRealTime         = time(NULL);
  gZLS_GlobalSimRealTimeLimit    = gZLS_GlobalSimRealTime + Delta_rt;
  gZLS_GlobalSimRealTimeDuration = Delta_rt;
  return(err);
}

time_t zlst_GetRealTime  ( void ){
  return(gZLS_GlobalSimRealTime);
}

time_t zlst_GetRealTimeLimit  ( void){
  return(gZLS_GlobalSimRealTimeLimit);
}

time_t zlst_GetRealTimeDuration  ( void){
  return(gZLS_GlobalSimRealTimeDuration);
}

time_t zlst_GetRealTimeRemaining  ( void){
  return(gZLS_GlobalSimRealTimeLimit - time(NULL));
}

//
//====================================================================================
//
int zlst_SetClockCycles    ( TimeCoords * tc,  Integer c){
  int err = noErr;
  
  tc->cycles= c;
  
  return(err);
}

int zlst_SetClockTimingCycles   ( TimeCoords * tc,  Integer c){
  int err = noErr;
  
  tc->timingCycles= c;
  
  return(err);
}
// 
// these call set and check level IO limits and sim limits, after a simulation step
//

int zlst_SetLevelClockTime  ( Integer level, Real t){
  
  int err = noErr;
  
  gLevels[level].levelClock.t = t;
  
  return(err);
}


int zlst_SetLevelClockEndTime  ( Integer level, Real t){
  
  int err = noErr;
  
  gLevels[level].levelClock.endTime = t;
  
  return(err);
}

int zlst_SetLevelClockFrameTime  ( Integer level, Real t){
  
  int err = noErr;
  
  if ( t > gLevels[level].levelClock.endTime)
    gLevels[level].levelClock.frameTime = gLevels[level].levelClock.endTime; 
  
  gLevels[level].levelClock.frameTime = t;
  
  return(err);
}


int zlst_SetLevelClockUserTime  ( Integer level, Real t){
  
  int err = noErr;
  
  if (t>gLevels[level].levelClock.endTime)
    gLevels[level].levelClock.userTime = gLevels[level].levelClock.endTime; 
  
  gLevels[level].levelClock.userTime = t;
  
  return(err);
}

int zlst_SetLevelClockDataTime  ( Integer level, Real t){
  
  int err = noErr;
  
  if (t>gLevels[level].levelClock.endTime)
    gLevels[level].levelClock.dataTime = gLevels[level].levelClock.endTime; 
  
  gLevels[level].levelClock.dataTime = t;
  
  return(err);
}

int zlst_SetLevelClockSaveTime  ( Integer level, Real t){
  
  int err = noErr;
  
  if (t>gLevels[level].levelClock.endTime)
    gLevels[level].levelClock.saveTime = gLevels[level].levelClock.endTime; 
  
  gLevels[level].levelClock.saveTime = t;
  
  return(err);
}


int zlst_SetLevelClockEndCycle  ( Integer level, Integer i ){
  
  int err = noErr;
  
  Integer   ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.endCycles = i*ntR;
  
  return(err);
}

int zlst_SetLevelClockCycles  ( Integer level, Integer i ){ // best to use 0 only
  
  int err = noErr;
  
  Integer   ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.cycles = i*ntR;
  
  return(err);
}

int zlst_SetLevelClockTimingCycles  ( Integer level, Integer i ){ 
  int err = noErr;
  
  Integer   ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.timingCycles = i*ntR;
  
  return(err);
}

int zlst_SetLevelClockFrameCycle  ( Integer level, Integer i ){
  
  int err = noErr;
  
  Integer     ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.frameCycles = i*ntR;
  
  return(err);
}



int zlst_SetLevelClockUserCycle  ( Integer level, Integer i ){
  
  int err = noErr;
  
  Integer     ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.userCycles= i*ntR;
  
  return(err);
}


int zlst_SetLevelClockDataCycle  ( Integer level, Integer i ){
  
  int err = noErr;
  
  Integer     ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.dataCycles= i*ntR;
  
  return(err);
}

int zlst_SetLevelClockSaveCycle  ( Integer level, Integer i ){
  
  int err = noErr;
  
  Integer     ntR;
  
  ntR = gLevels[level].levelTotalRefinement;
  gLevels[level].levelClock.saveCycles = i*ntR;
  
  return(err);
}


int zlst_SetAllLevelClocksTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockTime  ( level, t);
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksEndTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockEndTime  ( level, t);
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksFrameTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockFrameTime  ( level, t);
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksUserTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockUserTime  ( level, t);
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksDataTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockDataTime  ( level, t);
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksSaveTime  (Real t){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockSaveTime  ( level, t);
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksCycles  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockCycles  ( level,  i );
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksTimingCycles  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockTimingCycles  ( level,  i );
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksEndCycle  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockEndCycle  ( level,  i );
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksFrameCycle  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockFrameCycle  ( level,  i );
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksUserCycle  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockUserCycle  ( level,  i );
    
  } // level
  
  return(err);
}


int zlst_SetAllLevelClocksDataCycle  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockDataCycle  ( level,  i );
    
  } // level
  
  return(err);
}

int zlst_SetAllLevelClocksSaveCycle  ( Integer i ){
  
  int err = noErr;
  
  Integer   level;
  
  for (level = 0; level < gNLevels; level++){
    
    err = zlst_SetLevelClockSaveCycle  ( level,  i );
    
  } // level
  
  return(err);
}

//
//====================================================================================
//
//  Here is where the Simulation level timers are actually checked and reset as needed.
//

int zlst_CheckIOTimeLimits( TimeCoords * tc ){
  
  int event = kGoEventFlag;
  
  const Real kTimeTolerance = 1.00001; // to prevent timer failure due to roundoff in single or double.
  
  if ((tc->tFrame*kTimeTolerance) >= (tc->frameTime)) {
    tc->tFrame = 0.0;
    event |= kFrameEventFlag;
  }
  
  if ((tc->tUser*kTimeTolerance) >= (tc->userTime)) {
    tc->tUser = 0.0;
    event |= kUserEventFlag;
  }
  
  if ((tc->tData*kTimeTolerance) >= (tc->dataTime)) {
    tc->tData = 0.0;
    event |= kDataEventFlag;
  }
  
  if ((tc->tSave*kTimeTolerance) >= (tc->saveTime)) {
    tc->tSave = 0.0;
    event |= kSaveEventFlag;
  }
  
  if ((tc->t*kTimeTolerance) >= (tc->endTime)) {
    zlsio_Printf( " SIMULATION ENDED: at %8.4f = %8.4g%s \n\n",tc->t, tc->t*f_t0, s_Time);
    zlsio_Printf( "%s\n\n", DIVIDERLINE );
    event |= kEndEventFlag;
  }
  
  return(event);
}

int zlst_CheckIOCycleLimits( TimeCoords *tc ){
  int event = kGoEventFlag;
  
  if (tc->cycles == 0) { // trigger save initial state event, once off
    event |= kStartEventFlag;
  }
  
  if (tc->nCyclesFrame >= tc->frameCycles) {
    tc->nCyclesFrame = 0;
    event |= kFrameEventFlag;
  }
  
  if (tc->nCyclesUser >= tc->userCycles) {
    tc->nCyclesUser = 0;
    event |= kUserEventFlag;
  }
  
  if (tc->nCyclesData >= tc->dataCycles) {
    tc->nCyclesData = 0;
    event |= kDataEventFlag;
  }
  
  if (tc->nCyclesSave >= tc->saveCycles) {
    tc->nCyclesSave = 0;
    event |= kSaveEventFlag;
  }
  
  if (tc->cycles >= tc->endCycles) {
    zlsio_Printf( " SIMULATION CYCLE LIMIT: at cycle = %d \n\n",(int)tc->cycles);
    zlsio_Printf( "%s\n\n", DIVIDERLINE );
    event |= kEndEventFlag;
  }
  
  return(event);
}


int zlst_CheckRealTimeLimit(){
  int event = kGoEventFlag;
  
  Integer globalRTime = (Integer) zlst_GetRealTimeRemaining();   
  
#ifdef MPI  
  Integer localRTime = globalRTime;    
  MPI_Allreduce( &localRTime, &globalRTime, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);
#endif !MPI
  
#if (ZLS_DEBUG > 0)
  zlsio_Printf( "%s\n", DIVIDERLINE );
  zlsio_Printf( "Remaining: %d \n", globalRTime );
	zlsio_Printf( "%s\n\n", DIVIDERLINE );
#endif ZLS_DEBUG
  
  if ( globalRTime <= 0) {
    zlsio_Printf( " SIMULATION REALTIME LIMIT REACHED: %s \n", zlst_TheCurrentRealTime());
    zlsio_Printf( "%s\n\n", DIVIDERLINE );
    gZLS_GlobalSimRealTimeLimit = 0;
    event |= kEndEventFlag;
    event |= kSaveEventFlag;
  }
  
  return(event);
}



int zlst_CheckLevelIOTimeLimits  ( Integer level ){
  
  int code = kGoEventFlag;
  
  code = zlst_CheckIOTimeLimits ( &gLevels[level].levelClock );
  
  return(code);
}

int zlst_CheckLevelIOCycleLimits  ( Integer level ){
  
  int event = kGoEventFlag;
  
  event = zlst_CheckIOCycleLimits ( &gLevels[level].levelClock );
  
  return(event);
}

int   zlst_PrintEvent(Integer count, Integer level, const char * eventName);
int   zlst_PrintEvent(Integer count,  Integer level, const char * eventName){
  
  if ( f_t0 != 1.0 ) {
    zlsio_Printf( " #%s: %4.4d\n Cycle %6d Level %2d at t = %g%s = %g%s \n", eventName, count,
                 (int)gLevels[0].levelClock.cycles, level, zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr");
    zlsio_FilesPrintf( " #%s: %4.4d\n Cycle %6d Level %2d at t = %g%s = %g%s \n", eventName, count,
                      (int)gLevels[0].levelClock.cycles, level, zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr");
    
  } else {
    zlsio_Printf( " #%s: %4.4d Cycle %6d Level %2d at t = %g%s \n", eventName, count,
                 (int)gLevels[0].levelClock.cycles, level, zlst_GlobalSeconds, s_Time );
    zlsio_FilesPrintf( " #%s: %4.4d Cycle %6d Level %2d at t = %g%s \n", eventName, count,
                      (int)gLevels[0].levelClock.cycles, level, zlst_GlobalSeconds, s_Time );
  };
  
  return(noErr);
}


int zlst_CheckAllLevelEvents  ( void){
  
  int code           = goCode;
  int allEvents      = kGoEventFlag;
  int eventTriggered = kGoEventFlag;
  
  Integer level=0;
  Integer l=0;
  
  // update all level clock to the same real time before 
  // checking against real time ending event.
  zlst_SetRealTime();
  
  for (level = 0; level < gNLevels; level++){
    
    eventTriggered  = zlst_CheckLevelIOTimeLimits  ( level );
    eventTriggered |= zlst_CheckLevelIOCycleLimits ( level );
    eventTriggered |= zlst_CheckRealTimeLimit ();
    
    allEvents |= eventTriggered;
    
    if (eventTriggered & kFrameEventFlag) {
      // one level triggered an image frame, sych all the others, keep code
      for (l = 0; l < gNLevels; l++){
        gLevels[l].levelClock.nCyclesFrame = 0;
        gLevels[l].levelClock.tFrame = 0.0;
      }
    }
    
    if (eventTriggered & kSaveEventFlag) {
      // one level triggered a restart save, sych all the others, keep code
      for (l = 0; l < gNLevels; l++){
        gLevels[l].levelClock.nCyclesSave = 0;
        gLevels[l].levelClock.tSave = 0.0;
      }
    }
    
    if (eventTriggered & kUserEventFlag) {
      // one level triggered a user output, sych all the others, keep code
      for (l = 0; l < gNLevels; l++){
        gLevels[l].levelClock.nCyclesUser = 0;
        gLevels[l].levelClock.tUser = 0.0;
      }
    }
    
    if (eventTriggered & kDataEventFlag) {
      // one level triggered a data output, sych all the others, keep code
      for (l = 0; l < gNLevels; l++){
        gLevels[l].levelClock.nCyclesData = 0;
        gLevels[l].levelClock.tData = 0.0;
      }
    }
    
    if (eventTriggered & kEndEventFlag) {
      // one level finished, sych all the others, keep code
      for (l = 1; l < gNLevels; l++){
        gLevels[l].levelClock.t    = zlst_GlobalTime;
        gLevels[l].levelClock.endTime = zlst_GlobalTime;
      }
    }
    
  } // level
  
  Integer count = 0; // frame/file number count from level io state
  IOSettingsPtr io = gLevels[0].levelIO;
  
  if (allEvents & kFrameEventFlag) {
    if ( io != NULL ) count =  (int)io->ioState.ioFrameCount;
    for (l = gNLevels-1; l >= 0; l-- ){
      zlst_PrintEvent(count, l, "Frame");
      zlsio_SaveLevelImages( l, gLevels[0].levelClock.cycles);
    }
  }
  
  if (allEvents & kUserEventFlag) {
    if ( io != NULL ) count =  (int)io->ioState.ioUserCount;
    for (l = gNLevels-1; l >= 0; l-- ){
      zlst_PrintEvent(count, l, "User");
      zlsio_SaveLevelUserIO( l, gLevels[0].levelClock.cycles);
    }
  }
  
  if (allEvents & kDataEventFlag) {
    if ( io != NULL ) count =  (int)io->ioState.ioDataCount;
    for (l = gNLevels-1; l >= 0; l-- ){
      zlst_PrintEvent(count, l, "Data");
      zlsio_SaveLevelZoneData( l, gLevels[0].levelClock.cycles);
    }
  }
  
  if (allEvents & kSaveEventFlag) {
    if ( io != NULL ) count =  (int)io->ioState.ioSaveCount;
    zlst_PrintEvent(count, l,  "Save State");
    if ( io != NULL) {
      zlsio_SaveState ( io->restartName );
    }  else {
      zlsio_SaveState ( "Restart" );
    }
  }
  
  if (allEvents & kStartEventFlag) {
    zlst_PrintEvent(count, l, "Save Initial State");
    zlsio_WriteInitialOutput();
  }    
  
  if (allEvents & kEndEventFlag) {
    
    for (l = gNLevels-1; l >= 0; l-- ){
      zlsio_Printf( " #END  : Cycle %6d Level %2d at t = %g%s = %g%s \n", 
                   (int)gLevels[0].levelClock.cycles, l, zlst_GlobalSeconds, s_Time, zlst_GlobalYears, "yr");
      
      // save the final output if any are still outstanding (>0)  at the end 
      
      if (!(allEvents & kFrameEventFlag)&&(gLevels[0].levelClock.frameTime <= gLevels[0].levelClock.endTime)) {
        zlsio_SaveLevelImages( l, gLevels[0].levelClock.cycles);
      }
      
      // save the final userdump if a user event didn't happen at the end
      if (!(allEvents & kUserEventFlag)&&(gLevels[0].levelClock.userTime <= gLevels[0].levelClock.endTime)) {
        zlsio_SaveLevelUserIO( l, gLevels[0].levelClock.cycles);
      }
      
      // save the final data if a data event didn't happen at the end
      if (!(allEvents & kDataEventFlag)&&(gLevels[0].levelClock.dataTime <= gLevels[0].levelClock.endTime)) {
        zlsio_SaveLevelZoneData( l, gLevels[0].levelClock.cycles);
      }
      
    }
    // save the final restart if a save event didn't already happen at the end
    if (!(allEvents & kSaveEventFlag)&&(gLevels[0].levelClock.saveTime <= gLevels[0].levelClock.endTime)) {
      if ( io != NULL) {
        zlsio_SaveState ( io->restartName );
      }  else {
        zlsio_SaveState ( "Restart" );
      }
    }
    
    code = endCode;
  }
  
  
  return(code);
  
}
//====================================================================================
