#include "zls.h"


int zls_RangeOptions(char* opt, int *i, int *j) {
  int err = noErr;
  int ncp;
  char* cp = strchr(opt, ':');
  
  if (cp && isdigit(cp[1])) {
    *j= atoi(cp+1);
    ncp = strcspn(opt, ":");
    if ((ncp > 0)&&(isdigit(opt[0]))) {
      opt[ncp]='\0';
      *i = atoi(opt);
      return (2); // found two
    } else {
      *i = 0;
      return (3); // found only 1, second arg
    }
  } else {
    if (isdigit(opt[0])) {
      *i = atoi(opt);
      *j = 0;
      return (1); // found 1, first arg
    }
  }
  
  *i = 0;
  *j = 0;
  return (err);
  
}
//========================================

static char gTimeIntervalString[80];

char * zls_TimeIntervalString( Real seconds ){
  
  if ( seconds < 1e-12) seconds = 0.0;
  
  if ( seconds < 1e-9 ) { // less than nano - seconds - just use seconds
    sprintf(gTimeIntervalString, "%#12.5gs", seconds);
  } else if ( seconds < 1e-6 ) { // nano - seconds
    sprintf(gTimeIntervalString, "%#12.5gns", seconds*1.0e9);
  } else if  ( seconds < 0.001 ) { // micro - seconds
    sprintf(gTimeIntervalString, "%#12.5gs", seconds*1.0e6);
  } else if  ( seconds < 1.0 ) { //milli- seconds
    sprintf(gTimeIntervalString, "%#12.5gms", seconds*1000.0);
  } else if ( seconds < 120.0 ) { // seconds
    sprintf(gTimeIntervalString, "%#12.5gs", seconds);
  } else if ( seconds < 3600.0 ) { // minutes and  seconds
    Integer mins = seconds/60.0;
    seconds -= 60.0*mins;
    Integer iseconds = seconds;
    sprintf(gTimeIntervalString, "%3.2dm %3.2ds", mins, iseconds);
  } else if ( seconds < 86400.0 ) {  // hours, minutes, seconds
    Integer hours = seconds/3600.0;
    seconds -= 3600.0*hours;
    Integer mins  = seconds/60.0;
    seconds -= 60.0*mins;
    Integer iseconds = seconds;
    sprintf(gTimeIntervalString, "%3.2dh%3.2dm%3.2ds", hours, mins, iseconds);
  } else {  // days hours, minutes
    Integer days = seconds/86400.0;
    seconds -= 86400.0*days;
    Integer hours = seconds/3600.0;
    seconds -= 3600.0*hours;
    Integer mins  = seconds/60.0;
    sprintf(gTimeIntervalString, "%3.2dd%3.2dh%3.2dm", days, hours, mins);
  }
  
  return(gTimeIntervalString);
  
}


//========================================
// utility functions for directions.

static char gDirString[8];

char * zls_dirname( Integer direction){
  
  switch (direction){
      
    case  kDirRight:
      strcpy(gDirString, "Right");
      break;
    case  kDirLeft:
      strcpy(gDirString, "Left");
      break;
    case  kDirUp:
      strcpy(gDirString, "Up");
      break;
    case  kDirDown:
      strcpy(gDirString, "Down");
      break;
    case  kDirHither:
      strcpy(gDirString, "Hither");
      break;
    case  kDirYon:
      strcpy(gDirString, "Yon");
      break;
    case  kDirSuper:
      strcpy(gDirString, "Super");
      break;
    case  kDirSub:
      strcpy(gDirString, "Sub");
      break;
    default:
      ;
  }
  
  return(gDirString);
  
}

int zls_opposite(int direction){
  int opposite = -1;
  switch (direction){
    case  kDirRight: // increasing index, 1st dim
      opposite = kDirLeft;
      break;
    case  kDirLeft:// decreasing index, 1st dim
      opposite = kDirRight;
      break;
    case  kDirUp: // increasing index, 2nd dim
      opposite = kDirDown;
      break;
    case  kDirDown:// decreasing index, 2nd dim
      opposite = kDirUp;
      break;
    case  kDirHither:// increasing index, 3rd dim
      opposite = kDirYon;
      break;
    case  kDirYon: // decreasing index, 3rd dim
      opposite = kDirHither;
      break;
    case  kDirSuper:// increasing level
      opposite = kDirSub;
      break;
    case  kDirSub:// decreasing level
      opposite = kDirSuper;
      break;
    default:
      zlsio_ErrPrintf( " zls_opposite: Invalid Direction code: %d , no opposite.\n", direction);
  }
  return (opposite);
}


Real zls_AxisFiducialScale( Integer axis){
  
  Real fx = f_x0; // also answer for all 1D and CARTESIAN cases
#ifdef SYM1D
  if ( axis == kYDim ) { return (1.0); }
  if ( axis == kZDim ) { return (1.0); }
#endif SYM1D
#ifdef SYM2D
  if ( axis == kZDim ) { return (1.0); }
#ifdef SPHERICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_r0;
      break;
    case  kYDim:
      fx = f_phi0;
      break;
  }
#endif SPHERICAL
#ifdef POLARSPHERICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_r0;
      break;
    case  kYDim:
      fx = f_phi0;
      break;
  }
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_z0;
      break;
    case  kYDim:
      fx = f_r0;
      break;
  }
#endif CYLINDRICAL
#ifdef CIRCULAR
  switch (axis){
    default:
    case  kXDim:
      fx = f_r0;
      break;
    case  kYDim:
      fx = f_phi0;
      break;
  }
#endif CIRCULAR
#endif SYM2D
  
#ifdef SYM3D
  
#ifdef SPHERICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_r0;
      break;
    case  kYDim:
      fx = f_phi0;
      break;
    case  kZDim:
      fx = f_theta0;
  }
#endif SPHERICAL
#ifdef POLARSPHERICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_r0;
      break;
    case  kYDim:
      fx = f_theta0;
      break;
    case  kZDim:
      fx = f_phi0;
  }
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
  switch (axis){
    default:
    case  kXDim:
      fx = f_z0;
      break;
    case  kYDim:
      fx = f_r0;
      break;
    case  kZDim:
      fx = f_phi0;
  }
#endif CYLINDRICAL
#endif SYM3D
  return(fx);
}

//========================================
// Powerlaw integerator for use in initialisation, 
//integrates/averages over cell for correct
// cell values (ie not just central coordinate value)  
// slow. Special cases for alpha 
// = -3, -2, -1, 0, 1, 2, 3  
// x1 != x0, alpha may equal -1
//========================================

Real zls_IntegratePowerlaw( Real r0,  Real alpha,  Real x0,  Real x1){
  
  if ( x1 == x0 ) return (0.0);
  
  if ( alpha == -2.0 )  return  r0*r0 /(x0*x1); 
  if ( alpha == -1.0 )  return  r0* ( log(x0) - log(x1))/((x0 - x1)); 
  if ( alpha ==  0.0 )  return  1.0; 
  if ( alpha ==  1.0 )  return  ( x1*x1 - x0*x0) /(2.0*r0*(x1 - x0)); 
  if ( alpha ==  2.0 )  return  ( x1*x1*x1 - x0*x0*x0) /(3.0*r0*r0*(x1 - x0)); 
  
  if ( alpha == -3.0 )  return  r0*r0*r0*( x0 + x1)/(2.0*x0*x0*x1*x1);
  if ( alpha ==  3.0 )  return  ( x1*x1*x1*x1 - x0*x0*x0*x0) / (4.0*r0*r0*r0*(x1 - x0)); 
  
  // general case left.., v slow, so OK to wait for ifs above
  
  return  pow( r0, -alpha ) *
  (pow(x0, 1.0+alpha) - pow(x1, 1.0+alpha))/
  ( (1.0+alpha)*(x0 - x1) ); // radial integral average for powerlaw
  
}

Real zls_erf(Real x){
  //
  //  Abramowitz and Stegun 7.1.26;
  //
  //  err <= 1.5e-7;
  //
  float t, erf;
  
  const float p  =  0.3275911F;
  const float a1 =  0.254829592F;
  const float a2 = -0.284496736F;
  const float a3 =  1.421413741F;
  const float a4 = -1.453152027F;
  const float a5 =  1.061405429F;
  
  
  t = 1.0F/(1.0F+(p*x));
  erf = 1.0F - exp(-1.0F*(x*x))*(t*(a1+t*(a2+t*(a3+t*(a4+t*a5)))));
  
  return(erf);
}

Real zls_invNormal(Real x){
  //  Algorithm AS 241: The Percentage Points of the Normal Distribution
  //  Michael J. Wichura
  //  Applied Statistics, Vol. 37, No. 3. (1988), pp. 477-484.
  //  Stable URL: http://links.jstor.org/sici?sici=0035-9254%281988%2937%3A3%3C477%3AAA2TPP%3E2.0.CO%3B2-U
  //C
  //C	ALGORITHM AS241  APPL. STATIST. (1988) VOL. 37, NO. 3
  //C
  //C	Produces the normal deviate Z corresponding to a given lower
  //C	tail area of P; Z is accurate to about 1 part in 10**16.
  //C
  //
  // x = 0 -> 1.0
  //
  
  double ppnd16 = 0.0;
  double q, r;
  
  int ifault = 0;
  
  const double split1 = 0.425e0;
  const double split2 = 5.0e0;
  const double const1 = 0.180625e0;
  const double const2 = 1.6e0;
  //  c
  //  c	coefficients for p close to 0.5
  //  c
  const double a[8] = { 3.3871328727963666080e+0,
    1.3314166789178437745e+2,
    1.9715909503065514427e+3,
    1.3731693765509461125e+4,
    4.5921953931549871457e+4,
    6.7265770927008700853e+4,
    3.3430575583588128105e+4,
  2.5090809287301226727e+3};
  
  const double  b[8]  = { 1.0000000000000000000e+0,
    4.2313330701600911252e+1,
    6.8718700749205790830e+2,
    5.3941960214247511077e+3,
    2.1213794301586595867e+4,
    3.9307895800092710610e+4,
    2.8729085735721942674e+4,
  5.2264952788528545610e+3};
  //  
  //c
  //c	coefficients for p not close to 0, 0.5 or 1.
  //c
  //
  const double  c[8]  = {1.42343711074968357734e0,
    4.63033784615654529590e0,
    5.76949722146069140550e0,
    3.64784832476320460504e0,
    1.27045825245236838258e0,
    2.41780725177450611770e-1,
    2.27238449892691845833e-2,
  7.74545014278341407640e-4};
  
  const double  d[8]  = { 1.0000000000000000000e+0,
    2.05319162663775882187e0,
    1.67638483018380384940e0,
    6.89767334985100004550e-1,
    1.48103976427480074590e-1,
    1.51986665636164571966e-2,
    5.47593808499534494600e-4,
  1.05075007164441684324e-9};
  //  
  //c
  //c	coefficients for p near 0 or 1.
  //c
  //
  const double  e[8]  = { 6.65790464350110377720e0,
    5.46378491116411436990e0,
    1.78482653991729133580e0,
    2.96560571828504891230e-1,
    2.65321895265761230930e-2,
    1.24266094738807843860e-3,
    2.71155556874348757815e-5,
  2.01033439929228813265e-7};
  
  const double  f[8]  = { 1.0000000000000000000e+0,
    5.99832206555887937690e-1,
    1.36929880922735805310e-1,
    1.48753612908506148525e-2,
    7.86869131145613259100e-4,
    1.84631831751005468180e-5,
    1.42151175831644588870e-7,
  2.04426310338993978564e-15};
  //
  ifault = 0;
  q = x - 0.5e0;
  
  if (fabs(q) <= split1) {
    
    r = const1 - (q * q);
    
    ppnd16 = q * (((((((a[7]*r +a[6])*r +a[5])*r +a[4])*r +a[3])* r + a[2])*r +a[1])*r +a[0]) /
    (((((((b[7]*r +b[6])*r +b[5])*r +b[4])*r +b[3])* r + b[2])*r +b[1])*r +b[0]);
    
    
  } else {
    
    if (q < 0.0) {
      r = x;
    }    else  {
      r = 1.0 - x;
    }
    
    if (r <= 0.0) {
      ifault = -1;
      ppnd16 = 0.0;
      return(  ppnd16);
    }
    
    r = sqrt(-log(r));
    if (r <= split2) {
      
      r -= const2;
      ppnd16 = (((((((c[7]*r +c[6])*r +c[5])*r +c[4])*r +c[3])*r +c[2])*r +c[1])*r +c[0]) /
      (((((((d[7]*r +d[6])*r +d[5])*r +d[4])*r +d[3])* r + d[2])*r +d[1])*r +d[0]);
      
    }   else {
      
      r =- split2;
      ppnd16 = (((((((e[7]*r +e[6])*r +e[5])*r +e[4])*r +e[3])* r + e[2])*r +e[1])*r +e[0]) /
      (((((((f[7]*r +f[6])*r +f[5])*r +f[4])*r +f[3])* r +f[2])*r +f[1])*r +f[0]);
      
    }
    
    if (q < 0.0) ppnd16 = -ppnd16;
    
  }
  
  return(  ppnd16);
}

Real zls_RealSpH0402( Real x, Real y){
  
  Real sph0402 = 0.12544622691699172;
  sph0402 += 0.16726163588932229*cos(2.*y);
  sph0402 -= 0.29270786280631401*cos(4.*y);
  sph0402 *= cos(2.*x);
  
  return(sph0402);
}

Real zls_RealSpH0804( Real x, Real y){
  
  Real sph0804  = 0.09250160086809265;
  sph0804 += 0.105716115277820170*cos(2.*y);
  sph0804 -= 0.095144503750038154*cos(4.*y);
  sph0804 -= 0.274861899722332440*cos(6.*y);
  sph0804 += 0.171788687326457780*cos(8.*y);
  sph0804 *= cos(4.*x);
  
  return(sph0804);
}

Real zls_RealSpH1608( Real x, Real y){
  
  Real sph1608  = 0.066809098965275787;
  sph1608 += 0.071263038896294173*cos(2.*y);
  sph1608 -= 0.063040380562106384*cos(4.*y);
  sph1608 -= 0.145017792439311920*cos(6.*y);
  sph1608 -= 0.071470681783521138*cos(8.*y);
  sph1608 += 0.120432874591639340*cos(10.*y);
  sph1608 += 0.177801638005489210*cos(12.*y);
  sph1608 -= 0.216779174264950810*cos(14.*y);
  sph1608 += 0.060001378591191741*cos(16.*y);
  sph1608 *= cos(8.*x);
  
  return(sph1608);
}

Real zls_RealSpH3216( Real x, Real y){
  
  Real sph3216  = 0.047742216808322043;
  sph3216 += 0.049282288318267916*cos(2.*y);
  sph3216 -= 0.045600278271500772*cos(4.*y);
  sph3216 -= 0.097368710125620007*cos(6.*y);
  sph3216 -= 0.051684721198355127*cos(8.*y);
  sph3216 += 0.050843208816726465*cos(10.*y);
  sph3216 += 0.10390317650263479*cos(12.*y);
  sph3216 += 0.040074217003986952*cos(14.*y);
  sph3216 -= 0.080418141933328912*cos(16.*y);
  sph3216 -= 0.106964849785293890*cos(18.*y);
  sph3216 += 0.023643919186687215*cos(20.*y);
  sph3216 += 0.140037628741524840*cos(22.*y);
  sph3216 += 0.0035059183457339516*cos(24.*y);
  sph3216 -= 0.190258804389077910*cos(26.*y);
  sph3216 += 0.162098751976872960*cos(28.*y);
  sph3216 -= 0.056213893522394829*cos(30.*y);
  sph3216 += 0.0073780735248143213*cos(32.*y);
  sph3216 *= cos(16.*x);
  
  return(sph3216);
  
}


Real zls_RealSpH6432( Real x, Real y){
  
  Real sph6432  =0.033936472510172123;
  sph6432 += 0.034475146677000252*cos(2.*y);
  sph6432 -= 0.033028985750819432*cos(4.*y);
  sph6432 -= 0.068200306916148694*cos(6.*y);
  sph6432 -= 0.035679257628522644*cos(8.*y);
  sph6432 += 0.033097450770091295*cos(10.*y);
  sph6432 += 0.069275309603835139*cos(12.*y);
  sph6432 += 0.034847396425798230*cos(14.*y);
  sph6432 -= 0.036694672589310479*cos(16.*y);
  sph6432 -= 0.071058080055377143*cos(18.*y);
  sph6432 -= 0.029480107738304106*cos(20.*y);
  sph6432 += 0.046068107289204076*cos(22.*y);
  sph6432 += 0.072132306062889642*cos(24.*y);
  sph6432 += 0.015173099769357631*cos(26.*y);
  sph6432 -= 0.062636271165185666*cos(28.*y);
  sph6432 -= 0.066513834412514230*cos(30.*y);
  sph6432 += 0.015301989029389872*cos(32.*y);
  sph6432 += 0.081677687057397408*cos(34.*y);
  sph6432 += 0.037472539209144028*cos(36.*y);
  sph6432 -= 0.065877830998964143*cos(38.*y);
  sph6432 -= 0.074609662730029002*cos(40.*y);
  sph6432 += 0.041025650853025544*cos(42.*y);
  sph6432 += 0.095489507527944825*cos(44.*y);
  sph6432 -= 0.030125394956069642*cos(46.*y);
  sph6432 -= 0.109204896665583590*cos(48.*y);
  sph6432 += 0.059722350550766810*cos(50.*y);
  sph6432 += 0.100197552482745420*cos(52.*y);
  sph6432 -= 0.165475259265258200*cos(54.*y);
  sph6432 += 0.116117456678482050*cos(56.*y);
  sph6432 -= 0.047884066454184989*cos(58.*y);
  sph6432 += 0.012096781187917078*cos(60.*y);
  sph6432 -= 0.0017502110371764652*cos(62.*y);
  sph6432 += 0.00011203467828700155*cos(64.*y);
  sph6432 *= cos(32.*x);
  
  return(sph6432);
  
}

Real zls_RealSpH12864( Real x, Real y){
  
  Real sph12864  = 0.024059539180718144;
  sph12864 += 0.024248984371117500*cos(2.*y);
  sph12864 -= 0.023711097808703621*cos(4.*y);
  sph12864 -= 0.048176656257530347*cos(6.*y);
  sph12864 -= 0.024741578501070897*cos(8.*y);
  sph12864 += 0.023440298645222395*cos(10.*y);
  sph12864 += 0.048358367538770557*cos(12.*y);
  sph12864 += 0.024879868194429787*cos(14.*y);
  sph12864 -= 0.023764832905668390*cos(16.*y);
  sph12864 -= 0.048677701864014204*cos(18.*y);
  sph12864 -= 0.024331324189267325*cos(20.*y);
  sph12864 += 0.025028676883683193*cos(22.*y);
  sph12864 += 0.049109592291652431*cos(24.*y);
  sph12864 += 0.022695773749393498*cos(26.*y);
  sph12864 -= 0.027574184022851338*cos(28.*y);
  sph12864 -= 0.049508904524378377*cos(30.*y);
  sph12864 -= 0.019445712241173267*cos(32.*y);
  sph12864 += 0.031689143422720179*cos(34.*y);
  sph12864 += 0.049475149184389506*cos(36.*y);
  sph12864 += 0.013867724097668820*cos(38.*y);
  sph12864 -= 0.037474841013101330*cos(40.*y);
  sph12864 -= 0.048146974739962423*cos(42.*y);
  sph12864 -= 0.0050600066514052196*cos(44.*y);
  sph12864 += 0.044553610691091646*cos(46.*y);
  sph12864 += 0.043936112381683908*cos(48.*y);
  sph12864 -= 0.0078726361286263928*cos(50.*y);
  sph12864 -= 0.051498761606194470*cos(52.*y);
  sph12864 -= 0.034323541738673596*cos(54.*y);
  sph12864 += 0.02512270142519557*cos(56.*y);
  sph12864 += 0.054920609616957861*cos(58.*y);
  sph12864 += 0.016168570732322026*cos(60.*y);
  sph12864 -= 0.044483854844686447*cos(62.*y);
  sph12864 -= 0.048559129777358634*cos(64.*y);
  sph12864 += 0.012316801326541614*cos(66.*y);
  sph12864 += 0.05833472498550985*cos(68.*y);
  sph12864 += 0.024165396255350171*cos(70.*y);
  sph12864 -= 0.045984661310713268*cos(72.*y);
  sph12864 -= 0.051065058444648541*cos(74.*y);
  sph12864 += 0.021008628864245991*cos(76.*y);
  sph12864 += 0.063455841836238426*cos(78.*y);
  sph12864 += 0.0058695585752350846*cos(80.*y);
  sph12864 -= 0.064258242966176797*cos(82.*y);
  sph12864 -= 0.027672507670313910*cos(84.*y);
  sph12864 += 0.060175440864937886*cos(86.*y);
  sph12864 += 0.041841797267907315*cos(88.*y);
  sph12864 -= 0.058243058169578909*cos(90.*y);
  sph12864 -= 0.047765816447180616*cos(92.*y);
  sph12864 += 0.064151259359591209*cos(94.*y);
  sph12864 += 0.042454103895382962*cos(96.*y);
  sph12864 -= 0.080284413737503114*cos(98.*y);
  sph12864 -= 0.015048271864617261*cos(100.*y);
  sph12864 += 0.096183422064011304*cos(102.*y);
  sph12864 -= 0.052483273097736766*cos(104.*y);
  sph12864 -= 0.058654985806865470*cos(106.*y);
  sph12864 += 0.128240988778890920*cos(108.*y);
  sph12864 -= 0.122607133200293680*cos(110.*y);
  sph12864 += 0.07819795102506795*cos(112.*y);
  sph12864 -= 0.036683306675512571*cos(114.*y);
  sph12864 += 0.013092996493294941*cos(116.*y);
  sph12864 -= 0.0035826499146459212*cos(118.*y);
  sph12864 += 0.00074414451302166662*cos(120.*y);
  sph12864 -= 0.00011408685095316608*cos(122.*y);
  sph12864 += 0.000012219293153971771*cos(124.*y);
  sph12864 -= 8.1872369210172103e-7*cos(126.*y);
  sph12864 += 2.5889700085061863e-8*cos(128.*y);
  sph12864 *= cos(64.*x);
  
  return(sph12864);
  
}

Real zls_RealSpH256128( Real x, Real y){
  Real sph256128 = 0.017034876705636073;
  sph256128 +=0.017101680143697391*cos(2.*y);
  sph256128 -=0.016906692911897861*cos(4.*y);
  sph256128 -= 0.034079914984415403*cos(6.*y);
  sph256128 -=0.017288387866776873*cos(8.*y);
  sph256128 +=0.016759061855763619*cos(10.*y);
  sph256128 += 0.034111232952831235*cos(12.*y);
  sph256128 +=0.017411766166253863*cos(14.*y);
  sph256128 -=0.016715553555657604*cos(16.*y);
  sph256128 -= 0.034165607405229173*cos(18.*y);
  sph256128 -=0.017416187478372496*cos(20.*y);
  sph256128 +=0.016834036492934346*cos(22.*y);
  sph256128 += 0.034244176928208675*cos(24.*y);
  sph256128 +=0.017243609337375557*cos(26.*y);
  sph256128 -=0.017172855535594608*cos(28.*y);
  sph256128 -= 0.034344170800420741*cos(30.*y);
  sph256128 -=0.016831051433371823*cos(32.*y);
  sph256128 +=0.017789459018418021*cos(34.*y);
  sph256128 += 0.034454284782567553*cos(36.*y);
  sph256128 +=0.016107483981840239*cos(38.*y);
  sph256128 -=0.018737926715519912*cos(40.*y);
  sph256128 -= 0.03454836137176746*cos(42.*y);
  sph256128 -=0.014990423760702163*cos(44.*y);
  sph256128 +=0.020064763550869323*cos(46.*y);
  sph256128 += 0.034577127056762536*cos(48.*y);
  sph256128 +=0.013382870474971981*cos(50.*y);
  sph256128 -=0.021801955396340289*cos(52.*y);
  sph256128 -= 0.034457811374737853*cos(54.*y);
  sph256128 -=0.011171761534913703*cos(56.*y);
  sph256128 +=0.023955852649758024*cos(58.*y);
  sph256128 += 0.034061772332924947*cos(60.*y);
  sph256128 +=0.008229973366917583*cos(62.*y);
  sph256128 -=0.02649003874507744*cos(64.*y);
  sph256128 -= 0.033200998559725893*cos(66.*y);
  sph256128 -=0.0044251483683426588*cos(68.*y);
  sph256128 +=0.029300168207875645*cos(70.*y);
  sph256128 += 0.03161589364967577*cos(72.*y);
  sph256128 -=0.00035969319962685209*cos(74.*y);
  sph256128 -=0.032179281676486279*cos(76.*y);
  sph256128 -= 0.028969549668983966*cos(78.*y);
  sph256128 +=0.0061868914367051962*cos(80.*y);
  sph256128 +=0.034774198511972818*cos(82.*y);
  sph256128 += 0.024858312424955995*cos(84.*y);
  sph256128 -=0.012998008948474985*cos(86.*y);
  sph256128 -=0.036538731371466095*cos(88.*y);
  sph256128 -= 0.018855082142462174*cos(90.*y);
  sph256128 +=0.020513701422025782*cos(92.*y);
  sph256128 +=0.036699797641016872*cos(94.*y);
  sph256128 += 0.010609599027575603*cos(96.*y);
  sph256128 -=0.028100011124728378*cos(98.*y);
  sph256128 -=0.034270283299215403*cos(100.*y);
  sph256128 -= 0.000035112297283404929*cos(102.*y);
  sph256128 +=0.034622224712614433*cos(104.*y);
  sph256128 +=0.028168050040303922*cos(106.*y);
  sph256128 -= 0.012396558851580048*cos(108.*y);
  sph256128 -=0.038350457582135389*cos(110.*y);
  sph256128 -=0.017526758615720532*cos(112.*y);
  sph256128 += 0.025269290970819607*cos(114.*y);
  sph256128 +=0.037052706568521245*cos(116.*y);
  sph256128 +=0.0022863451187036685*cos(118.*y);
  sph256128 -= 0.035866476810717877*cos(120.*y);
  sph256128 -=0.028498390910024188*cos(122.*y);
  sph256128 +=0.015928460166668964*cos(124.*y);
  sph256128 += 0.040156872443132493*cos(126.*y);
  sph256128 +=0.011630853896308875*cos(128.*y);
  sph256128 -=0.032890046535726969*cos(130.*y);
  sph256128 -= 0.033686736096788137*cos(132.*y);
  sph256128 +=0.011522999436216981*cos(134.*y);
  sph256128 +=0.041578902710712335*cos(136.*y);
  sph256128 += 0.014043562642495312*cos(138.*y);
  sph256128 -=0.03384708280136892*cos(140.*y);
  sph256128 -=0.03415549341966046*cos(142.*y);
  sph256128 += 0.015049702796791471*cos(144.*y);
  sph256128 +=0.043105604177773809*cos(146.*y);
  sph256128 +=0.0075721812929705913*cos(148.*y);
  sph256128 -= 0.040110172672645067*cos(150.*y);
  sph256128 -=0.027438051451175701*cos(152.*y);
  sph256128 +=0.028085771457540316*cos(154.*y);
  sph256128 += 0.040612075314351152*cos(156.*y);
  sph256128 -=0.011509693319202643*cos(158.*y);
  sph256128 -=0.046164129827226635*cos(160.*y);
  sph256128 -= 0.005473325922505404*cos(162.*y);
  sph256128 +=0.045412804550290921*cos(164.*y);
  sph256128 +=0.020107071177338768*cos(166.*y);
  sph256128 -= 0.040784338971976129*cos(168.*y);
  sph256128 -=0.031218117043123893*cos(170.*y);
  sph256128 +=0.034850116772408503*cos(172.*y);
  sph256128 += 0.03883890438363032*cos(174.*y);
  sph256128 -=0.02978967162545955*cos(176.*y);
  sph256128 -=0.043593975783239371*cos(178.*y);
  sph256128 += 0.0272573087694161*cos(180.*y);
  sph256128 +=0.046062731192375743*cos(182.*y);
  sph256128 -=0.02847080588693408*cos(184.*y);
  sph256128 -= 0.046173553773489901*cos(186.*y);
  sph256128 +=0.034246464342139416*cos(188.*y);
  sph256128 +=0.042598832339481793*cos(190.*y);
  sph256128 -= 0.044586745451429627*cos(192.*y);
  sph256128 -=0.03225130675881525*cos(194.*y);
  sph256128 +=0.057258478559248281*cos(196.*y);
  sph256128 += 0.01061508275991408*cos(198.*y);
  sph256128 -=0.064923356038508245*cos(200.*y);
  sph256128 +=0.024838373537718717*cos(202.*y);
  sph256128 += 0.052298336156300237*cos(204.*y);
  sph256128 -=0.064270373271189177*cos(206.*y);
  sph256128 -=0.0019295482002975114*cos(208.*y);
  sph256128 += 0.069736840853352319*cos(210.*y);
  sph256128 -=0.071307833938880347*cos(212.*y);
  sph256128 +=0.0099897053879857973*cos(214.*y);
  sph256128 += 0.062853008822448617*cos(216.*y);
  sph256128 -=0.10338855266257569*cos(218.*y);
  sph256128 +=0.10271911766375012*cos(220.*y);
  sph256128 -= 0.077265686242447847*cos(222.*y);
  sph256128 +=0.047271933591066096*cos(224.*y);
  sph256128 -=0.024311835347572327*cos(226.*y);
  sph256128 += 0.010697128922418208*cos(228.*y);
  sph256128 -=0.0040661968108348888*cos(230.*y);
  sph256128 +=0.0013420036768745279*cos(232.*y);
  sph256128 -= 0.00038519688901410824*cos(234.*y);
  sph256128 +=0.000096055210250652424*cos(236.*y);
  sph256128 -=0.000020736273010354386*cos(238.*y);
  sph256128 += 3.8516560647187257e-6*cos(240.*y);
  sph256128 -=6.0996069220846402e-7*cos(242.*y);
  sph256128 +=8.1295866363851617e-8*cos(244.*y);
  sph256128 -= 8.954477056960964e-9*cos(246.*y);
  sph256128 +=7.9413975318909343e-10*cos(248.*y);
  sph256128 -=5.4540018389924537e-11*cos(250.*y);
  sph256128 += 2.723843007438512e-12*cos(252.*y);
  sph256128 -=8.8061301504085039e-14*cos(254.*y);
  sph256128 +=1.3840835712533051e-15*cos(256.*y);
  sph256128 *= cos(128.*x);
  
  return(sph256128);
  
}

Real zls_OverlapSin2D (Real ampl, Real f, Real phase, Real offset,
                       Real xin, Real yin,  Real  dx, Real dy,
                       Integer n ){
  //
  //  overlap of a ampl*sin (2pi*f*x + phase*pi))+offset wave and a square (xin, yin, dx, dy).;
  //  returns the faction of cell that lies below the sine curve,inside the cell.
  //     
  //   xin and yin are cell centre coordinates;
  //
  //  Divides square into n by n sub cells and counts cells whose centres;
  //  are inside the circle.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;
  //
  //
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  
  Integer i, j;
  
  const Integer nzones = n;
  Real dzone           = 1.0/nzones; // 1/[nzones]
  Real fzone           = 1.0/(nzones*nzones); // 1/(nzones x nzones)
  
  Real     s;
  Real     xs[nzones];
  Real     ys[nzones];
  Real     fracin;
  Real     deltax;
  Real     deltay;
  //
  //
  //      Assume all outside circle to start, fracin = 0;
  //
  //
  fracin = 0.0;
  //
  //  allow for rectangular cells;
  //
  deltax = dzone*dx;
  deltay = dzone*dy;
  //
  // set up sub cell centre coords - could be pre calculated possibly;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ( (xin -(0.5*dx)) + ( 0.5 +  (Real)(i-1) ) * deltax) ;
    xs[i] = ampl*sin(twopi*f*xs[i]+(phase*pi))+offset;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ( (yin -(0.5*dy)) + ( 0.5 +  (Real)(j-1) ) * deltay) ;
  }
  //
  //  count subcells inside circle..;
  //
  for ( i=0; i<nzones; i++){
    s = xs[i];
    for ( j=0; j<nzones; j++){
      if (s > ys[j]) fracin = fracin + fzone;
    }
  }
  
  fracin = fracin < 0.0? 0.0: fracin;
  fracin = fracin > 1.0? 1.0: fracin;
  
  //
  //  return fraction of cell in circle;
  //
  return(fracin);
}


Real zls_OverlapCos2D ( Real ampl, Real f, Real phase, Real offset,
                       Real xin, Real yin, Real  dx, Real dy,
                       Integer n ){
  //
  //  overlap of a ampl*cos (2pi*f*x + phase*pi))+offset wave and a square (xin, yin, dx, dy).;
  //  returns the faction of cell that lies below the sine curve,inside the cell.
  //     
  //   xin and yin are cell centre coordinates;
  //
  //  Divides square into n by n sub cells and counts cells whose centres;
  //  are inside the circle.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;
  //
  //
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  
  Integer i, j;
  
  Integer nzones = n;
  Real dzone     = 1.0/(Real)nzones; // 1/[nzones]
  Real fzone     = 1.0/(Real)(nzones*nzones); // 1/(nzones x nzones)
  
  Real     s;
  Real     xs[nzones];
  Real     ys[nzones];
  Real     fracin;
  Real     deltax;
  Real     deltay;
  //
  //
  //      Assume all outside circle to start, fracin = 0;
  //
  //
  fracin = 0.0;
  //
  //  allow for rectangular cells;
  //
  deltax = dzone*dx;
  deltay = dzone*dy;
  //
  // set up sub cell centre coords - could be pre calculated possibly;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ( (xin -(0.5*dx)) + ( 0.5 +  (Real)(i-1) ) * deltax);
    xs[i] = ampl*cos(twopi*f*xs[i]+(phase*pi))+offset;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ( (yin -(0.5*dy)) + ( 0.5 +  (Real)(j-1) ) * deltay) ;
  }
  //
  //  count subcells beneath curve..;
  //
  for ( i=0; i<nzones; i++){
    s = xs[i];
    for ( j=0; j<nzones; j++){
      if (s > ys[j]) fracin = fracin + fzone;
    }
  }
  
  fracin = fracin < 0.0? 0.0: fracin;
  fracin = fracin > 1.0? 1.0: fracin;
  
  //
  //  return fraction of cell in circle;
  //
  return(fracin);
}

Real zls_OverlapLine2D ( Real a, Real b,
                        Real xin, Real yin, Real  dx, Real dy,
                        Integer n ){
  //
  //  overlap of a a*x + b line and a square (xin, yin, dx, dy).;
  //  returns the faction of cell that lies below the line, inside the cell.
  //     
  //   xin and yin are cell centre coordinates;
  //
  //  Divides square into n by n sub cells and counts cells whose centres;
  //  are inside the circle.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;
  //
  //
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  
  Integer i, j;
  
  Integer nzones = n;
  Real dzone     = 1.0/(Real)nzones; // 1/[nzones]
  Real fzone     = 1.0/(Real)(nzones*nzones); // 1/(nzones x nzones)
  
  Real     s;
  Real     xs[nzones];
  Real     ys[nzones];
  Real     fracin;
  Real     deltax;
  Real     deltay;
  //
  //
  //      Assume all outside circle to start, fracin = 0;
  //
  //
  fracin = 0.0;
  //
  //  allow for rectangular cells;
  //
  deltax = dzone*dx;
  deltay = dzone*dy;
  //
  // set up sub cell centre coords - could be pre calculated possibly;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ( (xin -(0.5*dx)) + ( 0.5 +  (Real)(i-1) ) * deltax);
    xs[i] = a*xs[i]+b;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ( (yin -(0.5*dy)) + ( 0.5 +  (Real)(j-1) ) * deltay) ;
  }
  //
  //  count subcells beneath curve..;
  //
  for ( i=0; i<nzones; i++){
    s = xs[i];
    for ( j=0; j<nzones; j++){
      if (s > ys[j]) fracin = fracin + fzone;
    }
  }
  
  fracin = fracin < 0.0? 0.0: fracin;
  fracin = fracin > 1.0? 1.0: fracin;
  
  //
  //  return fraction of cell in circle;
  //
  return(fracin);
}



Real zls_OverlapCircle2D (Real rad,  Real x_0, Real y_0, 
                          Real xin, Real yin, 
                          Real  dx, Real dy,
                          Integer n){
  //
  //  overlap of a circle (rad, x_0, y_0) and a square (xin, yin, dx, dy).;
  //   xin and yin are cell centre coordinates;
  //
  //  3D version below adds z_0, zin and nz etc;
  //
  //  Divides square into n by n sub cells and counts cells whose centres;
  //  are inside the circle.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;
  //
  //
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  
  Integer i, j;
  
  Integer nzones = n;
  Real dzone     = 1.0/nzones; // 1/[nzones]
  Real fzone     = 1.0/(nzones*nzones); // 1/(nzones x nzones)
  
  Real     r, r2;
  Real     xs[nzones];
  Real     ys[nzones];
  Real     fracin;
  Real     deltax;
  Real     deltay;
  //
  //
  //      Assume all outside circle to start, fracin = 0;
  //
  //
  fracin = 0.0;
  r2     = rad*rad; // work in r^2 space;
  //
  //  allow for rectangular cells;
  //
  deltax = dzone*dx;
  deltay = dzone*dy;
  //
  // set up sub cell centre coords - could be pre calculated possibly;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ( (xin -(0.5*dx)) + ( 0.5 +  (Real)(i) ) * deltax) - x_0 ;
    xs[i] = xs[i]*xs[i]  ;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ( (yin -(0.5*dy)) + ( 0.5 +  (Real)(j) ) * deltay) - y_0 ;
    ys[j] = ys[j]*ys[j]  ;
  }
  //
  //  Special Cases
  //
  
  //  1) if all corners are inside, then the whole cube in (convex hull)
  //  2) if all corners are outside by more than a cell width (from dx dy dz), then the whole cube is out
  
  Integer corners1 = 0;
  Integer corners2 = 0;
  Real r3 = rad + (dx + dy)*0.75;  // 50% more than mean edge size
  r3 *= r3; // lol C
  
  
  r = xs[0] + ys[0];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[0] ;
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  
  r = xs[0] + ys[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
    
  if ( corners1 == 4 ) {
    return(1.0);  // all inside
  }
  
  if ( corners2 == 0 ) {
    return(0.0);  // all outside
  }
  
  //
  //  count subcells inside circle..;
  //
  for ( j=0; j<nzones; j++){
    for ( i=0; i<nzones; i++){
      r = xs[i] + ys[j];
      fracin += (r < r2) ? fzone:0.0;
    }
  }
  
  fracin = fracin < 0.0? 0.0: fracin;
  fracin = fracin > 1.0? 1.0: fracin;
  //
  //  return fraction of cell in circle;
  //
  return(fracin);
}


Real zls_OverlapSphere3D (Real  rad, Real  x_0, Real  y_0, Real z_0,
                          Real  xin, Real yin, Real zin, 
                          Real   dx, Real  dy, Real dz,
                          Integer n ){
  //
  //  overlap of a sphere (rad, x_0, y_0, z_0) and a cube (xin, yin, zin, dx, dy, dz).;
  //
  //   *****xin, yin and zin are cell center coordinates;********
  //
  //  Divides cube into n by n by n sub cells and counts cells whose centres;
  //  are inside the sphere.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;
  //
  //
  
  Integer    i, j, k;
  
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  
  Integer nzones = n;
  Real dzone     = 1.0/nzones; // 1/[nzones]
  Real vzone     = dzone*dzone*dzone; // 1/(nzones x nzones x nzones)
  
  Real     r, r2, xs[nzones], ys[nzones], zs[nzones];
  Real     fracin;
  
  Real     deltax;
  Real     deltay;
  Real     deltaz;
  
  //
  //
  //      Assume all outside circle to start, fracin = 0;
  //
  //
  
  fracin = 0.0;
  r2     = rad*rad; // work in r^2 space;
  //
  //  allow for rectangular cells;
  //
  deltax = dzone*dx;
  deltay = dzone*dy;
  deltaz = dzone*dz;
  //
  // set up sub cell centre coords - could be pre calculated possibly;
  //
  
  for( i = 0; i<nzones; i++){
    xs  [i] = ( (xin - dx*0.5) + ( 0.5 +  (Real)(i) ) * deltax) - x_0 ;
    xs  [i] = xs[i]*xs[i]  ;
  }
  
  for ( j=0; j<nzones; j++){
    ys  [j] = ( (yin - dy*0.5)+ ( 0.5 +  (Real)(j) ) * deltay) - y_0 ;
    ys  [j] = ys[j]*ys[j]  ;
  }
  
  for ( k=0; k<nzones; k++){
    zs  [k] = ( (zin- dz*0.5) + ( 0.5 +  (Real)(k) ) * deltaz) - z_0 ;
    zs  [k] = zs[k]*zs[k]  ;
  }
  
  //
  //  Special Cases
  //
  
  //  1) if all corners are inside, then the whole cube in (convex hull)
  //  2) if all corners are outside by more than a cell width (from dx dy dz), then the whole cube is out
  
  Integer corners1 = 0;
  Integer corners2 = 0;
  Real r3 = rad + (dx + dy + dz)*0.5;  // 50% more than mean edge size
  r3 *= r3; // lol C
  
  
  r = xs[0] + ys[0] + zs[0];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[0] + zs[0];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  
  r = xs[0] + ys[nzones-1] + zs[0];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[nzones-1] + zs[0];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  
  r = xs[0] + ys[0] + zs[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[0] + zs[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;

  r = xs[0] + ys[nzones-1] + zs[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  r = xs[nzones-1] + ys[nzones-1] + zs[nzones-1];
  corners1 += (r < r2) ? 1: 0;
  corners2 += (r < r3) ? 1: 0;
  
  if ( corners1 == 8 ) {
    return(1.0);  // all inside
  }
  
  if ( corners2 == 0 ) {
    return(0.0);  // all outside
  }
  
  // 
  // otherwise we have to count all sub cells.
  //
  //  count subcells inside sphere...;
  //
  for ( k=0; k<nzones; k++){
    for ( j=0; j<nzones; j++){
      for ( i=0; i<nzones; i++){
        r = xs[i] + ys[j] + zs[k];
        fracin += (r < r2) ? vzone:0.0;
      }
    }
  }
  
  fracin = fracin < 0.0? 0.0: fracin; // fix bit errors
  fracin = fracin > 1.0? 1.0: fracin;
  //
  //  return fraction of cell in circle;
  //
  return(fracin);
}



Real zls_MeanProjectedRadius2D ( Real * r2, Real * fi, Real * fj,  
                                Real xin, Real yin,
                                Real   dx, Real dy,
                                Integer n ){
  
  Real   meanr  = 0.0;
  Real   meanr2 = 0.0;
  
  if ( n <= 1){
    
    meanr  = zls_CvtCoordsToProjRadius2D( fi, fj, xin, yin);
    meanr2 = (meanr*meanr);
    
  } else {
    
    //
    Integer    i,j ;
    //
    //  set resolution, bigger nzones = slower but more accurate;
    //
    Integer nzones=n;
    Real dzone = 1.0/nzones; // 1/[nzones]
    Real vzone = dzone*dzone; // 1/(nzones x nzoness)
    
    Real     r, xs[nzones], ys[nzones];
    
    Real     deltax;
    Real     deltay;
    //
    //  allow for rectangular cells;
    //
    deltax = dzone*dx;
    deltay = dzone*dy;
    //
    // set up sub cell centre coords - could be pre calculated possibly;
    //
    for( i = 0; i<nzones; i++){
      xs  [i] = ( xin + ( 0.5 +  (Real)(i-1) ) * deltax)  ;
    }
    for ( j=0; j<nzones; j++){
      ys  [j] = ( yin + ( 0.5 +  (Real)(j-1) ) * deltay)  ;
    }
    //
    Real  meanfi  = 0.0;
    Real  meanfj  = 0.0;
    
    for ( j=0; j<nzones; j++){
      for ( i=0; i<nzones; i++){
        r = zls_CvtCoordsToProjRadius2D( fi, fj, xs[i], ys[j]);
        meanfi += *fi;
        meanfj += *fj;
        meanr  += r;
        meanr2 += (r*r);
      }
    }
    
    meanfi *= vzone;
    meanfj *= vzone;
    meanr  *= vzone;
    meanr2 *= vzone;
    
    *fi = meanfi;
    *fj = meanfj;
    
  }
  
  *r2 = meanr2;
  return(meanr);
}


Real zls_MeanProjectedRadius3D ( Real * r2, Real * fi, Real * fj, Real * fk,  
                                Real xin, Real yin, Real zin,
                                Real   dx, Real  dy, Real dz,
                                Integer n ){
  
  
  Real   meanr  = 0.0;
  Real   meanr2 = 0.0;
  
  if ( n <= 1){
    
    meanr  = zls_CvtCoordsToProjRadius3D( fi, fj, fk, xin, yin, zin);
    meanr2 = (meanr*meanr);
    
  } else {
    //
    Integer    i,j,k;
    //
    //  set resolution, bigger nzones = slower but more accurate;
    //
    Integer nzones=n;
    Real dzone = 1.0/nzones; // 1/[nzones]
    Real vzone = dzone*dzone*dzone; // 1/(nzones x nzones x nzones)
    
    Real     r, xs[nzones], ys[nzones], zs[nzones];
    
    Real     deltax;
    Real     deltay;
    Real     deltaz;
    //
    //  allow for rectangular cells;
    //
    deltax = dzone*dx;
    deltay = dzone*dy;
    deltaz = dzone*dz;
    //
    // set up sub cell centre coords - could be pre calculated possibly;
    //
    for( i = 0; i<nzones; i++){
      xs  [i] = ( xin + ( 0.5 +  (Real)(i-1) ) * deltax)  ;
    }
    for ( j=0; j<nzones; j++){
      ys  [j] = ( yin + ( 0.5 +  (Real)(j-1) ) * deltay)  ;
    }
    for ( k=0; k<nzones; k++){
      zs  [k] = ( zin + ( 0.5 +  (Real)(k-1) ) * deltaz)  ;
    }
    //
    Real  meanfi  = 0.0;
    Real  meanfj  = 0.0;
    Real  meanfk  = 0.0;
    Real   meanr  = 0.0;
    Real   meanr2 = 0.0;
    
    for ( k=0; k<nzones; k++){
      for ( j=0; j<nzones; j++){
        for ( i=0; i<nzones; i++){
          r = zls_CvtCoordsToProjRadius3D( fi, fj, fk, xs[i], ys[j], zs[k]);
          meanfi += *fi;
          meanfj += *fj;
          meanfk += *fk;
          meanr  += r;
          meanr2 += (r*r);
        }
      }
    }
    
    meanfi *= vzone;
    meanfj *= vzone;
    meanfk *= vzone;
    meanr  *= vzone;
    meanr2 *= vzone;
    
    *fi = meanfi;
    *fj = meanfj;
    *fk = meanfk;
    
  }
  
  *r2  = meanr2;
  return(meanr);
  
}

Real zls_SmoothSin2D ( Real ampl, Real f, Real phase, Real offset,
                      Real xin, Real yin, Real  dx, Real dy,
                      Integer n, Real hs ){
  Integer i,j;
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  Integer nzones    = n;
  Real   dzone      = 1.0/(Real)n; // 1/[nzones]
  const Real glimit = 3.1622777702; // root 10;
  //
  Real     xs[nzones], ys[nzones], wx[nzones], wy[nzones];
  Real     deltax;
  Real     deltay;
  Real     dhsx;
  Real     dhsy;
  //
  //
  //        norm   = invpi*(nzones/hs)**2;
  //
  //  allow for rectangular cells;
  //
  dhsx = dx*hs ;
  dhsy = dy*hs;
  //
  //  +/- 3.16 x kernel;
  //
  
  deltax = 2.0*glimit*dzone*dhsx;
  deltay = 2.0*glimit*dzone*dhsy;
  
  // 
  // set up sub cell centre coords- circle centred;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ((xin -(glimit*dhsx)) + ( 0.5 +  (Real)(i) ) * deltax);
    wx[i] = exp(-zls_sqr((xs[i]-xin)/dhsx));
    xs[i] = ampl*cos(twopi*f*xs[i]+(phase*pi))+offset;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ((yin -(glimit*dhsy)) + ( 0.5 +  (Real)(j) ) * deltay);
    wy[j] = exp(-zls_sqr((ys[j]-yin)/dhsy));
  }
  
  //
  //   evaluate kernel in gaussian centred coordinates;
  //
  Real norm   = 0.0;
  Real fracin = 0.0;
  for( i = 0; i<nzones; i++){
    Real s = xs[i];
    for( j = 0; j<nzones; j++){
      norm +=  wx[i]*wy[j];
      if (s > ys[j]) 
        fracin += wx[i]*wy[j];
    }
  }
  //
  fracin /= norm;
  fracin = (fracin < 0.0)? 0.0 : fracin;
  fracin = (fracin > 1.0)? 1.0 : fracin;
  
  return( fracin);
  
}


Real zls_SmoothCos2D ( Real ampl, Real f, Real phase, Real offset,
                      Real xin, Real yin, Real  dx, Real dy,
                      Integer n, Real hs ){
  Integer i,j;
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  Integer nzones    = n;
  Real   dzone      = 1.0/(Real)n; // 1/[nzones]
  const Real glimit = 3.1622777702; // root 10;
  //
  Real     xs[nzones], ys[nzones], wx[nzones], wy[nzones];
  Real     deltax;
  Real     deltay;
  Real     dhsx;
  Real     dhsy;
  //
  //
  //        norm   = invpi*(nzones/hs)**2;
  //
  //  allow for rectangular cells;
  //
  dhsx = dx*hs ;
  dhsy = dy*hs;
  //
  //  +/- 3.16 x kernel;
  //
  
  deltax = 2.0*glimit*dzone*dhsx;
  deltay = 2.0*glimit*dzone*dhsy;
  
  // 
  // set up sub cell centre coords- circle centred;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ((xin -(glimit*dhsx)) + ( 0.5 +  (Real)(i) ) * deltax) ;
    wx[i] = exp(-zls_sqr((xs[i]-xin)/dhsx));
    xs[i] = ampl*cos(twopi*f*xs[i]+(phase*pi))+offset;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ((yin -(glimit*dhsy)) + ( 0.5 +  (Real)(j) ) * deltay);
    wy[j] = exp(-zls_sqr((ys[j]-yin)/dhsy));
  }
  
  //
  //   evaluate kernel in gaussian centred coordinates;
  //
  Real norm   = 0.0;
  Real fracin = 0.0;
  for( i = 0; i<nzones; i++){
    Real s = xs[i];
    for( j = 0; j<nzones; j++){
      norm +=  wx[i]*wy[j];
      if (s > ys[j]) 
        fracin += wx[i]*wy[j];
    }
  }
  //
  fracin /= norm;
  fracin = (fracin < 0.0)? 0.0 : fracin;
  fracin = (fracin > 1.0)? 1.0 : fracin;
  
  return( fracin);
  
}


Real zls_SmoothLine2D ( Real a, Real b,
                       Real xin, Real yin, Real  dx, Real dy,
                       Integer n, Real hs ){
  //
  //  overlap of a a*x + b line and a square (xin, yin, dx, dy).;
  //  returns the faction of cell that lies below the line, inside the cell.
  
  Integer i,j;
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  Integer nzones    = n;
  Real   dzone      = 1.0/(Real)n; // 1/[nzones]
  const Real glimit = 3.1622777702; // root 10;
  //
  Real     xs[nzones], ys[nzones], wx[nzones], wy[nzones];
  Real     deltax;
  Real     deltay;
  Real     dhsx;
  Real     dhsy;
  //
  //
  //        norm   = invpi*(nzones/hs)**2;
  //
  //  allow for rectangular cells;
  //
  dhsx = dx*hs ;
  dhsy = dy*hs;
  //
  //  +/- 3.16 x kernel;
  //
  
  deltax = 2.0*glimit*dzone*dhsx;
  deltay = 2.0*glimit*dzone*dhsy;
  
  // 
  // set up sub cell centre coords- circle centred;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ((xin -(glimit*dhsx)) + ( 0.5 +  (Real)(i) ) * deltax) ;
    wx[i] = exp(-zls_sqr((xs[i]-xin)/dhsx));
    xs[i] = a*xs[i]+b;
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ((yin -(glimit*dhsy)) + ( 0.5 +  (Real)(j) ) * deltay);
    wy[j] = exp(-zls_sqr((ys[j]-yin)/dhsy));
  }
  
  //
  //   evaluate kernel in gaussian centred coordinates;
  //
  Real norm   = 0.0;
  Real fracin = 0.0;
  for( i = 0; i<nzones; i++){
    Real s = xs[i];
    for( j = 0; j<nzones; j++){
      norm +=  wx[i]*wy[j];
      if (s > ys[j]) 
        fracin += wx[i]*wy[j];
    }
  }
  //
  fracin /= norm;
  fracin = (fracin < 0.0)? 0.0 : fracin;
  fracin = (fracin > 1.0)? 1.0 : fracin;
  
  return( fracin);
  
}

Real zls_SmoothCircle2D (Real rad, Real  x_0, Real  y_0, 
                         Real xin, Real yin, 
                         Real  dx, Real  dy,  Integer n, Real hs){
  //
  //   smooth a circle (rad, x_0, y_0) and a square (xin, yin, dx, dy).;
  //   xin and yin are cell centre coordinates.  hs is gaussian width in cells;
  //   summing over +/- 3 hs;
  //
  //   Divides square and neighbours into n by n sub cells and weights cells;
  //   by gaussian and sums.  Pretty slow but safe enough,  try to minimise;
  //   calls to this routine. ;)
  //
  //
  Integer i,j;
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  Integer nzones    = n;
  Real   dzone      = 1.0/(Real)n; // 1/[nzones]
  const Real glimit = 3.1622777702; // root 10;
  //
  Real     r, r2;
  Real     xs[nzones], ys[nzones], wx[nzones], wy[nzones];
  Real     deltax;
  Real     deltay;
  Real     dhsx;
  Real     dhsy;
  //
  //
  r2     = rad*rad; // work in r^2 space;
  //        norm   = invpi*(nzones/hs)**2;
  //
  //  allow for rectangular cells;
  //
  dhsx = dx*hs ;
  dhsy = dy*hs;
  //
  //  +/- 3.16 x kernel;
  //
  
  deltax = 2.0*glimit*dzone*dhsx;
  deltay = 2.0*glimit*dzone*dhsy;
  
  // 
  // set up sub cell centre coords- circle centred;
  //
  for( i = 0; i<nzones; i++){
    xs[i] = ((xin -(glimit*dhsx)) + ( 0.5 +(Real)(i)) * deltax) - x_0 ;
    wx[i] = exp(-zls_sqr((xs[i]+x_0-xin)/dhsx));
    xs[i] = xs[i]*xs[i];
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ((yin -(glimit*dhsy)) + ( 0.5 +(Real)(j)) * deltay) - y_0 ;
    wy[j] = exp(-zls_sqr((ys[j]+y_0-yin)/dhsy));
    ys[j] = ys[j]*ys[j];
  }
  //
  //   evaluate kernel in gaussian centred coordinates;
  //
  Real norm   = 0.0;
  Real fracin = 0.0;
  for( j = 0; j<nzones; j++){
    for( i = 0; i<nzones; i++){
      norm +=  wx[i]*wy[j];
      r = xs[i]+ys[j];
      if (r<=r2) {
        fracin += wx[i]*wy[j];
      }
    }
  }
  //
  //  return fraction of cell in circle;
  //
  fracin /= norm;
  fracin = (fracin < 0.0)? 0.0 : fracin;
  fracin = (fracin > 1.0)? 1.0 : fracin;
  
  return( fracin);
}

Real zls_SmoothSphere3D (Real rad, Real x_0,  Real y_0,  Real z_0, 
                         Real xin, Real yin, Real zin, 
                         Real dx,  Real dy,  Real dz, Integer n, Real hs){
  //
  //
  //
  Integer    i,j,k;
  //
  //  set resolution, bigger nzones = slower but more accurate;
  //
  Integer nzones    = n;
  Real   dzone      = 1.0/(Real)n; // 1/[nzones]
  const Real glimit = 3.1622777702; // root 10;
  
  Real     r, r2;
  Real     xs[nzones], ys[nzones], zs[nzones];
  Real     wx[nzones], wy[nzones], wz[nzones];
  Real     deltax;
  Real     deltay;
  Real     deltaz;
  Real     dhsx;
  Real     dhsy;
  Real     dhsz;
  
  r2     = rad*rad; // // work in r^2 space
  //
  //  allow for rectangular grid cells;
  //
  dhsx = dx*hs;
  dhsy = dy*hs;
  dhsz = dz*hs;
  //
  //  +/- 3.16 x kernel;
  //
  
  deltax = 2.0*glimit*dzone*dhsx;
  deltay = 2.0*glimit*dzone*dhsy;
  deltaz = 2.0*glimit*dzone*dhsz;
  
  // 
  // set up sub cell centre coords- circle centred;
  //
  
  for( i = 0; i<nzones; i++){
    xs[i] = ((xin -(glimit*dhsx)) + ( 0.5 +  (Real)(i) ) * deltax) - x_0 ;
    wx[i] = exp(-zls_sqr((xs[i]+x_0-xin)/dhsx));
    xs[i] = xs[i]*xs[i];
  }
  for( j = 0; j<nzones; j++){
    ys[j] = ((yin -(glimit*dhsy)) + ( 0.5 +  (Real)(j) ) * deltay) - y_0 ;
    wy[j] = exp(-zls_sqr((ys[j]+y_0-yin)/dhsy));
    ys[j] = ys[j]*ys[j];
  }
  for( k = 0; k<nzones; k++){
    zs[k] = ((zin -(glimit*dhsz)) + ( 0.5 +  (Real)(k) ) * deltaz) - z_0 ;
    wz[k] = exp(-zls_sqr((zs[k]+z_0-zin)/dhsz));
    zs[k] = zs[k]*zs[k];
  }
  //
  //   evaluate kernel in gaussian centred coordinates;
  //
  Real     fracin = 0.0;
  Real     norm   = 0.0;
  for( k = 0; k<nzones; k++){
    for( j = 0; j<nzones; j++){
      for( i = 0; i<nzones; i++){
        norm += wx[i]*wy[j]*wz[k];
        
        r = xs[i]+ys[j]+zs[k];
        
        if (r<=r2) {
          fracin += wx[i]*wy[j]*wz[k];
        }
        
      }
    }
  }
  //
  //  return fraction of cell in circle;
  //
  fracin /= norm;
  fracin = (fracin < 0.0)? 0.0 : fracin;
  fracin = (fracin > 1.0)? 1.0 : fracin;
  
  return( fracin);
  
}

//
//
//  Some useful Num Rec routines;
//
//

/* numerical recipies random number gens - don't bank with them but they are fine for
 * randomising 2 and 3d fields
 */

float zls_ran0_(SInt32 *idum)
{
  /* System generated locals */
  float ret_val;
  
  /* Local variables */
  static int k;
  
  *idum ^= 123459876;
  k = *idum / 127773;
  *idum = (*idum - k * 127773) * 16807 - k * 2836;
  if (*idum < 0) {
    *idum += 2147483647;
  }
  ret_val = *idum * 4.6566128752457969e-10f;
  *idum ^= 123459876;
  return ret_val;
} /* ran0_ */


float zls_ran1_(SInt32 *idum)
{
  /* Initialized data */
  
  static int iv[32] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  static int iy = 0;
  
  /* System generated locals */
  int i__1;
  float ret_val, r__1;
  
  /* Local variables */
  static int j, k;
  
  if (*idum <= 0 || iy == 0) {
    
    i__1 = -(*idum);
    *idum = zls_max(i__1,1);
    for (j = 40; j >= 1; --j) {
      k = *idum / 127773;
      *idum = (*idum - k * 127773) * 16807 - k * 2836;
      if (*idum < 0) {
        *idum += 2147483647;
      }
      if (j <= 32) {
        iv[j - 1] = *idum;
      }
      
    }
    iy = iv[0];
  }
  k = *idum / 127773;
  *idum = (*idum - k * 127773) * 16807 - k * 2836;
  if (*idum < 0) {
    *idum += 2147483647;
  }
  j = iy / 67108864 + 1;
  iy = iv[j - 1];
  iv[j - 1] = *idum;
  
  r__1 = iy * 4.6566128752457969e-10f;
  ret_val = zls_min(r__1,.99999987999999995f);
  return ret_val;
} 

#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0F/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7F
#define RNMX (1.0F-EPS)

float zls_ran2_(SInt32 *idum) {
  
	SInt32 j;
	SInt32 k;
	static SInt32 idum2=123456789;
	static SInt32 iy=0;
	static SInt32 iv[NTAB];
	float temp;
  
	if (*idum <= 0) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		idum2=(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ1;
			*idum=IA1*(*idum-k*IQ1)-k*IR1;
			if (*idum < 0) *idum += IM1;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ1;
	*idum=IA1*(*idum-k*IQ1)-k*IR1;
	if (*idum < 0) *idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0) idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = *idum;
	if (iy < 1) iy += IMM1;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

int zls_splint( Real *y, Real x, Integer n, Real *xa, Real *ya, Real * y2a){
  int err = noErr;
  
  Integer k,khi,klo;
  Real a,b,h;
  
  klo=1;
  khi=n;
  
  do{
    
    k=(khi+klo)/2;
    
    
    if(xa[k]>x){
      khi=k;
    } else {
      klo=k;
    }
    
    
  } while ( khi-klo > 1) ;
  
  h=xa[khi]-xa[klo];
  
  if (h==0.0){ err = dimErr; return (err);}// 'bad xa input in splint';
  
  a=(xa[khi]-x)/h;
  b=(x-xa[klo])/h;
  
  *y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)*0.166666666666666666;
  
  return (err);
  
}

int zls_mkdir(char *fname) {
  int err = noErr;
  
  err = mkdir( fname, S_IRWXU | S_IRGRP | S_IXGRP);
  
  if ((err != noErr) && (errno == EEXIST)) {
    err = noErr;
  }
  return (err);
}

#ifdef OS_MACOSXG4

/* Table of sin(i degrees)
 * for 0 <= i <= 90
 */
static const double kZLS_sineTable[94] = {
  0.00000000000000000000E0,
  1.74524064372835128194E-2,
  3.48994967025009716460E-2,
  5.23359562429438327221E-2,
  6.97564737441253007760E-2,
  8.71557427476581735581E-2,
  1.04528463267653471400E-1,
  1.21869343405147481113E-1,
  1.39173100960065444112E-1,
  1.56434465040230869010E-1,
  1.73648177666930348852E-1,
  1.90808995376544812405E-1,
  2.07911690817759337102E-1,
  2.24951054343864998051E-1,
  2.41921895599667722560E-1,
  2.58819045102520762349E-1,
  2.75637355816999185650E-1,
  2.92371704722736728097E-1,
  3.09016994374947424102E-1,
  3.25568154457156668714E-1,
  3.42020143325668733044E-1,
  3.58367949545300273484E-1,
  3.74606593415912035415E-1,
  3.90731128489273755062E-1,
  4.06736643075800207754E-1,
  4.22618261740699436187E-1,
  4.38371146789077417453E-1,
  4.53990499739546791560E-1,
  4.69471562785890775959E-1,
  4.84809620246337029075E-1,
  5.00000000000000000000E-1,
  5.15038074910054210082E-1,
  5.29919264233204954047E-1,
  5.44639035015027082224E-1,
  5.59192903470746830160E-1,
  5.73576436351046096108E-1,
  5.87785252292473129169E-1,
  6.01815023152048279918E-1,
  6.15661475325658279669E-1,
  6.29320391049837452706E-1,
  6.42787609686539326323E-1,
  6.56059028990507284782E-1,
  6.69130606358858213826E-1,
  6.81998360062498500442E-1,
  6.94658370458997286656E-1,
  7.07106781186547524401E-1,
  7.19339800338651139356E-1,
  7.31353701619170483288E-1,
  7.43144825477394235015E-1,
  7.54709580222771997943E-1,
  7.66044443118978035202E-1,
  7.77145961456970879980E-1,
  7.88010753606721956694E-1,
  7.98635510047292846284E-1,
  8.09016994374947424102E-1,
  8.19152044288991789684E-1,
  8.29037572555041692006E-1,
  8.38670567945424029638E-1,
  8.48048096156425970386E-1,
  8.57167300702112287465E-1,
  8.66025403784438646764E-1,
  8.74619707139395800285E-1,
  8.82947592858926942032E-1,
  8.91006524188367862360E-1,
  8.98794046299166992782E-1,
  9.06307787036649963243E-1,
  9.13545457642600895502E-1,
  9.20504853452440327397E-1,
  9.27183854566787400806E-1,
  9.33580426497201748990E-1,
  9.39692620785908384054E-1,
  9.45518575599316810348E-1,
  9.51056516295153572116E-1,
  9.56304755963035481339E-1,
  9.61261695938318861916E-1,
  9.65925826289068286750E-1,
  9.70295726275996472306E-1,
  9.74370064785235228540E-1,
  9.78147600733805637929E-1,
  9.81627183447663953497E-1,
  9.84807753012208059367E-1,
  9.87688340595137726190E-1,
  9.90268068741570315084E-1,
  9.92546151641322034980E-1,
  9.94521895368273336923E-1,
  9.96194698091745532295E-1,
  9.97564050259824247613E-1,
  9.98629534754573873784E-1,
  9.99390827019095730006E-1,
  9.99847695156391239157E-1,
  1.00000000000000000000E0,
  9.99847695156391239157E-1,
  9.99390827019095730006E-1,
  9.98629534754573873784E-1,
};

/* Define one of the following to be 1:
 */
#define ACC11 1
#define ACC17 0

void zls_sincos(double x, double *s, double *c) {
  //
  // 20-40% faster than equivalent OSX sin and cos math lib calls on G4
  // 1.8 s / 10^7 calls vs 2.2s for system calls with ACC17 (eps !10^-16)
  //
  register int ix, ix0, ssign, csign, xsign;
  register double y, z, sx, sz, cx, cz;
  
  /* Make argument nonnegative.  */
  xsign = 1;
  if( x < 0.0 ) {
    xsign = -1;
    x = -x;
  }
  
  
  x *= 57.295779513082320876798; // cvt rads to deg RSS
  // minimal range reduction, one circle only  for speed. RSS 
  x = x > 360.0?x-360.0:x;
  
  //ix = (int)floor(x + 0.5);
  ix  = x;
  ix0 = ix; // delay evaluating z until later to improve pipelining, save ix before fiddling with it - RSS
  
  /* Look up the sine and cosine of the integer.
   */
  if( !(ix > 180) )
   {
     ssign = 1;
     csign = 1;
   }  else {
     ssign = -1;
     csign = -1;
     ix -= 180;
   }
  
  if( ix > 90 )  {
    csign = -csign;
    ix = 180 - ix;
  }
  
  sx = kZLS_sineTable[ix];
  sx = ssign < 0?-sx:sx;
  cx = kZLS_sineTable[ 90-ix ];
  cx = csign < 0?-cx:cx;
  
  z = x - (double)ix0;    /* the residual */
  y = z * z;
  
  /* Find sine and cosine
   * of the residual angle between -0.5 and +0.5 degree.
   */
  
  
#if ACC11
  sz = ( -8.86092781698004819918e-7 * y
        + 1.74532925198378577601e-2     ) * z;
  
  cz = 1.0 - ( -3.86631403698859047896e-9 * y
              + 1.52308709893047593702e-4     ) * y;
#endif
  
  
#if ACC17
  sz = ((  1.34959795251974073996e-11 * y
         - 8.86096155697856783296e-7     ) * y
        + 1.74532925199432957214e-2          ) * z;
  
  cz = 1.0 - ((  3.92582397764340914444e-14 * y
               - 3.86632385155548605680e-9     ) * y
              + 1.52308709893354299569e-4          ) * y;
#endif
  
  /* Combine the tabulated part and the calculated part
   * by trigonometry.
   */
  
  y = sx * cz  +  cx * sz;
  if( xsign < 0 ) y = - y;
  
  *s = y; /* sine */
  *c = cx * cz  -  sx * sz; /* cosine */
  
}
#endif OS_MACOSXG4

// private routine to help write a linked grid conf file template
// use at own risk !!!

void CreateLinkedGrid (char * gFile, Integer level, Integer dims, Integer nR, Integer levels,
                       Integer nx, Integer ny, Integer nz,
                       Real *xl, Real *xr, Real *yl, Real *yr, Real *zl, Real *zr,
                       Integer nXZones,Integer nYZones, Integer nZZones,
                       Integer periodicX, Integer periodicY, Integer periodicZ );

void CreateLinkedGrid (char * gFile, Integer level, Integer dims, Integer nR, Integer levels,
                       Integer nx, Integer ny, Integer nz,
                       Real *xl, Real *xr, Real *yl, Real *yr, Real *zl, Real *zr,
                       Integer nXZones,Integer nYZones, Integer nZZones,
                       Integer periodicX, Integer periodicY, Integer periodicZ ){
  
  
#define COMMENTLINE "#######################################"
  
  zFile f = NULL;
  char zfilename[DIRLEN];
  
  Integer i, j, k, direction;
  Integer zonep1, zone, zlink;
  Integer deltai, deltaj, deltak;
  Integer nLinks;
  Integer iLevel;
  
  
#ifdef ZLIB_COMPRESSED_IO
  sprintf(zfilename, "%s.gz", gFile);
#else
  sprintf(zfilename, "%s", gFile);
#endif !ZLIB_COMPRESSED_IO
  
  f = zlsio_open( zfilename, "w");
  
  time_t writeTime = time(NULL);
  zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
  zlsio_ZFPrintf( f, "#\n" );
  zlsio_ZFPrintf( f, "# ZLS Linked Grid Template\n" );
  if ( dims == k1D ){
    zlsio_ZFPrintf( f, "# 1D %2d cells.\n", nx);
    zlsio_ZFPrintf( f, "# in %2d zones.\n", nXZones);
    if (periodicX) zlsio_ZFPrintf( f, "# Periodic x-Axis.\n" );
  }
  if ( dims == k2D ){
    zlsio_ZFPrintf( f, "# 2D %2d x %2d cells.\n", nx, ny);
    zlsio_ZFPrintf( f, "# in %2d x %2d zones.\n", nXZones, nYZones);
    if (periodicX) zlsio_ZFPrintf( f, "# Periodic x-Axis.\n" );
    if (periodicY) zlsio_ZFPrintf( f, "# Periodic y-Axis.\n" );
  }
  if ( dims == k3D ){
    zlsio_ZFPrintf( f, "# 3D %2d x %2d x %2d cells.\n", nx, ny, nz);
    zlsio_ZFPrintf( f, "# in %2d x %2d x %2d zones.\n", nXZones, nYZones, nZZones);
    if (periodicX) zlsio_ZFPrintf( f, "# Periodic x-Axis.\n" );
    if (periodicY) zlsio_ZFPrintf( f, "# Periodic y-Axis.\n" );
    if (periodicZ) zlsio_ZFPrintf( f, "# Periodic z-Axis.\n" );
  }
  if ( levels > 1) {
    zlsio_ZFPrintf( f, "# Multi-Level %d levels \n", levels);
  }
  zlsio_ZFPrintf( f, "# ZLS Linked Grid Template\n" );
  zlsio_ZFPrintf( f, "# Written:\n# %s", ctime(&writeTime));
  zlsio_ZFPrintf( f, "#\n" );
  zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
  zlsio_ZFPrintf( f, "Global_Cell_Scale = 1.0\n" );
  //  if ( levels > 1) nR = nXZones;
  zlsio_ZFPrintf( f, "Refine_Cells    = %d\n", nR);
  if (nR < 5) { // 2 , 3 or 5
    zlsio_ZFPrintf( f, "Boundary_Cells  = %d\n", nR*2);
  } else { // 4 or more up to 8
    zlsio_ZFPrintf( f, "Boundary_Cells  = %d\n", nR);
  }
  zlsio_ZFPrintf( f, "Geometry        = Cartesian\n" );
  zlsio_ZFPrintf( f, "NLevels         = %d\n", levels);
  zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
  
  if ( dims == 1 ){
    
    zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n", level, nXZones);
    if ( levels > 1 ) {
      for ( iLevel = 1; iLevel < levels; iLevel++)
        zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n",iLevel, nXZones);
    }
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level Grid Cells\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", level, nx);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level Grid Coordinates\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xl    = %f\n", level, xl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_xr    = %f\n", level, xr[0]);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xKind = Linear\n", level);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    
    deltai = nx/nXZones;
    zone = 0;
    for ( i = 0; i < nXZones; i++){
      zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", level , zone, i*deltai);
      zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", level , zone, deltai);
      zone++;
      zlsio_ZFPrintf( f, "\n" );
    }
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Links \n" );
    
    nLinks = (nXZones-1);
    if (periodicX) nLinks += 1;
    zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n",level, nLinks);
    
    zlink = 0;        
    
    // rows
    zlsio_ZFPrintf( f, "# Rows \n" );
    zone = 0; zonep1 = 0; zlink = 0;
    
    for ( i = 0; i < nXZones-1; i++){
      zone   =  i   ;
      zonep1 = (i+1);
      direction = kDirRight;
      zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
      zlink++;
    }
    if (periodicX) {
      i = nXZones-1;
      zone   =  i   ;
      zonep1 =  0   ;
      direction = kDirRight;
      zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
      zlink++;
    }
    zlsio_ZFPrintf( f, "\n" );
    
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Remaining nonLink BCs \n" );
    if ( !periodicX ) {
      
      zone = 0;
      i    = 0; // first column
      zone = i;
      direction = kDirLeft;
      zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
      zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      zlsio_ZFPrintf( f, "\n" );
      i    = (nXZones-1); // last column
      zone = i;
      direction = kDirRight;
      zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
      zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      zlsio_ZFPrintf( f, "\n" );
    }
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    
    if ( levels > 1 ) {
      for ( iLevel = 1; iLevel < levels; iLevel++){
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "# Level %2.2d Grid Cells on super grid\n", iLevel);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        i = 0 ;
        
        deltai = (nx/nXZones);
        
        zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", iLevel, i*deltai); // sub level covers the refinedZone0 on the super grid
        zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", iLevel, deltai);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        deltai /= nXZones;
        zone = 0;
        for ( i = 0; i < nXZones; i++){
          zone   =  i ;
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_sid = %2.2d:%2.2d\n", iLevel , zone, iLevel-1, 0 );
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", iLevel , zone, i*deltai);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", iLevel , zone, deltai);
          zlsio_ZFPrintf( f, "#\n" );
        }
        
        nLinks = (nXZones-1);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n",iLevel, nLinks);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        zlsio_ZFPrintf( f, "#\n" );
        zone = 0; zonep1 = 0; zlink = 0;
        for ( i = 0; i < nXZones-1; i++){
          zone   =  i   ;
          zonep1 = (i+1);
          direction = kDirRight;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
          zlink++;
        }
        zlsio_ZFPrintf( f, "#\n" );
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
      } // iLevel loop
    } //levels > 1
    
  } // 1D
  
  if ( dims == 2 ){
    
    zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n",level, nXZones*nYZones);
    if ( levels > 1 ) {
      for ( iLevel = 1; iLevel < levels; iLevel++)
        zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n",iLevel, nXZones*nYZones);
    }
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level Grid Cells\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_j0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", level, nx);
    zlsio_ZFPrintf( f, "Level_%2.2d_nj    = %d\n", level, ny);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level Grid Coordinates\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xl    = %f\n", level, xl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_xr    = %f\n", level, xr[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_yl    = %f\n", level, yl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_yr    = %f\n", level, yr[0]);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xKind = Linear\n", level);
    zlsio_ZFPrintf( f, "Level_%2.2d_yKind = Linear\n", level);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    deltai = nx/nXZones;
    deltaj = ny/nYZones;
    zone = 0;
    for ( j = 0; j < nYZones; j++){
      zlsio_ZFPrintf( f, "# Row %d\n", j);
      for ( i = 0; i < nXZones; i++){
        zone   =  i + nXZones*j;
        zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", level , zone, i*deltai);
        zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_j0 = %d\n", level , zone, j*deltaj);
        zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", level , zone, deltai);
        zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nj = %d\n", level , zone, deltaj);
        zlsio_ZFPrintf( f, "#\n" );
      }
    }
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    nLinks = (2*nXZones*nYZones - nYZones - nXZones);
    if (periodicX)  nLinks += nYZones;
    if (periodicY)  nLinks += nXZones;
    zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n",level, nLinks);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    // rows
    zlsio_ZFPrintf( f, "# Rows \n" );
    zone = 0; zonep1 = 0; zlink = 0;
    for ( j = 0; j < nYZones; j++){
      for ( i = 0; i < nXZones-1; i++){
        zone   =  i    + nXZones*j;
        zonep1 = (i+1) + nXZones*j;
        direction = kDirRight;
        zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
        zlink++;
      }
      if (periodicX) {
        i = nXZones-1;
        zone   =  i    + nXZones*j;
        zonep1 =  0    + nXZones*j;
        direction = kDirRight;
        zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
        zlink++;
      }
      zlsio_ZFPrintf( f, "\n" );
    }
    
    // columns
    zlsio_ZFPrintf( f, "# Columns \n" );
    for ( i = 0; i < nXZones; i++){
      for ( j = 0; j < nYZones-1; j++){
        zone   =  i    + nXZones*j ;
        zonep1 =  i    + nXZones*(j+1);
        direction = kDirUp;
        zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
        zlink++;
      }
      if (periodicY) {
        j = nYZones-1;
        zone   =  i    + nXZones*j;
        zonep1 =  i    + nXZones*0;
        direction = kDirUp;
        zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
        zlink++;
      }
      zlsio_ZFPrintf( f, "\n" );
    }
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Remaining nonLink BCs \n" );
    
    
    if ( !periodicX ) {
      
      zone = 0;
      i    = 0; // inner i boundary
      for ( j = 0; j < nYZones; j++){
        zone   =  i    + nXZones*j;
        direction = kDirLeft;
        zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
        zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      }
      zlsio_ZFPrintf( f, "\n" );
      
      i    = (nXZones-1); // outer i boundary
      for ( j = 0; j < nYZones; j++){
        zone   =  i    + nXZones*j;
        direction = kDirRight;
        zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
        zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      }
      
      zlsio_ZFPrintf( f, "\n" );
    }
    
    if ( !periodicY ) {
      
      zone = 0;
      
      j    = 0; // inner j boundary
      for ( i = 0; i < nXZones; i++){
        zone   =  i    + nXZones*j;
        direction = kDirDown;
        zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
        zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      }
      
      zlsio_ZFPrintf( f, "\n" );
      
      j    = (nYZones-1); // outer j boundary
      for ( i = 0; i < nXZones; i++){
        zone   =  i    + nXZones*j;
        direction = kDirUp;
        zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
        zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
      }
      
      zlsio_ZFPrintf( f, "\n" );
    }
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    if ( levels > 1 ) {
      for ( iLevel = 1; iLevel < levels; iLevel++){
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "# Level %2.2d Grid Cells on super grid\n", iLevel);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        i = 0;
        j = 0;
        
        deltai = (nx/nXZones);
        deltaj = (ny/nYZones);
        
        zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", iLevel, i*deltai); // sub level covers the refinedZone0 on the super grid
        zlsio_ZFPrintf( f, "Level_%2.2d_j0    = %d\n", iLevel, j*deltaj);
        zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", iLevel, deltai);
        zlsio_ZFPrintf( f, "Level_%2.2d_nj    = %d\n", iLevel, deltaj);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        deltai /= nXZones;
        deltaj /= nYZones;
        zone = 0;
        for ( j = 0; j < nYZones; j++){
          zlsio_ZFPrintf( f, "# Row %d\n", j);
          for ( i = 0; i < nXZones; i++){
            zone   =  i + nXZones*j;
            zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_sid = %2.2d:%2.2d\n", iLevel , zone, iLevel-1, 0 );
            zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", iLevel , zone, i*deltai);
            zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_j0 = %d\n", iLevel , zone, j*deltaj);
            zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", iLevel , zone, deltai);
            zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nj = %d\n", iLevel , zone, deltaj);
            zlsio_ZFPrintf( f, "#\n" );
          }
        }
        
        nLinks = (2*nXZones*nYZones - nYZones - nXZones);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n",iLevel, nLinks);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        // rows
        zlsio_ZFPrintf( f, "# Rows \n" );
        zone = 0; zonep1 = 0; zlink = 0;
        for ( j = 0; j < nYZones; j++){
          for ( i = 0; i < nXZones-1; i++){
            zone   =  i    + nXZones*j;
            zonep1 = (i+1) + nXZones*j;
            direction = kDirRight;
            zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
            zlink++;
          }
          zlsio_ZFPrintf( f, "\n" );
        }
        
        // columns
        zlsio_ZFPrintf( f, "# Columns \n" );
        for ( i = 0; i < nXZones; i++){
          for ( j = 0; j < nYZones-1; j++){
            zone   =  i    + nXZones*j ;
            zonep1 =  i    + nXZones*(j+1);
            direction = kDirUp;
            zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
            zlink++;
          }
          zlsio_ZFPrintf( f, "\n" );
        }
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
      } // iLevel loop
    } //levels > 1
  } // 2D
  
  if ( dims == 3 ){
    zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n",level, nXZones*nYZones*nZZones);
    if ( levels > 1 ) {
      for ( iLevel = 1; iLevel < levels; iLevel++)
        zlsio_ZFPrintf( f, "Level_%2.2d_Zones = %d\n",iLevel, nXZones*nYZones*nZZones);
    }
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level 00 Grid Cells\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_j0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_k0    = %d\n", level, 0);
    zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", level, nx);
    zlsio_ZFPrintf( f, "Level_%2.2d_nj    = %d\n", level, ny);
    zlsio_ZFPrintf( f, "Level_%2.2d_nk    = %d\n", level, nz);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Level Grid Coordinates\n" );
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xl    = %f\n", level, xl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_xr    = %f\n", level, xr[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_yl    = %f\n", level, yl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_yr    = %f\n", level, yr[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_zl    = %f\n", level, zl[0]);
    zlsio_ZFPrintf( f, "Level_%2.2d_zr    = %f\n", level, zr[0]);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "Level_%2.2d_xKind = Linear\n", level);
    zlsio_ZFPrintf( f, "Level_%2.2d_yKind = Linear\n", level);
    zlsio_ZFPrintf( f, "Level_%2.2d_zKind = Linear\n", level);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    deltai = nx/nXZones;
    deltaj = ny/nYZones;
    deltak = nz/nZZones;
    zone = 0;
    for ( k = 0; k < nZZones; k++){
      zlsio_ZFPrintf( f, "# Plane %d\n", k);
      for ( j = 0; j < nYZones; j++){
        zlsio_ZFPrintf( f, "# Row %d\n", j);
        for ( i = 0; i < nXZones; i++){
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", level , zone, i*deltai);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_j0 = %d\n", level , zone, j*deltaj);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_k0 = %d\n", level , zone, k*deltak);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", level , zone, deltai);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nj = %d\n", level , zone, deltaj);
          zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nk = %d\n", level , zone, deltak);
          zone++;
          zlsio_ZFPrintf( f, "\n" );
        }
      }
    }
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    nLinks = ((nXZones-1)*(nYZones*nZZones) + (nXZones*(nYZones-1)*nZZones) + (nXZones*nYZones)*(nZZones-1));
    if (periodicX)  nLinks += nYZones*nZZones;
    if (periodicY)  nLinks += nXZones*nZZones;
    if (periodicZ)  nLinks += nXZones*nYZones;
    
    zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n",level, nLinks);
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    // rows
    zlsio_ZFPrintf( f, "# Rows \n" );
    zone = 0; zonep1 = 0; zlink = 0;
    for ( k = 0; k < nZZones; k++){
      for ( j = 0; j < nYZones; j++){
        for ( i = 0; i < nXZones-1; i++){
          zone   =  i    + nXZones*(j+ nYZones*k);
          zonep1 = (i+1) + nXZones*(j+ nYZones*k);
          direction = kDirRight;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
        if (periodicX) {
          i = nXZones-1;
          zone   =  i    + nXZones*(j+ nYZones*k);
          zonep1 =         nXZones*(j+ nYZones*k);
          direction = kDirRight;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
      }
    }
    zlsio_ZFPrintf( f, "\n" );
    
    // columns
    zlsio_ZFPrintf( f, "# Columns \n" );
    for ( k = 0; k < nZZones; k++){
      for ( i = 0; i < nXZones; i++){
        for ( j = 0; j < nYZones-1; j++){
          zone   =  i    + nXZones*( j     + nYZones*k);
          zonep1 =  i    + nXZones*( (j+1) + nYZones*k);
          direction = kDirUp;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
        if (periodicY) {
          j = nYZones-1;
          zone   =  i    + nXZones*( j     + nYZones*k);
          zonep1 =  i    + nXZones*( 0     + nYZones*k);
          direction = kDirUp;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
      }
    }
    zlsio_ZFPrintf( f, "\n" );
    
    
    // planes
    zlsio_ZFPrintf( f, "# Planes \n" );
    for ( j = 0; j < nYZones; j++){
      for ( i = 0; i < nXZones; i++){
        for ( k = 0; k < nZZones-1; k++){
          zone   =  i    + nXZones*( j + nYZones*k);
          zonep1 =  i    + nXZones*( j + nYZones*(k+1));
          direction = kDirHither;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
        if (periodicZ) {
          k = nZZones-1;
          zone   =  i    + nXZones*( j + nYZones*k);
          zonep1 =  i    + nXZones*( j + nYZones*(0));
          direction = kDirHither;
          zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, level, zone, zls_dirname(direction), level, zonep1);
          zlink++;
        }
      }
    }
    zlsio_ZFPrintf( f, "\n" );
    
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    zlsio_ZFPrintf( f, "# Remaining nonLink BCs \n" );
    
    if ( !periodicX ) {
      
      zone = 0;
      i    = 0; // inner i boundary
      for ( k = 0; k < nZZones; k++){
        for ( j = 0; j < nYZones; j++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirLeft;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
      i    = (nXZones-1); // outer i boundary
      for ( k = 0; k < nZZones; k++){
        for ( j = 0; j < nYZones; j++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirRight;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
    }
    
    if ( !periodicY ) {
      
      zone = 0;
      j    = 0; // inner j boundary
      for ( k = 0; k < nZZones; k++){
        for ( i = 0; i < nXZones; i++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirDown;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
      j    = (nYZones-1); // outer j boundary
      for ( k = 0; k < nZZones; k++){
        for ( i = 0; i < nXZones; i++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirUp;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
    }
    
    if ( !periodicZ ) {
      
      zone = 0;
      k    = 0; // inner k boundary
      for ( j = 0; j < nYZones; j++){
        for ( i = 0; i < nXZones; i++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirYon;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
      k    = (nZZones-1); // outer k boundary
      for ( j = 0; j < nYZones; j++){
        for ( i = 0; i < nXZones; i++){
          zone   =  i    + nXZones*( j + nYZones*k);
          direction = kDirHither;
          zlsio_ZFPrintf( f, "BCType_Zone_%2.2d:%2.2d_Dir_%s = Computed\n", level, zone, zls_dirname(direction));
          zlsio_ZFPrintf( f, "BCKind_Zone_%2.2d:%2.2d_Dir_%s = Free\n", level, zone, zls_dirname(direction));
        }
      }
      zlsio_ZFPrintf( f, "\n" );
    }
    
    
    zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
    
    if ( levels > 1 ) {
      
      Integer ks, js, is;  // super zone coords
      Integer k0, j0, i0;
      //      Integer nk, nj, ni;
      
      Integer nSupXZoneCells = nx/nXZones;
      Integer nSupYZoneCells = ny/nYZones;
      Integer nSupZZoneCells = nz/nZZones;
      
      Integer nXZoneCells    = nSupXZoneCells/nR;
      Integer nYZoneCells    = nSupYZoneCells/nR;
      Integer nZZoneCells    = nSupZZoneCells/nR;
      
      Integer sup_ni[nXZones];
      Integer sup_nj[nYZones];
      Integer sup_nk[nZZones];
      
      Integer n_ni[nXZones];
      Integer n_nj[nYZones];
      Integer n_nk[nZZones];
      
      Integer d_ni[nXZones];
      Integer d_nj[nYZones];
      Integer d_nk[nZZones];
      
      Integer sup_zone ; // super zone id
      
      for ( iLevel = 1; iLevel < levels; iLevel++){
        
        
        Integer deltaX0 =  (xr[iLevel-1]-xl[iLevel-1])/nXZones;
        Integer deltaX1 =  (xr[iLevel]-xl[iLevel])/nXZones;
        
        for ( i = 0; i < nXZones; i++){
          sup_ni[i] = ( xl[iLevel]+(i*deltaX1) - xl[iLevel-1])/deltaX0;
          n_ni[i]   = ((xl[iLevel]+(i*deltaX1))-(xl[iLevel-1]+sup_ni[i]*deltaX0))/deltaX1;
          d_ni[i]   = nR;
        }
        
        Integer deltaY0 =  (yr[iLevel-1]-yl[iLevel-1])/nYZones;
        Integer deltaY1 =  (yr[iLevel]-yl[iLevel])/nYZones;
        
        for ( j = 0; j < nYZones; j++){
          sup_nj[j] =  (yl[iLevel]+(j*deltaY1) - yl[iLevel-1])/deltaY0;
          n_nj[j]   = ((yl[iLevel]+(j*deltaY1))-(yl[iLevel-1]+sup_nj[j]*deltaY0))/deltaY1;
          d_nj[j]   = nR;
        }   
        
        Integer deltaZ0 =  (zr[iLevel-1]-zl[iLevel-1])/nZZones;
        Integer deltaZ1 =  (zr[iLevel]  -zl[iLevel]  )/nZZones;
        
        for ( k = 0; k < nZZones; k++){
          sup_nk[k] =  (zl[iLevel]+(k*deltaZ1) - zl[iLevel-1])/deltaZ0;
          n_nk[k]   = ((zl[iLevel]+(k*deltaZ1))-(zl[iLevel-1]+sup_nk[k]*deltaZ0))/deltaZ1;
          d_nk[k]   = nR;
        }   
        
        i0 =  sup_ni[0]*nSupXZoneCells+(n_ni[0]*nXZoneCells);
        j0 =  sup_nj[0]*nSupYZoneCells+(n_nj[0]*nYZoneCells);
        k0 =  sup_nk[0]*nSupZZoneCells+(n_nk[0]*nZZoneCells);
        
        //    for ( i = 0; i < nXZones; i++) zlsio_ZFPrintf( f, "%d " , sup_ni[i] );
        //    zlsio_ZFPrintf( f, "\n" );
        //    for ( j = 0; j < nYZones; j++) zlsio_ZFPrintf( f, "%d " , sup_nj[j] );
        //    zlsio_ZFPrintf( f, "\n" );
        //    for ( k = 0; k < nZZones; k++) zlsio_ZFPrintf( f, "%d " , sup_nk[k] );
        //    zlsio_ZFPrintf( f, "\n" );
        //
        
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "# Level %2.2d cells on level %2.2d\n", iLevel, iLevel-1);
        zlsio_ZFPrintf( f, "# \n");
        zlsio_ZFPrintf( f, "# Level gets all axis types and  \n");
        zlsio_ZFPrintf( f, "# coords from super-level \n");
        zlsio_ZFPrintf( f, "# \n");
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        zlsio_ZFPrintf( f, "Level_%2.2d_i0    = %d\n", 1, i0); 
        zlsio_ZFPrintf( f, "Level_%2.2d_j0    = %d\n", 1, j0); 
        zlsio_ZFPrintf( f, "Level_%2.2d_k0    = %d\n", 1, k0); 
        zlsio_ZFPrintf( f, "Level_%2.2d_ni    = %d\n", 1, nXZoneCells*nXZones);
        zlsio_ZFPrintf( f, "Level_%2.2d_nj    = %d\n", 1, nYZoneCells*nYZones);
        zlsio_ZFPrintf( f, "Level_%2.2d_nk    = %d\n", 1, nZZoneCells*nZZones);
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        for ( k = 0; k < nZZones; k++){
          for ( j = 0; j < nYZones; j++){
            for ( i = 0; i < nZZones; i++){
              
              zone = i + (nXZones*(j + nYZones*k));
              
              is = sup_ni[i]; js = sup_nj[j]; ks =  sup_nk[k];
              sup_zone  = is + (nXZones*(js + nYZones*ks));
              
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_sid = %2.2d:%2.2d\n", iLevel , zone, iLevel-1, sup_zone );
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_i0 = %d\n", iLevel , zone, n_ni[i]*nSupXZoneCells/d_ni[i]);
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_j0 = %d\n", iLevel , zone, n_nj[j]*nSupYZoneCells/d_nj[j]);
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_k0 = %d\n", iLevel , zone, n_nk[k]*nSupZZoneCells/d_nk[k]);
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_ni = %d\n", iLevel , zone, nXZoneCells);
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nj = %d\n", iLevel , zone, nYZoneCells);
              zlsio_ZFPrintf( f, "Zone_%2.2d:%2.2d_nk = %d\n", iLevel , zone, nZZoneCells);
              
              zlsio_ZFPrintf( f, "#\n" );
              
            }
          }
        }
        
        Integer nLinks   ;
        Integer zonep1   ;
        Integer direction;
        Integer zlink;
        
        nLinks = ((nXZones-1)*(nYZones*nZZones) + (nXZones*(nYZones-1)*nZZones) + (nXZones*nYZones)*(nZZones-1));
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        zlsio_ZFPrintf( f, "Level_%2.2d_Links = %d\n", iLevel, nLinks);
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
        // rows
        zlsio_ZFPrintf( f, "# Rows \n" );
        zone = 0; zonep1 = 0; zlink = 0;
        for ( k = 0; k < nZZones; k++){
          for ( j = 0; j < nYZones; j++){
            for ( i = 0; i < nXZones-1; i++){
              zone   =  i    + nXZones*(j+ nYZones*k);
              zonep1 = (i+1) + nXZones*(j+ nYZones*k);
              direction = kDirRight;
              zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
              zlink++;
            }
          }
        }
        zlsio_ZFPrintf( f, "#\n" );
        
        // columns
        zlsio_ZFPrintf( f, "# Columns \n" );
        for ( k = 0; k < nZZones; k++){
          for ( i = 0; i < nXZones; i++){
            for ( j = 0; j < nYZones-1; j++){
              zone   =  i    + nXZones*( j     + nYZones*k);
              zonep1 =  i    + nXZones*( (j+1) + nYZones*k);
              direction = kDirUp;
              zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
              zlink++;
            }
          }
        }
        zlsio_ZFPrintf( f, "#\n" );
        
        // planes
        zlsio_ZFPrintf( f, "# Planes \n" );
        for ( j = 0; j < nYZones; j++){
          for ( i = 0; i < nXZones; i++){
            for ( k = 0; k < nZZones-1; k++){
              zone   =  i    + nXZones*( j + nYZones*k);
              zonep1 =  i    + nXZones*( j + nYZones*(k+1));
              direction = kDirHither;
              zlsio_ZFPrintf( f, "Link_%2.2d_Zone_%2.2d:%2.2d_Dir_%s = %2.2d:%2.2d\n",zlink, iLevel, zone, zls_dirname(direction), iLevel, zonep1);
              zlink++;
            }
          }
        }
        zlsio_ZFPrintf( f, "#\n" );
        
        zlsio_ZFPrintf( f, "%s\n" , COMMENTLINE );
        
      }// iLevel loop
    } //levels > 1
    
  } // 3D
  
  if (( dims < 1 ) && ( dims > 3 )){
    fprintf( stderr, " ERROR: CreateLinkedGrid Dims Err: %d \n", dims);
  }
  zlsio_close(f);
  
}

// private utility routine to make tiffs of CTAB collection
// hard wired list, will need editing to work, not intended for 
// public use, so no zls_ namespace

void CreateTiffGradients (char * inDir, char * outDir);

void CreateTiffGradients (char * inDir, char * outDir){
#define CTAB "01.rss.ctab"
  int err = noErr;
  
  char tiffFileName [PATHLEN];
  char CTABFile [PATHLEN];
  TIFFCTab * outTIFFCTab;
  int nxOut = 256;
  int nyOut = 32;
  
  unsigned char  *  image8 = NULL;
  
  int i, j;
  
  image8 = (unsigned char *) malloc(nxOut*nyOut);
  outTIFFCTab = (TIFFCTab *) malloc (sizeof(TIFFCTab));
  
  for(j = 0;j<nyOut;j++){
    for(i = 0;i<nxOut;i++){
      image8[i + nxOut*j] = i;
    }//j
  }//i
  
  strcpy(CTABFile , "37.contour");
  sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
  err = zls_ReadCtab(tiffFileName, outTIFFCTab);
  sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
  zls_Write_tif8_NC (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
  
  strcpy(CTABFile , "40.bbyw");
  sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
  err = zls_ReadCtab(tiffFileName, outTIFFCTab);
  sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
  zls_Write_tif8_NC (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
  
  strcpy(CTABFile , "41.contour2");
  sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
  err = zls_ReadCtab(tiffFileName, outTIFFCTab);
  sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
  zls_Write_tif8_NC (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
  /*  
   strcpy(CTABFile , "01.rss");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "02.gray");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "03.gray_inverse");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "04.grey_mod");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "08.blue");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "09.green");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "10.red");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "11.blackbody");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "12.idlstd");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "13.ps");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "14.stern2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "15.ag1");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "16.ag2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "17.ag3");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "18.bluegold");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "19.goldblue");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "20.r1");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "21.r2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "22.r3");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "23.r4");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "24.p1");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "25.p2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "26.p3");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "27.tim");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "28.air");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "29.water");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "30.earth");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "31.123");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "32.barb");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "33.barb2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "34.blob1");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "35.blob2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "36.carl");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "37.ag4");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "38.blackbody");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "39.bright");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "40.contour");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "41.diffuse");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "42.lograinbow");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "43.magenta");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "44.ps");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "45.purple");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "46.rainbow");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "47.subdued");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "48.subdued2");
   sprintf(tiffFileName, "%s%s%s%s", inDir, PATHSEP, CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "%s%s%s%s", outDir, PATHSEP, CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "001_b-wlinear");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "002_blue_white");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "003_grn-red-blu-wht");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "004_redtemperature");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "005_blu_grn_red_ylw");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "006_stdgamma-ii");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "007_prism");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "008_red-purple");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "009_green_whtelin");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "010_grn_whtexp");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "011_green-pink");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "012_blue-red");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "013_16level");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "014_rainbow");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "015_steps");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "016_sternspecial");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "017_haze");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "018_blue-pastel-red");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "019_pastels");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "020_huesatlightness1");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "021_huesatlightness2");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "022_huesatvalue1");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "023_huesatvalue2");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "024_prpl-red_stripes");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "025_beach");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "026_macstyle");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "027_eosa");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "028_eosb");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "029_hardcandy");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "030_nature");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "031_ocean");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "032_peppermint");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "033_plasma");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "034_blue-red");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "035_rainbow");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "036_bluewaves");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "037_volcano");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "038_waves");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "039_rainbow18");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "040_rainbow_white");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   
   strcpy(CTABFile , "041_rainbow_black");
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, CTABEXT);
   err = zls_ReadCtab(tiffFileName, outTIFFCTab);
   sprintf(tiffFileName, "ctabs/idlcolors/%s%s", CTABFile, TIFFEXT);
   zls_Write_tif8 (tiffFileName, image8, outTIFFCTab, nxOut, nyOut );
   */  
  
  if (image8!=NULL)  free(image8) ; image8 = NULL; 
  if (outTIFFCTab!=NULL)  free(outTIFFCTab) ; outTIFFCTab = NULL; 
  
}

