#include "zls.h"


#ifndef SCALAR_LRQS
#ifndef VECTOR_LRQS
#define SCALAR_LRQS
#endif
#endif
//
//  Allocating/Releasing
//

//========================================
//  interpolation coefficients
//========================================
//
// just use the dimension and integer parts of r, cast for differnt dim Interp
//

int zls_NewInterp            (InPtr *f, RangePtr r){
#pragma unused(f, r)
  int err = nyiErr;
  
  return(err);
}

int zls_NewInterp1D          (InPtr *f, Counter ni){
  int err = noErr;
  
  *f = (InPtr) malloc(ni*sizeof(Interp));
  if (*f==NULL){
  err = memErr;}
  
  return(err);
}

//plain blocks
int zls_NewInterp2D          (InPtr *f, Counter ni, Counter nj){
  int err = noErr;
  
  *f = (InPtr) malloc(ni*nj*sizeof(Interp));
  if (*f==NULL){
  err = memErr;}
  
  return(err);
}

int zls_NewInterp3D          (InPtr *f, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  
  *f = (InPtr) malloc(ni*nj*nk*sizeof(Interp));
  if (*f==NULL){
  err = memErr;}
  
  return(err);
}


int zls_DisposeInterp        (InPtr f){
  int err = noErr;
  
  if (f!=NULL){
    free(f);
  } else { err = memErr;}
  
  return(err);  
}


// multidim arrs
int zls_NewInterp2DArr       (In2DArr *f, Counter nj, Counter ni){
  
  int err = noErr;
  
  *f = (In2DArr)zlsAlloc2DArr(sizeof(Interp), nj, ni);
  
  if (*f == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewInterp3DArr       (In3DArr *f, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *f = (In3DArr)zlsAlloc3DArr(sizeof(Interp),nk, nj, ni);
  
  if (*f == NULL) {
    err = memErr;
  }
  
  return(err);
}


int zls_DisposeInterp2DArr   (In2DArr f){
  int err = noErr;
  
  if (f != NULL) {
    zlsFree2DArr((void **) f);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeInterp3DArr   (In3DArr f){
  int err = noErr;
  
  if (f != NULL) {
    zlsFree3DArr((void ***) f);
  } else {err = memErr;}
  
  return(err);
}


//
//========================================
//
// Variable = quantity + coefficients - elements change with SYM flag dimension,
// Input ns should include boundary cells if needed, ready for sweeps
//
//========================================
//
//
// Single quantity variables:
//
//========================================

int zls_NewVar1DArr(V1DArr *v, Integer nBC, Counter ni){
  
  int err = noErr;
  
  v->n = ni;
  v->ns = nBC;
  v->ne = nBC+ni;
  v->nt = nBC+nBC+ni;
  
  v->x = NULL;  // get an axis from an owning grid
  
  v->v = (Var1 *) malloc((nBC+nBC+ni)*sizeof(Var1));
  
  if (v->v == NULL) {
    err = memErr;
  }
  
  return(err);
  
}

int zls_NewVar2DArr(V2DArr *v, Integer nBC, Counter nj, Counter ni){
  
  int err = noErr;
  
  v->n[kXDim] = ni;
  v->ns[kXDim] = nBC;
  v->ne[kXDim] = nBC+ni;
  v->nt[kXDim] = nBC+nBC+ni;
  v->x = NULL;  // get an axis from an owning grid
  
  v->n[kYDim] = nj;
  v->ns[kYDim] = nBC;
  v->ne[kYDim] = nBC+nj;
  v->nt[kYDim] = nBC+nBC+nj;
  v->y = NULL;  // get an axis from an owning grid
  
  v->v = (Var2**) zlsAlloc2DArr(sizeof(Var2), (Counter)v->nt[kYDim], (Counter)v->nt[kXDim]);
  
  if (v->v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVar3DArr(V3DArr *v, Integer nBC, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  v->n[kXDim] = ni;
  v->ns[kXDim] = nBC;
  v->ne[kXDim] = nBC+ni;
  v->nt[kXDim] = nBC+nBC+ni;
  v->x = NULL;  // get an axis from an owning grid
  v->n[kYDim] = nj;
  v->ns[kYDim] = nBC;
  v->ne[kYDim] = nBC+nj;
  v->nt[kYDim] = nBC+nBC+nj;
  v->y = NULL;  // get an axis from an owning grid
  v->n[kZDim] = nk;
  v->ns[kZDim] = nBC;
  v->ne[kZDim] = nBC+nk;
  v->nt[kZDim] = nBC+nBC+nk;
  v->z = NULL;  // get an axis from an owning grid
  
  v->v = (Var3***) zlsAlloc3DArr(sizeof(Var3), (Counter)v->nt[kZDim], (Counter)v->nt[kYDim], (Counter)v->nt[kXDim]);
  
  if (v->v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeVar1DArr  (V1DArr v){
  int err = noErr;
  if (v.v !=NULL) free(v.v); v.v = NULL;
  return(err);
}

int zls_DisposeVar2DArr  (V2DArr v){
  int err = noErr;
  if (v.v !=NULL) zlsFree2DArr((void **)v.v); v.v = NULL;
  return(err);
}

int zls_DisposeVar3DArr  (V3DArr v){
  int err = noErr;
  if (v.v !=NULL) zlsFree3DArr((void ***)v.v); v.v = NULL;
  return(err);
}

int zls_SetVar1DArrQ1DArr(V1DArr v, Q1DArr q){
  int err = noErr;
  
  Integer     i;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    for (i = v.ns; i<v.ne; i++){
      
      a = q[i];
      
      v.v[i].a      = a;
#ifndef GODUNOV        
      v.v[i].al    = a;
      v.v[i].ar    = a;
#ifdef PPM
      v.v[i].del   = 0.0;
      v.v[i].a6    = 0.0;
#endif PPM
#endif GODUNOV        
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar2DArrQ2DArr(V2DArr v, Q2DArr q){
  int err = noErr;
  Integer i, j;
  Integer ni, nj;
  Real      a;
  
  if ((v.v!=NULL)&&(q!=NULL)) {
    
    ni = v.nt[kXDim];
    nj = v.nt[kYDim];
    
    for (j = 0; j<nj; j++){
      
      for (i = 0; i<ni; i++){
        
        a = q[j][i];
        
        v.v[j][i].a[kXDim]    = a;
        v.v[j][i].a[kYDim]    = a;
        
#ifdef PLM
        v.v[j][i].al[kXDim]   = a;
        v.v[j][i].ar[kXDim]   = a;
#endif
#ifdef PPM
        v.v[j][i].al [kXDim]  = a;
        v.v[j][i].ar [kXDim]  = a;
        v.v[j][i].del[kXDim]  = 0.0;
        v.v[j][i].a6 [kXDim]  = 0.0;
#endif
        
#ifdef PLM
        v.v[j][i].al[kYDim]   = a;
        v.v[j][i].ar[kYDim]   = a;
#endif PLM
#ifdef PPM
        v.v[j][i].al [kYDim]  = a;
        v.v[j][i].ar [kYDim]  = a;
        v.v[j][i].del[kYDim]  = 0.0;
        v.v[j][i].a6 [kYDim]  = 0.0;
#endif PPM        
        
      } //i
    } //j
    
  } else { err = memErr; }
  return(err);
}

int zls_SetVar3DArrQ3DArr(V3DArr v, Q3DArr q){
  int err = noErr;
  Integer i, j, k;
  Integer ni, nj, nk;
  Real             a;
  
  if ((v.v!=NULL)&&(q!=NULL)) {
    
    ni = v.nt[kXDim];
    nj = v.nt[kYDim];
    nk = v.nt[kZDim];
    
    for (k = 0; k<nk; k++){
      
      for (j = 0; j<nj; j++){
        
        for (i = 0; i<ni; i++){
          
          a = q[k][j][i];
          
          v.v[k][j][i].a[kXDim]    = a;          
          v.v[k][j][i].a[kYDim]    = a;          
          
#ifdef PLM
          v.v[k][j][i].al [kXDim]  = a;
          v.v[k][j][i].ar [kXDim]  = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kXDim]  = a;
          v.v[k][j][i].ar [kXDim]  = a;
          v.v[k][j][i].del[kXDim]  = 0.0;
          v.v[k][j][i].a6 [kXDim]  = 0.0;
#endif
          
#ifdef PLM
          v.v[k][j][i].al [kYDim]  = a;
          v.v[k][j][i].ar [kYDim]  = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kYDim]  = a;
          v.v[k][j][i].ar [kYDim]  = a;
          v.v[k][j][i].del[kYDim]  = 0.0;
          v.v[k][j][i].a6 [kYDim]  = 0.0;
#endif
          
#ifdef PLM
          v.v[k][j][i].al [kZDim]  = a;
          v.v[k][j][i].ar [kZDim]  = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kZDim]  = a;
          v.v[k][j][i].ar [kZDim]  = a;
          v.v[k][j][i].del[kZDim]  = 0.0;
          v.v[k][j][i].a6 [kZDim]  = 0.0;
#endif
          
        } //i
      } //j
    } //k
    
  } else { err = memErr; }
  return(err);
}

//========================================


int zls_SetVar1DArrQ2DArrX(V1DArr v, Q2DArr q, Integer j){
  int err = noErr;
  
  Integer     i;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    Integer  jc = j + v.ns;
    for (i = 0; i<v.nt; i++){
      
      a = q[jc][i];
      v.v[i].a      = a;
      
#ifdef PLM
      v.v[i].al    = a;
      v.v[i].ar    = a;
#endif PLM
#ifdef PPM
      v.v[i].al    = a;
      v.v[i].ar    = a;
      v.v[i].del   = 0.0;
      v.v[i].a6    = 0.0;
#endif
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}



int zls_SetVar1DArrQ2DArrY(V1DArr v, Q2DArr q, Integer i){
  int err = noErr;
  
  Integer     j;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    Integer  ic = i + v.ns;
    for (j = 0; j<v.nt; j++){
      
      a = q[j][ic];
      
      v.v[j].a     = a;
      
#ifdef PLM
      v.v[j].al    = a;
      v.v[j].ar    = a;
#endif PLM
#ifdef PPM
      v.v[j].al    = a;
      v.v[j].ar    = a;
      v.v[j].del   = 0.0;
      v.v[j].a6    = 0.0;
#endif PPM
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar1DArrQ3DArrX(V1DArr v, Q3DArr q, Integer k, Integer j ){
  int err = noErr;
  
  Integer     i;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    Integer  kc = k + v.ns;
    Integer  jc = j + v.ns;
    
    for (i = 0; i<v.nt; i++){
      
      a = q[kc][jc][i];
      
      v.v[i].a     = a;
      
#ifdef PLM
      v.v[i].al    = a;
      v.v[i].ar    = a;
#endif
#ifdef PPM
      v.v[i].al    = a;
      v.v[i].ar    = a;
      v.v[i].del   = 0.0;
      v.v[i].a6    = 0.0;
#endif
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar1DArrQ3DArrY(V1DArr v, Q3DArr q, Integer k, Integer i){
  int err = noErr;
  
  Integer     j;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    Integer  kc = k + v.ns;
    Integer  ic = i + v.ns;
    
    for (j = 0; j<v.nt; j++){
      
      a = q[kc][j][ic];
      
      v.v[j].a     = a;
      
#ifdef PLM
      v.v[j].al    = a;
      v.v[j].ar    = a;
#endif
#ifdef PPM
      v.v[j].al    = a;
      v.v[j].ar    = a;
      v.v[j].del   = 0.0;
      v.v[j].a6    = 0.0;
#endif
      
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar1DArrQ3DArrZ(V1DArr v, Q3DArr q,  Integer j, Integer i){
  int err = noErr;
  
  Integer     k;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    
    Integer  jc = j + v.ns;
    Integer  ic = i + v.ns;
    
    for (k = 0; k<v.nt; k++){
      
      a = q[k][jc][ic];
      
      v.v[k].a       = a;
      
#ifdef PLM
      v.v[k].al    = a;
      v.v[k].ar    = a;
#endif
#ifdef PPM
      v.v[k].al    = a;
      v.v[k].ar    = a;
      v.v[k].del   = 0.0;
      v.v[k].a6    = 0.0;
#endif
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}


//========================================

int zls_SetVar1DArrAxis(V1DArr v, ACoordPtr x){ // dims must match
  int err = noErr;
  v.x = x;
  return(err);
}

int zls_SetVar2DArrAxis(V2DArr v, ACoordPtr x, ACoordPtr y){ // v provides dimensions
  int err = noErr;
  v.x = x;
  v.y = y;
  return(err);
}

int zls_SetVar3DArrAxis(V3DArr v, ACoordPtr x, ACoordPtr y, ACoordPtr z){
  int err = noErr;
  v.x = x;
  v.y = y;
  v.z = z;
  return(err);
}


int zls_SetVar1DArrQ(V1DArr v, QPtr q){
  int err = noErr;
  
  Integer     i;
  Real        a;
  
  if ((v.v!=NULL) &&(q!=NULL)){
    
    for (i = 0; i<v.nt; i++){
      
      a = q[i];
      
      v.v[i].a     = a;
      
#ifdef PLM
      v.v[i].al    = a;
      v.v[i].ar    = a;
#endif
#ifdef PPM
      v.v[i].al    = a;
      v.v[i].ar    = a;
      v.v[i].del   = 0.0;
      v.v[i].a6    = 0.0;
#endif
      
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar2DArrQ(V2DArr v, QPtr q){
  int err = noErr;
  Integer i, j;
  Integer ni, nj, jOffset;
  Real      a;
  
  if ((v.v!=NULL)&&(q!=NULL)) {
    
    ni = v.nt[kXDim];
    nj = v.nt[kYDim];
    
    for (j = 0; j<nj; j++){
      jOffset = j*ni;
      
      for (i = 0; i<ni; i++){
        
        a = q[i + jOffset];
        
        v.v[j][i].a[kXDim]   = a;
        v.v[j][i].a[kYDim]   = a;
        
#ifdef PLM
        v.v[j][i].al [kXDim] = a;
        v.v[j][i].ar [kXDim] = a;
#endif PLM
#ifdef PPM
        v.v[j][i].al [kXDim] = a;
        v.v[j][i].ar [kXDim] = a;
        v.v[j][i].del[kXDim] = 0.0;
        v.v[j][i].a6 [kXDim] = 0.0;
#endif PPM
        
#ifdef PLM
        v.v[j][i].al [kYDim] = a;
        v.v[j][i].al [kYDim] = a;
#endif PLM
#ifdef PPM
        v.v[j][i].al [kYDim] = a;
        v.v[j][i].al [kYDim] = a;
        v.v[j][i].del[kYDim] = 0.0;
        v.v[j][i].a6 [kYDim] = 0.0;
#endif PPM
        
      } //i
    } //j
    
  } else { err = memErr; }
  return(err);
}

int zls_SetVar3DArrQ(V3DArr v, QPtr q){
  int err = noErr;
  Integer i, j, k;
  Integer ni, nj, nk;
  Integer jOffset, kOffset;
  Real             a;
  
  if ((v.v!=NULL)&&(q!=NULL)) {
    
    ni = v.nt[kXDim];
    nj = v.nt[kYDim];
    nk = v.nt[kZDim];
    
    for (k = 0; k<nk; k++){
      kOffset = k*nj*ni;
      for (j = 0; j<nj; j++){
        jOffset = j*nj;
        for (i = 0; i<ni; i++){
          
          a = q[i + jOffset + kOffset];
          
          v.v[k][j][i].a[kXDim]     = a;
          v.v[k][j][i].a[kYDim]     = a;
          v.v[k][j][i].a[kZDim]     = a;
          
#ifdef PLM
          v.v[k][j][i].al [kXDim] = a;
          v.v[k][j][i].ar [kXDim] = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kXDim] = a;
          v.v[k][j][i].ar [kXDim] = a;
          v.v[k][j][i].del[kXDim] = 0.0;
          v.v[k][j][i].a6 [kXDim] = 0.0;
#endif PPM
          
#ifdef PLM
          v.v[k][j][i].al [kYDim] = a;
          v.v[k][j][i].ar [kYDim] = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kYDim] = a;
          v.v[k][j][i].al [kYDim] = a;
          v.v[k][j][i].del[kYDim] = 0.0;
          v.v[k][j][i].a6 [kYDim] = 0.0;
#endif PPM
          
#ifdef PLM
          v.v[k][j][i].al [kZDim] = a;
          v.v[k][j][i].ar [kZDim] = a;
#endif PLM
#ifdef PPM
          v.v[k][j][i].al [kZDim] = a;
          v.v[k][j][i].al [kZDim] = a;
          v.v[k][j][i].del[kZDim] = 0.0;
          v.v[k][j][i].a6 [kZDim] = 0.0;
#endif PPM
          
        } //i
      } //j
    } //k
    
  } else { err = memErr; }
  return(err);
}

//========================================
// Single quantity LR only variables basic create dispose routines: a,
// al ar independently of PPM/PLM/GODUNOV Used in WStar and main zone
// Field structures
//========================================


int zls_NewLRQ1DArr  (LRQ1DArr *v, Counter nti){
  
  int err = noErr;
  
  *v = (LRQ1 *) malloc((nti)*sizeof(LRQ1));
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
  
}

int zls_NewLRQ2DArr  (LRQ2DArr *v, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (LRQ2 **) zlsAlloc2DArr(sizeof(LRQ2), ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewLRQ3DArr  (LRQ3DArr *v, Counter ntk, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (LRQ3 ***) zlsAlloc3DArr(sizeof(LRQ3), ntk, ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeLRQ1DArr (LRQ1DArr v){
  int err = noErr;
  if (v !=NULL) free(v); v = NULL;
  return(err);
}

int zls_DisposeLRQ2DArr (LRQ2DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree2DArr((void **)v); v = NULL;
  return(err);
}

int zls_DisposeLRQ3DArr (LRQ3DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree3DArr((void ***)v); v = NULL;
  return(err);
}


int zls_NewSLRQ1DArr  (SLRQ1DArr *v, Counter nti){
  
  int err = noErr;
  
  *v = (SLRQ1 *) malloc((nti)*sizeof(SLRQ1));
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
  
}

int zls_NewSLRQ2DArr  (SLRQ2DArr *v, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (SLRQ2 **) zlsAlloc2DArr(sizeof(SLRQ2), ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewSLRQ3DArr  (SLRQ3DArr *v, Counter ntk, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (SLRQ3 ***) zlsAlloc3DArr(sizeof(SLRQ3), ntk, ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeSLRQ1DArr (SLRQ1DArr v){
  int err = noErr;
  if (v !=NULL) free(v); v = NULL;
  return(err);
}

int zls_DisposeSLRQ2DArr (SLRQ2DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree2DArr((void **)v); v = NULL;
  return(err);
}

int zls_DisposeSLRQ3DArr (SLRQ3DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree3DArr((void ***)v); v = NULL;
  return(err);
}


int zls_NewVLRQ1DArr  (VLRQ1DArr *v, Counter nti){
  
  int err = noErr;
  
  *v = (VLRQ1 *) malloc((nti)*sizeof(VLRQ1));
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
  
}

int zls_NewVLRQ2DArr  (VLRQ2DArr *v, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (VLRQ2 **) zlsAlloc2DArr(sizeof(VLRQ2), ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVLRQ3DArr  (VLRQ3DArr *v, Counter ntk, Counter ntj, Counter nti){
  
  int err = noErr;
  
  *v = (VLRQ3 ***) zlsAlloc3DArr(sizeof(VLRQ3), ntk, ntj, nti);
  
  if (*v == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeVLRQ1DArr (VLRQ1DArr v){
  int err = noErr;
  if (v !=NULL) free(v); v = NULL;
  return(err);
}

int zls_DisposeVLRQ2DArr (VLRQ2DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree2DArr((void **)v); v = NULL;
  return(err);
}

int zls_DisposeVLRQ3DArr (VLRQ3DArr v){
  int err = noErr;
  if (v !=NULL) zlsFree3DArr((void ***)v); v = NULL;
  return(err);
}


int zls_SetLRQ1DArrQArr   (LRQ1DArr v, QPtr q, Integer nBC, Counter ni){ 
  // set LRQ = q values, dims must match, BCs are offset if present
  // ni is size of q, nBC are the number of boundary cells in v
  
  int err = noErr;
  
  Integer     i;
  Real        a;
  
  if ((v!=NULL) &&(q!=NULL)){
    Integer is = nBC;
    Integer ie = ni + nBC;
    Integer ic;
    
    for (i = is; i<ie; i++){
      ic = i - nBC;
      a  = q[ic];
      v[i].a     = a;
      v[i].al    = a;
      v[i].ar    = a;      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetLRQ2DArrQArr   (LRQ2DArr v, Q2DArr q, Integer nBC, Counter nj, Counter ni){ 
  // no dimensions are checked
  
  int err = noErr;
  
  Integer     i,j;
  Real        a;
  
  if ((v!=NULL) &&(q!=NULL)){
    
    Integer is = nBC;
    Integer ie = ni + nBC;
    Integer js = nBC;
    Integer je = nj + nBC;
    Integer ic, jc;
    
    for (j = js; j<je; j++){
      jc = j - nBC;
      for (i = is; i<ie; i++){
        ic = i - nBC;
        
        a  = q[jc][ic];
        
#ifdef SCALAR_LRQS
        v[j][i].a             = a;
#endif SCALAR_LRQS
        
#ifdef VECTOR_LRQS
        v[j][i].a[kXDim]      = a;
        v[j][i].a[kYDim]      = a;
#endif VECTOR_LRQS
        
        v[j][i].al[kXDim]    = a;
        v[j][i].ar[kXDim]    = a;
        v[j][i].al[kYDim]    = a;
        v[j][i].ar[kYDim]    = a;
        
      }
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetLRQ3DArrQArr   (LRQ3DArr v, Q3DArr q, Integer nBC, Counter nk, Counter nj, Counter ni){
  // sets initial left and right values to same as q  
  int err = noErr;
  
  Integer     i,j,k;
  Real        a;
  
  if ((v!=NULL) &&(q!=NULL)){
    
    Integer is = nBC;
    Integer ie = ni + nBC;
    Integer js = nBC;
    Integer je = nj + nBC;
    Integer ks = nBC;
    Integer ke = nk + nBC;
    Integer ic, jc, kc;
    
    for (k = ks; k<ke; k++){
      kc = k - nBC;
      for (j = js; j<je; j++){
        jc = j - nBC;
        for (i = is; i<ie; i++){
          ic = i - nBC;
          
          a  = q[kc][jc][ic];
          
#ifdef SCALAR_LRQS
          v[k][j][i].a            = a;
#endif SCALAR_LRQS
          
#ifdef VECTOR_LRQS
          v[k][j][i].a [kXDim]    = a;
          v[k][j][i].a [kYDim]    = a;
          v[k][j][i].a [kZDim]    = a;
#endif VECTOR_LRQS
          
          v[k][j][i].al[kXDim]    = a;
          v[k][j][i].ar[kXDim]    = a;
          v[k][j][i].al[kYDim]    = a;
          v[k][j][i].ar[kYDim]    = a;
          v[k][j][i].al[kZDim]    = a;
          v[k][j][i].ar[kZDim]    = a;
          
        }
      }
    }
    
  } else {err = memErr; } 
  
  return(err);
}

int zls_SetVar1DArrtLRQ1DArr ( V1DArr v, LRQ1DArr lr ){
  int err = noErr;
  
  Integer     i;
  
  if ((v.v!=NULL) &&(lr!=NULL)){
    
    for (i = v.ns; i<v.ne; i++){
      
      v.v[i].a     = lr[i].a;
#ifndef GODUNOV        
      v.v[i].al    = lr[i].al;
      v.v[i].ar    = lr[i].ar;
#ifdef PPM
      v.v[i].del   = 0.0;
      v.v[i].a6    = 0.0;
#endif PPM
#endif GODUNOV        
      
    }
    
  } else {err = memErr; } 
  
  return(err);
}

//========================================


int zls_RepeatBoundaryVar1DArr   (V1DArr v){
  
  int err = noErr;
  
  Integer i, nG;
  Integer is, ie, ic;
  Real a;
  Var1 * va;
  
  if (v.v != NULL){
    va = v.v;
    
    nG = v.ns;
    is = v.ns;
    a  = va[is].a;
    
    for(i = 0; i < nG; i++){
      ic = is-i-1;
      va[ic].a     = a;
      
#ifndef GODUNOV
      va[ic].al   = a;
      va[ic].ar   = a;
#endif
    }
    
    ie = v.ne;
    a = va[ie-1].a;
    
    for(i = 0; i < nG; i++){
      ic = ie+i;
      va[ic].a     = a;
      
#ifndef GODUNOV
      va[ic].al   = a;
      va[ic].ar   = a;
#endif
    }
    
  } else {err = memErr;}
  
  return(err);         
  
}

int zls_RepeatBoundaryVar2DArr(V2DArr v){
  int err = noErr;
  
  Integer i, j, ic, jc;
  Integer is, js;
  Integer ie, je;
  Integer nG, nti, ntj;
  Real ax, ay;
  
  Var2 ** va;
  
  if (v.v != NULL){
    va = v.v;
    
    is  = v.ns[kXDim];
    ie  = v.ne[kXDim];
    nti = v.nt[kXDim];
    
    js  = v.ns[kYDim];
    je  = v.ne[kYDim];
    ntj = v.nt[kYDim];
    
    // left and right bands, whole width
    nG = v.ns[kXDim];
    
    for(j = 0; j < ntj; j++){
      
      ax = va[j][is].a[kXDim];
      ay = va[j][is].a[kYDim];
      
      for(i = 0; i < nG; i++) {
        ic = is-i-1;
        
        va[j][ic].a[kXDim]    = ax;
        va[j][ic].a[kYDim]    = ay;
        
        va[j][ic].al[kXDim]   = ax;
        va[j][ic].ar[kXDim]   = ax;
        va[j][ic].al[kYDim]   = ay;
        va[j][ic].ar[kYDim]   = ay;
        
      }
      
      ax = va[j][ie-1].a[kXDim];
      ay = va[j][ie-1].a[kYDim];
      for(i = 0; i < nG; i++){
        ic = ie+i;
        
        va[j][ic].a[kXDim]    = ax;
        va[j][ic].a[kYDim]    = ay;
        
#ifndef GODUNOV
        va[j][ic].al[kXDim]   = ax;
        va[j][ic].ar[kXDim]   = ax;
        va[j][ic].al[kYDim]   = ay;
        va[j][ic].ar[kYDim]   = ay;
#endif
      }
      
    }//j
    
    // bottom and top bands, whole width
    nG = v.ns[kYDim];
    
    for(i = 0; i < nti; i++){
      
      ax = va[js][i].a[kXDim];
      ay = va[js][i].a[kYDim];
      
      for(j = 0; j < nG; j++){
        jc = (js-j-1);
        
        va[jc][i].a [kXDim]   = ax;
        va[jc][i].a [kYDim]   = ay;
        
#ifndef GODUNOV
        va[jc][i].al[kXDim]   = ax;
        va[jc][i].ar[kXDim]   = ax;
        va[jc][i].al[kYDim]   = ay;
        va[jc][i].ar[kYDim]   = ay;
#endif
      }//j
      
      ax = va[je-1][i].a [kXDim];
      ay = va[je-1][i].a [kYDim];
      
      
      for(j = 0; j < nG; j++){
        jc =  (je+j);
        
        va[jc][i].a [kXDim]   = ax;
        va[jc][i].a [kYDim]   = ay;
        
#ifndef GODUNOV
        va[jc][i].al[kXDim]   = ax;
        va[jc][i].ar[kXDim]   = ax;
        va[jc][i].al[kYDim]   = ay;
        va[jc][i].ar[kYDim]   = ay;
#endif
      }//j
      
    }//i
    
  } else {err = memErr;}
  
  return(err);         
  
}

int zls_RepeatBoundaryVar3DArr( V3DArr v){
  
  int err = noErr;
  
  Integer i, j, k, ic, jc, kc;
  Integer is, js, ks;
  Integer ie, je, ke;
  Integer nG, nti, ntj, ntk;
  
  Real ax, ay, az;
  Var3 *** va;
  
  if (v.v != NULL){
    va = v.v;
    
    is = v.ns[kXDim];
    ie = v.ne[kXDim];
    nti = v.nt[kXDim];
    
    js = v.ns[kYDim];
    je = v.ne[kYDim];
    ntj = v.nt[kYDim];
    
    ks = v.ns[kZDim];
    ke = v.ne[kZDim];
    ntk = v.nt[kZDim];
    
    // left and right bands, whole jk end
    nG = v.ns[kXDim];
    
    for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
        
        ax = va[k][j][is].a [kXDim];
        ay = va[k][j][is].a [kYDim];
        az = va[k][j][is].a [kZDim];
        
        for(i = 0; i < nG; i++) {
          ic = is-i-1;
          
          va[k][j][ic].a [kXDim]   = ax;
          va[k][j][ic].a [kYDim]   = ay;
          va[k][j][ic].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[k][j][ic].al[kXDim]   = ax;
          va[k][j][ic].ar[kXDim]   = ax;
          va[k][j][ic].al[kYDim]   = ay;
          va[k][j][ic].ar[kYDim]   = ay;
          va[k][j][ic].al[kZDim]   = az;
          va[k][j][ic].ar[kZDim]   = az;
#endif
        }
        
        ax = va[k][j][ie-1].a [kXDim];
        ay = va[k][j][ie-1].a [kYDim];
        az = va[k][j][ie-1].a [kZDim];
        
        for(i = 0; i < nG; i++){
          ic = ie+i;
          
          va[k][j][ic].a [kXDim]   = ax;
          va[k][j][ic].a [kYDim]   = ay;
          va[k][j][ic].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[k][j][ic].al[kXDim]   = ax;
          va[k][j][ic].ar[kXDim]   = ax;
          va[k][j][ic].al[kYDim]   = ay;
          va[k][j][ic].ar[kYDim]   = ay;
          va[k][j][ic].al[kZDim]   = az;
          va[k][j][ic].ar[kZDim]   = az;
#endif
        }
        
      }//j
    }//k
    
    // bottom and top bands, whole ik area
    nG = v.ns[kYDim];
    
    for(k = 0; k < ntk; k++){
      for(i = 0; i < nti; i++){
        
        ax = va[k][js][i].a  [kXDim];
        az = va[k][js][i].a  [kYDim];
        ay = va[k][js][i].a  [kZDim];
        
        for(j = 0; j < nG; j++){
          jc = (js-j-1);
          
          va[k][jc][i].a [kXDim]   = ax;
          va[k][jc][i].a [kYDim]   = ay;
          va[k][jc][i].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[k][jc][i].al[kXDim]   = ax;
          va[k][jc][i].ar[kXDim]   = ax;
          va[k][jc][i].al[kYDim]   = ay;
          va[k][jc][i].ar[kYDim]   = ay;
          va[k][jc][i].al[kZDim]   = az;
          va[k][jc][i].ar[kZDim]   = az;
#endif
          
        }//j
        
        ax = va[k][je-1][i].a [kXDim];
        ay = va[k][je-1][i].a [kYDim];
        az = va[k][je-1][i].a [kZDim];
        
        for(j = 0; j < nG; j++){
          jc =  (je+j);
          
          va[k][jc][i].a [kXDim]   = ax;
          va[k][jc][i].a [kYDim]   = ay;
          va[k][jc][i].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[k][jc][i].al[kXDim]   = ax;
          va[k][jc][i].ar[kXDim]   = ax;
          va[k][jc][i].al[kYDim]   = ay;
          va[k][jc][i].ar[kYDim]   = ay;
          va[k][jc][i].al[kZDim]   = az;
          va[k][jc][i].ar[kZDim]   = az;
#endif
          
        }//j
      }//i
    }//k
    // hiher and yon bands, whole ij area
    nG = v.ns[kYDim];
    
    for(j = 0; j < ntj; j++){
      for(i = 0; i < nti; i++){
        
        ax = va[ks][j][i].a [kXDim];
        ay = va[ks][j][i].a [kYDim];
        az = va[ks][j][i].a [kYDim];
        
        for(k = 0; k < nG; k++){
          kc = (ks-k-1);
          
          va[kc][j][i].a [kXDim]   = ax;
          va[kc][j][i].a [kYDim]   = ay;
          va[kc][j][i].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[kc][j][i].al[kXDim]   = ax;
          va[kc][j][i].ar[kXDim]   = ax;
          va[kc][j][i].al[kYDim]   = ay;
          va[kc][j][i].ar[kYDim]   = ay;
          va[kc][j][i].al[kZDim]   = az;
          va[kc][j][i].ar[kZDim]   = az;
#endif
        }//k
        
        ax = va[ke-1][j][i].a [kXDim];
        ay = va[ke-1][j][i].a [kYDim];
        az = va[ke-1][j][i].a [kZDim];
        
        for(k = 0; k < nG; k++){
          kc =  (ke+k);
          
          va[kc][j][i].a [kXDim]   = ax;
          va[kc][j][i].a [kYDim]   = ay;
          va[kc][j][i].a [kZDim]   = az;
          
#ifndef GODUNOV
          va[kc][j][i].al[kXDim]   = ax;
          va[kc][j][i].ar[kXDim]   = ax;
          va[kc][j][i].al[kYDim]   = ay;
          va[kc][j][i].ar[kYDim]   = ay;
          va[kc][j][i].al[kZDim]   = az;
          va[kc][j][i].ar[kZDim]   = az;
#endif
        }//k
        
      }//i
    }//j
    
  } else {err = memErr;}
  
  return(err);         
  
}

