#ifndef ZLS_PROBLEM_H
#define ZLS_PROBLEM_H
#include <zls.h>


#define FYRIS_PROBLEMNAME "Mach Wedge"
// this identifier is recorded in state dumps so that 
// the reader can cross check consistency of the data. 
// define a new one for each different problem
#define MACHWEDGE
//#define DRIVEN_ISM
//#define ZLS_USER_BOUNDARIES
//#define ZLS_USER_UPDATE_ROUTINE



typedef struct prob {
//
// define global problem parameters here.  These will be available to all the code
// via the global level[0] structure
//
  Real gM0; // shock Mach number wrt to pre-shock medium, Mach 10
  Real gv0 ; //  shock  velocity wrt to preshock gass
  Real gT0; // wedge angle in degrees, 30 degrees default
  Real gEdgeX; // location of edge along x axis: default 1/6
//  
  Real gP0; // pre-shock Pressure = 1.0 default
  Real gD0; // pre-shock Density = 1.4 default
  Real gv0x; // pre-shock x-velocity = 0.0 default
  Real gv0y; // pre-shock y-velocity = 0.0 default
//  
  Real gP1; // post shock pressure, computed
  Real gD1; // post shock density, computed
  Real gv1x; // post shock x-velocity, computed
  Real gv1y; // post shock y-velocity, computed
//  
} Problem, * ProblemPtr;


int   zlsp_InitProblem    ( ProblemPtr   prob);
int   zlsp_NewProblem     ( ProblemPtr * prob);
int   zlsp_DisposeProblem ( ProblemPtr   prob);

int   zlsp_InitZoneData          ( ZonePtr z );
int   zlsp_InitLevelZones        ( Integer level );

int   zlsp_SetUpProblemFromConfigFile(char * confFile);

#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind );
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z);
#endif ZLS_USER_UPDATE_ROUTINE


#endif ZLS_PROBLEM_H
