      program riemann
C      
      implicit none
C
      INTEGER*4 iter, i
      INTEGER*4 n
      
      REAL*8 x0, t
      REAL*8 sp, spl, spr
      REAL*8 xl, xr, dx
      
      REAL*8 maxerr
      REAL*8 dl, Pl , vl, ul
      REAL*8 dr, Pr , vr, ur 
     
      REAL*8 d0,  p0,  v0,   e0
      REAL*8 d0l, p0l, v0l, e0l
      REAL*8 d0r, p0r, v0r, e0r

      REAL*8  deltav
      REAL*8  newtonUstar
      REAL*8  newtonStep
	
      REAL*8  Ustar
      REAL*8  Pstar
	
      REAL*8   Pstarl, Pstarr
      REAL*8  dPstarl, dPstarr
      REAL*8   astarl, astarr
      REAL*8   deltaP, deltaDP

	
      REAL*8  al, ar          ! Eulerian sound speed
      REAL*8  inv_al, inv_ar  ! Eulerian sound rates
      REAL*8  cl, cr          ! Lagrangian sound speed
      REAL*8  sl, sr          ! intermediate wave slopes
      REAL*8  z, zl, zr          ! shock mach number
  
      REAL*8  alphal ! (gamma+1)/(4a), predivide and collect constants
      REAL*8  alphar 
      
      REAL*8 sixth
C      
      REAL*8  gam
      REAL*8  alpha, beta, delta
      common /gg88/ alpha, beta, delta
      
      REAL*8  g0, g1, g2, g3, g4, g5, g6, g7
      common /gammas/ g0, g1, g2, g3, g4, g5, g6, g7
C      
      gam = 7.d0/5.d0
      write(*,*) ' Polytropic gamma : '
      read(*,*) gam
C
C Groth and Gottlieb solver constants
C
      alpha = ((gam+1.d0)/4.d0)
      beta  = ((gam-1.d0)/2.d0)
      delta = (2.d0*gam)/(gam-1.d0)
C
C Toro sampler constants
C
      g0 = (gam)
      g1 = (gam-1.d0)/(2.d0*gam)
      g2 = (gam+1.d0)/(2.d0*gam)
      g3 = (2.d0*gam)/(gam-1.d0)
      g4 = 2.d0/(gam-1.d0)
      g5 = 2.d0/(gam+1.d0)
      g6 = (gam-1.d0)/(gam+1.d0)
      g7 = (gam-1.d0)/2.d0

      write(*,*) 'x0, time t : '
      read(*,*) x0, t

      write(*,*) '  W: d, P, v : '
      read(*,*)    dl, Pl, vl

      write(*,*) ' 1 W: d, P, v : '
      read(*,*)    dr, Pr, vr 
        
      al      = sqrt(gam*Pl/(dl))
      ar      = sqrt(gam*Pr/(dr))
      
      inv_al  = 1.d0/al
      inv_ar  = 1.d0/ar
      deltav   = vr-vl
      
      cl       = dl*al 
      cr       = dr*ar	
      
      alphal = alpha*inv_al
      alphar = alpha*inv_ar

      ul = vl + al/beta
      ur = vr - ar/beta
      
      z = (ar/al)*((Pl/Pr)**(1.d0/delta))
      newtonUstar = (ul*z + ur)/(1.d0+z)

      maxerr = 1.d0
      do iter = 1, 40

        Ustar  = newtonUstar
				
        if ( Ustar .le. vl ) then
C left wave is shock
          sl     = (newtonUstar - vl)*alphal
          zl     = sl - dsqrt(1.d0+(sl*sl))
          Pstarl = Pl + cl*(newtonUstar - vl)*zl
          dPstarl = 2*cl*zl*zl*zl/(1.d0+zl*zl)
					
        else  
C left wave is rarefaction 
					
          astarl  = al - beta*(newtonUstar - vl)
          Pstarl  = Pl * ((astarl*inv_al) ** delta)
          dPstarl = -gam*Pstarl/astarl
					
        endif
				
        if ( Ustar .ge. vr ) then 
C right wave is shock
					
          sr     = (newtonUstar - vr)*alphar
          zr     = sr + dsqrt(1.d0+(sr*sr))
          Pstarr = Pr + cr*(newtonUstar - vr)*zr
          dPstarr = 2*cr*zr*zr*zr/(1.d0+zr*zr)
					
        else  
C right wave is rarefaction
					
          astarr  = ar + beta*(newtonUstar - vr)
          Pstarr  = Pr * ((astarr*inv_ar)**delta)
          dPstarr = gam*Pstarr/astarr
          
        endif
				

        deltaP     = ( Pstarl -  Pstarr)
        deltaDP    = (dPstarl - dPstarr)
        newtonStep = deltaP/deltaDP
        
        newtonUstar   =  newtonUstar - newtonStep 
        maxerr        = dabs(newtonStep/newtonUstar)
        
	        if (maxerr.lt.1.d-14) goto 26
      enddo
 26   continue
      
      Ustar  = newtonUstar
      Pstar  = 0.5*(Pstarl+Pstarr)
      
      write(*,*) 'xl, xr, n: '
      read(*,*) xl, xr, n
      
      dx = (xr-xl)/n
      
      call  WaveLocations( t, x0, Pstar, Ustar
     &           , Pl, vl, dl, al
     &           , Pr, vr, dr, ar
     &           , p0, v0, d0)


      do i = 0, n-1
      
C simple cell average of left and right coords, to 
C give a more accurate comparison with fyris cell integral
C values.  fyris produces cell integrals, not centred point 
C values.

C
C Evaluate the left, centre and right values for cells
C and compute the PPM integral
C

      spl = (xl+(i*dx)-x0)/t
      spr = (xl+((i+1.d0)*dx)-x0)/t
      
      sp  = (xl+((i+0.5d0)*dx)-x0)/t
      
      call  sample( spl, Pstar, Ustar
     &           , Pl, vl, dl, al
     &           , Pr, vr, dr, ar
     &           , p0l, v0l, d0l )
     
      e0l = p0l/(d0l*(gam-1.d0))

      call  sample( spr, Pstar, Ustar
     &           , Pl, vl, dl, al
     &           , Pr, vr, dr, ar
     &           , p0r, v0r, d0r )
     
      e0r = p0r/(d0r*(gam-1.d0))
      
      call  sample( sp, Pstar, Ustar
     &           , Pl, vl, dl, al
     &           , Pr, vr, dr, ar
     &           , p0, v0, d0 )
     
      e0 = p0/(d0*(gam-1.d0))
      
c
c convert left-central-right values to PPM integral average
c over cell
c    <a> = (1/6)*(4*a_c+a_l+a_r)
c
      sixth = 1.d0/6.d0
      d0 = sixth*(4.d0*d0+d0l+d0r)
      p0 = sixth*(4.d0*p0+p0l+p0r)
      v0 = sixth*(4.d0*v0+v0l+v0r)
      e0 = sixth*(4.d0*e0+e0l+e0r)
     
      write(*,'(5(1pg14.7,x))') (xl+((i+0.5)*dx))
     &                        , d0
     &                        , p0
     &                        , v0
     &                        , e0 

      enddo

      end

     
      subroutine sample(s, pm, um, 
     &     pl, vl, rl, cl, 
     &     pr, ur, rr, cr, 
     &     p0, v0, d0)
C
C     Riemann solution sampler at speed = s
C     returns p0, v0 and d0
C
      implicit none

      REAL*8 s, pm, um
      REAL*8 pl, vl, rl, cl
      REAL*8 pr, ur, rr, cr
      REAL*8 p0, v0, d0
C
      REAL*8 shl, shr, c, sr, sl, pml, pmr, cml, cmr, stl , str
      REAL*8 g0, g1, g2, g3, g4, g5, g6, g7
      common /gammas/ g0, g1, g2, g3, g4, g5, g6, g7
            

C
      if (s.le.um) then
C
C    to the left of the CD
C      
         if (pm.le.pl) then
C  left rarefaction      
           shl = vl - cl

           if (s.le.shl) then
C  in left state           
             v0 = vl
             p0 = pl
             d0 = rl
           else
	           cml = cl*(pm/pl)**g1
	           stl = um - cml

	           if (s.gt.stl) then
C  in left star state           
	              d0 = rl*(pm/pl)**(1.d0/g0)
	              p0 = pm
	              v0 = um
	           else
C  in left fan state    
                v0 = g5*(cl + g7*vl+s)
                c  = g5*(cl + g7*(vl-s))
                d0 = rl*(c/cl)**g4
                p0 = pl*(c/cl)**g3
             endif
           endif
         else
C  left shock
           pml = pm/pl
           sl  = vl - cl*sqrt(g2*pml + g1)
           
           if (s.le.sl) then
C  in left state     
               d0 = rl
               p0 = pl
               v0 = vl      
           else
C  in left star state
               d0 = rl*(pml+g6)/(pml*g6+1.0)
               v0 = um
               p0 = pm
           endif
         endif
      else
C
C    to the right of the CD
C      
         if(pm.gt.pr)then
C   right shock
            pmr = pm/pr
            sr  = ur + cr*sqrt(g2*pmr + g1)
                        
            if (s.ge.sr) then
C   in right state
                d0 = rr
                v0 = ur
                p0 = pr
            else
C   in right star state
                d0 = rr*(pmr + g6)/(pmr*g6 + 1.d0)
                v0 = um
                p0 = pm
            endif
         else
C   right rarefaction
            shr = ur + cr

            if (s.ge.shr) then
C   in right state
                d0 = rr
                v0 = ur
                p0 = pr
            else
                cmr = cr*(pm/pr)**g1
                str = um + cmr

                if (s.le.str) then
C   in right star state
		              d0 = rr*(pm/pr)**(1.d0/g0)
	                v0 = um
	                p0 = pm
                else
C  in right fan state    
	                v0 = g5*(-cr + g7*ur+s)
	                c  = g5*( cr - g7*(ur-s))
	                d0 = rr*(c/cr)**g4
	                p0 = pr*(c/cr)**g3
                endif
            endif
         endif
      endif
C      
      return
      end

      
      subroutine WaveLocations(t, x0, pm, um, 
     &     pl, vl, rl, cl, 
     &     pr, ur, rr, cr )
C
C     Riemann solution sampler at speed = s
C     returns p0, v0 and d0
C
      implicit none

      REAL*8 t, x0, pm, um
      REAL*8 pl, vl, rl, cl
      REAL*8 pr, ur, rr, cr
      REAL*8 rmr, rml
C
      REAL*8 shl, shr, sr, sl, pml, pmr, cml, cmr, stl , str
      REAL*8 g0, g1, g2, g3, g4, g5, g6, g7
      common /gammas/ g0, g1, g2, g3, g4, g5, g6, g7
      
      
10    format(" Exact Rieman Solution"/
     & "--------------------------------------------------"/
     &  " Gamma = ", 1pg12.5)
     
20    format(" x0    = ", 1pg12.5 / 
     &  " t     = ", 1pg12.5 /
     & "--------------------------------------------------")
     
       write(*, 10)  g0
       write(*, 20)  x0, t
                   
100   format("       R-Wave x = ",1pg12.5,1pg12.5)
101   format("       S-Wave x = ",1pg12.5)
102   format(" Contact Discont. x = ",1pg12.5)
103   format("     1 R-Wave x = ",1pg12.5,1pg12.5)
104   format("     1 S-Wave x = ",1pg12.5)

      if (pm.le.pl) then
        shl = vl - cl
        cml = cl*(pm/pl)**g1
        stl = um - cml
        write(*, 100)  x0+shl*t, x0+stl*t
	      rml = rl*(pm/pl)**(1.d0/g0)
      else 
        pml = pm/pl
        sl  = vl - cl*sqrt(g2*pml + g1)
        write(*, 101)  x0+sl*t
        rml = rl*(pml+g6)/(pml*g6+1.0)
      endif
      
      write(*, 102) x0+um*t

      if(pm.gt.pr)then
        pmr = pm/pr
        sr  = ur + cr*sqrt(g2*pmr + g1)
        write(*,104) x0+sr*t
        rmr = rr*(pmr+g6)/(pmr*g6+1.0)
      else
        shr = ur + cr
        cmr = cr*(pm/pr)**g1
        str = um + cmr
        write(*,103) x0+shr*t, x0+str*t
	      rmr = rr*(pm/pr)**(1.d0/g0)
      endif
200   format( "--------------------------------------------------")
       write(*, 200)
       
       
30    format(" d  = ", 1pg12.5, 1pg12.5, 1pg12.5, 1pg12.5 /
     &" P  = ", 1pg12.5, "      ", 1pg12.5, "      ", 1pg12.5 /
     &" v  = ", 1pg12.5, "      ", 1pg12.5, "      ", 1pg12.5 /
     &"--------------------------------------------------")

       write(*, 30)   rl, rml, rmr, rr
     &              , pl, pm, pr
     &              , vl, um, ur
     
      return
      end

