#include <zls.h>


int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  // Std SOD Shock problem		
  if (prob != NULL) {
    
//    zlsp_InitGlobalEOS_Gamma( NULL, 1.4 );
    
    prob->gX0 = 0.5; 
    prob->gX1 = 999.0; // off grid
    
    prob->gP0  = 1.0;
    prob->gd0  = 1.0; 
    prob->gv0  = 0.0; 
    
    prob->gP1  = 0.1;
    prob->gd1  = 0.125; 
    prob->gv1  = 0.0; 

    prob->gP2  = 0.1;
    prob->gd2  = 0.125; 
    prob->gv2  = 0.0; 
  }
	
  return (err);
}


int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
	
  *prob = (ProblemPtr)malloc(sizeof(Problem));
	
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    fprintf(stderr, " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
	
  return (err);
}


int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
	
  if ( prob != NULL ) {
    free(prob);
  }
	
  return (err);
}




int zlsp_InitZoneData( ZonePtr z){
  //
  // here we know the zone has BCs because we compile zls with that option.
  //
  int err = noErr;
	
  WHD       zoneW;
  ACoordPtr  xx;
  ProblemPtr prob = NULL;
  
	
  if (z != NULL){
		
    err     = zlsp_GetZoneWHD(&zoneW, z);
		
		
    if (err == noErr) {
      
      prob = gLevels[0].levelProblem;  // get the user's problem parameters
      
      
      Real x0  = prob->gX0;
      Real x1  = prob->gX1;
      
      Real d0  = prob->gd0;
      Real P0  = prob->gP0;
      Real v0  = prob->gv0;
      
      Real d1  = prob->gd1;
      Real P1  = prob->gP1;
      Real v1  = prob->gv1;

      Real d2  = prob->gd2;
      Real P2  = prob->gP2;
      Real v2  = prob->gv2;
      
      
#ifdef SYM1D
      Integer    i ;
      xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      
        for(i = 0; i< nti; i++){
          
          if (xx[i].l < x0) {
            zoneW.d [i]      =  d0;
            zoneW.P [i]      =  P0;
            zoneW.vx[i]      =  v0;
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  1.0;
#endif
          } else if (xx[i].l < x1) {
            zoneW.d [i]      =  d1;
            zoneW.P [i]      =  P1;
            zoneW.vx[i]      =  v1;
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  0.0;
#endif
          } else {
            zoneW.d [i]      =  d2;
            zoneW.P [i]      =  P2;
            zoneW.vx[i]      =  v2;
#if (NTRACERS > 0)    
            zoneW.tr[0][i]   =  0.0;
#endif
          }          
          
        }//i
        
      
        // set wind density after outburst to low value for BC  updates
#endif SYM1D
      
#ifdef SYM2D
       Integer    i, j ;
       xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      
      for(j = 0; j < ntj; j++){
        for(i = 0; i< nti; i++){
          
          if (xx[i].l < x0) {
            zoneW.d [j][i]      =  d0;
            zoneW.P [j][i]      =  P0;
            zoneW.vx[j][i]      =  v0;
            zoneW.vy[j][i]      =  0.0;
            
#if (NTRACERS > 0)    
            zoneW.tr[0][j][i]   =  1.0;
#endif
          } else if (xx[i].l < x1) {
            zoneW.d [j][i]      =  d1;
            zoneW.P [j][i]      =  P1;
            zoneW.vx[j][i]      =  v1;
            zoneW.vy[j][i]      =  0.0;
#if (NTRACERS > 0)    
            zoneW.tr[0][j][i]   =  0.0;
#endif
          } else {
            zoneW.d [j][i]      =  d2;
            zoneW.P [j][i]      =  P2;
            zoneW.vx[j][i]      =  v2;
            zoneW.vy[j][i]      =  0.0;
#if (NTRACERS > 0)    
            zoneW.tr[0][j][i]   =  0.0;
#endif
          }
          
        }//i
        
      } //j
        
        // set wind density after outburst to low value for BC  updates
#endif SYM2D
            
#ifdef SYM3D
        Integer    i, j, k ;
        xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      
      for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
	for(i = 0; i< nti; i++){
        
          if (xx[i].l >= x0){
	    zoneW.d [k][j][i]      =  d1;
	    zoneW.P [k][j][i]      =  P1;
	    zoneW.vx[k][j][i]      =  v1;
	    zoneW.vy[k][j][i]      =  0.0;
	    zoneW.vz[k][j][i]      =  0.0;
          
#if (NTRACERS > 0)    
	    zoneW.tr[0][k][j][i]   =  0.0;
#endif
          } else {
	    zoneW.d [k][j][i]      =  d0;
	    zoneW.P [k][j][i]      =  P0;
	    zoneW.vx[k][j][i]      =  v0;
	    zoneW.vy[k][j][i]      =  0.0;
	    zoneW.vz[k][j][i]      =  0.0;
          
#if (NTRACERS > 0)    
	    zoneW.tr[0][k][j][i]   =  1.0;
#endif
          }
          
	}//i
       
      } //j
      } //k
        
        // set wind density after outburst to low value for BC  updates
#endif SYM3D
      zlsp_ReleaseZoneWHD(zoneW);
        
    } // zoneW
		
  } //z
	
	
  return(err);
}



int zlsp_InitLevelZones(Integer level){
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if (z != NULL){
      err = zlsp_InitZoneData( z );
      if (err != noErr ) {return(err);}
      err = zls_SetZoneConstantBCValues( z );
      if (err != noErr ) {return(err);}
    } else { err = zonErr; }
  } // zone
  
  return(err);
  
}

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
	
  int err = noErr;
  struct ccl_t	config;
  const char		*str1;

  Integer level;
	
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " FYRIS Hydro Problem Initialising from %s \n", confFile);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      
      // Default Values: std SOD tube
      err = zlsp_InitProblem(prob);
      
      Real gamma = kZLS_eos_gamma;
      if ((str1= ccl_get( &config,"SOD_G"))){
        gamma = atof( str1);
      }
      
      if ( gamma != kZLS_eos_gamma) zlsp_InitGlobalEOS_Gamma(NULL, gamma);
      
      
      if ((str1= ccl_get( &config,"SOD_X0"))){
        prob->gX0 = atof( str1);
      }
      if ((str1= ccl_get( &config,"SOD_X1"))){
        prob->gX1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P0"))){
        prob->gP0 = atof( str1);
      }      
      if ((str1= ccl_get( &config,"SOD_D0"))){
        prob->gd0 = atof( str1);
      }
      if ((str1= ccl_get( &config,"SOD_V0"))){
        prob->gv0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P1"))){
        prob->gP1 = atof( str1);
      }
      if ((str1= ccl_get( &config,"SOD_D1"))){
        prob->gd1 = atof( str1);
      }      
      if ((str1= ccl_get( &config,"SOD_V1"))){
        prob->gv1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P2"))){
        prob->gP2 = atof( str1);
      }
      if ((str1= ccl_get( &config,"SOD_D2"))){
        prob->gd2 = atof( str1);
      }      
      if ((str1= ccl_get( &config,"SOD_V2"))){
        prob->gv2 = atof( str1);
      }
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
      for (level = 0; level < gNLevels; level++){
        err = zlsp_InitLevelZones( level ); // uses levelProblem
      }
      
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
      
      ccl_release(&config);
      
    }// config	
    
  } else {
    err = memErr;
    fprintf( stderr, " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

#ifdef ZLS_USER_BC_ROUTINE
//
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  Integer err = noErr;
	return (err);
}
#endif ZLS_USER_BC_ROUTINE


#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z){
  //
  // This routine is called just before sweeps are done, and after edge boundaries are set up.
  // The user has an opportunity to include any customised zone values, eg a stellar wind update or
  // coordinate evolution for a star.  Basicly for all boundary conditions that don't fit in the
  // grid boundary direction updates scheme.
  //  Many simulations will have nothing to do here so it is best to 
  // comment out as all the code unless a zone update is actually needed.
  // Define the global macro in zls_defs.h ZLS_USER_UPDATE_ROUTINE to enable this
  // routine.
  //
  int err = noErr;
  
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    
    prob = gLevels[0].levelProblem;  // get the user's problem parameters
    phys = gLevels[0].levelPhysics;  // get the global eos parameters
        
  } else { err = memErr; }//z
  
  
  return(err);
}
#endif ZLS_USER_UPDATE_ROUTINE

