#include "zls.h"

int  zlsbc_FillUpperZoneFromZone ( ZonePtr zDest, ZonePtr z , int boundaryExpand ){
  //
  // Same as zlsbc_FillSuperZoneFromZone except
  // zDest can be a proxy for the zone's actual super zone,
  // however zDest *must* share the exact
  // same coordinates as the actual z->zoneSuper.
  //
  // copy the cells in the active area of zone z +/- boundaryExpand cells,
  // downsample and insert in the zones owning superzone. 
  // For Eulerian codes, image processing etc, boundaryExpand = 0
  // For lagrangian codes, boundaryExpand = 1, and the zone z BCs must be equivalent to 
  // two super zone cells. ie zone nBCs = nR*2 or more.  If nR =  4, then lagrangian
  // zones on sub levels need 8 boundary cells.
  //
	Integer err = noErr;
	
#ifdef SYM1D	
	Integer i, id;	
	Integer l;
	
	Integer is;
	Integer ie;
	
	Integer ni;
	Integer nsi;
	
	Integer izd;
	Integer nti;
	Integer nsti;
	Integer iss; // super start offsets
	Integer ids; // data start offsets
	Q1DArr          q;
	Q1DArr     qSuper;
#endif SYM1D
	
#ifdef SYM2D	
	Integer i, j, id;	
	Integer l, m;
	
	Integer is, js;
	Integer ie, je;
	
	Integer ni, nj;
	Integer nsi, nsj;
	
	Integer izd, jzd;
	Integer nti, ntj;
	Integer nsti, nstj;
	Integer iss, jss; // super start offsets
	Integer ids, jds; // data start offsets
	Q2DArr          q;
	Q2DArr     qSuper;
#endif SYM2D
	
#ifdef SYM3D	
	Integer i, j, k, id;	
	Integer l, m, n;
	
	Integer is, js, ks;
	Integer ie, je, ke;
	
	Integer ni, nj, nk;
	Integer nsi, nsj, nsk;
	
	Integer izd, jzd, kzd;
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	Integer iss, jss, kss; // super start offsets
	Integer ids, jds, kds; // data start offsets
	Q3DArr          q;
	Q3DArr     qSuper;
#endif SYM3D
	
	Integer nR;
	Real    fx;
	
	ZonePtr    zSuper = zDest;
	Range         rng;
	
	if (z != NULL){
		
		nR     = z->zoneRefine;
		// zSuper = z->zoneSuper;
		
		if (zSuper != NULL){
			
			rng    = z->zoneSuperRange;
			// get target zone grid size parameters.
			// The z->zoneData[id] may be a bare Ptr to a block 
			// without BCs, or be expanded to include BCs 
#ifdef SYM1D
			ni  = z->zoneGrid.n[kXDim];
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data has no BCs
			
			nti = z->zoneGrid.nt[kXDim];
			ids = z->zoneGrid.ns[kXDim]-boundaryExpand*nR;
			
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			iss  = zSuper->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			
			// zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
      
			is -= boundaryExpand;
			ie += boundaryExpand;
      
#endif SYM1D
#ifdef SYM2D
      
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ids = z->zoneGrid.ns[kXDim]-(boundaryExpand*nR);
			jds = z->zoneGrid.ns[kYDim]-(boundaryExpand*nR);
			
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data with no BCs
			nsj = zSuper->zoneGrid.n[kYDim]; 
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data with BC, so size is nt not just n
			nstj = zSuper->zoneGrid.nt[kYDim]; // 
			iss  = zSuper->zoneGrid.ns[kXDim]; 
			jss  = zSuper->zoneGrid.ns[kYDim];
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
			
			is -= boundaryExpand;
			ie += boundaryExpand;
			js -= boundaryExpand;
			je += boundaryExpand;
      
#endif SYM2D
#ifdef SYM3D
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nk  = z->zoneGrid.n[kZDim];
			nsi = zSuper->zoneGrid.n[kXDim];
			nsj = zSuper->zoneGrid.n[kYDim];
			nsk = zSuper->zoneGrid.n[kZDim]; 
			
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ntk = z->zoneGrid.nt[kZDim];
			ids = z->zoneGrid.ns[kXDim]-(boundaryExpand*nR);
			jds = z->zoneGrid.ns[kYDim]-(boundaryExpand*nR);
			kds = z->zoneGrid.ns[kZDim]-(boundaryExpand*nR);
			
			nsti = zSuper->zoneGrid.nt[kXDim]; 
			nstj = zSuper->zoneGrid.nt[kYDim]; 
			nstk = zSuper->zoneGrid.nt[kZDim];
			iss  = zSuper->zoneGrid.ns[kXDim];
			jss  = zSuper->zoneGrid.ns[kYDim];
			kss  = zSuper->zoneGrid.ns[kZDim];
			
      // zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
			
			ks = rng.n0[kZDim];
			ke = ks+rng.rn[kZDim];
      
			is -= boundaryExpand;
			ie += boundaryExpand;
			js -= boundaryExpand;
			je += boundaryExpand;
			ks -= boundaryExpand;
			ke += boundaryExpand;
      
#endif SYM3D
			
			
			for (id = 0; id <  z->nQ; id++){
				
				if (z->zoneData[id]!=NULL){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef SYM1D
					q = z->zoneData[id];
#endif SYM1D
					
#ifdef SYM2D
					zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif SYM2D
					
#ifdef SYM3D
					zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef ZONEDATA_ARE_MULTIDARRAYS
					q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
					
					if (q!=NULL){
						
#ifdef SYM1D
						fx = 1.0/((Real)nR);
#endif SYM1D
#ifdef SYM2D
						fx = 1.0/((Real)nR*(Real)nR);
#endif SYM2D
#ifdef SYM3D
						fx = 1.0/((Real)nR*(Real)nR*(Real)nR);
#endif SYM3D
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
						qSuper = zSuper->zoneData[id];
#endif SYM1D
						
#ifdef SYM2D
						err = zls_SetQuantity2DArr(&qSuper,zSuper->zoneData[id], nstj, nsti);
#endif SYM2D
						
#ifdef SYM3D
						err = zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[id], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
#ifdef ZONEDATA_ARE_MULTIDARRAYS
						qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
						
						if (qSuper!=NULL){
#ifdef SYM1D
							for(i = is; i<ie; i++){ // step in x in Super
								izd = ids+((i-is)*nR);
								qSuper [i+iss] = 0.0;
								for( l = 0; l<nR; l++){
									qSuper [i+iss] += q [(izd+l)];
								} //l
								qSuper [i+iss] *= fx;
							} //i
#endif SYM1D
							
#ifdef SYM2D
							
							for(j = js; j<je; j++){
								jzd = jds+((j-js)*nR);
								for(i = is; i<ie; i++){ // step in x in Super
									izd = ids+((i-is)*nR);
                  
                  //  zlsio_Printf(": jzd %d izd %d j+jss %d  i+iss %d\n", jzd, izd, j+jss, i+iss); 
                  
									qSuper[j+jss][i+iss] = 0.0;
									for( l = 0; l<nR; l++){
										for( m = 0; m<nR; m++){
											qSuper[j+jss][i+iss] += q[(jzd+l)][(izd+m)];
										} //m
									} //l
                  
									qSuper[j+jss][i+iss] *= fx;
                  
								} //i
							} //j
#endif SYM2D
							
#ifdef SYM3D
							for(k = ks; k<ke; k++){
								kzd = kds+((k-ks)*nR);
								for(j = js; j<je; j++){
									jzd = jds+((j-js)*nR);
									for(i = is; i<ie; i++){ // step in x in Super
										izd = ids+((i-is)*nR);
										qSuper[k+kss][j+jss][i+iss] = 0.0;
										for( l = 0; l<nR; l++){
											for( m = 0; m<nR; m++){
												for( n = 0; n<nR; n++){
													qSuper[k+jss][j+jss][i+iss] += q[(kzd+l)][(jzd+m)][(izd+n)];
												} //n
											} //m
										} //l
										qSuper[k+kss][j+jss][i+iss] *= fx;
									} //i
								} //j
							} //k
#endif SYM3D
							
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
							zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
							zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						}
						
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
						zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
						zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
					} else { err = memErr; } // q, zSuper
					
				}
			} // var id loop
			
		}
		
  } else { err = memErr; } // z
	
	return (err);
	
}


int  zlsbc_FillSuperZoneFromZone ( ZonePtr z , int boundaryExpand ){
  //
  // copy the cells in the active area of zone z +/- boundaryExpand cells,
  // downsample and insert in the zones owning superzone. 
  // For Eulerian codes, image processing etc, boundaryExpand = 0
  // For lagrangian codes, boundaryExpand = 1, and the zone z BCs must be equivalent to 
  // two super zone cells. ie zone nBCs = nR*2 or more.  If nR =  4, then lagrangian
  // zones on sub levels need 8 boundary cells.
  //
	Integer err = noErr;
	
#ifdef SYM1D	
	Integer i, id;	
	Integer l;
	
	Integer is;
	Integer ie;
	
	Integer ni;
	Integer nsi;
	
	Integer izd;
	Integer nti;
	Integer nsti;
	Integer iss; // super start offsets
	Integer ids; // data start offsets
	Q1DArr          q;
	Q1DArr     qSuper;
#endif SYM1D
	
#ifdef SYM2D	
	Integer i, j, id;	
	Integer l, m;
	
	Integer is, js;
	Integer ie, je;
	
	Integer ni, nj;
	Integer nsi, nsj;
	
	Integer izd, jzd;
	Integer nti, ntj;
	Integer nsti, nstj;
	Integer iss, jss; // super start offsets
	Integer ids, jds; // data start offsets
	Q2DArr          q;
	Q2DArr     qSuper;
#endif SYM2D
	
#ifdef SYM3D	
	Integer i, j, k, id;	
	Integer l, m, n;
	
	Integer is, js, ks;
	Integer ie, je, ke;
	
	Integer ni, nj, nk;
	Integer nsi, nsj, nsk;
	
	Integer izd, jzd, kzd;
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	Integer iss, jss, kss; // super start offsets
	Integer ids, jds, kds; // data start offsets
	Q3DArr          q;
	Q3DArr     qSuper;
#endif SYM3D
	
	Integer nR;
	Real    fx;
	
	ZonePtr    zSuper;
	Range         rng;
	
	if (z != NULL){
		
		nR     = z->zoneRefine;
		zSuper = z->zoneSuper;
		
		if (zSuper != NULL){
			
			rng    = z->zoneSuperRange;
			// get target zone grid size parameters.
			// The z->zoneData[id] may be a bare Ptr to a block 
			// without BCs, or be expanded to include BCs 
#ifdef SYM1D
			ni  = z->zoneGrid.n[kXDim];
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data has no BCs
			
			nti = z->zoneGrid.nt[kXDim];
			ids = z->zoneGrid.ns[kXDim]-boundaryExpand*nR;
			
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			iss  = zSuper->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			
			// zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
      
			is -= boundaryExpand;
			ie += boundaryExpand;
      
#endif SYM1D
#ifdef SYM2D
      
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ids = z->zoneGrid.ns[kXDim]-(boundaryExpand*nR);
			jds = z->zoneGrid.ns[kYDim]-(boundaryExpand*nR);
			
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data with no BCs
			nsj = zSuper->zoneGrid.n[kYDim]; 
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data with BC, so size is nt not just n
			nstj = zSuper->zoneGrid.nt[kYDim]; // 
			iss  = zSuper->zoneGrid.ns[kXDim]; 
			jss  = zSuper->zoneGrid.ns[kYDim];
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
			
			is -= boundaryExpand;
			ie += boundaryExpand;
			js -= boundaryExpand;
			je += boundaryExpand;
      
#endif SYM2D
#ifdef SYM3D
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nk  = z->zoneGrid.n[kZDim];
			nsi = zSuper->zoneGrid.n[kXDim];
			nsj = zSuper->zoneGrid.n[kYDim];
			nsk = zSuper->zoneGrid.n[kZDim]; 
			
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ntk = z->zoneGrid.nt[kZDim];
			ids = z->zoneGrid.ns[kXDim]-(boundaryExpand*nR);
			jds = z->zoneGrid.ns[kYDim]-(boundaryExpand*nR);
			kds = z->zoneGrid.ns[kZDim]-(boundaryExpand*nR);
			
			nsti = zSuper->zoneGrid.nt[kXDim]; 
			nstj = zSuper->zoneGrid.nt[kYDim]; 
			nstk = zSuper->zoneGrid.nt[kZDim];
			iss  = zSuper->zoneGrid.ns[kXDim];
			jss  = zSuper->zoneGrid.ns[kYDim];
			kss  = zSuper->zoneGrid.ns[kZDim];
			
      // zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
			
			ks = rng.n0[kZDim];
			ke = ks+rng.rn[kZDim];
      
			is -= boundaryExpand;
			ie += boundaryExpand;
			js -= boundaryExpand;
			je += boundaryExpand;
			ks -= boundaryExpand;
			ke += boundaryExpand;
      
#endif SYM3D
			
			
			for (id = 0; id <  z->nQ; id++){
				
				if (z->zoneData[id]!=NULL){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef SYM1D
					q = z->zoneData[id];
#endif SYM1D
					
#ifdef SYM2D
					zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif SYM2D
					
#ifdef SYM3D
					zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef ZONEDATA_ARE_MULTIDARRAYS
					q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
					
					if (q!=NULL){
						
#ifdef SYM1D
						fx = 1.0/((Real)nR);
#endif SYM1D
#ifdef SYM2D
						fx = 1.0/((Real)nR*(Real)nR);
#endif SYM2D
#ifdef SYM3D
						fx = 1.0/((Real)nR*(Real)nR*(Real)nR);
#endif SYM3D
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
						qSuper = zSuper->zoneData[id];
#endif SYM1D
						
#ifdef SYM2D
						err = zls_SetQuantity2DArr(&qSuper,zSuper->zoneData[id], nstj, nsti);
#endif SYM2D
						
#ifdef SYM3D
						err = zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[id], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
#ifdef ZONEDATA_ARE_MULTIDARRAYS
						qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
						
						if (qSuper!=NULL){
#ifdef SYM1D
							for(i = is; i<ie; i++){ // step in x in Super
								izd = ids+((i-is)*nR);
								qSuper [i+iss] = 0.0;
								for( l = 0; l<nR; l++){
									qSuper [i+iss] += q [(izd+l)];
								} //l
								qSuper [i+iss] *= fx;
							} //i
#endif SYM1D
							
#ifdef SYM2D
							
							for(j = js; j<je; j++){
								jzd = jds+((j-js)*nR);
								for(i = is; i<ie; i++){ // step in x in Super
									izd = ids+((i-is)*nR);
                  
                  //  zlsio_Printf(": jzd %d izd %d j+jss %d  i+iss %d\n", jzd, izd, j+jss, i+iss); 
                  
									qSuper[j+jss][i+iss] = 0.0;
									for( l = 0; l<nR; l++){
										for( m = 0; m<nR; m++){
											qSuper[j+jss][i+iss] += q[(jzd+l)][(izd+m)];
										} //m
									} //l
                  
									qSuper[j+jss][i+iss] *= fx;
                  
								} //i
							} //j
#endif SYM2D
							
#ifdef SYM3D
							for(k = ks; k<ke; k++){
								kzd = kds+((k-ks)*nR);
								for(j = js; j<je; j++){
									jzd = jds+((j-js)*nR);
									for(i = is; i<ie; i++){ // step in x in Super
										izd = ids+((i-is)*nR);
										qSuper[k+kss][j+jss][i+iss] = 0.0;
										for( l = 0; l<nR; l++){
											for( m = 0; m<nR; m++){
												for( n = 0; n<nR; n++){
													qSuper[k+jss][j+jss][i+iss] += q[(kzd+l)][(jzd+m)][(izd+n)];
												} //n
											} //m
										} //l
										qSuper[k+kss][j+jss][i+iss] *= fx;
									} //i
								} //j
							} //k
#endif SYM3D
							
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
							zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
							zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						}
						
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
						zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
						zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
					} else { err = memErr; } // q, zSuper
					
				}
			} // var id loop
			
		}
		
  } else { err = memErr; } // z
	
	return (err);
	
}



int  zlsim_FillImageZoneFromZoneVarIndex ( ZonePtr iz,  ZonePtr z, Integer varIndex ){
	
	Integer err = noErr;
	
#ifdef SYM1D	
	Integer i, id;	
	Integer l;
	
	Integer is;
	Integer ie;
	
	Integer ni;
	Integer nsi;
	
	Integer izd;
	Integer nti;
	Integer nsti;
	Integer iss; // super start offsets
	Integer ids; // data start offsets
	Q1DArr          q;
	Q1DArr     qSuper;
#endif SYM1D
	
#ifdef SYM2D	
	Integer i, j, id;	
	Integer l, m;
	
	Integer is, js;
	Integer ie, je;
	
	Integer ni, nj;
	Integer nsi, nsj;
	
	Integer izd, jzd;
	Integer nti, ntj;
	Integer nsti, nstj;
	Integer iss, jss; // super start offsets
	Integer ids, jds; // data start offsets
	Q2DArr          q;
	Q2DArr     qSuper;
#endif SYM2D
	
#ifdef SYM3D	
	Integer i, j, k, id;	
	Integer l, m, n;
	
	Integer is, js, ks;
	Integer ie, je, ke;
	
	Integer ni, nj, nk;
	Integer nsi, nsj, nsk;
	
	Integer izd, jzd, kzd;
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	Integer iss, jss, kss; // super start offsets
	Integer ids, jds, kds; // data start offsets
	Q3DArr          q;
	Q3DArr     qSuper;
#endif SYM3D
	
	Integer nR;
	Real    fx;
	
	ZonePtr    zSuper;
	Range         rng;
	
	if (z != NULL){
		
		nR     = z->zoneRefine;
		zSuper = iz;  // iz is always a plain block zone
		
		if ( zSuper != NULL ){
			
			rng    = z->zoneSuperRange;
			
			// get target zone grid size parameters.
			// The z->zoneData[id] may be a bare Ptr to a block 
			// without BCs, or be expanded to include BCs 
#ifdef SYM1D
			ni  = z->zoneGrid.n[kXDim];
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data has no BCs
			
			nti = z->zoneGrid.nt[kXDim];
			ids = z->zoneGrid.ns[kXDim];
			
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			iss  = zSuper->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			
			// zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
#endif SYM1D
#ifdef SYM2D
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data has no BCs
			nsj = zSuper->zoneGrid.n[kYDim]; // size of data is n not nt
			
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ids = z->zoneGrid.ns[kXDim];
			jds = z->zoneGrid.ns[kYDim];
			
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			nstj = zSuper->zoneGrid.nt[kYDim]; // 
			iss  = zSuper->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			jss  = zSuper->zoneGrid.ns[kYDim];
			
			// zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
#endif SYM2D
#ifdef SYM3D
			ni  = z->zoneGrid.n[kXDim];
			nj  = z->zoneGrid.n[kYDim];
			nk  = z->zoneGrid.n[kZDim];
			nsi = zSuper->zoneGrid.n[kXDim]; // super zone data has no BCs
			nsj = zSuper->zoneGrid.n[kYDim]; // size of data is n not nt
			nsk = zSuper->zoneGrid.n[kZDim]; // size of data is n not nt
			
			nti = z->zoneGrid.nt[kXDim];
			ntj = z->zoneGrid.nt[kYDim];
			ntk = z->zoneGrid.nt[kZDim];
			ids = z->zoneGrid.ns[kXDim];
			jds = z->zoneGrid.ns[kYDim];
			kds = z->zoneGrid.ns[kZDim];
			
			nsti = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			nstj = zSuper->zoneGrid.nt[kYDim]; // 
			nstk = zSuper->zoneGrid.nt[kZDim]; // 
			iss  = zSuper->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			jss  = zSuper->zoneGrid.ns[kYDim];
			kss  = zSuper->zoneGrid.ns[kZDim];
			
			// zone range on super zone in purely active area coordinates
			
			is = rng.n0[kXDim];
			ie = is+rng.rn[kXDim];
			
			js = rng.n0[kYDim];
			je = js+rng.rn[kYDim];
			
			ks = rng.n0[kZDim];
			ke = js+rng.rn[kZDim];
#endif SYM3D
			
			
			id = varIndex;
			
			if ( id < z->nQ ) {
				
				if (z->zoneData[id]!=NULL){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef SYM1D
					q = z->zoneData[id];
#endif SYM1D
					
#ifdef SYM2D
					zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif SYM2D
					
#ifdef SYM3D
					zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef ZONEDATA_ARE_MULTIDARRAYS
					q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
					
					if (q!=NULL){
						
#ifdef SYM1D
						fx = 1.0/((Real)nR);
#endif SYM1D
#ifdef SYM2D
						fx = 1.0/((Real)nR*(Real)nR);
#endif SYM2D
#ifdef SYM3D
						fx = 1.0/((Real)nR*(Real)nR*(Real)nR);
#endif SYM3D
						if ( zSuper->zoneType == kImageZoneType) {
							//                                          
							// wrap temp qSuper arrays around destination zone data in 2D and 3D, if needed
							// and , just copy reference if 1D as usual
							//
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
							qSuper = zSuper->zoneData[0];
#endif SYM1D
							
#ifdef SYM2D
							err = zls_SetQuantity2DArr(&qSuper,zSuper->zoneData[0], nstj, nsti);
#endif SYM2D
							
#ifdef SYM3D
							err = zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[0], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
							
#ifdef ZONEDATA_ARE_MULTIDARRAYS
							qSuper = zSuper->zoneData[0];
#endif ZONEDATA_ARE_MULTIDARRAYS
							
							if (qSuper!=NULL){
#ifdef SYM1D
								for(i = is; i<ie; i++){ // step in x in Super
									izd = ids+((i-is)*nR);
									qSuper [i+iss] = 0.0;
									for( l = 0; l<nR; l++){
										qSuper [i+iss] += q [(izd+l)];
									} //l
									qSuper [i+iss] *= fx;
								} //i
#endif SYM1D
								
#ifdef SYM2D
								
								for(j = js; j<je; j++){
									jzd = jds+((j-js)*nR);
									for(i = is; i<ie; i++){ // step in x in Super
										izd = ids+((i-is)*nR);
										qSuper[j+jss][i+iss] = 0.0;
										for( l = 0; l<nR; l++){
											for( m = 0; m<nR; m++){
												qSuper[j+jss][i+iss] += q[(jzd+l)][(izd+m)];
											} //m
										} //l
										qSuper[j+jss][i+iss] *= fx;
									} //i
								} //j
#endif SYM2D
								
#ifdef SYM3D
								for(k = ks; k<ke; k++){
									kzd = kds+((k-ks)*nR);
									for(j = js; j<je; j++){
										jzd = jds+((j-js)*nR);
										for(i = is; i<ie; i++){ // step in x in Super
											izd = ids+((i-is)*nR);
											qSuper[k+kss][j+jss][i+iss] = 0.0;
											for( l = 0; l<nR; l++){
												for( m = 0; m<nR; m++){
													for( n = 0; n<nR; n++){
														qSuper[k+jss][j+jss][i+iss] += q[(kzd+l)][(jzd+m)][(izd+n)];
													} //n
												} //m
											} //l
											qSuper[k+kss][j+jss][i+iss] *= fx;
										} //i
									} //j
								} //k
#endif SYM3D
								
								
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
								zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
								zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
								
							}
							
						}  // iz is correct kind
						else { err = kindErr; }
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
						zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
						zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
					} else { err = memErr; } // q, zSuper
					
				}
			} // var id OK
			
		}
		
  } else { err = memErr; } // z
	
	return (err);
	
}

int  zlsim_FillImageZoneFromData ( ZonePtr iz,  QPtr data, GridCoords grid, Range destRng, Integer nR, Integer dataHasBCs ){
	//
	// zlsim_FillImageZoneFromData.
	//
	// like above, but uses provided grid to put plain block data, often
	// from outside a zone to into an image zone.  Most useful if the data
	// only represents a sub area of the image zone.  The image zone is assumed to 
	// have just one anon var, and the data input is a simple block
	// and the grid provided is consistent withe the image zone being a valid
	// super grid of the grid, ie boundary cells if any are satisfied.
	// 
	// nR is the refinement scaling from the data to the image zone, usually from 
	// the original zone the data is from.  If in doubt, use 1 for a 1:1 copy.
	// should be consistent with the original zonedata zone -> image zone 
	// coordinate relationship or the copy could screw up badly and give
	// strange results or worse.  See save level images for an example.
	//
	// Finally if the data is truely external it may or may not have BCs included.
	// Since this routine can't query data for size, the caller needs to pass
	//   dataHasBCs = kZLS_DataNoBCs if the data is just the active area, and 
	//   dataHasBCs = kZLS_DataHasBCs if the data includes BCs, consistent with zones.
	// 
	//
	// ie get data or compute data from a zone z.  Take nR and grid from z, which
	// already has iz as it's super zone, and pass the data, the gridCoords and nR.
	//
	// Designed this way so that data doesn't have to come from a zone for flexibility,
	// although it sometimes will.  Consider a Fill Zone with zone call if possible in that
	// case as it will be easier and less memory intensive
	//
	//
	// Useage: when a derived quantity is obtained from a zone and the results
	// are being merged into a single image zone, the data returned from a 
	// GetWQuantity call can be passed here, whether ic came from a literal var or
	// was computed on the fly.  Use the grid and superRange from the original zone to provide
	// the coordinates, and assume image zone coords are already set up properly in 
	// initiialization
	//
	
	Integer err = noErr;
	
#ifdef SYM1D	
	Integer i;	
	Integer l;
	
	Integer is;
	Integer ie;
	
	Integer ni;
	Integer nsi;
	
	Integer izd;
	Integer nti;
	Integer nsti;
	Integer iss; // super start offsets
	Integer ids; // data start offsets
	Q1DArr          q;
	Q1DArr     qSuper;
#endif SYM1D
	
#ifdef SYM2D	
	Integer i, j;	
	Integer l, m;
	
	Integer is, js;
	Integer ie, je;
	
	Integer ni, nj;
	Integer nsi, nsj;
	
	Integer izd, jzd;
	Integer nti, ntj;
	Integer nsti, nstj;
	Integer iss, jss; // super start offsets
	Integer ids, jds; // data start offsets
	Q2DArr          q;
	Q2DArr     qSuper;
#endif SYM2D
	
#ifdef SYM3D	
	Integer i, j, k;	
	Integer l, m, n;
	
	Integer is, js, ks;
	Integer ie, je, ke;
	
	Integer ni, nj, nk;
	Integer nsi, nsj, nsk;
	
	Integer izd, jzd, kzd;
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	Integer iss, jss, kss; // super start offsets
	Integer ids, jds, kds; // data start offsets
	Q3DArr          q;
	Q3DArr     qSuper;
#endif SYM3D
	
	if ( iz != NULL ){
		
		Real           fx;
		
		// get target zone grid size parameters.
		// The z->zoneData[id] may be a bare Ptr to a block 
		// without BCs, or be expanded to include BCs 
#ifdef SYM1D
		
		ni  = grid.n [kXDim];
		nti = ni;
		ids = 0;
		
		if ( dataHasBCs == kZLS_DataHasBCs ) {
			nti = grid.nt[kXDim];
			ids = grid.ns[kXDim];
		}
		
		nsi  = iz->zoneGrid.n [kXDim]; // if super zone data has no BCs
		nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BCs, so size is nt not just n
		iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
		
		// destination range on target zone in purely active area coordinates
		
		is = destRng.n0[kXDim];
		ie = is+destRng.rn[kXDim];
		
#endif SYM1D
#ifdef SYM2D
		ni  = grid.n [kXDim];
		nj  = grid.n [kYDim];
		nti = ni;
		ntj = nj;
		ids = 0;
		jds = 0;
		
		if ( dataHasBCs == kZLS_DataHasBCs ) {
			nti = grid.nt[kXDim];
			ntj = grid.nt[kYDim];
			ids = grid.ns[kXDim];
			jds = grid.ns[kYDim];
		} 
		
		nsi  = iz->zoneGrid.n [kXDim]; // super zone data has no BCs
		nsj  = iz->zoneGrid.n [kYDim]; // size of data is n not nt
		nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
		nstj = iz->zoneGrid.nt[kYDim]; // 
		iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
		jss  = iz->zoneGrid.ns[kYDim];
		
		// destination range on target zone in purely active area coordinates
		
		is = destRng.n0[kXDim];
		ie = is+destRng.rn[kXDim];
		
		js = destRng.n0[kYDim];
		je = js+destRng.rn[kYDim];
		
		
#endif SYM2D
#ifdef SYM3D
		ni  =grid.n[kXDim];
		nj  =grid.n[kYDim];
		nk  =grid.n[kZDim];
		
		nti = ni;
		ntj = nj;
		ntk = nk;
		
		ids = 0;
		jds = 0;
		kds = 0;
		
		if ( dataHasBCs == kZLS_DataHasBCs ) {
			
			nti =grid.nt[kXDim];
			ntj =grid.nt[kYDim];
			ntk =grid.nt[kZDim];
			ids =grid.ns[kXDim];
			jds =grid.ns[kYDim];
			kds =grid.ns[kZDim];
			
		}
		
		nsi = iz->zoneGrid.n[kXDim]; // super zone data has no BCs
		nsj = iz->zoneGrid.n[kYDim]; // size of data is n not nt
		nsk = iz->zoneGrid.n[kZDim]; // size of data is n not nt
		
		nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
		nstj = iz->zoneGrid.nt[kYDim]; // 
		nstk = iz->zoneGrid.nt[kZDim]; // 
		
		iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
		jss  = iz->zoneGrid.ns[kYDim];
		kss  = iz->zoneGrid.ns[kZDim];
		
		// destination range on target zone in purely active area coordinates
		
		is = destRng.n0[kXDim];
		ie = is+destRng.rn[kXDim];
		
		js = destRng.n0[kYDim];
		je = js+destRng.rn[kYDim];
		
		ks = destRng.n0[kZDim];
		ke = js+destRng.rn[kZDim];
#endif SYM3D
		
		
		if ( data != NULL ){
			//
			// provided data is assumed to always be a plain block
			// and may not be the same as the super zone format since it 
			// may even come from outsize the zone heirachy or even a file.
			// So wrap multi-d helper structures in 2D and 3D everytime.
			//
#ifdef SYM1D
			q = data;
#endif SYM1D
			
#ifdef SYM2D
			zls_SetQuantity2DArr(&q, data, ntj, nti); // sizses adjusted for data BCs as needed by dataHasBCs
#endif SYM2D
			
#ifdef SYM3D
			zls_SetQuantity3DArr(&q, data, ntk, ntj, nti); // sizes adjusted for data BCs as needed by dataHasBCs
#endif SYM3D
			
			if (q!=NULL){
				
#ifdef SYM1D
				fx = 1.0/((Real)nR);
#endif SYM1D
#ifdef SYM2D
				fx = 1.0/((Real)nR*(Real)nR);
#endif SYM2D
#ifdef SYM3D
				fx = 1.0/((Real)nR*(Real)nR*(Real)nR);
#endif SYM3D
				if ( iz->zoneType == kImageZoneType ) {
					//                                          
					// wrap temp qSuper arrays around destination zone data in 2D and 3D, *if needed*
					// and , just copy reference if 1D as usual
					//
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
					qSuper = iz->zoneData[0];
#endif SYM1D
					
#ifdef SYM2D
					err = zls_SetQuantity2DArr(&qSuper,iz->zoneData[0], nstj, nsti);
#endif SYM2D
					
#ifdef SYM3D
					err = zls_SetQuantity3DArr(&qSuper,iz->zoneData[0], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef ZONEDATA_ARE_MULTIDARRAYS
					qSuper = iz->zoneData[0];
#endif ZONEDATA_ARE_MULTIDARRAYS
					
					if (qSuper!=NULL){
#ifdef SYM1D
						for(i = is; i<ie; i++){ // step in x in Super
							izd = ids+((i-is)*nR);
							qSuper [i+iss] = 0.0;
							for( l = 0; l<nR; l++){
								qSuper [i+iss] += q [(izd+l)];
							} //l
							qSuper [i+iss] *= fx;
						} //i
#endif SYM1D
						
#ifdef SYM2D
						
						for(j = js; j<je; j++){
							jzd = jds+((j-js)*nR);
							for(i = is; i<ie; i++){ // step in x in Super
								izd = ids+((i-is)*nR);
								qSuper[j+jss][i+iss] = 0.0;
								for( l = 0; l<nR; l++){
									for( m = 0; m<nR; m++){
										qSuper[j+jss][i+iss] += q[(jzd+l)][(izd+m)];
									} //m
								} //l
								qSuper[j+jss][i+iss] *= fx;
							} //i
						} //j
#endif SYM2D
						
#ifdef SYM3D
						for(k = ks; k<ke; k++){
							kzd = kds+((k-ks)*nR);
							for(j = js; j<je; j++){
								jzd = jds+((j-js)*nR);
								for(i = is; i<ie; i++){ // step in x in Super
									izd = ids+((i-is)*nR);
									qSuper[k+kss][j+jss][i+iss] = 0.0;
									for( l = 0; l<nR; l++){
										for( m = 0; m<nR; m++){
											for( n = 0; n<nR; n++){
												qSuper[k+jss][j+jss][i+iss] += q[(kzd+l)][(jzd+m)][(izd+n)];
											} //n
										} //m
									} //l
									qSuper[k+kss][j+jss][i+iss] *= fx;
								} //i
							} //j
						} //k
#endif SYM3D
						
						//  only release if super zone data is plain not multi-d form
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
						zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
						zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
					}
					
				}   else { err = kindErr; } // iz is correct kind
				
				
				//  input data are assumed to a plain block, so multi-d structures from above always need to be released
#ifdef SYM2D
				zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
				zls_ReleaseQuantity3DArr(q);
#endif SYM3D
				
				
			} else { err = memErr; } // q
			
		} else { err = memErr; } // data
		
	} else { err = memErr; } // iz
  
  return (err);
  
}

int  zlsim_FillLevelImageFromDataAndZone ( ZonePtr iz,  QPtr data, Integer dataHasBCs, ZonePtr coordZone ){
	//
	// zlsim_FillLevelImageFromDataAndZone like zlsim_FillImageZoneFromData, but with simpler args, using the
	// coords zone to provide the grid, LevelRange and Refine.  Only works if iz is a genuine superzone
	// of the coords zone.  if not then setting up manually with zlsim_FillImageZoneFromData may be the
	// best option.  If working, one coordZone can be shared by a number of external data blocks, and 
	// control the fill mapping by proxy without having to create new zones for each QPtr.
	//
	// Useage: when a derived quantity is obtained from a zone and the results
	// are being merged into a single image zone, the data returned from a 
	// GetWQuantity call can be passed here, whether it came from a literal var or
	// was computed on the fly. 
	//
	
	Integer err = noErr;
	
#ifdef SYM1D	
	Integer i;	
	Integer l;
	
	Integer is;
	Integer ie;
	
	Integer ni;
	Integer nsi;
	
	Integer izd;
	Integer nti;
	Integer nsti;
	Integer iss; // super start offsets
	Integer ids; // data start offsets
	Q1DArr          q;
	Q1DArr     qSuper;
#endif SYM1D
	
#ifdef SYM2D	
	Integer i, j;	
	Integer l, m;
	
	Integer is, js;
	Integer ie, je;
	
	Integer ni, nj;
	Integer nsi, nsj;
	
	Integer izd, jzd;
	Integer nti, ntj;
	Integer nsti, nstj;
	Integer iss, jss; // super start offsets
	Integer ids, jds; // data start offsets
	Q2DArr          q;
	Q2DArr     qSuper;
#endif SYM2D
	
#ifdef SYM3D	
	Integer i, j, k;	
	Integer l, m, n;
	
	Integer is, js, ks;
	Integer ie, je, ke;
	
	Integer ni, nj, nk;
	Integer nsi, nsj, nsk;
	
	Integer izd, jzd, kzd;
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	Integer iss, jss, kss; // super start offsets
	Integer ids, jds, kds; // data start offsets
	Q3DArr          q;
	Q3DArr     qSuper;
#endif SYM3D
	
	if ( coordZone != NULL) {
		
		GridCoords grid = coordZone->zoneGrid;
		Range destRng   = coordZone->zoneLevelRange;
		Integer nR      = 1;
		
		if ( iz != NULL ){
			
			Real           fx;
			
			// get target zone grid size parameters.
			// The z->zoneData[id] may be a bare Ptr to a block 
			// without BCs, or be expanded to include BCs 
#ifdef SYM1D
			
			ni  = grid.n [kXDim];
			nti = ni;
			ids = 0;
			
			if ( dataHasBCs == kZLS_DataHasBCs ) {
				nti = grid.nt[kXDim];
				ids = grid.ns[kXDim];
			}
			
			nsi  = iz->zoneGrid.n [kXDim]; // if super zone data has no BCs
			nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BCs, so size is nt not just n
			iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			
			// destination range on target zone in purely active area coordinates
			
			is = destRng.n0[kXDim];
			ie = is+destRng.rn[kXDim];
			
#endif SYM1D
#ifdef SYM2D
			ni  = grid.n [kXDim];
			nj  = grid.n [kYDim];
			nti = ni;
			ntj = nj;
			ids = 0;
			jds = 0;
			
			if ( dataHasBCs == kZLS_DataHasBCs ) {
				nti = grid.nt[kXDim];
				ntj = grid.nt[kYDim];
				ids = grid.ns[kXDim];
				jds = grid.ns[kYDim];
			} 
			
			nsi  = iz->zoneGrid.n [kXDim]; // super zone data has no BCs
			nsj  = iz->zoneGrid.n [kYDim]; // size of data is n not nt
			nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			nstj = iz->zoneGrid.nt[kYDim]; // 
			iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			jss  = iz->zoneGrid.ns[kYDim];
			
			// destination range on target zone in purely active area coordinates
			
			is = destRng.n0[kXDim];
			ie = is+destRng.rn[kXDim];
			
			js = destRng.n0[kYDim];
			je = js+destRng.rn[kYDim];
			
			
#endif SYM2D
#ifdef SYM3D
			ni  =grid.n[kXDim];
			nj  =grid.n[kYDim];
			nk  =grid.n[kZDim];
			
			nti = ni;
			ntj = nj;
			ntk = nk;
			
			ids = 0;
			jds = 0;
			kds = 0;
			
			if ( dataHasBCs == kZLS_DataHasBCs ) {
				
				nti =grid.nt[kXDim];
				ntj =grid.nt[kYDim];
				ntk =grid.nt[kZDim];
				ids =grid.ns[kXDim];
				jds =grid.ns[kYDim];
				kds =grid.ns[kZDim];
				
			}
			
			nsi = iz->zoneGrid.n[kXDim]; // super zone data has no BCs
			nsj = iz->zoneGrid.n[kYDim]; // size of data is n not nt
			nsk = iz->zoneGrid.n[kZDim]; // size of data is n not nt
			
			nsti = iz->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			nstj = iz->zoneGrid.nt[kYDim]; // 
			nstk = iz->zoneGrid.nt[kZDim]; // 
			
			iss  = iz->zoneGrid.ns[kXDim]; // also we fill from ns, not 0
			jss  = iz->zoneGrid.ns[kYDim];
			kss  = iz->zoneGrid.ns[kZDim];
			
			// destination range on target zone in purely active area coordinates
			
			is = destRng.n0[kXDim];
			ie = is+destRng.rn[kXDim];
			
			js = destRng.n0[kYDim];
			je = js+destRng.rn[kYDim];
			
			ks = destRng.n0[kZDim];
			ke = ks+destRng.rn[kZDim];
#endif SYM3D
			
      
			if ( data != NULL ){
				//
				// provided data is assumed to always be a plain block
				// and may not be the same as the super zone format since it 
				// may even come from outsize the zone heirachy or even a file.
				// So wrap multi-d helper structures in 2D and 3D everytime.
				//
#ifdef SYM1D
				q = data;
#endif SYM1D
				
#ifdef SYM2D
				zls_SetQuantity2DArr(&q, data, ntj, nti); // sizses adjusted for data BCs as needed by dataHasBCs
#endif SYM2D
				
#ifdef SYM3D
				zls_SetQuantity3DArr(&q, data, ntk, ntj, nti); // sizes adjusted for data BCs as needed by dataHasBCs
#endif SYM3D
				
				if (q!=NULL){
					
#ifdef SYM1D
					fx = 1.0/((Real)nR);
#endif SYM1D
#ifdef SYM2D
					fx = 1.0/((Real)nR*(Real)nR);
#endif SYM2D
#ifdef SYM3D
					fx = 1.0/((Real)nR*(Real)nR*(Real)nR);
#endif SYM3D
					if ( iz->zoneType == kImageZoneType ) {
						//                                          
						// wrap temp qSuper arrays around destination zone data in 2D and 3D, *if needed*
						// and , just copy reference if 1D as usual
						//
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
						qSuper = iz->zoneData[0];
#endif SYM1D
						
#ifdef SYM2D
						err = zls_SetQuantity2DArr(&qSuper,iz->zoneData[0], nstj, nsti);
#endif SYM2D
						
#ifdef SYM3D
						err = zls_SetQuantity3DArr(&qSuper,iz->zoneData[0], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
#ifdef ZONEDATA_ARE_MULTIDARRAYS
						qSuper = iz->zoneData[0];
#endif ZONEDATA_ARE_MULTIDARRAYS
						
						if (qSuper!=NULL){
#ifdef SYM1D
							for(i = is; i<ie; i++){ // step in x in Super
								izd = ids+((i-is)*nR);
								qSuper [i+iss] = 0.0;
								for( l = 0; l<nR; l++){
									qSuper [i+iss] += q [(izd+l)];
								} //l
								qSuper [i+iss] *= fx;
							} //i
#endif SYM1D
							
#ifdef SYM2D
							
							for(j = js; j<je; j++){
								jzd = jds+((j-js)*nR);
								for(i = is; i<ie; i++){ // step in x in Super
									izd = ids+((i-is)*nR);
									qSuper[j+jss][i+iss] = 0.0;
									for( l = 0; l<nR; l++){
										for( m = 0; m<nR; m++){
											qSuper[j+jss][i+iss] += q[(jzd+l)][(izd+m)];
										} //m
									} //l
									qSuper[j+jss][i+iss] *= fx;
								} //i
							} //j
#endif SYM2D
							
#ifdef SYM3D
							for(k = ks; k<ke; k++){
								kzd = kds+((k-ks)*nR);
								for(j = js; j<je; j++){
									jzd = jds+((j-js)*nR);
									for(i = is; i<ie; i++){ // step in x in Super
										izd = ids+((i-is)*nR);
										qSuper[k+kss][j+jss][i+iss] = 0.0;
										for( l = 0; l<nR; l++){
											for( m = 0; m<nR; m++){
												for( n = 0; n<nR; n++){
													qSuper[k+jss][j+jss][i+iss] += q[(kzd+l)][(jzd+m)][(izd+n)];
												} //n
											} //m
										} //l
										qSuper[k+kss][j+jss][i+iss] *= fx;
									} //i
								} //j
							} //k
#endif SYM3D
							
							//  only release if super zone data is plain not multi-d form
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
							zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
							zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
							
						}
						
					}   else { err = kindErr; } // iz is correct kind
					
					
					//  input data are assumed to a plain block, so multi-d structures from above always need to be released
#ifdef SYM2D
					zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
					zls_ReleaseQuantity3DArr(q);
#endif SYM3D
					
					
				} else { err = memErr; } // q
				
			} else { err = memErr; } // data
			
		} else { err = memErr; } // iz
		
	} else { err = memErr; } // coordZone
	
	return (err);
	
}




int  zlsbc_FillZoneBCsFromNextZone ( ZonePtr z, int direction ){
	
	Integer err = noErr;
	
	Integer nbc;
	ZonePtr    zNext;
	
#ifdef SYM1D
	Integer i, id;	
	
	Integer is, ie;
	Integer nis, nie;
	Integer nti, nsti;
	
	Q1DArr     qNext;
	Q1DArr          q;
#endif SYM1D
	
#ifdef SYM2D
	Integer i, j, id;	
	
	Integer is, js;
	Integer ie, je;
	Integer nis, njs;
	Integer nie, nje;
	
	Integer nti, ntj;
	Integer nsti, nstj;
	
	Q2DArr     qNext;
	Q2DArr          q;
#endif SYM2D
	
#ifdef SYM3D
	Integer i, j, k, id;	
	
	Integer is, js, ks;
	Integer ie, je, ke;
	Integer nis, njs,nks;
	Integer nie, nje, nke;
	
	Integer nti, ntj, ntk;
	Integer nsti, nstj, nstk;
	
	Q3DArr     qNext;
	Q3DArr          q;
#endif SYM3D
	
	
	if (z != NULL){
		
		nbc    = z->zoneBCells;
		zNext = z->zoneNext[direction];
		
		if (zNext != NULL){
			
			// get target zone grid size parameters.  Common boundary dimension and nBC
			// must be the same, as this is not checked here.
#ifdef SYM1D
			is  =  z->zoneGrid.ns[kXDim];
			ie  =  z->zoneGrid.ne[kXDim];
			nti =  z->zoneGrid.nt[kXDim];
			
			nis  = zNext->zoneGrid.ns[kXDim];
			nie  = zNext->zoneGrid.ne[kXDim];
			nsti = zNext->zoneGrid.nt[kXDim]; // Next zone data has BC, so size is nt not just n
#endif SYM1D
			
#ifdef SYM2D
			is  =  z->zoneGrid.ns[kXDim];
			ie  =  z->zoneGrid.ne[kXDim];
			nti =  z->zoneGrid.nt[kXDim];
			
			js  =  z->zoneGrid.ns[kYDim];
			je  =  z->zoneGrid.ne[kYDim];
			ntj =  z->zoneGrid.nt[kYDim];
			
			nis  = zNext->zoneGrid.ns[kXDim];
			nie  = zNext->zoneGrid.ne[kXDim];
			nsti = zNext->zoneGrid.nt[kXDim]; // Next zone data has BC, so size is nt not just n
      
			njs  = zNext->zoneGrid.ns[kYDim];
			nje  = zNext->zoneGrid.ne[kYDim];
			nstj = zNext->zoneGrid.nt[kYDim]; // 
#endif SYM2D
			
#ifdef SYM3D
			is  =  z->zoneGrid.ns[kXDim];
			ie  =  z->zoneGrid.ne[kXDim];
			nti =  z->zoneGrid.nt[kXDim];
			
			js  =  z->zoneGrid.ns[kYDim];
			je  =  z->zoneGrid.ne[kYDim];
			ntj =  z->zoneGrid.nt[kYDim];
			
			ks  =  z->zoneGrid.ns[kZDim];
			ke  =  z->zoneGrid.ne[kZDim];
			ntk =  z->zoneGrid.nt[kZDim];
			
			nis  = zNext->zoneGrid.ns[kXDim];
			nie  = zNext->zoneGrid.ne[kXDim];
			nsti = zNext->zoneGrid.nt[kXDim]; // Next zone data has BCs, so size is nt not just n
			
			njs  = zNext->zoneGrid.ns[kYDim];
			nje  = zNext->zoneGrid.ne[kYDim];
			nstj = zNext->zoneGrid.nt[kYDim]; // 
			
			nks  = zNext->zoneGrid.ns[kZDim];
			nke  = zNext->zoneGrid.ne[kZDim];
			nstk = zNext->zoneGrid.nt[kZDim]; // 
      
#endif SYM3D
			
			for (id = 0; id <  z->nQ; id++){
				
				if (z->zoneData[id]!=NULL){
					
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
					q = z->zoneData[id];
#endif SYM1D
#ifdef SYM2D
					zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif SYM2D
#ifdef SYM3D
					zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
					
#ifdef ZONEDATA_ARE_MULTIDARRAYS
					q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
					
					if (q!=NULL){
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
						qNext = zNext->zoneData[id];
#endif SYM1D
#ifdef SYM2D
						err = zls_SetQuantity2DArr(&qNext,zNext->zoneData[id], nstj, nsti);
#endif SYM2D
#ifdef SYM3D
						err = zls_SetQuantity3DArr(&qNext,zNext->zoneData[id], nstk, nstj, nsti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
#ifdef ZONEDATA_ARE_MULTIDARRAYS
						qNext = zNext->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
						
						if (qNext!=NULL){
#ifdef SYM1D
							switch (direction) {
								case kDirRight:
									for(i = 0; i<nbc; i++){
										q[ie+i] = qNext[nis+i];
									} //i
									break;
                  case kDirLeft:
									for(i = 0; i<nbc; i++){
										q[i] = qNext[nie-nbc+i];
									} //i
									break;
                  default:
									zlsio_ErrPrintf( " WARNING: FillZoneBCsFromNextZone, bad direction %d \n", direction);
							}
#endif SYM1D
							
#ifdef SYM2D
							switch (direction) {
								case kDirRight:
									
									for(j = 0; j<ntj; j++){
										for(i = 0; i<nbc; i++){
											q[j][ie+i] = qNext[j][nis+i];
										} //i
									} //j
									
									break;
                  case kDirLeft:
									
									for(j = 0; j<ntj; j++){
										for(i = 0; i<nbc; i++){
											q[j][i]    = qNext[j][nie-nbc+i];
										} //i
									} //j
									
									break;
                  case kDirUp:
									
									for(j = 0; j<nbc; j++){
										for(i = 0; i<nti; i++){
											q[je+j][i] = qNext[njs+j][i];
										} //i
									} //j
									
									break;
                  case kDirDown:
									
									for(j = 0; j<nbc; j++){
										for(i = 0; i<nti; i++){
											q[j][i]    = qNext[nje-nbc+j][i];
										} //i
									} //j
									
									break;	
                  default:
									zlsio_ErrPrintf( " WARNING: FillZoneBCsFromNextZone, bad direction %d \n", direction);
							}
#endif SYM2D
							
#ifdef SYM3D
#if (ZLS_DEBUG > 2)
							zlsio_Printf( " Zone link %d to %d, direction %d \n", z->zoneID, zNext->zoneID, direction);
#endif ZLS_DEBUG			
							switch (direction) {
                  
								case kDirRight: // larger i
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " i -> , ks, ke : %d ,%d \n", 0, ntk);
									zlsio_Printf( " i -> , js, je : %d ,%d \n", 0, ntj);
									zlsio_Printf( " i -> , is, ie : %d ,%d \n", is, ie);
#endif ZLS_DEBUG			
									for(k = 0; k<ntk; k++){
										for(j = 0; j<ntj; j++){
											for(i = 0; i<nbc; i++){
												q[k][j][ie+i] = qNext[k][j][nis+i];
											} //i
										} //j
									} //k
                  break;
                  case kDirLeft: // smaller i
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " <- i , ks, ke : %d ,%d \n", 0, ntk);
									zlsio_Printf( " <- i , js, je : %d ,%d \n", 0, ntj);
									zlsio_Printf( " <- i , is, ie : %d ,%d \n", is, ie);
#endif ZLS_DEBUG			
									for(k = 0; k<ntk; k++){
										for(j = 0; j<ntj; j++){
											for(i = 0; i<nbc; i++){
												q[k][j][i] = qNext[k][j][nie-nbc+i];
											} //i
										} //j
									} //k
                  break;
									
                  case kDirUp: // larger j
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " j -> , ks, ke : %d ,%d \n",  0, ntk);
									zlsio_Printf( " j -> , js, je : %d ,%d \n", js, je);
									zlsio_Printf( " j -> , is, ie : %d ,%d \n",  0, nti);
#endif ZLS_DEBUG			
									for(k = 0; k<ntk; k++){
										for(j = 0; j<nbc; j++){
											for(i = 0; i<nti; i++){
												q[k][je+j][i] = qNext[k][njs+j][i];
											} //i
										} //j
									} //k
                  break;
                  case kDirDown: // smaller j
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " <- j , ks, ke : %d ,%d \n", 0, ntk);
									zlsio_Printf( " <- j , js, je : %d ,%d \n", js, je);
									zlsio_Printf( " <- j , is, ie : %d ,%d \n", 0, nti);
#endif ZLS_DEBUG			
									for(k = 0; k<ntk; k++){
										for(j = 0; j<nbc; j++){
											for(i = 0; i<nti; i++){
												q[k][j][i] = qNext[k][nje-nbc+j][i];
											} //i
										} //j
									} //k
                  break;	
									
                  case kDirHither: // larger k
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " k -> , ks, ke : %d ,%d \n", ks, ke);
									zlsio_Printf( " k -> , js, je : %d ,%d \n", 0, ntj);
									zlsio_Printf( " k -> , is, ie : %d ,%d \n", 0, nti);
#endif ZLS_DEBUG			
									for(k = 0; k<nbc; k++){
										for(j = 0; j<ntj; j++){
											for(i = 0; i<nti; i++){
												q[ke+k][j][i] = qNext[nks+k][j][i];
											} //i
										} //j
									} //k
                  break;
                  case  kDirYon: // smaller k
#if (ZLS_DEBUG > 2)
									zlsio_Printf( " <- k , ks, ke : %d ,%d \n", ks, ke);
									zlsio_Printf( " <- k , js, je : %d ,%d \n", 0, ntj);
									zlsio_Printf( " <- k , is, ie : %d ,%d \n", 0, nti);
#endif ZLS_DEBUG			
									for(k = 0; k<nbc; k++){
										for(j = 0; j<ntj; j++){
											for(i = 0; i<nti; i++){
												q[k][j][i] = qNext[nke-nbc+k][j][i];
											} //i
										} //j
									} //k
                  break;	
									
                  default:
									zlsio_ErrPrintf( " WARNING: Fill2DZoneBCsFromNextZone, bad direction %d \n", direction);
							}
#endif SYM3D
							
							
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
							zls_ReleaseQuantity2DArr(qNext);
#endif SYM2D
#ifdef SYM3D
							zls_ReleaseQuantity3DArr(qNext);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						}
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
						zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
						zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
						
					} else { err = memErr; } // q, zNext
					
				}
			} // var id loop
			
		}
		
	} else { err = memErr; } // z
	
	return (err);
	
}


int  zlsbc_ImposeZoneLinks ( ZonePtr z,  int direction ){
	Integer err = noErr;
	
	if (z != NULL){
    Integer bndryType;  // what sort of boundary, super, link, calc
		bndryType = z->zoneBCType[direction]&kBoundaryTypeFlagNextZ; 
		if  (bndryType == kBoundaryTypeFlagNextZ) {
			if (z->zoneNext[direction] != NULL)
				err = zlsbc_FillZoneBCsFromNextZone(z, direction); 
		}
		
	} else { err = memErr; } // z
	
	return (err);
}


int  zls_FillZoneFromSuperZone1D( ZonePtr z, int inclBCs );
int  zls_FillZoneFromSuperZone1D( ZonePtr z, int inclBCs ){
#ifndef SYM1D
  
  return( dimErr );
  
#else
  
	Integer err = noErr;
	
	Integer    i,   id;	 //  loop counters
	Integer    l;        //  sub cell loop counters
	
	Integer         nR;
	Integer         nG;
	Integer       nsbc;   
	Range          rng;
	Integer     interp;
	Real          f, a;
	
	Integer  izd;     // i and j in zone data
	Integer  isd;     // i and j in super data
	
	InPtr           fx;		// interpolation fraction array
	V1DArr        v1dx;   // extracted vars/coeffs for 
	
	Integer   ia,  i_0,  nir; // super range in active area coord
	Integer  izs;     // zone start offsets
	Integer  iss;     // super start offsets
	
	Integer   nsi ;     // super data full range for making vars
	Integer  nti  ;     // zonedata full range
	Integer nsti  ;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q1DArr      qSuper;
	Q1DArr           q;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nG     = z->zoneBCells;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          iss    = zSuper->zoneGrid.ns[kXDim];
          izs    = z->zoneGrid.ns[kXDim];
          nsi   = zSuper->zoneGrid.n[kXDim]; 
          
          // get Super zone area parameters
          nti    = z->zoneGrid.nt[kXDim];
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          
          
          if (inclBCs == kInterpBCsFromSuper){
            //
            // zone range on super zone in active area coordinates, expanded to include 
            // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
            // cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
            // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
            // range by 2 nsbcs.
            // 
            // coordinate in q and vars = i + i_0, j + j_0
            //
            ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;                      // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
            
            izs = 0;
            
          } else {
            ia  = rng.n0[kXDim];       // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;              // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim];       // nir is the range of the i loop in super cells
          }
          
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              if (zSuper->zoneData[id]!=NULL){
                
                q      = z->zoneData[id];
                qSuper = zSuper->zoneData[id];
                
                if (q != NULL){
                  if (qSuper != NULL){
                    
                    switch (interp){
                        
                      case  kInterpNN: // simple rebin
                        
                        for(i = 0; i < nir; i++){
                          izd = izs+(i*nR);
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            q [(izd+l)] =  qSuper [isd];
                          } //l
                        } // i
                        
                        break; // interp == kInterpNN
                        
                        case kInterpPLM: // Piecewise Linear Method interpolation 
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewInterp1D (&fx, nsti);
                        
                        if (err == noErr) {
                          
                          zls_SetVar1DArrQ1DArr            ( v1dx, qSuper );
                          err = zlsppm_CalcCoeffsVar1DArr  ( v1dx, NULL);
                          err = zlsppm_CalcLinInterpVar1D   ( v1dx, fx, nR);
                          for(i = 0; i<nir; i++){ // step in x in Super
                            izd = izs+(i*nR); 
                            isd = (i+i_0);
                            a   = v1dx.v[isd].a;
                            for( l = 0; l<nR; l++){
                              f = fx[isd].c[l];
                              q[(izd+l)] = a  + f; 
                            } //l
                          } //i
                          
                          zls_DisposeVar1DArr(v1dx);
                          zls_DisposeInterp(fx); fx = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill1DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break;
                        
                        case kInterpPPM: // PPM interpolation same as ppmlr
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewInterp1D (&fx, nsti);
                        
                        if (err == noErr) {
                          
                          zls_SetVar1DArrQ1DArr            ( v1dx, qSuper);
                          err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                          err = zlsppm_CalcInterpVar1D     ( v1dx, fx, nR);
                          for(i = 0; i<nir; i++){ // step in x in Super
                            izd = izs+(i*nR); 
                            isd = (i+i_0);
                            a   = v1dx.v[isd].a;
                            for( l = 0; l<nR; l++){
                              f = fx[isd].c[l];
                              q[(izd+l)] = a  + f; 
                            } //l
                          } //i
                          
                          zls_DisposeVar1DArr(v1dx);
                          zls_DisposeInterp(fx); fx = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill1DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break; // kInterpPPM
                        
                        default:
                        
                        err = nyiErr;
                        zlsio_Printf( " Fill1DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                        
                    }
                    
                    
                  } // qSuper		
                  
                } // q
                
              } // zSuperdata
            } // zdata
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // zSuper
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
  
#endif SYM1D
	
}


int  zls_FillZoneFromSuperZone2D( ZonePtr z, int inclBCs );
int  zls_FillZoneFromSuperZone2D( ZonePtr z, int inclBCs ){
#ifndef SYM2D
  return( dimErr );
#else
	Integer err = noErr;
	
	Integer    i,    j,   id;	 //  loop counters
	Integer    l,    m;          //  sub cell loop counters
	
	Integer         nR;
	Integer         nG;
	Integer       nsbc;   
	Range          rng;
	Integer     interp;
	
	Real f, a;
	
	Integer  izd,  jzd;     // i and j in zone data
	Integer  isd,  jsd;     // i and j in super data
	
	InPtr        fx, fy;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
	V1DArr        v1dy;		// x and y rows of the  super zone
	
	Integer   ia,  i_0,  nir; // super range in active area coord
	Integer   ja,  j_0,  njr; // super range in active area coord
	Integer  izs,  jzs;     // zone start offsets
	Integer  iss,  jss;     // super start offsets
	
	Integer   nsi, nsj;     // super data full range for making vars
	Integer  nti,  ntj;     // zonedata full range
	Integer nsti, nstj;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q2DArr      qSuper;
	Q2DArr           q;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nG     = z->zoneBCells;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          izs    = z->zoneGrid.ns[kXDim];
          jzs    = z->zoneGrid.ns[kYDim];
          nti    = z->zoneGrid.nt[kXDim];
          ntj    = z->zoneGrid.nt[kYDim];
          
          // get Super zone area parameters
          iss    = zSuper->zoneGrid.ns[kXDim];
          jss    = zSuper->zoneGrid.ns[kYDim];
          nsi    = zSuper->zoneGrid.n[kXDim]; 
          nsj    = zSuper->zoneGrid.n[kYDim];
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          nstj   = zSuper->zoneGrid.nt[kYDim]; // 
          
          
          if (inclBCs == kInterpBCsFromSuper){
            //
            // zone range on super zone in active area coordinates, expanded to include 
            // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
            // cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
            // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
            // range by 2 nsbcs.
            // 
            // coordinate in q and vars = i + i_0, j + j_0
            //
            ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;                      // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
            
            ja  = rng.n0[kYDim]-nsbc;          // ja is coord of start of copy are in active coordinates
            j_0  = ja+jss;      // i_0 is now absolute coord in super of start of area 
            njr = rng.rn[kYDim]+nsbc+nsbc;	   // njr is the range of the j loop in super cells
            
            izs = 0;
            jzs = 0;
            
          } else {
            ia  = rng.n0[kXDim];       // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;              // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim];       // nir is the range of the i loop in super cells
            
            ja  = rng.n0[kYDim];       // ja is coord of start of copy are in active coordinates
            j_0  = ja+jss;   // i_0 is now absolute coord in super of start of area 
            njr = rng.rn[kYDim];	   // njr is the range of the j loop in super cells
          }
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              if (zSuper->zoneData[id]!=NULL){
                
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
                zls_SetQuantity2DArr(&qSuper,zSuper->zoneData[id], nstj, nsti);
#endif ZONEDATA_ARE_PLAINBLOCKS
                
#ifdef ZONEDATA_ARE_MULTIDARRAYS
                q = z->zoneData[id];
                qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
                
                if (q != NULL){
                  if (qSuper != NULL){
                    
                    switch (interp){
                        
                      case  kInterpNN: // simple rebin
                        
                        for(j = 0; j < njr; j++){
                          jzd = jzs+(j*nR);
                          jsd = (j+j_0);
                          for(i = 0; i < nir; i++){
                            izd = izs+(i*nR);
                            isd = (i+i_0);
                            for( l = 0; l<nR; l++){
                              for( m = 0; m<nR; m++){
                                q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                              } //m
                            } //l
                          } // i
                        } //j
                        
                        break; // interp == kInterpNN
                        
                        case kInterpPLM: // Piecewise Linear Method interpolation 
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        
                        if (err == noErr) {
                          
                          for(j = 0; j<njr; j++){// rows, set and mult by x coeffs
                            jzd = jzs+(j*nR);
                            jsd = (j+ja);
                            zls_SetVar1DArrQ2DArrX       ( v1dx, qSuper, jsd);
                            err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                            err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                            for(i = 0; i<nir; i++){ // step in x in Super
                              izd = izs+(i*nR); 
                              isd = (i+i_0);
                              a = v1dx.v[isd].a;
                              for( l = 0; l<nR; l++){
                                for( m = 0; m<nR; m++){
                                  f = fx[isd].c[m];
                                  q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                                } //m
                              } //l
                            } //i
                          } //j
                          
                          for(i = 0; i<nir; i++){// rows, set and mult by x coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            err = zls_SetVar1DArrQ2DArrY    ( v1dy, qSuper, isd);
                            err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                            err = zlsppm_CalcLinInterpVar1D ( v1dy, fy, nR);
                            for(j = 0; j< njr; j++){ // step in y in Super
                              jzd = jzs+(j*nR);
                              jsd = (j+j_0);
                              for( l = 0; l<nR; l++){
                                f = fy[jsd].c[l];
                                for( m = 0; m<nR; m++){
                                  q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                                } //m
                              } //l
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill2DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break;
                        
                        case kInterpPPM: // PPM interpolation same as ppmlr
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        
                        if (err == noErr) {
                          
                          for(j = 0; j<njr; j++){// rows, set and mult by x coeffs
                            jzd = jzs+(j*nR);
                            jsd = (j+ja);
                            zls_SetVar1DArrQ2DArrX       ( v1dx, qSuper, jsd);
                            err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                            err = zlsppm_CalcInterpVar1D    ( v1dx, fx, nR);
                            for(i = 0; i<nir; i++){ // step in x in Super
                              izd = izs+(i*nR); 
                              isd = (i+i_0);
                              a = v1dx.v[isd].a;
                              for( l = 0; l<nR; l++){
                                for( m = 0; m<nR; m++){
                                  f = fx[isd].c[m];
                                  q[(jzd+l)][(izd+m)] = a  + f; // step in x in z
                                } //m
                              } //l
                            } //i
                          } //j
                          
                          for(i = 0; i<nir; i++){// rows, set and mult by x coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            zls_SetVar1DArrQ2DArrY       ( v1dy, qSuper, isd);
                            err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                            err = zlsppm_CalcInterpVar1D    ( v1dy, fy, nR);
                            for(j = 0; j< njr; j++){ // step in y in Super
                              jzd = jzs+(j*nR);
                              jsd = (j+j_0);
                              for( l = 0; l<nR; l++){
                                f = fy[jsd].c[l];
                                for( m = 0; m<nR; m++){
                                  q[(jzd+l)][(izd+m)] += f; // step in y in z
                                } //m
                              } //l
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill2DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break; // kInterpPPM
                        
                        default:
                        
                        err = nyiErr;
                        zlsio_Printf( " Fill2DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                        
                    }
                    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                    zls_ReleaseQuantity2DArr(qSuper);
#endif ZONEDATA_ARE_PLAINBLOCKS
                    
                  } // qSuper		
                  
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                  zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
                  
                } // q
                
              } // zSuperdata
            } // zdata
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // zSuper
      
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
#endif SYM2D
  
}

int  zls_FillZoneFromSuperZone3D( ZonePtr z,  int inclBCs );
int  zls_FillZoneFromSuperZone3D( ZonePtr z,  int inclBCs ){
#ifndef SYM3D
  return( dimErr );
#else
	Integer err = noErr;
	
	Integer    i,    j,    k,   id;	 //  loop counters
	Integer    l,    m,    n ;          //  sub cell loop counters
	
	Integer         nR;
	Integer         nG;
	Integer       nsbc;   
	Integer     interp;
	Range          rng;
	
	Real  f, a;
	
	Integer  izd,  jzd,   kzd;     // i and j and k in zone data
	Integer  isd,  jsd,   ksd;     // i and j and k in super data
	
	InPtr        fx, fy, fz;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
	V1DArr        v1dy;		// x , y , z rows of the  super zone
	V1DArr        v1dz;		
	
	Integer   ia,  i_0,  nir; // super range in active area coord
	Integer   ja,  j_0,  njr; // super range in active area coord
	Integer   ka,  k_0,  nkr; // super range in active area coord
	Integer  izs,  jzs,  kzs;     // zone start offsets
	Integer  iss,  jss,  kss;     // super start offsets
	
	Integer   nsi, nsj, nsk;     // super data full range for making vars
	Integer  nti,  ntj, ntk;     // zonedata full range
	Integer nsti, nstj, nstk;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q3DArr      qSuper;
	Q3DArr           q;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nG     = z->zoneBCells;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          nti    = z->zoneGrid.nt[kXDim];
          ntj    = z->zoneGrid.nt[kYDim];
          ntk    = z->zoneGrid.nt[kZDim];
          
          izs    = z->zoneGrid.ns[kXDim];
          jzs    = z->zoneGrid.ns[kYDim];
          kzs    = z->zoneGrid.ns[kZDim];
          
          // get Super zone area parameters
          
          iss    = zSuper->zoneGrid.ns[kXDim];
          jss    = zSuper->zoneGrid.ns[kYDim];
          kss    = zSuper->zoneGrid.ns[kZDim];
          
          nsi   = zSuper->zoneGrid.n[kXDim]; 
          nsj   = zSuper->zoneGrid.n[kYDim];
          nsk   = zSuper->zoneGrid.n[kZDim];
          
          
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          nstj   = zSuper->zoneGrid.nt[kYDim]; // 
          nstk   = zSuper->zoneGrid.nt[kZDim]; // 
          
          if (inclBCs == kInterpBCsFromSuper){
            //
            // zone range on super zone in active area coordinates, expanded to include 
            // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
            // cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
            // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
            // range by 2 nsbcs.
            // 
            // coordinate in q and vars = i + i_0, j + j_0, k + k_0
            //
            ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;                     // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
            
            ja  = rng.n0[kYDim]-nsbc;          // ja is coord of start of copy are in active coordinates
            j_0 = ja+jss;                      // i_0 is now absolute coord in super of start of area 
            njr = rng.rn[kYDim]+nsbc+nsbc;	   // njr is the range of the j loop in super cells
            
            ka  = rng.n0[kZDim]-nsbc;          // ja is coord of start of copy are in active coordinates
            k_0 = ka+kss;                      // k_0 is now absolute coord in super of start of area 
            nkr = rng.rn[kZDim]+nsbc+nsbc;	   // nkr is the range of the k loop in super cells
            
            izs = 0;
            jzs = 0;
            kzs = 0;
            
          } else {
            ia  = rng.n0[kXDim];       // ia is now  coord in super of start of area in active areas
            i_0  = ia+iss;             // i_0 is now absolute coord in super of start of area
            nir = rng.rn[kXDim];       // nir is the range of the i loop in super cells
            
            ja  = rng.n0[kYDim];       // ja is coord of start of copy are in active coordinates
            j_0  = ja+jss;             // j_0 is now absolute coord in super of start of area 
            njr = rng.rn[kYDim];	   // njr is the range of the j loop in super cells
            
            ka  = rng.n0[kZDim];       // ka is coord of start of copy are in active coordinates
            k_0 = ka+kss;              // k_0 is now absolute coord in super of start of area 
            nkr = rng.rn[kZDim];	   // nkr is the range of the k loop in super cells
          }
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              if (zSuper->zoneData[id]!=NULL){
                
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
                zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[id], nstk, nstj, nsti);
#endif ZONEDATA_ARE_PLAINBLOCKS
                
#ifdef ZONEDATA_ARE_MULTIDARRAYS
                q = z->zoneData[id];
                qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
                
                if (q != NULL){
                  if (qSuper != NULL){
                    
                    switch (interp){
                        
                      case  kInterpNN: // simple rebin
                        
                        for(k = 0; k < nkr; k++){
                          kzd = kzs+(k*nR);
                          ksd = (k+k_0);
                          for(j = 0; j < njr; j++){
                            jzd = jzs+(j*nR);
                            jsd = (j+j_0);
                            for(i = 0; i < nir; i++){
                              izd = izs+(i*nR);
                              isd = (i+i_0);
                              for( l = 0; l<nR; l++){
                                for( m = 0; m<nR; m++){
                                  for( n = 0; n<nR; n++){
                                    q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
                                  } //n
                                } //m
                              } //l
                            } // i
                          } //j
                        } //k
                        
                        break; // interp == kInterpNN
                        
                        case kInterpPLM: // Piecewise Linear Method interpolation 
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewVar1DArr (&v1dz, nG, nsk);
                        v1dz.x = zSuper->zoneGrid.z;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        err += zls_NewInterp1D (&fz, nstk);
                        
                        if (err == noErr) {
                          
                          for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(j = 0; j<njr; j++){// rows, set and mult by x coeffs
                              jzd = jzs+(j*nR);
                              jsd = (j+ja);
                              zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                              for(i = 0; i<nir; i++){ // step in x in Super
                                izd = izs+(i*nR); 
                                isd = (i+i_0);
                                a = v1dx.v[isd].a;
                                for( l = 0; l<nR; l++){
                                  for( m = 0; m<nR; m++){
                                    for( n = 0; n<nR; n++){
                                      f = fx[isd].c[n];
                                      q[(kzd+l)][(jzd+m)][(izd+n)] =  a  + f; // step in x in z
                                    } //n
                                  } //m
                                } //l
                              } //i
                            } //j
                          } //k
                          
                          for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(i = 0; i<nir; i++){// rows, set and mult by y coeffs
                              izd = izs+(i*nR); 
                              isd = (i+ia);
                              err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dy, fy, nR);
                              for(j = 0; j< njr; j++){ // step in y in Super
                                jzd = jzs+(j*nR);
                                jsd = (j+j_0);
                                for( l = 0; l<nR; l++){
                                  for( m = 0; m<nR; m++){
                                    f = fy[jsd].c[m];
                                    for( n = 0; n<nR; n++){
                                      q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                    } //n
                                  } //m
                                } //l
                              } //j
                            } //i
                          } // k
                          
                          for(i = 0; i<nir; i++){// rows, set and mult by z coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            for(j = 0; j<njr; j++){// rows, set and mult by z coeffs
                              jzd = jzs+(j*nR); 
                              jsd = (j+ja);
                              err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dz, fz, nR);
                              for(k = 0; k< nkr; k++){ // step in z in Super
                                kzd = kzs+(k*nR);
                                ksd = (k+k_0);
                                for( l = 0; l<nR; l++){
                                  f = fz[ksd].c[l];
                                  for( m = 0; m<nR; m++){
                                    for( n = 0; n<nR; n++){
                                      q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                    } //n
                                  } //m
                                } //l
                              } //k
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dz);
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          zls_DisposeInterp(fy); fz = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break;
                        
                        case kInterpPPM: // PPM interpolation same as ppmlr
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewVar1DArr (&v1dz, nG, nsk);
                        v1dz.x = zSuper->zoneGrid.z;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        err += zls_NewInterp1D (&fz, nstk);
                        
                        if (err == noErr) {
                          
                          for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(j = 0; j<njr; j++){// rows, set and mult by x coeffs
                              jzd = jzs+(j*nR);
                              jsd = (j+ja);
                              zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dx, fx, nR);
                              for(i = 0; i<nir; i++){ // step in x in Super
                                izd = izs+(i*nR); 
                                isd = (i+i_0);
                                a = v1dx.v[isd].a;
                                for( l = 0; l<nR; l++){
                                  for( m = 0; m<nR; m++){
                                    for( n = 0; n<nR; n++){
                                      f = fx[isd].c[n];
                                      q[(kzd+l)][(jzd+m)][(izd+n)] =  a  +  f; // step in x in z
                                    } //n
                                  } //m
                                } //l
                              } //i
                            } //j
                          } //k
                          
                          for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(i = 0; i<nir; i++){// rows, set and mult by y coeffs
                              izd = izs+(i*nR); 
                              isd = (i+ia);
                              err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dy, fy, nR);
                              for(j = 0; j< njr; j++){ // step in y in Super
                                jzd = jzs+(j*nR);
                                jsd = (j+j_0);
                                for( l = 0; l<nR; l++){
                                  for( m = 0; m<nR; m++){
                                    f = fy[jsd].c[m];
                                    for( n = 0; n<nR; n++){
                                      q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                    } //n
                                  } //m
                                } //l
                              } //j
                            } //i
                          } // k
                          
                          for(i = 0; i<nir; i++){// rows, set and mult by z coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            for(j = 0; j<njr; j++){// rows, set and mult by z coeffs
                              jzd = jzs+(j*nR); 
                              jsd = (j+ja);
                              err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dz, fz, nR);
                              for(k = 0; k< nkr; k++){ // step in z in Super
                                kzd = kzs+(k*nR);
                                ksd = (k+k_0);
                                for( l = 0; l<nR; l++){
                                  f = fz[ksd].c[l];
                                  for( m = 0; m<nR; m++){
                                    for( n = 0; n<nR; n++){
                                      q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                    } //n
                                  } //m
                                } //l
                              } //k
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dz);
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          zls_DisposeInterp(fy); fz = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break; // kInterpPPM
                        
                        default:
                        
                        err = nyiErr;
                        zlsio_Printf( " Fill2DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                        
                    }
                    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
                    zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
                    zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
                    
                  } // qSuper		
                  
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
                  zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
                  zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
                  
                } // q
                
              } // zSuperdata
            } // zdata
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // zSuper
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
  
#endif SYM3D
  
}

int  zlsbc_FillZoneFromSuperZone( ZonePtr z,  int inclBCs ){
  int err = noErr;
#ifdef SYM1D
  err = zls_FillZoneFromSuperZone1D( z, inclBCs ); 
#endif SYM1D
#ifdef SYM2D
  err = zls_FillZoneFromSuperZone2D( z, inclBCs ); 
#endif SYM2D
#ifdef SYM3D
  err = zls_FillZoneFromSuperZone3D( z, inclBCs ); 
#endif SYM3D
  return(err);
}

//

int  zlsbc_CalcZoneBCs ( ZonePtr z, int direction, int bcKind ){
	
	Integer err = noErr;
	
	Real ratio = 1.0;
	Real a     = 1.0;
	
	Integer    id;  
	QArr        q;
#ifdef SYM1D
	Integer    i;	//  loop counters
#endif SYM1D
#ifdef SYM2D
	Integer    i,    j;	//  loop counters
#endif SYM2D
#ifdef SYM3D
	Integer    i,    j,    k;	//  loop counters
#endif SYM3D
	
	if (z != NULL){
		
		// get  zone area parameters    
		Integer nbc   = z->zoneBCells;
#ifdef SYM1D
		
		Integer is    = z->zoneGrid.ns[kXDim];
		Integer ie    = z->zoneGrid.ne[kXDim];
		
#endif SYM1D
		
#ifdef SYM2D
		Integer is    = z->zoneGrid.ns[kXDim];
		Integer ie    = z->zoneGrid.ne[kXDim];
		Integer nti   = z->zoneGrid.nt[kXDim];
		
		Integer je    = z->zoneGrid.ne[kYDim];
		Integer js    = z->zoneGrid.ns[kYDim];
		Integer ntj   = z->zoneGrid.nt[kYDim];
#endif SYM2D
		
#ifdef SYM3D
		
		Integer is    = z->zoneGrid.ns[kXDim];
		Integer ie    = z->zoneGrid.ne[kXDim];
		Integer nti   = z->zoneGrid.nt[kXDim];
		
		Integer je    = z->zoneGrid.ne[kYDim];
		Integer js    = z->zoneGrid.ns[kYDim];
		Integer ntj   = z->zoneGrid.nt[kYDim];
		
		Integer ke    = z->zoneGrid.ne[kZDim];
		Integer ks    = z->zoneGrid.ns[kZDim];
		Integer ntk   = z->zoneGrid.nt[kZDim];
#endif SYM3D
		
		for (id = 0; id <  z->nQ; id++){
			if (z->zoneData[id]!=NULL){
				
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
				q = z->zoneData[id];
#endif SYM1D
#ifdef SYM2D
				zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif SYM2D
#ifdef SYM3D
				zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
				
				
#ifdef ZONEDATA_ARE_MULTIDARRAYS
				q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
				
				if (q!=NULL){
					
					switch (bcKind) {
              
						case  kBCCKindFree:
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < nbc; i++){
										q[i] = q[is];
									}; //i
									break;
                  case kDirRight:
									for(i = 0; i < nbc; i++){
										q[i+ie] = q[ie-1];
									}; //i
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nbc; i++){
											q[j][i] = q[j][is];
										}; //i
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nbc; i++){
											q[j][i+ie] = q[j][ie-1];
										}; //i
									} //j
									break;
                  case kDirDown:
									for(i = 0; i < nti; i++){
										for(j = 0; j < nbc; j++){
											q[j][i] = q[js][i];
										} //j
									} //i
									break;
                  case kDirUp:
									for(i = 0; i < nti; i++){
										for(j = 0; j < nbc; j++){
											q[j+je][i] = q[je-1][i];
										}; //j
									} // i
									break;
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][i] = q[k][j][is];
											} //i
										} //j
									} //k
									break;
                  case kDirRight: 
									for( k = 0; k < ntk; k++ ){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][i+ie] = q[k][j][ie-1];
											} //i
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ //
										for(i = 0; i < nti; i++){
											for(j = 0; j < nbc; j++){
												q[k][j][i] = q[k][js][i];
											} //j
										} //i
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){
										for(i = 0; i < nti; i++){
											for(j = 0; j < nbc; j++){ 
												q[k][j+je][i] = q[k][je-1][i];
											} //j
                    } //i
									} //k
									break;
                  case kDirYon:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < nbc; k++){ 
												q[k][j][i] = q[ks][j][i];
											} //k
										} //j 
									} //i
									break;
                  case kDirHither:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < nbc; k++){ 
												q[k+ke][j][i] = q[ke-1][j][i];
											} //k
										} //j
									} //i
							} // direction switch 
#endif SYM3D
							break;
              
              
						case  kBCCKindFreer:
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < (nbc-1); i++){
										q[i] = q[is-1];
									}; //i
									break;
                  case kDirRight:
									for(i = 0; i < (nbc-1); i++){
										q[i+ie+1] = q[ie];
									}; //i
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < (nbc-1); i++){
											q[j][i] = q[j][is-1];
										}; //i
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < (nbc-1); i++){
											q[j][i+ie+1] = q[j][ie];
										}; //i
									} //j
									break;
                  case kDirDown:
									for(i = 0; i < nti; i++){
										for(j = 0; j < (nbc-1); j++){
											q[j][i] = q[js-1][i];
										} //j
									} //i
									break;
                  case kDirUp:
									for(i = 0; i < nti; i++){
										for(j = 0; j < (nbc-1); j++){
											q[j+je+1][i] = q[je][i];
										}; //j
									} // i
									break;
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < (nbc-1); i++){
												q[k][j][i] = q[k][j][is-1];
											} //i
										} //j
									} //k
									break;
                  case kDirRight: 
									for( k = 0; k < ntk; k++ ){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < (nbc-1); i++){
												q[k][j][i+ie+1] = q[k][j][ie];
											} //i
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ //
										for(i = 0; i < nti; i++){
											for(j = 0; j < (nbc-1); j++){
												q[k][j][i] = q[k][js-1][i];
											} //j
										} //i
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){
										for(i = 0; i < nti; i++){
											for(j = 0; j < (nbc-1); j++){ 
												q[k][j+je+1][i] = q[k][je][i];
											} //j
                    } //i
									} //k
									break;
                  case kDirYon:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < (nbc-1); k++){ 
												q[k][j][i] = q[ks-1][j][i];
											} //k
										} //j 
									} //i
									break;
                  case kDirHither:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < (nbc-1); k++){ 
												q[k+ke+1][j][i] = q[ke][j][i];
											} //k
										} //j
									} //i
							} // direction switch 
#endif SYM3D
							break;
              
						case  kBCCKindFreeOutOnly:
							
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < nbc; i++){
										q[i] = q[is];
									}; //i
									if (id == kWVx){ // enforce 0 or outward only velocities
										for(i = 0; i < nbc; i++){
											a   = q[i];
											q[i] = (a > 0.0)? -a: a;
										} //i
                  };
									break;
                  case kDirRight:
									for(i = 0; i < nbc; i++){
										q[i+ie] = q[ie-1];
									}; //i
									if (id == kWVx){ // enforce 0 or outward only velocities
										for(i = 0; i < nbc; i++){
											a   = q[i+ie];
											q[i+ie] = (a < 0.0)? -a: a;
										} //i
									};
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nbc; i++){
											q[j][i] = q[j][is];
										}; //i
										if (id == kWVx){ // enforce 0 or outward only velocities
											for(i = 0; i < nbc; i++){
												a = q[j][i] ;
												q[j][i]  = (a > 0.0)? -a: a;
											} //i
										}
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nbc; i++){
											q[j][i+ie] = q[j][ie-1];
										}; //i
										if (id == kWVx){ // enforce 0 or outward only velocities
											for(i = 0; i < nbc; i++){
												a  = q[j][i+ie];
												q[j][i+ie] = (a < 0.0)? -a: a;
											} //i
										}
									} //j
									break;
                  case kDirDown:
									for(i = 0; i < nti; i++){
										for(j = 0; j < nbc; j++){
											q[j][i] = q[js][i];
										} //j
										if (id == kWVy){ // enforce 0 or outward only velocities
											for(j = 0; j < nbc; j++){
												a = q[j][i];
												q[j][i] = (a > 0.0)? -a: a;
											} //j
										}
									} //i
									break;
                  case kDirUp:
									for(i = 0; i < nti; i++){
										for(j = 0; j < nbc; j++){
											q[j+je][i] = q[je-1][i];
										}; //j
										if (id == kWVy){ // enforce 0 or outward only velocities
											for(j = 0; j < nbc; j++){
												a = q[j+je][i];
												q[j+je][i] = (a < 0.0)? -a: a;
											} //j
										}
									} // i
									break;
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][i] = q[k][j][is];
											} //i
											if (id == kWVx){ // enforce 0 or outward only velocities
												for(i = 0; i < nbc; i++){
													a = q[k][j][i] ;
													q[k][j][i]  = (a > 0.0)? -a: a;
												} //i
											}
										} //j
									} //k
									break;
                  case kDirRight: 
									for( k = 0; k < ntk; k++ ){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][i+ie] = q[k][j][ie-1];
											} //i
											if (id == kWVx){ // enforce 0 or outward only velocities
												for(i = 0; i < nbc; i++){
													a  = q[k][j][i+ie];
													q[k][j][i+ie] = (a < 0.0)? -a: a;
												} //i
											}
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ //
										for(i = 0; i < nti; i++){
											for(j = 0; j < nbc; j++){
												q[k][j][i] = q[k][js][i];
											} //j
											if (id == kWVy){ // enforce 0 or outward only velocities
												for(j = 0; j < nbc; j++){
													a = q[k][j][i];
													q[k][j][i] = (a > 0.0)? -a: a;
												} //j
											}
										} //i
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){
										for(i = 0; i < nti; i++){
											for(j = 0; j < nbc; j++){ 
												q[k][j+je][i] = q[k][je-1][i];
											} //j
											if (id == kWVy){ // enforce 0 or outward only velocities
												for(j = 0; j < nbc; j++){
													a = q[k][j+je][i];
													q[k][j+je][i] = (a < 0.0)? -a: a;
												} //j
											}
                    } //i
									} //k
									break;
                  case kDirYon:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < nbc; k++){ 
												q[k][j][i] = q[ks][j][i];
											} //k
											if (id == kWVz){ // enforce 0 or outward only velocities
												for(k = 0; k < nbc; k++){
													a = q[k][j][i];
													q[k][j][i] = (a > 0.0)? -a: a;
												} //k
											}
										} //j 
									} //i
									break;
                  case kDirHither:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											for(k = 0; k < nbc; k++){ 
												q[k+ke][j][i] = q[ke-1][j][i];
											} //k
											if (id == kWVz){ // enforce 0 or outward only velocities
												for(k = 0; k < nbc; k++){
													a = q[k+ke][j][i];
													q[k+ke][j][i] = (a < 0.0)? -a: a;
												} //k
											}
										} //j
									} //i
							} // direction switch 
#endif SYM3D
							break;
							
						case  kBCCKindExtrapol:
							
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									if ((id == kWDens)||(id == kWPres)){
										ratio = q[is+1];
										ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[is]/ratio:1.0;
										for(i = 0; i < nbc; i++){
											q[i] = ratio*q[is];
										} //i
									} else  {
										for(i = 0; i < nbc; i++){
                      ratio = q[is+1]-q[is];
											q[i]  = q[is]-ratio;
										} //i
									}
									break;
                  case kDirRight:
									if ((id == kWDens)||(id == kWPres)){
										ratio  = q[ie-2];
										ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[ie-1]/ratio:1.0;
										for(i = 0; i < nbc; i++){
											q[i+ie] = ratio*q[ie-1];
										} //i
									} else  {
										for(i = 0; i < nbc; i++){
                      ratio   = q[ie-2]-q[ie-1];
											q[i+ie] = q[ie-1]-ratio;
										} //i
									} 
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ 
										
										if ((id == kWDens)||(id == kWPres)){
											ratio = q[j][is+1];
											ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[j][is]/ratio:1.0;
											for(i = 0; i < nbc; i++){
												q[j][i] = ratio*q[j][is];
											} //i
										} else  {
											for(i = 0; i < nbc; i++){
	                      ratio   = q[j][is+1]-q[j][is];
                        q[j][i] = q[j][is]-ratio;
											} //i
										} 
										
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ 
										
										if ((id == kWDens)||(id == kWPres)){
											ratio  = q[j][ie-2];
											ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[j][ie-1]/ratio:1.0;
											for(i = 0; i < nbc; i++){
												q[j][i+ie] = ratio*q[j][ie-1];
											} //i
										} else  {
											for(i = 0; i < nbc; i++){
	                      ratio      = q[j][ie-2]-q[j][ie-1];
                        q[j][i+ie] = q[j][ie-1]-ratio;
											} //i
										} 
										
									} //j
									break;
                  case kDirDown:
									for(i = 0; i < nti; i++){
										
										if ((id == kWDens)||(id == kWPres)){
											ratio = q[js+1][i];
											ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[js][i]/ratio:1.0;
											for(j = 0; j < nbc; j++){ 
												q[j][i] = ratio*q[js][i];
											} //j
										} else  {
											for(j = 0; j < nbc; j++){ 
                        ratio   = q[js+1][is]-q[js][i];
                        q[j][i] = q[js][i]-ratio;
											} //j
										} 
										
									} //i
									break;
                  case kDirUp:
									for(i = 0; i < nti; i++){
										
										if ((id == kWDens)||(id == kWPres)){
											ratio  = q[je-2][i];
											ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[je-1][i]/ratio:1.0;
											for(j = 0; j < nbc; j++){ 
												q[j+je][i] = q[je-1][i];
											} //j
										} else  {
											for(j = 0; j < nbc; j++){ 
                        ratio      = q[je-2][i]-q[je-1][i];
                        q[j+je][i] = q[je-1][i]-ratio;
											} //j
										} 
										
									} //i
									break;
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ // left col, copy left edge cell
										for(j = 0; j < ntj; j++){ // left col, copy left edge cell
											
											if ((id == kWDens)||(id == kWPres)){
												ratio = q[k][j][is+1];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[k][j][is]/ratio:1.0;
												for(i = 0; i < nbc; i++){
													q[k][j][i] = ratio*q[k][j][is];
												} //i
											} else  {
												for(i = 0; i < nbc; i++){
                          ratio      = q[k][j][is+1]-q[k][j][is];
                          q[k][j][i] = q[k][j][is]-ratio;
												} //i
											} 
											
										} //j
									} //k
									break;
                  case kDirRight:
									for( k = 0; k < ntk; k++ ){ // right col, copy right edge cell
										for(j = 0; j < ntj; j++){ // right col, copy right edge cell
											
											if ((id == kWDens)||(id == kWPres)){
												ratio  = q[k][j][ie-2];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[k][j][ie-1]/ratio:1.0;
												for(i = 0; i < nbc; i++){
													q[k][j][i+ie] = ratio*q[k][j][ie-1];
												} //i
											} else  {
												for(i = 0; i < nbc; i++){
                          ratio         = q[k][j][ie-2]-q[k][j][ie-1];
                          q[k][j][i+ie] = q[k][j][ie-1]-ratio;
												} //i
											} 
											
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ //
										for(i = 0; i < nti; i++){
											
											if ((id == kWDens)||(id == kWPres)){
												ratio  = q[k][js+1][i];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[k][js][i]/ratio:1.0;
												for(j = 0; j < nbc; j++){ 
													q[k][j][i] = ratio*q[k][js][i];
												} //j
											} else  {
												for(j = 0; j < nbc; j++){ 
                          ratio      = q[k][js+1][i]-q[k][js][i];
                          q[k][j][i] = q[k][js][i]-ratio;
												} //j
											} 
											
										} //i
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){
										for(i = 0; i < nti; i++){
											
											if ((id == kWDens)||(id == kWPres)){
												ratio  = q[k][je-2][i];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)? q[k][je-1][i]/ratio:1.0;
												for(j = 0; j < nbc; j++){ 
													q[k][j+je][i] = ratio*q[k][je-1][i];
												} //j
											} else  {
												for(j = 0; j < nbc; j++){ 
                          ratio         = q[k][je-2][i]-q[k][je-1][i];
                          q[k][j+je][i] = q[k][je-1][i]-ratio;
												} //j
											} 
											
										} //j
									} //k
									break;
                  case kDirYon:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											
											if ((id == kWDens)||(id == kWPres)){
												ratio = q[ks+1][j][i];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[ks][j][i]/ratio:1.0;
												for(k = 0; k < nbc; k++){ 
													q[k][j][i] = ratio*q[ks][j][i];
												} //k
											} else  {
												for(k = 0; k < nbc; k++){ 
                          ratio      = q[ks+1][j][i]-q[ks][j][i];
                          q[k][j][i] = q[ks][j][i]-ratio;
												} //k
											} 
											
										} //j
									} //i
									break;
                  case kDirHither:
									for(j = 0; j < ntj; j++){ 
										for(i = 0; i < nti; i++){
											
											if ((id == kWDens)||(id == kWPres)){
												ratio = q[ke-2][j][i];
												ratio = (zls_tol(ratio, kZLS_Small) != 0.0)?q[ke-1][j][i]/ratio:1.0;
												for(k = 0; k < nbc; k++){ 
													q[k+ke][j][i] = ratio*q[ke-1][j][i];
												} //k
											} else  {
												for(k = 0; k < nbc; k++){ 
                          ratio         = q[ke-2][j][i]-q[ke-1][j][i];
                          q[k+ke][j][i] = q[ke-1][j][i]-ratio;
												} //k
											} 
											
										} //j
									} //i
							} // direction switch
#endif SYM3D
							break; 
							
							
						case kBCCKindConstant: 
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < nbc; i++){
										q[i] = z->zoneBCValues[direction][id];
									} //i
									break;
                  case kDirRight:
									for(i = 0; i < nbc; i++){
										q[ie+i] = z->zoneBCValues[direction][id];
									} //i
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ // left col
										for(i = 0; i < nbc; i++){
											q[j][i] = z->zoneBCValues[direction][id];
										} //i
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ // rgt col
										for(i = 0; i < nbc; i++){
											q[j][ie+i] = z->zoneBCValues[direction][id];
										} //i
									} //j
									break;
                  case kDirDown:
									for(j = 0; j < nbc; j++){ // bottom row
										for(i = 0; i < nti; i++){
											q[j][i] = z->zoneBCValues[direction][id];
										} //i
									} //j
									break;
                  case kDirUp:
									for(j = 0; j < nbc; j++){ // top row
										for(i = 0; i < nti; i++){
											q[je+j][i]  = z->zoneBCValues[direction][id];
										} //i
									} //j
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][i] = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
									break;
                  case kDirRight:
									for(k = 0; k < ntk; k++){
										for(j = 0; j < ntj; j++){ // rgt col
											for(i = 0; i < nbc; i++){
												q[k][j][ie+i] = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){
										for(j = 0; j < nbc; j++){ // bottom row
											for(i = 0; i < nti; i++){
												q[k][j][i] = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){
										for(j = 0; j < nbc; j++){ // top row
											for(i = 0; i < nti; i++){
												q[k][je+j][i]  = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
									break;
                  case kDirYon:
									for(k = 0; k < nbc; k++){
										for(j = 0; j < ntj; j++){ // bottom row
											for(i = 0; i < nti; i++){
												q[k][j][i] = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
									break;
                  case kDirHither:
									for(k = 0; k < nbc; k++){
										for(j = 0; j < ntj; j++){ // top row
											for(i = 0; i < nti; i++){
												q[ke+k][j][i]  = z->zoneBCValues[direction][id];
											} //i
										} //j
									} //k
							} // direction switch
#endif SYM3D
							break;
							
						case kBCCKindReflected:
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < nbc; i++){
										if (id == kWVx) {
											q[is-i-1] = -q[is+i];
										} else {
											q[is-i-1] =  q[is+i];
										}
									} //i
									break;
                  case kDirRight:
									for(i = 0; i < nbc; i++){
										if (id == kWVx) {
											q[ie+i] = -q[ie-i-1];
										} else {
											q[ie+i] =  q[ie-i-1];
										}
									} //i
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirDown:
									for(j = 0; j < nbc; j++){ // bottom row, reverse y
										for(i = 0; i < nti; i++){
											if (id == kWVy) {
												q[js-j-1][i] = -q[js+j][i];
											} else {
												q[js-j-1][i] =  q[js+j][i];
											}
										} //i
									} //j
									break;
                  case kDirUp:
									for(j = 0; j < nbc; j++){ // top row, reverse y
										for(i = 0; i < nti; i++){
											if (id == kWVy) {
												q[je+j][i] = -q[je-j-1][i];
											} else {
												q[je+j][i] =  q[je-j-1][i];
											}
										} //i
									} //j
									break;
                  case kDirLeft:
									for(j = 0; j < ntj; j++){ // left col, reverse x
										for(i = 0; i < nbc; i++){
											if (id == kWVx) {
												q[j][is-i-1] = -q[j][is+i];
											} else {
												q[j][is-i-1] =  q[j][is+i];
											}
										} //i
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ // rgt col, reverse x
										for(i = 0; i < nbc; i++){
											if (id == kWVx) {
												q[j][ie+i] = -q[j][ie-i-1];
											} else {
												q[j][ie+i] =  q[j][ie-i-1];
											}
										} //i
									} //j
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ // left col, reverse x
											for(i = 0; i < nbc; i++){
												if (id == kWVx) {
                          q[k][j][is-i-1] = -q[k][j][is+i];
												} else {
													q[k][j][is-i-1] =  q[k][j][is+i];
												}
											} //i
										} //j
									} //k
									break;
                  case kDirRight:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ // rgt col, reverse x
											for(i = 0; i < nbc; i++){
												if (id == kWVx) {
													q[k][j][ie+i] = -q[k][j][ie-i-1];
												} else {
													q[k][j][ie+i] =  q[k][j][ie-i-1];
												}
											} //i
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < nbc; j++){ // bottom row, reverse y
											for(i = 0; i < nti; i++){
												if (id == kWVy) {
													q[k][js-j-1][i] = -q[k][js+j][i];
												} else {
													q[k][js-j-1][i] =  q[k][js+j][i];
												}
											} //i
										} //j
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < nbc; j++){ // top row, reverse y
											for(i = 0; i < nti; i++){
												if (id == kWVy) {
													q[k][je+j][i] = -q[k][je-j-1][i];
												} else {
													q[k][je+j][i] =  q[k][je-j-1][i];
												}
											} //i
										} //j
									} //k
									break;
                  case kDirYon:
									for(k = 0; k < nbc; k++){ 
										for(j = 0; j < ntj; j++){
											for(i = 0; i < nti; i++){
												if (id == kWVz) {
													q[ks-k-1][j][i] = -q[ks+k][j][i];
												} else {
													q[ks-k-1][j][i] =  q[ks+k][j][i];
												}
											} //i
										} //j
									} //k
									break;
                  case kDirHither:
									for(k = 0; k < nbc; k++){ 
										for(j = 0; j < ntj; j++){
											for(i = 0; i < nti; i++){
												if (id == kWVz) {
													q[ke+k][j][i] = -q[ke-k-1][j][i];
												} else {
													q[ke+k][j][i] =  q[ke-k-1][j][i];
												}
											} //i
										} //j
									} //k
							} // direction switch
#endif SYM3D
							break;
							
						case kBCCKindPeriodic:
#ifdef SYM1D
							switch (direction)  {
								case kDirLeft:
									for(i = 0; i < nbc; i++){
										q[is-i-1] = q[ie-i-1];
									} //i
									break;
                  case kDirRight:
									for(i = 0; i < nbc; i++){
										q[ie+i] = q[is+i];
									} //i
									break;
							} // direction switch
#endif SYM1D
#ifdef SYM2D
							switch (direction)  {
								case kDirLeft:
									for(j = 0; j < ntj; j++){ // left col, copy right edge cell
										for(i = 0; i < nbc; i++){
											q[j][is-i-1] = q[j][ie-i-1];
										} //i
									} //j
									break;
                  case kDirRight:
									for(j = 0; j < ntj; j++){ // rgt col, copy left edge cell
										for(i = 0; i < nbc; i++){
											q[j][ie+i] = q[j][is+i];
										} //i
									} //j
									break;
                  case kDirDown:
									for(j = 0; j < nbc; j++){ // bottom row, copy top edge cells
										for(i = 0; i < nti; i++){
											q[js-j-1][i] = q[je-j-1][i];
										} //i
									} //j
									break;
                  case kDirUp:
									for(j = 0; j < nbc; j++){ // top row, copy bot edge cell
										for(i = 0; i < nti; i++){
											q[je+j][i] = q[js+j][i];
										} //i
									} //j
									break;
							} // direction switch
#endif SYM2D
#ifdef SYM3D
							switch (direction)  {
								case kDirLeft:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][is-nbc+i] = q[k][j][ie-nbc+i];
											} //i
										} //j
									} //k
									break;
                  case kDirRight:
									for(k = 0; k < ntk; k++){
										for(j = 0; j < ntj; j++){ 
											for(i = 0; i < nbc; i++){
												q[k][j][ie+i] = q[k][j][is+i];
											} //i
										} //j
									} //k
									break;
                  case kDirDown:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < nbc; j++){ // bottom row, copy top edge cells
											for(i = 0; i < nti; i++){
												q[k][js-nbc+j][i] = q[k][je-nbc+j][i];
											} //i
										} //j
									} //k
									break;
                  case kDirUp:
									for(k = 0; k < ntk; k++){ 
										for(j = 0; j < nbc; j++){ // top row, copy top edge cell
											for(i = 0; i < nti; i++){
												q[k][je+j][i] = q[k][js+j][i];
											} //i
										} //j
									} //k
									break;
                  case kDirYon:
									for(k = 0; k < nbc; k++){ 
										for(j = 0; j < ntj; j++){ // bottom row, copy top edge cells
											for(i = 0; i < nti; i++){
												q[ks-nbc+k][j][i] = q[ke-nbc+k][j][i];
											} //i
										} //j
									} //k
									break;
                  case kDirHither:
									for(k = 0; k < nbc; k++){ 
										for(j = 0; j < ntj; j++){ // top row, copy top edge cell
											for(i = 0; i < nti; i++){
												q[ke+k][j][i] = q[ks+k][j][i];
											} //i
										} //j
									} //k
									break;
									
							} // direction switch
#endif SYM3D
							break;
						default:
							err = nyiErr;
							zlsio_ErrPrintf( "ERROR: Calc2DZoneBCs: Unknown Boundary Type Not Implemented Yet: %d \n", bcKind);
					}
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
					zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
					zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
				} else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop
	} else { err = memErr; } // z
	return (err);
}


int  zlsbc_ImposeZoneBCs ( ZonePtr z, int direction ){
	
	Integer err = noErr;
	
	int bndryType;  // what sort of boundary, super, link, calc
	int bndryKind;   // what sub stype of calculated bndry - const, periodic etc
	
	if (z != NULL){
		// get  zone area parameters
		
    // mask off type for the given direction
    bndryType  = z->zoneBCType[direction]&kBoundaryTypeFlagComputed; 
    bndryKind  = z->zoneBCType[direction]&kBCCKindMask; 
    
    if  (bndryType == kBoundaryTypeFlagComputed) {
#ifdef ZLS_USER_BOUNDARIES
      if (bndryKind == kBCCKindUser) {
        err = zlsp_CalcUserZoneBCs(z, direction, bndryKind);
        if (err != noErr){
          zlsio_ErrPrintf( " ERROR: zlsbc_ImposeZoneBCs from zlsp_CalcUserZoneBCs: Z %d  dir %d kind %d E %d \n",z->zoneRankID, direction,bndryKind, err);
          return(err);
        } 
      } else {
        err = zlsbc_CalcZoneBCs(z, direction, bndryKind); 
        if (err != noErr){
          zlsio_ErrPrintf( " ERROR: zlsbc_ImposeZoneBCs from zlsbc_CalcZoneBCs: Z %d  dir %d kind %d E %d \n",z->zoneRankID, direction,bndryKind, err);
          return(err);
        }
      }
#else
      err = zlsbc_CalcZoneBCs(z, direction, bndryKind); 
      if (err != noErr){
        zlsio_ErrPrintf( " ERROR: zlsbc_ImposeZoneBCs from zlsbc_CalcZoneBCs: Z %d  dir %d kind %d E %d \n",z->zoneRankID, direction,bndryKind, err);
        return(err);
      }
#endif !ZLS_USER_BOUNDARIES
      
    }  // bndryType
    
  } else { err = memErr; } // z
	
	return (err);
	
}

#ifdef ZLS_USER_BOUNDARIES
int  zlsbc_ImposeZoneUserBCs ( ZonePtr z, int direction ){
	
	Integer err = noErr;
	
	int bndryType;  // what sort of boundary, super, link, calc
	int bndryKind;   // what sub stype of calculated bndry - const, periodic etc
  
	if (z != NULL){
    
    // get  zone area parameters
    // mask off type for the given direction
    
    bndryType  = z->zoneBCType[direction]&kBoundaryTypeFlagUserBC; 
    bndryKind  = z->zoneBCType[direction]&kBCCKindMask; 
		
    if  (bndryType == kBoundaryTypeFlagUserBC) {
      err = zlsp_CalcUserZoneBCs (z, direction, bndryKind ); // see zls_problem.c
      if (err != noErr){
        zlsio_ErrPrintf( " ERROR: zlsbc_ImposeZoneUserBCs from zlsp_CalcUserZoneBCs: Z %d  dir %d kind %d E %d \n",z->zoneRankID, direction,bndryKind, err);
        return(err);
      }
    }
		
	} else { err = memErr; } // z
	
	return (err);
	
}
#endif ZLS_USER_BOUNDARIES




int  zlsbc_FillZoneBCsFromSuperZone1D ( ZonePtr z , int iteration){
#ifndef SYM1D
  return( dimErr );
#else 
  
	Integer err = noErr;
	//
	
	Integer    i,   id;	 //  loop counters
	Integer    l      ;          //  sub cell loop counters
	
	Integer         nR;
	Integer       nsbc;   
	Range          rng;
	Integer     interp;
	
	Integer  izd;     // i and j in zone data
	Integer  isd;     // i and j in super data
	
	InPtr            f;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
	
	Integer   ia,  i_0, nir; // super range in active area coord
	
	Integer  izs ;     // zone start offsets
	Integer  iss ;     // super start offsets
	
	Integer  nti,  nsti;     // zonedata full range
	
	ZonePtr     zSuper = NULL;
	Q1DArr      qSuper = NULL;
	Q1DArr           q;
  Real             a;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          
          nti    = z->zoneGrid.nt[kXDim];
          izs    = z->zoneGrid.ns[kXDim];
          
          // get Super zone area parameters
          
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          iss    = zSuper->zoneGrid.ns[kXDim]; // also we offset from ns, not 0
          
          //
          // zone range on super zone in active area coordinates, expanded to include 
          // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
          // cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
          // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
          // range by 2 nsbcs.
          // 
          // coordinate in q and vars = i + i_0, j + j_0
          //
          
          
          ia   = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
          i_0  = ia+iss;                      // i_0 is now absolute coord in super of start of area
          nir  = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
#ifdef LAGRANGIAN
          nsbc -= (iteration>0)?1:0; // if the iteration is > 0 then reduce the BCs filled by 1 super cell
          // required that nsbc >1 to start with, used by lagrangian recursive sweeps.
#endif LAGRANGIAN
          
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              
              q = z->zoneData[id];
              
              if (q!=NULL){
                
                switch (interp){
                    
                  case  kInterpNN: // simple rebin
                    
                    qSuper = zSuper->zoneData[id];
                    
                    if (qSuper != NULL){
                      
                      if (nsbc > 1){
                        
                        for(i = 0; i < nsbc; i++){
                          izd = (i*nR);
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            q[(izd+l)] =  qSuper[isd];
                          } //l
                        }
                        for(i = (nir-nsbc); i<nir; i++){
                          izd = (i*nR); // right column
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            q[(izd+l)] =  qSuper[isd];
                          } //m
                        } // is over 0 .. nsbc-1, left and right cols
                        
                      } else { // nsbc = 1, remove nsbc loops
                        
                        i   = 0;
                        izd = 0;
                        isd = i_0;
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] =  qSuper[isd];
                        } //l
                        
                        i = nir-1;
                        izd = (i*nR); 
                        isd = (i+i_0);
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] =  qSuper[isd];
                        } //l
                      }
                      
                    } //qSuper
                    
                    break; // interp == kInterpNN
                    
                    case kInterpPLM: // Piecewise Linear Method interpolation 
                    
                    err     = zls_NewInterp1D (&f, nsti);
                    
                    if (f!=NULL) {
                      
                      err = zls_GetZoneVar1D          ( &v1dx, zSuper, id ); //  j in active area - 0-nyIn
                      err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL );
                      err = zlsppm_CalcLinInterpVar1D   ( v1dx, f, nR );
                      
                      for(i = 0; i<nsbc; i++){ // step in x in Super
                        izd = (i*nR); 
                        isd = (i+i_0);
                        a   = qSuper[isd];
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] =  a + (f[isd].c[l]); // step in x in z
                        } //l
                      } //i
                      
                      for(i = (nir-nsbc); i<nir; i++){ // step in x in Super
                        izd = (i*nR); 
                        isd = (i+i_0);
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] += (f[isd].c[l]); // step in x in z
                        } //l
                      } //i
                      
                      zls_DisposeVar1DArr( v1dx );
                      zls_DisposeInterp( f ); f = NULL;
                      
                    } else { err = memErr;}; //!f
                    
                    break;
                    
                    case kInterpPPM: // PPM interpolation same as ppmlr
                    
                    err     = zls_NewInterp1D (&f, nsti);
                    
                    if (f!=NULL) {
                      
                      err = zls_GetZoneVar1D          ( &v1dx, zSuper, id ); //  j in active area - 0-nyIn
                      err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL );
                      err = zlsppm_CalcInterpVar1D      ( v1dx, f, nR );
                      
                      for(i = 0; i<nsbc; i++){ // step in x in Super
                        izd = (i*nR); 
                        isd = (i+i_0);
                        a   = qSuper[isd];
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] = a + (f[isd].c[l]); // step in x in z
                        } //l
                      } //i
                      
                      for(i = (nir-nsbc); i<nir; i++){ // step in x in Super
                        izd = (i*nR); 
                        isd = (i+i_0);
                        for( l = 0; l<nR; l++){
                          q[(izd+l)] += (f[isd].c[l]); // step in x in z
                        } //l
                      } //i
                      
                      zls_DisposeVar1DArr( v1dx );
                      zls_DisposeInterp( f ); f = NULL;
                      
                    } else { err = memErr;}; //!f
                    
                    break; // kInterpPPM
                    
                    default:
                    
                    err = nyiErr;
                    zlsio_Printf( " Fill1DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                    
                }
                
              } else { err = memErr; } // q, zSuper
              
            }
            
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // super
      
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
#endif SYM1D
  
}


int  zlsbc_FillZoneBCsFromSuperZone2D ( ZonePtr z , int iteration){
  
#ifndef SYM2D
  return( dimErr );
#else 
  
	Integer err = noErr;
	//
	
	Integer    i,    j,   id;	 //  loop counters
	Integer    l,    m;          //  sub cell loop counters
	
	Integer         nR;
	Integer       nsbc;   
	Range          rng;
	Integer     interp;
  Real             a;
  Real             f;
	
  Integer  izd,  jzd;     // i and j in zone data
  Integer  isd,  jsd;     // i and j in super data
	
	InPtr           fx;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
  V1DArr        v1dy;		// x and y rows of the  super zone
	
  Integer   ia, i_0,   nir; // super range in active area coord
	Integer   ja,  j_0,  njr; // super range in active area coord
	Integer  izs,  jzs;     // zone start offsets
	Integer  iss,  jss;     // super start offsets
	
	Integer  nti,  ntj;     // zonedata full range
	Integer nsti, nstj;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q2DArr      qSuper;
	Q2DArr           q;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          
          nti    = z->zoneGrid.nt[kXDim];
          ntj    = z->zoneGrid.nt[kYDim];
          izs    = z->zoneGrid.ns[kXDim];
          jzs    = z->zoneGrid.ns[kYDim];
          
          // get Super zone area parameters
          
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          nstj   = zSuper->zoneGrid.nt[kYDim]; // 
          iss    = zSuper->zoneGrid.ns[kXDim]; // also we offset from ns, not 0
          jss    = zSuper->zoneGrid.ns[kYDim];
          
          //
          // zone range on super zone in active area coordinates, expanded to include 
          // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
          // cells equivalent to nBCs in current zone, ie nR * nsbc = nBC.
          // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
          // range by 2 nsbcs.
          // 
          // coordinate in q and vars = i + i_0, j + j_0
          //
          
          
          ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in super active areas
          i_0 = ia+iss;                      // i_0 is now absolute coord in super of start of area
          nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
          
          ja  = rng.n0[kYDim]-nsbc;          // ja is coord of start of copy are in active coordinates
          j_0 = ja+jss;                     // j_0 is now absolute coord in super of start of area 
          njr = rng.rn[kYDim]+nsbc+nsbc;	   // njr is the range of the j loop in super cells
          
#ifdef LAGRANGIAN
          //      iteration = 0;
          nsbc -= (iteration>0)?1:0; // if the iteration is > 0 then reduce the BCs filled by 1 super cell
          // required that nsbc >1 to start with, used by lagrangian recursive sweeps.
#endif LAGRANGIAN
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              
#ifdef ZONEDATA_ARE_PLAINBLOCKS
              zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
              
#ifdef ZONEDATA_ARE_MULTIDARRAYS
              q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
              
              if (q!=NULL){
                
                switch (interp){
                    
                  case  kInterpNN: // simple rebin
                    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                    zls_SetQuantity2DArr(&qSuper,zSuper->zoneData[id], nstj, nsti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
                    qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
                    
                    if (qSuper != NULL){
                      
                      if (nsbc > 1){
                        
                        for(j = 0; j < njr; j++){
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for(i = 0; i < nsbc; i++){
                            izd = (i*nR);
                            isd = (i+i_0);
                            for( l = 0; l<nR; l++){
                              for( m = 0; m<nR; m++){
                                q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                              } //m
                            } //l
                          }
                          for(i = (nir-nsbc); i<nir; i++){
                            izd = (i*nR); // right column
                            isd = (i+i_0);
                            for( l = 0; l<nR; l++){
                              for( m = 0; m<nR; m++){
                                q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                              } //m
                            } //l
                          } // is over 0 .. nsbc-1, left and right cols
                        } //j
                        
                        for(j = 0; j < nsbc; j++){ 
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for(i = 0; i < nir; i++){
                            izd = (i*nR); 
                            isd = (i+i_0);
                            for( l = 0; l<nR; l++){
                              for( m = 0; m<nR; m++){
                                q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                              } //m
                            } //l
                          } //i
                        } // j
                        
                        for(j = (njr-nsbc); j<njr; j++){ 
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for(i = 0; i<nir; i++){ 
                            izd = (i*nR); 
                            isd = (i+i_0);
                            for( l = 0; l<nR; l++){
                              for( m = 0; m<nR; m++){
                                q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                              } //m
                            } //l
                          } //i
                        } // j over 0...nsbc-1 top and bottom rows
                        
                      } else { // nsbc = 1, remove nsbc loops
                        
                        for(j = 0; j<njr; j++){
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          
                          i   = 0;
                          izd = 0;
                          isd = i_0;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                            } //m
                          } //l
                          
                          i = nir-1;
                          izd = (i*nR); 
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                            } //m
                          } //l
                        } //j
                        
                        j   = 0;
                        jzd = 0; // bottom row
                        jsd = j_0;
                        for(i = 0; i < nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                            } //m
                          } //l
                        } //i
                        
                        j   = njr-1;
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        for(i = 0; i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] =  qSuper[jsd][isd];
                            } //m
                          } //l
                        } //i
                      }
                      
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                      zls_ReleaseQuantity2DArr(qSuper);
#endif ZONEDATA_ARE_PLAINBLOCKS
                      
                    } //qSuper
                    
                    break; // interp == kInterpNN
                    
                    case kInterpPLM: // Piecewise Linear Method interpolation 
                    
                    err     = zls_NewInterp1D (&fx, nsti);
                    if (fx!=NULL) {
                      
                      for(j = 0; j<nsbc; j++){ // bottom
                        
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                        
                        for(i = 0; i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f  = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        
                        zls_DisposeVar1DArr(v1dx);
                        
                      } //j
                      
                      
                      for(j = nsbc; j<(nir-nsbc); j++){  // middle
                        
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                        
                        for(i = 0; i<nsbc; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f  = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] = a +  f; // step in x in z
                            } //m
                          } //l
                        } //i
                        
                        for(i = (nir-nsbc); i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f  = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] = a +  f; // step in x in z
                            } //m
                          } //l
                        } //i
                        
                        zls_DisposeVar1DArr(v1dx);
                        
                      } //j
                      
                      for(j = (njr-nsbc); j<njr; j++) { // top
                        
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                        
                        for(i = 0; i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f  = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] = a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        
                        zls_DisposeVar1DArr(v1dx);
                        
                      } //j
                      
                      zls_DisposeInterp(fx); fx = NULL;
                    } else { err = memErr;}; //!fx
                    
                    err     = zls_NewInterp1D (&fx, nstj);
                    
                    if (fx != NULL) {
                      for(i = 0; i<nsbc; i++){ // left
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dy, fx, nR);
                        for(j = 0; j< njr; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f  = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] += f; // step in y in z
                            } //m
                          } //l
                        } //j
                        zls_DisposeVar1DArr(v1dy);
                      } //i
                      
                      for(i = nsbc; i<(nir-nsbc); i++){ // middle
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dy, fx, nR);
                        for(j = 0; j< nsbc; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f  = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] += f; // step in y in z
                            } //m
                          } //l
                        } //j
                        for(j = (njr-nsbc); j< njr; j++){ 
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f  = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        zls_DisposeVar1DArr(v1dy);
                      } //i
                      
                      for(i = (nir-nsbc); i<nir; i++){ // right
                        
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                        err = zlsppm_CalcLinInterpVar1D ( v1dy, fx, nR);
                        
                        for(j = 0; j< njr; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f  = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        
                        zls_DisposeVar1DArr(v1dy);
                        
                      } //i
                      
                      zls_DisposeInterp(fx); fx = NULL;
                    } else { err = memErr;} //!fx
                    
                    break;
                    
                    case kInterpPPM: // PPM interpolation same as ppmlr
                    
                    err     = zls_NewInterp1D (&fx, nsti);
                    
                    if (fx!=NULL) {
                      
                      for(j = 0; j<nsbc; j++){ // bottom
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                        err = zlsppm_CalcInterpVar1D ( v1dx, fx, nR);
                        for(i = 0; i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        zls_DisposeVar1DArr(v1dx);
                      } //j
                      
                      for(j = nsbc; j<(nir-nsbc); j++){ // middle
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr    ( v1dx, NULL);
                        err = zlsppm_CalcInterpVar1D         ( v1dx, fx, nR);
                        for(i = 0; i<nsbc; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        for(i = (nir-nsbc); i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        zls_DisposeVar1DArr(v1dx);
                      } //j
                      
                      for(j = (njr-nsbc); j<njr; j++){ // top
                        jzd = (j*nR);
                        jsd = (j+j_0);
                        err = zls_GetZone2DVar1DX    ( &v1dx, zSuper, id, j+ja); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr    ( v1dx, NULL);
                        err = zlsppm_CalcInterpVar1D         ( v1dx, fx, nR);
                        for(i = 0; i<nir; i++){ // step in x in Super
                          izd = (i*nR); 
                          isd = (i+i_0);
                          a = v1dx.v[isd].a;
                          for( l = 0; l<nR; l++){
                            for( m = 0; m<nR; m++){
                              f = fx[isd].c[m];
                              q[(jzd+l)][(izd+m)] =  a + f; // step in x in z
                            } //m
                          } //l
                        } //i
                        zls_DisposeVar1DArr(v1dx);
                      } //j
                      
                      zls_DisposeInterp(fx); fx = NULL;
                    } else { err = memErr;}; //!fx
                    
                    err     = zls_NewInterp1D (&fx, nstj);
                    
                    if (fx!=NULL) {
                      
                      for(i = 0; i<nsbc; i++){
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr     ( v1dy, NULL);
                        err = zlsppm_CalcInterpVar1D          ( v1dy, fx, nR);
                        for(j = 0; j< njr; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        zls_DisposeVar1DArr(v1dy);
                      } //i
                      
                      for(i = nsbc; i<(nir-nsbc); i++){
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                        err = zlsppm_CalcInterpVar1D ( v1dy, fx, nR);
                        for(j = 0; j< nsbc; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        for(j = (njr-nsbc); j< njr; j++){
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        zls_DisposeVar1DArr(v1dy);
                      } //i
                      
                      for(i = (nir-nsbc); i<nir; i++){
                        izd = (i*nR); 
                        isd = (i+i_0);
                        err = zls_GetZone2DVar1DY    ( &v1dy, zSuper, id, i+ia ); //  j in active area - 0-nyIn
                        err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                        err = zlsppm_CalcInterpVar1D ( v1dy, fx, nR);
                        for(j = 0; j< njr; j++){ // step in y in Super
                          jzd = (j*nR);
                          jsd = (j+j_0);
                          for( l = 0; l<nR; l++){
                            f = fx[jsd].c[l];
                            for( m = 0; m<nR; m++){
                              q[(jzd+l)][(izd+m)] +=  f; // step in y in z
                            } //m
                          } //l
                        } //j
                        zls_DisposeVar1DArr(v1dy);
                      } //i
                      
                      zls_DisposeInterp(fx); fx = NULL;
                      
                    } else { err = memErr;} //!fx
                    
                    break; // kInterpPPM
                    
                    default:
                    
                    err = nyiErr;
                    zlsio_Printf( " Fill2DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                    
                }
                
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
                
              } else { err = memErr; } // q, zSuper
              
            }
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // zSuper
      
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
  
#endif SYM2D
  
}



int  zlsbc_FillZoneBCsFromSuperZone3D ( ZonePtr z , int iteration){
  //
  // in 3D the complex logic of the 1D and 2D versions gets too hard to write reliably,
  // so we choose a simpler scheme that may do a little more work but is easier to 
  // follow.  Based on the full fill 3D from super plus a 3D range limit on the innermost loops
  // Because the tests aren't done until the inner loops, more unnecessary work is done
  // but the alternative is too hard to debug
  //
  
#ifndef SYM3D
  return( dimErr );
#else
	Integer err = noErr;
	
	Integer    i,    j,    k,   id;	 //  loop counters
	Integer    l,    m,    n ;          //  sub cell loop counters
	
	Integer         nR;
	Integer         nG;
	Integer       nsbc;   
	Integer     interp;
	Range          rng;
	
	Real  f, a;
	
	Integer  izd,  jzd,   kzd;     // i and j and k in zone data
	Integer  isd,  jsd,   ksd;     // i and j and k in super data
	
	InPtr        fx, fy, fz;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
	V1DArr        v1dy;		// x , y , z rows of the  super zone
	V1DArr        v1dz;		
	
	Integer   ia,  i_0,  nir; // super range in active area coord
	Integer   ja,  j_0,  njr; // super range in active area coord
	Integer   ka,  k_0,  nkr; // super range in active area coord
	Integer  izs,  jzs,  kzs;     // zone start offsets
	Integer  iss,  jss,  kss;     // super start offsets
	
	Integer   nsi, nsj, nsk;     // super data full range for making vars
	Integer  nti,  ntj, ntk;     // zonedata full range
	Integer nsti, nstj, nstk;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q3DArr      qSuper;
	Q3DArr           q;
	
	if (z != NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zSuper = z->zoneSuper;
      interp = z->zoneBCInterp;
      
      if (zSuper != NULL){
#ifdef MPI    
        if (zSuper->zoneRankID == gMPIrank) {
#endif MPI    
          
          rng    = z->zoneSuperRange;
          nR     = z->zoneRefine;
          nG     = z->zoneBCells;
          nsbc   = z->zoneSuperBCells;
          
          // get  zone area parameters
          
          izs    = z->zoneGrid.ns[kXDim];
          jzs    = z->zoneGrid.ns[kYDim];
          kzs    = z->zoneGrid.ns[kZDim];
          
          nti    = z->zoneGrid.nt[kXDim];
          ntj    = z->zoneGrid.nt[kYDim];
          ntk    = z->zoneGrid.nt[kZDim];
          
          // get Super zone area parameters
          
          iss    = zSuper->zoneGrid.ns[kXDim];
          jss    = zSuper->zoneGrid.ns[kYDim];
          kss    = zSuper->zoneGrid.ns[kZDim];
          
          nsi   = zSuper->zoneGrid.n[kXDim]; 
          nsj   = zSuper->zoneGrid.n[kYDim];
          nsk   = zSuper->zoneGrid.n[kZDim];
          
          nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
          nstj   = zSuper->zoneGrid.nt[kYDim]; // 
          nstk   = zSuper->zoneGrid.nt[kZDim]; // 
          
          //
          // zone range on super zone in active area coordinates, expanded to include 
          // cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
          // cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
          // superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
          // range by 2 nsbcs.
          // 
          // coordinate in q and vars = i + i_0, j + j_0, k + k_0
          //
          ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
          i_0  = ia+iss;                     // i_0 is now absolute coord in super of start of area
          nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
          
          ja  = rng.n0[kYDim]-nsbc;          // ja is coord of start of copy are in active coordinates
          j_0 = ja+jss;                      // i_0 is now absolute coord in super of start of area 
          njr = rng.rn[kYDim]+nsbc+nsbc;	   // njr is the range of the j loop in super cells
          
          ka  = rng.n0[kZDim]-nsbc;          // ja is coord of start of copy are in active coordinates
          k_0 = ka+kss;                      // k_0 is now absolute coord in super of start of area 
          nkr = rng.rn[kZDim]+nsbc+nsbc;	   // nkr is the range of the k loop in super cells
          
#ifdef LAGRANGIAN
          iteration = 0;
          nsbc -= (iteration>0)?1:0; // if the iteration is > 0 then reduce the BCs filled by 1 super cell
          // required that nsbc >1 to start with, used by lagrangian recursive sweeps.
#endif LAGRANGIAN
          
          izs = 0;
          jzs = 0;
          kzs = 0;
          
          for (id = 0; id <  z->nQ; id++){
            
            if (z->zoneData[id]!=NULL){
              if (zSuper->zoneData[id]!=NULL){
                
#ifdef ZONEDATA_ARE_PLAINBLOCKS
                zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
                zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[id], nstk, nstj, nsti);
#endif ZONEDATA_ARE_PLAINBLOCKS
                
#ifdef ZONEDATA_ARE_MULTIDARRAYS
                q = z->zoneData[id];
                qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
                
                if (q != NULL){
                  if (qSuper != NULL){
                    
                    switch (interp){
                        
                      case  kInterpNN: // simple rebin
                        
                        for(k = 0; k < nkr; k++){
                          kzd = kzs+(k*nR);
                          ksd = (k+k_0);
                          for(j = 0; j < njr; j++){
                            jzd = jzs+(j*nR);
                            jsd = (j+j_0);
                            for(i = 0; i < nir; i++){
                              izd = izs+(i*nR);
                              isd = (i+i_0);
                              
                              if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                  (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                
                                for( l = 0; l<nR; l++){
                                  for( m = 0; m<nR; m++){
                                    for( n = 0; n<nR; n++){
                                      q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
                                    } //n
                                  } //m
                                } //l
                                
                              }
                              
                            } // i
                          } //j
                        } //k
                        
                        break; // interp == kInterpNN
                        
                        case kInterpPLM: // Piecewise Linear Method interpolation 
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewVar1DArr (&v1dz, nG, nsk);
                        v1dz.x = zSuper->zoneGrid.z;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        err += zls_NewInterp1D (&fz, nstk);
                        
                        if (err == noErr) {
                          
                          for(k = 0; k<nkr; k++){ // rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(j = 0; j<njr; j++){ // rows, set and mult by x coeffs
                              jzd = jzs+(j*nR);
                              jsd = (j+ja);
                              zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
                              for(i = 0; i<nir; i++){ // step in x in Super
                                izd = izs+(i*nR); 
                                isd = (i+i_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  a = v1dx.v[isd].a;
                                  for( l = 0; l<nR; l++){
                                    for( m = 0; m<nR; m++){
                                      for( n = 0; n<nR; n++){
                                        f = fx[ksd].c[n];
                                        q[(kzd+l)][(jzd+m)][(izd+n)] =  a  + f; // step in x in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //i
                            } //j
                          } //k
                          
                          for(k = 0; k<nkr; k++){ // rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(i = 0; i<nir; i++){ // rows, set and mult by y coeffs
                              izd = izs+(i*nR); 
                              isd = (i+ia);
                              err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dy, fy, nR);
                              for(j = 0; j< njr; j++){ // step in y in Super
                                jzd = jzs+(j*nR);
                                jsd = (j+j_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  for( l = 0; l<nR; l++){
                                    for( m = 0; m<nR; m++){
                                      f = fy[ksd].c[m];
                                      for( n = 0; n<nR; n++){
                                        q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //j
                            } //i
                          } // k
                          
                          for(i = 0; i<nir; i++){ // rows, set and mult by z coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            for(j = 0; j<njr; j++){ // rows, set and mult by z coeffs
                              jzd = jzs+(j*nR); 
                              jsd = (j+ja);
                              err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
                              err = zlsppm_CalcLinInterpVar1D ( v1dz, fz, nR);
                              for(k = 0; k< nkr; k++){ // step in z in Super
                                kzd = kzs+(k*nR);
                                ksd = (k+k_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  for( l = 0; l<nR; l++){
                                    f = fz[ksd].c[l];
                                    for( m = 0; m<nR; m++){
                                      for( n = 0; n<nR; n++){
                                        q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //k
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dz);
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          zls_DisposeInterp(fy); fz = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break;
                        
                        case kInterpPPM: // PPM interpolation same as ppmlr
                        
                        err  = zls_NewVar1DArr (&v1dx, nG, nsi);
                        v1dx.x = zSuper->zoneGrid.x;
                        err += zls_NewVar1DArr (&v1dy, nG, nsj);
                        v1dy.x = zSuper->zoneGrid.y;
                        err += zls_NewVar1DArr (&v1dz, nG, nsk);
                        v1dz.x = zSuper->zoneGrid.z;
                        err += zls_NewInterp1D (&fx, nsti);
                        err += zls_NewInterp1D (&fy, nstj);
                        err += zls_NewInterp1D (&fz, nstk);
                        
                        if (err == noErr) {
                          
                          for(k = 0; k<nkr; k++){ // rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(j = 0; j<njr; j++){ // rows, set and mult by x coeffs
                              jzd = jzs+(j*nR);
                              jsd = (j+ja);
                              zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dx, fx, nR);
                              for(i = 0; i<nir; i++){ // step in x in Super
                                izd = izs+(i*nR); 
                                isd = (i+i_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  a = v1dx.v[isd].a;
                                  for( l = 0; l<nR; l++){
                                    for( m = 0; m<nR; m++){
                                      for( n = 0; n<nR; n++){
                                        f = fx[ksd].c[n];
                                        q[(kzd+l)][(jzd+m)][(izd+n)] =  a  +  f; // step in x in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //i
                            } //j
                          } //k
                          
                          for(k = 0; k<nkr; k++){ // rows, set and mult by y coeffs
                            kzd = kzs+(k*nR); 
                            ksd = (k+ka);
                            for(i = 0; i<nir; i++){ // rows, set and mult by y coeffs
                              izd = izs+(i*nR); 
                              isd = (i+ia);
                              err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dy, fy, nR);
                              for(j = 0; j< njr; j++){ // step in y in Super
                                jzd = jzs+(j*nR);
                                jsd = (j+j_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  
                                  for( l = 0; l<nR; l++){
                                    for( m = 0; m<nR; m++){
                                      f = fy[ksd].c[m];
                                      for( n = 0; n<nR; n++){
                                        q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //j
                            } //i
                          } // k
                          
                          for(i = 0; i<nir; i++){ // rows, set and mult by z coeffs
                            izd = izs+(i*nR); 
                            isd = (i+ia);
                            for(j = 0; j<njr; j++){ // rows, set and mult by z coeffs
                              jzd = jzs+(j*nR); 
                              jsd = (j+ja);
                              err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
                              err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
                              err = zlsppm_CalcInterpVar1D ( v1dz, fz, nR);
                              for(k = 0; k< nkr; k++){ // step in z in Super
                                kzd = kzs+(k*nR);
                                ksd = (k+k_0);
                                
                                if ((i < nsbc) || ( j < nsbc) || ( k < nsbc) ||
                                    (i >= nir-nsbc) || ( j >= njr-nsbc) || ( k >= nkr-nsbc)) {
                                  
                                  
                                  for( l = 0; l<nR; l++){
                                    f = fz[ksd].c[l];
                                    for( m = 0; m<nR; m++){
                                      for( n = 0; n<nR; n++){
                                        q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
                                      } //n
                                    } //m
                                  } //l
                                  
                                }
                                
                              } //k
                            } //j
                          } //i
                          
                          zls_DisposeVar1DArr(v1dz);
                          zls_DisposeVar1DArr(v1dy);
                          zls_DisposeVar1DArr(v1dx);
                          
                          zls_DisposeInterp(fx); fx = NULL;
                          zls_DisposeInterp(fy); fy = NULL;
                          zls_DisposeInterp(fy); fz = NULL;
                          
                        } else {
                          zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
                          return(err);
                        }
                        
                        break; // kInterpPPM
                        
                        default:
                        
                        err = nyiErr;
                        zlsio_Printf( " Fill2DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
                        
                    }
                    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
                    zls_ReleaseQuantity2DArr(qSuper);
#endif SYM2D
#ifdef SYM3D
                    zls_ReleaseQuantity3DArr(qSuper);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
                    
                  } // qSuper		
                  
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
                  zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
                  zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
                  
                } // q
                
              } // zSuperdata
            } // zdata
          } // var id loop
#ifdef MPI			
        }
#endif MPI
        
      } // zSuper
#ifdef MPI			
    }
#endif MPI
		
	} else { err = memErr; } // z
	
	return (err);
  
#endif SYM3D
  
}



int  zlsbc_FillZoneBCsFromSuperZone ( ZonePtr z , int iteration){
  int err = noErr;
#ifdef SYM1D
  err = zlsbc_FillZoneBCsFromSuperZone1D ( z , iteration); 
#endif SYM1D
#ifdef SYM2D
  err = zlsbc_FillZoneBCsFromSuperZone2D ( z , iteration); 
#endif SYM2D
#ifdef SYM3D
  err = zlsbc_FillZoneBCsFromSuperZone3D ( z , iteration); 
#endif SYM3D
  return(err);
}


int  old_zls_FillZoneBCsFromSuperZone3D ( ZonePtr z, int iteration );
int  old_zls_FillZoneBCsFromSuperZone3D ( ZonePtr z, int iteration ){
  
#ifndef SYM3D
  return( dimErr );
#else 
	
	Integer err = noErr;
	//
	
	Integer    i,    j,    k,   id;	 //  loop counters
	Integer    l,    m,    n ;          //  sub cell loop counters
	
	Integer         nR;
	Integer         nG;
	Integer       nsbc;   
	Range          rng;
	Integer		  interp;
	
	Real  f, a;
	
	Integer  izd,  jzd,   kzd;     // i and j and k in zone data
	Integer  isd,  jsd,   ksd;     // i and j and k in super data
	
	InPtr        fx, fy, fz;		// interpolation fraction array
	V1DArr        v1dx;     // extracted vars/coeffs for 
	V1DArr        v1dy;		// x , y , z rows of the  super zone
	V1DArr        v1dz;		
	
	Integer   ia,  i_0,  nir; // super range in active area coord
	Integer   ja,  j_0,  njr; // super range in active area coord
	Integer   ka,  k_0,  nkr; // super range in active area coord
	Integer  izs,  jzs,  kzs;     // zone start offsets
	Integer  iss,  jss,  kss;     // super start offsets
	
	
	Integer   nsi, nsj, nsk;     // super data full range for making vars
	Integer  nti,  ntj, ntk;     // zonedata full range
	Integer nsti, nstj, nstk;     // super data full range for making vars
	
	ZonePtr     zSuper;
	Q3DArr      qSuper;
	Q3DArr           q;
	
	if (z != NULL){
		
		zSuper = z->zoneSuper;
		interp = z->zoneBCInterp;
		
		if (zSuper != NULL){
			
			rng    = z->zoneSuperRange;
			nR     = z->zoneRefine;
			nG     = z->zoneBCells;
			nsbc   = z->zoneSuperBCells;
			
			// get  zone area parameters
			iss    = zSuper->zoneGrid.ns[kXDim];
			jss    = zSuper->zoneGrid.ns[kYDim];
			kss    = zSuper->zoneGrid.ns[kZDim];
			
			izs    = z->zoneGrid.ns[kXDim];
			jzs    = z->zoneGrid.ns[kYDim];
			kzs    = z->zoneGrid.ns[kZDim];
			
			nsi   = zSuper->zoneGrid.n[kXDim]; 
			nsj   = zSuper->zoneGrid.n[kYDim];
			nsk   = zSuper->zoneGrid.n[kZDim];
			
			// get Super zone area parameters
			nti    = z->zoneGrid.nt[kXDim];
			ntj    = z->zoneGrid.nt[kYDim];
			ntk    = z->zoneGrid.nt[kZDim];
			
			nsti   = zSuper->zoneGrid.nt[kXDim]; // Super zone data has BC, so size is nt not just n
			nstj   = zSuper->zoneGrid.nt[kYDim]; // 
			nstk   = zSuper->zoneGrid.nt[kZDim]; // 
			
			//
			// zone range on super zone in active area coordinates, expanded to include 
			// cells in the superzone sufficient to cover the BCs in the zone, nsbc = number of superzone
			// cells equal to BCs in zone, ie nR * nsbc = nBC in zone.
			// superzone nBC > nsbc so this is always safe.  Subtract nsbc from the left and extend
			// range by 2 nsbcs.
			// 
			// coordinate in q and vars = i + i_0, j + j_0, k + k_0
			//
			ia  = rng.n0[kXDim]-nsbc;          // ia is now  coord in super of start of area in active areas
			i_0  = ia+iss;                     // i_0 is now absolute coord in super of start of area
			nir = rng.rn[kXDim]+nsbc+nsbc;     // nir is the range of the i loop in super cells
			
			ja  = rng.n0[kYDim]-nsbc;          // ja is coord of start of copy are in active coordinates
			j_0 = ja+jss;                      // i_0 is now absolute coord in super of start of area 
			njr = rng.rn[kYDim]+nsbc+nsbc;	   // njr is the range of the j loop in super cells
			
			ka  = rng.n0[kZDim]-nsbc;          // ja is coord of start of copy are in active coordinates
			k_0 = ka+kss;                      // k_0 is now absolute coord in super of start of area 
			nkr = rng.rn[kZDim]+nsbc+nsbc;	   // nkr is the range of the k loop in super cells
			
#ifdef LAGRANGIAN
			nsbc -= (iteration>0)?1:0; // if the iteration is > 0 then reduce the BCs filled by 1 super cell
      // required that nsbc >1 to start with, used by lagrangian recursive sweeps.
#endif LAGRANGIAN
			izs = 0;
			jzs = 0;
			kzs = 0;
			
			for (id = 0; id <  z->nQ; id++){
				
				if (z->zoneData[id]!=NULL){
					if (zSuper->zoneData[id]!=NULL){
						
#ifdef ZONEDATA_ARE_PLAINBLOCKS
						zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
						zls_SetQuantity3DArr(&qSuper,zSuper->zoneData[id], nstk, nstj, nsti);
#endif ZONEDATA_ARE_PLAINBLOCKS
						
#ifdef ZONEDATA_ARE_MULTIDARRAYS
						q = z->zoneData[id];
						qSuper = zSuper->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
						
						if (q != NULL){
							if (qSuper != NULL){
								
								switch (interp){
                    
									case  kInterpNN: // simple rebin
										
										for(k = 0; k < nkr; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = 0; j < njr; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = 0; i < nsbc; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										
										for(k = 0; k < nkr; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = 0; j < njr; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = (nir-nsbc); i < nir; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										for(k = 0; k < nkr; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = 0; j < nsbc; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = 0; i < nir; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										for(k = 0; k < nkr; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = (njr-nsbc); j < njr; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = 0; i < nir; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										for(k = 0; k < nsbc; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = 0; j < njr; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = 0; i < nir; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										for(k = (nkr-nsbc); k < nkr ; k++){
											kzd = kzs+(k*nR);
											ksd = (k+k_0);
											for(j = 0; j < njr; j++){
												jzd = jzs+(j*nR);
												jsd = (j+j_0);
												for(i = 0; i < nir; i++){
													izd = izs+(i*nR);
													isd = (i+i_0);
													
													for( l = 0; l<nR; l++){
														for( m = 0; m<nR; m++){
															for( n = 0; n<nR; n++){
																q[(kzd+l)][(jzd+m)][(izd+n)] =  qSuper[ksd][jsd][isd];
															} //n
														} //m
													} //l
													
												} // i
											} //j
										} //k
										
										
										break; // interp == kInterpNN
										
                    case kInterpPLM: // Piecewise Linear Method interpolation 
										
										err  = zls_NewVar1DArr (&v1dx, nG, nsi);
										v1dx.x = zSuper->zoneGrid.x;
										err += zls_NewVar1DArr (&v1dy, nG, nsj);
										v1dy.x = zSuper->zoneGrid.y;
										err += zls_NewVar1DArr (&v1dz, nG, nsk);
										v1dz.x = zSuper->zoneGrid.z;
										err += zls_NewInterp1D (&fx, nsti);
										err += zls_NewInterp1D (&fy, nstj);
										err += zls_NewInterp1D (&fz, nstk);
										
										if (err == noErr) {
											
											for(k = 0; k<nkr; k++){ // left
												kzd = kzs+(k*nR); 
												ksd = (k+ka);
												for(j = 0; j<njr; j++){ // rows, set and mult by x coeffs
													jzd = jzs+(j*nR);
													jsd = (j+ja);
													zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
													err = zlsppm_CalcLinInterpVar1D ( v1dx, fx, nR);
													for(i = 0; i<nir; i++){ // step in x in Super
														if ((i < nsbc)||(i >= (nir - nsbc))){
															izd = izs+(i*nR); 
															isd = (i+i_0);
															
															a = v1dx.v[isd].a;
															for( n = 0; n<nR; n++){
																f = fx[isd].c[n];
																if (f != 0.0){
																	for( l = 0; l<nR; l++){
																		for( m = 0; m<nR; m++){
																			q[(kzd+l)][(jzd+m)][(izd+n)] = a + f; // step in x in z
																		} //m
																	} //l
																} // f
															} //n
														} // bc
														
													} //i
												} //j
											} //k
											
											for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
												kzd = kzs+(k*nR); 
												ksd = (k+ka);
												for(i = 0; i<nir; i++){// rows, set and mult by y coeffs
													izd = izs+(i*nR); 
													isd = (i+ia);
													err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
													err = zlsppm_CalcLinInterpVar1D ( v1dy, fy, nR);
													for(j = 0; j< njr; j++){ // step in y in Super
														if ((j < nsbc)||(j >= (njr - nsbc))){
															jzd = jzs+(j*nR);
															jsd = (j+j_0);
															
															for( m = 0; m<nR; m++){
																f = fy[jsd].c[m];
																if (f != 0.0){
																	for( l = 0; l<nR; l++){
																		for( n = 0; n<nR; n++){
																			q[(jzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
																		} //n
																	} //l
																} //f
															} //m
														} // bc
														
													} //j
												} //i
											} // k
											
											for(i = 0; i<nir; i++){// rows, set and mult by z coeffs
												izd = izs+(i*nR); 
												isd = (i+ia);
												for(j = 0; j<njr; j++){// rows, set and mult by z coeffs
													jzd = jzs+(j*nR); 
													jsd = (j+ja);
													err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
													err = zlsppm_CalcLinInterpVar1D ( v1dz, fz, nR);
													for(k = 0; k< nkr; k++){ // step in z in Super
														if ((k < nsbc)||(k >= (nkr - nsbc))){
															kzd = kzs+(k*nR);
															ksd = (k+k_0);
															for( l = 0; l<nR; l++){
																f = fz[ksd].c[l];
																if (f != 0.0){
																	for( m = 0; m<nR; m++){
																		for( n = 0; n<nR; n++){
																			q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
																		} //n
																	} //m
																}
															} //l
														} 
													} //k
												} //j
											} //i
											
											zls_DisposeVar1DArr(v1dz);
											zls_DisposeVar1DArr(v1dy);
											zls_DisposeVar1DArr(v1dx);
											
											zls_DisposeInterp(fx); fx = NULL;
											zls_DisposeInterp(fy); fy = NULL;
											zls_DisposeInterp(fy); fz = NULL;
											
										} else {
											zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
											return(err);
										}
                    
                    break;
										
                    case kInterpPPM: // PPM interpolation same as ppmlr
										
										err  = zls_NewVar1DArr (&v1dx, nG, nsi);
										v1dx.x = zSuper->zoneGrid.x;
										err += zls_NewVar1DArr (&v1dy, nG, nsj);
										v1dy.x = zSuper->zoneGrid.y;
										err += zls_NewVar1DArr (&v1dz, nG, nsk);
										v1dz.x = zSuper->zoneGrid.z;
										err += zls_NewInterp1D (&fx, nsti);
										err += zls_NewInterp1D (&fy, nstj);
										err += zls_NewInterp1D (&fz, nstk);
										
										if (err == noErr) {
											
											for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
												kzd = kzs+(k*nR); 
												ksd = (k+ka);
												for(j = 0; j<njr; j++){// rows, set and mult by x coeffs
													jzd = jzs+(j*nR);
													jsd = (j+ja);
													zls_SetVar1DArrQ3DArrX       ( v1dx, qSuper, ksd, jsd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dx, NULL);
													err = zlsppm_CalcInterpVar1D ( v1dx, fx, nR);
													for(i = 0; i<nir; i++){ // step in x in Super
														if ((i < nsbc)||(i >= (nir - nsbc))){
															izd = izs+(i*nR); 
															isd = (i+i_0);
															
															a = v1dx.v[isd].a;
															for( n = 0; n<nR; n++){
																f = fx[ksd].c[n];
																if ( f!= 0.0){
																	for( l = 0; l<nR; l++){
																		for( m = 0; m<nR; m++){
																			q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in x in z
																		} //m
																	} //l
																} // f
															} //n
														}
														
													} //i
												} //j
											} //k
											
											for(k = 0; k<nkr; k++){// rows, set and mult by y coeffs
												kzd = kzs+(k*nR); 
												ksd = (k+ka);
												for(i = 0; i<nir; i++){// rows, set and mult by y coeffs
													izd = izs+(i*nR); 
													isd = (i+ia);
													err = zls_SetVar1DArrQ3DArrY    ( v1dy, qSuper, ksd, isd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dy, NULL);
													err = zlsppm_CalcInterpVar1D ( v1dy, fy, nR);
													for(j = 0; j< njr; j++){ // step in y in Super
														if ((j < nsbc)||(j >= (njr - nsbc))){
															jzd = jzs+(j*nR);
															jsd = (j+j_0);
															
															for( m = 0; m<nR; m++){
																f = fy[ksd].c[m];
																if (f != 0.0) {
																	for( l = 0; l<nR; l++){
																		for( n = 0; n<nR; n++){
																			q[(jzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
																		} //n
																	} //l
																} // f
															} //m
														} //  bc
														
													} //j
												} //i
											} // k
											
											for(i = 0; i<nir; i++){// rows, set and mult by z coeffs
												izd = izs+(i*nR); 
												isd = (i+ia);
												for(j = 0; j<njr; j++){// rows, set and mult by z coeffs
													jzd = jzs+(j*nR); 
													jsd = (j+ja);
													err = zls_SetVar1DArrQ3DArrZ    ( v1dz, qSuper, jsd, isd);
													err = zlsppm_CalcCoeffsVar1DArr ( v1dz, NULL);
													err = zlsppm_CalcInterpVar1D    ( v1dz, fz, nR);
													for(k = 0; k< nkr; k++){ // step in z in Super
														if ((k < nsbc)||(k >= (nkr - nsbc))){
															kzd = kzs+(k*nR);
															ksd = (k+k_0);
															for( l = 0; l<nR; l++){
																f = fz[ksd].c[l];
																if ( f != 0.0) {
																	for( m = 0; m<nR; m++){
																		for( n = 0; n<nR; n++){
																			q[(kzd+l)][(jzd+m)][(izd+n)] += f; // step in y in z
																		} //n
																	} //m
																}
															} //l
														} 
													} //k
												} //j
											} //i
											
											zls_DisposeVar1DArr(v1dz);
											zls_DisposeVar1DArr(v1dy);
											zls_DisposeVar1DArr(v1dx);
											
											zls_DisposeInterp(fx); fx = NULL;
											zls_DisposeInterp(fy); fy = NULL;
											zls_DisposeInterp(fy); fz = NULL;
											
										} else {
											zlsio_ErrPrintf( " ERROR: NewFill3DZoneFromSuperZone temp 1d arrays failed: %d \n\n", interp); 
											return(err);
										}
                    
                    break; // kInterpPPM
										
                    default:
										
										err = nyiErr;
										zlsio_Printf( " Fill2DZoneFromSuperZone: Unknown Interpolation code: %d \n",interp);
										
								}
								
#ifdef ZONEDATA_ARE_PLAINBLOCKS
								zls_ReleaseQuantity3DArr(qSuper);
#endif ZONEDATA_ARE_PLAINBLOCKS
								
							} // qSuper		
							
#ifdef ZONEDATA_ARE_PLAINBLOCKS
							zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
							
						} // q
						
					} // zSuperdata
				} // zdata
			} // var id loop
			
		} // zSuper
		
	} else { err = memErr; } // z
	
	return (err);
	
#endif SYM3D
  
}


int  zlsbc_ImposeZoneSuperBCs ( ZonePtr z, int iteration ){
  //
	// set iteration to 0 for normal operation
	// ie  zlsbc_ImposeZoneSuperBCs ( z,0)
	// the iteration count is used in recursive lagrangian calculation by refined sweeps.
	// otherwise it is set to 0.
	//
	Integer err = noErr;
	int bndryType;  // what sort of boundary, super, link, calc
  
  Integer dirIndex;
	
	if (z != NULL){
		
		bndryType = kBoundaryTypeFlagNULL;
		for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) 
			bndryType |= (z->zoneBCType[dirIndex]&kBoundaryTypeFlagSuperZ); 
		
		if (bndryType == kBoundaryTypeFlagSuperZ) {  // one call fills all boundaries if any boundary need filling from super, individual BCs are overwritten later
#ifdef SYM1D
			err = zlsbc_FillZoneBCsFromSuperZone1D( z, iteration ); 
#endif SYM1D
#ifdef SYM2D
			err = zlsbc_FillZoneBCsFromSuperZone2D( z, iteration ); 
#endif SYM2D
#ifdef SYM3D
			err = zlsbc_FillZoneBCsFromSuperZone3D( z, iteration ); 
#endif SYM3D
		} 
		
	} else { err = memErr; } // z
	
	return (err);
}

//====================================================================================
//
//  Level dynamic updates/sweeps.
//
//  These are algorithm level calls and only process zones with zoneType == kAlgorithmZoneType
//  Image rouines to transfer data into image zones are elsewhere but are also
//  based on similar calls, but allow transfers between zones of different types.
//
//====================================================================================

#ifdef LAGRANGIAN
static const Integer kZLS_SuperCopyBoundary = 1;
#else
static const Integer kZLS_SuperCopyBoundary = 0;
#endif


int zlsbc_FillSuperLevelFromLevel( Integer level ){
  //
  int     err=noErr;
  Integer      zone;
  ZonePtr         z;
  //		 
  
  if  ((level < gNLevels)&&(level > 0)){ // level 0 has no super
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){
            //  err = zlsbc_FillSuperZoneFromZone( z , kZLS_SuperCopyBoundary );
            err = zlsbc_FillSuperZoneFromZone( z , 0 );
            if (err!= noErr) {return (err);}
          } // only process algorithm appropriate zones
#ifdef MPI	
        }
#endif MPI
      } // z
    } // zone
  } else { err = levErr;}//level
  return (err);
}

int zlsbc_ImposeLevelSuperBCs( Integer level, Integer iteration){
  //
  int     err=noErr;
  Integer      zone;
  ZonePtr         z;
  //		 
  if  ((level < gNLevels)&&(level > 0)){ 
    
    // then apply all computed BCs
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){ 
            err = zlsbc_ImposeZoneSuperBCs( z, iteration ); 
            if (err!= noErr) {return (err);}
          } // only process algorithm appropriate zones
#ifdef MPI	
        }
#endif MPI
      } else  { err = zonErr;}//level// z
    } // zone
    
  } else { err = levErr;}//level
  
  return (err);
}

int  zlsbc_ImposeLevelLinks ( Integer level){
  
  Integer err = noErr;
  Integer zone; 
  ZonePtr         z;
  
  if  ((level < gNLevels)&&(level >= 0)){ 
    
    // apply all the links in a given direction at a time
    Integer dirIndex;
		for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
      
      
      for ( zone = 0; zone < gLevels[level].nZones; zone++){
        z =  gLevels[level].levelZones[zone];
        if (z != NULL){
#ifdef MPI	
          if (z->zoneRankID == gMPIrank) {
#endif MPI
            if ( z->zoneType == kAlgorithmZoneType ){ 
              err = zlsbc_ImposeZoneLinks ( z, dirIndex );
              if (err!= noErr) {return (err);}
            } // only process algorithm appropriate zones
#ifdef MPI	
          }
#endif MPI
        } else  { err = zonErr;} //level// z
      } // zone
      
		} // dir
    
  } else { err = levErr;}//level
  
  return (err);
}

int zlsbc_ImposeLevelBCs( Integer level ){
  //
  int     err=noErr;
  Integer      zone;
  ZonePtr         z;
  //		 
  if  ((level < gNLevels)&&(level >= 0)){ 
    
    // then apply all computed BCs
    Integer dirIndex;
		for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
      
      for ( zone = 0; zone < gLevels[level].nZones; zone++){
        z =  gLevels[level].levelZones[zone];
        if (z != NULL){
#ifdef MPI	
          if (z->zoneRankID == gMPIrank) {
#endif MPI
            if ( z->zoneType == kAlgorithmZoneType ){ 
              err = zlsbc_ImposeZoneBCs( z , dirIndex ); // see zls_boundary.c
              if (err!= noErr) {return (err);}
            } // only process algorithm appropriate zones
#ifdef MPI	
          }
#endif MPI
        } else  { err = zonErr;}//level// z
      } // zone
      
    }
    
  } else { err = levErr;}//level
  
  return (err);
}

#ifdef ZLS_USER_BOUNDARIES
int zlsbc_ImposeLevelUserBCs( Integer level ){
  //
  int     err=noErr;
  Integer      zone;
  ZonePtr         z;
  //		 
  if  ((level < gNLevels)&&(level >= 0)){ 
    
    /// then apply all User BCs
    Integer dirIndex;
    for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
      
      for ( zone = 0; zone < gLevels[level].nZones; zone++){
        z =  gLevels[level].levelZones[zone];
        if (z != NULL){
#ifdef MPI	
          if (z->zoneRankID == gMPIrank) {
#endif MPI
            if ( z->zoneType == kAlgorithmZoneType ){ 
              err = zlsbc_ImposeZoneUserBCs( z, dirIndex ); 
              if (err != noErr){
                zlsio_ErrPrintf( " ERROR: zlsbc_ImposeLevelUserBCs from zlsbc_ImposeZoneUserBCs: Z %d  dir %d E %d \n",z->zoneRankID, dirIndex, err);
                return(err);
              }
            } // only process algorithm appropriate zones
#ifdef MPI	
          }
#endif MPI
        } else  { err = zonErr;}//level// z
      } // zone
      
    }
    
  } else { err = levErr;}//level
  return (err);
}
#endif ZLS_USER_BOUNDARIES

#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsbc_ImposeLevelUserUpdates( Integer level ){
  //
  int     err=noErr;
  Integer      zone;
  ZonePtr         z;
  //		 
  if  ((level < gNLevels)&&(level >= 0)){ 
    /// then apply all user special updates
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){ 
            err = zlsp_UserZoneUpdate( z ); // see zls_problem.c
            if (err!= noErr) {return (err);}
          } // only process algorithm appropriate zones
#ifdef MPI	
        }
#endif MPI
      } else  { err = zonErr;}//level// z
    } // zone
  } else { err = levErr;}//level
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE

int zlsbc_ImposeAllLevelBCs( Integer level, int iteration ){
  //
  int            err = noErr;
  //		 
#if (ZLS_DEBUG > 1)
  zlsa_PrintLevelLinks (level);
#endif (ZLS_DEBUG > 1)
  
  // first set BCs from super level if any
  if ( level > 0 ) err =  zlsbc_ImposeLevelSuperBCs ( level, iteration );
  if (err!= noErr) {return (err);}
  
  
#ifdef ZLS_USER_UPDATE_ROUTINE	
  //
  // set any userupdate that may affect *active area* of level, which should then flow on to 
  // any BC symmetries and links below.
  //
  // This routine should *only* set active area cells.
  // 
  if (level == (gNLevels-1)) {
    
    err = zlsbc_ImposeLevelUserUpdates( level );
    
    if (err!= noErr) {return (err);}
    
#ifdef STELLAR_SNe
// need to turn off and SNe so they dont repeat, but also need to make sure that
// all globals are set properly.
// even if mpi is running, all the zones on this process are done, so we can kill the
// star in this process.
// in pthreads, all zones are done by here, so we can do the same!

    Integer  starID = 0;
    
  ProblemPtr prob = gLevels[0].levelProblem;  // get the user's problem parameters
  for ( starID = 0; starID < prob->gNstars; starID++){
    if ( prob->gStarStatus[starID] == kStarSNe ) { // if it went SNe in this update...
      prob->gStarStatus[starID] = kStarDead ; // stop it supernovaing again!
    }
  }
  
#endif STELLAR_SNe
  }

#endif ZLS_USER_UPDATE_ROUTINE	
  //
  // then do all BCs, so level is treated with links and symmetry correctly.
  //
  // First links
#ifdef MPI
  err = zls_MPI_ImposeLevelLinks ( level );
  if (err!= noErr) {return (err);}
#else
  err = zlsbc_ImposeLevelLinks ( level );
  if (err!= noErr) {return (err);}
#endif !MPI
  
  // then normal BCs..
  
  err = zlsbc_ImposeLevelBCs ( level );
  if (err!= noErr) {return (err);}
  
#ifdef ZLS_USER_BOUNDARIES
  //      
  // these come last, and may *only* set BC cells, and can be used to override
  // the normal BC conditions, ie for jet inlets and so on.
  //  Anything that affects the active area must be done above in 
  // zlsbc_ImposeLevelUserUpdates or risk having inconsistent flow accross linked
  // sub zones.
  //
  err = zlsbc_ImposeLevelUserBCs ( level );
  if (err != noErr){
    zlsio_ErrPrintf( " ERROR: zlsbc_ImposeAllLevelBCs from zlsbc_ImposeLevelUserBCs: L %d E %d \n",level,  err);
    return(err);
  }
#endif ZLS_USER_BOUNDARIES
  
  return (err);
}


int  zlsbc_PrintLevelBCsAndLinks ( Integer level){
  
  Integer err = noErr;
  Integer zone; 
  ZonePtr         z;
  ZonePtr         zNext;
  ZonePtr         zSuper;
	Integer bndryType;  // what sort of boundary, super, link, calc
	Integer bndryKind;  // subtype
	Integer dirIndex;  
  
  if  ((level < gNLevels)&&(level >= 0)){ 
    
    // apply all the links
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){ 
            
            for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
              zlsio_Printf( "\n Zone Boundary: Zone: %2.2d:%2.2d Direction: %s\n", level, zone, zls_dirname(dirIndex));
              
              zlsio_Printf( "   BC: %d, 0x%x \n", z->zoneBCType[dirIndex] ,z->zoneBCType[dirIndex] );
              
              bndryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagSuperZ; 
              if (bndryType == kBoundaryTypeFlagSuperZ) { 
                zSuper = z->zoneSuper;
                zlsio_Printf( "   Super Bdry: Zone: 0x%x Zone id: %d, Direction %s : Super Zone 0x%x id: %d,\n\n",z, z->zoneID,  
                             zls_dirname(dirIndex), z->zoneSuper , z->zoneSuper->zoneID);
              } 
              
              bndryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagNextZ; 
              if ( bndryType == kBoundaryTypeFlagNextZ) {
                zNext =  z->zoneNext[dirIndex];
                zlsio_Printf( "   Link Bdry: Zone: %x Zone id: %d, next Zone id: %d, Direction %s \n",z, z->zoneID,  zNext->zoneID, zls_dirname(dirIndex) );
                zlsio_Printf( "    Link: 0x%x %d \n",  z->zoneNext[dirIndex], z->zoneNext[dirIndex]->zoneID );
                zlsio_Printf( "    nextBC: %d, %x, Link: 0x%x %d \n\n", zNext->zoneBCType[zls_opposite(dirIndex)], 
                             zNext->zoneBCType[zls_opposite(dirIndex)] , 
                             zNext->zoneNext[zls_opposite(dirIndex)],
                             zNext->zoneNext[zls_opposite(dirIndex)]->zoneID  );
              }
              
              bndryType  = z->zoneBCType[dirIndex]&kBoundaryTypeFlagComputed; 
              bndryKind  = z->zoneBCType[dirIndex]&kBCCKindMask; 
              
              if ( bndryType == kBoundaryTypeFlagComputed) {
                zlsio_Printf( "   Computed Bdry: Zone: %x Zone id: %d, Direction %s \n",z, z->zoneID,  zls_dirname(dirIndex) );
                zlsio_Printf( "     Kind: 0x%x %d \n\n", bndryKind, bndryKind);
              }
              
              bndryType  = z->zoneBCType[dirIndex]&kBoundaryTypeFlagUserBC; 
              bndryKind  = z->zoneBCType[dirIndex]&kBCCKindMask; 
              
              if ( bndryType == kBoundaryTypeFlagUserBC) {
                zlsio_Printf( " User Bdry: Zone: %x Zone id: %d, Direction %s \n",z, z->zoneID,  zls_dirname(dirIndex) );
                zlsio_Printf( " Kind: 0x%x %d \n\n", bndryKind, bndryKind);
              }
              
            }
            
            if (err!= noErr) {return (err);}
            
          } // only process algorithm appropriate zones
#ifdef MPI	
        }
#endif MPI
      } else  { err = zonErr;} //level// z
    } // zone
    
  } else { err = levErr;}//level
  
  return (err);
}


