#include "zls.h"

#define THETA_EPSILON (1.0e-10)

int    zls_SetRangeDims (RangePtr r, Integer ndims, Integer nbc){
  
  int err = noErr;
  Integer dim;
  
  if (r != NULL) {
    r->nD = ndims;
    r->nG = nbc;
    for (dim = 0; dim<kMaxCoordDims; dim++){ // init, Set Range Dims *must* be first call 
      r->type[dim] = kAxisTypeSpatial;
      r->kind[dim] = kAxisKindLinear;
      r->n0[dim] = 0;
      r->rn[dim] = 0;
      r->x0[dim] = 0.0;
      r->rx[dim] = 0.0;
    }
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_SetRangeType (RangePtr r, Integer dim, Integer type){
  
  int err = noErr;
  
  if (r != NULL) {
    r->type[dim] = type;
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_SetRangeKind (RangePtr r, Integer dim, Integer kind){
  
  int err = noErr;
  
  if (r != NULL) {
    r->kind[dim] = kind;
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_SetRange (RangePtr r, Integer dim, Integer n0, Integer n1, Real x0, Real x1){
  int err = noErr;
  
  if (r != NULL) {
    
    r->n0[dim] = n0;
    r->rn[dim] = n1 - n0;
    r->x0[dim] = x0;
    
    if (r->kind[dim] == kAxisKindLog) {
      r->rx[dim] = (x1/x0); // for log axes, a range is a ratio
    } else {
      r->rx[dim] = (x1 - x0);  // for lin axes, a range is a difference
    }
    
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_SetRangeDelta(RangePtr r, Integer dim, Integer n0, Integer n, Real x0, Real x){
  
  int err = noErr;
  if (r != NULL) {
    r->n0[dim] = n0;
    r->rn[dim] = n;
    r->x0[dim] = x0;
    if (r->kind[dim] == kAxisKindLog) {
      r->rx[dim] = ((x0 + x)/x0);
    } else {
      r->rx[dim] = x;
    }
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_ScaleRangeCells(RangePtr r, Integer dim,  Integer n){
  
  int err = noErr;
  if (r != NULL) {
    r->n0[dim] *= n;
    r->rn[dim] *= n;
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_PrintRange     ( FILE *f, RangePtr r){ 
  int err = noErr;
  int i;
  
  if (r != NULL) {
    
    if ((r->nD<1)||(r->nD>3)) {
      zlsio_ErrPrintf( " Range Dimension Error: %d \n", r->nD);
      err = dimErr ;
    } else {
      
      //  zlsio_FPrintf( f, "      Range nD: %d \n", r->nD);
      //  zlsio_FPrintf( f, "      Range nG: %d \n", r->nG);
      for (i = 0; i<r->nD; i++){
        
        if ( i == 0 )           zlsio_FPrintf( f, " First  Axis Range:\n" );
        if ( i == 1 )           zlsio_FPrintf( f, " Second Axis Range:\n" );
        if ( i == 2 )           zlsio_FPrintf( f, " Third  Axis Range:\n" );
        
        zlsio_FPrintf( f, "      Index       n0[%d]: %5d \n", i, r->n0[i]);
        zlsio_FPrintf( f, "      Range  Delta n[%d]: %5d \n", i, r->rn[i]);
#ifdef CARTESIAN    
        zlsio_FPrintf( f, "      Spatial     x0[%d]: %#5g \n", i, r->x0[i]);
        zlsio_FPrintf( f, "      Range  Delta x[%d]: %#5g \n", i, r->rx[i]);
#endif CARTESIAN    
#ifdef SPHERICAL    
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_FPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_FPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_FPrintf( f, "      Phi    Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeTheta) {
          zlsio_FPrintf( f, "      Theta Angle t0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_FPrintf( f, "      Theta  Delta t[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif SPHERICAL
#ifdef POLARSPHERICAL   
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_FPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePolarTheta) {
          zlsio_FPrintf( f, "      Polar Theta  t0[%d]: %#5g (%#5g) \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      P. Theta Del. t[%d]: %#5g (%#5g) \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_FPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_FPrintf( f, "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif POLARSPHERICAL
#ifdef CYLINDRICAL    
        if ( r->type[i] == kAxisTypeSpatial) {
          zlsio_FPrintf( f, "      Axial dist. z0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      Axial  Delta z[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_FPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_FPrintf( f, "      Phi Angle   p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_FPrintf( f, "      Range  Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CYLINDRICAL
#ifdef CIRCULAR   
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_FPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_FPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_FPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_FPrintf( f, "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CIRCULAR
      }
      
    }
    
  } else { err = memErr ; }
  
  return(err);                 
}

int    zls_ZPrintRange     ( zFile f, RangePtr r){ 
  int err = noErr;
  int i;
  
  if (r != NULL) {
    
    if ((r->nD<1)||(r->nD>3)) {
      zlsio_ErrPrintf( " Range Dimension Error: %d \n", r->nD);
      err = dimErr ;
    } else {
      
      //  zlsio_FPrintf( f, "      Range nD: %d \n", r->nD);
      //  zlsio_FPrintf( f, "      Range nG: %d \n", r->nG);
      for (i = 0; i<r->nD; i++){
        
        if ( i == 0 )           zlsio_ZFPrintf( f, " First  Axis Range:\n" );
        if ( i == 1 )           zlsio_ZFPrintf( f, " Second Axis Range:\n" );
        if ( i == 2 )           zlsio_ZFPrintf( f, " Third  Axis Range:\n" );
        
        zlsio_ZFPrintf( f, "      Index       n0[%d]: %5d \n", i, r->n0[i]);
        zlsio_ZFPrintf( f, "      Range  Delta n[%d]: %5d \n", i, r->rn[i]);
#ifdef CARTESIAN    
        zlsio_ZFPrintf( f, "      Spatial     x0[%d]: %#5g \n", i, r->x0[i]);
        zlsio_ZFPrintf( f, "      Range  Delta x[%d]: %#5g \n", i, r->rx[i]);
#endif CARTESIAN    
#ifdef SPHERICAL    
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_ZFPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_ZFPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_ZFPrintf( f, "      Phi    Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeTheta) {
          zlsio_ZFPrintf( f, "      Theta Angle t0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_ZFPrintf( f, "      Theta  Delta t[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif SPHERICAL
#ifdef POLARSPHERICAL   
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_ZFPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePolarTheta) {
          zlsio_ZFPrintf( f, "      Polar Theta  t0[%d]: %#5g (%#5g) \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      P. Theta Del. t[%d]: %#5g (%#5g) \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_ZFPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_ZFPrintf( f, "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif POLARSPHERICAL
#ifdef CYLINDRICAL    
        if ( r->type[i] == kAxisTypeSpatial) {
          zlsio_ZFPrintf( f, "      Axial dist. z0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      Axial  Delta z[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_ZFPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_ZFPrintf( f, "      Phi Angle   p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_ZFPrintf( f, "      Range  Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CYLINDRICAL
#ifdef CIRCULAR   
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_ZFPrintf( f, "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_ZFPrintf( f, "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_ZFPrintf( f, "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_ZFPrintf( f, "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CIRCULAR
      }
      
    }
    
  } else { err = memErr ; }
  
  return(err);                 
}


int    zls_UserLogPrintRange     ( RangePtr r){ 
  int err = noErr;
  int i;
  
  if (r != NULL) {
    
    if ((r->nD<1)||(r->nD>3)) {
      zlsio_ErrPrintf( " Range Dimension Error: %d \n", r->nD);
      err = dimErr ;
    } else {
      
      //  zlsio_UserLogPrintf(  "      Range nD: %d \n", r->nD);
      //  zlsio_UserLogPrintf(  "      Range nG: %d \n", r->nG);
      for (i = 0; i<r->nD; i++){
        
        if ( i == 0 )           zlsio_UserLogPrintf( " First  Axis Range:\n" );
        if ( i == 1 )           zlsio_UserLogPrintf( " Second Axis Range:\n" );
        if ( i == 2 )           zlsio_UserLogPrintf( " Third  Axis Range:\n" );
        
        zlsio_UserLogPrintf( "      Index       n0[%d]: %5d \n", i, r->n0[i]);
        zlsio_UserLogPrintf( "      Range  Delta n[%d]: %5d \n", i, r->rn[i]);
#ifdef CARTESIAN    
        zlsio_UserLogPrintf(  "      Spatial     x0[%d]: %#5g \n", i, r->x0[i]);
        zlsio_UserLogPrintf(  "      Range  Delta x[%d]: %#5g \n", i, r->rx[i]);
#endif CARTESIAN    
#ifdef SPHERICAL    
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_UserLogPrintf(  "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_UserLogPrintf(  "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_UserLogPrintf(  "      Phi    Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeTheta) {
          zlsio_UserLogPrintf(  "      Theta Angle t0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_UserLogPrintf(  "      Theta  Delta t[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif SPHERICAL
#ifdef POLARSPHERICAL   
        if ( r->type[i] == kAxisTypeSRadius) {
          zlsio_UserLogPrintf(  "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePolarTheta) {
          zlsio_UserLogPrintf(  "      Polar Theta  t0[%d]: %#5g (%#5g) \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      P. Theta Del. t[%d]: %#5g (%#5g) \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_UserLogPrintf(  "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_UserLogPrintf(  "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif POLARSPHERICAL
#ifdef CYLINDRICAL    
        if ( r->type[i] == kAxisTypeSpatial) {
          zlsio_UserLogPrintf(  "      Axial dist. z0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      Axial  Delta z[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_UserLogPrintf(  "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_UserLogPrintf(  "      Phi Angle   p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_UserLogPrintf(  "      Range  Delta p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CYLINDRICAL
#ifdef CIRCULAR   
        
        if ( r->type[i] == kAxisTypeCRadius) {
          zlsio_UserLogPrintf(  "      Radius      r0[%d]: %#5g \n", i, r->x0[i]);
          zlsio_UserLogPrintf(  "      Radius Delta r[%d]: %#5g \n", i, r->rx[i]);
        }
        
        if ( r->type[i] == kAxisTypePhi) {
          zlsio_UserLogPrintf(  "      Phi   Angle p0[%d]: %#5g (%#5g) \n", i, invpi*r->x0[i], r->x0[i]);
          zlsio_UserLogPrintf(  "      Phi   Delta  p[%d]: %#5g (%#5g) \n", i, invpi*r->rx[i], r->rx[i]);
        }
        
#endif CIRCULAR
      }
      
    }
    
  } else { err = memErr ; }
  
  return(err);                 
}



//
// Axis array routines;
//


int    zls_NewAxis  (ACoordPtr *x,  Integer nbc, Integer ni){  
  // usually dynamically allocated, but could be fixed array too.
  // grid coords use dynamics axes.
  int err = noErr;
  ACoordPtr tempX= NULL;
  
  tempX = (ACoordPtr) malloc((ni+nbc+nbc)*sizeof(AxisCoords));
  if (tempX == NULL) {
    err = memErr;
  }
  
  *x = tempX;
  
  return(err);
}

int    zls_DisposeAxis (ACoordPtr x){ // only if dynamically allocated
  
  int err = noErr;
  
  if (x != NULL) free(x);
  
  return(err);                 
}


int    zls_CopyAxis  (ACoordPtr dst,  ACoordPtr src, Integer nbc, Integer ni){  
  
  int err = noErr;
  
  Size nCells = (ni+nbc+nbc);
  if (( dst != NULL) && ( src != NULL )) 
    memcpy(dst, src, nCells*sizeof(AxisCoords));
  
  return(err);
  
}


// calculating coeffs

int    zls_InitAxis    (ACoordPtr x, Integer type, Integer kind, Integer nbc, Integer ni, Real ix, Real ox){ // Resulting Axes include coords of bCells
  
  int err = noErr;
  
  Integer i, n;
  Real    x0;
  Real    dex;
  
  Real del, f;
  
  n   = ni + nbc + nbc;
#ifdef CARTESIAN  
  if (type != kAxisTypeSpatial){
    zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Type %d, using Spatial \n", type);
    type = kAxisTypeSpatial;
  }
#endif CARTESIAN  
  
#ifdef SPHERICAL  
  switch (type){
    default:
      zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Type %d, using Radius \n",type);
    case kAxisTypeSRadius: // == 1st coord, spherical dirvergent radial coord
      type = kAxisTypeSRadius;
      break;
    case kAxisTypePhi: // == kAxisTypeAzimuth, 2nd coord, in x-y plane
    case kAxisTypeTheta: // == kAxisTypeAltitude, 3rd coord, positive 0 -> pi/2 -> z-axis, to preserve rh coordinates
      kind = kAxisKindLinear; // enforce linear angular scales.
  }
#endif SPHERICAL  
  
#ifdef POLARSPHERICAL 
  switch (type){
    default:
      zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Type %d, using Radius \n",type);
    case kAxisTypeSRadius: // == 1st coord, spherical dirvergent radial coord
      type = kAxisTypeSRadius;
      break;
    case kAxisTypePhi: // == kAxisTypeAzimuth, 3rd coord, in x-y plane
    case kAxisTypePolarTheta: // == kAxisTypeAltitude, 2nd coord, positive 0 -> pi from z-axis down
      kind = kAxisKindLinear; // enforce linear angular scales.
  }
#endif POLARSPHERICAL 
  
#ifdef CYLINDRICAL  
  switch (type){
    default:
      zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Type %d, using Spatial \n",type);
    case kAxisTypeSpatial: // == 1st coord, cylindrical z-spatial coord along x direction
      type = kAxisTypeSpatial;
      break;
    case kAxisTypeCRadius: // == 2nd coord, cylindrical dirvergent radial coord
      type = kAxisTypeCRadius;
      break;
    case kAxisTypePhi: // == kAxisTypeAzimuth, 3nd coord, in x-y plane
      kind = kAxisKindLinear; // enforce linear angular scales.
  }
#endif CYLINDRICAL  
  
#ifdef CIRCULAR 
  switch (type){
    default:
      zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Type %d, using Radial \n",type);
    case kAxisTypeCRadius: // == 2nd coord, cylindrical dirvergent radial coord
      type = kAxisTypeCRadius;
      break;
    case kAxisTypePhi: // == kAxisTypeAzimuth, 3nd coord, in x-y plane
      kind = kAxisKindLinear; // enforce linear angular scales.
  }
#endif CIRCULAR 
  
  
  switch (kind){
    default:
      zlsio_ErrPrintf( " Warning: zls_InitAxis Unknown Axis Kind %d, using linear \n", kind);
    case kAxisKindLinear:
      
      del = (ox - ix);
      for(i = 0;i<=n;i++){
        f  = ix+(((double)(i-nbc))/((double)(ni)))*del;
        if ( i <  n) x[i  ].l  = f;
        if ( i >  0) x[i-1].r  = f;
      }
      
      for(i = 0;i<n;i++){
        //        dex = (ox - ix)/ni;
        //        x0  = ix - (dex*nbc);
        //        x[i].l  = x0;
        //        x0     += dex;
        //        x[i].r  = x0;
#ifdef PLM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PLM        
#ifdef PPM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PPM        
      }; //i
      
      break;      
      case kAxisKindLog:
      
      dex = (log(ox) - log(ix))/ni;
      x0  = log(ix) - (dex*nbc);
      //      x0  = zls_tolx(x0,kZLS_Small);
      
      for(i = 0;i<n;i++){
        x[i].l     = exp(x0);
        x0        += dex;
        x[i].r     = exp(x0);
#ifdef PLM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PLM        
#ifdef PPM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PPM        
      }; //i
      
      break;
      
  }; //switch
  
  
  zls_CalcAxisDeltas       ( x, type, ni);
  
  return(err); 
  
}

int    zls_SetAxisRange    (ACoordPtr x, Integer dim, RangePtr r){ // Resulting Axes include coords of bCells
  
  int err = noErr;
  
  Integer i, n, ni, nbc;
  Real           x0;
  Real          dex;
  Real          ix, ox;
  Real          f, del;
  
  nbc = r->nG;
  ni  = r->rn[dim];
  n   = r->rn[dim] + nbc + nbc;
  
  ix = r->x0[dim];
  ox = r->x0[dim]+r->rx[dim];
  
  switch (r->kind[dim]){
    default:
      zlsio_ErrPrintf( " Warning: zls_SetAxisRange Unknown Axis Kind %d, using linear \n", r->kind[dim]);
    case kAxisKindLinear:
      r->kind[dim] = kAxisKindLinear;
      
      
      del = (ox - ix);
      for(i = 0;i<=n;i++){
        f  = ix+(((double)(i-nbc))/((double)(ni)))*del;
        if ( i <  n) x[i  ].l  = f;
        if ( i >  0) x[i-1].r  = f;
      }
      
      
      //      dex = r->rx[dim]/r->rn[dim];
      //      x0  = r->x0[dim] - (dex*nbc);
      
      for(i = 0;i<n;i++){
        //        x[i].l  = x0;
        //        x0     += dex;
        //        x[i].r  = x0;
#ifdef PLM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PLM        
#ifdef PPM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PPM        
      }; //i
      
      
      
      
      break;
      
      case kAxisKindLog:
      
      dex = log(r->rx[dim])/r->rn[dim];
      x0  = log(r->x0[dim]) - (dex*nbc);
      //      x0  = zls_tolx(x0,kZLS_Small);
      
      for(i = 0;i<n;i++){
        x[i].l     = exp(x0);
        x0        += dex;
        x[i].r     = exp(x0);
#ifdef PLM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PLM        
#ifdef PPM        
        x[i].pc0 = 0.0;
        x[i].pc1 = 0.0;
        x[i].pc2 = 0.0;
        x[i].pc3 = 0.0;
        x[i].pc4 = 0.0;
#endif PPM        
      } //i
      
  }; // switch
  
  
  zls_CalcAxisDeltas ( x, r->type[dim], n);
  
  return(err);
  
}

int    zls_CopyAxisRange ( ACoordPtr dst, ACoordPtr src, Integer dim, RangePtr r ){ 
  
  int err = noErr;
  
  Integer nbc = r->nG;
  Integer n0  = r->n0[dim] + nbc;
  Integer n   = r->rn[dim] + nbc + nbc;
  
  Integer i;
  for ( i = 0; i < n; i++ ) {
    dst[i].l     = src[i+n0-nbc].l; 
    dst[i].r     = src[i+n0-nbc].r;  
#ifdef PLM        
    dst[i].pc0 = 0.0;
    dst[i].pc1 = 0.0;
    dst[i].pc2 = 0.0;
    dst[i].pc3 = 0.0;
    dst[i].pc4 = 0.0;
#endif PLM        
#ifdef PPM        
    dst[i].pc0 = 0.0;
    dst[i].pc1 = 0.0;
    dst[i].pc2 = 0.0;
    dst[i].pc3 = 0.0;
    dst[i].pc4 = 0.0;
#endif PPM        
  }
  
  zls_CalcAxisDeltas ( dst, r->type[dim], n);
  
  return(err);
  
}

int    zls_RescaleAxis          (ACoordPtr x, Integer type, Integer ni, Real s){ // 
  int err = noErr;
  Integer i;
  for(i = 0;i<ni;i++){
    x[i].l  *= s; 
    x[i].r  *= s; 
  } //i
  zls_CalcAxisDeltas ( x, type, ni);
  return(err);
}

int    zls_CalcAxisDeltas       (ACoordPtr x, Integer type, Integer ni){
  // compute delta x, area, volume and sine/cos tables for an axis, different treatments for 
  // different geometries, particularly dA and dVol
  int err = noErr;
  Integer i;
  
  
#ifdef CARTESIAN
  Real   dx;
  for(i = 0;i<ni;i++){
    x[i].c    = 0.5*(x[i].r + x[i].l); 
    dx        = x[i].r - x[i].l ;
    x[i].dx   = dx ;
    x[i].idx  = 1.0/dx ;
    x[i].dA   = 1.0;
    x[i].dVol = dx;
  }; //i
#endif CARTESIAN
  
#ifdef SPHERICAL
  Real r, dr, dphi, dtheta;
  switch (type) {
    case kAxisTypeSRadius:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dr        = x[i].r - x[i].l ;
        x[i].dx   = dr ;
        x[i].idx  = 1.0/dr ;
        r         = x[i].l ;
        x[i].dA   = (r*(r+dr)+dr*dr*third);
        x[i].dVol = dr* x[i].dA;
      }; //i
      break;
      case kAxisTypePhi:
      for(i = 0;i<ni;i++){
        
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dphi       = x[i].r - x[i].l;
        x[i].dx   = dphi ;
        x[i].idx  = 1.0/dphi ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        
        x[i].dA   = 1.0;
        x[i].dVol = dphi* x[i].dA;
        
      }; //i
      break;
      case kAxisTypeTheta:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dtheta    = x[i].r - x[i].l;
        x[i].dx   = dtheta ;
        x[i].idx  = 1.0/dtheta ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        
        x[i].dVol = ( x[i].sr - x[i].sl );
        x[i].dA = zls_abs(dtheta) > THETA_EPSILON ? x[i].dVol/dtheta:x[i].cl;
      }; //i
      break;
      default:
      zlsio_ErrPrintf(  " ERROR: zls_CalcAxisDeltas Bad axis type %d \n",type);
      err = dimErr;
      return(err);
  }; // switch type
#endif SPHERICAL
  
  
#ifdef POLARSPHERICAL
  Real r, dr, dphi, dtheta;
  switch (type) {
    case kAxisTypeSRadius:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dr        = x[i].r - x[i].l ;
        x[i].dx   = dr ;
        x[i].idx  = 1.0/dr ;
        r         = x[i].l ;
        x[i].dA   = (r*(r+dr)+dr*dr*third);
        x[i].dVol = dr* x[i].dA;
      }; //i
      break;
      case kAxisTypePhi:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dphi      = x[i].r - x[i].l ;  // radians
        x[i].dx   = dphi ;
        x[i].idx  = 1.0/dphi ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        
        x[i].dA   = 1.0;
        x[i].dVol = dphi* x[i].dA;
      }; //i
      break;
      case kAxisTypePolarTheta:
      for(i = 0;i<ni;i++){
        x[i].c  = 0.5*(x[i].r + x[i].l); 
        dtheta  = x[i].r - x[i].l ;  // radians
        x[i].dx = dtheta ;
        x[i].idx  = 1.0/dtheta ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        
        x[i].dVol = ( x[i].cl - x[i].cr );
        x[i].dA = zls_abs(dtheta) > THETA_EPSILON ?x[i].dVol/dtheta:x[i].sl;
      }; //i
      break;
      default:
      zlsio_ErrPrintf(  " ERROR: zls_CalcAxisDeltas Bad axis type %d \n",type);
      err = dimErr;
      return(err);
  }; // switch type
#endif POLARSPHERICAL
  
#ifdef CYLINDRICAL
  Real r, dr, dz, dphi;
  switch (type) {
    case kAxisTypeSpatial:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dz        = x[i].r - x[i].l ;
        x[i].dx   = dz ;
        x[i].idx  = 1.0/dz ;
        x[i].dA   = 1.0;
        x[i].dVol = dz* x[i].dA;
      }; //i
      break;
      case kAxisTypeCRadius:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dr        = x[i].r - x[i].l ;
        x[i].dx   = dr ;
        x[i].idx  = 1.0/dr ;
        r         = x[i].l ;
        x[i].dA   = (r+(dr*0.5));
        x[i].dVol = dr* x[i].dA;
      }; //i
      break;
      case kAxisTypePhi:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dphi      = x[i].r - x[i].l ;  // radians
        x[i].dx   = dphi ;
        x[i].idx  = 1.0/dphi ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        /*        
         x[i].sl = sin(x[i].l); // sin of left coord
         x[i].sr = sin(x[i].r); // sin of right coord
         x[i].sc = sin(x[i].c); // sin of centre coord
         
         x[i].cl = cos(x[i].l); // cos of left coord
         x[i].cr = cos(x[i].r); // cos of right coord
         x[i].cc = cos(x[i].c); // cos of centre coord
         */        
        x[i].dA   = 1.0;
        x[i].dVol = dphi* x[i].dA;
      }; //i
      
      break;
      default:
      zlsio_ErrPrintf(  " ERROR: zls_CalcAxisDeltas Bad axis type %d \n",type);
      err = dimErr;
      return(err);
  }; // switch type
#endif CYLINDRICAL
  
  
#ifdef CIRCULAR
  Real r, dr, dphi;
  switch (type) {
    case kAxisTypeCRadius:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dr        = x[i].r - x[i].l ;
        x[i].dx   = dr ;
        x[i].idx  = 1.0/dr ;
        r         = x[i].l ;
        x[i].dA   = (r+(dr*0.5));
        x[i].dVol = dr* x[i].dA;
      }; //i
      break;
      case kAxisTypePhi:
      for(i = 0;i<ni;i++){
        x[i].c    = 0.5*(x[i].r + x[i].l); 
        dphi      = x[i].r - x[i].l ;  // radians
        x[i].dx   = dphi ;
        x[i].idx  = 1.0/dphi ;
        
        zls_sincos(x[i].l, &x[i].sl, &x[i].cl); // 11 decimal custom routine
        zls_sincos(x[i].c, &x[i].sc, &x[i].cc);
        zls_sincos(x[i].r, &x[i].sr, &x[i].cr);
        /*        
         x[i].sl = sin(x[i].l); // sin of left coord
         x[i].sr = sin(x[i].r); // sin of right coord
         x[i].sc = sin(x[i].c); // sin of centre coord
         
         x[i].cl = cos(x[i].l); // cos of left coord
         x[i].cr = cos(x[i].r); // cos of right coord
         x[i].cc = cos(x[i].c); // cos of centre coord
         */        
        x[i].dA   = 1.0;
        x[i].dVol = dphi* x[i].dA;
      }; //i
      break;
      default:
      zlsio_ErrPrintf(  " ERROR: zls_CalcAxisDeltas Bad axis type %d \n",type);
      err = dimErr;
      return(err);
  }; // switch type
#endif CIRCULAR
  
#ifndef GODUNOV
  zls_CalcAxisParaCoeffs(x, ni);
#endif GODUNOV
  return(err);
}

int    zls_CalcAxisGeometrySource   ( ACoordPtr x, ACoordPtr x0, Integer type, Integer ni){ 
  int err = noErr;
#ifdef CARTESIAN  
  // nothing to do for straight cartesian axes, dA = 1.0 in all cases
#endif CARTESIAN    
#ifdef SPHERICAL    
  Integer    i;
  Real   r, dr, dtheta;
  switch (type) {
    case kAxisTypeSRadius:
      for(i = 0;i<ni;i++){
        r       =   x0[i].l;
        dr      = ( x[i].l - r );
        x[i].dA = ( r*(r+dr)+dr*dr*third );
      }; //i
      break;
      case kAxisTypeTheta:
      for(i = 0;i<ni;i++){ // area changes as theta approaches the pole as d dt cos t
        dtheta  =   x[i].l - x0[i].l ;
        x[i].dA = zls_abs(dtheta) > THETA_EPSILON ?( x[i].sl - x0[i].sl )/dtheta:x0[i].cl;
      }; //i
      break;
      default:
      ; // nothing dA = 1.0 for phi, and that is already set
  } // switch
#endif SPHERICAL    
#ifdef POLARSPHERICAL    
  Integer    i;
  Real   r, dr, dtheta;
  switch (type) {
    case kAxisTypeSRadius:
      for(i = 0;i<ni;i++){
        r       =  x0[i].l;
        dr      = ( x[i].l - r );
        x[i].dA = ( r*(r+dr)+dr*dr*third );
      }; //i
      break;
      case kAxisTypePolarTheta:
      for(i = 0;i<ni;i++){
        dtheta  =   x[i].l - x0[i].l ;
        x[i].dA = zls_abs(dtheta) > THETA_EPSILON ?( x[i].cl - x0[i].cl )/dtheta:x0[i].sl;
      }; //i
      break;
      default:
      ; // nothing dA = 1.0 for phi, and that is already set
  } // switch
#endif POLARSPHERICAL  
#ifdef CIRCULAR    
  Integer    i;
  Real   r, dr;
  switch (type) {
    case kAxisTypeCRadius:
      for(i = 0;i<ni;i++){
        r       =   x0[i].l;
        dr      = ( x[i].l - r );
        x[i].dA = ( r+(dr*0.5));
      }; //i
      break;
      default:
      ; // nothing dA = 1.0 for phi, and that is already set
  } // switch
#endif CIRCULAR    
#ifdef CYLINDRICAL    
  Integer    i;
  Real   r, dr;
  switch (type) {
    case kAxisTypeCRadius:
      for(i = 0;i<ni;i++){
        r       =   x0[i].l;
        dr      = ( x[i].l - r );
        x[i].dA = ( r+(dr*0.5));
      }; //i
      break;
      default:
      ; // nothing dA = 1.0 for z and phi, and that is already set
  } // switch
#endif CYLINDRICAL  
  
  return(err);
}

//int zls_CalcAxisParaCoeffs(ACoordPtr x, Integer ni){
//  int err = noErr;
//  
//  Real a[ni];  Real ai[ni];
//
//  Real b[ni];  Real bi[ni];
//
//  Real c[ni];  Real ci[ni];
//  Real d[ni];
//  
//  Integer n;
//    
//  x[0  ].pc0 = 0.5;
//  x[0  ].pc1 = +0.16666666666666666;
//  x[0  ].pc2 = -0.16666666666666666;
//  x[0  ].pc3 = 0.5;
//  x[0  ].pc4 = 0.5;
//  
//  x[ni-1].pc0 = 0.5;
//  x[ni-1].pc1 = +0.16666666666666666;
//  x[ni-1].pc2 = -0.16666666666666666;
//  x[ni-1].pc3 = 0.5;
//  x[ni-1].pc4 = 0.5;
//  
//  x[ni-2].pc0 = 0.5;
//  x[ni-2].pc1 = +0.16666666666666666;
//  x[ni-2].pc2 = -0.16666666666666666;
//  x[ni-2].pc3 = 0.5;
//  x[ni-2].pc4 = 0.5;
//  
//  for( n = 0 ; n < (ni-1); n++){
//         a [n]  = x[n].dx + x[n+1].dx;
//         ai[n] = 1.0/(a[n]+kZLS_Small);
//         b [n]  = a[n] + x[n].dx;
//         bi[n] = 1.0/(b[n]+kZLS_Small);
//         c [n]  = a[n] + x[n+1].dx;
//         ci[n] = 1.0/(c[n]+kZLS_Small);
//  }
//
////C
////C     constants for equation 1.6
////C     a(j+.5)  = a(j) + C1 * (a(j+1)-a(j))  + C2 * da(j) + C3 * da(j+1)
////C
//
//  for( n = 1 ; n < (ni-2); n++){
//         d[n]      = 1.0 / (a[n-1] + a[n+1] + kZLS_Small);
//         x[n].pc0  = x[n].dx * ai[n] + 2.0 * x[n+1].dx * x[n].dx *
//                     d[n] * ai[n] * ( a[n-1] * bi[n] - a[n+1] * ci[n] );
//         x[n].pc2  = - d[n] *  x[n].dx  * a[n-1] * bi[n];
//         x[n].pc1  =   d[n] *  x[n+1].dx * a[n+1] * ci[n];
//  }
//
////C
////C     constants for equation 1.7
////C     da(j) = D1 * (a(j+1) - a(j)) + D2 * (a(j) - a(j-1))
////C
//
//  for( n = 1 ; n < (ni-1); n++){
//         d[n]      = x[n].dx / ( a[n-1] + x[n+1].dx + kZLS_Small );
//         x[n].pc3  = d[n] * b[n-1] * ai[n  ];
//         x[n].pc4  = d[n] * c[n  ] * ai[n-1];
//  }
//
//  return(err);
//}

int zls_CalcAxisParaCoeffs(ACoordPtr x, Integer ni){
  int err = noErr;
  
  register Integer i, l;
#ifndef GODUNOV  
  
  // these are the coefficients for recontructing the lft and right
  // cell values if they are not known, by interpolating with neighbours
  // on the grid.  since the Godunov method has a single value for the cell
  // it is not needed for that case at all.
  
#ifdef PLM    
  
  //  standard CW grid interpolation coeffs
  
  Real dxm, dx0, dxp;
  Real ddxm, ddxp, ddx2;
  Real iddxm, iddxp; // inverses
  Real i0 ; //  1.0/(dx0+ddxp);
  Real i1 ; //1.0/(ddxp+dxp);
  Real i2 ; // 1.0/(ddxm+ddx2);
  Real i3  ; //  dx0/dx;
  Real dx, dx2;
  
  x[0].pc0 = 0.5;
  //  x[0].pc1 = -0.16666666666666666;
  //  x[0].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[0].pc1 = +0.16666666666666666;
  x[0].pc2 = -0.16666666666666666;
  x[0].pc3 = 0.5;
  x[0].pc4 = 0.5;
  
  x[ni-1].pc0 = 0.5;
  //  x[ni-1].pc1 = -0.16666666666666666;
  //  x[ni-1].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[ni-1].pc1 = +0.16666666666666666;
  x[ni-1].pc2 = -0.16666666666666666;
  x[ni-1].pc3 = 0.5;
  x[ni-1].pc4 = 0.5;
  
  x[ni-2].pc0 = 0.5;
  //  x[ni-2].pc1 = -0.16666666666666666;
  //  x[ni-2].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[ni-2].pc1 = +0.16666666666666666;
  x[ni-2].pc2 = -0.16666666666666666;
  x[ni-2].pc3 = 0.5;
  x[ni-2].pc4 = 0.5;
  
  dxm   =  x[0].dx; // intial deltas for i = 1, calcs work up to i < n-2
  dx0   =  x[1].dx;
  dxp   =  x[2].dx;
  dx2   =  x[3].dx;
  
  dx    =  dxp + dx0 + dxm;
  ddxm  =  dx0 + dxm;
  ddxp  =  dxp + dx0;
  ddx2  =  dx2 + dxp;
  
  iddxm =  1.0/ddxm;
  iddxp =  1.0/ddxp;
  
  i0    = 1.0/(dx0+ddxp);
  i1    = 1.0/(ddxp+dxp);
  i2    = 1.0/(ddxm+ddx2);
  i3    = dx0/dx;
  
  // optimised loop, minimize memory accesses and divisions
  x++ ; 
  // shift to x[1];
  l  = ni-3;
  
  for (i = 1; i<l; i++){
    // eqn 1.6    
    x->pc0  =  dx0*iddxp*(1.0+i2*(dxp+dxp)*((ddxm*i0) - (ddx2*i1)));
    //    x->pc1  = -i2*dx0*ddxm*i0;
    //    x->pc2  =  i2*dxp*ddx2*i1;
    //
    //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
    //  delta and 2 applies to the j+1 delta.
    //
    x->pc1  =  i2*dxp*ddx2*i1;
    x->pc2  = -i2*dx0*ddxm*i0;
    // eqn1.7    
    x->pc3  = i3*(ddxm+dxm)*iddxp;
    x->pc4  = i3*(ddxp+dxp)*iddxm;
    
    // shuffle along and do the minimum 5 necessary divisions
    x++;
    
    dx   -= (dxm-dx2);
    dxm   = dx0;
    dx0   = dxp;
    dxp   = dx2;
    
    ddxm  = ddxp;
    ddxp  = ddx2;
    
    dx2   = x[2].dx;
    ddx2 -= (dx0-dx2);
    
    iddxm = iddxp;
    iddxp = 1.0/ddxp;
    
    i0   = 1.0/(dx0+ddxp);
    i1   = 1.0/(ddxp+dxp);
    i2   = 1.0/(ddxm+ddx2);
    i3   = dx0/dx;
    
  }
  
#endif PLM  
  
#ifdef PPM  
  
  //  standard CW PPM coeffs, coeff 1 and 2 swapped so the second is -ve
  
  Real dxm, dx0, dxp;
  Real ddxm, ddxp, ddx2;
  Real iddxm, iddxp; // inverses
  Real i0 ; //  1.0/(dx0+ddxp);
  Real i1 ; //1.0/(ddxp+dxp);
  Real i2 ; // 1.0/(ddxm+ddx2);
  Real i3  ; //  dx0/dx;
  Real dx, dx2;
  
  x[0].pc0 = 0.5;
  //  x[0].pc1 = -0.16666666666666666;
  //  x[0].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[0].pc1 = +0.16666666666666666;
  x[0].pc2 = -0.16666666666666666;
  x[0].pc3 = 0.5;
  x[0].pc4 = 0.5;
  
  x[ni-1].pc0 = 0.5;
  //  x[ni-1].pc1 = -0.16666666666666666;
  //  x[ni-1].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[ni-1].pc1 = +0.16666666666666666;
  x[ni-1].pc2 = -0.16666666666666666;
  x[ni-1].pc3 = 0.5;
  x[ni-1].pc4 = 0.5;
  
  x[ni-2].pc0 = 0.5;
  //  x[ni-2].pc1 = -0.16666666666666666;
  //  x[ni-2].pc2 = +0.16666666666666666;
  //
  //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
  //  delta and 2 applies to the j+1 delta.
  //
  x[ni-2].pc1 = +0.16666666666666666;
  x[ni-2].pc2 = -0.16666666666666666;
  x[ni-2].pc3 = 0.5;
  x[ni-2].pc4 = 0.5;
  
  dxm =  x[0].dx; // intial deltas for i = 1, calcs work up to i < n-2
  dx0 =  x[1].dx;
  dxp =  x[2].dx;
  dx2 =  x[3].dx;
  
  dx  =  dxp + dx0 + dxm; // -1, 0, +1
  ddxm  =  dx0 + dxm; // -1, 0
  ddxp  =  dxp + dx0; // 0, +1
  ddx2  =  dx2 + dxp; // +1,+2
  
  iddxm =  1.0/ddxm;
  iddxp =  1.0/ddxp;
  
  i0   = 1.0/(dx0+ddxp); // 0, 0, +1
  i1   = 1.0/(ddxp+dxp); // 0, +1, +1
  i2   = 1.0/(ddxm+ddx2); // -1, 0, +1, +2
  i3   = dx0/dx;
  
  // optimised loop, minimize memory accesses and divisions
  x++ ; 
  // shift to x[1];
  l  = ni-3;
  
  for (i = 1; i<l; i++){
    // eqn 1.6    
    x->pc0  =  dx0*iddxp*(1.0+i2*(dxp+dxp)*((ddxm*i0) - (ddx2*i1)));
    //    x->pc1  = -i2*dx0*ddxm*i0;
    //    x->pc2  =  i2*dxp*ddx2*i1;
    //
    //  Important!!!!!! have swapped coeff 1 and 2, so 1 is applied to the current
    //  delta and 2 applies to the j+1 delta.
    //
    x->pc1  =  i2*dxp*ddx2*i1;
    x->pc2  = -i2*dx0*ddxm*i0;
    // eqn1.7    
    x->pc3  = i3*(ddxm+dxm)*iddxp;
    x->pc4  = i3*(ddxp+dxp)*iddxm;
    
    // shuffle along and do the minimum 5 necessary divisions
    x++;
    
    dx   -= (dxm-dx2);
    dxm   = dx0;
    dx0   = dxp;
    dxp   = dx2;
    
    ddxm  = ddxp;
    ddxp  = ddx2;
    
    dx2   = x[2].dx;
    ddx2 -= (dx0-dx2);
    
    iddxm = iddxp;
    iddxp = 1.0/ddxp;
    
    i0   = 1.0/(dx0+ddxp);
    i1   = 1.0/(ddxp+dxp);
    i2   = 1.0/(ddxm+ddx2);
    i3   = dx0/dx;
    
  }
  
#endif PPM   
  
#endif !GODUNOV  
  
  return(err);
  
}

int    zls_PrintAxisCoords(FILE *f, ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type ){
  int err = noErr;
  int i;
  if (x != NULL) {
    
#ifdef CARTESIAN
    i = 0;
    zlsio_FPrintf( f, "      X_0   : %3d \n", i);
    zlsio_FPrintf( f, "       l:%#18.16g\n       c:%#18.16g\n       r:%#18.16g \n", x[i].l, x[i].dx, x[i].r);
    i = ns;
    zlsio_FPrintf( f, "      X_s   : %3d \n", i);
    zlsio_FPrintf( f, "       l:%#18.16g\n       c:%#18.16g\n       r:%#18.16g \n", x[i].l, x[i].dx, x[i].r);
    i = ne-1;
    zlsio_FPrintf( f, "      X_e-1 : %3d \n", i);
    zlsio_FPrintf( f, "       l:%#18.16g\n       c:%#18.16g\n       r:%#18.16g \n", x[i].l, x[i].dx, x[i].r);
    i = nt-1;
    zlsio_FPrintf( f, "      X_t-1 : %3d \n", i);
    zlsio_FPrintf( f, "       l:%#18.16g\n       c:%#18.16g\n       r:%#18.16g \n", x[i].l, x[i].dx, x[i].r);
#endif CARTESIAN
    
#ifdef SPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_FPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_FPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_FPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_FPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_FPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypeTheta:
        i = 0;
        zlsio_FPrintf( f, "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif SPHERICAL
#ifdef POLARSPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_FPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_FPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_FPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_FPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_FPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypePolarTheta:
        i = 0;
        zlsio_FPrintf( f, "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    switch (type) {
      case kAxisTypeSpatial:
        i = 0;
        zlsio_FPrintf( f, "      Z_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_FPrintf( f, "      Z_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_FPrintf( f, "      Z_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_FPrintf( f, "      Z_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypeCRadius:
        i = 0;
        zlsio_FPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_FPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_FPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_FPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_FPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CYLINDRICAL
#ifdef CIRCULAR
    switch (type) {
      case kAxisTypeCRadius:
        i = 0;
        zlsio_FPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_FPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_FPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_FPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_FPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_FPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_FPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_FPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_FPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_FPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_FPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_FPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_FPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_FPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_FPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_FPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CIRCULAR
    
  } else {err =memErr;}
  
  return(err);
}


int    zls_ZPrintAxisCoords( zFile f, ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type ){
  int err = noErr;
  int i;
  if (x != NULL) {
    
#ifdef CARTESIAN
    i = 0;
    zlsio_ZFPrintf( f, "         X_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = ns;
    zlsio_ZFPrintf( f, "         X_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = ne-1;
    zlsio_ZFPrintf( f, "         X_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = nt-1;
    zlsio_ZFPrintf( f, "         X_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
#endif CARTESIAN
    
#ifdef SPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_ZFPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_ZFPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_ZFPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_ZFPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_ZFPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypeTheta:
        i = 0;
        zlsio_ZFPrintf( f, "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif SPHERICAL
#ifdef POLARSPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_ZFPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_ZFPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_ZFPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_ZFPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_ZFPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypePolarTheta:
        i = 0;
        zlsio_ZFPrintf( f, "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    switch (type) {
      case kAxisTypeSpatial:
        i = 0;
        zlsio_ZFPrintf( f, "      Z_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_ZFPrintf( f, "      Z_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_ZFPrintf( f, "      Z_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_ZFPrintf( f, "      Z_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypeCRadius:
        i = 0;
        zlsio_ZFPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_ZFPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_ZFPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_ZFPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_ZFPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CYLINDRICAL
#ifdef CIRCULAR
    switch (type) {
      case kAxisTypeCRadius:
        i = 0;
        zlsio_ZFPrintf( f, "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_ZFPrintf( f, "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_ZFPrintf( f, "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_ZFPrintf( f, "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_ZFPrintf( f, "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_ZFPrintf( f, "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_ZFPrintf( f, "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_ZFPrintf( f, "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_ZFPrintf( f, "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_ZFPrintf( f, "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_ZFPrintf( f, "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_ZFPrintf( f, "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_ZFPrintf( f, "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_ZFPrintf( f, "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CIRCULAR
    
  } else {err =memErr;}
  
  return(err);
}


int    zls_UserLogPrintAxisCoords( ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type ){
  int err = noErr;
  int i;
  if (x != NULL) {
    
#ifdef CARTESIAN
    i = 0;
    zlsio_UserLogPrintf(  "         X_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = ns;
    zlsio_UserLogPrintf(  "         X_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = ne-1;
    zlsio_UserLogPrintf(  "         X_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
    i = nt-1;
    zlsio_UserLogPrintf(  "         X_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
#endif CARTESIAN
    
#ifdef SPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_UserLogPrintf(  "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_UserLogPrintf(  "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_UserLogPrintf(  "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_UserLogPrintf(  "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_UserLogPrintf(  "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypeTheta:
        i = 0;
        zlsio_UserLogPrintf(  "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif SPHERICAL
#ifdef POLARSPHERICAL
    switch (type) {
      case kAxisTypeSRadius:
        i = 0;
        zlsio_UserLogPrintf(  "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_UserLogPrintf(  "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_UserLogPrintf(  "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_UserLogPrintf(  "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_UserLogPrintf(  "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
      case kAxisTypePolarTheta:
        i = 0;
        zlsio_UserLogPrintf(  "      T_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      T_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      T_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      T_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
    switch (type) {
      case kAxisTypeSpatial:
        i = 0;
        zlsio_UserLogPrintf(  "      Z_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_UserLogPrintf(  "      Z_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_UserLogPrintf(  "      Z_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_UserLogPrintf(  "      Z_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypeCRadius:
        i = 0;
        zlsio_UserLogPrintf(  "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_UserLogPrintf(  "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_UserLogPrintf(  "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_UserLogPrintf(  "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_UserLogPrintf(  "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CYLINDRICAL
#ifdef CIRCULAR
    switch (type) {
      case kAxisTypeCRadius:
        i = 0;
        zlsio_UserLogPrintf(  "      R_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ns;
        zlsio_UserLogPrintf(  "      R_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = ne-1;
        zlsio_UserLogPrintf(  "      R_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        i = nt-1;
        zlsio_UserLogPrintf(  "      R_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].l, x[i].dx, x[i].r);
        break;
      case kAxisTypePhi:
        i = 0;
        zlsio_UserLogPrintf(  "      P_0   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ns;
        zlsio_UserLogPrintf(  "      P_s   : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = ne-1;
        zlsio_UserLogPrintf(  "      P_e-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = nt-1;
        zlsio_UserLogPrintf(  "      P_t-1 : %5d %#18.16g %#18.16g %#18.16g (pi) \n", i, x[i].l/pi, x[i].dx/pi, x[i].r/pi);
        i = 0;
        zlsio_UserLogPrintf(  "      s_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ns;
        zlsio_UserLogPrintf(  "      s_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      s_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      s_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].sl, x[i].sc, x[i].sr);
        i = 0;
        zlsio_UserLogPrintf(  "      c_0   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ns;
        zlsio_UserLogPrintf(  "      c_s   : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = ne-1;
        zlsio_UserLogPrintf(  "      c_e-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        i = nt-1;
        zlsio_UserLogPrintf(  "      c_t-1 : %5d %#18.16g %#18.16g %#18.16g \n", i, x[i].cl, x[i].cc, x[i].cr);
        break;
    }; // switch
#endif CIRCULAR
    
  } else {err =memErr;}
  
  return(err);
}

int    zls_NewGridCoords     (GCoordPtr *g){ // if dynamically allocating, usually not necessary 
  
  int err = noErr;
  Integer i;
  
  GCoordPtr tempG = NULL;
  
  tempG = (GCoordPtr) malloc( sizeof(GridCoords) );
  if (tempG == NULL) {
    err = memErr;
  } else {
    
    tempG->nD  = 0;
    tempG->nG  = 0; 
    
    for (i = 0; i< kMaxCoordDims; i++){
      tempG->type[i] = kAxisTypeSpatial;
      tempG->kind[i] = kAxisKindLinear;
      
      tempG->n[i]  = 1; //  number of quantity cells, no bCells in Q arrays
      tempG->ns[i] = 0; //  start of active cells, each axis, steps over bCells
      tempG->ne[i] = 0; //  end of active cells, each axis
      tempG->nt[i] = 1; //  total # of cells, each axis
      
      tempG->xl[i] = 0.0; //
      tempG->xr[i] = 0.0; //  inner (left) x, outer (right) x of grid - range
      tempG->dx[i] = 0.0; //  simple global dx, range/nt, dxs in detail are in axes
    }
    
    tempG->x   = NULL;
    tempG->y   = NULL;
    tempG->z   = NULL;
    tempG->x0  = NULL;
    tempG->y0  = NULL;
    tempG->z0   = NULL;
    
  }
  
  *g = tempG;
  
  return(err);                 
  
}

int    zls_DisposeGridCoords (GCoordPtr g){ // *only* if dynamically allocated, usually not necessary 
  
  int err = noErr;
  Integer i;
  
  if (g != NULL) {
    
    g->nD      = 0;
    g->nG      = 0; 
    
    for (i = 0; i < kMaxCoordDims; i++){
      
      g->type[i] = 0;
      g->kind[i] = 0;
      g->n[i]    = 0;     
      g->ns[i]   = 0; 
      g->ne[i]   = 0; 
      g->nt[i]   = 0; 
      
      g->xl[i]   = 0.0;  
      g->xr[i]   = 0.0;  
      g->dx[i]   = 0.0; 
      
    } 
    
    zls_DisposeAxis (g->x);  g->x = NULL;
    zls_DisposeAxis (g->y);  g->y = NULL;
    zls_DisposeAxis (g->z);  g->z = NULL;
    
    zls_DisposeAxis(g->x0); g->x0 = NULL;
    zls_DisposeAxis(g->y0); g->y0 = NULL;
    zls_DisposeAxis(g->z0); g->z0 = NULL;
    
    free(g);
    
  }
  
  return(err);
  
}

int    zls_ClearGridCoords (GCoordPtr g){ // just zot grid and the axes, no free g
  
  int err = noErr;
  int i;
  
  if (g != NULL) {
    
    g->x = NULL;
    g->y = NULL;
    g->z = NULL;
    
    g->x0 = NULL;
    g->y0 = NULL;
    g->z0 = NULL;
    
    g->nD = 0;
    g->nG = 0; 
    
    for (i = 0; i< kMaxCoordDims; i++){
      
      g->type[i] = 0;
      g->kind[i] = 0;
      g->n[i]    = 1;     
      g->ns[i]   = 0; 
      g->ne[i]   = 0; 
      g->nt[i]   = 1; 
      
      g->xl[i]   = 0.0;  
      g->xr[i]   = 0.0;  
      g->dx[i]   = 0.0; 
      
    } 
  }
  
  return(err);
  
}

int zls_InitGridCoords (GCoordPtr g, Integer ndims, Integer nbc,  Integer allocAxes, RangePtr r){  // types and kinds in range
  
  int err = noErr;
  Integer  ni, nj, nk;
  Real    ix, ox;
  
  if ((g != NULL)&&(r != NULL)) {
    
    // clear g initially
    zls_ClearGridCoords (g);
    
    g->nG   = nbc;    // number of boundary cells, same all axes 
    g->nD   = ndims;  // number of dimensions, 1, 2 or 3
    
    switch (ndims){ // fall through switch!!!
        
      case k3D:
        
        nk      = r->rn[kZDim];
        ix      = r->x0[kZDim];
        if (r->kind[kZDim] == kAxisKindLog) {
          ox      = r->x0[kZDim]*r->rx[kZDim]; // in log axes, rx is a ratio
        } else {
          ox      = r->x0[kZDim]+r->rx[kZDim];
        };
        
        g->type[kZDim]  = r->type[kZDim];
        g->kind[kZDim]  = r->kind[kZDim];
        
        g->n[kZDim]     = nk;       // number of cells, each axis, no bCells
        g->ns[kZDim]    = nbc;        // start index, including bCells i = is
        g->ne[kZDim]    = nk+nbc;   // end index, including bCells, i < ie
        g->nt[kZDim]    = nk+nbc+nbc;
        
        g->xl[kZDim]    = ix;       // just the active area, axes include bCells
        g->xr[kZDim]    = ox;       // just the active area, axes include bCells
        if (r->kind[kZDim] == kAxisKindLog) {
          g->dx[kZDim]  = exp(log(r->rx[kZDim])/nk);  
        } else {
          g->dx[kZDim]  = r->rx[kZDim]/nk;    // just the active area, axes include bCells
        };
        
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis(&(g->z), nbc, nk) ;     // initial axes if using lagrangian/remapping, NULL otherwise
          zls_SetAxisRange(g->z, kZDim,  r);  // Resulting Axes include coords of bCells
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis(&(g->z0), nbc, nk);     // initial axes if using lagrangian/remapping, NULL otherwise
            zls_SetAxisRange(g->z0, kZDim, r);  // Resulting Axes include coords of bCells
          } else {
            g->z0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        }; //axes
        
        case k2D:
        
        nj      = r->rn[kYDim];
        ix      = r->x0[kYDim];
        if (r->kind[kYDim] == kAxisKindLog) {
          ox      = r->x0[kYDim]*r->rx[kYDim]; // in log axes, rx is a ratio
        } else {
          ox      = r->x0[kYDim]+r->rx[kYDim];
        };
        
        g->type[kYDim]  = r->type[kYDim];
        g->kind[kYDim]  = r->kind[kYDim];
        
        g->n[kYDim]     = nj;       // number of cells, each axis, no bCells
        g->ns[kYDim]    = nbc;        // start index, including bCells i = is
        g->ne[kYDim]    = nj+nbc;   // end index, including bCells, i < ie
        g->nt[kYDim]    = nj+nbc+nbc;
        
        g->xl[kYDim]   = ix;        // just the active area, axes include bCells
        g->xr[kYDim]   = ox;        // just the active area, axes include bCells
        if (r->kind[kYDim] == kAxisKindLog) {
          g->dx[kYDim]   = exp(log(r->rx[kYDim])/nj); 
        } else {
          g->dx[kYDim]   = r->rx[kYDim]/nj;
        };
        
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis(&(g->y), nbc, nj);            // initial axes if using lagrangian/remapping, NULL otherwise
          zls_SetAxisRange(g->y, kYDim, r);      // Resulting Axes include coords of bCells
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis(&(g->y0), nbc, nj);            // initial axes if using lagrangian/remapping, NULL otherwise
            zls_SetAxisRange(g->y0, kYDim, r);  // Resulting Axes include coords of bCells
          } else {
            g->y0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        }; //axes
        
        case k1D:
        
        ni      = r->rn[kXDim];
        ix      = r->x0[kXDim];
        
        if (r->kind[kXDim] == kAxisKindLog) {
          ox      = r->x0[kXDim]*r->rx[kXDim]; // in log axes, rx is a ratio
        } else {
          ox      = r->x0[kXDim]+r->rx[kXDim];
        };
        
        g->type[kXDim]  = r->type[kXDim];
        g->kind[kXDim]  = r->kind[kXDim];
        
        g->n[kXDim]     = ni;       // number of cells, each axis, no bCells
        g->ns[kXDim]    = nbc;        // start index, including bCells i = is
        g->ne[kXDim]    = ni+nbc;   // end index, including bCells, i < ie
        g->nt[kXDim]    = ni+nbc+nbc;
        
        g->xl[kXDim]   = ix;        // just the active area, axes include bCells
        g->xr[kXDim]   = ox;        // just the active area, axes include bCells
        if (r->kind[kXDim] == kAxisKindLog) {
          g->dx[kXDim]   = exp(log(r->rx[kXDim])/ni); 
        } else {
          g->dx[kXDim]   = r->rx[kXDim]/ni;   // just the active area, axes include bCells
        };
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis(&(g->x), nbc, ni) ;            // initial axes if using lagrangian/remapping, NULL otherwise
          zls_SetAxisRange(g->x, kXDim, r);  // Resulting Axes include coords of bCells
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis(&(g->x0), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
            zls_SetAxisRange(g->x0, kXDim, r); // Resulting Axes include coords of bCells
          } else {
            g->x0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        } //axes
        
        break;
        default:
        err   = dimErr;
        zlsio_ErrPrintf( " ERROR, Dimension Error in zls_InitGridCoords: %d \n\n",ndims);
        return(err);
    } // switch
    
    
  } else { err = memErr;}
  
  return(err);                 
}

int    zls_SetGridRange       (GCoordPtr g, Integer dim, Integer useAxes,   RangePtr r){ // used grid bc settings
  
  int err = noErr;
  Real ix, ox;
  Integer ni, nbc;
  
  if ((g!=NULL)&&(r!=NULL)){
    
    if ((dim < g->nD)){
      
      nbc     = g->nG;
      ni      = r->rn[dim];
      ix      = r->x0[dim];
      
      if (r->kind[dim] == kAxisKindLog) {
        ox      = ix*r->rx[dim]; // in log axes, rx is a ratio
      } else {
        ox      = ix+r->rx[dim]; // in lin axes, rx is a difference
      }
      
      g->type[dim]  = r->type[dim];
      g->kind[dim]  = r->kind[dim];
      
      g->n[dim]     = ni;       // number of quantity cells, each axis, no bCells
      g->ns[dim]    = nbc;        // variable start index, including bCells, used in axes i = is
      g->ne[dim]    = ni+nbc;     // variable end index, including bCells, i < ie
      g->nt[dim]    = ni+nbc+nbc;
      
      g->xl[dim]   = ix;        // just the active area, axes include bCells
      g->xr[dim]   = ox;        // just the active area, axes include bCells
      if (r->kind[dim] == kAxisKindLog) {
        g->dx[dim]   = exp(log(r->rx[dim])/ni); 
      } else {
        g->dx[dim]   = r->rx[dim]/ni;
      }
      
      if (useAxes == kAllocateAxes) {
        
        switch (dim){
            
          case kZDim:
            zls_NewAxis(&(g->z), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
            zls_InitAxis(g->z,  g->type[kZDim], g->kind[kZDim],  nbc, ni, ix, ox);  // Resulting Axes include coords of bCells
            
            if (kAxesType == kLagrangeGrid) {
              zls_NewAxis(&(g->z0), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
              zls_InitAxis(g->z0, g->type[kZDim], g->kind[kZDim], nbc, ni, ix, ox);   // Resulting Axes include coords of bCells
            } else {
              g->z0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
            }
            break;
            case kYDim:           
            zls_NewAxis(&(g->y), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
            zls_InitAxis(g->y, g->type[kYDim],  g->kind[kYDim], nbc, ni, ix, ox);  // Resulting Axes include coords of bCells
            
            if (kAxesType == kLagrangeGrid) {
              zls_NewAxis(&(g->y0), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
              zls_InitAxis(g->y0, g->type[kYDim], g->kind[kYDim], nbc, ni, ix, ox);   // Resulting Axes include coords of bCells
            } else {
              g->y0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
            }
            break;
            case kXDim:
            zls_NewAxis(&(g->x), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
            zls_InitAxis(g->x,g->type[kXDim],  g->kind[kXDim],  nbc, ni, ix, ox);  // Resulting Axes include coords of bCells
            
            if (kAxesType == kLagrangeGrid) {
              zls_NewAxis(&(g->x0), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
              zls_InitAxis(g->x0, g->type[kXDim],  g->kind[kXDim], nbc, ni, ix, ox);   // Resulting Axes include coords of bCells
            } else {
              g->x0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
            }
            
            break;
            default:
            err   = dimErr;
            zlsio_ErrPrintf( " Dimension Error in zls_SetGridRange: %d\n\n",dim);
            return(err);
        } // switch
        
      }
    } else {
      err   = dimErr;
      zlsio_ErrPrintf( " Dimension Error in zls_SetGridRange: %d %d\n\n",dim, r->nD);
      return(err);
    }
    
  } //g r
  
  return(err);
}

int zls_GetRangeDimsFromGrid(RangePtr r, GCoordPtr g){
  // get dimensions and axis types even before actual sizes and deltas have been determined  
  int err = noErr;
  Integer dim;
  
  if ((g != NULL)&&(r != NULL)) {
    if (g->nD > 0){
      r->nD        = g->nD;
      r->nG        = g->nG;
      for ( dim = 0; dim < g->nD; dim++) {
        r->type[dim]   = g->type[dim];
        r->kind[dim]   = g->kind[dim];
      }
    } else { err = dimErr;}
  } else { err = memErr;}
  
  return(err);                 
}

int zls_GetRangeFromGrid(RangePtr r, Integer dim, GCoordPtr g){
  
  int err = noErr;
  
  if ((g != NULL)&&(r != NULL)) {
    if (g->nD > (dim-1)){
      r->nD        = g->nD;
      r->nG        = g->nG;
      r->type[dim] = g->type[dim];
      r->kind[dim] = g->kind[dim];
      r->n0[dim]   = 0;
      r->rn[dim]   = g->n[dim];
      r->x0[dim]   =  g->xl[dim] ;
      if (r->kind[dim] == kAxisKindLog) {
        r->rx[dim]  =  g->xr[dim]/g->xl[dim]; // in log axes, rx is a ratio
      } else {
        r->rx[dim]  =  g->xr[dim]-g->xl[dim]; // in lin axes, rx is a difference
      }
    } else { err = dimErr;}
  } else { err = memErr;}
  
  return(err);                 
}

int zls_GetSubRangeFromGrid(RangePtr r, Integer dim, GCoordPtr g, Integer n0, Integer n1){
  
  int err = noErr;
  
  Integer nbc;
  Integer n;
  
  if ((g != NULL)&&(r != NULL)) {
    nbc    =  g->nG;
    r->nG  =  nbc;
    n      = n1-n0;
    r->n0[dim] = n0;
    r->rn[dim] = n;
    
    if (r->nD > (dim)){
      
      r->type[dim]   = g->type[dim];
      r->kind[dim]   = g->kind[dim];
      
      if (dim == kXDim) {
        if (g->x !=NULL){
          r->x0[dim] =  g->x[n0+nbc].l; 
          if (r->kind[dim] == kAxisKindLog) {
            r->rx[dim] = g->x[n1+nbc].l / r->x0[dim];
          } else {
            r->rx[dim] = g->x[n1+nbc].l - r->x0[dim];
          }
        } else {
          r->x0[dim] =  g->xl[dim]+(n0*g->dx[dim]);
          r->rx[dim] =  r->x0[dim]+g->dx[dim]*n;
        }
      }
      if (dim == kYDim) {
        if (g->y !=NULL){
          r->x0[dim] =  g->y[n0+nbc].l; 
          if (r->kind[dim] == kAxisKindLog) {
            r->rx[dim] = g->y[n1+nbc].l / r->x0[dim];
          } else {
            r->rx[dim] = g->y[n1+nbc].l - r->x0[dim];
          }
        } else {
          r->x0[dim] = g->xl[dim]+(n0*g->dx[dim]);
          r->rx[dim] = g->dx[dim]*(n1+nbc);
        }
      }
      if (dim == kZDim) {
        if (g->z !=NULL){
          r->x0[dim] = g->z[n0+nbc].l; 
          if (r->kind[dim] == kAxisKindLog) {
            r->rx[dim] = g->z[n1+nbc].l / r->x0[dim];
          } else {
            r->rx[dim] = g->z[n1+nbc].l - r->x0[dim];
          }
        } else {
          r->x0[dim] = g->xl[dim]+(n0*g->dx[dim]);
          r->rx[dim] = g->dx[dim]*(n1+nbc);
        }
      }
    }
    
    if (r->nD > 3){
      err = dimErr;
    }
    
    if (r->nD <= 0){
      err = dimErr;
    }
    
  } else { err = memErr;}
  
  return(err);                 
}


int zls_GetSubRangeFromGridDelta(RangePtr r, Integer dim, GCoordPtr g, Integer n0, Integer deltan){
  
  int err = noErr;
  
  Integer nbc;
  Integer n1;
  
  if ((g != NULL)&&(r != NULL)) {
    nbc    =  g->nG;
    r->nG  =  nbc;
    
    n1         = n0 + deltan;
    r->n0[dim] = n0;
    r->rn[dim] = deltan;
    
    
    if (r->nD > (dim)){
      
      r->type[dim]   = g->type[dim];
      r->kind[dim]   = g->kind[dim];
      
#if (ZLS_DEBUG > 0)
      zlsio_Printf( "   zls_GetSubRangeFromGridDelta: n0 %d deltan %d \n", n0, deltan);
#endif                
      
      if ((n0 == 0) && (deltan == (Integer) g->n[dim])) {// common shortcut, range == whole grid
        r->x0[dim] = g->xl[dim]; 
        if (r->kind[dim] == kAxisKindLog) {
          r->rx[dim] = g->xr[dim] /g->xl[dim];
        } else {
          r->rx[dim] =g->xr[dim] -g->xl[dim];
        }
#if (ZLS_DEBUG > 0)
        zlsio_Printf( "   Shortcut:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
        return(err);
      }
      
      switch (dim) {
        case  kXDim:
          if (g->x !=NULL){
            r->x0[dim] =  g->x[n0+nbc].l; 
            if (r->kind[dim] == kAxisKindLog) {
              r->rx[dim] = g->x[n1+nbc].l / g->x[n0+nbc].l;
            } else {
              r->rx[dim] = g->x[n1+nbc].l - g->x[n0+nbc].l;
            };
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Axes X:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          } else {// this occurs when a level is used that has no axes 
            if (r->kind[dim] == kAxisKindLog) {
              r->x0[dim] =  g->xl[dim]*pow(g->dx[dim],n0) ; // this occurs when a level is used that has no axes 
              r->rx[dim] = pow(g->dx[dim],deltan); // if using xl, then we are starting on the active border
            } else {
              r->x0[dim] = g->xl[dim]+(g->dx[dim]*n0) ; // this occurs when a level is used that has no axes 
              r->rx[dim] = (g->dx[dim]*deltan); // if using xl, then we are starting on the active border
            };
            r->x0[dim] = r->x0[dim] ;
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Coords X:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          };
          break;
          case  kYDim:
          if (g->y !=NULL){
            r->x0[dim] = g->y[n0+nbc].l;
            if (r->kind[dim] == kAxisKindLog) {
              r->rx[dim] = g->y[n1+nbc].l / g->y[n0+nbc].l;
            } else {
              r->rx[dim] = g->y[n1+nbc].l - g->y[n0+nbc].l;
            };
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Axes Y:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          } else {
            if (r->kind[dim] == kAxisKindLog) {
              r->x0[dim] =  g->xl[dim]*pow(g->dx[dim],n0) ; // this occurs when a level is used that has no axes 
              r->rx[dim] = pow(g->dx[dim],deltan); // if using xl, then we are starting on the active border
            } else {
              r->x0[dim] = g->xl[dim]+(g->dx[dim]*n0) ; // this occurs when a level is used that has no axes 
              r->rx[dim] = (g->dx[dim]*deltan); // if using xl, then we are starting on the active border
            };
            r->x0[dim] = r->x0[dim];
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Coords Y:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          };
          break;
          case  kZDim:
          if (g->z !=NULL){
            r->x0[dim] = g->z[n0+nbc].l ;
            if (r->kind[dim] == kAxisKindLog) {
              r->rx[dim] = g->z[n1+nbc].l / g->z[n0+nbc].l;
            } else {
              r->rx[dim] = g->z[n1+nbc].l - g->z[n0+nbc].l;
            };
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Axes Z:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          } else {
            if (r->kind[dim] == kAxisKindLog) {
              r->x0[dim] = g->xl[dim]*pow(g->dx[dim],n0); // this occurs when a level is used that has no axes 
              r->rx[dim] = pow(g->dx[dim],deltan); // if using xl, then we are starting on the active border
            } else {
              r->x0[dim] = g->xl[dim]+(g->dx[dim]*n0); // this occurs when a level is used that has no axes 
              r->rx[dim] = (g->dx[dim]*deltan); // if using xl, then we are starting on the active border
            };
            r->x0[dim] = r->x0[dim];
#if (ZLS_DEBUG > 0)
            zlsio_Printf( "   From Super Coords Z:  zls_GetSubRangeFromGridDelta: x0 %g rx %g \n",  r->x0[dim], r->rx[dim]);
#endif                
          };
          break;
          default:
          err = dimErr;
      } // switch
      
    }; //r->nD
    
    if (r->nD > 3){
      err = dimErr;
    }
    
    if (r->nD <= 0){
      err = dimErr;
    }
    
  } else { err = memErr;}
  
  return(err);                 
}


int zls_InitSubGridCoords (GridCoords * g, GridCoords * sg, Integer allocAxes, RangePtr r){  // types and kinds in range
  
  int err = noErr;
  Integer  ni0, nj0, nk0;
  Integer  ni, nj, nk;
  Real     ix, ox;
  
  if ((g != NULL)&&(sg != NULL)&&(r != NULL)) {
    
    // clear g initially
    zls_ClearGridCoords (g);
    
    Integer nbc   = sg->nG;
    Integer ndims = sg->nD;
    g->nG   = nbc;  // number of boundary cells, same all axes 
    g->nD   = ndims;  // number of dimensions, 1, 2 or 3
    
    switch ( ndims ){ // fall through switch!!!
        
      case k3D:
        
        nk      = r->rn[kZDim];
        nk0     = r->n0[kZDim]+nbc;
        
        ix      = sg->z[nk0          ].l;
        ox      = sg->z[nk0 + nk - 1 ].r;
        
        g->type[kZDim]  = sg->type[kZDim];
        g->kind[kZDim]  = sg->kind[kZDim];
        
        g->n [kZDim]    = nk;       // number of cells, each axis, no bCells
        g->ns[kZDim]    = nbc;        // start index, including bCells i = is
        g->ne[kZDim]    = nk+nbc;   // end index, including bCells, i < ie
        g->nt[kZDim]    = nk+nbc+nbc;
        
        g->xl[kZDim]    = ix;       // just the active area, axes include bCells
        g->xr[kZDim]    = ox;       // just the active area, axes include bCells
        g->dx[kZDim]    = sg->dx[kZDim];  
        
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis(&(g->z), nbc, nk) ;     // initial axes if using lagrangian/remapping, NULL otherwise
          if (sg->z != NULL) {
            zls_CopyAxisRange( g->z   , sg->z, kZDim, r  );      // Resulting Axes include coords of bCells
          } else {
            zls_SetAxisRange ( g->z    , kZDim, r  );      // Resulting Axes include coords of bCells
          }
          
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis(&(g->z0), nbc, nk);     // initial axes if using lagrangian/remapping, NULL otherwise
            if (sg->z != NULL) {
              zls_SetAxisRange ( g->z0    , kZDim, r  );      // Resulting Axes include coords of bCells
            }
            
          } else {
            g->z0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        }; //axes
        
        case k2D:
        
        nj      = r->rn[kYDim];
        nj0     = r->n0[kYDim]+nbc;
        
        ix      = sg->y[nj0          ].l;
        ox      = sg->y[nj0 + nj - 1 ].r;
        
        g->type[kYDim]  = sg->type[kYDim];
        g->kind[kYDim]  = sg->kind[kYDim];
        
        g->n [kYDim]    = nj;       // number of cells, each axis, no bCells
        g->ns[kYDim]    = nbc;        // start index, including bCells i = is
        g->ne[kYDim]    = nj+nbc;   // end index, including bCells, i < ie
        g->nt[kYDim]    = nj+nbc+nbc;
        
        g->xl[kYDim]    = ix;       // just the active area, axes include bCells
        g->xr[kYDim]    = ox;       // just the active area, axes include bCells
        g->dx[kYDim]    = sg->dx[kYDim];  
        
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis       ( &(g->y), nbc  , nj );            // initial axes if using lagrangian/remapping, NULL otherwise
          if (sg->y != NULL) {
            zls_CopyAxisRange ( g->y   , sg->y, kYDim, r  );      // Resulting Axes include coords of bCells
          } else {
            zls_SetAxisRange ( g->y   , kYDim, r  );      // Resulting Axes include coords of bCells
          }
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis       ( &(g->y0), nbc  , nj );            // initial axes if using lagrangian/remapping, NULL otherwise
            if (sg->y != NULL) {
              zls_CopyAxisRange ( g->y0   , sg->y, kYDim, r  );      // Resulting Axes include coords of bCells
            } else {
              zls_SetAxisRange ( g->y0   , kYDim, r  );      // Resulting Axes include coords of bCells
            }
            
          } else {
            g->y0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        }; //axes
        
        case k1D:
        
        
        ni      = r->rn[kXDim];
        ni0     = r->n0[kXDim]+nbc;
        
        ix      = sg->x[ni0          ].l;
        ox      = sg->x[ni0 + ni - 1 ].r;
        
        g->type[kXDim]  = sg->type[kXDim];
        g->kind[kXDim]  = sg->kind[kXDim];
        
        g->n [kXDim]    = ni;       // number of cells, each axis, no bCells
        g->ns[kXDim]    = nbc;      // start index, including bCells i = is
        g->ne[kXDim]    = ni+nbc;   // end index, including bCells, i < ie
        g->nt[kXDim]    = ni+nbc+nbc;
        
        g->xl[kXDim]    = ix;       // just the active area, axes include bCells
        g->xr[kXDim]    = ox;       // just the active area, axes include bCells
        g->dx[kXDim]    = sg->dx[kXDim];  
        
        if (allocAxes == kAllocateAxes) {
          
          zls_NewAxis       ( &(g->x), nbc  , ni );  // initial axes if using lagrangian/remapping, NULL otherwise
          if (sg->x != NULL) {
            zls_CopyAxisRange ( g->x   , sg->x, kXDim, r  );      // Resulting Axes include coords of bCells
          } else {
            zls_SetAxisRange ( g->x    , kXDim, r  );      // Resulting Axes include coords of bCells
          }
          
          if (kAxesType == kLagrangeGrid) {
            zls_NewAxis       ( &(g->x0), nbc, ni);            // initial axes if using lagrangian/remapping, NULL otherwise
            if (sg->x != NULL) {
              zls_CopyAxisRange ( g->x0   , sg->x, kXDim, r  );      // Resulting Axes include coords of bCells
            } else {
              zls_SetAxisRange ( g->x0   , kXDim, r  );      // Resulting Axes include coords of bCells
            }
          } else {
            g->x0 = NULL; // initial axes if using lagrangian/remapping, NULL otherwise
          } //kAxesType
        } //axes
        
        break;
        default:
        err   = dimErr;
        zlsio_ErrPrintf( " ERROR, Dimension Error in zls_InitSubGridCoords: %d \n\n", ndims);
        return(err);
    } // switch
    
    
  } else { err = memErr;}
  
  return(err);                 
}



int    zls_CloneGridCoords (GCoordPtr g1, GCoordPtr g0){ 
  
  int err = noErr;
  ACoordPtr tempX = NULL;
  
  if ((g1 != NULL)&&(g0 != NULL)) {
    
    *g0 = *g1;
    
    if (g1->nD == 1) {
      if (g1->x !=NULL) {
        err = zls_NewAxis(&tempX, g1->nG, g1->nt[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kXDim], g1->kind[kXDim], g1->nG,  g1->n[kXDim],   g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x = tempX;
      }
      
      if (g1->x0 !=NULL) {
        err = zls_NewAxis(&tempX, g1->nG, g1->nt[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kXDim], g1->kind[kXDim],  g1->nG,   g1->n[kXDim],  g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x0 = tempX;
      }
      
      g1->y = NULL;
      g1->z = NULL;
      g1->y0 = NULL;
      g1->z0 = NULL;
      
    } //nD1
    
    if (g1->nD == 2) {
      
      if (g1->x !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kXDim], g1->kind[kXDim], g1->nG,  g1->n[kXDim],   g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x = tempX;
      }
      
      if (g1->y !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kYDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kYDim], g1->kind[kYDim], g1->nG,  g1->n[kYDim],   g1->xl[kYDim], g1->xr[kYDim]); // Resulting Axes include coords of bCells
        g1->y = tempX;
      }
      
      if (g1->x0 !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kXDim], g1->kind[kXDim], g1->nG,   g1->n[kXDim],  g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x0 = tempX;
      }
      
      if (g1->y0 !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kYDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kYDim], g1->kind[kYDim], g1->nG,  g1->n[kYDim],  g1->xl[kYDim], g1->xr[kYDim]); // Resulting Axes include coords of bCells
        g1->y0 = tempX;
      }
      
      g1->z = NULL;
      g1->z0 = NULL;
      
    } //nD2
    
    
    if (g1->nD == 3) {
      
      if (g1->x !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX,  g1->type[kXDim], g1->kind[kXDim], g1->nG,  g1->n[kXDim],   g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x = tempX;
      }
      
      if (g1->y !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kYDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kYDim], g1->kind[kYDim], g1->nG,  g1->n[kYDim],   g1->xl[kYDim], g1->xr[kYDim]); // Resulting Axes include coords of bCells
        g1->y = tempX;
      }
      
      if (g1->z !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kZDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kZDim], g1->kind[kZDim], g1->nG,  g1->n[kZDim],   g1->xl[kZDim], g1->xr[kZDim]); // Resulting Axes include coords of bCells
        g1->z = tempX;
      }
      
      if (g1->x0 !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kXDim]);
        if (tempX != NULL) zls_InitAxis (tempX, g1->type[kXDim], g1->kind[kXDim], g1->nG,   g1->n[kXDim],  g1->xl[kXDim], g1->xr[kXDim]); // Resulting Axes include coords of bCells
        g1->x0 = tempX;
      }
      
      if (g1->y0 !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kYDim]);
        if (tempX != NULL) zls_InitAxis (tempX,  g1->type[kYDim], g1->kind[kYDim],  g1->nG,  g1->n[kYDim],  g1->xl[kYDim], g1->xr[kYDim]); // Resulting Axes include coords of bCells
        g1->y0 = tempX;
      }
      
      if (g1->z0 !=NULL) {
        err = zls_NewAxis(&tempX,  g1->nG, g1->n[kZDim]);
        if (tempX != NULL) zls_InitAxis (tempX,   g1->type[kZDim], g1->kind[kZDim],  g1->nG,  g1->n[kZDim],  g1->xl[kZDim], g1->xr[kZDim]); // Resulting Axes include coords of bCells
        g1->z0 = tempX;
      }
      
    } // nD3
    
  } // ggs
  
  return(err);                 
}


int    zls_PrintGridCoords     (FILE *f,  GCoordPtr g){ 
  int err = noErr;
  int i;
  
  if (g != NULL) {
    
    zlsio_FPrintf( f, "     Grid Dimensions     : %d \n", g->nD);
    zlsio_FPrintf( f, "     Grid Boundary Cells : %d \n", g->nG);
    
    for (i = 0; i<g->nD; i++){
      switch (g->type[i]) {
        default:
        case kAxisTypeSpatial:
          zlsio_FPrintf( f, "     Grid Dimension Type: Spatial, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeCRadius:
          zlsio_FPrintf( f, "     Grid Dimension Type: Circular Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeSRadius:
          zlsio_FPrintf( f, "     Grid Dimension Type: Spherical Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_FPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_FPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_FPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_FPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypePhi:
          zlsio_FPrintf( f, "     Grid Dimension Type: Azimuthal Phi %d \n", g->type[i]);
          zlsio_FPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_FPrintf( f, "         lphi[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_FPrintf( f, "         rphi[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_FPrintf( f, "         dphi[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypeTheta:
          zlsio_FPrintf( f, "     Grid Dimension Type: Altitude Theta %d \n", g->type[i]);
          zlsio_FPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_FPrintf( f, "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_FPrintf( f, "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_FPrintf( f, "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypePolarTheta:
          zlsio_FPrintf( f, "     Grid Dimension Type: Polar-Colatitude Theta %d \n", g->type[i]);
          zlsio_FPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_FPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_FPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_FPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_FPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_FPrintf( f, "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_FPrintf( f, "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_FPrintf( f, "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
      }; // switch
    }; // i
    
    switch (g->nD) {
        
      case k3D:
        if ( g->z != NULL) {
          zlsio_FPrintf( f, "    Z-Axis allocated\n" );
          //  zls_PrintAxisCoords( f, g->z, g->ns[kZDim], g->ne[kZDim], g->nt[kZDim], g->type[kZDim]);
        }
        if (g->z0 != NULL) {
          zlsio_FPrintf( f, "    Z0-Axis allocated\n" );
        }
        case k2D:
        if ( g->y != NULL) {
          zlsio_FPrintf( f, "    Y-Axis allocated\n" );
          // zls_PrintAxisCoords( f, g->y, g->ns[kYDim],g->ne[kYDim],g->nt[kYDim], g->type[kYDim]);
        }
        if (g->y0 != NULL) {
          zlsio_FPrintf( f, "    Y0-Axis allocated\n" );
        }
        case k1D:
        if ( g->x != NULL) {
          zlsio_FPrintf( f, "    X-Axis allocated\n" );
          //zls_PrintAxisCoords( f, g->x, g->ns[kXDim],g->ne[kXDim],g->nt[kXDim], g->type[kXDim]);
        }
        if (g->x0 != NULL) {
          zlsio_FPrintf( f, "    X0-Axis allocated\n" );
        }
        break;
        default:
        zlsio_ErrPrintf( " Error: Dimensions %d\n\n", g->nD);
        
    }
    
    
  } else { err = memErr ; }
  
  return(err);                 
}


int    zls_ZPrintGridCoords     ( zFile f,  GCoordPtr g){ 
  int err = noErr;
  int i;
  
  if (g != NULL) {
    
    zlsio_ZFPrintf( f, "     Grid nD: %d \n", g->nD);
    zlsio_ZFPrintf( f, "     Grid nG: %d \n", g->nG);
    
    for (i = 0; i<g->nD; i++){
      switch (g->type[i]) {
        default:
        case kAxisTypeSpatial:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Spatial, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeCRadius:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Circular Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeSRadius:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Spherical Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_ZFPrintf( f, "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_ZFPrintf( f, "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_ZFPrintf( f, "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_ZFPrintf( f, "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypePhi:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Azimuthal Phi %d \n", g->type[i]);
          zlsio_ZFPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_ZFPrintf( f, "         lphi[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_ZFPrintf( f, "         rphi[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_ZFPrintf( f, "         dphi[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypeTheta:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Altitude Theta %d \n", g->type[i]);
          zlsio_ZFPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_ZFPrintf( f, "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_ZFPrintf( f, "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_ZFPrintf( f, "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypePolarTheta:
          zlsio_ZFPrintf( f, "     Grid Dimension Type: Polar-Colatitude Theta %d \n", g->type[i]);
          zlsio_ZFPrintf( f, "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_ZFPrintf( f, "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_ZFPrintf( f, "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_ZFPrintf( f, "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_ZFPrintf( f, "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_ZFPrintf( f, "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_ZFPrintf( f, "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_ZFPrintf( f, "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
      }; // switch
    }; // i
    
#if (ZLS_DEBUG>0)   
    switch (g->nD) {
        
      case k3D:
        if ( g->z != NULL) {
          zlsio_ZFPrintf( f, "    Z-Axis allocated\n" );
          zls_ZPrintAxisCoords( f, g->z, g->ns[kZDim], g->ne[kZDim], g->nt[kZDim], g->type[kZDim]);
        }
        if (g->z0 != NULL) {
          zlsio_ZFPrintf( f, "    Z0-Axis allocated\n" );
        }
        case k2D:
        if ( g->y != NULL) {
          zlsio_ZFPrintf( f, "    Y-Axis allocated\n" );
          zls_ZPrintAxisCoords( f, g->y, g->ns[kYDim],g->ne[kYDim],g->nt[kYDim], g->type[kYDim]);
        }
        if (g->y0 != NULL) {
          zlsio_ZFPrintf( f, "    Y0-Axis allocated\n" );
        }
        case k1D:
        if ( g->x != NULL) {
          zlsio_ZFPrintf( f, "    X-Axis allocated\n" );
          zls_ZPrintAxisCoords( f, g->x, g->ns[kXDim],g->ne[kXDim],g->nt[kXDim], g->type[kXDim]);
          
        }
        if (g->x0 != NULL) {
          zlsio_ZFPrintf( f, "    X0-Axis allocated\n" );
        }
        break;
        default:
        zlsio_ErrPrintf( " Error: Dimensions %d\n\n", g->nD);
        
    }
    
#endif
    
  } else { err = memErr ; }
  
  return(err);                 
}


int    zls_UserLogPrintGridCoords     ( GCoordPtr g){ 
  int err = noErr;
  int i;
  
  if (g != NULL) {
    
    zlsio_UserLogPrintf(  "     Grid nD: %d \n", g->nD);
    zlsio_UserLogPrintf(  "     Grid nG: %d \n", g->nG);
    
    for (i = 0; i<g->nD; i++){
      switch (g->type[i]) {
        default:
        case kAxisTypeSpatial:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Spatial, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lx[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rx[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dx[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_UserLogPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeCRadius:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Circular Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_UserLogPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypeSRadius:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Spherical Radius, %d \n", g->type[i]);
          if (  g->kind[i] == kAxisKindLinear ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Linear %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else  if (  g->kind[i] == kAxisKindLog ){
            zlsio_UserLogPrintf(  "     Grid Dimension Kind: Logarithmic %d \n", g->kind[i]);
            zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
            zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
            zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
            zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
            zlsio_UserLogPrintf(  "           lr[%d]: %#18.16g \n",i, g->xl[i]);
            zlsio_UserLogPrintf(  "           rr[%d]: %#18.16g \n",i, g->xr[i]);
            zlsio_UserLogPrintf(  "           dr[%d]: %#18.16g \n",i, g->dx[i]);
          } else {
            zlsio_UserLogPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
            zlsio_ErrPrintf(  " ERROR: Unknown Grid Dimension Kind %d \n",g->kind[i]);
          };
          break;
          case kAxisTypePhi:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Azimuthal Phi %d \n", g->type[i]);
          zlsio_UserLogPrintf(  "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_UserLogPrintf(  "         lphi[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_UserLogPrintf(  "         rphi[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_UserLogPrintf(  "         dphi[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypeTheta:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Altitude Theta %d \n", g->type[i]);
          zlsio_UserLogPrintf(  "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_UserLogPrintf(  "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_UserLogPrintf(  "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_UserLogPrintf(  "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
          case kAxisTypePolarTheta:
          zlsio_UserLogPrintf(  "     Grid Dimension Type: Polar-Colatitude Theta %d \n", g->type[i]);
          zlsio_UserLogPrintf(  "     Grid Dimension Kind: %d \n", g->kind[i]);
          zlsio_UserLogPrintf(  "           n [%d]: %5d \n", i, g->n[i]);
          zlsio_UserLogPrintf(  "           ns[%d]: %5d \n",i, g->ns[i]);
          zlsio_UserLogPrintf(  "           ne[%d]: %5d \n",i, g->ne[i]);
          zlsio_UserLogPrintf(  "           nt[%d]: %5d \n",i, g->nt[i]);
          zlsio_UserLogPrintf(  "          lth[%d]: %#18.16g pi\n",i, g->xl[i]/pi);
          zlsio_UserLogPrintf(  "          rth[%d]: %#18.16g pi\n",i, g->xr[i]/pi);
          zlsio_UserLogPrintf(  "          dth[%d]: %#18.16g pi\n",i, g->dx[i]/pi);
          break;
      }; // switch
    }; // i
    
#if (ZLS_DEBUG>0)   
    switch (g->nD) {
        
      case k3D:
        if ( g->z != NULL) {
          zlsio_UserLogPrintf(  "    Z-Axis allocated\n" );
          zls_UserLogPrintAxisCoords(  g->z, g->ns[kZDim], g->ne[kZDim], g->nt[kZDim], g->type[kZDim]);
        }
        if (g->z0 != NULL) {
          zlsio_UserLogPrintf(  "    Z0-Axis allocated\n" );
        }
        case k2D:
        if ( g->y != NULL) {
          zlsio_UserLogPrintf(  "    Y-Axis allocated\n" );
          zls_UserLogPrintAxisCoords(  g->y, g->ns[kYDim],g->ne[kYDim],g->nt[kYDim], g->type[kYDim]);
        }
        if (g->y0 != NULL) {
          zlsio_UserLogPrintf(  "    Y0-Axis allocated\n" );
        }
        case k1D:
        if ( g->x != NULL) {
          zlsio_UserLogPrintf(  "    X-Axis allocated\n" );
          zls_UserLogPrintAxisCoords(  g->x, g->ns[kXDim],g->ne[kXDim],g->nt[kXDim], g->type[kXDim]);
          
        }
        if (g->x0 != NULL) {
          zlsio_UserLogPrintf(  "    X0-Axis allocated\n" );
        }
        break;
        default:
        zlsio_ErrPrintf( " Error: Dimensions %d\n\n", g->nD);
        
    }
    
#endif
    
  } else { err = memErr ; }
  
  return(err);                 
}

int zls_GetGridDataDims  ( Counter * nx, Counter * ny, Counter * nz, GCoordPtr g ){
  int err = noErr;
  //
  // get dimensions from grids, works for all dims and grids.  If a dim is 0 (ie Z in a 2D sim)
  // then a value of 1 is returned if requested so a single row of pixels may be created.
  // based on grids not zones so grids from other structures such as levels can be used too.
  //
  *nx    = g->nt[kXDim]>0? g->nt[kXDim]: 1UL; 
  *ny    = g->nt[kYDim]>0? g->nt[kYDim]: 1UL; 
  *nz    = g->nt[kZDim]>0? g->nt[kZDim]: 1UL; 
  
  return(err);
}


int zls_GetGridImageDims( Counter * nx, Counter *ny, GCoordPtr g, Counter slice ){
  int err = noErr;
  //
  // get dimensions from grids, works for all dims and grids.  If a dim is 0 (ie Z in a 2D sim)
  // then a value of 1 is returned if requested so a single row of pixels may be created.
  // based on grids not zones so grids from other structures such as levels can be used too.
  //
#ifdef FILES_NO_BCS    
  switch ((int)slice ) {
    case kSliceY:
    case kSliceYSum:
      *nx    = g->n[kYDim]>0?g->n[kYDim]: 1UL;  
      *ny    = 1UL;   
      break;
    case kSliceYZ:
    case kSliceYZSum:
      *nx    = g->n[kYDim]>0?g->n[kYDim]: 1UL;  
      *ny    = g->n[kZDim]>0?g->n[kZDim]: 1UL;   
      break;
    case kSliceYX:
    case kSliceYXSum:
      *nx    = g->n[kYDim]>0?g->n[kYDim]: 1UL;  
      *ny    = g->n[kXDim]>0?g->n[kXDim]: 1UL;   
      break;
    case kSliceZ:
    case kSliceZSum:
      *nx    = g->n[kZDim]>0?g->n[kZDim]: 1UL;   
      *ny    = 1UL; 
      break;
    case kSliceZX:
    case kSliceZXSum:
      *nx    = g->n[kZDim]>0?g->n[kZDim]: 1UL;   
      *ny    = g->n[kXDim]>0?g->n[kXDim]: 1UL; 
      break;
    case kSliceZY:
    case kSliceZYSum:
      *nx    = g->n[kZDim]>0?g->n[kZDim]: 1UL;   
      *ny    = g->n[kYDim]>0?g->n[kYDim]: 1UL; 
      break;
    case kSliceX:
    case kSliceXSum:
      *nx    = g->n[kXDim]>0?g->n[kXDim]: 1UL;  
      *ny    = 1UL;  
      break;
    case kSliceXY:
    case kSliceXYSum:
      *nx    = g->n[kXDim]>0?g->n[kXDim]: 1UL;  
      *ny    = g->n[kYDim]>0?g->n[kYDim]: 1UL;  
      break;
    case kSliceXZ:
    case kSliceXZSum:
      *nx    = g->n[kXDim]>0?g->n[kXDim]: 1UL;  
      *ny    = g->n[kZDim]>0?g->n[kZDim]: 1UL;  
      break;
    case kSliceNone:
    default:
      *nx    = g->n[kXDim]>0?g->n[kXDim]: 1UL;  
      *ny    = g->n[kYDim]>0?g->n[kYDim]: 1UL;  
  }
#endif FILES_NO_BCS    
  
#ifdef FILES_HAVE_BCS    
  switch ( slice ) {
    case kSliceY:
    case kSliceYSum:
      *nx    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL; 
      *ny    = 1UL;   
      break;
    case kSliceYZ:
    case kSliceYZSum:
      *nx    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL; 
      *ny    = g->nt[kZDim]>0?g->nt[kZDim]: 1UL;   
      break;
    case kSliceYX:
    case kSliceYXSum:
      *nx    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL; 
      *ny    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL;   
      break;
    case kSliceZ:
    case kSliceZSum:
      *nx    = g->nt[kZDim]>0?g->nt[kZDim]: 1UL;   
      *ny    = 1UL; 
      break;
    case kSliceZX:
    case kSliceZXSum:
      *nx    = g->nt[kZDim]>0?g->nt[kZDim]: 1UL;   
      *ny    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL; 
      break;
    case kSliceZY:
    case kSliceZYSum:
      *nx    = g->nt[kZDim]>0?g->nt[kZDim]: 1UL;   
      *ny    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL; 
      break;
    case kSliceX:
    case kSliceXSum:
      *nx    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL;  
      *ny    = 1UL; 
      break;
    case kSliceXY:
    case kSliceXYSum:
      *nx    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL;  
      *ny    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL; 
      break;
    case kSliceXZ:
    case kSliceXZSum:
      *nx    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL;  
      *ny    = g->nt[kZDim]>0?g->nt[kZDim]: 1UL; 
      break;
    case kSliceNone:
    default:
      *nx    = g->nt[kXDim]>0?g->nt[kXDim]: 1UL;  
      *ny    = g->nt[kYDim]>0?g->nt[kYDim]: 1UL;  
  }
#endif FILES_HAVE_BCS    
  
  return(err);
}

int zls_GetGridCubeDims  ( Counter * nx, Counter * ny, Counter * nz, GCoordPtr g ){
  int err = noErr;
  //
  // get dimensions from grids, works for all dims and grids.  If a dim is 0 (ie Z in a 2D sim)
  // then a value of 1 is returned if requested so a single row of pixels may be created.
  // based on grids not zones so grids from other structures such as levels can be used too.
  //
#ifdef FILES_NO_BCS    
  *nx    = g->n[kXDim]>0? g->n[kXDim]: 1UL; 
  *ny    = g->n[kYDim]>0? g->n[kYDim]: 1UL; 
  *nz    = g->n[kZDim]>0? g->n[kZDim]: 1UL; 
#endif FILES_NO_BCS    
  
#ifdef FILES_HAVE_BCS    
  *nx    = g->nt[kXDim]>0? g->nt[kXDim]: 1UL; 
  *ny    = g->nt[kYDim]>0? g->nt[kYDim]: 1UL; 
  *nz    = g->nt[kZDim]>0? g->nt[kZDim]: 1UL; 
#endif FILES_HAVE_BCS    
  
  return(err);
}



int zls_GetGridDirBCDims  ( Counter * ni, Counter * nj, Counter * nk, GCoordPtr g, Integer direction ){
  int err = noErr;
  //
  // works for all dims and grids.  If a dim is 0 (ie Z in a 2D sim)
  // then a value of 1 is returned if requested so a single row of cells may be created.
  // based on grids not zones so grids from other structures such as levels can be used too.
  //
  
  *ni = 1;
  *nj = 1;
  *nk = 1;
  
  switch (direction){
      
    case  kDirRight:
    case  kDirLeft:
#ifdef SYM1D
      *ni = g->nG;
#endif SYM1D
#ifdef SYM2D
      *ni = g->nG;
      *nj = g->nt[kYDim];
#endif SYM2D
#ifdef SYM3D
      *ni = g->nG;
      *nj = g->nt[kYDim];
      *nk = g->nt[kZDim];
#endif SYM3D
      break;
    case  kDirUp:
    case  kDirDown:
#ifdef SYM2D
      *ni = g->nt[kXDim];
      *nj = g->nG;
#endif SYM2D
#ifdef SYM3D
      *ni = g->nt[kXDim];
      *nj = g->nG;
      *nk = g->nt[kZDim];
#endif SYM3D
      break;
    case  kDirHither:
    case  kDirYon:
#ifdef SYM3D
      *ni = g->nt[kXDim];
      *nj = g->nt[kYDim];
      *nk = g->nG;
#endif SYM3D
      break;
    default:
      ;
  }
  
  return(err);
  
}

